/* clamdif.c -- mvmf cusp interface to clamd

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Provides the glue for a mvmf "cusp" interface to clamd

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-symtab.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
// #include "clamdif.h"
#include "clamdif-version.h"

#include "data.h"
#include "errwarn.h"


/* Local definitions */

    /* The default clamd host and port number */
#define	CLAMDHOST	"127.0.0.1"
#define	CLAMDPORT	3310


    /* Operating modes */
typedef enum {
	OM_NONE,			/* None specified */
	OM_NORMAL,			/* Normal */
	OM_VERSION,			/* Print version info */
} OPMODE;


    /* A handle for use in talking to a clamd connection */
typedef struct {
	/* Typically (always?) the input and the output BIO use the same
	   fd.  Take care e.g. not to close the same one twice.
	*/
    BIO		ch_in;			/* Input BIO handle */
    BIO		ch_out;			/* Output.. */
}  CLAMHAND;


/* External data referenced */


extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */


/* External routines used */



/* Local data publicly available */

    /* This is required to satisfy linking */
	void	*OmmP;			/* The open mail message handle */


/* Local routines and forward references */

    /* Initialize */
static	BOOL	clamdif_init PROTO( (void) );

    /* Normal operation */
static	STSCODE	clamdif_normal PROTO( (void) );

    /* Close a clamd handle */
static	BOOL	close_clamd PROTO( (CLAMHAND *chP) );

    /* Open a connection to clamd via tcp */
static	CLAMHAND *open_clamd_tcp PROTO( (int portN) );

    /* Emit usage info */
static	void	usage();



/* Private data */

static	char	*Clamd_hostP = CLAMDHOST;
static	int	Clamd_port = CLAMDPORT;
   
static	OPMODE	Opmode = OM_NONE;	/* Operating mode */

static	char	*STnameP = NULL;	/* Name of any suspicious thing found */

int
main( argc, argv )
	int		argc;
	char		**argv;
{
	char		usageOK;
	int		optch;
	STSCODE		status;

    usageOK = TRUE;

    while ( usageOK &&
	   ( optch = getopt( argc, argv, "h:p:v" ) ) != EOF ) {

	switch ( optch ) {
	    case	'h':		/* Host to connect to clamd on */
		Clamd_hostP = optarg;
		break;

	    case	'p':		/* Port to connect to clamd on */
		Clamd_port = atoi( optarg );
		break;

	    case	'v':		/* Print version info */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-v\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_VERSION;
		break;

	    default:
		usageOK = FALSE;
		break;
	}
    }

    /* If no operating mode selected, set to normal */
    if ( usageOK && ( Opmode == OM_NONE ) )
	Opmode = OM_NORMAL;

    /* No more command line arguments allowed */
    if ( usageOK && ( optind < argc ) )
	usageOK = FALSE;

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    if ( !clamdif_init() ) {
	exit( 1 );
    }

    /* Proceed now depending on the mode */
    switch ( Opmode ) {
	case	OM_NORMAL:
	    status = clamdif_normal();
	    break;

	case	OM_VERSION:
	    printf( "clamdif version %s\n", CLAMDIF_VERSION );
	    status = SCOK;
	    break;

	default:
	    warning( "Error: unknown Opmode %d.  This is a bug.", Opmode );
	    status = SCERR;
	    break;
    }

    /* Exit appropriately */
    switch ( status ) {
	case	SCOK:
	    exit( 0 );

	case	SCNOTDONE:		/* Some problem */
	    exit( 1 );

	case	SCBREAK:		/* Some timeout */
	    exit( 2 );

	default:			/* Unknown */
	    exit( 101 );
    }
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/


static void
usage()
{
    printf( "usage: clamdif [flags]\n" );
    printf( "    Flags are:\n" );
    printf( "       -h name          Host to connect to clamd on.\n" );
    printf( "       -p port          TCP port to connect to clamd on.\n" );
    printf( "       -v               Give version information.\n" );
    printf( "\n" );
}
/*

*//* clamdif_init()

	Do general initialization

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
clamdif_init NOARGLIST
{
	char		path[ PATH_MAX +1 ];

    /* Other signals that need to be different than default */
    signal( SIGPIPE, SIG_IGN );		/* Default terminates process; we
					   can check write failure just
					   fine, thank you.
					*/

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();

    /* Initialize lib stuff */
    if ( !data_init() )
	return ( FALSE );

    /* Init some general-purpose BSTR(s) */
    Gbs1P = bstr_new( "Gbs1P", 0, 0 );
#if 0
    Gbs2P = bstr_new( "Gbs2P", 0, 0 );
    Gbs3P = bstr_new( "Gbs3P", 0, 0 );
#endif

    if ( ( Gbs1P == NULL ) ||
#if 0
         ( Gbs2P == NULL ) ||
	 ( Gbs3P == NULL ) ||
#endif
	 FALSE ) {
	fprintf( stderr, "Can't create general BSTRs\n" );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* clamdif_normal()

	Run in normal mode

Accepts :



Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
clamdif_normal NOARGLIST
{
	STSCODE		status;
	int		len;
	int		i, n;
	int		sport;		/* Stream port */
	CLAMHAND	*chP;		/* Main clamd handle */
	CLAMHAND	*chsP;		/* Handle for the streamed data */
	MML_IN		*inP;		/* An input handle */
	UBYTE		buf[1024];

    status = SCNOTDONE;

    /* Connect to clamd */
    chP = open_clamd_tcp( Clamd_port );
    if ( chP == NULL )
	return ( SCNOTDONE );

    /* Make an input stack for easy receiving */
    if ( ( inP = in_new() ) == NULL ) {
	close_clamd( chP );
	return ( SCNOTDONE );
    }
    if ( !in_stream_push_bio( inP, &chP->ch_in, "clamd input" ) )
	status = SCNOTDONE;
    else {
	status = SCOK;

	/* Setup a stream to clamd */
	if ( ! ( bio_putn( &chP->ch_out, (UBYTE *)"STREAM\n", 7 ) &&
	         bio_drain( &chP->ch_out ) ) ) {
	    chsP = NULL;
	    status = SCNOTDONE;
	}
	else {
	    bio_timeout_set( &chP->ch_in, 10, 0L );
	    len = igetbtrec( inP, Gbs1P, (TXR_END_CRLF)|(TXR_END_LF), NULL );
	    if ( len < 0 ) {
		if ( chP->ch_in.bio_result == IOR_TIMEOUT ) {
		    warning( "timeout getting stream port from clamd." );
		    status = SCBREAK;
		}
		else {
		    warning( "error getting stream port from clamd." );
		    status = SCERR;
		}
	    }
	    else if ( ( len < 6 ) ||
		      ( strnicmp( (char *)Gbs1P->bs_bP, "PORT ", 5 ) != 0 ) ) {
		warning( "clamd gave unexpected response to STREAM: \"%s\"",
			  Gbs1P->bs_bP );
		status = SCNOTDONE;
	    }
	    else {
		sport = atoi( &Gbs1P->bs_bP[5] );
		if ( sport < 1024 ) {
		    warning( "clamd gave unusual port number %d for STREAM",
			    sport );
		    status = SCNOTDONE;
		}
		else {
		    /* Get another connection to stream on. */
		    chsP = open_clamd_tcp( sport );
		    if ( chsP == NULL ) {
			warning ( "can not connect to port %d for streaming.",
				   sport );
			status = SCNOTDONE;
		    }
		}
	    }
	}
    }

    /* Copy stdin to the stream channel */
    while ( status == SCOK ) {
	len = read( fileno(stdin), &buf[0], sizeof(buf) );
	if ( len <= 0 ) {
	    if ( len < 0 ) {
		warning( "Error reading stdin to give to clamd stream." );
		status = SCNOTDONE;
	    }
	    break;
	}
	if ( !bio_putn( &chsP->ch_out, &buf[0], len ) ) {
	    warning( "clamd stream output error." );
	    status = SCNOTDONE;
	    break;
	}
    }
    if ( status == SCOK ) {
	if ( !bio_drain( &chsP->ch_out ) ) {
	    warning( "clamd stream drain error." );
	    status = SCNOTDONE;
	}
    }

    /* OK to get rid of the stream channel now no matter what */
    if ( chsP != NULL )
	if ( !close_clamd( chsP ) ) {
TODO		/* Note this or something */
	    ;
	}


    if ( status == SCOK ) {
	/* Get response from clamd.  The response isn't the friendliest
	   thing in the world to parse, but what the hey.
	*/
	for ( ; ; ) {
	    len = igetbtrec( inP, Gbs1P, (TXR_END_CRLF)|(TXR_END_LF), NULL );
	    if ( len < 0 ) {
		if ( chP->ch_in.bio_result == IOR_TIMEOUT ) {
		    warning( "timeout getting stream port from clamd." );
		    status = SCBREAK;
		}
		else
		    status = SCNOTDONE;
		break;
	    }

	    if ( ( len >= 9 ) &&
		 ( strnicmp( (char *)Gbs1P->bs_bP, "stream: ", 8 ) == 0 ) ) {

		/* look for exactly "stream: OK" */
		if ( ( len == 10 ) &&
		     ( strnicmp( (char *)&Gbs1P->bs_bP[8], "OK", 2 ) == 0 ) ) {
		    /* This is a good result.  We're done. */
		    STnameP = NULL;
		    break;
		}

		/* Look for "stream: xxxx FOUND" */
		i = Gbs1P->bs_bC - 6;
		if ( strnicmp( (char *)&Gbs1P->bs_bP[i], " FOUND", 6 ) == 0 ) {
		    /* The thing's name is between "stream: " and " FOUND" */
		    n = i - 8;
		    STnameP = emalloc( "clamdif_normal", "name of thing",
				        n + 1 );
		    if ( STnameP == NULL ) {
			status = SCERR;
			break;
		    }
		    memcpy( STnameP, &Gbs1P->bs_bP[8], n );
		    STnameP[n] = NUL;
		    break;
		}
	    }
	    
	    warning( "unrecognized clamd response: \"%s\"", 
		      Gbs1P->bs_bP );
	}
    }

    close_clamd( chP );

    /* Print the name of the thing found, if any */
    if ( status == SCOK ) {
	printf( "%s\n", STnameP == NULL ? "" : STnameP );
    }

    return ( status );
}
/*

*//* close_clamd( chP )

	Close and discard an open clamd handle

Accepts :

	chP		Ptr to previously opened/allocated CLAMHAND

Returns :

	<value>		TRUE if success.
			FALSE if not.

Notes :


*/

static BOOL
close_clamd ARGLIST( (chP) )
    FARG( CLAMHAND	*chP )		/* The clamd handle */
{
	BOOL		okF;

    /* Close the output side */
    okF = ( close( chP->ch_out.bio_fd ) == 0 );

    /* Only close the input side if it's a different fd */
    if ( chP->ch_in.bio_fd != chP->ch_out.bio_fd ) {
	if ( close( chP->ch_in.bio_fd ) != 0 )
	    okF = FALSE;
    }

    /* Clean up the BIO elements */
    bio_finit( &chP->ch_in );
    bio_finit( &chP->ch_out );

    /* Get rid of the handle */
    dealloc( chP );

    /* Done */
    return ( okF );
}
/*

*//* open_clamd_tcp( portN )

	Open a TCP connection to clamd using the specified port

Accepts :

	portN		The TCP port to use

Returns :

	<value>		CLAMHAND ptr for the open connection
			NULL if error.

Notes :

	In this stage of the program's life, it's assumed we know what
	IP address on which to find clamd.  There may be embellishment
	of connecting to clamd in the future.

*/

static CLAMHAND *
open_clamd_tcp ARGLIST( (portN) )
    FARG( int		portN )		/* The port number */
{
	int		fd;		/* Fd to use */
	CLAMHAND	*chP;		/* clam handle to return */
	struct hostent	*hP;		/* for host */
	struct sockaddr_in sin;

    hP = gethostbyname( Clamd_hostP );
    if ( hP == NULL ) {
	warning( "open_clamd_tcp, can't access host \"%s\"", Clamd_hostP );
	return ( NULL );
    }

    fd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( fd < 0 ) {
	warning( "open_clamd_tcp, socket() failed, errno= %d", errno );
	return ( NULL );
    }

    sin.sin_family = hP->h_addrtype;
    memcpy( (char *)&sin.sin_addr, hP->h_addr, hP->h_length );
    sin.sin_port = htons( portN );
    if ( connect( fd, (struct sockaddr *)&sin, sizeof(sin) ) < 0 ) {
	warning( "open_clamd_tcp, connect failed, errno= %d", errno );
	close( fd );
	return ( NULL );
    }

    /* Allocate and set up the clam handle */
    chP = emalloc( "open_clamd_tcp", "clam handle", sizeof(CLAMHAND) );
    if ( chP == NULL ) {
	warning( "open_clamd_tcp, can't allocate clam handle." );
	close( fd );
	return ( NULL );
    }

    /* Initialize the two BIO channels.   Remember they both use the
       same fd.
    */
    bio_init( &chP->ch_in, fd );
    bio_init( &chP->ch_out, fd );

    /* It's OK to start logging now */
    // log_begin();

    return ( chP );
}
