/* mfl-sieve.h -- parsing and execution for sieve constructs

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Parsing / execution / evaluation for "sieve"-like language constructs

*/

#ifndef	H_MFL_SIEVE		/* For multiple inclusions */
#define	H_MFL_SIEVE


#ifdef	SIEVE			/* Only if SIEVE is wanted */

   /* Bitmasks for built-in statement classes */
#define	SBIS_ACTION	0x0001		/* Action */
#define	SBIS_CONTROL	0x0002		/* Control */
#define	SBIS_TEST	0x0004		/* Test */
    

    /* Info about a sieve capability */
typedef struct {
    char	*cap_nameP;		/* Capability name */
    BOOL	cap_enabledF;		/* TRUE if enabled */
}  SIEVE_CAP;


    /* option classes.  These must be defined from 0 thru max as
       they will also be used as array indexes.  (0 and max will
       be wasted values).  Perhaps #defines would be better.
    */
typedef enum {
	SOC_MIN=0,			/* Hack to make sure values can be
					   used as array indexes
					*/

	SOC_ADDR_PART,			/* Address-part */
	SOC_COMPARATOR,			/* Comparator */
	SOC_COPY,			/* :copy */
	SOC_INDEX,			/* :index */
	SOC_IP,				/* :ip */
	SOC_LAST,			/* :last */
	SOC_MATCH_TYPE,			/* Match-type */
	SOC_NEWNAME,			/* :newname */
	SOC_NEWVALUE,			/* :newvalue */
	SOC_OVER_UNDER,			/* Over/under (for "size") */

#ifdef	VACATION	/* classes for vacation */

	SOC_ADDRESSES,			/* :addresses */
	SOC_DAYS,			/* :days */
	SOC_FROM,			/* :from */
	SOC_HANDLE,			/* :handle */
	SOC_MIME,			/* :mime */
	SOC_SUBJECT,			/* :subject */

#endif	/* VACATION */


	SOC_MAX				/* Max bound on array of classes */

}  SIEVE_OPT_CLASS;


    /* Some option values are one of a number of choices (e.g.  off/on
       or yes/no/maybe.  This type enumerates all of the choices of
       options that have a "pick one of" sort of value.  The values
       don't matter as long as they are distinct; we'll just lump them
       all together in a big enum here.  One benefit is that different
       classes of values have different values; one drawback is that
       the different classes aren't uniquely typed.  Another approach
       would be to divide each class into different enum types, but
       that comes with its own drawbacks.
    */
typedef enum {
	SOV_NONE,			/* No chosen value (or not set) */
	SOV_SET,			/* Value has been set */

	SOV_MATCH_IS,			/* match-type "is" */
	SOV_MATCH_CONTAINS,		/* match-type "contains" */
	SOV_MATCH_MATCHES,		/* match-type "matches" */
	SOV_MATCH_REGEX,		/* match-type "regex" */
	SOV_MATCH_COUNT,		/* match-type "count" (relational) */
	SOV_MATCH_VALUE,		/* match-type "value" (relational) */

	SOV_ADDR_ALL,			/* address-part "all" */
	SOV_ADDR_DOMAIN,		/* address-part "domain" */
	SOV_ADDR_LOCAL,			/* address-part "localpart" */

	SOV_COMP_OCTET,			/* comparator "i;octet" */
	SOV_COMP_ASC_CASEMAP,		/* comparator "i;ascii-casemap" */
	SOV_COMP_ASC_NUMERIC,		/* comparator "i;ascii-numeric" */

	SOV_OVUN_OVER,			/* over/under "over" */
	SOV_OVUN_UNDER,			/* over/under "under" */

	SOV_COPY,			/* if :copy is set */

	    /* Relational values for options like :count and :value */
	SOV_REL_LT,			/* relational "lt" */
	SOV_REL_LE,			/* relational "le" */
	SOV_REL_EQ,			/* relational "eq" */
	SOV_REL_NE,			/* relational "ne" */
	SOV_REL_GE,			/* relational "ge" */
	SOV_REL_GT,			/* relational "gt" */

}  SIEVE_OPT_ONEOF;


    /* Definition of a built-in statement option */
typedef	struct {
    char		*od_nameP;	/* Option name */
    SIEVE_OPT_CLASS	od_class;	/* Option class */
    SIEVE_OPT_ONEOF	od_value;	/* Any "one-of" value specified */
    SIEVE_CAP		*od_capP;	/* Any capability required */
    STSCODE		(*od_parse)();	/* Any extended parsing routine */
}  SIEVE_OPTION;

    
    /* A parsed reference to a statement option.  This contains info
       about an option parsed; the actual option information is evaluated
       from this info at run-time.  "OPTREF" is probably not a good choice
       of name.  If I think of a better one I will probably replace it.
    */
typedef struct {
    SIEVE_OPTION	*or_optP;	/* Ptr to the option definition */
    long		or_int;		/* Any int value */
    PTNODE		*or_valP;	/* Any extended value expression */
}  SIEVE_OPTREF;


    /* Evaluated option value (data).  This will contain info about a
       specific option that was referened at parse time (via a SIEVE_OPTREF)
       once that reference is evaluated at run-time.  There can only be
       one option value per option class, so each sieve statement execution
       will use an array of these, one per class, and pay attention only
       to the classes that are relevant to each statement.
    */
typedef struct {
    SIEVE_OPT_CLASS	ov_class;	/* The option value class */
    SIEVE_OPT_ONEOF	ov_oov;		/* Any "one-of" values set.  This
					   will always be set to SOV_NONE
					   for an uninitialized value,
					   even if the value is not a
					   "one-of" type.
					*/
    union {
      ULONG	ovv_uint;	/* Any unsigned value */
      long	ovv_int;	/* Any integer value */
      UBYTE	*ovv_bP;	/* Any byte-string value */
      UBYTE	**ovv_bPP;	/* Any list of byte-string values */
    }			ov_val;		/* Additional value */
}  SIEVE_OPTVAL;



    /* Extended arguments to a sieve built-in statement */
typedef struct {
    int		ba_optC;		/* Number of options (tagged
						arguments) */
    int		ba_optL;		/* Length of options array */
    SIEVE_OPTREF *ba_optP;		/* Ptr to options */
}  SIEVE_BISARG;



/* Declarations for functions contained in this module: */

    /* Execute a sieve block */
STSCODE	exec__sieve PROTO( (PTNODE *treeP, VALUE *rvalP) );

    /* Execute a sieve statement */
STSCODE exec_sieve PROTO( (PTNODE *treeP, VALUE *rvalP) );

    /* Pick up Sieve statement arguments (tagged args) */
STSCODE	exec_sieve_get_bisargs PROTO( (PTNODE *treeP,
				       SIEVE_OPTVAL *optvalsP) );

    /* Execute a cross product of two string lists */
STSCODE	exec_sieve_stringx2 PROTO( (SIEVE_OPTVAL *optvalP,
				    PTNODE *list1P, PTNODE *list2P,
				    STSCODE (*rtc)()) );

    /* Execute a cross product of two string lists in "count" mode */
STSCODE	exec_sieve_stringx2count PROTO( (SIEVE_OPTVAL *optvalP,
					 PTNODE *list1P, PTNODE *list2P,
					 STSCODE (*rtc)()) );

    /* Execute sieve builtin statement: "addheader" */
STSCODE	exec_svbis_addheader PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "address" */
STSCODE	exec_svbis_address PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "allof" */
STSCODE	exec_svbis_allof PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "anyof" */
STSCODE	exec_svbis_anyof PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "c" */
STSCODE	exec_svbis_c PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "deleteheader" */
STSCODE	exec_svbis_deleteheader PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "discard" */
STSCODE	exec_svbis_discard PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "dnsbl" */
STSCODE	exec_svbis_dnsbl PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "else" */
STSCODE	exec_svbis_else PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "elsif" */
STSCODE	exec_svbis_elsif PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "envelope" */
STSCODE	exec_svbis_envelope PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "exists" */
STSCODE	exec_svbis_exists PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "false" */
STSCODE	exec_svbis_false PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "fileinto" */
STSCODE	exec_svbis_fileinto PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "header" */
STSCODE	exec_svbis_header PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "if" */
STSCODE	exec_svbis_if PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "keep" */
STSCODE	exec_svbis_keep PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "not" */
STSCODE	exec_svbis_not PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "redirect" */
STSCODE	exec_svbis_redirect PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "reject" */
STSCODE	exec_svbis_reject PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "replaceheader" */
STSCODE	exec_svbis_replaceheader PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "require" */
STSCODE	exec_svbis_require PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "sieve" */
STSCODE	exec_svbis_sieve PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "size" */
STSCODE	exec_svbis_size PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "stop" */
STSCODE	exec_svbis_stop PROTO( (PTNODE *nodeP, VALUE *rvalP) );

    /* Execute sieve builtin statement: "true" */
STSCODE	exec_svbis_true PROTO( (PTNODE *nodeP, VALUE *rvalP) );


    /* Parse built-in "sieve" introducer */
STSCODE	parse_bis_sieve PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse a sieve element */
STSCODE	parse_sieve PROTO( (MML_IN *inP, PTNODE **treePP, UWORD oktypes) );

    /* Parse a sieve compound element */
STSCODE	parse_sieve_compound PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse a sieve compound test group */
STSCODE	parse_sieve_compound_test PROTO( (MML_IN *inP, PTNODE **treePP) );

    /* Parse a sieve string or string list */
STSCODE	parse_sieve_string PROTO( (MML_IN *inP, PTNODE **treePP, BOOL multiF) );

    /* Parse tagged arguments for a statement */
STSCODE	parse_sieve_tagged_args PROTO( (MML_IN *inP, PTNODE *nodeP,
					SIEVE_OPTION **optsPP) );

    /* Parse a comparator value for a tagged argument */
STSCODE	parse_sieve_targ_comparator PROTO( (MML_IN *inP, SIEVE_OPTREF *optrefP) );

    /* Parse an int value for a tagged argument */
STSCODE	parse_sieve_targ_int PROTO( (MML_IN *inP, SIEVE_OPTREF *optrefP) );

    /* Parse a relational value for a tagged argument */
STSCODE	parse_sieve_targ_rel PROTO( (MML_IN *inP, SIEVE_OPTREF *optrefP) );

    /* Parse a string value for a tagged argument */
STSCODE	parse_sieve_targ_string PROTO( (MML_IN *inP, SIEVE_OPTREF *optrefP) );

    /* Parse a string list value for a tagged argument */
STSCODE	parse_sieve_targ_string_list PROTO( (MML_IN *inP,
					     SIEVE_OPTREF *optrefP) );

    /* Parse sieve builtin statement: "addheader" */
STSCODE	parse_svbis_addheader PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "address" */
STSCODE	parse_svbis_address PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "allof" */
STSCODE	parse_svbis_allof PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "anyof" */
STSCODE	parse_svbis_anyof PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "c" */
STSCODE	parse_svbis_c PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "deleteheader" */
STSCODE	parse_svbis_deleteheader PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "discard" */
STSCODE	parse_svbis_discard PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "dnsbl" */
STSCODE	parse_svbis_dnsbl PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "else" */
STSCODE	parse_svbis_else PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "elsif" */
STSCODE	parse_svbis_elsif PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "envelope" */
STSCODE	parse_svbis_envelope PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "exists" */
STSCODE	parse_svbis_exists PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "false" */
STSCODE	parse_svbis_false PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "fileinto" */
STSCODE	parse_svbis_fileinto PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "header" */
STSCODE	parse_svbis_header PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "if" */
STSCODE	parse_svbis_if PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "keep" */
STSCODE	parse_svbis_keep PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "not" */
STSCODE	parse_svbis_not PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "redirect" */
STSCODE	parse_svbis_redirect PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "reject" */
STSCODE	parse_svbis_reject PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "replaceheader" */
STSCODE	parse_svbis_replaceheader PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "require" */
STSCODE	parse_svbis_require PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "sieve" */
STSCODE	parse_svbis_sieve PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "size" */
STSCODE	parse_svbis_size PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "stop" */
STSCODE	parse_svbis_stop PROTO( (MML_IN *inP, PTNODE *nodeP) );

    /* Parse sieve builtin statement: "true" */
STSCODE	parse_svbis_true PROTO( (MML_IN *inP, PTNODE *nodeP) );


    /* Free the extended args for a sieve built-in statement */
BOOL	sieve_bis_args_free PROTO( (SIEVE_BISARG *args) );

    /* Get or make the extended args for a sieve built-in statement */
SIEVE_BISARG *sieve_bis_args_get PROTO( (PTNODE *nodeP) );

    /* See if a capability is enabled */
BOOL	sieve_cap_enabled PROTO( (SIEVE_CAP *capP) );

    /* Finalize (clear, remove) one sieve statement options value */
void	sieve_optval_finit PROTO( (SIEVE_OPTVAL *optvalP) );

    /* Initialize sieve statement options values array */
void	sieve_optvals_init PROTO( (SIEVE_OPTVAL *optvalsP,
				   SIEVE_OPTION **optsPP) );

    /* Finalize (clear, remove) sieve statement options values array */
void	sieve_optvals_finit PROTO( (SIEVE_OPTVAL *optvalsP,
				   SIEVE_OPTION **optsPP) );

    /* Print a description of a sieve PN_SVBIS parse tree node */
void	sieve_node_print_svbis PROTO( (FILE *fP, PTNODE *nodeP) );


#endif	/* SIEVE */

#endif	/* H_MFL_SIEVE */
