/* msgpart.h -- stuff for manipulating message parts

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

"msgpart" provides a semi-opaque interface to manipulate parts of
an email message.  Also included are some string comparison
routines for use in testing header components.

*/


#ifndef	H_MSGPART			/* For multiple inclusions */
#define	H_MSGPART


    /* Character bitmasks for mail character table.  Some of these duplicate
       generic character info (e.g. iscntrl()) but since we have a table,
       why not optimize some lookups.
    */
#define	MC_CTL		0x0001		/* CTL (RFC822 and friends) */
#define	MC_SPC		0x0002		/* SPACE */
#define	MC_ALPHA	0x0004		/* US ASCII alpha */
#define	MC_DIGIT	0x0008		/* US ASCII digit */
#define	MC_ATOM		0x0010		/* atom char (RFC2822 "atext") */
#define	MC_TSPL		0x0020		/* "tspecials" (RFC2045) */
#define	MC_NASC		0x0040		/* Non-ASCII (for convenience) */
#define	MC_NWSCTL	0x0080		/* NO-WS-CTL (RFC2822) */

    /* Bits for selecting address part(s) */
#define	AP_LOCAL	0x0001	/* Local part (e.g. username */
#define	AP_DOMAIN	0x0002	/* Domain part */
#define	AP_COMMENTS	0x0004	/* Comments inclosed in parens */
#define	AP_DISPLAY	0x0008	/* Display name */


    /* Bits for fetching bytes from headers (e.g. msgpart_hdrchar()) */
#define	HC_PAREN	0x0001	/* Skip parenthesized comment strings */
#define	HC_QUOTED	0x0002	/* Skip quoted strings */

    /* Info that we keep about each header line */
typedef struct mphl {
    REFSTR	*mh_hP;		/* Ptr to raw header buffer */
}  MPHL;


    /* MIME info -- generally stuff found in headers that we extract
       and put here for easy access.
    */
typedef struct {
    char	*mi_versionP;	/* MIME version */
    char	*mi_typeP;	/* Media type */
    char	*mi_subtypeP;	/* Media subtype */
    char	*mi_boundaryP;	/* Boundary string if multipart */
    int		mi_boundaryL;	/* Length of boundary string */
}  MIMEINFO;


    /* A message container;
       Has info about how to access the message.
       In this initial version, the container is designed just
       to access the incoming message.  However we'd like to plan
       for other message containers at some point (e.g. if we want
       to manipulate folders in addition to the incoming message).
       Having a container object *may* help in that effort but
       will undoubtedly need work when and if we come to that.
    */
typedef struct msgcon {
    FILE	*mc_ifP;	/* File pointer where the message is */
    MML_IN	*mc_inP;	/* Alternatively, stream where the message is */
    FILE	*mc_tmpfP;	/* File ptr to any temp copy of the message */
    FILE	*mc_usefP;	/* File ptr to use to reference the body
				   (this is a copy of either mc_ifP or
				    mc_tmpfP, depending on whether a temp
				    file is being used)
				*/
    char	*mc_tmpfnameP;	/* Name of the temporary copy */
    BOOL	mc_uncachedF;	/* If container has been prepped for uncached
				   content (see msgcon_uncache())
				*/
    int		mc_prereadC;	/* If >= 0, data in the general read buffer
				   has data that's already been read, and
				   this is the bytecount to be used.  (Used
				   by msgcon_read() )
				*/
}  MSGCON;



    /* A mail message object.  This contains a reference to a top
       message part, a reference to a message container, and various
       parameters about the message.
    */
typedef struct mmsg {
    MSGCON	*mm_conP;	/* Ptr to message container */
    struct msgpart *mm_topmpP;	/* The top messagepart */
    long	mm_cachemax;	/* How much body to cache, approximately */
    long	mm_cacheC;	/* How much body is cached */
    BOOL	mm_oqfromF;	/* Whether to do "From " quoting on output */
    UBYTE	mm_cnlC;	/* Consecutive newline counter.  Keeps
				   track of consecutive newlines, up
				   to some small number; used to determine
				   whether blank lines are needed at
				   various point(s).
				*/
}  MMSG;



    /* A part of a message body that's a byte string.  The byte
       string is entirely or partially cached in memory, with any
       overflow being cached via the message container.
    */
typedef struct {
    REFSTR	*ps_partP;	/* Cached part (all or part) */
    BSTR	*ps_partbsP;	/* Underlying BSTR for ps_partP */
    off_t	ps_nc_start;	/* File offset to the start of any
				   non-cached part.
				*/
    off_t	ps_nc_end;	/* File offset to the end of any
				   non-cached part.
				*/
} MSGPSTR;


    /* A message part-- header + body.
       This may be a recursive structure, as the "body" may itself
       be one or more message parts.
    */
typedef struct msgpart {
    struct msgpart *mp_parentP;	/* Link upwards to the parent msgpart */
    MMSG	*mp_mmsgP;	/* Ptr to the message object it's in */
    LL_NODE	*mp_slnodeP;	/* Ptr to the node in the parent's child
				   list (from this perspective, the sibling
				   list node) */
    BOOL	mp_multiF;	/* TRUE if multipart, FALSE if not */
    BOOL	mp_deleteF;	/* TRUE to delete this part */
    ULONG	mp_hdrsize;	/* Size of all headers, approximately */
    LL_HEAD	mp_hdrs;	/* Headers in memory */
    MIMEINFO	mp_mime;	/* Extracted MIME information */

	/* Parts of the body that are byte strings */
    MSGPSTR	mp_body;	/* Body or multipart prolog */
    MSGPSTR	mp_epilog;	/* Any epilog to multipart */

	/* Body stuff if multipart: */
    LL_HEAD	mp_parts;	/* Linked list of msgparts */

}  MSGPART;




/* Declarations for functions contained in this module: */

/* Note- some of these things pass around the MSGPART pointer seemingly
   needlessly, but this is just in case we ever want to reference it.
*/

    /* Skip over the header name in an ESTR */
int	hdr_name_skip_estr PROTO( (ESTR *hstrP, int bX) );

    /* Skip over the header name in a string */
UBYTE	*hdr_name_skip_str PROTO( (UBYTE *hstrP) );

    /* Remove a mmesage object */
BOOL	mmsg_free PROTO( (MMSG *msgP) );

    /* Set up a new mmsg object */
MMSG	*mmsg_new PROTO( (MSGCON *conP) );

    /* Setup a mmsg from an already-open file */
MMSG	*mmsg_openfp PROTO( (FILE *fP) );

    /* Setup a mmsg from an input stream */
MMSG	*mmsg_openip PROTO( (MML_IN *inP) );

    /* Free a message container object */
BOOL	msgcon_free PROTO( (MSGCON *conP) );

    /* Setup a new message container */
MSGCON	*msgcon_new PROTO( (void) );

    /* Read a source line from message container */
int	msgcon_readline PROTO( (MSGCON *conP, UWORD rctl) );

    /* Append a source line from message container */
int	msgcon_readlinea PROTO( (MSGCON *conP, UWORD rctl) );

    /* Prepare a message container for uncached content */
BOOL	msgcon_uncache PROTO( (MSGCON *conP) );

    /* Make data available as uncached data (add to uncached data) */
BOOL	msgcon_uncache_data PROTO( (MSGCON *conP,
				    UBYTE *byteP, int byteC,
				    BOOL newlineF) );

    /* Add a child messagepart to a messagepart */
BOOL	msgpart_add PROTO( (MSGPART *mpP, MSGPART *childP, BOOL endF) );

    /* Return a date/time string formatted for use in message parts */
char	*msgpart_date_string PROTO( (time_t *timeP) );

    /* Same as msgpart_date_string, but for the current date/time */
char	*msgpart_date_string_now PROTO( (void) );

    /* Remove a previously created messagepart (and its children) */
BOOL	msgpart_free PROTO( (MSGPART *mpP) );

    /* Add a header string */
BOOL	msgpart_hdr_add PROTO( (MSGPART *mpP, ESTR *hstrP,
				LL_NODE *beforeP) );

    /* Add a header given name and value */
BOOL	msgpart_hdr_add2 PROTO( (MSGPART *mpP, ESTR *hnameP, ESTR *hvalP,
				 LL_NODE *beforeP) );

    /* Get next address part(s) from header string */
char	*msgpart_hdr_addr PROTO( (MSGPART *mpP, LL_NODE *hdrP, UWORD parts) );

    /* Change a header */
BOOL	msgpart_hdr_change PROTO( (MSGPART *mpP, LL_NODE *hdrP,
				ESTR *newnameP, ESTR *newvalP ) );

    /* Delete a header, return previous */
LL_NODE *msgpart_hdr_del PROTO( (MSGPART *mpP, LL_NODE *hdrP) );

    /* Get [next] header */
LL_NODE	*msgpart_hdr_next PROTO( (MSGPART *mpP, LL_NODE *hdrP, ESTR *hnameP) );

    /* Get references to header parts */
BOOL	msgpart_hdr_parts PROTO( (MSGPART *mpP, LL_NODE *hdrP,
				 REFSTR **namePP, REFSTR **tailPP) );

    /* Get [prev] header */
LL_NODE	*msgpart_hdr_prev PROTO( (MSGPART *mpP, LL_NODE *hdrP, ESTR *hnameP) );

    /* Get [next/prev] header */
LL_NODE	*msgpart_hdr_step PROTO( (MSGPART *mpP, LL_NODE *hdrP, ESTR *hnameP,
				  BOOL forwardF) );

    /* Get raw string from header */
REFSTR	*msgpart_hdr_str PROTO( (MSGPART *mpP, LL_NODE *hdrP) );

    /* Get raw string from header tail */
ESTR	*msgpart_hdr_tail PROTO( (MSGPART *mpP, LL_NODE *hdrP) );

    /* Get a byte from a header string as dictated by HC_XXXX flagbits */
int	msgpart_hdrchar PROTO( (REFSTR *rsP, UWORD flags) );

    /* Get an option=value pair from a header string */
BOOL	msgpart_hdroptpair PROTO( (REFSTR *rsP,
				   char *optP, int optL,
				   char *valP, int valL,
				   BOOL semireqF) );

    /* Get a token from a header string */
int	msgpart_hdrtoken PROTO( (REFSTR *rsP, char *tokP, int tokL) );

    /* Glean information from messagepart headers */
BOOL	msgpart_learn_headers PROTO( (MSGPART *mpP) );

    /* Allocate a new msgpart struct */
MSGPART	*msgpart_new PROTO( (MMSG *msgP, MSGPART *parentP) );

    /* Read messagepart headers */
BOOL	msgpart_read_headers PROTO( (MSGPART *mpP) );

    /* Read an entire messagepart */
BOOL	msgpart_read_msgpart PROTO( (MSGPART *mpP) );

    /* Read multipart body */
BOOL	msgpart_read_multipart PROTO( (MSGPART *mpP) );

    /* Read implied multipart body (e.g. message/rfc822) */
BOOL	msgpart_read_multipart_implied PROTO( (MSGPART *mpP) );

    /* Get next address part(s) from a refstr */
char	*msgpart_refstr_addr PROTO( (MSGPART *mpP, REFSTR *rsP, UWORD parts) );

    /* Get the size of a messagepart's body (maybe including children) */
ULONG	msgpart_size_body PROTO( (MSGPART *mpP, BOOL childrenF) );

    /* Get the size of a messagepart's headers */
ULONG	msgpart_size_hdr PROTO( (MSGPART *mpP) );

    /* Write body to a file stream */
BOOL	msgpart_write_body PROTO( (MSGPART *mpP, FILE *ofP, BOOL forceF ) );

    /* Write header(s) to a file stream */
BOOL	msgpart_write_hdrs PROTO( (MSGPART *mpP, FILE *ofP, BOOL embedF) );

    /* Write reference headers to a file stream */
BOOL	msgpart_write_references PROTO( (MSGPART *mpP, FILE *ofP) );


#endif	/* H_MSGPART */
