/* msgstore.h -- definitions for the msgstore module

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Functions dealing with mail message storage

*/


#ifndef	H_MSGSTORE			/* For multiple inclusions */
#define	H_MSGSTORE

    /* Parameters for dotlocking... */
#define	DOTLOCK_TRYEVERY	1	/* Sleep this long between tries */
#define	DOTLOCK_MAXTRIES	30	/* Try this many times */
#define	DOTLOCK_MAXAGE		(2*3600)  /* Delete troublesome dotlock
					     older than this many seconds
					  */


    /* This has to be done here so we can use it in function pointer
        prototype in structure member; dumb language has dumb
	idiosyncracies.
    */
typedef struct msmsg MSMSG;


    /* General (underlying) msgstore type.  msgstore operations go
       through the generic type handler, with reference to more detail
       provided by a specific type handler.

       e.g. a generic type might be "file" with specific type handlers
       for "maildir" and "mbox".

       The general type also provides a small number of interface
       functions that are common to all types. 
    */
typedef struct {
    char	*gt_nameP;		/* Name of type */

	/* Handlers */
    BOOL	(*gt_attr_int)();	/* Set an integer attribute */
    BOOL	(*gt_attr_str)();	/* Set a string attribute */

    BOOL	(*gt_folder_preacc) PROTO( (MSMSG *msmsgP) );
					/* Any preprocessing for folder
					   access.  Must set up specific
					   type, and instantiate general
					   and specific data; also does
					   anything else necessary for
					   the general type.
					*/
}  MSGT;

    /* Specific msgstore type.  Each specific type is a set of handlers
       providing an interface to mailstore access for that type.
       Every specific type is expected to make use of common functions
       provided by its associated general type; however, that aspect
       is private to the specific type.  Each general type *may*
       specify an extended interface (the "specific type interface")
       between general handlers and a specific type; that interface
       may be accessed via the st_gtifP member.
    */
typedef struct msst {
    char	*st_nameP;		/* Name of the type. */

    MSGT	*st_msgtP;		/* Ptr to general type */

    void	*st_stifP;		/* Specific type interface used
					   by general type functions.
					*/

    void	*st_gtdataP;		/* Ptr to [template] general data */
    void	*st_stdataP;		/* Ptr to specific data */
    

	/* Handlers */
    void	*(*st_gtdata_get) PROTO( (void *srcP) );
					/* Get general data for type */

    void	(*st_gtdata_release) PROTO( (void *dataP) );
					/* Release general data for type */

    void	*(*st_stdata_get) PROTO( (void *srcP) );
					/* Get specific data for type */

    void	(*st_stdata_release) PROTO( (void *dataP) );
					/* Release specific data for type */

    BOOL	(*st_attr_int_set) PROTO( (void *gtdataP, void *stdataP,
					   char *nameP, long val) );
					/* Set an integer attribute */

    BOOL	(*st_attr_str_set) PROTO( (void *gtdataP, void *stdataP,
					   char *nameP, char *valP) );
					/* Set a string attribute */

    BOOL	(*st_msg_write) PROTO( (OMM *ommP, MSMSG *msmsgP) );
					/* Write a message */

}  MSST;


    /* Msgstore namespace.  A namespace is used by a person to specify
       a unique combination of specific type and type parameters.  It
       is located primarily by its mailbox prefix string.
    */
typedef struct msns {
    UBYTE	*ns_prefixP;		/* Prefix for this namespace */
    int		ns_prefixlen;		/* # of bytes in the prefix */
    ATTR_STR	ns_transhook;		/* MFL hook: tail name translation */
    MSST	*ns_msstP;		/* Ptr to specific storage type */
    BOOL	ns_inboxF;		/* INBOX associated with this? */
    BOOL	ns_mustsstF;		/* Must be specific msst? */
    void	*ns_gtdataP;		/* Ptr to general type data */
    void	*ns_stdataP;		/* Ptr to specific type data */
}  MSNS;


    /* Control for a folder being accessed via msgstore */
typedef struct {
    MSNS	*mf_msnsP;		/* Namespace for folder */
    UBYTE	*mf_fnameP;		/* The raw folder name.  This is
					   NOT allocated; it is a pointer
					   to an incoming name for reference
					   only.
					*/
    UBYTE	*mf_ftailP;		/* Folder tail: the part of the
					   folder name that follows the
					   namespace prefix.  This string
					   is NOT allocated; it points into
					   mf_fnameP, for convenience.
					*/
}  MSFOLDER;


    /* Control for a message being accessed via msgstore.  Note the
       typedef has been done above.
    */
/* typedef */	struct msmsg {
    MSFOLDER	*mm_folderP;		/* Ptr to folder info */
    MSST	*mm_msstP;		/* Ptr to specific type in use
					   (may be different than the
					   default namespace msst!)
					*/
    void	*mm_gtdataP;		/* Ptr to general type data */
    void	*mm_stdataP;		/* Ptr to specific type data */
} /*  MSMSG */ ;



/* Declarations for functions contained in this module: */

    /* Derive an MSFOLDER from a folder name */
MSFOLDER *msfolder_fname PROTO( (UBYTE *fnameP) );

    /* Release an MSFOLDER previously gotten */
void	msfolder_release PROTO( (MSFOLDER *msfP) );

    /* See if message is autoreplyable */
BOOL	msg_autoreplyable PROTO( (OMM *ommP) );

    /* Bounce mail to the sender, with reason */
BOOL	msg_bounce PROTO( (OMM *ommP, char *reasonP) );

    /* Forward mail to another recipient */
BOOL	msg_forward PROTO( (OMM *ommP, char *rnameP, BOOL keepfromF) );

    /* Write current message to a pipe */
BOOL	msg_pipe PROTO( (OMM *ommP, char *cmdP) );

    /* Begin sending a message */
FILE	*msg_send_begin PROTO( (OMM *ommP, char *fromP, char *toP) );

    /* Finish sending a message */
BOOL	msg_send_end PROTO( (OMM *ommP, FILE *msgfP) );

    /* Write current message to a folder */
BOOL	msg_write PROTO( (OMM *ommP, char *fnameP) );

    /* Initialize msgstore stuff */
BOOL	msgstore_init PROTO( (void) );

    /* Get an MSMSG from a raw folder name */
MSMSG	*msmsg_fname PROTO( (UBYTE *fnameP) );

    /* Get translated folder tail from a folder */
UBYTE	*msmsg_folder_tail PROTO( (MSMSG *msmsgP) );

    /* Release an MSMSG previously gotten */
void	msmsg_release PROTO( (MSMSG *msmsgP) );

    /* Look up a namespace by specific prefix */
MSNS	*msns_find PROTO( (UBYTE *prefixP) );

    /* Look up a namespace, by path (prefix superstring) */
MSNS	*msns_find_path PROTO( (UBYTE *pathP) );

    /* Register a namespace */
MSNS	*msns_register PROTO( (UBYTE *prefixP) );

    /* Look up a specific msgstore type */
MSST	*msst_find PROTO( (UBYTE *nameP) );

    /* Register a specific msgstore type */
BOOL	msst_register PROTO( (MSST *newmsstP) );

#endif	/* H_MSGSTORE */
