/* mfl-bif-control.c -- MFL functions: misc

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to mvmf controls.

Many $control_xx() functions have a parallel $admin_xx() for admin mode.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "controls.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */


/* External routines used */


/* Local routines and forward references */

static	STSCODE	bif_int_set PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, CNTRLTYPE cat) );

static	STSCODE	bif_int_setlim PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, CNTRLTYPE cat) );

static	STSCODE	bif_string_set PROTO( (char *fnameP, VALUE *rvalP,
				   PTNODE *treeP, CNTRLTYPE cat) );

    /* Built-in functions contained herein */
static	STSCODE	bif_admin_int_set();
static	STSCODE	bif_admin_int_setlim();
static	STSCODE	bif_admin_string_set();
static	STSCODE	bif_control_int_get();
static	STSCODE	bif_control_int_set();
static	STSCODE	bif_control_int_setlim();
static	STSCODE	bif_control_string_get();
static	STSCODE	bif_control_string_set();


/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_control[] = {
/*  function name               handler function		*/
{  "$admin_int_set",		bif_admin_int_set	},
{  "$admin_int_setlim",		bif_admin_int_setlim	},
{  "$admin_string_set",		bif_admin_string_set	},
{  "$control_int_get",		bif_control_int_get	},
{  "$control_int_set",		bif_control_int_set	},
{  "$control_int_setlim",	bif_control_int_setlim	},
{  "$control_string_get",	bif_control_string_get	},
{  "$control_string_set",	bif_control_string_set	},

{  NULL,		NULL			}
		};


/* Private data */


/*

*//* bif_control_init(args)

	Initialize the control built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_control_init NOARGLIST
{
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_admin_int_set -- Set an integer control on behalf of admin
     bif_control_int_set -- Set an integer control on behalf of user

    BOOL  $admin_int_set( name, value )
    BOOL  $control_int_set( name, value )

*/

static STSCODE
bif_admin_int_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Only allowed if admin mode is enabled */
    if ( ! cntrl_int_get( estr_tmp_str( "admin" ), 0 ) ) {
	val_set_int( rvalP, 4, FALSE );
	return ( SCOK );
    }
    else
	return ( bif_int_set( fnameP, rvalP, treeP, CA_ADMIN ) );
}


static STSCODE
bif_control_int_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( bif_int_set( fnameP, rvalP, treeP, CA_USER ) );
}



/*

*//* bif_control_int_get -- Get an integer control value

    int  $control_int_get( name )

*/

static STSCODE
bif_control_int_get ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*cnameP;	/* Control name */
	CNTRL		*cntrlP;
	CNTRLINT	*ciP;
	VALUE		nameval;	/* VALUE for evaluated name */

    /* Evaluate and access the control name */
    cnameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( cnameP == NULL )
	status = SCNOTDONE;
    else {
	if ( ( cntrlP = cntrl_find( cnameP ) ) == NULL ) 
	    status = SCNOTDONE;
	else if ( cntrlP->ctl_type != CTL_INT ) {
	    warning( "%s: control \"%s\" is the wrong type for this function.",
			    fnameP, cnameP->es_strP );
	    status = SCNOTDONE;
	}
	else {
	    ciP = (CNTRLINT *)cntrlP->ctl_infoP;
	    val_set_int( rvalP, 4, ciP->cv_int );
	    status = SCOK;
	}
	val_release( &nameval );
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}



/*

*//* bif_admin_int_setlim -- Set integer control limits on behalf of admin
     bif_control_int_setlim -- Set integer control limits on behalf of user

    BOOL  $admin_int_setlim( name, low, high )
    BOOL  $control_int_setlim( name, low, high )

*/

static STSCODE
bif_admin_int_setlim ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Only allowed if admin mode is enabled */
    if ( ! cntrl_int_get( estr_tmp_str( "admin" ), 0 ) ) {
	val_set_int( rvalP, 4, FALSE );
	return ( SCOK );
    }
    else
	return ( bif_int_setlim( fnameP, rvalP, treeP, CA_ADMIN ) );
}


static STSCODE
bif_control_int_setlim ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( bif_int_setlim( fnameP, rvalP, treeP, CA_USER ) );
}



/*

*//* bif_admin_string_set -- Set a string control on behalf of admin
     bif_control_string_set -- Set a string control on behalf of user

    BOOL  $admin_string_set( name, value )
    BOOL  $control_string_set( name, value )

*/

static STSCODE
bif_admin_string_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    /* Only allowed if admin mode is enabled */
    if ( ! cntrl_int_get( estr_tmp_str( "admin" ), 0 ) ) {
	val_set_int( rvalP, 4, FALSE );
	return ( SCOK );
    }
    else
	return ( bif_string_set( fnameP, rvalP, treeP, CA_ADMIN ) );
}


static STSCODE
bif_control_string_set ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
    return ( bif_string_set( fnameP, rvalP, treeP, CA_USER ) );
}


/*

*//* bif_control_string_get -- Get a string control value

    string  $control_string_get( name )

*/

static STSCODE
bif_control_string_get ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*cnameP;	/* Control name */
	CNTRL		*cntrlP;
	CNTRLSTR	*csP;
	VALUE		nameval;	/* VALUE for evaluated name */

    /* Evaluate and access the control name */
    cnameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( cnameP == NULL )
	status = SCNOTDONE;
    else {
	if ( ( cntrlP = cntrl_find( cnameP ) ) == NULL ) 
	    status = SCNOTDONE;
	else if ( cntrlP->ctl_type != CTL_STR ) {
	    warning( "%s: control \"%s\" is the wrong type for this function.",
			    fnameP, cnameP->es_strP );
	    status = SCNOTDONE;
	}
	else {
	    csP = (CNTRLSTR *)cntrlP->ctl_infoP;
	    val_set_string( rvalP, estr_tmp_str( csP->cs_strP ) );
	    status = SCOK;
	}
	val_release( &nameval );
    }

    if ( status != SCOK )
	val_set_string( rvalP, NULL );

    return ( status );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* bif_int_set

	Common code for
		bif_admin_int_set
		bif_control_int_set

*/

static STSCODE
bif_int_set ARGLIST( (fnameP, rvalP, treeP, cat) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( CNTRLTYPE	cat )		/* Access type */
{
	BOOL		result;
	STSCODE		status;
	long		val;
	ESTR		*cnameP;	/* Control name */
	CNTRL		*cntrlP;
	CNTRLINT	*ciP;
	VALUE		nameval;	/* VALUE for evaluated name */

    /* Get the value first */
    status = func_treearg_long( fnameP, treeP, 1, &val );
    if ( status != SCOK )
	result = FALSE;
    else {
	/* Evaluate and access the control name */
	cnameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
	if ( cnameP == NULL )
	    result = FALSE;
	else {
	    if ( ( cntrlP = cntrl_find( cnameP ) ) == NULL )
		result = FALSE;
	    else if ( cntrlP->ctl_type != CTL_INT ) {
		warning( "%s: control \"%s\" is the wrong type for this function.",
			    fnameP, cnameP->es_strP );
		result = FALSE;
	    }
	    else {
		ciP = (CNTRLINT *)cntrlP->ctl_infoP;
		result = cntrl_intc_set( ciP, cat, val );
	    }
	    val_release( &nameval );
	}
    }

    val_set_int( rvalP, 4, result );

    return ( status );
}




/*

*//* bif_int_setlim

	Common code for
		bif_admin_int_setlim
		bif_control_int_setlim

*/

static STSCODE
bif_int_setlim ARGLIST( (fnameP, rvalP, treeP, cat) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( CNTRLTYPE	cat )		/* Access type */
{
	BOOL		result;
	STSCODE		status;
	long		low, high;
	ESTR		*cnameP;	/* Control name */
	CNTRL		*cntrlP;
	CNTRLINT	*ciP;
	VALUE		nameval;	/* VALUE for evaluated name */

    /* Get the values first */
    status = func_treearg_long( fnameP, treeP, 1, &low );
    if ( status == SCOK )
	status = func_treearg_long( fnameP, treeP, 2, &high );
    if ( status != SCOK )
	result = FALSE;
    else {
	/* Evaluate and access the control name */
	cnameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
	if ( cnameP == NULL )
	    result = FALSE;
	else {
	    if ( ( cntrlP = cntrl_find( cnameP ) ) == NULL )
		result = FALSE;
	    else if ( cntrlP->ctl_type != CTL_INT ) {
		warning( "%s: control \"%s\" is the wrong type for this function.",
			    fnameP, cnameP->es_strP );
		result = FALSE;
	    }
	    else {
		ciP = (CNTRLINT *)cntrlP->ctl_infoP;
		result = cntrl_intc_setlim( ciP, cat, low, high );
	    }
	    val_release( &nameval );
	}
    }

    val_set_int( rvalP, 4, result );

    return ( status );
}



/*

*//* bif_string_set

	Common code for
		bif_admin_string_set
		bif_control_string_set

*/

static STSCODE
bif_string_set ARGLIST( (fnameP, rvalP, treeP, cat) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
   NFARG( PTNODE	*treeP )	/* Argument tree */
    FARG( CNTRLTYPE	cat )		/* Access type */
{
	BOOL		result;
	ESTR		*cnameP;	/* Control name */
	ESTR		*cvalP;		/* Control value */
	CNTRL		*cntrlP;
	CNTRLSTR	*csP;
	VALUE		nameval;	/* VALUE for evaluated name */
	VALUE		valval;		/* VALUE for new value */

    /* Evaluate and access the control name */
    cnameP = func_treearg_estr( fnameP, treeP, 0, &nameval );
    if ( cnameP == NULL )
	result = FALSE;
    else {
	if ( ( cntrlP = cntrl_find( cnameP ) ) == NULL ) 
	    result = FALSE;
	else if ( cntrlP->ctl_type != CTL_STR ) {
	    warning( "%s: control \"%s\" is the wrong type for this function.",
			    fnameP, cnameP->es_strP );
	    result = FALSE;
	}
	else {
	    csP = (CNTRLSTR *)cntrlP->ctl_infoP;

	    /* Get value */
	    cvalP = func_treearg_estr( fnameP, treeP, 1, &valval );
	    if ( cvalP == NULL )
		result = FALSE;
	    else {
		result = cntrl_strc_set( csP, cat, cvalP );
		val_release( &valval );
	    }
	}
	val_release( &nameval );
    }

    val_set_int( rvalP, 4, result );

    return ( SCOK );
}
