/* mfl-bif-cusp.c -- MFL functions: cusp stuff

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to CUSPs .  

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "controls.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "mvmfmisc.h"
#include "errwarn.h"

/* Local definitions */

    /* Information about a running CUSP */
typedef struct {
    pid_t	rc_pid;			/* Process id */

#ifdef	RCUSP_BIO_OUT
	/* Would prefer this, but msgpart functions need a FILE handle.
	   may convert those some day.
	*/
    BIO		rc_out;			/* Output side */
#else
    FILE	*rc_outP;
#endif	/* RCUSP_BIO_OUT */

    BIO		rc_in;			/* Input side */
    BIO		rc_errin;		/* Error input side */
    MML_IN	*rc_inP;		/* stream for reading from cusp */
    CUSP	*rc_cuspP;		/* Cusp definition */
}  RCUSP;

/* External data referenced */

extern	OMM	*OmmP;			/* Open mail message */
extern	SYMTAB	*STcusp;		/* Symbol table for CUSP names */

/* External routines used */


/* Local routines and forward references */

    /* Close running cusp */
static	int	cusp_close PROTO( (RCUSP *rcuspP) );

    /* Lookup cusp by name */
static	CUSP	*cusp_lookup PROTO( (ESTR *cnameP) );

    /* Open (run) a cusp */
static	RCUSP	*cusp_open PROTO( (CUSP *cuspP, RCUSP *rcuspP) );

    /* Read a line of text from cusp stdout */
static	STSCODE	cusp_read_line PROTO( (RCUSP *rcuspP, BSTR *bsP) );

    /* Register a cusp by name */
static	CUSP	*cusp_register PROTO( (ESTR *cnameP) );

    /* Write end to cusp (drain, close output) */
static	STSCODE	cusp_write_end PROTO( (RCUSP *rcuspP) );

    /* Get ptr to $CUSP$ structure when MFL caller passes a ($CUSP$ *) */
static	UBYTE	*get_cusparg PROTO( (char *fnameP, PTNODE *treeP, int argN) );

    /* Open a CUSP when MFL caller passes its name. */
static	STSCODE	open_cusparg PROTO( (char *fnameP, PTNODE *treeP,
				      int argN, RCUSP *rcuspP) );

    /* Open a CUSP when MFL caller passes its name, return DVS. */
static	DVS	*open_cusparg_dvs PROTO( (char *fnameP, PTNODE *treeP, int argN) );

    /* Given a DVS pointing to internal cusp info, set $CUSP$ value */
static	void	val_set_cusp_dvs PROTO( (VALUE *rvalP, DVS *dvsP) );

    /* Built-in functions contained herein */
static	STSCODE	bif_cusp_close();
static	STSCODE	bif_cusp_define();
static	STSCODE	bif_cusp_open();
static	STSCODE	bif_cusp_read_line();
static	STSCODE	bif_cusp_write_end();
static	STSCODE	bif_cusp_write_message();
static	STSCODE	bif_cusp_write_string();
static	STSCODE	bif_cuspu_message();

/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_cusp[] = {
/*  function name               handler function		*/
{  "$cusp_close",		bif_cusp_close		},
{  "$cusp_define",		bif_cusp_define		},
{  "$cusp_open",		bif_cusp_open		},
{  "$cusp_read_line",		bif_cusp_read_line	},
{  "$cusp_write_end",		bif_cusp_write_end	},
{  "$cusp_write_message",	bif_cusp_write_message	},
{  "$cusp_write_string",	bif_cusp_write_string	},
{  "$cuspu_message",		bif_cuspu_message	},

{  NULL,		NULL			}
		};


/* Private data */

static	TYPE	*Ty_cuspP;	/*   ($CUSP$) */
static	TYPE	*Ty_cusppP;	/*   ($CUSP$ *) */

static	int	Ty_cusp_x_c_stat;    /* offset of c_stat in $CUSP$ */
static	int	Ty_cusp_x_c_priv1;   /* offset of c_priv1 in $CUSP$ */
static	int	Ty_cusp_x_rcusp;     /* where we put RCUSP (c_priv1 plus
					 any appropriate alignment padding)
				     */

/*

*//* bif_cusp_init(args)

	Initialize the cusp built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_cusp_init NOARGLIST
{
	int		i;
	TNSYM		*tnsymP;
	TYPSYM		*tpsymP;
	TYPE		*typeP;
	AGGMEM		*agmemP;
	char		mfl[1000];	/* Buffer for mfl code */

    /* Initialize the $CUSP$ type that we use to pass info between scripts
       and cuspy functions.  This struct has user-available members and
       private data that we represent as a character array.  The private
       data really has to be in some sense visible to scripts in case of
       copying to some other MFL variable.

       The c_priv1 element is where we store our internal RCUSP (plus
         offset padding for alignment, thus the slop bytes reserved).

    */
    i = snprintf( &mfl[0], sizeof(mfl), "\n\
typedef struct $cusp$ {\n\
    int c_stat;\n\
    unsigned char c_priv1[%d];\n\
} $CUSP$;\n\
",
	sizeof(RCUSP) + sizeof(char *) );
    if ( i >= sizeof(mfl) ) {
	warning( "bif_cusp_init: buffer overflow making $CUSP$" );
	return ( FALSE );
    }
    mfl_run_str( NULL, estr_tmp_str( &mfl[0] ), estr_tmp_str( "Mc" ) );

    tnsymP = type_aggname( TC_STRUCT, "$cusp$" );
    if ( ( tnsymP == NULL ) ||
         ( ( typeP = tnsymP->tn_typeP ) == NULL ) ||
	 ( ( typeP->ty_flags & TF_DEFINED ) == 0 ) ) {
	warning( "bif_cusp_init: unable to find \"$cusp$\" struct." );
	return ( FALSE );
    }
    Ty_cuspP = typeP;			/* ($CUSP$) type */
    tpsymP = type_ptr( Ty_cuspP );	/* ($CUSP$ *) type */
    if ( tpsymP == NULL ) {
	warning( "bif_cusp_init: can't find/make '$CUSP$ *' type." );
	return ( FALSE );
    }
    Ty_cusppP = &(tpsymP->ts_type);

    /* Remember offset(s) for future reference */
    if ( ( agmemP = type_agg_member( Ty_cuspP, "c_stat" ) ) == NULL ) {
	warning( "bif_cusp_init: can't find $CUSP$ member 'c_stat'" );
	return ( FALSE );
    }
    Ty_cusp_x_c_stat = agmemP->am_offs;

    if ( ( agmemP = type_agg_member( Ty_cuspP, "c_priv1" ) ) == NULL ) {
	warning( "bif_cusp_init: can't find $CUSP$ member 'c_priv1'" );
	return ( FALSE );
    }
    Ty_cusp_x_c_priv1 = agmemP->am_offs;

    /* We put an RCUSP in priv1, but we want to make sure it's aligned
       on a pointer boundary just in case.
    */
    Ty_cusp_x_rcusp = ( Ty_cusp_x_c_priv1 + sizeof(char *) -1 ) / sizeof(char *);
    Ty_cusp_x_rcusp *= sizeof(char *);

    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_cusp_close -- Close a cusp

    BOOL  $cusp_close( $CUSP$ *cuspP )

    Closes and waits for cusp to terminate

*/

static STSCODE
bif_cusp_close ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		result;
	UBYTE		*ucuspP;	/* Ptr to user $cusp$ struct */
	RCUSP		*rcuspP;	/* Ptr to RCUSP within ucusp */

    if ( ( ucuspP = get_cusparg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	status = SCOK;
	rcuspP = (RCUSP *)(ucuspP + Ty_cusp_x_rcusp);

	result = cusp_close( rcuspP );
	status = SCOK;
	val_set_int( rvalP, 4, result );
    }

    if ( status != SCOK )
	val_set_int( rvalP, 4, -1 );

    return ( status );
}
/*

*//* bif_cusp_define -- define a cusp

    int $cusp_define( string name, string path )

    Defines a new cusp

    returns TRUE if OK, false if not

*/

static STSCODE
bif_cusp_define ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	ESTR		*cnameP;	/* cusp name */
	ESTR		*cpathP;	/* CUSP path */
	CUSP		*cuspP;
	VALUE		cnameval, cpathval;

    cnameP = cpathP = NULL;

    /* Need pipe permission or admin mode to do this. */
    if ( ! ( cntrl_int_get( estr_tmp_str( "pipe_allow" ), 0 ) ||
	     cntrl_int_get( estr_tmp_str( "admin" ), 0 ) ) ) {
TODO	/* Might we want a different control?  But pipe seems fine. */
	warning( "%s: not authorized.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Get cusp name and path */
	cnameP = func_treearg_estr( fnameP, treeP, 0, &cnameval );
	if ( cnameP == NULL ) {
	    warning( "%s requires a CUSP name.", fnameP );
	    status = SCNOTDONE;
	}
	else if ( ( cpathP = func_treearg_estr( fnameP, treeP, 1, &cpathval ) )
			== NULL ) {
	    warning( "%s requires a path name.", fnameP );
	    status = SCNOTDONE;
	}
	else if ( cusp_lookup( cnameP ) != NULL ) {
	    warning( "%s: CUSP is already defined, \"%s\"...",
		      fnameP, tmpstr_estr( cnameP ) );
	    status = SCNOTDONE;
	}
	else if ( ( cuspP = cusp_register( cnameP ) ) != NULL ) {
	    status = SCOK;

	    /* Fill in the path */
	    cuspP->cp_pathP = newstr_n( fnameP, "cusp path",
				        (char *)cpathP->es_strP,
					estr_len( cpathP ) );
	    if ( cuspP->cp_pathP == NULL )
		status = SCERR;
	}
    }

    /* Release values as required */
    if ( cpathP != NULL )
	val_release( &cpathval );
    if ( cnameP != NULL )
	val_release( &cnameval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    
    return ( status );
}
/*

*//* bif_cusp_open -- open a cusp

    $CUSP$  *$cusp_open( string name )

    Returns ptr to a $CUSP$ struct for the cusp

*/

static STSCODE
bif_cusp_open ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	DVS		*dvsP;		/* DVS for cusp */

    dvsP = open_cusparg_dvs( fnameP, treeP, 0 );
    if ( dvsP == NULL ) {
	val_set_int( rvalP, 4, 0 );
    }
    else {
	/* Return the ptr to the $CUSP$ within the DVS */
	val_set_cusp_dvs( rvalP, dvsP );
    }

    return ( SCOK );
}
/*

*//* bif_cusp_read_line -- read a line of output from cusp

    string  *$cusp_read_line( $CUSP$ *cuspP )

    Reads line from cusp, returns string pointer, or NULL if no input.

*/

static STSCODE
bif_cusp_read_line ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		len;
	UBYTE		*ucuspP;	/* Ptr to user $cusp$ struct */
	RCUSP		*rcuspP;	/* Ptr to RCUSP within ucusp */

    if ( ( ucuspP = get_cusparg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	rcuspP = (RCUSP *)(ucuspP + Ty_cusp_x_rcusp);
	status = cusp_read_line( rcuspP, Gbs1P );
	if ( status == SCOK ) {
	    val_set_string_ptr( rvalP,
			        estr_tmp_nstr( Gbs1P->bs_bP, Gbs1P->bs_bC ) );
	}
    }

    if ( status != SCOK )
	val_set_string_ptr( rvalP, NULL );

    return ( SCOK );
}
/*

*//* bif_cusp_write_end -- End the writing to a cusp

    BOOL  $cusp_write_end( $CUSP$ *cuspP )

    Flushes the output to a CUSP and closes just the output side.
    Is a no-op if cusp is already write-closed.

*/

static STSCODE
bif_cusp_write_end ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	UBYTE		*ucuspP;	/* Ptr to user $cusp$ struct */
	RCUSP		*rcuspP;	/* Ptr to RCUSP within ucusp */

    if ( ( ucuspP = get_cusparg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	status = SCOK;
	rcuspP = (RCUSP *)(ucuspP + Ty_cusp_x_rcusp);

	status = cusp_write_end( rcuspP );
	if ( status == SCNOTDONE )
	    /* Special marker to return non-success to MFL, but
	       success to the interpreter (to let MFL continue).
	    */
	    status = SCDONE;
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    if ( status == SCDONE )
	status = SCOK;

    return ( SCOK );
}
/*

*//* bif_cusp_write_message -- Write currently open message to cusp

    BOOL  $cusp_write_message( $CUSP$ *cuspP )

    Sends the current message to the cusp

*/

static STSCODE
bif_cusp_write_message ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	BOOL		okF;
	UBYTE		*ucuspP;	/* Ptr to user $cusp$ struct */
	RCUSP		*rcuspP;	/* Ptr to RCUSP within ucusp */

    if ( OmmP == NULL ) {
	warning( "%s: no mail message is open.", fnameP );
	status = SCNOTDONE;
    }
    else if ( ( ucuspP = get_cusparg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else {
	status = SCOK;
	rcuspP = (RCUSP *)(ucuspP + Ty_cusp_x_rcusp);

#ifdef	RCUSP_BIO_OUT
	warning( "can't output message via BIO." );
	status = SCNOTDONE;
#else
	if ( rcuspP->rc_outP == NULL ) {
	    warning( "%s: cusp output is closed.", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    okF = msgpart_write_hdrs( OmmP->om_mmsgP->mm_topmpP,
				       rcuspP->rc_outP, FALSE );
	    okF = okF && msgpart_write_body( OmmP->om_mmsgP->mm_topmpP,
					     rcuspP->rc_outP, FALSE );

	    if ( okF )
		status = SCOK;
	    else
		status = SCNOTDONE;
	}
#endif	/* RCUSP_BIO_OUT */

    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( SCOK );
}
/*

*//* bif_cusp_write_string -- Write a string to cusp

    BOOL  $cusp_write_string( $CUSP$ *cuspP, string s )

    Send a string to the cusp

*/

static STSCODE
bif_cusp_write_string ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		len;
	ESTR		*strP;		/* For the string.. */
	UBYTE		*ucuspP;	/* Ptr to user $cusp$ struct */
	RCUSP		*rcuspP;	/* Ptr to RCUSP within ucusp */
	VALUE		strval;

    if ( ( ucuspP = get_cusparg( fnameP, treeP, 0 ) ) == NULL ) {
	status = SCNOTDONE;
	strP = NULL;
    }
    else {
	/* Get the string to write */
	strP = func_treearg_estr( fnameP, treeP, 1, &strval );
	if ( strP == NULL )
	    status = SCNOTDONE;
	else {
	    status = SCOK;
	    rcuspP = (RCUSP *)(ucuspP + Ty_cusp_x_rcusp);

	    len = estr_len( strP );

#ifdef	RCUSP_BIO_OUT
	    if ( !bio_putn( &rcuspP->rc_out, (UBYTE *)strP->es_strP, len ) ) {
		status = SCNOTDONE;
	    }
#else
	    if ( rcuspP->rc_outP == NULL ) {
		warning( "%s: cusp output is closed.", fnameP );
		status = SCNOTDONE;
	    }
	    else {
		if ( fwrite( strP->es_strP, 1, len, rcuspP->rc_outP ) != len ) {
		    status = SCNOTDONE;
		}
	    }
#endif	/* RCUSP_BIO_OUT */
	}
    }

    if ( strP != NULL )
	val_release( &strval );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    /* output errors still allow MFL to proceed */
    if ( status == SCNOTDONE )
	status = SCOK;

    return ( SCOK );
}
/*

*//* bif_cuspu_message -- unified write message, return result

    string  *$cuspu_message( string name )

    returns a string with combination of exit status and first stdout line,
    like this:

        "0 some-output"

    The "unified" functions don't give many options for return info.  But
    some special returns here are:

        "" -- failure to execute
	"999 xxx" -- weird status code had to be replaced by 999

*/

static STSCODE
bif_cuspu_message ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		result;
	int		len;
	BOOL		okF;
	RCUSP		rcusp;
	RCUSP		*rcuspP;
	char		stsbuf[10];

    if ( OmmP == NULL ) {
	warning( "%s: no mail message is open.", fnameP );
	rcuspP = NULL;
    }
    else if ( ( status = open_cusparg( fnameP, treeP, 0, &rcusp ) ) != SCOK )
	rcuspP = NULL;
    else {
	rcuspP = &rcusp;
    
#ifdef	RCUSP_BIO_OUT
	warning( "can't output message via BIO." );
	status = SCNOTDONE;
#else
	if ( rcuspP->rc_outP == NULL ) {
	    warning( "%s: cusp output is closed.", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    okF = msgpart_write_hdrs( OmmP->om_mmsgP->mm_topmpP,
				       rcuspP->rc_outP, FALSE );
	    okF = okF && msgpart_write_body( OmmP->om_mmsgP->mm_topmpP,
					     rcuspP->rc_outP, FALSE );

	    cusp_write_end( rcuspP );

	    if ( okF )
		status = SCOK;
	    else
		status = SCNOTDONE;
	}
#endif	/* RCUSP_BIO_OUT */

    }

    if ( status == SCOK ) {
	status = cusp_read_line( rcuspP, Gbs1P );
	if ( status != SCOK ) {
	    Gbs1P->bs_bC = 0;
	    bstr_nulterm( Gbs1P );
	}
    }
    if ( rcuspP != NULL ) {
	/* Get and set exit status */
	result = cusp_close( rcuspP );

	if ( snprintf( &stsbuf[0], sizeof(stsbuf), "%d ", result ) >= sizeof(stsbuf) )
	    strcpy( &stsbuf[0], "999 " );
	len = strlen( &stsbuf[0] );

	/* Must slide existing text down by len bytes */
	if ( Gbs1P->bs_bL < ( Gbs1P->bs_bC + len + 1 ) ) {
	    if ( !bstr_grow( Gbs1P, len + 1 ) )
		status = SCERR;
	}
	if ( status != SCERR ) {
	    memmove( &Gbs1P->bs_bP[len], &Gbs1P->bs_bP[0], Gbs1P->bs_bC );
	    memcpy( &Gbs1P->bs_bP[0], &stsbuf[0], len );
	    Gbs1P->bs_bC += len;
	    bstr_nulterm( Gbs1P );
	}
    }

    if ( rcuspP == NULL ) {
	val_set_string( rvalP, estr_tmp_str( "" ) );
    }
    else if ( status == SCERR ) {
	val_set_string( rvalP, estr_tmp_str( "999 " ) );
    }
    else
	val_set_string( rvalP, estr_tmp_nstr( Gbs1P->bs_bP, Gbs1P->bs_bC ) );

    return ( status == SCERR ? status : SCOK );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* cusp_close( rcuspP )

	Closes a running cusp

Accepts :

	rcuspP		Ptr to the running cusp info

Returns :

	<value>		< 0 if some error
			status code from cusp otherwise

Notes :


*/

static int
cusp_close ARGLIST( (rcuspP) )
    FARG( RCUSP		*rcuspP )	/* Ptr to running cusp info */
{
	int		waitstat, exitcode;
	pid_t		epid;		/* Pid of child that exited */

    /* Wait for the child to exit */
    epid = waitpid( rcuspP->rc_pid, &waitstat, 0 );
    if ( epid != rcuspP->rc_pid ) {
	warning( "cusp_close, expected child %d got %d", rcuspP->rc_pid, epid );
	return ( -1 );
    }
    exitcode = WEXITSTATUS( waitstat );

    /* Close up the cusp */
    in_destroy( rcuspP->rc_inP );
    rcuspP->rc_inP = NULL;

#ifdef	RCUSP_BIO_OUT
	if ( rcuspP->rc_out.bio_fd >= 0 )
	    close( rcuspP->rc_out.bio_fd );
	bio_finit( &rcuspP->rc_out.bio_fd );
#else
	if ( rcuspP->rc_outP != NULL ) {
	    fclose( rcuspP->rc_outP );
	    rcuspP->rc_outP = NULL;
	}
#endif	/* RCUSP_BIO_OUT */

    if ( rcuspP->rc_in.bio_fd >= 0 )
	close( rcuspP->rc_in.bio_fd );
    bio_finit( &rcuspP->rc_in );

    if ( rcuspP->rc_errin.bio_fd >= 0 )
	close( rcuspP->rc_errin.bio_fd );
    bio_finit( &rcuspP->rc_errin );

    return ( exitcode );
}
/*

*//* cusp_lookup( cnameP )

	Look up a cusp by name

Accepts :

	cnameP		Name of cusp

Returns :

	<value>		Ptr to CUSP if found
			NULL if not

Notes :

	Just does a symbol table lookup, but is provided as a wrapper
	in case any other functionality is needed.

*/

static CUSP *
cusp_lookup ARGLIST( (cnameP) )
    FARG( ESTR		*cnameP )	/* CUSP name */
{
	int		cnlen;		/* Cusp name length */
	char		*cnstrP;	/* Ptr to cusp name */
	char		cname[PATH_MAX +1];

    /* If passed a NUL-terminated string, use it directly.  Otherwise
       we'll make one in a local buffer.  This is just a hack because
       we don't have an ESTR interface to the symtab, which would
       be better.
    */
    if ( cnameP->es_len == -1 )
	cnstrP = (char *)cnameP->es_strP;
    else {
	cnlen = estr_len( cnameP );
	if ( cnlen >= sizeof(cname) )
	    return ( NULL );		/* Too long, no match. */
	memcpy( &cname[0], cnameP->es_strP, cnlen );
	cname[ cnlen ] = NUL;
	cnstrP = &cname[0];
    }

    /* Now look it up */
    return ( (CUSP *)symtab_lookup( STcusp, cnstrP ) );
}
/*

*//* cusp_open( cuspP, rcuspP )

	Open (run) a CUSP

Accepts :

	cuspP		Ptr to CUSP info
	rcuspP		If non-NULL, an RCUSP struct to use

Returns :

	<value>		Ptr to RCUSP
			FALSE if not.

Notes :

	If rcuspP is passed as NULL, a new one will be allocated.

*/

static RCUSP *
cusp_open ARGLIST( (cuspP, rcuspP) )
   NFARG( CUSP		*cuspP )	/* Ptr to CUSP info */
    FARG( RCUSP		*rcuspP )	/* Any RCUSP to use */
{
	BOOL		okF;
	int		coutfd;		/* fd for output to child's stdin */
	int		cinfd;		/* fd for input from child's stdout */
	int		cerrfd;		/* fd for input from child's stderr */
	RCUSP		*rrcuspP;	/* The RCUSP to return */
	pid_t		pid;
	char		*argv[2];	/* Arguments for execv */

    /* Fire up a connection to the cusp */
    argv[0] = cuspP->cp_pathP;
    argv[1] = NULL;
    pid = pexec_std( cuspP->cp_pathP, &argv[0],
		      &coutfd, &cinfd, &cerrfd );
    if ( pid < 0 )
	return ( NULL );

    okF = TRUE;

    /* Need an RCUSP struct if caller hasn't supplied one */
    if ( (rrcuspP = rcuspP ) == NULL ) {
	rrcuspP = emalloc( "cusp_open", "rcusp", sizeof(RCUSP) );
    }

    if ( rrcuspP == NULL )
	okF = FALSE;
    else {				/* Fill in the RCUSP info */
	rrcuspP->rc_pid = pid;
	rrcuspP->rc_cuspP = cuspP;

	bio_init( &rrcuspP->rc_in, cinfd );
	bio_init( &rrcuspP->rc_errin, cerrfd );

#ifdef	RCUSP_BIO_OUT
	bio_init( &rrcuspP->rc_out, coutfd );
#else
	rrcuspP->rc_outP = fdopen( coutfd, "w" );
	if ( rrcuspP->rc_outP == NULL ) {
	    warning( "cusp_open can't fdopen output." );
	    okF = FALSE;
	}
#endif	/* RCUSP_BIO_OUT */

	if ( okF ) {
	    /* Set up input stream to use */
	    rrcuspP->rc_inP = in_new();
	    if ( rrcuspP->rc_inP == NULL )
		okF = FALSE;
	}
    }

    /* Time to be done.  If not OK, undo some things. */
    if ( !okF ) {
	close( coutfd );
	close( cinfd );
	close( cerrfd );

	if ( rrcuspP != NULL ) {
	    if ( rrcuspP->rc_inP != NULL ) {
		in_destroy( rrcuspP->rc_inP );
	    }

	    if ( rcuspP == NULL ) {
		/* We allocated this, so free it. */
		dealloc( rrcuspP );
	    }

	    rrcuspP = NULL;
	}
    }

    return ( rrcuspP );
}
/*

*//* cusp_read_line( rcuspP, bsP )

	Read a line of text from cusp's stdout

Accepts :

	rcuspP		Ptr to RCUSP handle for the CUSP
	bsP		Where to put the text

Returns :

	<value>		status code

Notes :


*/

static STSCODE
cusp_read_line ARGLIST( (rcuspP, bsP) )
   NFARG( RCUSP		*rcuspP )	/* CUSP handle */
    FARG( BSTR		*bsP )		/* Where to put the data */
{
	STSCODE		status;
	int		len;

    /* Push the stdout-from-cusp stream onto cusp's input stack;
       remember BIO streams do not self-pop, so we can push/pop here.
    */
    if ( !in_stream_push_bio( rcuspP->rc_inP, &rcuspP->rc_in,
				  "cusp stdin" ) )
	status = SCNOTDONE;
    else {
TODO	/* Hack timeout in; it should be settable. */
	bio_timeout_set( &rcuspP->rc_in, 5, 0L );
	len = igetbtrec( rcuspP->rc_inP, Gbs1P,
			     (TXR_END_CRLF)|(TXR_END_LF), NULL );
	if ( len < 0 ) {
	    /* Timeout or error */
	    status = SCNOTDONE;

TODO	    /* Should return the status in cusp result area */
	}
	else
	    status = SCOK;

	/* Remove stream from cusp's input stack */
	in_stream_pop( rcuspP->rc_inP );
    }

    return ( status );
}
/*

*//* cusp_register( cnameP )

	Register a new cusp 

Accepts :

	cnameP		Name of cusp

Returns :

	<value>		Ptr to CUSP if successful
			NULL if not

Notes :

	This function does NOT check to see if the name is already
	registered; it's assumed that callers want more discrete
	control over that check.

*/

static CUSP *
cusp_register ARGLIST( (cnameP) )
    FARG( ESTR		*cnameP )	/* CUSP name */
{
	int		cnlen;		/* Cusp name length */
	char		*cnstrP;	/* Ptr to cusp name */
	CUSP		*cuspP;
	char		cname[PATH_MAX +1];

    /* Need a NUL-terminated name; same as in cusp_lookup. */
    if ( cnameP->es_len == -1 )
	cnstrP = (char *)cnameP->es_strP;
    else {
	cnlen = estr_len( cnameP );
	if ( cnlen >= sizeof(cname) ) {
	    warning( "cusp_register: name too long, \"%s\"...",
		      tmpstr_estr( cnameP ) );
	    return ( NULL );
	}
	memcpy( &cname[0], cnameP->es_strP, cnlen );
	cname[ cnlen ] = NUL;
	cnstrP = &cname[0];
    }

    /* Now register it */
    cuspP = (CUSP *)symtab_enter( STcusp, cnstrP, sizeof(CUSP) );

    if ( cuspP != NULL ) {
	/* Initialize the cusp struct */
	cuspP->cp_pathP = NULL;
    }

    return ( cuspP );
}
/*

*//* cusp_write_end( rcuspP )

	Finish output to a CUSP

Accepts :

	rcuspP		Ptr to RCUSP handle for the CUSP

Returns :

	<value>		status code, including:
			SCOK if ok
			SCNOTDONE if output problem or closed

Notes :

	Drains the output and closes it.  Just like MFL $cusp_write_end .

	It is not considered an error to use this when output is
	already closed.

*/

static STSCODE
cusp_write_end ARGLIST( (rcuspP) )
    FARG( RCUSP		*rcuspP )	/* CUSP handle */
{
	STSCODE		status;

    status = SCOK;			/* Assume success */

#ifdef	RCUSP_BIO_OUT
    if ( rcuspP->rc_out.bio_fd >= 0 ) {
	/* Still open, go ahead and finish it. */
	if ( !bio_drain( &rcuspP->rc_out ) )
	    status = SCNOTDONE;
	if ( close( rcuspP->rc_out.bio_fd ) != 0 )
	    status = SCNOTDONE;
	rcuspP->rc_out.bio_fd = -1;
    }
#else
    if ( rcuspP->rc_outP != NULL ) {
	/* Still open, finish it */
	if ( fflush( rcuspP->rc_outP ) < 0 )
	    status = SCNOTDONE;
	if ( fclose( rcuspP->rc_outP ) != 0 )
	    status = SCNOTDONE;
	rcuspP->rc_outP = NULL;
    }
#endif	/* RCUSP_BIO_OUT */

    return ( status );
}
/*

*//* get_cusparg( fnameP, treeP, argN )

	Get the cusp pointer passed to a built-in-function

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the cusp is passed in

Returns :

	<value>		Ptr to cusp object if found
			NULL if not

Notes :

	Error/warning messages are given here
*/

static UBYTE *
get_cusparg ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	UBYTE		*dataP;
	PTNODE		*cuspargP;	/* Subtree for header arg */
	VALUE		cuspval;	/* To get the value.. */

    if ( ( cuspargP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: arg #%d (header pointer) not supplied.",
			fnameP, argN );
	return ( NULL );
    }

    /* Evaluate */
    if ( eval_tree( cuspargP, &cuspval ) != SCOK )
	return ( NULL );
    if ( cuspval.v_typeP != Ty_cusppP ) {
	warning( "%s: arg #%d not $CUSP$ pointer.", fnameP, argN );
	return ( NULL );
    }

    val_rvalue( &cuspval );
    dataP = (UBYTE *)valptr_resolve( &cuspval.v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "%s: arg #%d bad $CUSP$ pointer.", fnameP, argN );
	return ( NULL );
    }

    val_release( &cuspval );

    return ( dataP );
}
/*

*//* open_cusparg( fnameP, treeP, argN, rcuspP )

	Open a CUSP

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the cusp name is passed in
	rcuspP		The RCUSP struct to initialize

Returns :

	<value>		status code

Notes :

	Error/warning messages are given here
*/

static STSCODE
open_cusparg ARGLIST( (fnameP, treeP, argN, rcuspP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
   NFARG( int		argN )		/* Arg number */
    FARG( RCUSP		*rcuspP )	/* RCUSP thing to use */
{
	STSCODE		status;
	ESTR		*cnameP;	/* Cusp name */
	DVS		*dvsP;
	CUSP		*cuspP;		/* Underlying CUSP */
	VALUE		cnameval;	/* To get the cusp name.. */

    /* Get the cusp name */
    cnameP = func_treearg_estr( fnameP, treeP, argN, &cnameval );
    if ( cnameP == NULL ) {
	warning( "%s requires a CUSP name.", fnameP );
	return ( SCNOTDONE );
    }

    cuspP = cusp_lookup( cnameP );
    if ( cuspP == NULL ) {
	warning ( "%s: no CUSP \"%s\"...", fnameP, tmpstr_estr( cnameP ) );
	status = SCNOTDONE;
    }
    else {
	rcuspP = cusp_open( cuspP, rcuspP );
	if ( rcuspP == NULL ) {
	    warning( "%s: can't open CUSP \"%s\"...", fnameP,
			tmpstr_estr( cnameP ) );
	    status = SCNOTDONE;
	}
	else
	    status = SCOK;
    }

    val_release( &cnameval );

    return ( status );
}
/*

*//* open_cusparg_dvs( fnameP, treeP, argN )

	Open a CUSP and attach it to a DVS

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the cusp name is passed in

Returns :

	<value>		Ptr to DVS if OK
			NULL if not

Notes :

	The DVS data pointer points to an MFL $CUSP$ struct, which
	contains (at x_rcusp offset) our RCUSP data.

	Error/warning messages are given here
*/

static DVS *
open_cusparg_dvs ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	ESTR		*cnameP;	/* Cusp name */
	DVS		*dvsP;
	RCUSP		*rcuspP;	/* Running cusp info */
	CUSP		*cuspP;		/* Underlying CUSP */
	VALUE		cnameval;	/* To get the cusp name.. */

    /* Get the cusp name */
    cnameP = func_treearg_estr( fnameP, treeP, argN, &cnameval );
    if ( cnameP == NULL ) {
	warning( "%s requires a CUSP name.", fnameP );
	return ( NULL );
    }

    rcuspP = NULL;
    dvsP = NULL;

    cuspP = cusp_lookup( cnameP );
    if ( cuspP == NULL ) {
	warning ( "%s: no CUSP \"%s\"...", fnameP, tmpstr_estr( cnameP ) );
    }
    else if ( ( dvsP = dvs_new( Ty_cuspP->ty_size, Ty_cuspP->ty_size) )
		== NULL ) {
	warning( "%s: can't get DVS.", fnameP );
    }
    else {
	rcuspP = cusp_open( cuspP,
			(RCUSP *)(dvsP->dv_dataP + Ty_cusp_x_rcusp) );
	if ( rcuspP == NULL ) {
	    warning( "%s: can't open CUSP \"%s\"...", fnameP,
		      tmpstr_estr( cnameP ) );
	    dvs_release( dvsP );
	    dvsP = NULL;
	}
    }

    val_release( &cnameval );

    return ( dvsP );
}
/*

*//* val_set_cusp_dvs(rvalP, dvsP)

	Set a value to a $CUSP$ type given the DVS pointer

Accepts :

	rvalP		Where to built the result value
	dvsP		Ptr to DVS that already points to the cusp info

Returns :

	<nothing>

Notes :

	Constructs a VALUE in rvalP this value will contain a reference to
	the mfl-visible $CUSP$ struct within an internal-purpose $CUSPi$
	that is pointed to by the given DVS

*/

static void
val_set_cusp_dvs ARGLIST( (rvalP, dvsP) )
   NFARG( VALUE		*rvalP )	/* Value to return/build */
    FARG( DVS		*dvsP )		/* Ptr to DVS */
{
	VALPTR		*vptrP;

    /* The VALUE will reference the $CUSP$ within the DVS $CUSPi$ */
    rvalP->v_flags = VF_RVALUE;
    rvalP->v_typeP = Ty_cusppP;
    rvalP->v_data.d_ptr.p_byteX = 0;
    rvalP->v_data.d_ptr.p_ptr.p_genP = dvsP;
}
