/* mfl-bif-db.c -- MFL functions: db stuff

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to "db" file accesses.  "db" files are those
that implement some kind of indexed or keyed data access via local files,
but do not include more complex things like SQL or LDAP.  Support is
included for db types:

    cdb

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "cdb.h"

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */


/* External routines used */


/* Local routines and forward references */

static	void	*get_cdbarg PROTO( (char *fnameP, PTNODE *treeP, int argN) );
static	int	open_fnamearg_cdb PROTO( (char *fnameP, PTNODE *treeP, int argN) );

    /* Built-in functions contained herein */
static	STSCODE	bif_cdb_close();
static	STSCODE	bif_cdb_get_string();
static	STSCODE	bif_cdb_open();
static	STSCODE	bif_cdbu_get_string();

/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_db[] = {
/*  function name               handler function		*/
{  "$cdb_close",		bif_cdb_close		},
{  "$cdb_get_string",		bif_cdb_get_string	},
{  "$cdb_open",			bif_cdb_open		},
{  "$cdbu_get_string",		bif_cdbu_get_string	},

{  NULL,		NULL			}
		};


/* Private data */

static	TYPE	*Ty_cdbP;	/*   ($CDB$) */
static	TYPE	*Ty_cdbpP;	/*   ($CDB$ *) */
static	TYPE	*Ty_cdbiP;	/*   ($CDBi$) */
static	TYPE	*Ty_cdbipP;	/*   ($CDBi$ *) */

static	int	Ty_cdb_x_c_stat;   /* offset of c_stat in $CDB$ */
static	int	Ty_cdbi_x_ci_cdb;  /* offset of ci_cdb in $CDBi$ */

/*

*//* bif_db_init(args)

	Initialize the db built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_db_init NOARGLIST
{
	int		i;
	TNSYM		*tnsymP;
	TYPSYM		*tpsymP;
	TYPE		*typeP;
	AGGMEM		*agmemP;
	char		mfl[1000];	/* Buffer for mfl code */

    /* Initialize the $CDB$ type that's available to scripts. */
    strcpy( &mfl[0], "\n\
typedef struct $cdb$ {\n\
    int c_stat;\n\
} $CDB$;\n\
" );
    mfl_run_str( NULL, estr_tmp_str( &mfl[0] ), estr_tmp_str( "Mc" ) );
    tnsymP = type_aggname( TC_STRUCT, "$cdb$" );
    if ( ( tnsymP == NULL ) ||
         ( ( typeP = tnsymP->tn_typeP ) == NULL ) ||
	 ( ( typeP->ty_flags & TF_DEFINED ) == 0 ) ) {
	warning( "bif_db_init: unable to find \"$cdb$\" struct." );
	return ( FALSE );
    }
    Ty_cdbP = typeP;		/* ($CDB$) type */
    tpsymP = type_ptr( Ty_cdbP );  /* ($CDB$ *) type */
    if ( tpsymP == NULL ) {
	warning( "bif_db_init: can't find/make '$CDB$ *' type." );
	return ( FALSE );
    }
    Ty_cdbpP = &(tpsymP->ts_type);

    /* Remember offset(s) for future reference */
    if ( ( agmemP = type_agg_member( Ty_cdbP, "c_stat" ) ) == NULL ) {
	warning( "bif_db_init: can't find $CDB$ member 'c_stat'" );
	return ( FALSE );
    }
    Ty_cdb_x_c_stat = agmemP->am_offs;



    /* The $CDBi$ mfl type is used internally to encapsulate both
       internal cdb information and the script-visible  $CDB$
      structure.
    */
    i = snprintf( &mfl[0], sizeof(mfl), "\n\
typedef struct $cdbi$ {\n\
    unsigned char ci_cdbstruct[%d];\n\
    $CDB$ ci_cdb;\n\
} $CDBi$;\n\
",
        sizeof(struct cdb) );
    if ( i >= sizeof(mfl) ) {
	warning( "bif_db_init: buffer overflow making $CDBi$" );
	return ( FALSE );
    }
    mfl_run_str( NULL, estr_tmp_str( &mfl[0] ), estr_tmp_str( "Mc" ) );

    tnsymP = type_aggname( TC_STRUCT, "$cdbi$" );
    if ( ( tnsymP == NULL ) ||
         ( ( typeP = tnsymP->tn_typeP ) == NULL ) ||
	 ( ( typeP->ty_flags & TF_DEFINED ) == 0 ) ) {
	warning( "bif_db_init: unable to find \"$cdbi$\" struct. " );
	return ( FALSE );
    }
    Ty_cdbiP = typeP;			/* ($CDBi$) type */
    tpsymP = type_ptr( Ty_cdbiP );	/* ($CDBi$ *) type */
    if ( tpsymP == NULL ) {
	warning( "bif_db_init: can't find/make '$CDBi$ *' type." );
	return ( FALSE );
    }
    Ty_cdbipP = &(tpsymP->ts_type);

    /* Remember offset(s) for future reference */
    if ( ( agmemP = type_agg_member( Ty_cdbiP, "ci_cdb" ) ) == NULL ) {
	warning( "bif_db_init: can't find $CDBi$ member 'ci_cdb'" );
	return ( FALSE );
    }
    Ty_cdbi_x_ci_cdb = agmemP->am_offs;
    
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_cdb_close -- close an open CDB

    BOOL   $cdb_close( cdbP )

    Closes a previously opened CDB

    cdbP is ($CDB$ *), the result of a prior $cdb_open()

*/

static STSCODE
bif_cdb_close ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	struct cdb	*cdbP;

    status = SCOK;

    if ( ( cdbP = get_cdbarg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else if ( cdbP->fd >= 0 ) {
	cdb_free( cdbP );
	close( cdbP->fd );
	memset( (UBYTE *)cdbP, 0, sizeof(struct cdb) );
	cdbP->fd = -1;
    }

    val_set_int( rvalP, 4, status == SCOK ? 1 : 0 );

    return ( SCOK );
}


/*

*//* bif_cdb_get_string -- return a string from a CDB lookup

    string  *$cdb_get_string( cdbP, keyP )

    Looks up a key, returns a string pointer.
      returns NULL if key not found.

    cdbP is ($CDB$ *), the result of a prior $cdb_open()

    keyP is a (string)

*/

static STSCODE
bif_cdb_get_string ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		r;
	unsigned int	dlen;
	off_t		dpos;
	UBYTE		*strP;
	struct cdb	*cdbP;
	ESTR		*keyP;
	VALUE		keyval;

    status = SCOK;
    keyP = NULL;

    if ( ( cdbP = get_cdbarg( fnameP, treeP, 0 ) ) == NULL )
	status = SCNOTDONE;
    else if ( ( keyP = func_treearg_estr( fnameP, treeP, 1, &keyval ) )
			== NULL ) {
	warning( "%s requires a key string second argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Got key, do lookup */
	r = cdb_find( cdbP, keyP->es_strP, estr_len( keyP ) );
	if ( r <= 0 )
	    status = SCNOTDONE;
	else {
	    /* Found key; read & make string */
	    dlen = cdb_datalen( cdbP );
	    dpos = cdb_datapos( cdbP );
	    strP = tmpstr( dlen );
	    if ( strP == NULL )
		status = SCNOTDONE;
	    else {
		r = cdb_read( cdbP, strP, dlen, dpos );
		if ( r != 0 ) {
		    warning( "%s: key found, but error reading data.",
				fnameP );
		    status = SCNOTDONE;
		}
	    }
	}
    }

    if ( status == SCOK ) {
	/* Return the string */
	status = val_set_string_ptr( rvalP, estr_tmp_nstr( strP, dlen ) );
    }

    if ( status != SCOK ) {
	/* Return null string */
	val_set_string_ptr( rvalP, NULL );
	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    /* Cleanup and return */
    if ( keyP != NULL )
	val_release( &keyval );

    return ( status );
}


/*

*//* bif_cdb_open -- open a CDB

    $CDB   *$cdb_open( fnameP )

    Open a CDB, return handle

    fnameP is the name of the CDB file to open

*/

static STSCODE
bif_cdb_open ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		cfd;		/* File descriptor for cdb */
	void		*cdbP;
	DVS		*dvsP;		/* DVS dynamic storage thing */

    status = SCOK;
    dvsP = NULL;
    cfd = -1;

    /* Get & open the cdb filename, arg #0 */
    cfd = open_fnamearg_cdb( fnameP, treeP, 0 );
    if ( cfd < 0 ) {
	status = SCNOTDONE;
    }
    else {
	dvsP = dvs_new( Ty_cdbiP->ty_size, Ty_cdbiP->ty_size );
	if ( dvsP == NULL ) {
	    warning( "%s: can't alloc DVS", fnameP );
	    status = SCNOTDONE;
	}
	else {
	    cdbP = (struct cdb *)(dvsP->dv_dataP + 0);
	    cdb_init( cdbP, cfd );
	}
    }

    if ( status == SCOK ) {
	/* Construct the return value, referencing the $CDB$ within
	   the DVS $CDBi$
	*/
	rvalP->v_flags = VF_RVALUE;
	rvalP->v_typeP = Ty_cdbpP;
	rvalP->v_data.d_ptr.p_byteX = Ty_cdbi_x_ci_cdb;
	rvalP->v_data.d_ptr.p_ptr.p_genP = dvsP;
    }
    else {
	/* Clean up any allocated/initialized stuff */
	if ( dvsP != NULL )
	    dvs_release( dvsP );

	if ( cfd >= 0 )
	    close( cfd );

	/* Give a result value of 0, which will coerce to a NULL ptr */
	val_set_int( rvalP, 4, 0 );
    }

    return ( SCOK );
}


/*

*//* bif_cdbu_get_string -- return a string from a CDB

    string  *$cdbu_get_string( fname, key )

    Opens CDB, looks up a key, closes CDB, returns a string pointer.
      Returns NULL if key not found or if CDB not available.

    fname is a (string)

    key is a (string)

*/

static STSCODE
bif_cdbu_get_string ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		r;
	int		cfd;
	unsigned int	dlen;
	off_t		dpos;
	UBYTE		*strP;
	ESTR		*keyP;
	VALUE		keyval;
	struct cdb	cdbs;

    status = SCOK;
    keyP = NULL;

    cfd = open_fnamearg_cdb( fnameP, treeP, 0 );
    if ( cfd < 0 )
	status = SCNOTDONE;
    else if ( ( keyP = func_treearg_estr( fnameP, treeP, 1, &keyval ) )
			== NULL ) {
	warning( "%s requires a key string second argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Got key, do lookup */
	cdb_init( &cdbs, cfd );
	r = cdb_find( &cdbs, keyP->es_strP, estr_len( keyP ) );
	if ( r <= 0 )
	    status = SCNOTDONE;
	else {
	    /* Found key; read & make string */
	    dlen = cdb_datalen( &cdbs );
	    dpos = cdb_datapos( &cdbs );
	    strP = tmpstr( dlen );
	    if ( strP == NULL )
		status = SCNOTDONE;
	    else {
		r = cdb_read( &cdbs, strP, dlen, dpos );
		if ( r != 0 ) {
		    warning( "%s: key found, but error reading data.",
				fnameP );
		    status = SCNOTDONE;
		}
	    }
	}
	cdb_free( &cdbs );
    }

    if ( status == SCOK ) {
	/* Return the string */
	status = val_set_string_ptr( rvalP, estr_tmp_nstr( strP, dlen ) );
    }

    if ( status != SCOK ) {
	/* Return null string ptr */
	val_set_string_ptr( rvalP, NULL );
	if ( status == SCNOTDONE )
	    status = SCOK;
    }

    /* Cleanup and return */
    if ( keyP != NULL )
	val_release( &keyval );

    if ( cfd >= 0 )
	close( cfd );

    return ( status );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* get_cdbarg( fnameP, treeP, argN )

	Get the $CDB$ pointer passed to a built-in-function

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the cdb ptr is passed in

Returns :

	<value>		Ptr to $CDB$ object if found
			NULL if not

Notes :

	Error/warning messages are given here
*/

static void *
get_cdbarg ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	UBYTE		*dataP;
	PTNODE		*cdbargP;	/* Subtree for cdb arg */
	VALUE		cdbval;		/* To get the value.. */
	void		*cdbP;

    if ( ( cdbargP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: arg #%d (cdb pointer) not supplied.",
			fnameP, argN );
	return ( NULL );
    }

    /* Evaluate */
    if ( eval_tree( cdbargP, &cdbval ) != SCOK )
	return ( NULL );
    if ( cdbval.v_typeP != Ty_cdbpP ) {
	warning( "%s: arg #%d not $CDB$ pointer.", fnameP, argN );
	return ( NULL );
    }

    val_rvalue( &cdbval );
    dataP = (UBYTE *)valptr_resolve( &cdbval.v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "%s: arg #%d bad $CDB$ pointer.", fnameP, argN );
	return ( NULL );
    }
    dataP -= Ty_cdbi_x_ci_cdb;
    cdbP = (void *)dataP;

    val_finish( &cdbval );

    return ( cdbP );
}
/*

*//* open_fnamearg_cdb( fnameP, treeP, argN )

	Open a CDB file whose name is passed as a function argument

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the cdb ptr is passed in

Returns :

	<value>		A file descriptor if successfully opened,
			otherwise -1

Notes :

	Error/warning messages may be given here, particularly for a
	MFL invocation error.  Failure to open the CDB file will not
	generate a message.

*/

static int
open_fnamearg_cdb ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	int		cfd;		/* File descriptor for cdb */
	char		*ddirP;		/* Ptr to data directory */
	char		*cfnameP;	/* Evaluated cdb filename */
	VALUE		cfnameval;
	char		cfpath[ PATH_MAX +1 ];

    /* Get the cdb filename to open, as nul-terminated string */
    cfnameP = func_treearg_str( fnameP, treeP, argN, &cfnameval, PATH_MAX );
    if ( cfnameP == NULL ) {
	warning( "%s requires a valid filename.", fnameP );
	cfd = -1;
    }
    else {
	ddirP = data_directory( TRUE );	/* Make/get data directory path */
	if ( ddirP == NULL )
	    cfd = -1;
	else if ( snprintf( &cfpath[0], sizeof(cfpath), "%s/%s",
			        ddirP, cfnameP ) >= sizeof(cfpath) )
	    cfd = -1;
	else
	    cfd = open( &cfpath[0], O_RDONLY );

	/* Free up the evaluated name */
	val_release( &cfnameval );
    }

    return ( cfd );
}
