/* mfl-bif-dns.c -- MFL functions: DNS stuff

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

MFL built-in-functions relating to DNS operations

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"

#include "mvmfmisc.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */


/* External routines used */

extern	TYPSYM	*TystrpP;		/* (string *) */

/* Local routines and forward references */

static	STSCODE	dns_return PROTO( (VALUE *rvalP) );
static	int	get_qtarg PROTO( (char *fnameP, PTNODE *treeP, int argN) );

    /* Built-in functions contained herein */
static	STSCODE	bif_dns_query();
static	STSCODE	bif_dns_result();
static	STSCODE	bif_dns_result_next();

/* Local data publicly available */

    /* Table of built-in functions to define */
	BIFINF	Bifinf_dns[] = {
/*  function name               handler function		*/
{  "$dns_query",		bif_dns_query	},
{  "$dns_result",		bif_dns_result	},
{  "$dns_result_next",		bif_dns_result_next },

{  NULL,		NULL			}
		};


/* Private data */



/*

*//* bif_dns_init(args)

	Initialize the DNS built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called as part of bif module(s) initialization to do any
	extra setup required after functions have been registered.

*/

BOOL
bif_dns_init NOARGLIST
{
    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_dns_query -- do general DNS query

    string  *$dns_query( name, qtype )

    Does a DNS query:

      name is the name to look up;

      qtype is the query type.

    Returns a string pointer, format depending on result type.

*/

static STSCODE
bif_dns_query ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	int		qtype;		/* Query type */
	char		*nameP;
	MML_DNS		*dnsP;		/* DNS handle */
	VALUE		nameval;

    status = SCOK;
    nameP = NULL;

    /* Get the name */
    if ( ( nameP = func_treearg_str( fnameP, treeP, 0, &nameval, 255 ) )
			== NULL ) {
	warning( "%s requires a valid query name as first argument.", fnameP );
	status = SCNOTDONE;
    }
    else {
	/* Got name, get query type */
	qtype = get_qtarg( fnameP, treeP, 1 );
	if ( qtype < 0 ) {
	    warning( "%s: unknown query type.", fnameP );
	    status = SCNOTDONE;
	}
	else if ( ( dnsP = dns_handle() ) == NULL ) {
	    warning( "%s: can't access DNS.", fnameP );
	    status = SCNOTDONE;
	}
	else if ( !mdns_query( dnsP, nameP, ns_c_in, qtype ) ) {
	    status = SCNOTDONE;
	}
	else {
TODO	    /* We'll attempt to resolve one CNAME level... may want
	       to figure out some real way to deal with this some day.
	    */
	    if ( ( dnsP->dn_anstype != ns_t_cname ) ||
		 ( mdns_query( dnsP, dnsP->dn_ansP, ns_c_in, qtype ) ) ) {
		status = dns_return( rvalP );
	    }
	    else
		status = SCNOTDONE;
	}

	/* Done with name's value */
	val_release( &nameval );
    }

    if ( status != SCOK ) {
	val_set_string_ptr( rvalP, NULL );
	if ( status != SCERR )
	    status = SCOK;
    }

    return ( status );
}
/*

*//* bif_dns_result -- return last DNS query result

    string  *$dns_result()

    Gives the same result as last DNS lookup.

*/

static STSCODE
bif_dns_result ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;

    status = dns_return( rvalP );
    if ( status != SCOK )
	val_set_string_ptr( rvalP, NULL );

    return ( status == SCERR ? SCERR : SCOK );
}
/*

*//* bif_dns_result_next -- return next DNS query result

    string  *$dns_result_next()

    Steps to the next result of current DNS query, returns it.

*/

static STSCODE
bif_dns_result_next ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	MML_DNS		*dnsP;

    if ( ( dnsP = DnsP ) == NULL )
	status = SCNOTDONE;
    else if ( !mdns_ans_next( dnsP ) )
	status = SCNOTDONE;
    else
	status = dns_return( rvalP );

    if ( status != SCOK )
	val_set_string_ptr( rvalP, NULL );

    return ( status == SCERR ? SCERR : SCOK );
}




/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* dns_return( rvalP )

	Provides the current DNS return value as a string pointer value

Accepts :

	rvalP		Where to put the VALUE

Returns :

	<value>		SCOK if valid return
			SCNOTDONE if not (e.g. no more returns)
			perhaps other status if some problem.

Notes :

	We only know about a few return types, as required for
	whatever need I happen to have.

	returns that have multiple values are formatted rigidly for
	access to MFL users.  See inline.

	rvalP is not guaranteed to be set if return is other than SCOK

*/

STSCODE
dns_return ARGLIST( (rvalP) )
    FARG( VALUE		*rvalP )	/* Where to put the return */
{
	STSCODE		status;
	int		i;
	char		*strP;		/* A result string */
	MML_DNS		*dnsP;
	UBYTE		ip_octets[4];

    /* No value if DNS not open.  Here we touch the global handle. */
    if ( ( dnsP = DnsP ) == NULL )
	status = SCNOTDONE;
    else {
	switch ( dnsP->dn_anstype ) {
	    case ns_t_a:		/* IPv4 address */
		mdns_ip4_iptoba( ntohl( dnsP->dn_ipaddr ), &ip_octets[0] );
		/* Format as "n.n.n.n" */
		i = 20;			/* 20 chars should be excessive */
		strP = tmpstr( i +1 );
		if ( ( strP == NULL ) ||
		     ( snprintf( strP, i, "%d.%d.%d.%d",
				   ip_octets[0],
				   ip_octets[1],
				   ip_octets[2],
				   ip_octets[3] ) >= i ) ) {
		    status = SCNOTDONE;
		}
		else
		    status = SCOK;
	        break;

	    case ns_t_ptr:		/* PTR string */
	    case ns_t_txt:		/* PTR string */
		/* Format as a simple string */
		strP = tmpstr( strlen( dnsP->dn_ansP ) + 1 );
		if ( strP == NULL )
		    status = SCNOTDONE;
		else {
		    strcpy( strP, dnsP->dn_ansP );
		    status = SCOK;
		}
		break;

	    case ns_t_mx:		/* MX */
		/* Format as "nnn ssss" -- 3 digit priority + string */
		i = strlen( dnsP->dn_ansP ) + 5;
		strP = tmpstr( i + 1 );
		if ( ( strP == NULL ) ||
		     ( snprintf( strP, i, "%03d %s",
					dnsP->dn_val1,
					dnsP->dn_ansP ) >= i ) ) {
		    status = SCNOTDONE;
		}
		else
		    status = SCOK;
		break;

	    default:
		status = SCNOTDONE;
		break;
	}
    }

    if ( status == SCOK ) {
	val_set_string_ptr( rvalP, estr_tmp_str( strP ) );
    }
#if 0
    else {
	val_set_string_ptr( rvalP, NULL );
    }
#endif

    return ( status );
}

/*

*//* get_qtarg( fnameP, treeP, argN )

	Get the query type passed to a DNS function

Accepts :

	fnameP		Function name (for messages)
	treeP		Ptr to argument tree passed to bif
	argN		Argument number that the query type is passed in

Returns :

	<value>		>= 0: Query type
			-1 if error or no such type.

Notes :

	Error/warning messages are given here
*/

static int
get_qtarg ARGLIST( (fnameP, treeP, argN) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Passed tree */
    FARG( int		argN )		/* Arg number */
{
	int		qt;		/* Query type */
	ESTR		*qtP;		/* Query type string */
	VALUE		qtval;		/* To hold evaluated value */

    /* Get the query type name */
    if ( ( qtP = func_treearg_estr( fnameP, treeP, argN, &qtval ) ) == NULL ) {
	warning( "%s requires a query name", fnameP );
	qt = -1;
    }
    else {
	qt = mdns_query_type_str( ns_c_in, tmpstr_estr( qtP ) );
	val_release( &qtval );
    }

    return ( qt );
}
