/* mfl-eval.c -- Tree evaluation routines

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

There are multiple modules that "run" the parse tree.  This
one contains evaluation functions which, in general, produce
results from terms in the tree.  This is distinct from,
for example, execution functions, which would generally be
at a higher level (e.g. executing control statements as opposed
to evaluating expressions).


*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include <mml/mml.h>

#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-exec.h"
#include "mfl-func.h"
#include "mfl-sieve.h"

#include "errwarn.h"

/* Local definitions */

    /* Hack to define this */
#ifndef	MAXFLOAT
#define MAXFLOAT        ((float)3.40282346638528860e+38)
#endif	/* MAXFLOAT */

/* External data referenced */

extern	BISINFO	Bistbl[];		/* Built-in statement table */
extern	OPINFO	Optbl[];		/* Operator table */
extern	SYMTAB	*STglobal;		/* Global variables symbab */
extern	SYMTAB	*STtpool;		/* Type pool symtab */
extern	TNSYM	*Tnflt4;
extern	TNSYM	*Tnflt8;
extern	TNSYM	*Tnint1;
extern	TNSYM	*Tnint2;
extern	TNSYM	*Tnint4;

extern	TYPSYM	*Tyint1pP;


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Common code for increment/decrement operations */
static	STSCODE	evs_incdec PROTO( ( OPINFO *opP, VALUE *rvalP,
				    VALUE *valP, int delta,
				    BOOL preF ) );


/* Private data */



/*

*//* eval_addr( rvalP, opP, valP )

	Execute "address of" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_addr ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    /* Just return the lvalue */
    return( val_lvalue( rvalP, valP ) );
}
/*

*//* eval_aref( rvalP, opP, val1P, val2P )

	Execute "array reference" node (binary operator)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_aref ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;

    /* Use the plus operation to perform the pointer arithmetic */
    status = eval_plus( rvalP, opP, val1P, val2P );

    /* If successful addition, dereference the result. */
    if ( status == SCOK )
        status = val_deref( rvalP, rvalP );

    return( status );
}
/*

*//* eval_arith( rvalP, opP, val1P, val2P )

	Execute binary arithmetic operator node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_arith ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;
	long		dint;		/* Long integer */
	double		dfloat;		/* Long float */
	TYPE		*ntypeP;	/* A nother type */

    /* Get the common numeric type. */
    ntypeP = val_common_type( val1P, val2P );
    if ( ( ntypeP == NULL ) ||
         ( ( ( ntypeP->ty_class == TC_INT ) ||
 	     ( ntypeP->ty_class == TC_UINT ) ) &&
           ( opP->op_ieval == NULL ) ) ||
         ( ( ntypeP->ty_class == TC_FLOAT ) &&
	   ( opP->op_feval == NULL ) ) ) {
       	warning( "eval_arith: invalid arguments to \"%s\"", opP->op_str );
	status = SCNOTDONE;
    }
    else {
    	/* Make them both compatible common types */
        val_coerce( val1P, ntypeP, FALSE );
	val_coerce( val2P, ntypeP, FALSE );
	memcpy( rvalP, val1P, sizeof(VALUE) );
	val_access( rvalP );
	status = SCOK;

	/* Perform the operation depending on type */
	switch( ntypeP->ty_class ) {
	    case TC_INT:
	    case TC_UINT:
	    	dint = (*opP->op_ieval)( val_get_int( val1P ),
				         val_get_int( val2P ) );
	        val_put_int( rvalP, dint );
		break;


	    case TC_FLOAT:
		switch( ntypeP->ty_size ) {
		    case 4:
		        dfloat = (*opP->op_feval)( rvalP->v_data.d_float,
					val2P->v_data.d_float );
			rvalP->v_data.d_float = (float)dfloat;
			break;

		    case 8:
		        dfloat = (*opP->op_feval)( rvalP->v_data.d_double,
					val2P->v_data.d_double );
			rvalP->v_data.d_double = dfloat;
			break;
		}
		break;
	}

	/* Make result attributes reflect the source */
	val_attrprop( rvalP, val1P, val2P );
    }

    return( status );
}
/*

*//* eval_assign( rvalP, opP, val1P, val2P )

	Execute "assign" binary operator node (assignment operator)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_assign ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;

    /* Assign the value */
    status = val_assign( val1P, val2P );

    /* If successsful, copy the result into the result term
       and take an access.
    */
    if ( status == SCOK ) {
	memcpy( rvalP, val1P, sizeof( VALUE ) );
	val_access( rvalP );
    }

    return( status );
}
/*

*//* eval_bitcomp( rvalP, opP, valP )

	Execute "bitwise complement" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_bitcomp ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
	STSCODE		status;

    /* Value must be an integer */
    if ( val_isint( valP, TRUE ) ) {
        /* Copy source to result */
	memcpy( rvalP, valP, sizeof(VALUE) );
	val_access( rvalP );		/* Probably not necessary */

	/* Make its value real */
	val_rvalue( rvalP );

	/* Complement the result */
	val_put_int( rvalP, ~val_get_int( rvalP ) );

	/* Ensure propagation of attributes from source */
	val_attrprop( rvalP, valP, valP );

	status = SCOK;
    }
    else
	status = SCNOTDONE;

    return( status );
}
/*

*//* eval_comma( rvalP, opP, val1P, val2P )

	Execute "comma" binary operator node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :

	Comma separates two expressions.  Each is evaluated in
	turn, resulting in the value of the second.

*/

STSCODE
eval_comma ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
    /* Both subexpressions should already have been evaluated... 
       return the value of the second one. */
    memcpy( rvalP, val2P, sizeof( VALUE ) );
    val_access( rvalP );

    return( SCOK );
}
/*

*//* eval_condex( rvalP, opP, val1P, cvnodeP )

	Execute "condex" binary operator node (conditional expression)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	cvnodeP		Ptr to conditional value tree 

Returns :

	<value>		SCxx status code

Notes :

	This has a slightly different calling sequence from other
	binary operators- the parsing is done specially.

*/

STSCODE
eval_condex ARGLIST( (rvalP, opP, val1P, cvnodeP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( PTNODE	*cvnodeP )	/* Ptr to conditional value */
{
	STSCODE		status;

    /* right side must be an EXPR placeholder. */
    if ( ( cvnodeP == NULL ) || ( cvnodeP->pt_type != PN_EXPR ) ) {
	warning( "Internal error in eval_condex()!" );
	status = SCNOTDONE;
    }
    else {
	/* Left side must be convertable to int */
	if ( val_isint( val1P, FALSE ) ) {
	    /* Make sure we have the value */
	    val_rvalue( val1P );

	    /* Process left or right subtree of c.v. depending on
	       this value's truth. */
	    if ( val_get_int( val1P ) )
		status = eval_tree( cvnodeP->pt_leftP, rvalP );
	    else
		status = eval_tree( cvnodeP->pt_rightP, rvalP );
	}
	else
	    status = SCNOTDONE;
    }

    return( status );
}
/*

*//* eval_deref( rvalP, opP, valP )

	Execute "pointer dereference" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_deref ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    /* Simply hand this off to the dereferencing routine */
    return( val_deref( rvalP, valP ) );
}
/*

*//* eval_land( rvalP, opP, val1P, term2P )

	Execute binary(infix) logical "AND" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	term2P		Ptr to tree for second term

Returns :

	<value>		SCxx status code

Notes :

	Second term is only evaluated if necessary

*/

STSCODE
eval_land ARGLIST( (rvalP, opP, val1P, term2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( PTNODE	*term2P )	/* Ptr to the second term's tree */
{
	STSCODE		status;
	WORD		result;		/* Boolean result */

    /* Init the result to the first term. */
    val_rvalue( val1P );
    result = val_get_int( val1P );
    if ( result ) {
	/* First term is TRUE- result of AND now entirely depends
	   on the second term.
	*/
	status = eval_tree_word( term2P, &result );
	if ( status != SCOK )
	    result = FALSE;
    }
    else {
	/* First term is FALSE-- entire AND is false. */
	status = SCOK;
    }

    val_set_int( rvalP, 4, result );	/* Return the result. */

    return( status );
}
/*

*//* eval_lor( rvalP, opP, val1P, term2P )

	Execute binary(infix) logical "OR" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	term2P		Ptr to tree for second term

Returns :

	<value>		SCxx status code

Notes :

	Second term is only evaluated if necessary

*/

STSCODE
eval_lor ARGLIST( (rvalP, opP, val1P, term2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( PTNODE	*term2P )	/* Ptr to the second term's tree */
{
	STSCODE		status;
	WORD		result;		/* Boolean result */

    /* Init the result to the first term. */
    val_rvalue( val1P );
    result = val_get_int( val1P );
    if ( result ) {
	/* First term is TRUE -- entire OR is true, and there's no
	   need to evaluate the second term.
	*/
	status = SCOK;
    }
    else {
	/* First term is FALSE- result of OR now entirely depends
	   on the second term.
	*/
	status = eval_tree_word( term2P, &result );
	if ( status != SCOK )
	    result = FALSE;
    }

    val_set_int( rvalP, 4, result );	/* Return the result. */

    return( status );
}
/*

*//* eval_mineq( rvalP, opP, val1P, val2P )

	Execute "mineq" binary operator node ( "-=" )

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_mineq ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;		/* Status code */
	VALUE		rvalue;		/* Result value */

    /* Save away the destination to preserve its lvalue */
    memcpy( &rvalue, val1P, sizeof( VALUE ) );

    /* Do the subtraction */
    status = eval_minus( rvalP, opP, val1P, val2P );

    /* Do the assignment */
    if ( status == SCOK )
	status = val_assign( &rvalue, rvalP );

    return( status );
}
/*

*//* eval_minus( rvalP, opP, val1P, val2P )

	Execute "minus" binary operator node (subtraction)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_minus ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;
	TYPE		*type1P;	/* First type */
	TYPE		*type2P;	/* Second type */
	TYPE		*ntypeP;	/* A nother type */

    /* Assume success */
    status = SCOK;

    /* Get types */
    type1P = val1P->v_typeP;
    type2P = val2P->v_typeP;

    /* Check something-minus-ptr */
    if ( ( type2P->ty_class == TC_PTR ) ||
         ( type2P->ty_class == TC_DIM ) ) {
	/* Must be ptr-minus-ptr */
	if ( ( type1P->ty_class != TC_PTR ) &&
	     ( type1P->ty_class != TC_DIM ) ) {
	    warning( "Invalid pointer subtraction." );
	    status = SCNOTDONE;
	}
	else if ( ( type1P->ty_nextP != type2P->ty_nextP ) ||
		  ( type1P->ty_size != type2P->ty_size ) ) {
	    warning( "Subtraction of incompatible pointers." );
	    status = SCNOTDONE;
	}

	/* Can only subtract pointers to things within the same
	   storage element.
	*/
	else if ( val1P->v_data.d_ptr.p_ptr.p_genP !=
		  val2P->v_data.d_ptr.p_ptr.p_genP ) {
	    warning( "Subtraction of pointers to different objects." );
	    status = SCNOTDONE;
	}

	else {

#if 0
	    /* rvalue the terms */
	    val_rvalue( val1P );
	    val_rvalue( val2P );
#endif

	    /* Get referenced type to get subtraction scale */
	    ntypeP = (TYPE *)type2P->ty_nextP;

	    /* Do the subtraction; yield an int. */
	    rvalP->v_flags = VF_RVALUE;
	    rvalP->v_typeP = Tnint4->tn_typeP;
	    rvalP->v_data.d_long = (long)val1P->v_data.d_ptr.p_byteX -
					 val2P->v_data.d_ptr.p_byteX;
	    
	    /* Propagate attributes */
	    val_attrprop( rvalP, val1P, val2P );
	}
    }
    else {				/* Right term is not a pointer */
        if ( ( type1P->ty_class == TC_PTR ) ||
	     ( type1P->ty_class == TC_DIM ) ) {
	    
	    /* Subtraction of non-pointer from pointer.  Make sure
	       the second term is an integer. */
	    if ( type2P->ty_class != TC_INT ) {
	    	warning( "Subtraction of non-integer from pointer." );
		status = SCNOTDONE;
	    }

	    else {
		/* Use the "plus" logic rather than duplicate it all
		   here.
		*/

		val_rvalue( val2P );	/* Need second term's value */
		val2P->v_data.d_long = -val2P->v_data.d_long;
		status = eval_plusptr( rvalP, opP, val1P, val2P );
	    }
	}
	else
	    /* Subtraction of two non-pointers. */
	    status = eval_arith( rvalP, opP, val1P, val2P );
    }

    return( status );
}
/*

*//* eval_node(treeP, rvalP)

	Evaluate a node in a tree

Accepts :

	treeP		Ptr to node to evaluate
	rvalP		Where to return the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

STSCODE
eval_node ARGLIST( (treeP, rvalP) )
   NFARG( PTNODE	*treeP )	/* Tree to evaluate */
    FARG( VALUE		*rvalP )	/* Where to return the value */
{
	STSCODE		status;		/* Status code */
	int		len;
	int		n;
	VSYM		*vsymP;		/* Ptr to global variable entry */

    status = SCOK;			/* Presume good status */

    switch( treeP->pt_type ) {
	case	PN_INTEGER:		/* Integer constant */
	    rvalP->v_flags = VF_CONST|VF_RVALUE;
	    rvalP->v_typeP = Tnint4->tn_typeP;
	    rvalP->v_data.d_long = treeP->pt_data.pv_long;
	    break;

				  	/* Float constant */
	case	PN_FLOAT:
	    rvalP->v_flags = VF_CONST|VF_RVALUE;
	    rvalP->v_typeP = Tnflt8->tn_typeP;
	    rvalP->v_data.d_double = treeP->pt_data.pv_float;
	    break;

	case	PN_SYM:
	    vsymP = (VSYM *)symtab_lookup( STglobal, treeP->pt_info.od_P );
	    if ( vsymP == NULL ) {
	    	warning( "variable %s is not declared.", treeP->pt_info.od_P );
		status = SCNOTDONE;
	    }
	    else {
		/* Make a new reference to the symbol's value */
		val_ref( rvalP, &vsymP->vs_value );
	    }
	    break;


	case	PN_LITSTR:
	    /* Make a free-standing NUL-terminated string using DVS */
	    len = strlen( (char *)treeP->pt_info.od_P );

	    rvalP->v_flags = VF_RVALUE;
	    rvalP->v_typeP = &Tyint1pP->ts_type;
	    rvalP->v_data.d_ptr.p_ptr.p_dvsP = dvs_new( len + 1, 0 );
	    if ( rvalP->v_data.d_ptr.p_ptr.p_dvsP == NULL ) {
		warning( "eval_node: dvs_new failed for litstring." );
		status = SCERR;
	    }
	    else {
		rvalP->v_data.d_ptr.p_byteX = 0;

		/* Copy the string */
		strcpy( (char *)rvalP->v_data.d_ptr.p_ptr.p_dvsP->dv_dataP,
				(char *)treeP->pt_info.od_P );
	    }
	    break;


	case	PN_ARG:
	    n = treeP->pt_data.pv_int;
	    status = mfl_func_arg( rvalP, n );
	    break;


	default:			/* Something bad */
	    warning(
"internal error at eval_node - invalid node type %d found",
	    		treeP->pt_type );
	    status = SCERR;
	    break;
    }

    return( status );
}
/*

*//* eval_operator(nodeP, rvalP)

	Evaluate an operator node in a tree

Accepts :

	nodeP		Ptr to node to evaluate
	rvalP		Where to return the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

STSCODE
eval_operator ARGLIST( (nodeP, rvalP) )
   NFARG( PTNODE	*nodeP )	/* Node to evaluate */
    FARG( VALUE		*rvalP )	/* Where to return the value */
{
	STSCODE		status;		/* Status code */
	int		opX;		/* Operator table index */
	OPINFO		*opP;		/* Ptr to operator data */
	VALUE		value1;
	VALUE		value2;

    opX = nodeP->pt_data.pv_int;	/* Get operator table index */
    opP = &Optbl[opX];			/* Get ptr to operator entry */

    if ( nodeP->pt_leftP == NULL ) {     /* Prefix operator */
 	status = eval_tree( nodeP->pt_rightP, &value2 );
	if ( status == SCOK ) { 
	    status = (*opP->op_preeval)( rvalP, opP, &value2 );
	    val_release( &value2 );
	}
    }
    else if ( nodeP->pt_rightP == NULL ) {    /* Postfix operator */
	status = eval_tree( nodeP->pt_leftP, &value1 );
	if ( status == SCOK ) {
	    status = (*opP->op_posteval)( rvalP, opP, &value1 );
	    val_release( &value1 );
	}
    }
    else if ( ( opP->op_prio == ORMEMBER &&
    	        *opP->op_str != '[' ) ) {
    	/* Aggregate membership operation */
	if ( nodeP->pt_rightP->pt_type != PN_SYM ) {
	    warning( "right side of \"%s\" must be a symbol.",
	    	opP->op_str );
	    status = SCNOTDONE;
	}
	else {
	    /* Evaluate left side. */
	    status = eval_tree( nodeP->pt_leftP, &value1 );

	    /* Do the membership operation. */
	    if ( status == SCOK ) {
	    	status = (*opP->op_eval)( rvalP, &value1, 
				   nodeP->pt_rightP->pt_info.od_P );
		val_release( &value1 );
	    }
	}
    }
    else if ( ( opP->op_prio == ORCOND ) ||
	      ( opP->op_prio == ORLAND ) ||
	      ( opP->op_prio == ORLOR ) ) {
	/* Conditional or boolean operator; don't evaluate right side */
	status = eval_tree( nodeP->pt_leftP, &value1 );
	if ( status == SCOK ) {
	    status = (*opP->op_eval)( rvalP, opP, &value1,
					nodeP->pt_rightP );
	    val_release( &value1 );
	}
    }
    else {				/* Binary (infix) operator */
	status = eval_tree( nodeP->pt_leftP, &value1 );
	if ( status == SCOK ) {
 	    status = eval_tree( nodeP->pt_rightP, &value2 );
	    if ( status == SCOK ) {
		status = (*opP->op_eval)( rvalP, opP, &value1, &value2 );
		val_release( &value2 );
	    }
	    val_release( &value1 );
	}
    }

    return( status );
}
/*

*//* eval_opeq( rvalP, opP, val1P, val2P )

	Execute "opeq" binary operator node (generic operator=)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_opeq ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;		/* Status code */
	VALUE		rvalue;		/* Result value */

    /* Save away the destination to preserve its lvalue */
    memcpy( &rvalue, val1P, sizeof( VALUE ) );

    /* Perform the operation */
    status = eval_arith( rvalP, opP, val1P, val2P );

    /* Do the assignment */
    if ( status == SCOK )
	status = val_assign( &rvalue, rvalP );

    return( status );
}
/*

*//* eval_plus( rvalP, opP, val1P, val2P )

	Execute "plus" binary operator node (addition)

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_plus ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
    /* If either term is a string, yield a string. */
    if ( ( val1P->v_typeP->ty_class == TC_STRING ) ||
         ( val2P->v_typeP->ty_class == TC_STRING ) )
	return ( eval_plusstr(rvalP, opP, val1P, val2P) );

    /* Check for the addition of a pointer and something */
    if ( ( val1P->v_typeP->ty_class == TC_PTR ) ||
         ( val1P->v_typeP->ty_class == TC_DIM ) )
	return ( eval_plusptr( rvalP, opP, val1P, val2P ) );

    /* Check for the addition of something and a pointer */
   if ( ( val2P->v_typeP->ty_class == TC_PTR ) ||
    	( val2P->v_typeP->ty_class == TC_DIM ) )
	return ( eval_plusptr( rvalP, opP, val2P, val1P ) );

    /* Just do arithmetic */
    return ( eval_arith( rvalP, opP, val1P, val2P ) );
}
/*

*//* eval_plusptr( rvalP, opP, val1P, val2P )

	Execute "plus" binary operator node (addition) for pointer types

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value, which must be a pointer type.
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :

	This function is only called when val1P represents a pointer type.
	Caller must ensure this.

*/

STSCODE
eval_plusptr ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;
	int		i;
	TYPSYM		*tpsymP;	/* Type pool symbol */
	TYPE		*ntypeP;	/* A nother type */
	REFSTR		*srsP, *rsP;	/* Source and new refstrs */
	VALPTR		*vptrP;
	FSV		*fsvP;

    /* Second term can't be a pointer */
    if ( ( val2P->v_typeP->ty_class == TC_PTR ) ||
	 ( val2P->v_typeP->ty_class == TC_DIM ) ) {
	warning( "illegal addition of two pointers." );
	return ( SCNOTDONE );
    }

    /* Make sure right hand value is of the proper type
       (note this gets the rvalue)
    */
    status = val_coerce( val2P, Tnint4->tn_typeP, FALSE );
    if ( status != SCOK ) {
        warning( "Illegal addition." );
	return ( SCNOTDONE );
    }

    val_rvalue( val1P );		/* Must have actual pointer value */

    /* Init the result. */
    memcpy( rvalP, val1P, sizeof(VALUE) );
    /* Note: val_access later as necessary */

TODO		/* This is out now that rvalue is done above. */
#if 0
    if ( val1P->v_typeP->ty_class == TC_DIM ) {
TODO	/* Save some dimension info from the array */

	/* We have the real pointer available but of the
	   wrong reference type (it needs to be a PTR rather
	   than a DIM.  Ignore that for now- we're going to
	   re-type it once we are successful.  However
	   we are going to make it be an rvalue (with
	   the value being the pointer already in place).
	*/
	rvalP->v_flags &= ~VF_LVALUE;
	rvalP->v_flags |= VF_RVALUE;
    }
#endif

    if ( rvalP->v_data.d_ptr.p_byteX < 0 ) {
	/* Special pointer types. */
	switch ( rvalP->v_data.d_ptr.p_byteX ) {
	    case PO_UNDEF:		/* Not defined */
		warning( "Addition to undefined pointer." );
		status = SCNOTDONE;
		break;

	    case PO_SIMPLE:		/* v_rvalue element */
		warning( "Pointer addition out of bounds." );
		status = SCNOTDONE;
		break;

	    case PO_PRIVATE:		/* Something for this type */
		warning( "Undefined addition to private pointer." );
		status = SCNOTDONE;
		break;

	    case PO_REFSTR:		/* A string */
		/* We should never get here... this would be a string
		   addition rather than a pointer addition.
		*/
		warning( "Addition to string in eval_plusptr shouldn't happen." );
		status = SCERR;
		break;


	    case PO_FSV:		/* Free-standing value */
		/* At present this is a special case for string pointers. */
		vptrP = (VALPTR *)valptr_resolve( &rvalP->v_data.d_ptr );
		if ( vptrP == NULL ) {
		    warning( "Null pointer reference." );
		    status = SCNOTDONE;
		}
		else if ( vptrP->p_byteX != PO_REFSTR ) {
		    warning( "eval_plusptr can't handle FSV." );
		    status = SCNOTDONE;
		}
		else if ( ( srsP = (REFSTR *)vptrP->p_ptr.p_genP ) == NULL ) {
		    warning( "eval_plusptr null string pointer reference." );
		    status = SCNOTDONE;
		}
		else if ( ( rsP = refstr_ref( srsP, 0, 0 ) ) == NULL ) {
		    warning( "eval_plusptr: refstr_new failed." );
		    status = SCERR;
		}
		else if ( ( fsvP = fsv_new() ) == NULL ) {
		    warning( "eval_plusptr: fsv_new failed for string." );
		    status = SCERR;
		}
		else {
		    /* Need actual copy of refstr offsets */
		    rsP->rs_startX = srsP->rs_startX;
		    rsP->rs_endX = srsP->rs_endX;

		    vptrP = (VALPTR *)&fsvP->fv_data;
		    vptrP->p_byteX = PO_REFSTR;
		    vptrP->p_ptr.p_genP = rsP;

		    fsvP->fv_ptrF = TRUE;

		    rvalP->v_data.d_ptr.p_ptr.p_genP = fsvP;

		    if ( !refstr_move( rsP, val2P->v_data.d_long ) ) {
TODO			/* string didn't move, but it's probably OK. */
			status = SCOK;
		    }
		    else
			status = SCOK;
		}
		break;

	    default:			/* unknown */
		warning( "eval_plusptr: unknown pointer type %d",
				rvalP->v_data.d_ptr.p_byteX );
		status = SCERR;
		break;
	}
    }

    else {
	/* Get pointed-to type to determine its size */
	ntypeP = (TYPE *)val1P->v_typeP->ty_nextP;

	/* Add the offset, scaled by size, to the pointer */
	rvalP->v_data.d_ptr.p_byteX += 
	    (ntypeP->ty_size * val2P->v_data.d_long);

	/* Check underflow */
	if ( rvalP->v_data.d_ptr.p_byteX < 0 ) {
	    warning( "Pointer out of bounds." );
	    rvalP->v_data.d_ptr.p_byteX = 0;
	    status = SCNOTDONE;
	}

TODO	/* Should check overflow here */

	/* If array ref, convert to pointer ref */
	if ( ( status == SCOK ) && ( val1P->v_typeP->ty_class == TC_DIM ) ) {
	    if ( ( tpsymP = type_ptr( ntypeP ) ) == NULL )
		status = SCERR;
	    else
		rvalP->v_typeP = &(tpsymP->ts_type);
	}

	if ( status == SCOK ) {
	    /* Propagate attributes */
	    val_attrprop( rvalP, val1P, val2P );

	    /* OK to access the new value now */
	    val_access( rvalP );
	}
    }

    return( status );
}
/*

*//* eval_plusstr( rvalP, opP, val1P, val2P )

	Execute "plus" binary operator node (addition) for strings

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :

	Adds two terms yielding a string result.  Note: one of the terms
	must be a string to begin with.

*/

STSCODE
eval_plusstr ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;
	int		rlen;		/* Result/source lengths */
	VALUE		result;		/* Result string value */
	REFSTR		*rrsP;		/* Result string */
	ESTR		*sesP;		/* Source ESTR */

    /* From the first term, make a string value for the result.  It's OK to
       steal the first term's string if we can.
    */
    status = val_string_val( &result, val1P, TRUE );
    if ( status == SCOK ) {
	/* Get access to the refstr */
	rrsP = (REFSTR *)valptr_resolve( &result.v_data.d_ptr );
	if ( rrsP == NULL ) {
	    warning( "eval_plusstr: non-string result." );
	    status = SCERR;
	}
    }

    if ( status == SCOK ) {
	/* Append second string based on its type */
	rlen = refstr_len( rrsP );
	switch ( val2P->v_typeP->ty_class ) {
	    /* Types where we can generate a simple string */
	    case TC_INT:
	    case TC_UINT:
	    case TC_ENUM:
	    case TC_FLOAT:
		sesP = val_estr_tmp( val2P );
		if ( sesP == NULL ) {
		    warning( "eval_plusstr: programming error." );
		    status = SCERR;
		}
		else
		    refstr_replacee( rrsP, rlen, rlen, sesP );
		break;


	    /* Types where we might be able to access the source string.
	       (similar to the INT cases, but error is a different
	       classification.)
	    */
	    case TC_PTR:
	    case TC_STRING:
		sesP = val_estr_tmp( val2P );
		if ( sesP == NULL ) {
		    warning( "eval_plusstr: can't get string from second term pointer." );
		    status = SCNOTDONE;
		}
		else
		    refstr_replacee( rrsP, rlen, rlen, sesP );
		break;
		    

	    default:
		warning( "eval_plusstr: can't deal with second term type %d",
			    val2P->v_typeP->ty_class );
		status = SCNOTDONE;
		break;
	}
    }

    if ( status == SCOK )
	memcpy( rvalP, &result, sizeof(VALUE) );

    return ( status );
}
/*

*//* eval_pluseq( rvalP, opP, val1P, val2P )

	Execute "pluseq" binary operator node ( += operator )

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_pluseq ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;		/* Status code */
	VALUE		rvalue;		/* Result value */

    /* Save away the destination to preserve its lvalue */
    memcpy( &rvalue, val1P, sizeof( VALUE ) );

    /* Do the addition */
    status = eval_plus( rvalP, opP, val1P, val2P );

    /* Do the assignment */
    if ( status == SCOK )
	status = val_assign( &rvalue, rvalP );

    return( status );
}
/*

*//* eval_postdecr( rvalP, opP, valP )

	Execute "post-decrement" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_postdecr ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    return( evs_incdec( opP, rvalP, valP, -1, FALSE ) );
}
/*

*//* eval_postincr( rvalP, opP, valP )

	Execute "post-increment" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_postincr ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    return( evs_incdec( opP, rvalP, valP, 1, FALSE ) );
}
/*

*//* eval_predecr( rvalP, opP, valP )

	Execute "pre-decrement" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_predecr ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    return( evs_incdec( opP, rvalP, valP, -1, TRUE ) );
}
/*

*//* eval_preincr( rvalP, opP, valP )

	Execute "pre-increment" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_preincr ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
    return( evs_incdec( opP, rvalP, valP, 1, TRUE ) );
}
/*

*//* eval_preminus( rvalP, opP, valP )

	Execute "prefix minus" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_preminus ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
	VALUE		zerovalue;	/* Zero to subtract from */

    /* Setup a zero value */
    zerovalue.v_flags = VF_CONST|VF_RVALUE;
    zerovalue.v_typeP = Tnint4->tn_typeP;
    zerovalue.v_data.d_long = 0;

    /* Perform binary subtraction from zero */
    return( eval_minus( rvalP, opP, &zerovalue, valP ) );
}
/*

*//* eval_prenot( rvalP, opP, valP )

	Execute "prefix not" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_prenot ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{
	STSCODE		status;

    /* Value must be convertable to integer */
    if ( val_isint( valP, FALSE ) ) {
        /* Need rvalue of source */
	val_rvalue( valP );

        /* Make result be an int-2 */
	rvalP->v_flags = VF_RVALUE;
	rvalP->v_typeP = Tnint2->tn_typeP;
	rvalP->v_data.d_word = !(int)val_get_int( valP );

	/* Propagate type attributes */
	val_attrprop( rvalP, valP, valP );

	status = SCOK;
    }
    else
	status = SCNOTDONE;

    return( status );
}
/*

*//* eval_preplus( rvalP, opP, valP )

	Execute "prefix plus" node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	valP		Ptr to the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_preplus ARGLIST( (rvalP, opP, valP ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
    FARG( VALUE		*valP )		/* Ptr to the value */
{

    /* Prefix plus is a no-op */
    memcpy( rvalP, valP, sizeof( VALUE ) );
    val_access( rvalP );
    return( SCOK );
}
/*

*//* eval_rel( rvalP, opP, val1P, val2P )

	Execute relational binary operator node

Accepts :

	rvalP		Where to put the result
	opP		Ptr to operator table entry
	val1P		Ptr to the first value
	val2P		Ptr to the second value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_rel ARGLIST( (rvalP, opP, val1P, val2P ) )
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*val1P )	/* Ptr to the first value */
    FARG( VALUE		*val2P )	/* Ptr to the second value */
{
	STSCODE		status;
	long		dint;		/* Long integer */
	double		dfloat;		/* Long float */
	TYPE		*ntypeP;	/* A nother type */

    /* If either type is a string type, do a string operation */
    if ( type_is_string( val1P->v_typeP ) ||
         type_is_string( val2P->v_typeP ) ) {
	if ( opP->op_seval == NULL ) {
	    warning( "\"%s\" not supported for strings.", opP->op_str );
	    status = SCNOTDONE;
	}
	else {
	    status = (*opP->op_seval)( rvalP, val1P, val2P );
	}
	return ( status );
    }

    /* Get the common numeric type. */
    ntypeP = val_common_type( val1P, val2P );

    /* If the common type is pointer, use unsigned long */
    if ( ( ntypeP != NULL ) && ( ntypeP->ty_class == TC_PTR ) ) {
    	ntypeP = type_uint( 4 );
    }

    if ( ( ntypeP == NULL ) ||
         ( ( ntypeP->ty_class == TC_INT ) && ( opP->op_ieval == NULL ) ) ||
         ( ( ntypeP->ty_class == TC_FLOAT ) && ( opP->op_feval == NULL ) ) ) 
{
       	warning( "eval_rel: invalid arguments to \"%s\"", opP->op_str );
	status = SCNOTDONE;
    }
    else {
    	/* Make them both compatible common types */
        val_coerce( val1P, ntypeP, FALSE );
	val_coerce( val2P, ntypeP, FALSE );

	/* The result type is an integer-2 */
	rvalP->v_flags = VF_RVALUE;
	rvalP->v_typeP = Tnint2->tn_typeP;

	status = SCOK;

	/* Perform the operation depending on type */
	switch( ntypeP->ty_class ) {
	    case TC_INT:
	    case TC_UINT:
	    case TC_ENUM:
	    	dint = (*opP->op_ieval)( val_get_int( val1P ),
					 val_get_int( val2P ) );
		break;


	    case TC_FLOAT:
		switch( ntypeP->ty_size ) {
		    case 4:
		        dfloat = (*opP->op_feval)( val1P->v_data.d_float,
					val2P->v_data.d_float );
			break;

		    case 8:
		        dfloat = (*opP->op_feval)( val1P->v_data.d_double,
					val2P->v_data.d_double );
			break;
		}
		dint = (long)dfloat;
		break;
	}

	val_put_int( rvalP, dint );

	/* Make result attributes reflect the source */
	val_attrprop( rvalP, val1P, val2P );
    }

    return( status );
}
/*

*//* eval_tree(treeP, rvalP)

	Evaluate an expression tree

Accepts :

	treeP		Ptr to top node of expression
	rvalP		Where to return the value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
eval_tree ARGLIST( (treeP, rvalP) )
   NFARG( PTNODE	*treeP )	/* Tree to evaluate */
    FARG( VALUE		*rvalP )	/* Where to return the value */
{
	STSCODE		status;		/* Status code */

    if ( treeP == NULL ) {		/* No tree?? */
	val_set_int( rvalP, 4, 0 );	/* Make constant zero */
	return( SCOK );
    }

    status = SCOK;				/* Presume good status */

    switch( treeP->pt_type ) {
	case	PN_LSEQ:			/* Left-wise sequencing */
	    status = exec_treeval( treeP->pt_rightP, rvalP, FALSE );
	    if ( status == SCOK )
	    	status = exec_treeval( treeP->pt_leftP, rvalP, FALSE );
	    break;


	case	PN_INTEGER:		/* Value node types */
	case	PN_FLOAT:
	case	PN_SYM:
	case	PN_LITSTR:
	case	PN_ARG:
	    status = eval_node( treeP, rvalP );
	    break;


    	case	PN_EXPR:			/* Expression placeholder */
	    status = eval_tree( treeP->pt_leftP, rvalP );
	    break;

	case	PN_OPR:			/* Operator */
	    status = eval_operator( treeP, rvalP );
	    break;


	case	PN_CAST:
	    status = eval_tree( treeP->pt_rightP, rvalP );
	    if ( status == SCOK )
	    	status = val_coerce( rvalP, treeP->pt_info.od_P, TRUE );
	    break;


	case	PN_DECL:
	    warning( "eval_tree: inappropriate PN_DECL node found." );
	    status = SCERR;
	    break;


	case	PN_TYPEREF:
	    warning( "eval_tree: inappropriate PN_TYPEREF node found." );
	    status = SCERR;
	    break;


	case	PN_TYPEDEF:
	    warning( "eval_tree: inappropriate PN_TYPEDEF node found." );
	    status = SCERR;
	    break;


	case	PN_PTR:
	    warning( "eval_tree: inappropriate PN_PTR node found." );
	    status = SCERR;
	    break;


	case	PN_DIM:
	    warning( "eval_tree: inappropriate PN_DIM node found." );
	    status = SCERR;
	    break;


	case	PN_FUNCDEF:
	    warning( "eval_tree: inappropriate PN_FUNCDEF node found." );
	    status = SCERR;
	    break;


	case	PN_FUNCREF:
	    status = eval_funcref( rvalP, treeP->pt_leftP, treeP->pt_rightP );
	    break;


	case	PN_BIS:			/* Built-in statement */
	    status = (*Bistbl[treeP->pt_data.pv_int].bis_exec)(
	    			treeP, rvalP );
	    break;


#ifdef	SIEVE
	case	PN_SVBIS:		/* A naked sieve statement */
	    status = exec_sieve( treeP, rvalP );
	    break;
#endif	/* SIEVE */


	default:			/* Something bad */
	    warning(
"internal error at eval_tree - invalid node type %d found",
	    		treeP->pt_type );
	    status = SCERR;
	    break;
    }

    return( status );
}
/*

*//* eval_tree_nil(treeP)

	Evaluate an expression tree, return no value

Accepts :

	treeP		Ptr to top node of expression

Returns :

	<value>		SCxx value

Notes :


*/

STSCODE
eval_tree_nil ARGLIST( (treeP) )
    FARG( PTNODE	*treeP )	/* Tree to evaluate */
{
	STSCODE		status;
	VALUE		val;

    /* Evaluate the tree */
    status = eval_tree( treeP, &val );

    if ( status == SCOK )
	val_release( &val );

    /* Return the status */
    return( status );
}
/*

*//* eval_tree_long(treeP, longP)

	Evaluate an expression tree, return long value

Accepts :

	treeP		Ptr to top node of expression
	longP		Where to put the long result

Returns :

	<value>		SCxx value

Notes :


*/

STSCODE
eval_tree_long ARGLIST( (treeP, longP) )
   NFARG( PTNODE	*treeP )	/* Tree to evaluate */
    FARG( long		*longP )	/* Where to put the result */
{
	STSCODE		status;
	VALUE		val;

    /* Evaluate the tree */
    status = eval_tree( treeP, &val );

    if ( status == SCOK ) {
    	/* Force the value to be long */
	status = val_coerce( &val, Tnint4->tn_typeP, TRUE );

	/* Return it and clean it */
	*longP = val.v_data.d_long;
	val_release( &val );
    }
    else
	*longP = 0L;

    /* Return the status */
    return( status );
}
/*

*//* eval_tree_word(treeP, wordP)

	Evaluate an expression tree, return word value

Accepts :

	treeP		Ptr to top node of expression
	wordP		Where to put the word result

Returns :

	<value>		SCxx value

Notes :


*/

STSCODE
eval_tree_word ARGLIST( (treeP, wordP) )
   NFARG( PTNODE	*treeP )	/* Tree to evaluate */
    FARG( WORD		*wordP )	/* Where to put the result */
{
	STSCODE		status;
	VALUE		val;

    /* Evaluate the tree */
    status = eval_tree( treeP, &val );

    if ( status == SCOK ) {
    	/* Force the value to be word */
	status = val_coerce( &val, Tnint2->tn_typeP, TRUE );

	/* Return the word, release the value */
	*wordP = val.v_data.d_word;
	val_release( &val );
    }
    else
	*wordP = 0;

    /* Return the status */
    return( status );
}





/***********************************************************************
 *                                                                     *
 *                       Simple support routines                       *
 *                                                                     *
 ***********************************************************************/


/* Following are some simple support routines that are used indirectly by
   generic eval functions, generally as pointed to by an operator table
   entry.
*/



/*
   Character test functions, for when a function (not a macro) is
   required, e.g. for calling indirectly.
*/

BOOL evalc_iscontrol( ch ) UBYTE ch;	{ return ( iscntrl( ch ) != 0 ); }
BOOL evalc_isdigit( ch ) UBYTE ch;	{ return ( isdigit( ch ) != 0 ); }
BOOL evalc_ishighbit( ch ) UBYTE ch;	{ return ( ( ch & 0x80 ) != 0 ); }
BOOL evalc_islower( ch ) UBYTE ch;	{ return ( islower( ch ) != 0 ); }
BOOL evalc_isupper( ch ) UBYTE ch;	{ return ( isupper( ch ) != 0 ); }



/*
   Floating point arithmentic
*/
double evals_fadd( d1, d2 ) double d1, d2;  { return d1 + d2; }
double evals_fmult( d1, d2 ) double d1, d2; { return d1 * d2; }
double evals_fsub( d1, d2 ) double d1, d2;  { return d1 - d2; }
double evals_fdiv( d1, d2 ) double d1, d2;
{ 
    if ( d2 == 0.0 ) {
    	warning( "floating point division by zero." );
	return( MAXFLOAT );
    }
    else
    	return d1 / d2;
}

/*
  Integer arithmetic
*/

long evals_iadd( l1, l2 ) long l1, l2; { return l1 + l2; }
long evals_ibitand( l1, l2 ) long l1, l2; { return l1 & l2; }
long evals_ibitor( l1, l2 ) long l1, l2; { return l1 | l2; }
long evals_ilshift( l1, l2 ) long l1, l2; { return l1 << l2; }
long evals_imod( l1, l2 ) long l1, l2;  { return l1 % l2; }
long evals_imult( l1, l2 ) long l1, l2; { return l1 * l2; }
long evals_irshift( l1, l2 ) long l1, l2; { return l1 >> l2; }
long evals_isub( l1, l2 ) long l1, l2; { return l1 - l2; }
long evals_idiv( l1, l2 ) long l1, l2;
{ 
    if ( l2 == 0 ) {
    	warning( "integer division by zero." );
	return( 0x7fffffff );
    }
    else
    	return l1 / l2;
}



/* 
  Floating point relational
*/

double evals_feq( d1, d2 ) double d1, d2; 
   { return (double)( d1 == d2 ); }

double evals_fge( d1, d2 ) double d1, d2; 
   { return (double)( d1 >= d2 ); }

double evals_fgt( d1, d2 ) double d1, d2; 
   { return (double)( d1 > d2 ); }

double evals_fle( d1, d2 ) double d1, d2; 
   { return (double)( d1 <= d2 ); }

double evals_flt( d1, d2 ) double d1, d2; 
   { return (double)( d1 < d2 ); }

double evals_fne( d1, d2 ) double d1, d2; 
   { return (double)( d1 != d2 ); }


/*
  Integer relational
*/

long evals_ieq( l1, l2) long l1, l2; { return l1 == l2; }
long evals_ige( l1, l2) long l1, l2; { return l1 >= l2; }
long evals_igt( l1, l2) long l1, l2; { return l1 > l2; }
long evals_ile( l1, l2) long l1, l2; { return l1 <= l2; }
long evals_ilt( l1, l2) long l1, l2; { return l1 < l2; }
long evals_ine( l1, l2) long l1, l2; { return l1 != l2; }


/*
  String relational
*/

STSCODE scmp_eq( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result == 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_eq( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result == 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmp_ge( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result <= 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_ge( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result <= 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmp_gt( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result < 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_gt( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result < 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmp_le( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result >= 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_le( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result >= 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmp_lt( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result > 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_lt( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result > 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmp_ne( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				0 );
    val_set_int( rvalP, 4, ( result != 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE scmpi_ne( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_STRCMP,
				MF_IGNCASE );
    val_set_int( rvalP, 4, ( result != 0 ) ? TRUE : FALSE );
    return status;
}

STSCODE sregex_eq( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_REGEX,
				0 );
    val_set_int( rvalP, 4, result );
    return status;
}

STSCODE sregex_ne( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_REGEX,
				0 );
    val_set_int( rvalP, 4, !result );
    return status;
}

STSCODE swild_eq( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_WILD,
				0 );
    val_set_int( rvalP, 4, result );
    return status;
}

STSCODE swildi_eq( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_WILD,
				MF_IGNCASE );
    val_set_int( rvalP, 4, result );
    return status;
}

STSCODE swild_ne( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_WILD,
				0 );
    val_set_int( rvalP, 4, !result );
    return status;
}

STSCODE swildi_ne( rvalP, val1P, val2P )
    VALUE *rvalP, *val1P, *val2P;
{
    STSCODE status;
    int  result;
    status = val_string_cmp( &result, val2P, val1P, SM_WILD,
				MF_IGNCASE );
    val_set_int( rvalP, 4, !result );
    return status;
}




/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* evs_incdec( opP, rvalP, valP, delta, preF )

	Common code for increment/decrement functions

Accepts :

	opP		Ptr to operator table entry
	rvalP		Ptr to result value
	valP		Ptr to value
	delta		How much to increment/decement by
	preF		TRUE if pre-increment, FALSE if post

Returns :

	<value>		SCxx status code

Notes :

	preF determines what value to return-- the value before
	the operation or the value after.

*/


static	STSCODE
evs_incdec ARGLIST( ( opP, rvalP, valP, delta, preF) )
   NFARG( OPINFO	*opP )		/* Ptr to operator table entry */
   NFARG( VALUE		*rvalP )	/* Ptr to result value */
   NFARG( VALUE		*valP )		/* Ptr to value */
   NFARG( int		delta )		/* Increment to use */
    FARG( BOOL		preF )		/* TRUE if preincrement */
{
	STSCODE		status;		/* Status code */
	VALUE		incvalue;	/* Increment value */
	VALUE		pvalue;		/* Preserved value */
	VALUE		rvalue;		/* Result value */

    /* If not a pre-operation, return the current value. */
    if ( !preF ) {
    	memcpy( rvalP, valP, sizeof( VALUE ) );
	val_access( rvalP );

	/* Could rvalue here, but why not keep the lvalue */
    }

    /* Construct a constant integer term for the increment */
    incvalue.v_flags = VF_CONST|VF_RVALUE;
    incvalue.v_typeP = Tnint1->tn_typeP;
    incvalue.v_data.d_byte = delta;

    /* Save away the destination to preserve its lvalue */
    memcpy( &pvalue, valP, sizeof( VALUE ) );

    /* Do the addition */
    status = eval_plus( &rvalue, opP, valP, &incvalue );

    /* Do the assignment to update the value */
    if ( status == SCOK ) {
	status = val_assign( &pvalue, &rvalue );

	/* If it's a pre-op, return the resulting value. */
	if ( preF && ( status == SCOK ) ) {
    	    memcpy( rvalP, &rvalue, sizeof( VALUE ) );
	    val_access( rvalP );

	    /* Could rvalue here too, but why not keep the lvalue. */
	}

	val_release( &rvalue );
    }

    if ( status != SCOK ) {
	if ( !preF ) {
	    /* Release the access we took when we came in */
	    val_release( rvalP );
	}

	/* Clear returned value for good measure */
	val_nill( rvalP );
    }

    return( status );
}
