/* mfl-exec.c -- Tree execution routines

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

There are multiple modules that "run" the parse tree.  This
one contains execution functions which are, mainly, higher
level control processing sorts of things.  This is distinct
from something like evaluation routines which would be more about
producing results from parts of (or terms in) the parse tree.

The lines that separate these "run" modules are not always distinct.


*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-exec.h"
#include "mfl-eval.h"
#include "mfl-type.h"
#include "mfl-parse.h"
#include "mfl-func.h"
#include "mfl-sieve.h"

#include "errwarn.h"

/* Local definitions */



/* External data referenced */

extern	SYMTAB	*STglobal;		/* Global variable names */
extern	SYMTAB	*STtname;		/* Type names */
extern	TNSYM	*Tnint4;		/* Ptr to int*4 value */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */



/*

*//* exec__break( nodeP, rvalP )

	Execute "break" statement node

Accepts :

	nodeP		Ptr to "break" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	returns SCBREAK to exit the current loop level.

*/

STSCODE
exec__break ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "return" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    val_set_int( rvalP, 4, 0 );

    return ( SCBREAK );
}
/*

*//* exec__c( nodeP, rvalP )

	Execute "C" statement node

Accepts :

	nodeP		Ptr to PN_BIS/C parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec__c ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "c" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Just evaluate the tree */
    return ( eval_tree( nodeP->pt_leftP, rvalP ) );
}
/*

*//* exec__continue( nodeP, rvalP )

	Execute "continue" statement node

Accepts :

	nodeP		Ptr to "continue" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	returns SCBREAK to exit the current loop level.

*/

STSCODE
exec__continue ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "return" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    val_set_int( rvalP, 4, 0 );

    return ( SCCONTINUE );
}
/*

*//* exec__do( nodeP, rvalP )

	Execute a "do" statement

Accepts :

	nodeP		Ptr to "do" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec__do ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "do" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	WORD		tval;		/* Testvalue */

    /* Loop until the do condition fails or something bad happens */
    for( status = SCOK; status == SCOK; ) {

	/* Perform the loop body */
	status = eval_tree( nodeP->pt_rightP, rvalP );

	/* An eval result of "continue" becomes OK now */
	if ( status == SCCONTINUE )
	    status = SCOK;

	if ( status == SCOK ) {
	    /* Test the do condition */
	    status = eval_tree_word( nodeP->pt_leftP, &tval );

	    if ( !tval )
	    	break;

	    /* Release the value that we obtained; only the final value
	       is propagated back up.
	    */
	    val_release( rvalP );
	}
    }

    /* If we broke out via "break" change status to OK */
    if ( status == SCBREAK )
	status = SCOK;

    return( status );
}
/*

*//* exec__for( nodeP, rvalP )

	Execute a "for" statement

Accepts :

	nodeP		Ptr to "for" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec__for ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "for" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	BOOL		loopedF;	/* If we did any loops. */
	PTNODE		*ttreeP;	/* Loop test expr */
	PTNODE		*ptreeP;	/* post-loop expr */
	PTNODE		*btreeP;	/* Loop body expr */
	WORD		tval;		/* Testvalue */

    loopedF = FALSE;

    /* Get the loop components */
    if ( ( nodeP->pt_leftP == NULL ) || ( nodeP->pt_rightP == NULL ) ||
         ( nodeP->pt_leftP->pt_type != PN_EXPR ) ||
	 ( nodeP->pt_rightP->pt_type != PN_EXPR ) ) {
	warning( "internal error in exec__for!" );
	status = SCNOTDONE;
    }
    else {
        /* Get the components */
	if ( ( ttreeP = nodeP->pt_leftP->pt_rightP ) == NULL )
	    tval = TRUE;
	ptreeP = nodeP->pt_rightP->pt_leftP;
	btreeP = nodeP->pt_rightP->pt_rightP;

    	/* Perform the intialization clause */
	status = eval_tree_nil( nodeP->pt_leftP->pt_leftP );

	/* Loop until the test fails or until error */

	for( ; status == SCOK; loopedF = TRUE ) {
	    /* Do loop test */
	    if ( ttreeP != NULL )
	    	status = eval_tree_word( ttreeP, &tval );
	    if ( ( status != SCOK ) || !tval )
	    	break;

	    /* If we went through this loop before, release the prior
	       result.  Only the last one is returned.
	    */
	    if ( loopedF )
		val_release( rvalP );

	    /* Do loop body, storing in result term */
	    status = eval_tree( btreeP, rvalP );

	    /* If a "continue" status bubbled up, change it to OK */
	    if ( status == SCCONTINUE )
		status = SCOK;

	    if ( status == SCOK )
	    	/* Do post-loop */
		status = eval_tree_nil( ptreeP );
	}
    }

    /* If we never executed a loop, init the value to zero. */
    if ( !loopedF )
	val_set_int( rvalP, 4, 0 );

    /* If we exited because of break, it's OK */
    if ( status == SCBREAK )
	status = SCOK;

    return( status );
}
/*

*//* exec__if( nodeP, rvalP )

	Execute an "if" statement

Accepts :

	nodeP		Ptr to "if" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec__if ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "if" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	WORD		tval;		/* Testvalue */

    /* Apply the condition */
    status = eval_tree_word( nodeP->pt_leftP, &tval );
    
    /* Proceed according to truthness */
    if ( status == SCOK ) {
        /* Focus on the if body node */
	nodeP = nodeP->pt_rightP;

    	if ( tval )		    	/* True */
	    status = eval_tree( nodeP->pt_leftP, rvalP );
	else				/* False */
	    status = eval_tree( nodeP->pt_rightP, rvalP );
    }

    return( status );
}
/*

*//* exec__pv$( nodeP, rvalP )

	Execute a "pv$" statement

Accepts :

	nodeP		Ptr to "pv$" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	This is a hackish statement that prints its values.

*/

STSCODE
exec__pv_ ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "pv$" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	int		i;
	char		*fmtP;		/* Format string */
	UBYTE		*strP;		/* Scratch string */
	REFSTR		*rsP;
	BSTR		*bstrP;

    /* Evaluate the thing to print */
    status = eval_tree( nodeP->pt_leftP, rvalP );
    val_rvalue( rvalP );

    if ( status == SCOK ) {
        /* If no format given, give annotated output */
	if ( nodeP->pt_rightP == NULL ) {
	    val_print( stdout, rvalP, 0, "" );
	}
	else {
	    /* Format given -- use it */
	    fmtP = (char *)nodeP->pt_rightP->pt_info.od_P;
	    switch( rvalP->v_typeP->ty_class ) {
	   	case TC_PTR:  		/* Ptrs */
		case TC_STRUCT:
		case TC_UNION:
TODO		    /* Need to resolve the pointer here and use it */
		    printf( fmtP, rvalP->v_data.d_ptr.p_ptr.p_genP );
		    break;


		case TC_STRING:
		    rsP = (REFSTR *)valptr_resolve( &rvalP->v_data.d_ptr );
		    if ( rsP == NULL )
			strP = NULL;
		    else {
			i = refstr_len( rsP );
TODO			/* May want to sanity check this length */
			strP = tmpstr( i + 1 );
			if ( strP != NULL ) {
			    bstrP = refstr_bstr( rsP );
			    strncpy( (char *)strP,
				     (char *)&bstrP->bs_bP[refstr_startx( rsP )],
				     i );
			    strP[i] = NUL;
			}
		    }
		    if ( strP == NULL )
			strP = (UBYTE *)"<>";
		    printf( fmtP, strP );
		    break;

			
		case TC_FLOAT:
		    if ( rvalP->v_typeP->ty_size == 4 )
		    	printf( fmtP, rvalP->v_data.d_float );
		    else
		    	printf( fmtP, rvalP->v_data.d_double );
		    break;

		case TC_INT:
		case TC_UINT:
		case TC_ENUM:
		    switch( rvalP->v_typeP->ty_size ) {
		    	case 1:
			    printf( fmtP, rvalP->v_data.d_byte );
			    break;

		    	case 2:
			    printf( fmtP, rvalP->v_data.d_word );
			    break;

		    	case 4:
			    printf( fmtP, rvalP->v_data.d_long );
			    break;
		    }
		    break;

		default:
		    warning( "error in pv$: type=%d",
		    		    rvalP->v_typeP->ty_class );
		    break;
	    }
	}
    }

    return( status );
}
/*

*//* exec__return( nodeP, rvalP )

	Execute "return" function node

Accepts :

	nodeP		Ptr to "return" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	returns SCDONE to exit the current function level.

*/

STSCODE
exec__return ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "return" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* If there's an expression to return, evaluate it */
    if ( nodeP->pt_leftP != NULL ) {
	status = eval_tree( nodeP->pt_leftP, rvalP );
	if ( status == SCOK )
	    status = SCRETURN;
    }
    else {
	val_set_int( rvalP, 4, 0 );
	status = SCRETURN;
    }

    return ( status );
}
/*

*//* exec__sizeof( nodeP, rvalP )

	Execute "sizeof" function node

Accepts :

	nodeP		Ptr to "sizeof" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	While this returns a value and looks like it should be
	an eval_ function, it's parsed as a built-in function and
	thus is treated as an exec__ function.

*/

STSCODE
exec__sizeof ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "sizeof" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	long		lval;
	VALUE		sval;

    /* Argument may be either a type reference or an expression. */
    if ( nodeP->pt_leftP->pt_type == PN_TYPEREF ) {
        lval = ((TYPE *)nodeP->pt_leftP->pt_info.od_P)->ty_size;
	status = SCOK;
    }
    else {
	/* Evaluate the term to get the size of; either an lvalue or
	   an evaluated term (with its own size).

	   Note that this interpreter has no way of evaluating the
	   type of an expression without also evaluating the expression.
	   So an expression in sizeof can easily have side-effects.
	*/
	status = eval_tree( nodeP->pt_leftP, &sval );
	if ( status == SCOK ) {
	    if ( ( sval.v_flags & VF_LVALUE ) != 0 ) {
		lval = sval.v_typeP->ty_size;
	    }
	    else {
		/* This is the same thing as the lvalue case, but I want
		   to keep the code paths distinct as separate things.
		*/
		lval = sval.v_typeP->ty_size;
	    }
	    val_release( &sval );
	}
    }

    val_set_int( rvalP, 4, lval );

    return( status );
}
/*

*//* exec__while( nodeP, rvalP )

	Execute a "while" statement

Accepts :

	nodeP		Ptr to "while" parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec__while ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "while" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	BOOL		loopedF;	/* If any loops were executed */
	WORD		tval;		/* Testvalue */

    /* Loop until the while condition fails or something bad happens */
    for( loopedF = FALSE, status = SCOK; status == SCOK; loopedF = TRUE ) {
    	/* Test the while condition */
	status = eval_tree_word( nodeP->pt_leftP, &tval );

	if ( status == SCOK ) {
	    if ( !tval )
	    	break;

	    /* If this isn't the first time through the loop, release the
	       previous value/result
	    */
	    if ( loopedF )
		val_release( rvalP );

	    /* Perform the loop body */
	    status = eval_tree( nodeP->pt_rightP, rvalP );

	    /* A "continue" changes into "OK" */
	    if ( status == SCCONTINUE )
		status = SCOK;
	}
    }

    /* If we never executed a loop, init the value to zero. */
    if ( !loopedF )
	val_set_int( rvalP, 4, 0 );

    /* If we finished because of break, it's OK */
    if ( status == SCBREAK )
	status = SCOK;

    return( status );
}
/*

*//* exec_decl( treeP )

	Execute a declaration

Accepts :

	treeP		Ptr to the PN_DECL that anchors this declaration

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec_decl ARGLIST( (treeP) )
    FARG( PTNODE	*treeP )	/* Address of type node */
{
	STSCODE		status;		/* Status value */
	TYPE		*typeP;		/* Type definition */
	PTNODE		*trnodeP;	/* Addr of typeref node */
	VSYM		*vsymP;		/* Gbl variable symbol entry */

    /* Root node shall point leftwards to a left-linked list of
         PN_SYM nodes.

       Each SYM node's pt_info generic pointer points to the variable
       name.

       On the right side of the PN_SYM node is a PN_TYPEREF node.  This
	 typeref node has its generic (info) pointer pointing to a TYPE;
	 its right side pointing to any initialization data;
	 its left side should be NULL.
    */
    for( status = SCOK, treeP = treeP->pt_leftP;
           ( treeP != NULL ) && ( status == SCOK );
	      treeP = treeP->pt_leftP ) {
	if ( treeP->pt_type != PN_SYM ) {
	    warning( "exec_decl: non- PN_SYM node found." );
	    status = SCERR;
	    break;
	}

	/* Get type reference node */
	if ( ( ( trnodeP = treeP->pt_rightP ) == NULL ) ||
	       ( trnodeP->pt_type != PN_TYPEREF ) ) {
	    warning( "exec_decl: missing PN_TYPEREF node." );
	    status = SCERR;
	    break;
	}

	/* New type */
	typeP = (TYPE *)trnodeP->pt_info.od_P;

	/* See if this global name already exists. */
	vsymP = (VSYM *)symtab_lookup( STglobal, treeP->pt_info.od_P );
	if ( vsymP != NULL ) {

#ifdef	MFL_ALLOW_REDECLARATIONS
	    warning( "variable \"%s\" being redeclared.",
		       (char *)treeP->pt_info.od_P );
	    symtab_remove( STglobal, treeP->pt_info.od_P );
#else
	    warning( "variable \"%s\" already exists.",
		       (char *)treeP->pt_info.od_P );
	    status = SCNOTDONE;
	    break;
#endif	/* MFL_ALLOW_REDECLARATIONS */
	}

	/* Enter the symbol */
	vsymP = (VSYM *)symtab_enter( STglobal, treeP->pt_info.od_P,
					    sizeof( VSYM ) );

	if ( vsymP == NULL ) {
	    warning( "can't create entry for variable %s", 
					(char *)treeP->pt_info.od_P );
	    status = SCNOTDONE;
	    break;
	}
	else {
	    /* Instantiate the value */
	    status = val_init( &vsymP->vs_value, typeP );
	    if ( status != SCOK )
		break;
	}

	/* Check for function instantiation here.  This is kind of
	   a special-case hack test blah.
	*/
	if ( ( status == SCOK ) && ( typeP->ty_class == TC_FUNCT ) ) {
	    status = exec_decl_value_func( vsymP, treeP, trnodeP );
	    continue;
	}

	/* Do any initial value setup (i.e., initialization) here */
	if ( status == SCOK )
	    status = exec_decl_value( vsymP, &vsymP->vs_value,
				      trnodeP->pt_rightP );
    }

    return ( status );
}
/*

*//* exec_decl_value( vsymP, itreeP, topF )

	Setup value for declaration

Accepts :

	vsymP		Ptr to symbol table entry for variable
	valP		Ptr to value to set up
	itreeP		Ptr to initialization tree if any

Returns :

	<value>		Status code

Notes :

	This routine may recurse to follow initialization data for
	subvalues (array members, things pointed to, etc).  Only the
	top invocation level will receive the vsymP pointer.

	Note that the valP passed may either be the value to copy
	into or value that's used as the target as an assignment
	(i.e., a pointer to another value).  The latter case will
	be an lvalue.

*/

STSCODE
exec_decl_value ARGLIST( (vsymP, valP, itreeP) )
   NFARG( VSYM		*vsymP )	/* Ptr to symbol table entry */
   NFARG( VALUE		*valP )		/* Ptr to target value thing */
    FARG( PTNODE	*itreeP )	/* Ptr to any subtree */
{
	STSCODE		status;
	TYPE		*typeP;
	VALUE		refval;		/* A reference for assignment */
	VALUE		tmpval;

    status = SCOK;

    /* Proceed by type */
    typeP = valP->v_typeP;
    switch ( typeP->ty_class ) {
	case  TC_VOID:
	    if ( itreeP != NULL )
		warning( "may not initialize a void type." );
		status = SCNOTDONE;
		break;

	case  TC_UINT:			/* Scalars */
	case  TC_INT:
	case  TC_ENUM:
	case  TC_FLOAT:
	case  TC_STRING:
	case  TC_PTR:
	    if ( itreeP != NULL ) {
		/* Just a simple check for sublist */
		if ( itreeP->pt_type == PN_LSEQ ) {
		    warning( "bad initializer data." );
		    status = SCNOTDONE;
		}
		else {
		    /* For a scalar, we just do an evaluation and
		       then either a value copy or an assignment
		       depending on whether the destination is an
		       lvalue (meaning it's assignable, not the
		       target value itself).

		       This could be more efficient if done on a
		       type-by-type basis.  In fact it was that
		       way at first.  However this is more generic
		       even if slower.
		    */
		    status = eval_tree( itreeP, &tmpval );
		    if ( status == SCOK ) {
			/* Do the copy or assignment */
			if ( ( valP->v_flags & VF_LVALUE ) == 0 ) {
			    /* This is a direct value.  Make a
			       reference to it and then do the assign.
			    */
			    refval.v_flags = VF_LVALUE;
			    refval.v_typeP = valP->v_typeP;

			    if ( vsymP != NULL ) {
				/* Direct reference to value */
				refval.v_data.d_ptr.p_byteX = PO_SIMPLE;
				refval.v_data.d_ptr.p_ptr.p_genP =
					&valP->v_data;
			    }
			    else {
				/* Somewhere else.  Just use the ptr.
				   No need to fudge with access/release
				   since this is private scope.
				*/
				refval.v_data.d_ptr.p_ptr =
				    valP->v_data.d_ptr.p_ptr;
			    }

			    /* This assign will now reference valP */
			    status = val_assign( &refval, &tmpval );
			}
			else
			    status = val_assign( valP, &tmpval );
		    }
		}
	    }
	    break;


	case  TC_DIM:
	    status = exec_decl_value_array( vsymP, valP, itreeP );
	    break;


	case  TC_STRUCT:
	    status = exec_decl_value_struct( vsymP, valP, itreeP );
	    break;


	case  TC_UNION:
	    if ( itreeP != NULL ) {
		warning( "don't yet know how to initialize this type.  Sorry." );
		status = SCNOTDONE;
	    }
	    break;


	case  TC_FUNCT:
TODO	    warning( "unimplemented function prototype/reference." );
	    break;


	default:
	    warning( "exec_decl_value: unknown type %d", typeP->ty_class );
	    status = SCERR;
	    break;
    }

    return( status );
}
/*

*//* exec_decl_value_array( vsymP, valP, itreeP )

	Setup value for declaration of an array

Accepts :

	vsymP		Ptr to symbol table entry for variable
	valP		Ptr to value to set up
	itreeP		Ptr to initialization tree if any

Returns :

	<value>		Status code

Notes :

	Called as part of the value initialization stuff for an array
	type.  Initial value, if present, must point to an
	initialization list (that is, a left-wise sequence of LSEQ
	nodes).

*/

STSCODE
exec_decl_value_array ARGLIST( (vsymP, valP, itreeP) )
   NFARG( VSYM		*vsymP )	/* Ptr to symbol table entry */
   NFARG( VALUE		*valP )		/* Ptr to target value thing */
    FARG( PTNODE	*itreeP )	/* Ptr to any subtree */
{
	STSCODE		status;
	int		memX;		/* Array member index */
	UBYTE		*dataP;		/* Ptr to base data */
	VALUE		memval;		/* Member value reference */

    if ( itreeP == NULL )
	return ( SCOK );		/* No initial data */

    status = SCOK;

    /* Get array address */
    dataP = valptr_resolve( &valP->v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "can't initialize array: bad reference." );
	return ( SCNOTDONE );
    }

    /* Walk the array members and the initialization list */
    for ( memX = 0;
	   (status == SCOK ) && ( itreeP != NULL) &&
	   (memX < valP->v_typeP->ty_dim);
	      ++memX, itreeP = itreeP->pt_leftP ) {

	/* Make sure initialization node is a list sequencer */
	if ( itreeP->pt_type != PN_LSEQ ) {
	    warning( "bad array initializer tree." );
	    status = SCERR;		/* Hard error here. */
	    break;
	}

	/* Construct a value reference */
	memval.v_flags = VF_LVALUE;
	memval.v_typeP = (TYPE *)valP->v_typeP->ty_nextP;
	memval.v_data.d_ptr.p_ptr.p_genP =
			(void *)(dataP + memX * memval.v_typeP->ty_size);
	memval.v_data.d_ptr.p_byteX = PO_SIMPLE;

	/* Recurse to initialize the member */
	status = exec_decl_value( (VSYM *)NULL, &memval, itreeP->pt_rightP );
    }

    /* All done here */
    if ( status == SCOK ) {
	if ( memX < valP->v_typeP->ty_dim ) {
TODO	    /* Do we really need this warning? */
	    warning(
   "exec_decl_value_array: too few initializer values for array." );
	}
	else if ( itreeP != NULL ) {
TODO	    /* Do we really need this warning? */
	    warning(
   "exec_decl_value_array: too many initializer values for array." );
	}
    }

    return ( status );
}
/*

*//* exec_decl_value_func( vsymP, treeP, trnodeP )

	Setup initial function definition

Accepts :

	vsymP		Ptr to the symbol table entry for the function name
	treeP		Ptr to the PN_DECL node for this declaration
	trnodeP		Ptr to the PN_TYPEREF node for this declaration

Returns :

	<value>		status code


Notes :

	Picked up whole and removed from exec_decl as a special case.

*/

STSCODE
exec_decl_value_func ARGLIST( (vsymP, treeP, trnodeP) )
   NFARG( VSYM		*vsymP )	/* Ptr to symbol table entry */
   NFARG( PTNODE	*treeP )	/* Ptr to PN_DECL node */
    FARG( PTNODE	*trnodeP )	/* Ptr to the PN_TYPEREF node */
{
	STSCODE		status;

    if ( func_register( (char *)treeP->pt_info.od_P,
		         vsymP, FUNC_MFL,
		         vsymP->vs_value.v_typeP,
			 (FUNCFARGS *)trnodeP->pt_xinfo.od_P,
		          trnodeP->pt_rightP ) == NULL ) {
	status = SCNOTDONE;

	/* tree freeing will remove the body tree and args */
    }
    else {
	/* Remove stuff from the tree so it doesn't get
	   deleted out from under the function.
	*/
	trnodeP->pt_rightP = NULL;	/* body tree */
	trnodeP->pt_xinfo.od_P = NULL;	/* Function arguments. */

	status = SCOK;
    }

    return ( status );
}
/*

*//* exec_decl_value_struct( vsymP, valP, itreeP )

	Setup value for declaration of a struct

Accepts :

	vsymP		Ptr to symbol table entry for variable
	valP		Ptr to value to set up
	itreeP		Ptr to initialization tree if any

Returns :

	<value>		Status code

Notes :

	Called as part of the value initialization stuff for a struct
	type.  Initial value, if present, must point to an
	initialization list (that is, a left-wise sequence of LSEQ
	nodes).

*/

STSCODE
exec_decl_value_struct ARGLIST( (vsymP, valP, itreeP) )
   NFARG( VSYM		*vsymP )	/* Ptr to symbol table entry */
   NFARG( VALUE		*valP )		/* Ptr to target value thing */
    FARG( PTNODE	*itreeP )	/* Ptr to any subtree */
{
	STSCODE		status;
	int		memX;		/* Member index */
	UBYTE		*dataP;		/* Ptr to base data */
	TYPEAG		*agP;		/* Ptr to the aggregate */
	AGGMEM		*agmemP;	/* Ptr to each member */
	VALUE		memval;		/* Member value reference */

    if ( itreeP == NULL )
	return ( SCOK );		/* No initial data */

    status = SCOK;

    /* Get struct address */
    dataP = valptr_resolve( &valP->v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "can't initialize struct: bad reference." );
	return ( SCNOTDONE );
    }

    /* Walk the members and the initialization list */
    agP = (TYPEAG *)valP->v_typeP->ty_nextP;
    for ( memX = 0;
	   (status == SCOK ) && (memX < agP->ag_memC) && ( itreeP != NULL);
	      ++memX, itreeP = itreeP->pt_leftP ) {

	/* Make sure initialization node is a list sequencer */
	if ( itreeP->pt_type != PN_LSEQ ) {
	    warning( "bad struct initializer tree." );
	    status = SCERR;		/* Hard error here. */
	    break;
	}

	/* Construct a value reference */
	agmemP = &agP->ag_memP[memX];

	memval.v_flags = VF_LVALUE;
	memval.v_typeP = agmemP->am_typeP;
	memval.v_data.d_ptr.p_ptr.p_genP = (void *)(dataP + agmemP->am_offs);
	memval.v_data.d_ptr.p_byteX = PO_SIMPLE;

	/* Recurse to initialize the member */
	status = exec_decl_value( (VSYM *)NULL, &memval, itreeP->pt_rightP );
    }

    /* All done here */
    if ( status == SCOK ) {
	if ( memX < agP->ag_memC ) {
TODO	    /* Do we really need this warning? */
	    warning(
   "exec_decl_value_struct: too few initializer values for struct." );
	}
	else if ( itreeP != NULL ) {
TODO	    /* Do we really need this warning? */
	    warning(
   "exec_decl_value_struct: too many initializer values for struct." );
	}
    }

    return ( status );
}
/*

*//* exec_tree( treeP, printF )

	Execute a parse tree

Accepts :

	treeP		Ptr to top node of tree
	printF		Whether to print the value afterward

Returns :

	<value>		SCxx status code

Notes :

	Deletes the tree after executing it

*/

STSCODE
exec_tree ARGLIST( (treeP, printF) )
   NFARG( PTNODE	*treeP )	/* Address of tree node */
    FARG( BOOL		printF )	/* Whether to print the value */
{
	STSCODE		status;		/* Status code */
	VALUE		val;		/* Scratch place for value */

    status = exec_treeval( treeP, &val, printF );
    if ( treeP != NULL ) {
	/* Delete the tree from memory */
	parsetree_free( treeP );
    }
    if ( status == SCOK )
	val_finish( &val );

    return( status );
}
/*

*//* exec_treeval( treeP, rvalP, printF )

	Execute a parse tree, also get value, maybe print result

Accepts :

	treeP		Ptr to "if" parse tree node
	rvalP		Where to put the result
	printF		TRUE to print the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec_treeval ARGLIST( (treeP, rvalP, printF ) )
   NFARG( PTNODE	*treeP )	/* Address of top node */
   NFARG( VALUE		*rvalP )	/* Where to put the result */
    FARG( BOOL		printF )	/* Whether to print the value */
{
	STSCODE		status;		/* Status code */

    if ( treeP != NULL ) {
	/* The root node dictates the interpretation of the tree. */
	switch( treeP->pt_type ) {
    	    case	PN_TYPEDEF:	/* new type definition */
		status = exec_typedef( treeP );
		val_set_int( rvalP, 4, status );
		break;


	    case	PN_DECL:	/* Variable declaration */
		status = exec_decl( treeP );
		val_set_int( rvalP, 4, status );
		break;


	    default:
		status = eval_tree( treeP, rvalP );
		if ( ( status == SCOK ) && printF ) {
		    val_print( stdout, rvalP, 0, "" );
		}
		break;
	}
    }
    else {
	/* No tree */
	status = SCOK;
	val_set_int( rvalP, 4, 0 );
    }

    return( status );
}
/*

*//* exec_typedef( treeP )

	Execute an typedef tree

Accepts :

	treeP		Ptr to parse tree node at top of typedef

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec_typedef ARGLIST( (treeP) )
    FARG( PTNODE	*treeP )	/* Address of typedef tree node */
{
	STSCODE		status;		/* Status value */
	TNSYM		*tnsymP;	/* Typename symbol ptr */
	PTNODE		*trnodeP;	/* Addr of typeref node */
	TYPE		*typeP;		/* Type definition */
	TYPEAG		*agP;		/* Aggregate ptr */

    /* The root node points left to a declaration tree (for
       declaring variables that represent a new type).
    */
    treeP = treeP->pt_leftP;
    if ( treeP->pt_type != PN_DECL ) {
    	warning( "internal error in exec_typedef." );
	status = SCERR;
    }
    else
    	status = SCOK;

    /* The PN_DECL node chains leftwise to PN_SYM/PN_TYPEREF pairs. */
    for( treeP = treeP->pt_leftP;
    		( treeP != NULL ) && ( status == SCOK );
    		treeP = treeP->pt_leftP ) {

	if ( treeP->pt_type != PN_SYM ) {
	    warning( "exec_typedef: non- PN_SYM node found." );
	    status = SCERR;
	    break;
	}

	/* Get type reference node */
	if ( ( ( trnodeP = treeP->pt_rightP ) == NULL ) ||
	       ( trnodeP->pt_type != PN_TYPEREF ) ) {
	    warning( "exec_typedef: missing PN_TYPEREF node." );
	    status = SCERR;
	    break;
	}

	/* See if type is already defined */
	tnsymP = (TNSYM *)symtab_lookup( STtname, treeP->pt_info.od_P );
	if ( tnsymP != NULL )
	    warning( "type \"%s\" is already defined -- not replaced.",
				treeP->pt_info.od_P );
	else {
	    tnsymP = (TNSYM *)symtab_enter( STtname, treeP->pt_info.od_P,
						   sizeof(TNSYM) );

	    /* Get the type ptr, which has already been declared. */
	    tnsymP->tn_typeP = typeP = (TYPE *)trnodeP->pt_info.od_P;
	    ++typeP->ty_refC;

	    /* If it's an aggregate with no name, give it a name. */
	    if ( ( typeP->ty_flags & TF_AGGR ) != 0 ) {
		agP = (TYPEAG *)typeP->ty_nextP;
		if ( agP->ag_tnsymP == NULL )
		    agP->ag_tnsymP = tnsymP;
	    }
	}
    }

    return( status );
}
