/* mfl-func.c -- MFL functions

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains code to deal with MFL functions (built-in and otherwise).

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>
#include "dnsbl.h"
#include "data.h"

#include "mvmf.h"
#include "controls.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-func.h"
#include "mfl-exec.h"
#include "mfl-eval.h"
#include "mfl-type.h"
#include "mfl-parse.h"
#include "mfl-misc.h"

#include "mvmfmisc.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */

extern	int	Mfl_nestC;		/* Nesting level */
extern	BOOL	Mfl_traceF;		/* Whether tracing */
extern	SYMTAB	*STglobal;		/* Global variables */
extern	SYMTAB	*STtname;		/* Type names symbol table */
extern	TNSYM	*Tnstring;
extern	TYPSYM	*TyfuncP;		/* Type symbol for function */
extern	TYPSYM	*TystrpP;		/* (string *) */



/* External routines used */

extern	char	*getenv();

/* Local data publicly available */

	FUNCFARGS *FargsP = NULL;	/* Any function args in scope */


/* Local routines and forward references */


static	BOOL	funcfargsdel();		/* Delete FARGS info */

/* Private data */

static	VALUE	*As_P;			/* Argument stack for MFL functions */
static	int	As_L;			/* Argument stack size */
static	int	Asf_argC;		/* Size of current argument frame */
static	int	Asf_localC;		/* Extra stack elements in use
					   in the current argument frame */
static	int	Asf_X;			/* Index to current argument frame */
static	LL_HEAD	Hookpath;		/* Directory path for hooks */

/*

*//* eval_funcref( rvalP, fnnodeP, argtreeP )

	Evaluate a function reference

Accepts :

	rvalP		Where to put the return value
	fnnodeP		Ptr to parse node containing function name
	argtreeP	Ptr to arguments tree

Returns :

	<value>		STSCODE status code

Notes :


*/

STSCODE
eval_funcref ARGLIST( (rvalP, fnnodeP, argtreeP) )
   NFARG( VALUE		*rvalP )	/* Where to put return value */
   NFARG( PTNODE	*fnnodeP )	/* node containing function name */
    FARG( PTNODE	*argtreeP )	/* Ptr to arguments tree */
{
	STSCODE		status;
	FUNC		*funcP;		/* Ptr to function info */
	char		*fnameP;	/* Function name */
	STSCODE		(*rtc)();	/* Ptr to built-in function */

    /* Do some arg checks */
    if ( ( fnnodeP == NULL ) || ( fnnodeP->pt_type != PN_SYM ) ) {
	error( "eval_funcref: non-symbol function name node." );
	return ( SCERR );
    }

    /* Lookup the function */
    fnameP = (char *)fnnodeP->pt_info.od_P;
    funcP = func_lookup( fnameP );
    if ( funcP == NULL ) {
	warning( "eval_funcref: no such function \"%s\"", 
			fnameP );
	return ( SCNOTDONE );
    }

    /* Process based on function type */
    switch ( funcP->fn_type ) {
	case FUNC_NOTDEF:
	    warning( "eval_funcref: function named \"%s\" found, but with NOTDEF type.",
			fnameP );
	    status = SCNOTDONE;
	    break;

	case FUNC_BUILTIN:
	    if ( Mfl_traceF &&
		 ( cntrl_int_get( estr_tmp_str( "mfl_trace_bif" ), 0 ) != 0 ) )
		log_priv( "mfl-trace: calling bif \"%s\"", fnameP );

	    rtc = (STSCODE (*)())funcP->fn_valP;
	    status = (*rtc)( fnameP, rvalP, argtreeP );

	    if ( Mfl_traceF &&
		 ( cntrl_int_get( estr_tmp_str( "mfl_trace_bif" ), 0 ) != 0 ) )
		log_priv( "mfl-trace: returned from bif \"%s\" status %d",
			   fnameP, status );
	    break;

	case FUNC_MFL:
	    if ( Mfl_traceF &&
		 ( cntrl_int_get( estr_tmp_str( "mfl_trace_bif" ), 0 ) != 0 ) )
		log_priv( "mfl-trace: calling func \"%s\"", fnameP );

	    /* Hand off to mfl function invoker */
	    status = mfl_func_exec( rvalP, funcP, argtreeP );

	    if ( Mfl_traceF &&
		 ( cntrl_int_get( estr_tmp_str( "mfl_trace_bif" ), 0 ) != 0 ) )
		log_priv( "mfl-trace: returned from func \"%s\" status %d",
			   fnameP, status );
	    break;

	default:
	    error( "eval_funcref: function \"%s\" has unknown type %d.",
		     fnameP, funcP->fn_type );
	    status = SCERR;
	    break;
    }

    return ( status );
}
/*

*//* fargs_find( fargsP, anameP )

	Find a name in a formal arguments list

Accepts :

	fargsP		Ptr to function arguments block
	anameP		Name to look up

Returns :

	<value>		-1 if not found,
			argument number if found

Notes :


*/

int
fargs_find ARGLIST( (fargsP, anameP) )
   NFARG( FUNCFARGS	*fargsP )	/* The arguments block */
    FARG( char		*anameP )	/* The name to find */
{
	int		argN;		/* Argument number */

    /* Search the table */
    if ( fargsP != NULL )
	for ( argN = 0; argN < fargsP->ffl_argC; ++argN )
	    if (strcmp( anameP, fargsP->ffl_argP[argN].fa_nameP ) == 0 )
		return ( argN );

    return ( -1 );
}
/*

*//* func_lookup( fnameP )

	Lookup a function by name

Accepts :

	fnameP		Name of function

Returns :

	<value>		PTR to FUNC data storage if found,
			NULL if not.

Notes :


*/

FUNC *
func_lookup ARGLIST( (fnameP) )
    FARG( char		*fnameP )	/* Function name */
{
	VSYM		*vsymP;		/* Ptr to variable symtab entry */
	FUNC		*funcP;		/* Ptr to function info */

    /* Look up the variable and make sure everything looks like
       a function. 
    */
    vsymP = (VSYM *)symtab_lookup( STglobal, fnameP );
    if ( ( vsymP == NULL ) ||
         ( vsymP->vs_value.v_typeP == NULL ) ||
	 ( vsymP->vs_value.v_typeP->ty_class != TC_FUNCT ) ||
	 ( vsymP->vs_value.v_data.d_ptr.p_byteX != PO_PRIVATE ) )
	return ( NULL );

    funcP = (FUNC *)vsymP->vs_value.v_data.d_ptr.p_ptr.p_genP;

    return ( funcP );
}
/*

*//* func_register( fnameP, vsymP, ftype, fvtypeP, fargsP, fvalP )

	Register a function 

Accepts :

	fnameP		Name of function
	vsymP		Ptr to function name symbol table entry if known
	ftype		Type of function (FUNC_XXX)
	fvtypeP		Ptr to value type for function name
	fargsP		Any formal arguments definition
	fvalP		Ptr to function value

Returns :

	<value>		PTR to FUNC structure if successful,
			NULL if not.

Notes :

	vsymP is the pointer to the function name's variable symbol
	table entry if known before hand.  If this is null, the
	symbol will be registered by this function.

	fvtypeP is the type that the name should be registered as.
	(e.g., function returning int)
	The first type reference should have class TC_FUNC.

	fargsP is a formal arguments list.  

	fvalP is an opaque pointer at this point, it is the body
	or value of the function (e.g. address of C routine,
	of MFL tree, of handle of some sort).

*/

FUNC *
func_register ARGLIST( (fnameP, vsymP, ftype, fvtypeP, fargsP, fvalP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VSYM		*vsymP )	/* Name's symbol table entry if known */
   NFARG( FUNCTYPE	ftype )		/* Function type */
   NFARG( TYPE		*fvtypeP )	/* Function variable type */
   NFARG( FUNCFARGS	*fargsP )	/* Formal arguments list */
    FARG( void		*fvalP )	/* Function value */
{
	FUNC		*funcP;		/* Ptr to function data */

    /* Register the symbol if not already done by caller */
    if ( vsymP == NULL ) {
	/* Make sure this name isn't already a global variable */
	if ( symtab_lookup( STglobal, fnameP ) != NULL ) {
	    warning( "func_register: duplicate name \"%s\"", fnameP );
	    return ( NULL );
	}

	/* Make new symbol */
	vsymP = (VSYM *)symtab_enter( STglobal, fnameP, sizeof( VSYM ) );
	if ( vsymP == NULL ) {
	    warning(
	"func_register: can't create entry for \"%s\" in STglobal table.",
		    fnameP );
	    return ( NULL );
	}
    }

    /* Make the function information block */
    funcP = (FUNC *)emalloc( "func_register", "function info block",
			    sizeof(FUNC) );
    if ( funcP == NULL )
	return ( NULL );

    /* Fill in the info. */
    funcP->fn_type = ftype;
    funcP->fn_valP = fvalP;
    if ( fargsP != NULL )
	funcP->fn_fargs = *fargsP;
    else {
	funcP->fn_fargs.ffl_argC = 0;
	funcP->fn_fargs.ffl_argL = 0;
	funcP->fn_fargs.ffl_argP = NULL;
    }

    /* Fill in the type (default to function type) */
    if ( fvtypeP == NULL )
	fvtypeP = &(TyfuncP->ts_type);

    /* Instantiate the symbol value */
    val_init( &vsymP->vs_value, fvtypeP );

    /* Other stuff for the symbol and value */
    vsymP->vs_value.v_data.d_ptr.p_byteX = PO_PRIVATE;
    vsymP->vs_value.v_data.d_ptr.p_ptr.p_genP = funcP;

    /* Put the return type in the function definition. */
    if ( fvtypeP->ty_class == TC_FUNCT )
	funcP->fn_rtypeP = fvtypeP->ty_nextP;
    else
	funcP->fn_rtypeP = NULL;

    return ( funcP );
}
/*

*//* func_treearg( treeP, argN, treePP )

	Get the nth argument to a function out of a parse tree

Accepts :

	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)

Returns :

	<value>		Ptr to the nth argument subtree,
			NULL if no such argument

Notes :

	See parse_func_call for a description of the arguments tree.

*/

PTNODE *
func_treearg ARGLIST( (treeP, argN) )
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
    FARG( int		argN )		/* Argument # to get */
{
    /* Idiot checks */
    if ( ( argN < 0 ) || ( treeP == NULL ) )
	return ( NULL );

    /* Arguments are in an LSEQ list with trees on the right side of
       each LSEQ node, and with the last argument being the final
       (leftwise) node.  So if there is only one argument there is
       no LSEQ node.
    */
    while ( ( argN-- > 0 ) && ( treeP != NULL ) )
	/* Select the next leftward branch unless we're at the end */
	if ( treeP->pt_type == PN_LSEQ )
	    treeP = treeP->pt_leftP;
	else
	    treeP = NULL;

    /* If we have an LSEQ, return the right branch.  Otherwise
       return what we have.
    */
    if ( ( treeP != NULL ) && ( treeP->pt_type == PN_LSEQ ) )
	treeP = treeP->pt_rightP;

    return ( treeP );
}
/*

*//* func_treearg_estr( fnameP, treeP, argN, rvalP )

	Evaluate nth argument of a function, return ESTR

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	rvalP		Where to put the evaluation VALUE

Returns :

	<value>		Ptr to ESTR encapsulating the string

Notes :

	Appropriate warning messages are issued if result
	  can not be coerced to a string.

	Returned value is a temporary estr (see estr_tmp_xxx)

	If the return is non-null (i.e., a string is returned),
	  rvalP will have a valid VALUE that needs to be released
	  by the caller at some point.  If the return is null,
	  no such rvalP is returned.

*/

ESTR *
func_treearg_estr ARGLIST( (fnameP, treeP, argN, rvalP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
    FARG( VALUE		*rvalP )	/* Where to return the VALUE */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */
	ESTR		*esP;

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be supplied.", fnameP, argN+1 );
	return ( NULL );
    }

    /* Evaluate it */
    status = eval_tree( argP, rvalP );
    if ( status != SCOK )
	return ( NULL );

    /* Turn it into an ESTR */
    esP = val_estr_tmp( rvalP );
    if ( esP == NULL )
	val_release( rvalP );

    return ( esP );
}
/*

*//* func_treearg_estr_new( fnameP, treeP, argN )

	Evaluate nth argument of a function, return ESTR with string copy

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)

Returns :

	<value>		Ptr to ESTR encapsulating the string

Notes :

	Appropriate warning messages are issued if result
	  can not be coerced to a string.

	Returned value is a temporary estr (see estr_tmp_xxx) that
	  encapsulates a new copy of the evaluated string.  Because
	  a copy is made, the evaluated argument is released here.

	Caller should arrange to deallocate the new string when done
	  with it.  (But not the estr)

	Use only when you know that a new copy of the string will be
	  required; making copies is inefficient, and if what you
	  might want is just the evaluated string, don't use this
	  function.

*/

ESTR *
func_treearg_estr_new ARGLIST( (fnameP, treeP, argN ) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
    FARG( int		argN )		/* Argument number */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */
	ESTR		*esP;
	VALUE		strval;		/* Evaluated string value */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be supplied.", fnameP, argN+1 );
	return ( NULL );
    }

    /* Evaluate it */
    status = eval_tree( argP, &strval );
    if ( status != SCOK )
	return ( NULL );

    /* Turn it into an ESTR and make a copy */
    esP = val_estr_tmp( &strval );
    if ( esP != NULL )
	esP = estr_tmp_estr_new( esP );

    /* Done with value. */
    val_release( &strval );

    return ( esP );
}
/*

*//* func_treearg_estr_tmpstr( fnameP, treeP, argN )

	Evaluate nth argument of a function, return ESTR with tmpstr copy

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)

Returns :

	<value>		Ptr to ESTR encapsulating the string

Notes :

	Appropriate warning messages are issued if result
	  can not be coerced to a string.

	Returned value is a temporary estr (see estr_tmp_xxx) that
	  encapsulates a tmpstr copy of the evaluated string.  Because
	  a copy is made, the evaluated argument is released here.

	Because a tmpstr copy is used, this differs from
	  func_treearg_estr_new in that:
	   - the copied string does not have to be deallocated;
	   - the copied string must be consumed before the tmpstr
	     ring rolls around (i.e., soon).  Don't use when the string
	     has to be visible for long.

	Use only when you know that a copy of the string will be required;
	  making copies is inefficient, and if what you might want is just
	  the evaluated string, don't use this function.

*/

ESTR *
func_treearg_estr_tmpstr ARGLIST( (fnameP, treeP, argN ) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
    FARG( int		argN )		/* Argument number */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */
	ESTR		*esP;
	VALUE		strval;		/* Evaluated string value */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be supplied.", fnameP, argN+1 );
	return ( NULL );
    }

    /* Evaluate it */
    status = eval_tree( argP, &strval );
    if ( status != SCOK )
	return ( NULL );

    /* Turn it into an ESTR and make a copy */
    esP = val_estr_tmp( &strval );
    if ( esP != NULL )
	esP = estr_tmp_estr_tmpstr( esP );

    /* Done with value. */
    val_release( &strval );

    return ( esP );
}
/*

*//* func_treearg_long( fnameP, treeP, argN, longP )

	Evaluate nth argument of a function, return long

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	longP		Where to return the evaluated long value

Returns :

	<value>		Status code

Notes :

	Appropriate warning messages are issued if result is not SCOK

*/

STSCODE
func_treearg_long ARGLIST( (fnameP, treeP, argN, longP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
    FARG( long		*longP )	/* Where to return the long int */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be a number.", fnameP, argN+1 );
	return ( SCNOTDONE );
    }

    /* Evaluate it */
    status = eval_tree_long( argP, longP );

    return ( status );
}
/*

*//* func_treearg_str( fnameP, treeP, argN, rvalP, maxlen )

	Evaluate nth argument of a function, return NUL-terminated string

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	rvalP		Where to put the evaluation VALUE
	maxlen		Biggest string length we'll take

Returns :

	<value>		Ptr to nul-terminated string with the value

Notes :

	Appropriate warning messages are issued if result
	  can not be coerced to a string.

	The returned value is a pointer to the NUL-terminated string
	  that is the value.  This might be a pointer to the actual
	  string value if it's already NUL-terminated, or it might
	  be a pointer to a tmpstr() that has been allocated to
	  hold a copy, if not.

	Since this function might return a tmpstr, which is simply
	  plucked out of a round-robin pool, caller must be sure to
	  use it up quickly (and not be in a mode where consuming
	  tmpstrs will matter greatly).

	maxlen specifies the maximum number of bytes to accept if
	  a tmpstr copy has to be made.  Caller may set this to 0
	  to indicate unlimited.

	If the return is non-null (i.e., a string is returned),
	  rvalP will have a valid VALUE that needs to be released
	  by the caller at some point.  If the return is null,
	  no such rvalP is returned.

*/

char *
func_treearg_str ARGLIST( (fnameP, treeP, argN, rvalP, maxlen) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
   NFARG( VALUE		*rvalP )	/* Where to return the VALUE */
    FARG( int		maxlen )	/* Maximum length if temporary */
{
	STSCODE		status;
	int		slen;		/* Length of string */
	char		*strP;		/* Ptr to nul-terminated string */
	ESTR		*esP;

    /* Get the string if we can. */
    esP = func_treearg_estr( fnameP, treeP, argN, rvalP );
    if ( esP == NULL )
	return ( NULL );

    /* If it's already nul-terminated, we're great, return it. */
    if ( esP->es_len == -1 )
	return ( (char *)esP->es_strP );

    slen = estr_len( esP );
    if ( esP->es_strP[slen] == NUL )
	return ( (char *)esP->es_strP );

    /* Nope, need a copy to be safe. */
    if ( ( maxlen > 0 ) && ( slen > maxlen ) ) {
	/* Larger than wanted.  Give warning and don't do it. */
	warning( "%s: string too large, \"%s\"...",
		 fnameP, tmpstr_estr( esP ) );
	strP = NULL;
    }
    else if ( ( strP = tmpstr( slen ) ) == NULL ) {
	warning( "%s: can't get tmpstr for \"%s\"...", 
		 fnameP, tmpstr_estr( esP ) );
    }
    else {
	memcpy( strP, esP->es_strP, slen );
	strP[ slen ] = NUL;
    }

    /* If we are left stringless, abort */
    if ( strP == NULL )
	val_release( rvalP );

    return ( strP );
}
/*

*//* func_treearg_string( fnameP, treeP, argN, strvalP )

	Evaluate nth argument of a function, return string

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	strvalP		Where to return the string value

Returns :

	<value>		Status code

Notes :

	Appropriate warning messages are issued if result is not SCOK

	This function finds the nth argument and evaluates it and
	  verifies that the result is a string type.  It does no further
	  access to that string-- that's left to the caller.

	This function does NOT coerce the result to a string; i.e.,
	  the argument has to be a string, and not something that can
	  be converted to one.  If the latter is wanted, do the
	  evaluation another way e.g. via func_treearg_string_coerce

*/

STSCODE
func_treearg_string ARGLIST( (fnameP, treeP, argN, strvalP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
    FARG( VALUE		*strvalP )	/* Where to return string value */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: string argument %d must be supplied.", fnameP, argN+1 );
	return ( SCNOTDONE );
    }

    /* Evaluate it */
    status = eval_tree( argP, strvalP );
    if ( status != SCOK )
	return ( status );

    if ( strvalP->v_typeP->ty_class != TC_STRING ) {
	warning( "%s: argument %d must be a string.", fnameP, argN +1 );
	val_release( strvalP );
	return ( SCNOTDONE );
    }

    return ( status );
}
/*

*//* func_treearg_string_coerce( fnameP, treeP, argN, strvalP )

	Evaluate nth argument of a function, forced to a string

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	strvalP		Where to return the string value

Returns :

	<value>		Status code

Notes :

	Appropriate warning messages are issued if result is not SCOK

	This function finds the nth argument and evaluates it and forces
	  the result to a string type.  It does no further access to that
	  string-- that's left to the caller.

*/

STSCODE
func_treearg_string_coerce ARGLIST( (fnameP, treeP, argN, strvalP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
    FARG( VALUE		*strvalP )	/* Where to return string value */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */
	VALUE		sval;		/* The initial value */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be supplied.", fnameP, argN+1 );
	return ( SCNOTDONE );
    }

    /* Evaluate it to the caller's value, hoping it is a string */
    status = eval_tree( argP, strvalP );
    if ( status != SCOK )
	return ( status );

    /* If it's a string already, great, return it. */
    if ( strvalP->v_typeP->ty_class == TC_STRING )
	return ( SCOK );

    /* Otherwise see if we can force it to be a string. */
    memcpy( &sval, strvalP, sizeof(VALUE) );
    status = val_string_val( strvalP, &sval, TRUE );
    val_release( &sval );

    return ( status );
}
/*

*//* func_treearg_word( fnameP, treeP, argN, wordP )

	Evaluate nth argument of a function, return word

Accepts :

	fnameP		The function name (for error messages)
	treeP		Ptr to the argument tree for a function
	argN		Argument number to get (0 is first)
	wordP		Where to return the evaluated word int

Returns :

	<value>		Status code

Notes :

	Appropriate warning messages are issued if result is not SCOK

*/

STSCODE
func_treearg_word ARGLIST( (fnameP, treeP, argN, wordP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( PTNODE	*treeP )	/* Ptr to arguments tree */
   NFARG( int		argN )		/* Argument number */
    FARG( WORD		*wordP )	/* Where to return the word int */
{
	STSCODE		status;
	PTNODE		*argP;		/* Ptr to argument subtree */

    /* Get the argument subtree */
    if ( ( argP = func_treearg( treeP, argN ) ) == NULL ) {
	warning( "%s: argument %d must be a number.", fnameP, argN +1 );
	return ( SCNOTDONE );
    }

    /* Evaluate it */
    status = eval_tree_word( argP, wordP );

    return ( status );
}
/*

*//* mfl_func_arg( rvalP, argN) )

	Reference an MFL function actual argument

Accepts :

	*rvalP		Where to return the value
	argN		Argument number (0 through n)

Returns :

	<value>		Status code

Notes :


*/

STSCODE
mfl_func_arg ARGLIST( (rvalP, argN) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( int		argN )		/* Argument number */
{
    /* Make sure the value is in scope. */
    if ( ( As_P == NULL ) || ( argN > Asf_argC ) ) {
	warning( "mfl_func_arg: invalid reference to argument #%d", argN );
	return (SCNOTDONE);
    }

    /* Reference the value. */
    val_ref( rvalP, &As_P[ Asf_X + argN ] );

    return ( SCOK );
}
/*

*//* mfl_func_arg_frame_grow( argN )

	Grows the current argument stack frame

Accepts :

	argN		The number of local slots needed

Returns :

	<value>		Status code

Notes :

	Grows the stack frame by increasing the local value count.

	Does not give a warning message: caller should do that if error.

*/

STSCODE
mfl_func_arg_frame_grow ARGLIST( (argC) )
    FARG( int		argC )		/* Number of values needed */
{
    if ( ( Asf_X + Asf_argC + Asf_localC + argC ) > As_L )
	return ( SCNOTDONE );
    else {
	Asf_localC += argC;
	return ( SCOK );
    }
}
/*

*//* mfl_func_arg_frame_mark()

	Puts a "mark" on the stack frame

Accepts :

	<nothing>

Returns :

	<value>		Index after the mark
			-1 if error

Notes :

	The MFL argument stack is allocated with a fixed size
	on startup.  It's used to hold actual function argument
	values and potentially other function-local values.

	An argument stack frame is bounded by the start point
	in this stack and the number of values taken by this
	frame.  A new frame is created by marking the old frame
	information and then using the new frame at the new
	top of the stack.  The "mark" enables us to pop back the
	prior stack frame when the new one is done.  

	Note that the marking done by this routine does not completely
	establish a new stack frame.  It merely places the mark.  This
	lets us do evaluations in the old frame context while building
	a new one.

*/

int
mfl_func_arg_frame_mark NOARGLIST
{
	int		curlen;
	VALUE		*valP;

    /* Get current stack size: start of frame plus count of things
       on the frame.
    */
    curlen = Asf_X + Asf_argC + Asf_localC;

    /* The mark is 3 values */
    if ( mfl_func_arg_frame_grow( 3 ) != SCOK )
	return ( -1 );

    /* The mark consists of these values:

       - the current frame start;
       - the current argument count;
       - the current local count (minus the length of the mark).

    */

    valP = &As_P[ curlen ];
    val_set_int( valP++, 4, Asf_X );
    val_set_int( valP++, 4, Asf_argC );
    val_set_int( valP++, 4, Asf_localC - 3 );

    return ( curlen +3 );
}
/*

*//* mfl_func_arg_frame_pop()

	Removes the current execution frame

Accepts :

	<nothing>

Returns :

	<value>		Status code

Notes :


*/

STSCODE
mfl_func_arg_frame_pop NOARGLIST
{
	int		newX;		/* New base index */
	int		newargC;	/* New arg count */
	int		newlocalC;	/* New local count */
	VALUE		*valP;

    /* When we grew the frame, we placed marker values that let us
       unwind the stack.  (see the mfl_func_arg_frame_mark comments)
    */
    if ( Asf_X < 3 ) {
	warning( "mfl_func_arg_frame_pop: no current frame!" );
	return (SCERR);
    }

    valP = &As_P[ Asf_X -1 ];
    newlocalC = val_get_int( valP-- );
    newargC = val_get_int( valP-- );
    newX = val_get_int( valP );

    /* Complain if they look troublesome */
    if ( ( newX < 0 ) ||
         ( newargC < 0 ) ||
	 ( newargC > 100 ) ||
	 ( newlocalC < 0 ) ||
	 ( newlocalC > 100 ) ) {
	warning( "mfl_func_arg_frame_pop: bad marker values." );
	return ( SCERR );
    }

    Asf_X = newX;
    Asf_argC = newargC;
    Asf_localC = newlocalC;

    return ( SCOK );
}
/*

*//* mfl_func_arg_frame_set( frameX )

	Sets the new stack frame previously marked

Accepts :

	frameX		New frame start

Returns :

	<value>		Status code

Notes :

	Caller has established the new frame beginning via
	mfl_func_arg_frame_mark().  This function assumes that
	any stack locations used at this point are function arguments.
	Caller is free to change that..

*/

STSCODE
mfl_func_arg_frame_set ARGLIST( (frameX) )
    FARG( int		frameX )	/* New frame point */
{
	int		curlen;

    /* Get current stack size */
    curlen = Asf_X + Asf_argC + Asf_localC;

    if ( curlen < frameX ) {
	warning( "mfl_func_arg_frame_set: bad index." );
	return ( SCNOTDONE );
    }

    Asf_X = frameX;			/* Set new frame */

    /* Assume that all locations used are arguments.  */
    Asf_argC = curlen - Asf_X;
    Asf_localC = 0;

    return ( SCOK );
}
/*

*//* mfl_func_bifreg(biP)

	Register built-in functions from a BIFINF table

Accepts :

	biP		Ptr to the BIFINF table

Returns :

	<value>		status code

Notes :

	BIFINF table ends with a NULL name entry.

*/

STSCODE
mfl_func_bifreg ARGLIST( (biP) )
    FARG( BIFINF	*biP )		/* Ptr to the table */
{
    /* Walk the table registering functions */
    for ( ; biP->bif_nameP != NULL; ++biP ) {
	if ( func_register( biP->bif_nameP,  NULL,
			      FUNC_BUILTIN,
			      (TYPE *)NULL,
			      (FUNCFARGS *)NULL,
			      (void *)biP->bif_bif ) == NULL ) {
	    return ( FALSE );
	}
    }

    return ( TRUE );
}
/*

*//* mfl_func_exec( rvalP, funcP, argtreeP )

	Execute (evaluate) an MFL function

Accepts :

	rvalP		Where to put the return value
	funcP		Ptr to function descriptor block
	argtreeP	Ptr to arguments tree

Returns :

	<value>		STSCODE status code

Notes :


*/

STSCODE
mfl_func_exec ARGLIST( (rvalP, funcP, argtreeP) )
   NFARG( VALUE		*rvalP )	/* Where to put return value */
   NFARG( FUNC		*funcP )	/* Function info */
    FARG( PTNODE	*argtreeP )	/* Ptr to arguments tree */
{
	STSCODE		status;
	int		argN;
	int		frameX;		/* Start of new argument frame */
	PTNODE		*panodeP;	/* Passed argument heads */
	PTNODE		*pargP;		/* Passed argument branch */
	VALUE		*valP;

    /* Mark the new argument frame */
    frameX = mfl_func_arg_frame_mark();
    if ( frameX < 0 ) {
	warning( "mfl_func_exec: argument stack overflow." );
	return ( SCNOTDONE );
    }

    /* Make room for new values */
    status = mfl_func_arg_frame_grow( funcP->fn_fargs.ffl_argC );
    if ( status != SCOK ) {
	/* Can't grow- let's set new frame and pop it just to be anal */
	mfl_func_arg_frame_set( frameX );
	mfl_func_arg_frame_pop();
	return ( status );
    }

    /* Make sure all arguments in the frame are initialized.  We're
       going to walk through those arguments in parallel with the
       passed arguments.
    */
    for ( argN = 0, panodeP = argtreeP; ; ++argN) {

	/* Find the passed argument branch and skip to the next.
	   Arguments are passed as right branches off of a leftwise
	   sequence of LSEQ nodes, except that the last (or only) arg
	   doesn't have an LSEQ.
	*/
	if ( panodeP != NULL ) {
	    if ( panodeP->pt_type == PN_LSEQ ) {
		pargP = panodeP->pt_rightP;
		panodeP = panodeP->pt_leftP;
	    }
	    else {
		pargP = panodeP;
		panodeP = NULL;
	    }
	}
	else
	    pargP = NULL;

	if ( argN >= funcP->fn_fargs.ffl_argC ) {
	    /* No more formal arguments defined. */
	    if ( panodeP == NULL )
		break;			/* No more passed arguments either. */
	    else {
TODO		/* More arguments passed than defined.. do we care? */
		;
	    }
	}
	else {
	    /* Instantiate this actual argument. */
	    valP = &As_P[ frameX + argN ];

	    /* If there's an actual argument, evaluate and assign it */
	    if ( pargP != NULL ) {
		status = eval_tree( pargP, valP );
		if ( status == SCOK )
		    /* Coerce to the correct type */
		    status = val_coerce( valP,
				funcP->fn_fargs.ffl_argP[argN].fa_typeP,
				FALSE );
		if ( status != SCOK )
		    break;
	    }
	    else {
		/* No actual argument, just init the value */
		val_init( valP, funcP->fn_fargs.ffl_argP[argN].fa_typeP );

TODO		/* Should we issue a warning about the missing arg? */
	    }
	}
    }

    /* Finish new stack frame */
    mfl_func_arg_frame_set( frameX );

    if ( status == SCOK ) {
	/* Exec the function body */
	status = eval_tree( (PTNODE *)funcP->fn_valP, rvalP );

	/* status may be "SCRETURN" here as a result of a return
           statement.
	*/
	if ( status == SCRETURN )
	    status = SCOK;

	/* Coerce the return type if we need to */
	if ( ( status == SCOK ) && ( funcP->fn_rtypeP != NULL ) ) {
	    val_coerce( rvalP, funcP->fn_rtypeP, FALSE );
	}

    }

    /* Must release any actual arguments that were initialized in the
       argument accepting loop, above.  argN is still set to one
       above the largest argument number initialized.
    */
    while ( --argN >= 0 ) {
	val_release( &As_P[ frameX + argN ] );
    }

    /* Pop the arguments frame */
    if ( mfl_func_arg_frame_pop() != SCOK )
	status = SCERR;

    return ( status );
}
/*

*//* mfl_func_hook( fnameP )

	Exec an mfl function if it's defined

Accepts :

	fnameP		Name of the function

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Nonexistent function is an OK result

	Uses a stack buffer for an mfl command string.

*/

BOOL
mfl_func_hook ARGLIST( (fnameP) )
    FARG( char		*fnameP )
{
	BOOL		retval;
	char		mflcmd[MFL_TOKSIZE +1];

    if ( !mfl_func_hook_find( fnameP ) )
	return ( TRUE );

    if ( snprintf( &mflcmd[0], sizeof(mflcmd),
			"%s();\n", fnameP ) >= sizeof(mflcmd) ) {
	warning( "mfl_func_hook: function name \"%s\" too long.", fnameP );
	return ( FALSE );
    }

    retval = mfl_run_str( NULL,
			  estr_tmp_str( &mflcmd[0] ),
			  estr_tmp_str( "Mc" ) );

    return ( retval );
}
/*

*//* mfl_func_hook_dir_add( dirnameP )

	Add a directory to the hook path

Accepts :

	dirnameP	Directory name to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

*/

BOOL
mfl_func_hook_dir_add ARGLIST( (dirnameP) )
    FARG( ESTR		*dirnameP )	/* Directory name to add */
{
    /* Hand off to generic path-adding function */
    return ( path_dir_add( &Hookpath, dirnameP ) );
}
/*

*//* mfl_func_hook_find( hnameP )

	Find a hook function

Accepts :

	hnameP		Name of the hook function

Returns :

	<value>		TRUE if the hook function exists
			FALSE if not.

Notes :

	Hook functions may be defined by the admin (or user) in
	a startup file, or in directories along the hook function
	path.  This function will try to find a specific hook function,
	presumably when it's needed, by:

	  - seeing if the function is already defined;

	  - if not, searching each directory in the hook path to see if
	    there's a file with the same base name (the name without the
	    "$hook_" prefix) as the hook.  If there is, the file is run as
	    MFL code, and the check is again made to see if the hook has
	    been defined.

	The passed hook name MUST start with "$hook_" .

	The files in the hook path must NOT include the "$hook_" prefix.

*/

BOOL
mfl_func_hook_find ARGLIST( (hnameP) )
    FARG( char		*hnameP )	/* The hook name */
{
	char		hfname[ PATH_MAX +1 ];
	FILE		*fP;
	LL_NODE		*nodeP;

    /* Verify that name starts with "$hook_" */
    if ( strncmp( hnameP, "$hook_", 6 ) != 0 )
	return ( FALSE );

    /* If function already exists, we can rejoice */
    if ( func_lookup( hnameP ) != NULL )
	return ( TRUE );

    /* Search along the path. */
    for ( nodeP = Hookpath.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {

	/* See if we can find the function in its correspondingly-named file
	   (name without the "$hook"_ prefix)
	*/
	if ( snprintf( &hfname[0], sizeof(hfname),
		       "%s/%s",
		       (char *)nodeP->lln_valP,
		       hnameP +6 ) >= sizeof(hfname) ) {
	    warning( "mfl_func_hook_find: file path too large: \"%s/%s\"",
		      (char *)nodeP->lln_valP,
		      hnameP + 6 );
	    continue;
	}
	if ( ( fP = fopen( &hfname[0], "r" ) ) == NULL )
	    continue;

	/* Attempt to run this file. */
	mfl_run_fp( NULL, fP, estr_tmp_str( "Mc" ), &hfname[0] );
	fclose( fP );

	/* And now check for the function */
	if ( func_lookup( hnameP ) != NULL )
	    return ( TRUE );
    }

    /* Sadly, no such hook function can be found. */
    return ( FALSE );
}
/*

*//* mfl_func_init()

	Initialize the functions module

Accepts :


Returns :

	<value>		TRUE if OK status,
			FALSE if not.

Notes :


*/

BOOL
mfl_func_init NOARGLIST
{
	int		i;

    /* Allocate the mfl functions argument stack */
    i = 10L * 1024;
    As_P = (VALUE *)emalloc( "mfl_func_init",
					"argument stack",
					i * sizeof(VALUE) );
    if ( As_P == NULL )
	return ( FALSE );
    As_L = i;
    Asf_X = Asf_argC = 0;

    /* Initialize the hooks path */
    ll_init( &Hookpath );

    return( TRUE );
}
/*

*//* parse_func_call( inP, symtokP, treePP )

	Parses a function call, once the open paren has been seen

Accepts :

	inP		Ptr to input stack
	symtokP		Ptr to the function name symbol token
	treePP		Where to put the resulting parse tree

Returns :

	<value>		Status code

Notes :

	Caller has already parsed the function name and the opening
	parenthesis.

	Builds a parse tree as follows:

                PN_FUNCREF
                    /\                  PN_SYM is function name
                   /  \
                  /    \                arg list is per
             PN_SYM    arg list          parse_expr_list()

*/

STSCODE
parse_func_call ARGLIST( (inP, symtokP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TOKEN		*symtokP )	/* Function name symbol */
    FARG( PTNODE	**treePP )	/* Where to put tree node */
{
	STSCODE		status;
	PTNODE		*frefP;		/* Node for function reference */
	TOKEN		*tokP;

    /* Make the function reference node */
    status = parsetree_mknode( &frefP, PN_FUNCREF );
    if ( status != SCOK )
	return ( status );

    mfl_tokreserve( symtokP );		/* Don't lose this token */

    /* The argument list will just be a single expression tree.
       Expressions can have comma-separated subexpressions, we'll
       just use that fact here.  Parse the argument(s) expression
       and hang it on the right side of the function reference node.
    */
    status = parse_expr_list( inP, &frefP->pt_rightP );
    if ( status == SCOK ) {
	/* Get the closing paren. */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( tokP->tk_type != TK_CHAR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
	    mfl_warning( inP,
	       "missing closing parenthesis in call to function \"%s\"",
	       symtokP->tk_bstrP->bs_bP );
	    status = SCNOTDONE;
	}
    }

    /* If we're OK, hang the function symbol on the left side of
       the function reference node.
    */
    if ( status == SCOK )
	status = parsetree_mktknode( &frefP->pt_leftP, symtokP );

    /* Wrap up */
    if ( status == SCOK ) {
	/* Good to go: give back pointer to the new function reference */
	*treePP = frefP;
    }
    else {
	/* Some error.  Delete any tree that we created. */
	parsetree_free( frefP );
	*treePP = NULL;
    }

    mfl_tokrelease( symtokP );		/* Done with this token now */

    return ( status );
}
/*

*//* parse_func_def( inP, treePP, vnameP )

	Parse a function definition

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the resulting parse tree
	vnameP		Ptr to function variable name
	
Returns :

	<value>		SCxx status code

Notes :

	This parses a function declaration/definition.  A function
	is defined if it includes a body: this can only happen if
	we're not in any nesting levels.

	This is called out of the parse_type mainline (it's forked into
	this module to keep function-related stuff together).  The
	calling/return interface is dictated by the parse_type stuff.

	Builds a parse tree as follows:

                   PN_FUNCDEF
                    / |                 PN_SYM is function name
                   /  |
                  /   funcargs          pt_info.od_P has any FUNCFARGS block  
             PN_SYM


*/

STSCODE
parse_func_def ARGLIST( (inP, treePP, vnameP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to put definition tree */
    FARG( char		*vnameP )	/* Variable name */
{
	STSCODE		status;		/* Status code */
	int		n;		/* General */
	TYPE		*btypeP;	/* Base type */
	TYPE		*argtypeP;	/* argument type ptr */
	TNSYM		*tsymP;		/* type name symbol ptr */
	TOKEN		*tokP;
	PTNODE		*fdnodeP;	/* The funcdef node */
	FUNCFARGS	*fargsP;	/* Function arguments */
	FUNCFARGS	fargs;		/* Function arguments temporary */

    /* Attempt to parse prototype argument declarations */
    fargs.ffl_argC = fargs.ffl_argL = 0;
    fargs.ffl_argP = NULL;
    for( status = SCOK; status == SCOK; ) {
	tokP = mfl_sigtoken( inP );
	if ( ( tokP != NULL ) && ( tokP->tk_type == TK_ELLIPSIS ) ) {
	    /* Ellipsis is the last thing */
	    tokP = mfl_sigtoken( inP );
	    break;
	}
	if ( tokP->tk_type != TK_SYM )
	    break;
	tsymP = (TNSYM *)symtab_lookup( STtname, tokP->tk_bstrP->bs_bP );
	if ( tsymP == NULL ) {
	    mfl_warning( inP,
	"bad function argument definition.  A type is needed, not \"%s\".",
		(char *)tokP->tk_bstrP->bs_bP );
	    status = SCNOTDONE;
	    break;
	}

	/* Parse the type */
	status = parse_type_base( inP, tsymP, &btypeP );
	if ( status == SCOK ) {
	    status = parse_type_usage( inP, btypeP, &argtypeP, &vnameP,
				        (ODINFO *)NULL );
	    if ( ( status == SCOK ) && ( vnameP == NULL ) ) {
		mfl_warning( inP, "missing formal argument name.");
		status = SCNOTDONE;
	    }
	}
	/* Check for duplicate argname */
	if ( ( status == SCOK ) && ( fargs_find( &fargs, vnameP ) >= 0 ) ) {
	    mfl_warning( inP, "Duplicate formal argument name \"%s\".",
					vnameP );
	    status = SCNOTDONE;
	}
	if ( status != SCOK )
	    break;

	/* Remember this argument */
	if ( fargs.ffl_argC == fargs.ffl_argL ) {
	    /* Need to expand or create the array */
	    n = fargs.ffl_argL + 5;
	    if ( fargs.ffl_argL == 0 )
		fargs.ffl_argP = (FUNCFARG *)emalloc( "parse_func_def",
					  "formal argument list",
					  n * sizeof(FUNCFARG) );
	    else
		fargs.ffl_argP = (FUNCFARG *)erealloc( "parse_func_def",
						"formal argument list",
						fargs.ffl_argP,
						n * sizeof(FUNCFARG) );
	    if ( fargs.ffl_argP == NULL ) {
		status = SCERR;
		fargs.ffl_argC = fargs.ffl_argL = 0;
		break;
	    }
	    fargs.ffl_argL = n;
	}

	/* Add new argument */
	n = fargs.ffl_argC++;
	fargs.ffl_argP[n].fa_nameP = vnameP;
	fargs.ffl_argP[n].fa_typeP = argtypeP;

	/* Look for comma */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( tokP->tk_type != TK_OPR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != ',' ) )
	    break;
    }

    if ( status == SCOK ) {
	/* Closing paren must occur now. */
	if ( ( tokP == NULL ) ||
             ( tokP->tk_type != TK_CHAR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
	    mfl_warning( inP, "unclosed paren in function definition" );
	    status = SCNOTDONE;
	}
    }

    /* If bad status here here we can just return. */
    if ( status != SCOK ) {
	return ( status );
    }

    /* Can't define a function inside any other element */
    if ( ( status == SCOK ) && ( Mfl_nestC > 0 ) ) {
	mfl_warning( inP, "Function definition not allowed here." );
	status = SCNOTDONE;
    }

    /* If OK status here, make a copy of function arguments block if any */
    if ( ( status == SCOK ) && ( fargs.ffl_argC > 0 ) ) {
	fargsP = (FUNCFARGS *)emalloc( "parse_func_def", "fargs block",
				      sizeof(FUNCFARGS) );
	if ( fargsP == NULL )
	    status = SCERR;
	else
	    *fargsP = fargs;
    }
    else
	fargsP = NULL;
	
    /* Create the funcdef node */
    if ( status == SCOK )
	status = parsetree_mknode( &fdnodeP, PN_FUNCDEF );
    else
	fdnodeP = NULL;

    /* If still OK, hook the function args into the info pointer */
    if ( status == SCOK ) {
	fdnodeP->pt_info.od_P = fargsP;
	fdnodeP->pt_info.od_del = funcfargsdel;
    }
    else if ( fargsP != NULL ) {
	/* Error- remove functions arg storage */
	funcfargsdel( fargsP );
    }

    /* Wrap up. */
    *treePP = fdnodeP;

    return ( status );
}
/*

*//* parse_func_value( inP, trnodeP )

	Parse any initializer (i.e., body) for a function

Accepts :

	inP		Ptr to input stack
	trnodeP		Ptr to the PN_TYPEREF node we're building
	
Returns :

	<value>		SCxx status code

Notes :

	Called by declaration parsing stuff to add any initial value
	to a PN_TYPEREF node.  In the case of a function, the initial
	value is the function body.

	The body (if any) is linked to the right node of the PN_TYPEREF
	node.

*/

STSCODE
parse_func_value ARGLIST( (inP, trnodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*trnodeP )	/* PN_TYPEREF node */
{
	STSCODE		status;		/* Status code */
	TOKEN		*tokP;
	PTNODE		*bodyP;		/* The function body */
	FUNCFARGS	*oldfargsP;	/* Old global fargs */

    /* Check for opening brace for body */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCOK );		/* EOF, catch it later. */

    if ( ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	/* No body-- take back this token and be done. */
	mfl_tokback( inP, tokP );
	return ( SCOK );
    }

    /* Make formal arguments visible to parsing scope.  First push
       back anything already there (this shouldn't happen, but who
       knows, someday we may have nested functions).
    */
    oldfargsP = FargsP;
    FargsP = (FUNCFARGS *)trnodeP->pt_xinfo.od_P;

    /* Get the body */
    status = parse_compound( inP, &bodyP );

    /* Pop the function args back. */
    FargsP = oldfargsP;

    /* Return it on the right side if success. */
    if ( status == SCOK )
	trnodeP->pt_rightP = bodyP;

    return ( status );
}
/*

*//* funcfargsdel( fargsP )

	Deallocate a FUNCFARGS element

Accepts :

	fargsP		Ptr to FUNCFARGS block

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


	Specified as the "info deletion routine" for a parse node
	when pt_info.od_P points to a FUNCFARGS block

*/

BOOL
funcfargsdel ARGLIST( (fargsP) )
    FARG( FUNCFARGS	*fargsP )	/* FUNCFARGS thing */
{
    dealloc( fargsP->ffl_argP );
    dealloc( fargsP );

    return ( TRUE );
}
