/* mfl-misc.c -- Data, initialization, and misc functions for mfl

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains global data used by the mail filter language;

Also contains misc general functions (such as initialization) for mfl.


*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-in.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>
#include "data.h"

#include "mvmf.h"
#include "controls.h"

#include "mfl.h"
#include "mfl-type.h"
#include "mfl-token.h"
#include "mfl-eval.h"
#include "mfl-exec.h"
#include "mfl-parse.h"
#include "mfl-pp.h"
#include "mfl-func.h"
#include "mfl-sieve.h"
#include "mfl-misc.h"

#include "errwarn.h"

/* Local definitions */

    /* Force C_STMT_TERM to be FALSE if not set; typically the configure
       process will undef it rather than make it 0.
    */
#ifndef	C_STMT_TERM
# define C_STMT_TERM	FALSE
#endif

    /* Built-in types info */
typedef struct {
    char	*bit_symP;		/* Symbol */
    UBYTE	bit_flags;		/* Flags */
    UBYTE	bit_class;		/* Type class */
    BYTE	bit_size;		/* Size */
    TNSYM	**bit_tnsymP;		/* Where to put TNSYM addr */
}  BITINFO;



/* External data referenced */

    /* External built-in function tables for registration,
       and corresponding module initialization function.
    */
extern	BIFINF	Bifinf_std[];		/* Standard functions */
extern	BOOL	mfl_func_init();

extern	BIFINF	Bifinf__[];		/* Misc (unclassified) stuff */
extern	BOOL	bif__init();
extern	BIFINF	Bifinf_control[];	/* string stuff */
extern	BOOL	bif_control_init();
extern	BIFINF	Bifinf_db[];		/* DB stuff */
extern	BOOL	bif_db_init();
extern	BIFINF	Bifinf_cusp[];		/* CUSP stuff */
extern	BOOL	bif_cusp_init();
extern	BIFINF	Bifinf_dns[];		/* DNS stuff */
extern	BOOL	bif_dns_init();
extern	BIFINF	Bifinf_msgpart[];	/* message part stuff */
extern	BOOL	bif_msgpart_init();
extern	BIFINF	Bifinf_str[];		/* string stuff */
extern	BOOL	bif_str_init();


/* External routines used */



/* Local data publicly available */

	int	Align2 = 2;		/* Alignment for 2-byte things */
	int	Align4 = 4;		/* Alignment for >2-byte things */
	BOOL	MexpandF = TRUE;	/* Whether expanding macros */

	int	Mfl_checkF = FALSE;	/* For check mode (parse and report
					   errors there, but don't exec).
					*/
	int	Mfl_nestC;		/* Nesting level */
	BOOL	Mfl_testF = FALSE;	/* For test mode (don't do
					   delivery, but report what would
					   happen..)
					*/
	BOOL	Mfl_traceF = FALSE;	/* Tracing enabled/disabled */


    /* Symbol tables */
	SYMTAB	*STbis = NULL;		/* Built-in statement */
	SYMTAB	*STcusp = NULL;		/* CUSPs */
	SYMTAB	*STglobal = NULL;	/* Global variables */
	SYMTAB	*STpreproc = NULL;	/* Preprocessor */
	SYMTAB	*STstruct = NULL;	/* Struct names */
	SYMTAB	*STtname = NULL;	/* Type names */
	SYMTAB	*STtpool = NULL;	/* Type pool */


    /* Where to put TNSYM info */
	TNSYM	*Tnflt4;		/* Ptr to float4 entry */
	TNSYM	*Tnflt8;		/* Ptr to float8 entry */
	TNSYM	*Tnint1;		/* Ptr to int1 entry */
	TNSYM	*Tnint2;		/* Ptr to int2 entry */
	TNSYM	*Tnint4;		/* Ptr to int4 entry */
	TNSYM	*Tnstring;		/* Ptr to string entry */

    /* Predefined types */
        TYPSYM	*TyfuncP;		/* (function) */
	TYPSYM	*Tyint1pP;		/* (int1 *) */
	TYPSYM	*TystrpP;		/* (string *) */

	/* Table for operators.  Note Well:  because of the lookup
	   method used, any operator must appear BEFORE an operator
	   of which it is an initial substring.  That is, "!" must
	   appear before "!=", etc.  They need not be adjacent.  */

	OPINFO	Optbl[] = {

/*
  operator	priority	eval	pre-eval	post-eval
  			   float-opr    int-opr       string-opr
*/


{ "(",		1,		NULL,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },

{ "?",		ORCOND,		eval_condex,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },

/* Note - the following evcond reference is just a placeholder to
   make the parser work.  This operator will not be called by itself. */
/*{ ":",		ORCOND,		evcond,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },
*/


{ "!",		ORPREFIX,	NULL,	eval_prenot,	NULL,
			NULL,	NULL,	    NULL  },

{ "~",		ORPREFIX,	NULL,	eval_bitcomp,	NULL,
			    NULL,	NULL,	    NULL  },

{ ",",		ORCOMMA,	eval_comma,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },

{ "=",		ORASSIGN,	eval_assign,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },

{ "==",		ORRELEQ,	eval_rel,	NULL,	NULL,
			    evals_feq,	    evals_ieq,	    scmp_eq  },

{ "==^",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    scmpi_eq  },

{ "!=",		ORRELEQ,	eval_rel,	NULL,	NULL,
			    evals_fne, evals_ine,	    scmp_ne  },

{ "!=^",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	   NULL,	    scmpi_ne  },

{ "=.",		ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    sregex_eq  },

{ "!=.",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    sregex_ne  },

{ "=?",		ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    swild_eq  },

{ "=?^",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    swildi_eq  },

{ "!=?",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    swild_ne  },

{ "!=?^",	ORRELEQ,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    swildi_ne  },

{ "<",		ORRELREL,	eval_rel,	NULL,	NULL,
			    evals_flt,    evals_ilt,	    scmp_lt  },

{ "<^",		ORRELREL,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    scmpi_lt  },

{ "<=",		ORRELREL,	eval_rel,	NULL,	NULL,
			    evals_fle, evals_ile,	    scmp_le  },

{ "<=^",	ORRELREL,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    scmpi_le  },

{ "<<",		ORSHIFT,	eval_arith,	NULL,	NULL,
			    NULL,	evals_ilshift,	    NULL  },

{ "<<=",	ORASSIGN,	eval_opeq,	NULL,	NULL,
			    NULL,	evals_ilshift,	    NULL  },


{ ">",		ORRELREL,	eval_rel,	NULL,	NULL,
			    evals_fgt, evals_igt,	    scmp_gt  },

{ ">^",		ORRELREL,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    scmpi_gt  },

{ ">=",		ORRELREL,	eval_rel,	NULL,	NULL,
			    evals_fge, evals_ige,	    scmp_ge  },

{ ">=^",	ORRELREL,	eval_rel,	NULL,	NULL,
			    NULL,	    NULL,	    scmpi_ge  },

{ ">>",		ORSHIFT,	eval_arith,	NULL,	NULL,
			    NULL,	evals_irshift,	    NULL  },

{ ">>=",	ORASSIGN,	eval_opeq,	NULL,	NULL,
			    NULL,	evals_irshift,	    NULL  },


{ "+",		ORADD,		eval_plus,	eval_preplus,	NULL,
			    evals_fadd, evals_iadd,	    NULL  },

{ "+=",		ORASSIGN,	eval_pluseq,	NULL,	NULL,
			    evals_fadd, evals_iadd,	    NULL  },

{ "++",		ORPREFIX,	NULL,	eval_preincr,	eval_postincr,
			    evals_fadd, evals_iadd,	    NULL  },


{ "-",		ORADD,		eval_minus,	eval_preminus,	NULL,
			    evals_fsub, evals_isub,	    NULL  },

{ "-=",		ORASSIGN,	eval_mineq,	NULL,	NULL,
			    evals_fsub, evals_isub,	    NULL  },

{ "--",		ORPREFIX,	NULL,    eval_predecr, eval_postdecr,
			    evals_fadd, evals_iadd,	    NULL  },


{ "*",		ORMULT,		eval_arith,	eval_deref,	NULL,
			    evals_fmult, evals_imult,	    NULL  },

{ "*=",		ORASSIGN,	eval_opeq,	NULL,	NULL,
			    evals_fmult, evals_imult,	    NULL  },

{ "/",		ORMULT,		eval_arith,	NULL,	NULL,
			    evals_fdiv, evals_idiv,	    NULL  },

{ "/=",		ORASSIGN,	eval_opeq,	NULL,	NULL,
			    evals_fdiv, evals_idiv,	    NULL  },

{ "%",		ORMULT,		eval_arith,	NULL,	NULL,
			    NULL,	evals_imod,	    NULL  },

{ "%=",		ORASSIGN,	eval_opeq,	NULL,	NULL,
			    NULL,	evals_imod,	    NULL  },


{ "[",		ORMEMBER,	eval_aref,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },


{ "&",		ORBITAND,	eval_arith,	eval_addr,	NULL,
			    NULL,	evals_ibitand,	    NULL  },

{ "&&",		ORLAND,		eval_land,	NULL,	NULL,
			    NULL,	NULL,	        NULL  },

{ "&=",		ORASSIGN,	eval_opeq,	NULL,	NULL,
			    NULL,	evals_ibitand,	    NULL  },


{ "|",		ORBITOR,	eval_arith,	NULL,	NULL,
			    NULL,	evals_ibitor,	    NULL  },

{ "||",		ORLOR,		eval_lor,	NULL,	NULL,
			    NULL,	NULL,	        NULL  },

{ "|=",		ORASSIGN,	eval_opeq,	NULL,	NULL,
			    NULL,	evals_ibitor,	    NULL  },


{ ".",		ORMEMBER,	val_member,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },

{ "->",		ORMEMBER,	val_pmember,	NULL,	NULL,
			    NULL,	NULL,	    NULL  },


{ NULL,		0,		NULL,	NULL,	NULL,  NULL }
			   };



    /* Table of built-in executable statements */
	BISINFO Bistbl[] = {

/*
  statement	parse rtc	exec rtc	stmt?	term?
*/

{ "break",	parse_bis_break, exec__break,	TRUE,	FALSE	},
{ "C",		parse_bis_c,	exec__c,	TRUE,	TRUE	},
{ "continue",	parse_bis_continue, exec__continue,	TRUE,	FALSE	},
{ "do",		parse_bis_do,	exec__do,	TRUE,	C_STMT_TERM	},
{ "else",	NULL,		NULL,		TRUE,	FALSE   },
{ "for",	parse_bis_for,	exec__for,	TRUE,	C_STMT_TERM	},
{ "if",		parse_bis_if,	exec__if,	TRUE,	C_STMT_TERM	},
{ "pv$",	parse_bis_pv_,	exec__pv_,	TRUE,	C_STMT_TERM	},
{ "return",	parse_bis_return, exec__return,	TRUE,	FALSE	},

#ifdef	SIEVE
{ "sieve",	parse_bis_sieve, exec__sieve,	TRUE,	TRUE	},
#endif	/* SIEVE */

{ "sizeof",	parse_bis_sizeof, exec__sizeof,	FALSE,	TRUE	},
{ "while",	parse_bis_while,  exec__while,	TRUE,	C_STMT_TERM	},

{ NULL,		NULL,		NULL,		FALSE,	FALSE	}
			    };



/* Local routines and forward references */



/* Private data */



    /* Table of built-in types (etc) */
static	BITINFO	Bittbl[] = {
{ "enum",	TMENUM,		0,		-1,	NULL },
{ "struct",	TMSTRUCT,	0,		-1,	NULL },
{ "union",	TMUNION,	0,		-1,	NULL },
{ "typedef",	TMTYPEDEF,	0,		-1,	NULL },
{ "unsigned",	TMUNSIGNED,	0,		-1,	NULL },
{ "signed",	TMSIGNED,	0,		-1,	NULL },
{ "short",	TMSHORT,	0,		-1,	NULL },
{ "long",	TMLONG,		0,		-1,	NULL },

/* { "void",	0,		TC_INT,		0,	NULL },    */
{ "char",	0,		TC_INT,		1,	&Tnint1 },
{ "int",	0,		TC_INT,		2,	&Tnint2 },
{ "$int4$",	0,		TC_INT,		4,	&Tnint4 },
{ "float",	0,		TC_FLOAT,	4,	&Tnflt4 },
{ "double",	0,		TC_FLOAT,	8,	&Tnflt8 },
{ "string",	0,		TC_STRING,  sizeof( VALPTR ),	&Tnstring },

{ NULL,		0,		0,		0,	NULL }
			  };



    /* Table of initializers for built-in-function stuff.
       (might want to autogenerate this one day, a la plugin definitions
       via standardized filenames etc.)
    */
static struct {
	BIFINF	*fm_biftblP;		/* Table of function definitions */
	BOOL	(*fm_init)();		/* Module initialization */
}  Funcmodtbl[] = {

/*	BIFtable		init function		*/
{	&Bifinf__[0],		bif__init		},
{	&Bifinf_control[0],	bif_control_init	},
{	&Bifinf_db[0],		bif_db_init		},
{	&Bifinf_cusp[0],	bif_cusp_init		},
{	&Bifinf_dns[0],		bif_dns_init		},
{	&Bifinf_msgpart[0],	bif_msgpart_init	},
{	&Bifinf_str[0],		bif_str_init		},

{	NULL,			NULL			}
	};


    /* Table of integer-value macros to define for MFL access
       (tempting to use cpp stringification for this but can't
        be sure that's widely available; what the hey, this is ok.
    */
static struct {
	char	*pp_nameP;		/* Name */
	int	pp_value;		/* integer value */
} Mflppintdefs[] = {
/*	name			definition		*/
{	"$MEX_OK",		MEX_OK			},
{	"$MEX_HARD",		MEX_HARD		},
{	"$MEX_SOFT",		MEX_SOFT		},
{	"$MEX_OKCONT",		MEX_OKCONT		},

#ifdef	QMAIL
{	"$QMEX_OK",		QMEX_OK			},
{	"$QMEX_STOP",		QMEX_STOP		},
{	"$QMEX_HARD",		QMEX_HARD		},
{	"$QMEX_SOFT",		QMEX_SOFT		},
#endif	/* QMAIL */

{	NULL,			0			}
};



    /* Table of integer controls to define */
static	CNTRLDEF_INT Ctdef_mfl_int[] = {
/*
 ------- name -------	val	admin	admin	user	user
				 low	 high	 low	high
*/
{ "mfl_mode_sieve",	  0,	    0,	    1,	   0,	   1	},
{ "mfl_trace_bif",	  0,	    0,	    1,	   0,	   1	},
{ "mfl_trace_functions",  0,	    0,	    1,	   0,	   1	},

{ NULL,			  0,	    0,	    0,	   0,	   0	}
		};


    /* Table of string controls to define */
static	CNTRLDEF_STR Ctdef_mfl_str[] = {
/*
 ------- name -------		--------- val --------		adminF
{ "some_thing",			"some_value",			 FALSE	},
*/

{ NULL,			  NULL,				     0	}
		};



       

/*

*//* mfl_init()

	Initialize the language

Accepts :


Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
mfl_init NOARGLIST
{
	int		i;		/* Scratch */
	BISINFO		*bisP;		/* Built-in-statements info */
	BITINFO		*bitP;		/* Built-in-types info */
	BISSYM		*bissymP;	/* Built-in statements table entry */
	TNSYM		*tnsymP;	/* Type name table entry */
	TYPSYM		*tpsymP;	/* Type pool table entry */
	TYPE		btype;		/* Type block */
	char		mflstr[1000];	/* A string for MFL execution */

    /* Define controls used by mfl */
    if ( !cntrl_table_def_int( &Ctdef_mfl_int[0] ) ) {
	warning( "mfl_init: problem defining integer controls." );
	return ( SCERR );
    }
    if ( !cntrl_table_def_str( &Ctdef_mfl_str[0] ) ) {
	warning( "mfl_init: problem defining string controls." );
	return ( SCERR );
    }

    /* Create the cusp symbol table */
    STcusp = symtab_create( 0, HSZCUSP,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STcusp == NULL )
	return (SCERR);

    /* Create the global variables symbol table */
    STglobal = symtab_create( 0, HSZGLOBAL,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STglobal == NULL )
	return (SCERR);

    /* Create the preprocessor symbol table */
    STpreproc = symtab_create( 0, HSZPP,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STpreproc == NULL )
	return (SCERR);

    /* Create the types symbol table */
    STtpool = symtab_create( 0, HSZTPOOL,
				type_symtab_hash,
				type_symtab_comp,
				type_symtab_keyassoc,
				type_symtab_keydis );
    if ( STtpool == NULL )
	return (SCERR);

    /* Create the aggregate (struct/union/enum) names symbol table */
    STstruct = symtab_create( 0, HSZSTRUCT,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STstruct == NULL )
	return (SCERR);

    /* Create the typenames symbol table */
    STtname = symtab_create( 0, HSZTYPE,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STtname == NULL )
	return (SCERR);

    /* Put the built-in types into the typenames table */
    for( bitP = &Bittbl[0]; bitP->bit_symP != NULL; ++bitP ) {
        tnsymP = (TNSYM *)symtab_enter( STtname, bitP->bit_symP,
					   sizeof( TNSYM ) );
	if ( tnsymP == NULL )
	    return (SCERR);

	tnsymP->tn_flags = bitP->bit_flags | TF_BUILTIN;
	if ( bitP->bit_size == -1 )
	    tnsymP->tn_typeP = NULL;
	else {
	    btype.ty_class = bitP->bit_class;
	    btype.ty_flags = TF_BUILTIN;
	    btype.ty_size = bitP->bit_size;
	    btype.ty_dim = 0;
	    btype.ty_nextP = NULL;
	    tpsymP = (TYPSYM *)symtab_enter( STtpool,
	    			&btype, sizeof(TYPSYM) );
	    if ( tpsymP == NULL )
		return (SCERR);
	    tnsymP->tn_typeP = &(tpsymP->ts_type);
	    tpsymP->ts_type.ty_refC = 1;

	    if ( bitP->bit_tnsymP != NULL )
	    	*(bitP->bit_tnsymP) = tnsymP;
	}
    }

    /* Enter the char * type (Tyint1pP) */
    if ( ( Tyint1pP = type_ptr( Tnint1->tn_typeP ) ) == NULL )
	return ( SCERR );
    Tyint1pP->ts_type.ty_refC = 1;
    
    /* Enter the string * type (TystrpP) */
    if ( ( TystrpP = type_ptr( Tnstring->tn_typeP ) ) == NULL )
	return ( SCERR );
    TystrpP->ts_type.ty_refC = 1;
    
    /* Enter the function type (TyfuncP) */
    btype.ty_flags = 0;
    btype.ty_size = 0;
    btype.ty_nextP = NULL;
    btype.ty_class = TC_FUNCT;
    TyfuncP = (TYPSYM *)symtab_enter( STtpool, &btype, sizeof(TYPSYM) );
    if ( TyfuncP == NULL )
	return ( SCERR );
    TyfuncP->ts_type.ty_refC = 1;
    

    /* Create the builtin statements symbol table */
    STbis = symtab_create( 0, HSZBIS,
				symtab_hash_str, symtab_comp_str,
				symtab_keyassoc_str, symtab_keydis_str );
    if ( STbis == NULL )
	return (SCERR);

    /* Built-in executable statement keywords go in type table too */
    for( i = 0, bisP = &Bistbl[0]; bisP->bis_symP != NULL; ++i, ++bisP ) {
    	/* Enter the symbol */
        bissymP = (BISSYM *)symtab_enter( STbis, bisP->bis_symP,
					   sizeof( BISSYM ) );
	if ( bissymP == NULL )
	    return ( SCERR );
	/* Fill in the entry */
	bissymP->bss_num = i;
    }


    /* Initialize the preprocessor */
    if ( !pp_init() )
	return ( SCERR );

    /* Initialize the main functions module */
    if ( !mfl_func_init() )
	return ( SCERR );

    /* Initialize built-in function module(s) via the init table */
    for ( i = 0; Funcmodtbl[i].fm_biftblP != NULL; ++i ) {
	/* Register built-in-function names in this module */
	if ( !mfl_func_bifreg( Funcmodtbl[i].fm_biftblP ) )
	    return ( SCERR );

	/* Do any module initialization */
	if ( Funcmodtbl[i].fm_init != NULL )
	    if ( !(*Funcmodtbl[i].fm_init)() )
	        return ( SCERR );
    }


    /* Misc data */
    Mfl_nestC = 0;			/* Nesting count */

    /* Initialize MFL/user preprocessor constants */
    for ( i = 0; Mflppintdefs[i].pp_nameP != NULL; ++i ) {
	if ( snprintf( &mflstr[0], sizeof(mflstr), "%sdefine %s %d",
		        MFL_PP_INTRODUCER,
			Mflppintdefs[i].pp_nameP,
			Mflppintdefs[i].pp_value ) >= sizeof(mflstr) ) {
	    warning( "mfl_init, Mflppintdefs string too large." );
	    return ( SCERR );
	}

	if ( mfl_run_str( NULL,
			  estr_tmp_str( &mflstr[0] ),
			  estr_tmp_str( "Mc" ) ) != SCOK ) {
	    warning( "mfl_init: error defining %s",
		     Mflppintdefs[i].pp_nameP );
	    return ( SCERR );
	}
    }

    return (SCOK);
}
/*

*//* mfl_array_from_list_strings( anameP, arrayL, listP )

	Make an MFL array from an internal linked list of strings

Accepts :

	anameP		The MFL variable name prefix for the array
	arrayL		The length of the array to create
	listP		Pointer to linked list containing strings

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Creates an MFL array named with the string in anameP, and two
	  variables with that base name and a suffix:
	    'L' for the length of the array;
	    'C' for the number of strings in the array.

	If arrayL is larger than the size required for the lists's
	  contents, arrayL is used.

	Does this via executing MFL code.

	Uses Gbs1P buffer

*/

BOOL
mfl_array_from_list_strings ARGLIST( (anameP, arrayL, listP) )
   NFARG( char		*anameP )	/* Name for array */
   NFARG( int		arrayL )	/* Length for array */
    FARG( LL_HEAD	*listP )	/* List for strings */
{
	BOOL		okF;
	int		i;
	int		nodeC;
	LL_NODE		*nodeP;

    /* Count the nodes */
    for ( nodeC = 0, nodeP = listP->llh_firstP; nodeP != NULL;
		++nodeC, nodeP = nodeP->lln_nextP ) {
	;
    }
    if ( nodeC > arrayL )
	arrayL = nodeC;			/* Must have size >= # of strings */
    if ( arrayL == 0 )
	arrayL = 1;			/* Can't have empty array */

    /* Create the variables for the array and the array length & count */
    Gbs1P->bs_bC = 0;
    okF = bstr_printf( Gbs1P, "string %s[%d];", anameP, arrayL );
    okF = okF && mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
			       estr_tmp_str( "Mc" ) );

    Gbs1P->bs_bC = 0;
    okF = okF && bstr_printf( Gbs1P, "int %sC = %d;", anameP, nodeC );
    okF = okF && mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
			       estr_tmp_str( "Mc" ) );

    Gbs1P->bs_bC = 0;
    okF = okF && bstr_printf( Gbs1P, "int %sL = %d;", anameP, arrayL );
    okF = okF && mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
			       estr_tmp_str( "Mc" ) );

    /* Assign the values to the array elements */
    for ( i = 0, nodeP = listP->llh_firstP; nodeP != NULL;
		++i, nodeP = nodeP->lln_nextP ) {
	Gbs1P->bs_bC = 0;
	okF = okF && bstr_printf( Gbs1P, "%s[%d] = \"", anameP, i );
	okF = okF && bstr_quote_str( Gbs1P, (UBYTE *)nodeP->lln_valP,
				     QF_DQUOTE );
	okF = okF && bstr_cat( Gbs1P, "\";" );
	if ( okF ) {
	    bstr_nulterm( Gbs1P );
	    okF = okF && mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
				       estr_tmp_str( "Mc" ) );
	}
    }

    return ( okF );
}
/*

*//* mfl_program_args( argC, argV, argN )

	Make program arguments available to MFL

Accepts :

	argC		argc from main()
	argV		argv from main()
	argN		Index of next argument (e.g., after options)

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Creates MFL variables $ArgC, $ArgV, $ArgN

	Does so via executing MFL code.  I really should have better
	glue code, but hey, when you can use an interpreter, there's
	no great shame in using it.

	Uses Gbs1P buffer

*/

BOOL
mfl_program_args ARGLIST( (argC, argV, argN) )
   NFARG( int		argC )		/* argc from main */
   NFARG( char		**argV )	/* argv from main */
    FARG( int		argN )		/* Next arg */
{
	BOOL		okF;
	int		i;

    /* Create mfl code that makes the variables. */
    Gbs1P->bs_bC = 0;
    okF = okF && bstr_printf( Gbs1P, "  int $ArgC = %d;\n", argC );
    okF = okF && bstr_printf( Gbs1P, "  int $ArgN = %d;\n", argN );
    okF = okF && bstr_printf( Gbs1P, "  string $ArgV[%d] = {\n", argC );
    for ( i = 0; okF && ( i < argC ); ++i ) {
	okF = bstr_printf( Gbs1P, "    \"" );
	okF = okF && bstr_quote_str( Gbs1P, argV[i], QF_DQUOTE );
	okF = okF && bstr_printf( Gbs1P, "\"%s\n",
			    ( ( i + 1 ) < argC ) ? "," : "" );
    }
    okF = okF && bstr_printf( Gbs1P, "  };\n" );

    if ( okF ) {
	bstr_nulterm( Gbs1P );
	if ( mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
			  estr_tmp_str( "Mc" ) ) != SCOK )
	    okF = FALSE;
    }

    return ( okF );
}
/*

*//* mfl_run_fname(rvalP, fnesP, msesP)

	Run mfl code that's in a named file

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	fnesP		Name of the file to run
	msesP		Modes string

Returns :

	<value>		status code

Notes :

	msesP is a string with characters indicating some operational
	  modes.  An 'O' character indicates that any following lowercase
	  characters, up to the next uppercase character, affect how to deal
	  with file opening issues, as follows:

		'e'	Return done (SCDONE) if file can't be opened
		'w'	Give warning if file can't be opened

	msesP is also passed along to mfl_run_fp().

	See notes at mfl_run_fp() as well.

	The rvalP value (if non-NULL) is set to -1 if the file can't be
	  opened or accessed, 1 if the final status was SCOK, and to 0
	  otherwise.

*/

STSCODE
mfl_run_fname ARGLIST( (rvalP, fnesP, msesP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( ESTR		*fnesP )	/* Name of file to open */
    FARG( ESTR		*msesP )	/* Modes characters string */
{
	STSCODE		status;
	UBYTE		ch;
	int		len;
	int		retval;		/* Value to return via rvalP */
	BOOL		warnedF;
	BOOL		mrelF;		/* If mode char is relevant */
	UBYTE		*strP;
	FILE		*fP;		/* File pointer */
	char		fname[PATH_MAX +1 ]; /* Actual filename */

    /* Get a nul-terminated copy of fname, which is what fopen() needs. */
    len = estr_len( fnesP );
    if ( ( len + 1 ) >= sizeof(fname) ) {
	warning( "mfl_run_fname: filename too long." );
	status = SCNOTDONE;
	retval = -1;
    }
    else {
	memcpy( &fname[0], fnesP->es_strP, len );
	fname[len] = NUL;

	fP = fopen( &fname[0], "r" );
	if ( fP != NULL ) {
	    /* Successful open; hand off to mfl_run_fp */
	    status = mfl_run_fp( rvalP, fP, msesP, &fname[0] );
	    retval = ( status == SCOK ? 1 : 0 );
	    fclose( fP );
	}
	else {
	    /* Can't open.  Work with the modes string in continuing. */
	    retval = -1;
	    warnedF = FALSE;
	    mrelF = FALSE;
	    status = SCNOTDONE;
	    strP = msesP->es_strP;
	    len = estr_len( msesP );
	    while ( len-- > 0 ) {
		ch = *strP++;
		if ( !mrelF ) {
		    /* Not in relevant area; just wait for an 'O' */
		    mrelF = ( ch == 'O' );
		    continue;
		}

		/* In relevant area; be done if out of it. */
		if ( !islower(ch) )
		    break;

		switch ( ch ) {
		    case 'e':		/* Return error (done) */
			status = SCDONE;
			break;

		    case 'w':		/* Issue warning */
			if ( !warnedF ) {
			    warning( "mfl_run_fname: Can't open file \"%s\"",
						&fname[0] );
			    warnedF = TRUE;
			}
			break;

		    default:
			break;
		}
	    }
	}
    }

    if ( rvalP != NULL )
	val_set_int( rvalP, 4, retval );

    if ( status == SCNOTDONE )
	status = SCOK;
	
    return ( status );
}
/*

*//* mfl_run_fp(rvalP, fP, msesP, fnameP)

	Run mfl code that's in an opened file

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	fP		File pointer for the open file
	msesP		Modes string
	fnameP		The name of the file or input (for logging purposes)

Returns :

	<value>		status code

Notes :

	The file is NOT closed by this function.

	fnameP is a NUL-terminated string for exposition.  It may be passed
	  as NULL.

	The msesP string is used to affect operation and is passed
	  along to mfl_run_in().

	See notes at calls mfl_run_in().

*/

STSCODE
mfl_run_fp ARGLIST( (rvalP, fP, msesP, fnameP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( FILE		*fP )		/* Input file */
   NFARG( ESTR		*msesP )	/* Modes string */
    FARG( char		*fnameP )	/* Name of the input */
{
	STSCODE		status;
	MML_IN		*inP;		/* Input stack */

    /* Create input stack */
    if ( ( inP = in_new() ) == NULL ) {
	warning( "mfl_run_fp: in_new() failed." );
	return ( SCERR );
    }

    /* Push the string */
    if ( !in_stream_push_file_nc( inP, fP, fnameP ) ) {
	warning( "mfl_run_fp: in_stream_push_fp() failed." );
	in_destroy( inP );
	return ( SCERR );
    }

    /* mfl-interp the input stack */
    status = mfl_run_in( rvalP, inP, msesP );

    /* Remove the input stack */
    in_destroy( inP );

    /* Be done */
    return ( status );
}
/*

*//* mfl_run_in(rvalP, inP, msesP)

	Run mfl code that's in an input stack

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	inP		Ptr to input stack to get MFL code from
	msesP		Modes string

Returns :

	<value>		status code

Notes :

	If rvalP is non-NULL, this provides a VALUE to use for
	the result of tree execution.  Only the final VALUE is returned
	in a multi-statement string.

	Note Well: if caller supplies this, the final value is
	not released.  Caller may need to call val_finish() or
	val_release() .

	If Mfl_checkF is true (parse-only mode), only parses, doesn't
	exec.

	modesP is a string of modes characters.  An uppercase character
	  indicates what area the immediately following string of lowercase
	  characters applies to.  Here:

	  'M' specifies mfl modes:
	      'c' starts in c-like mode
	      'p' signifies parse-only (forces Mfl_checkF for the duration)
	      's' starts in sieve mode

*/

STSCODE
mfl_run_in ARGLIST( (rvalP, inP, msesP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( ESTR		*msesP )	/* Modes string */
{
	STSCODE		status;
	BOOL		valF;		/* TRUE if any value present */
	BOOL		checkF;		/* Saved value of Mfl_checkF */
	BOOL		mrelF;		/* If in the relevant modes area */
	char		mflmode;	/* 0 for C-like, 1 for sieve, -1 unk */
	UBYTE		ch;
	int		len;
	UBYTE		*strP;		/* Generic string pointer */
	PTNODE		*treeP;		/* Parse tree */
	VALUE		*valP;		/* VALUE pointer to use */
	VALUE		val;		/* Scratch value if needed */
	TOKEN		*pbtokP;	/* Saved token context head */

    /* Save current pushedback token state */
    pbtokP = mfl_tokback_save();

    /* Save Mfl_checkF and interpret any modes characters */
    checkF = Mfl_checkF;
    mflmode = -1;			/* Dunno the mfl mode yet */
    strP = msesP->es_strP;
    len = estr_len( msesP );
    mrelF = FALSE;

    while ( len-- > 0) {
	ch = *strP++;
	if ( !mrelF ) {
	    /* Just looking for the 'M' area */
	    mrelF = ( ch == 'M' );
	    continue;
	}

	/* Be done if the area is ended (no more lowercase letters) */
	if ( !islower( ch ) )
	    break;
	switch ( ch ) {
	    case 'c':			/* C-like mode */
		mflmode = 0;
		break;

	    case 'p':			/* Force parse-only mode */
		Mfl_checkF = TRUE;
		break;

	    case 's':			/* Start in sieve mode */
		mflmode = 1;
		break;

	    default:			/* Ignore anything unrecognized */
		break;
	}
    }

    /* If we didn't get an explicit mfl mode, consult a control */
    if ( ( mflmode < 0 ) )
	mflmode = cntrl_int_get( estr_tmp_str( "mfl_mode_sieve" ), 0 );

    /* Loop parsing and executing until the input is used up. */
    valP = ( rvalP != NULL ) ? rvalP : &val;
    for ( valF = FALSE; ; ) {
	if ( mflmode )
	    status = parse_sieve( inP, &treeP,
				 SBIS_ACTION | SBIS_CONTROL | SBIS_TEST );
	else
	    status = mfl_parse( inP, &treeP );

	if ( status == SCERR ) {
	    /* SCERR almost certainly means EOF.  (other causes of
	       SCERR trigger an exit before we get to see it..)
	    */
	    status = SCOK;
	    break;
	}

	/* Any non-OK status is done. */
	if ( status != SCOK )
	    break;

	if ( valF ) {
	    /* finish up any prior value */
	    val_finish( valP );
	    valF = FALSE;
	}

	/* If not checking, exec this tree */
	if ( !Mfl_checkF )
	    status = exec_treeval( treeP, valP, FALSE );
	if ( treeP != NULL )
	    parsetree_free( treeP );

	if ( status != SCOK )
	    break;
    }

    /* Restore Mfl_checkF value */
    Mfl_checkF = checkF;

    /* Restore pushed-back token state */
    mfl_tokback_restore( pbtokP );

    /* Finish value if not giving it back to caller */
    if ( valF && ( rvalP == NULL ) )
	val_finish( valP );

    return ( status );
}
/*

*//* mfl_run_str(rvalP, mflesP, msesP)

	Run mfl code that's in a string

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	mflesP		Ptr to string containing the MFL code
	msesP		Modes string

Returns :

	<value>		status code

Notes :

	Mostly just calls mfl_run_in().  Notes there apply here.

        When calling this with a multiline literal inside of C code
	(inside of an mfl-using application written in C, such as mvmda),
	it's a good idea to terminate each line with a '\n\' sequence.
	Some compilers require the backslash to continue a quoted string
	across a line break; and we want to be sure to have newlines in
	the mfl code so that error reporting will give a line number.
	Also line breaks are necessary to make mfl preprocessor recognize
	'@' on the beginning of a line.

	The MFL ESTR can be temporary; if it is, caller shouldn't attempt to
	  reference it after calling this function, since it's likely that
	  its temporary life will have been long exceeded while the code has
	  been run.  The strings pointed to by the ESTR must not be
	  temporary.

*/

STSCODE
mfl_run_str ARGLIST( (rvalP, mflesP, msesP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( ESTR		*mflesP )	/* String to run */
    FARG( ESTR		*msesP )	/* Modes string */
{
	STSCODE		status;
	MML_IN		*inP;		/* Input stack */

    /* Create input stack */
    if ( ( inP = in_new() ) == NULL ) {
	warning( "mfl_run_str: in_new() failed." );
	return ( SCERR );
    }

    /* Push the string */
    if ( !in_stream_push_str( inP,
			      mflesP->es_strP, estr_len( mflesP ),
			      "[internal]" ) ) {
	warning( "mfl_run_str: in_stream_push_str() failed." );
	in_destroy( inP );
	return ( SCERR );
    }

    /* mfl-interp the input stack */
    status = mfl_run_in( rvalP, inP, msesP );

    /* Remove the input stack */
    in_destroy( inP );

    /* Be done */
    return ( status );
}
