/* mfl-parse.c -- language parsing for mfl

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains language parsing code for mfl

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-parse.h"
#include "mfl-type.h"
#include "mfl-func.h"
#include "mfl-eval.h"
#include "mfl-sieve.h"

#include "errwarn.h"


/* Local definitions */



/* External data referenced */

extern	int	Align2;			/* 2-byte entity alignment */
extern	int	Align4;			/* >2-byte entity alignment */
extern	BISINFO	Bistbl[];		/* Builtin statement table */
extern	FUNCFARGS *FargsP;		/* Any function arguments.. */
extern	int	Mfl_nestC;		/* Nesting level */
extern	OPINFO	Optbl[];		/* Operator table */
extern	SYMTAB	*STglobal;		/* Global variables symtab */
extern	SYMTAB	*STstruct;		/* Structure names */
extern	SYMTAB	*STtname;		/* Type names */
extern	SYMTAB	*STtpool;		/* Type pool */
extern	TNSYM	*Tnint2;
extern	TNSYM	*Tnint4;


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	STSCODE	xsizeof PROTO( (MML_IN *inP, int prio, PTNODE *nodeP) );


/* Private data */

static	PTNODE	*PnodelistP = NULL;	/* Free parse tree nodes */

/*

*//* mfl_parse( inP, treePP )

	Parse (top level parse)

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the top tree node

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
mfl_parse ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to store the tree */
{
	STSCODE		status;
	int		tmod;		/* Type modifier code */
	TNSYM		*tsymP;		/* Typename symbol */
	TOKEN		*tokP;		/* Ptr to token */

    /* Get the first token and find out what it is */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCERR );

    /* Check for type definition */
    if ( ( tokP->tk_type == TK_SYM ) &&
         ( ( tsymP = (TNSYM *)symtab_lookup( STtname,
				tokP->tk_bstrP->bs_bP ) ) != NULL ) ) {
	tmod = tsymP->tn_flags & TF_MOD;
	if ( tmod == TMTYPEDEF )
	    status = parse_typedef( inP, treePP );
	else
	    status = parse_decl_stmt( inP, tsymP, treePP );
    }
    else if ( ( tokP->tk_type == TK_CHAR ) &&
	      ( tokP->tk_bstrP->bs_bP[0] == '{' ) ) {
        /* Parse compound statement */
	status = parse_compound( inP, treePP );
    }
    else if ( ( tokP->tk_type == TK_BIS ) &&
    	       Bistbl[tokP->tk_val.tv_int].bis_stmtF ) {
    	/* Built-in statement */
    	status = parse_bis( inP, treePP, &tokP->tk_val );
    }
    else {
	mfl_tokback( inP, tokP );
	status = parse_expr_stmt( inP, treePP );
    }

    if ( status == SCNOTDONE ) {
    	/* Error in parsing -- skip to next semicolon. */
	for( ; ; ) {
	    tokP = mfl_sigtoken( inP );
	    if ( tokP == NULL )
		break;
	    if ( ( tokP->tk_type == TK_CHAR ) &&
	         ( tokP->tk_bstrP->bs_bP[0] == ';' ) )
	    	break;
	}
	status = SCOK;
    }

    return( status );
}
/*

*//* mfl_parse_soc( inP, treePP )

	Parse single or compound statement

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the top tree node

Returns :

	<value>		SCxx status code

Notes :

	Called when expecting a single or compound statement, e.g.
	the body of a 'while' .  The point here is that a single
	statement must be terminated with a semicolon, whereas a
	compound statement is not.  The generic mfl_parse would see
	a compound statement as a part of an expression which must
	end with a semicolon.
*/

STSCODE
mfl_parse_soc ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to store the tree */
{
	STSCODE		status;
	TOKEN		*tokP;		/* Ptr to token */

    /* Get the first token to see if it's a compound statement. */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCERR );

    if ( ( tokP->tk_type == TK_CHAR ) &&
	      ( tokP->tk_bstrP->bs_bP[0] == '{' ) ) {
	status = parse_compound( inP, treePP );
    }
    else {
	/* Not compound -- push back and do statement parse. */
	mfl_tokback( inP, tokP );
	status = mfl_parse( inP, treePP );
    }

    return( status );
}
/*

*//* parse_bis( inP, treePP, tkvalP )

	Parse built in statement or function

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the top tree node
	tkvalP		Token value pointer
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_bis ARGLIST( (inP, treePP, tkvalP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to store tree */
    FARG( TOKVAL	*tkvalP )	/* Token value ptr */
{
	STSCODE		status;
	PTNODE		*nodeP;

    /* Make sure it is parseable */
    if ( Bistbl[tkvalP->tv_int].bis_parse == NULL ) {
    	mfl_warning( inP, "invalid use of \"%s\"",
		Bistbl[tkvalP->tv_int].bis_symP );
	status = SCNOTDONE;
    }
    else {
	/* Build the node to hold the built-in statement */
	status = parsetree_mknode( &nodeP, PN_BIS );
	if ( status == SCOK ) {
    	    nodeP->pt_data.pv_int = tkvalP->tv_int;

    	    /* Call the parsing routine to handle it */
	    status = (*Bistbl[tkvalP->tv_int].bis_parse)( inP, nodeP );

	    if ( status != SCOK ) {
		/* Failure -- get rid of node. */
		parsetree_free( nodeP );
	    }
	}
    }

    *treePP = ( status == SCOK ? nodeP : NULL );

    return( status );
}
/*

*//* parse_bis_break( inP, nodeP )

	Parse built in statement: "break"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
parse_bis_break ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
    /* Just need to end the statement */
    return ( parse_semi( inP ) );
}
/*

*//* parse_bis_c( inP, nodeP )

	Parse MFL builtin statement: C

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_BIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Parses a C-like statement group that must be enclosed
	in curly braces.    Attaches the parse tree to the
	left, like this:

			PN_BIS
			  /
			 /
			/
		  parse tree

*/

STSCODE
parse_bis_c ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Next token must be opening brace */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	mfl_warning( inP, "'{' required for C block, got \"%s\".",
		     tokP == NULL ? "<NUL>" :
			(char *)&tokP->tk_bstrP->bs_bP[0] );
	status = SCNOTDONE;
    }
    else {
	/* Parse the compound statement and attach it to the left. */
	status = parse_compound( inP, &nodeP->pt_leftP );
    }

    return ( status );
}
/*

*//* parse_bis_continue( inP, nodeP )

	Parse built in statement: "continue"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
parse_bis_continue ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
    /* Just need to end the statement */
    return ( parse_semi( inP ) );
}
/*

*//* parse_bis_do( inP, nodeP )

	Parse built in statement: "do"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :

	Things are attached just like "while" :

 		PN_BIS
		 /\
		/  \
	       /    \
    (condition)     (body of do)


*/

STSCODE
parse_bis_do ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Collect the body of the loop */
    status = mfl_parse_soc( inP, &nodeP->pt_rightP );

    if ( status == SCOK ) {
	/* Now need the word "while" */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
             ( tokP->tk_type != TK_BIS ) ||
	     ( strcmp( (char *)tokP->tk_bstrP->bs_bP, "while" ) != 0 ) ) {
	    mfl_warning( inP, "\"while\" expected." );
	    status = SCNOTDONE;
	}
	else {
	    /* Parse the test portion, complete with surrounding parens */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
                 ( tokP->tk_type != TK_OPR ) ||
	         ( tokP->tk_bstrP->bs_bP[0] != '(' ) ) {
		mfl_warning( inP, "parenthesized condition required." );
	        status = SCNOTDONE;
	    }
	    else {
		/* Parse the test into the left subtree of the BIS node */
		status = parse_expr_right( inP, 0, &nodeP->pt_leftP );
		if ( status == SCOK ) {
		    /* Make sure it ends with paren */
		    tokP = mfl_sigtoken( inP );
		    if ( ( tokP == NULL ) ||
	                 ( tokP->tk_type != TK_CHAR ) ||
			 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
			mfl_warning( inP, "unclosed parenthesized condition" );
			status = SCNOTDONE;
		    }
		    else {
			/* Need final semicolon */
			status = parse_semi( inP );
		    }
		}
	    }
	}
    }

    return( status );
}
/*

*//* parse_bis_for( inP, nodeP )

	Parse built in statement: "for"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


	 		PN_BIS
			 /\	            (i) == initialization clause
			/  \
		       /    \               (t) == test clause
		  PN_EXPR    PN_EXPR
		    /\        /\            (p) == post-loop close
		   /  \      /  \
		  /    \    /    \          (b) == loop body
		(i)    (t) (p)   (b)

	    (condition)     (body of for)

*/

STSCODE
parse_bis_for ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	PTNODE		*itreeP;	/* Initialization expr */
	PTNODE		*ttreeP;	/* Test expr */
	PTNODE		*ptreeP;	/* Post-loop expr */
	PTNODE		*btreeP;	/* Body expr */
	PTNODE		*lnodeP;	/* Left expr node */
	PTNODE		*rnodeP;	/* Right expr node */
	TOKEN		*tokP;		/* Token */

    itreeP = ttreeP = ptreeP = btreeP = NULL;

    /* Parse the paren-enclosed triple-expression control parms */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
	 ( tokP->tk_type != TK_OPR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '(' ) ) {
    	mfl_warning( inP, "parenthesized control info required." );
	status = SCNOTDONE;
    }
    else {
    	/* Parse the initial, test, and post expressions */
	if ( ( status = parse_expr_stmt( inP, &itreeP ) ) == SCOK )
	    if ( ( status = parse_expr_stmt( inP, &ttreeP ) ) == SCOK )
		if ( ( status = parse_expr_right( inP, 0, &ptreeP ) ) == SCOK ) {
		    /* Get trailing paren */
		    tokP = mfl_sigtoken( inP );
		    if ( ( tokP == NULL ) ||
		         ( tokP->tk_type != TK_CHAR ) ||
			 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
		    	mfl_warning( inP, "missing paren in FOR control." );
			status = SCNOTDONE;
		    }
		    else
		    	status = SCOK;
		}

	    
	if ( status == SCOK )
	    /* Get the body of the loop */
	    status = mfl_parse_soc( inP, &btreeP );

	if ( status == SCOK ) {
	    /* Allocate link nodes */
	    if ( ( status = parsetree_mknode( &lnodeP, PN_EXPR ) ) == SCOK )
	    	if ( ( status = parsetree_mknode( &rnodeP, PN_EXPR ) ) != SCOK )
		    parsetree_free( lnodeP );
	}

	if ( status == SCOK ) {
	    /* Link everything together */
	    lnodeP->pt_leftP = itreeP;
	    lnodeP->pt_rightP = ttreeP;
	    rnodeP->pt_leftP = ptreeP;
	    rnodeP->pt_rightP = btreeP;

	    nodeP->pt_leftP = lnodeP;
	    nodeP->pt_rightP = rnodeP;
	}
    }

    if ( status != SCOK ) {
    	/* Free up any trees that we allocated */
	if ( itreeP != NULL )
	    parsetree_free( itreeP );
	if ( ttreeP != NULL )
	    parsetree_free( ttreeP );
	if ( ptreeP != NULL )
	    parsetree_free( ptreeP );
	if ( btreeP != NULL )
	    parsetree_free( btreeP );
    }

    return( status );
}
/*

*//* parse_bis_if( inP, nodeP )

	Parse built in statement: "if"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


 		PN_BIS
		 /\
		/  \
	       /    \
    (condition)     PN_EXPR
    		      /\
		     /  \
		    /    \
	(true branch)    (false branch)

*/

STSCODE
parse_bis_if ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse the test portion.  Make sure it has parens. */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_OPR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '(' ) ) {
    	mfl_warning( inP, "parenthesized condition required." );
	status = SCNOTDONE;
    }
    else {
    	/* Parse the test into the left subtree of the BIS node */
	status = parse_expr_right( inP, 0, &nodeP->pt_leftP );
	if ( status == SCOK ) {
	    /* Make sure it ends with paren */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_CHAR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
	    	mfl_warning( inP, "unclosed parenthesized condition" );
		status = SCNOTDONE;
	    }
	}
    }

    if ( status == SCOK ) {
    	/* Make the PN_EXPR container for the IF body */
	status = parsetree_mknode( &nodeP->pt_rightP, PN_EXPR );
	if ( status == SCOK ) {
	    /* Shift focus to the EXPR node */
	    nodeP = nodeP->pt_rightP;

	    /* Collect the TRUE portion of the if body */
	    status = mfl_parse_soc( inP, &nodeP->pt_leftP );
	}
    }

    if ( status == SCOK ) {
    	/* Look now for an ELSE clause */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP != NULL ) &&
	     ( tokP->tk_type == TK_BIS ) &&
	     ( strcmp( (char *)tokP->tk_bstrP->bs_bP, "else" ) == 0 ) ) {
	    /* Get the false part */
	    status = mfl_parse_soc( inP, &nodeP->pt_rightP );
	}
	else {
	    /* No ELSE seen... push back this token */
	    mfl_tokback( inP, tokP );
	}
    }

    /* Done */
    return( status );
}
/*

*//* parse_bis_pv_( inP, nodeP )

	Parse built in statement: "pv$"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


 		PN_BIS
		 /\
		/  \
	       /    \
  (thing to print)  (format specification)


*/

STSCODE
parse_bis_pv_ ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Get first token */
    status = SCOK;
    tokP = mfl_sigtoken( inP );

    /* If it's a string, it's a format spec */
    if ( ( tokP != NULL ) && ( tokP->tk_type == TK_LITSTR ) ) {
    	/* Bind the format spec to the right side */
	status = parsetree_mktknode( &nodeP->pt_rightP, tokP );

	/* Check for optional comma */
	if ( status == SCOK ) {
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_OPR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ',' ) )
	    	mfl_tokback( inP, tokP );
	}
    }
    else
    	/* Not string -- push back token for evaluation */
	mfl_tokback( inP, tokP );
	
    /* Parse expression in left side */
    if ( status == SCOK )
	status = mfl_parse( inP, &nodeP->pt_leftP );

    return( status );
}
/*

*//* parse_bis_return( inP, nodeP )

	Parse built in statement: "return"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


 		PN_BIS
		 /
		/
	       /
          data (an expression tree)

*/

STSCODE
parse_bis_return ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;

    /* Parse the expression and link it to left of the BIS node */
    status = mfl_parse( inP, &nodeP->pt_leftP );

    return ( status );
}
/*

*//* parse_bis_sizeof( inP, nodeP )

	Parse built in statement: "sizeof"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


 		PN_BIS
		 /
		/
	       /
          data (an expression or a PN_TYPEREF)

*/

STSCODE
parse_bis_sizeof ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
    /* Hand off to recursive routine */
    return( xsizeof( inP, ORMEMBER, nodeP ) );
}
/*

*//* parse_bis_while( inP, nodeP )

	Parse built in statement: "while"

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :


 		PN_BIS
		 /\
		/  \
	       /    \
    (condition)     (body of while)


*/

STSCODE
parse_bis_while ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse the test portion.  As with IF, it must have parens. */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_OPR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '(' ) ) {
    	mfl_warning( inP, "parenthesized condition required." );
	status = SCNOTDONE;
    }
    else {
    	/* Parse the test into the left subtree of the BIS node */
	status = parse_expr_right( inP, 0, &nodeP->pt_leftP );
	if ( status == SCOK ) {
	    /* Make sure it ends with paren */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_CHAR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
	    	mfl_warning( inP, "unclosed parenthesized condition" );
		status = SCNOTDONE;
	    }
	}
    }

    if ( status == SCOK ) {
        /* Collect the body of the loop. */
	status = mfl_parse_soc( inP, &nodeP->pt_rightP );
    }

    return( status );
}
/*

*//* parse_compound( inP, treePP )

	Parse compound statement

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :

	A compound statement is a series of statements enclosed
	in curly braces.

 	The opening brace has already been parsed.  This builds a
	tree of PN_LSEQ (left sequential) nodes with a statement
	hanging on each one.

*/

STSCODE
parse_compound ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put tree node */
{
	STSCODE		status;
	PTNODE		*headP;
	PTNODE		*tailP;
	PTNODE		*nodeP;
	PTNODE		*seqP;
	TOKEN		*tokP;

    headP = tailP = NULL;

    ++Mfl_nestC;			/* Bump nesting level */

    /* Continue looking for end brace and parsing statements. */
    for( status = SCOK ; status == SCOK; ) {
	tokP = mfl_sigtoken( inP );
	if ( tokP == NULL ) {
	    mfl_warning( inP, "missing '}'." );
	    status = SCNOTDONE;
	    break;
	}
	if ( ( tokP->tk_type == TK_CHAR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == '}' ) )
	    break;

	/* Not at end.  Parse a new statement */
	mfl_tokback( inP, tokP );

	status = mfl_parse( inP, &nodeP );
	if ( status == SCOK ) {
	    /* Link the new node in */
	    if ( headP == NULL )
	    	/* Try to optimize one-statement groups */
		headP = nodeP;
	    else {
	        /* Make LSEQ block */
		status = parsetree_mknode( &seqP, PN_LSEQ );
		if ( status == SCOK ) {
	    	    /* See if head needs a LSEQ */
		    if ( tailP == NULL ) {
			/* Yes, it does. */
		    	seqP->pt_rightP = headP;
			headP = seqP;
		    }
		    else {
			/* Otherwise tail needs the LSEQ */
			seqP->pt_rightP = tailP->pt_leftP;
			tailP->pt_leftP = seqP;
		    }
		    tailP = seqP;

		    /* Terminate left-list with new node */
		    tailP->pt_leftP = nodeP;
		}
	    }
	}
    }

    /* If error, free any tree that we made */
    if ( ( status != SCOK ) && ( headP != NULL ) )
    	parsetree_free( headP );

    *treePP = ( status == SCOK ? headP : NULL );

    --Mfl_nestC;			/* One nesting gone */

    return( status );
}
/*

*//* parse_decl_stmt( inP, tsymP, treePP )

	Parse a statement that is a declaration (of a variable, etc)

Accepts :

	inP		Ptr to input stack
	tsymP		Symbol for type declaring keyword (e.g., "int")
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_decl_stmt ARGLIST( (inP, tsymP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TNSYM		*tsymP )	/* Type name symbol */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;
	TYPE		*btypeP;	/* Base type */
	TYPE		*typeP;		/* Type reference */
	ODINFO		info;		/* Any extended information */
	char		*vnameP;	/* Variable name ptr */
	PTNODE		*treeP;		/* ptr to header of tree */
	PTNODE		*vnodeP;	/* ptr to a variable node */
	PTNODE		*trnodeP;	/* ptr to a typeref node */
	TOKEN		*tokP;		/* Token ptr */
	PTNODE		**linkP;	/* Where link to next var is */

    /* Parse the base type */
    status = parse_type_base( inP, tsymP, &btypeP );

    /* Make the declaration evaluation tree header */
    parsetree_mknode( &treeP, PN_DECL );
    trnodeP = NULL;
    vnodeP = NULL;
    linkP = &treeP->pt_leftP;

    /* Get variable declarations and link them together at treeP */
    for( status = SCOK ; status == SCOK ; ) {
	/* Parse and internalize a type usage */
	status = parse_type_usage( inP, btypeP, &typeP, &vnameP, &info );

	/* Make sure there's a variable name, etc. */
	if ( ( status == SCOK ) && ( typeP != NULL ) ) {
	    if ( vnameP == NULL ) {
		trnodeP = NULL;
	    	/* Note that it is not an error for there to be no variable
	           name to declare.  That is, for instance, a way of declaring
	           an aggregate without an instantiation.  It *is* an error
	           if there was a comma without a symbol name. */
		if ( vnodeP != NULL )
		    mfl_warning( inP, "no declaration after comma." );
	    }
	    else {
	    	/* Link this into the tree */
		parsetree_mknode( &trnodeP, PN_TYPEREF );
		trnodeP->pt_info.od_P = (void *)typeP;
		trnodeP->pt_info.od_del = NULL;

		/* Store any extended info */
		trnodeP->pt_xinfo = info;

		parsetree_mknode( &vnodeP, PN_SYM );
		vnodeP->pt_rightP = trnodeP;
		vnodeP->pt_info.od_P = (void *)vnameP;

		*linkP = vnodeP;
		linkP = &vnodeP->pt_leftP;
	    }

	    /* Check for initializer */
	    if ( ( status == SCOK ) && ( trnodeP != NULL ) )
		status = parse_decl_value( inP, trnodeP );

	    /* Get the following token to look at. */
	    tokP = mfl_sigtoken( inP );

	    /* Check for semicolon or comma */
	    if ( ( tokP != NULL ) &&
	         ( tokP->tk_type == TK_CHAR ) &&
		 ( tokP->tk_bstrP->bs_bP[0] == ';' ) )
	    	break;
	    else if ( ( tokP == NULL ) ||
		      ( tokP->tk_type != TK_OPR ) ||
		      ( tokP->tk_bstrP->bs_bP[0] != ',' ) ) {
		mfl_tokback( inP, tokP );
	    	mfl_warning( inP, "syntax error" );
		status = SCNOTDONE;
	    }

	    /* Comma: make sure something came before it. */
	    if ( vnodeP == NULL ) {
		mfl_warning( inP, "no declaration before comma." );
		status = SCNOTDONE;
	    }
	}
	else if ( status == SCOK ) {
	    /* no type? */
	    mfl_warning( inP, "internal error (no type!) in parse_decl_stmt." );
	    status = SCNOTDONE;
	}
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( treeP != NULL ) )
    	parsetree_free( treeP );

    /* Return the tree if any */
    *treePP = ( status == SCOK ? treeP : NULL );

    return( status );
}
/*

*//* parse_decl_value( inP, trnodeP )

	Parse any declaration initializer

Accepts :

	inP		The input block
	trnodeP		Ptr to the PN_TYPEREF node we're building

Returns :

	<value>		status code

Notes :


*/

STSCODE
parse_decl_value ARGLIST( (inP, trnodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*trnodeP )	/* PN_TYPEREF node */
{
	STSCODE		status;
	TYPE		*typeP;

    /* Have to deal with annoying initializer syntax inconsistencies.
       A scalar initializer can *optionally* be enclosed in braces
       if it's at the top level, but not if that scalar initializer
       is inside some kind of compound (aggregate).  So scalar types
       are handled one way, other types another.
    */
    typeP = (TYPE *)trnodeP->pt_info.od_P;
    switch (typeP->ty_class) {
	case	TC_UINT:		/* Scalar types */
	case	TC_INT:
	case	TC_FLOAT:
	case	TC_ENUM:
	case	TC_PTR:
	case	TC_STRING:
	    status = parse_decl_value_scalar( inP, trnodeP );
	    break;

	case	TC_FUNCT:		/* Function */
	    status = parse_func_value( inP, trnodeP );
	    break;

	default:			/* Anything else... */
	    status = parse_decl_value_aggr( inP, trnodeP );
	    break;
    }

    return ( status );
}
/*

*//* parse_decl_value_aggr( inP, trnodeP )

	Parse an optional declaration initializer for an aggregate

Accepts :

	inP		The input block
	trnodeP		Ptr to the PN_TYPEREF node we're building

Returns :

	<value>		status code

Notes :

	An initializer for an aggregate is an equal sign ('=')
	followed by an appropriate list of one or more values
	enclosed in curly braces.  The list may contain sublists
	as per parse_decl_value_list().  See that routine for
	the parse tree format.

*/

STSCODE
parse_decl_value_aggr ARGLIST( (inP, trnodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*trnodeP )	/* PN_TYPEREF node */
{
	STSCODE		status;
	TOKEN		*tokP;
	PTNODE		*lnodeP;	/* Ptr to list node being made */

    /* Check for '=' */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCOK );
    if ( ( tokP->tk_type != TK_OPR ) ||
         ( tokP->tk_bstrP->bs_bP[0] != '=' ) ||
	 ( tokP->tk_bstrP->bs_bP[1] != NUL ) ) {
	mfl_tokback( inP, tokP );
	return ( SCOK );
    }

    /* Check for the start of the list */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
         ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	/* No list: syntax error. */
	mfl_warning( inP, "missing '{' for initializer." );
	lnodeP = NULL;
	status = SCNOTDONE;
    }
    else {
	/* Use parse_decl_value_list to get the initializers. */
	status = parse_decl_value_list( inP, &lnodeP );
    }

    if ( ( status != SCOK ) && ( lnodeP != NULL ) ) {
	parsetree_free( lnodeP );
	lnodeP = NULL;
    }

    trnodeP->pt_rightP = lnodeP;

    return ( status );
}
/*

*//* parse_decl_value_list( inP, treePP )

	Parse declaration initializer expression list(s)

Accepts :

	inP		The input block
	treePP		Ptr to the list pointer variable

Returns :

	<value>		status code

Notes :

	This is a subcomponent of parse_decl_value routines.  It is
	called to parse a list of initializer values inside curly
	braces.  Sublists are allowed: i.e. if an opening curly bracket
	is encountered, it recursively calls itself.

	The opening curly bracket is assumed to have already
	been parsed.

	This could probably be merged into parse_expr_list but for
	now I want to leave open the open of making it more
	special.

	Each value list is represented by a left-sequence list; that
	is, each node is a PN_LSEQ node with its right branch pointing
	to the value and its left branch pointing to any next list
	element.  A sublist is in term represented by that same kind
	of list.  So:

		         PN_LSEQ
			   /\
			  /  \
			 /   first expression
		        /
		     PN_LSEQ
		       /\
		      /  \
		     /   second expression
		    /
		  ...
		  /  
		last expression


*/

STSCODE
parse_decl_value_list ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put the result tree */
{
	STSCODE		status;
	TOKEN		*tokP;
	PTNODE		*hnodeP;	/* Ptr to list header node */
	PTNODE		*lnodeP;	/* Ptr to last LSEQ node we made */
	PTNODE		*nodeP;		/* Ptr to LSEQ node under construction */

    hnodeP = lnodeP = NULL;

    /* Loop through the elements of the list. */
    for ( status = SCOK; status == SCOK; ) {

	/* Peek at the next token */
	if ( ( tokP = mfl_sigtoken( inP ) ) == NULL )
	    break;

	/* If it's a new sublist, get that */
	if ( ( tokP->tk_type == TK_CHAR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == '{' ) ) {

	    /* Make new LSEQ node to hold this term. */
	    status = parsetree_mknode( &nodeP, PN_LSEQ );
	    if ( status == SCOK ) {
		/* Recurse to get the sublist. */
		status = parse_decl_value_list( inP, &nodeP->pt_rightP );
	    }
	}

	else {
	    /* Generic term.  First push back this token. */
	    mfl_tokback( inP, tokP );

	    /* Make new LSEQ node to hold this term. */
	    status = parsetree_mknode( &nodeP, PN_LSEQ );
	    if ( status == SCOK ) {
		status = parse_expr_right( inP, ORCOMMA, &nodeP->pt_rightP );
		if ( status != SCOK )
		    parsetree_free( nodeP );
	    }
	}

	if ( status == SCOK ) {
	    /* Our new LSEQ term is in nodeP.  Link it into the
	       last node we made, if any-- otherwise remember it
	       as the list head.
	    */
	    if ( lnodeP != NULL )
		lnodeP->pt_leftP = nodeP;
	    else
		hnodeP = nodeP;

	    lnodeP = nodeP;
	}

	/* There might now be a comma and another term.  Otherwise
	   we're done.
	*/
	if ( status == SCOK ) {
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_OPR ) ||
	         ( tokP->tk_bstrP->bs_bP[0] != ',' ) )
		break;
	}
    }

    /* Verify that we have the closing brace */
    if ( ( status == SCOK ) &&
         ( ( tokP == NULL ) ||
	   ( tokP->tk_type != TK_CHAR ) ||
	   ( tokP->tk_bstrP->bs_bP[0] != '}' ) ) ) {
	mfl_warning( inP, "unterminated initializer list." );
	status = SCNOTDONE;
    }

    /* Check for null list */
    if ( ( status == SCOK ) && ( hnodeP == NULL ) ) {
	mfl_warning( inP, "invalid empty initializer list." );
	status = SCNOTDONE;
    }

    if ( ( status != SCOK ) && ( hnodeP != NULL ) ) {
	parsetree_free( hnodeP );
	hnodeP = NULL;
    }
    *treePP = hnodeP;
    return ( status );
}
/*

*//* parse_decl_value_scalar( inP, trnodeP )

	Parse an optional declaration initializer for a scalar

Accepts :

	inP		The input block
	trnodeP		Ptr to the PN_TYPEREF node we're building

Returns :

	<value>		status code

Notes :

	An initializer for a scalar is an equal sign ('=') followed
	by an expression optionally enclosed in curly braces.

*/

STSCODE
parse_decl_value_scalar ARGLIST( (inP, trnodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*trnodeP )	/* PN_TYPEREF node */
{
	STSCODE		status;
	TOKEN		*tokP;
	PTNODE		*lnodeP;	/* Ptr to list node being made */

    /* Check for '=' */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCOK );
    if ( ( tokP->tk_type != TK_OPR ) ||
         ( tokP->tk_bstrP->bs_bP[0] != '=' ) ||
	 ( tokP->tk_bstrP->bs_bP[1] != NUL ) ) {
	mfl_tokback( inP, tokP );
	return ( SCOK );
    }

    /* Check for optional curly braces */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL ) {
	mfl_warning( inP, "missing value for initializer." );
	lnodeP = NULL;
	status = SCNOTDONE;
    }
    else if ( ( tokP->tk_type == TK_CHAR ) &&
	      ( tokP->tk_bstrP->bs_bP[0] == '{' ) ) {

	/* The ORCOMMA priority will cause parsing to stop at a comma.
	   An expression containing commas would return only the last
	   value and would be valid, but would be confusing here if
	   allowed, so we don't allow it. 
	*/
	status = parse_expr_right( inP, ORCOMMA, &lnodeP );

	/* Get trailing '}' */
	if ( ( status == SCOK ) &&
	     ( ( ( tokP = mfl_sigtoken( inP ) ) == NULL ) ||
	       ( tokP->tk_type != TK_CHAR ) ||
	       ( tokP->tk_bstrP->bs_bP[0] != '}' ) ) ) {
	    mfl_warning( inP, "missing '}' in initializer." );
	    status = SCNOTDONE;
	}
    }
    else {
	/* No opening brace: just get the expression.  As per the
	   comment above, stop at a comma.  Particularly in this
	   case as a comma might introduce another variable
	   declaration.
	*/
	mfl_tokback( inP, tokP );
	status = parse_expr_right( inP, ORCOMMA, &lnodeP );
    }

    if ( ( status != SCOK ) && ( lnodeP != NULL ) ) {
	parsetree_free( lnodeP );
	lnodeP = NULL;
    }

    if ( ( status == SCOK ) && ( lnodeP == NULL ) ) {
	mfl_warning( inP, "invalid null initial value." );
	status = SCNOTDONE;
    }

    trnodeP->pt_rightP = lnodeP;

    return ( status );
}
/*

*//* parse_expr_cond( inP, treePP )

	Parse the two branches of a conditional expression

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :

	Test expression has already been parsed.. this routine
	deals with the   a:b  portion

*/

STSCODE
parse_expr_cond ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;
	PTNODE		*nodeP;
	PTNODE		*leftP;		/* Left side */
	PTNODE		*rightP;	/* Right side */
	TOKEN		*tokP;		/* Token */

    leftP = rightP = NULL;

    /* Get the left sub-expression */
    status = parse_expr_right( inP, 0, &leftP );
    if ( status == SCOK ) {
	/* Make sure the next thing is ':' */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( tokP->tk_type != TK_CHAR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != ':' ) ) {
	    mfl_warning( inP, "missing ':' in conditional expression." );
	    status = SCNOTDONE;
	}
	else {
	    /* Get right part */
	    status = parse_expr_right( inP, ORCOND, &rightP );
	    if ( status == SCOK )
		/* Make binary EXPR node for the reduced cond. opr. term */
		status = parsetree_mknode( &nodeP, PN_EXPR );
	    if ( status == SCOK ) {
		/* Assemble the term */
		nodeP->pt_leftP = leftP;
		nodeP->pt_rightP = rightP;
	    }
	}
    }

    if ( status == SCOK )
	*treePP = nodeP;
    else {
	/* Clean up allocated trees */
	if ( leftP != NULL )
	    parsetree_free( leftP );
	if ( rightP != NULL )
	    parsetree_free( rightP );
    }

    return( status );
}
/*

*//* parse_expr_list( inP, treePP )

	Parse one or more expressions separated by commas

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :

	"treePP" provides a pointer to return the resultant parse
	tree into.  If there's only a single expression in the
	expression "list", we just return a pointer to it:

		treePP -> PN_whatever

	If there are multiple expressions, we return a pointer to an LSEQ
	list, like this:

		treePP->  LSEQ
			   /\
			  /  \
			 /   first expression
		        /
		      LSEQ
		       /\
		      /  \
		     /   second expression
		    /
		  ...
		  /  
		last expression

	This looks a lot like a compound statement.

*/

STSCODE
parse_expr_list ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;
	PTNODE		*headP;		/* Ptr to the entire list, we'll
					    return this. */
	PTNODE		*nodeP;		/* New node being made */
	PTNODE		*seqP;		/* Ptr to an LSEQ node */
	PTNODE		**tailPP;	/* Any pointer to the end node */
	TOKEN		*tokP;

    headP = NULL;			/* No list yet */
    tailPP = NULL;			/* Nothing points to the end */

    /* Loop gathering statements until we reach an end brace */
    for ( status = SCOK; status == SCOK; ) {
	status = parse_expr_right( inP, ORCOMMA, &nodeP );

	if ( status != SCOK )
	    break;

	/* Avoid making an LSEQ for the last node in this block.
	   We do this by remembering the linkage around the last
	   parsed node and relinking if there turn out to be more
	   statements after that one.
	*/
	if ( headP == NULL ) {
	    /* This is the first statement, just remember it */
	    headP = nodeP;
	    tailPP = &headP;		/* Pointer to last node ptr */
	}
	else {
	    /* This is not the first statement, which means that the
	       last one we parsed is hanging at the end and needs to
	       be replaced with an LSEQ.
	    */
	    status = parsetree_mknode( &seqP, PN_LSEQ );
	    if ( status == SCOK ) {
		/* Unlink the tail node from the end and replace
		   it with the lseq sequence
		*/
		seqP->pt_rightP = *tailPP;
		*tailPP = seqP;

		/* Point the LSEQ to the new tail, and remember the
		   tail pointer.
		*/
		seqP->pt_leftP = nodeP;
		tailPP = &(seqP->pt_leftP);
	    }
	}

	/* Comma continues, non-comma ends */
	if ( status == SCOK ) {
	    tokP = mfl_sigtoken( inP );

	    if ( tokP == NULL )
		break;

	    if ( ( tokP->tk_type != TK_OPR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ',' ) ) {
		mfl_tokback( inP, tokP );
		break;
	    }

	    /* continue to next expression in list */
	}
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( headP != NULL ) )
    	parsetree_free( headP );

    /* Return the tree if any */
    *treePP = ( status == SCOK ? headP : NULL );

    return( status );
}
/*

*//* parse_expr_preop(inP, lprio, tokP, treePP)

	Parse prefix operator in an expresion

Accepts :

	inP		Ptr to input stack
	lprio		Priority of any operator to the left of this
	tokP		The prefix operator token already parsed
	treePP		Where to put the result
	
Returns :

	<value>		SCxx status code

Notes :

	Caller has already recognized the prefix operator token.

	Remember that tokP generally has a limited lifetime.  Caller
	can not expect it to be preserved.

*/

STSCODE
parse_expr_preop ARGLIST( (inP, lprio, tokP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( int		lprio )		/* Left operator priority */
   NFARG( TOKEN		*tokP )		/* Token pointer */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;
	TNSYM		*tsymP;		/* type name symbol ptr */
	PTNODE		*nodeP;

    status = SCOK;
    nodeP = NULL;

    /* Check for valid prefix operator */
    if ( *tokP->tk_bstrP->bs_bP == '(' ) {
    	/* Do left parenthesis magic */

	/* Must check for typecast */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP != NULL ) &&
	     ( tokP->tk_type == TK_SYM ) &&
	     ( ( tsymP = (TNSYM *)symtab_lookup( STtname,
	     		   tokP->tk_bstrP->bs_bP ) ) != NULL ) ) {
	    status = parse_typecast( inP, &nodeP, tsymP );
	}
	else {
	    /* Not typecast */
	    mfl_tokback( inP, tokP );

	    /* Get the enclosed expression */
	    status = parse_expr_right( inP, 0, &nodeP );

	    if ( status == SCOK ) {
		/* Check right paren */
		tokP = mfl_sigtoken( inP );
		if ( ( tokP == NULL ) ||
		   ( tokP->tk_type != TK_CHAR ) ||
		   ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
		    mfl_warning( inP, "unmatched left parenthesis" );
		    mfl_tokback( inP, tokP );
		    status = SCNOTDONE;
		}
	    }
	}
    }
    else {
        /* Not paren -- see if this can be a prefix operator */
	if ( Optbl[tokP->tk_val.tv_int].op_preeval == NULL )  {
    	    mfl_warning( inP, "invalid prefix operator." );
	    status = SCNOTDONE;
	}
	else {
            /* Prefix operator it is.. stack it and gather its argument */
	    parsetree_mktknode( &nodeP, tokP );
	    status = parse_expr_right( inP, ORPREFIX, &nodeP->pt_rightP );
	}
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( nodeP != NULL ) )
    	parsetree_free( nodeP );

    /* Return the tree if any */
    *treePP = ( status == SCOK ? nodeP : NULL );

    return( status );
}
/*

*//* parse_expr_right( inP, lprio, treePP )

	Parse the right side of an expression

Accepts :

	inP		Ptr to input stack
	lprio		Priority of the operator on the left of the
			 subexpression that is about to be parsed.
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :

	The relative operator priorities determines how the tree
	is constructed- higher priorities bind deeper (i.e. execute
	first).

*/

STSCODE
parse_expr_right ARGLIST( (inP, lprio, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( int		lprio )		/* Left operator priority */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;
	OPINFO		*opP;		/* Operator ptr */
	TOKEN		*tokP;		/* Token */
	TOKEN		*optokP;	/* Operator token */

	PTNODE		*leftP;		/* Ptr to left node */
	PTNODE		*rightP;	/* Ptr to right node */
	PTNODE		*oprP;		/* Ptr to operator node. */

    *treePP = NULL;

    /* Get the left side of the right-hand subexpression.  This will
       be the next token... */
    tokP = mfl_sigtoken( inP );

    if ( tokP == NULL )
	status = SCERR;

    /* If the first thing is a prefix operator, handle it specially. */
    else if ( tokP->tk_type == TK_OPR ) {
    	/* Do prefix operator */
        status = parse_expr_preop( inP, lprio, tokP, &leftP );
	if ( status != SCOK )
	    return( status );
    }
    else if ( tokP->tk_type == TK_CHAR ) {
	/* If it's a curly brace, take a compound statement here. */
	if ( tokP->tk_bstrP->bs_bP[0] == '{' )
	    status = parse_compound( inP, &leftP );
	else {
	    /* Random character terminates the expression */
	    in_charback( inP, tokP->tk_bstrP->bs_bP[0] );
	    return( SCOK );
	}
    }
    else if ( tokP->tk_type == TK_BIS ) {
	/* Built-in statement; take it if it's allowed as a term. */
	if ( Bistbl[tokP->tk_val.tv_int].bis_termF )
	    status = parse_bis( inP, &leftP, &tokP->tk_val );
	else {
	    mfl_warning( inP, "\"%s\" not allowed as an expression term.",
			 Bistbl[tokP->tk_val.tv_int].bis_symP );
	    status = SCNOTDONE;
	}
    }
    else if ( ( tokP->tk_type == TK_SYM ) ) {
	/* Check for a function call. */
	optokP = mfl_sigtoken( inP );
	if ( ( optokP != NULL ) &&
	     ( optokP->tk_type == TK_OPR ) &&
	     ( optokP->tk_bstrP->bs_bP[0] == '(' ) ) {
	    /* Parse the function call. */
	    status = parse_func_call( inP, tokP, &leftP );
	}
	else {
	    /* Not a function call... */
	    mfl_tokback( inP, optokP );
	    status = parsetree_mktknode( &leftP, tokP );
	}
    }
    else
    	/* First thing is an operand. */
	status = parsetree_mktknode( &leftP, tokP );

    /* Eat up any operators that come next */
    for( ; status == SCOK; ) {
	/* Get the operator that is presumed to be next */
	optokP = mfl_sigtoken( inP );
	if ( ( optokP == NULL ) || ( optokP->tk_type != TK_OPR ) ) {
	    mfl_tokback( inP, optokP );
	    break;
	}
	opP = &Optbl[optokP->tk_val.tv_int];

	/* Check for postfix operator; postfix operator will always
	   bind highest at this point, so we'll do this before the
	   operator priority check.
	*/
	if ( opP->op_posteval != NULL ) {
	    /* make a new left node out of this operator */
	    status = parsetree_mktknode( &oprP, optokP );
	    if ( status == SCOK ) {
	    	oprP->pt_leftP = leftP;
		leftP = oprP;
#if 0
		lprio = opP->op_prio;	/* bugfix/mm */
#endif
	    }

	    /* Continue looking for operators. */
	    continue;
	}

	/* Done at this level if previous grouping should be honored
	   because of finding a lower precedence operator, or in the
	   case of left-associative operators (everything except
	   assignment and conditional expressions), 
	   of finding an equal precedence operator. */
	if ( ( opP->op_prio < lprio ) ||
	     ( ( opP->op_prio != ORASSIGN ) &&
	       ( optokP->tk_bstrP->bs_bP[0] != '?' ) &&
	       ( opP->op_prio == lprio ) )
	     ) {
	    /* Can't use this term */
	    mfl_tokback( inP, optokP );
	    break;
	}

	/* Found an operator -- get the right side now.  First
	   reserve the token so it doesn't get reused while we
	   hang on to it
	*/
	mfl_tokreserve( optokP );

    if ( optokP->tk_bstrP->bs_bP[0] == '[' ) {	/* Array ref? */
	    status = parse_expr_right( inP, 0, &rightP );
	    if ( status == SCOK ) {
	    	/* Look for terminating ']' */
		tokP = mfl_sigtoken( inP );
		if ( ( tokP == NULL ) ||
		     ( tokP->tk_type != TK_CHAR ) ||
		     ( tokP->tk_bstrP->bs_bP[0] != ']' ) ) {
		    mfl_warning( inP, "missing ']'." );
		    mfl_tokback( inP, tokP );
		    status = SCNOTDONE;
		}
		else
		    tokP->tk_type = TK_OPR;
	    }
	}
	else if ( optokP->tk_bstrP->bs_bP[0] == '?' ) {	/* Conditional operator? */
	    /* Reduce the b:c part of the conditional expression. */
	    status = parse_expr_cond( inP, &rightP );
	    /* Get the right side as a conditional expression term,
	       verifying that the result is a ":" operator. */
	}
	else				/* Not special */
	    status = parse_expr_right( inP, opP->op_prio, &rightP );

	/* Now we can enable release of the operator token */
	mfl_tokrelease( optokP );

	if ( status == SCOK ) {
	    /* Make sure right term wasn't a null expression */
	    if ( rightP == NULL ) {
		mfl_warning( inP, "\"%s\" is not a postfix operator.",
			     optokP->tk_bstrP->bs_bP );
		status = SCNOTDONE;
	    }
	    else {
		/* Got operator and left and right terms --
	           replace left term with binary operator tree */
		status = parsetree_mktknode( &oprP, optokP );
		if ( status == SCOK ) {
	            oprP->pt_leftP = leftP;
		    oprP->pt_rightP = rightP;
		    leftP = oprP;
		}
	    }
	}
    }

    /* Skipped all operators that can be applied at this priority */
    *treePP = leftP;

    return( status );
}
/*

*//* parse_expr_stmt( inP, treePP )

	Parse a C statement that is an expression

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_expr_stmt ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put the tree */
{
	STSCODE		status;

    status = parse_expr_right( inP, 0, treePP );
    if ( status == SCOK )
	status = parse_semi( inP );

    return( status );
}
/*

*//* parse_semi( inP )

	Parse the semicolon that is expected next

Accepts :

	inP		Ptr to input stack
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_semi ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	TOKEN		*tokP;

    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL ) {
	mfl_warning( inP, "semicolon expected, but no input present." );
	return ( SCNOTDONE );
    }
    else if ( ( tokP->tk_type != TK_CHAR ) ||
	      ( tokP->tk_bstrP->bs_bP[0] != ';' ) ) {
	mfl_warning( inP, "semicolon expected, got type %s, \"%s\"",
		     mfl_toktype( tokP->tk_type ),
		     tokP->tk_rawF ?
		         tokP->tk_rawP->bs_bP : tokP->tk_bstrP->bs_bP );
	return ( SCNOTDONE );
    }

    return ( SCOK );
}
/*

*//* parse_type_agg( inP, typePP, tclass )

	Parse type: aggregate

Accepts :

	inP		Ptr to input stack
	typePP		Where to put the type
	tclass		Type class
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_type_agg ARGLIST( (inP, typePP, tclass) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TYPE		**typePP )	/* Where to put type */
    FARG( TYPECLASS	tclass )	/* Type class */
{
	STSCODE		status;		/* Status code */
	int		size;		/* Aggregate size */
	int		align;		/* Alignment */
	int		memC;		/* Number of members */
	BOOL		newF;		/* If new */
	PTNODE		*treeP;		/* ptr to tree */
	PTNODE		*nodeP;		/* Ptr to last node of tree */
	PTNODE		*snodeP;	/* Sym node ptr */
	PTNODE		**linkP;	/* Where to link to */
	TYPE		*typeP;		/* Ptr to type */
	TYPE		*etypeP;	/* Element type */
	TNSYM		*tsymP;		/* Ptr to type symbol */
	TYPEAG		*agP;		/* Aggregate ptr */
	AGGMEM		*memP;		/* Aggregate member ptr */
	TOKEN		*tokP;		/* Token */

    /* Start with an empty tree */
    treeP = NULL;

    /* parse the opening part */
    status = parse_type_aggname( inP, &typeP, &newF, tclass );

    /* Check for definition of new aggregate */
    if ( ( status == SCOK ) && newF ) {
	/* Raise the nesting level */
	++Mfl_nestC;

    	/* Collect the members of the aggregate */
	for( linkP = &treeP; status == SCOK; linkP = &nodeP->pt_rightP ) {
	    /* Get next type token */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP != NULL ) &&
	         ( tokP->tk_type == TK_CHAR ) &&
		 ( tokP->tk_bstrP->bs_bP[0] == '}' ) )
	    	break;
	    if ( ( tokP != NULL ) &&
	         ( tokP->tk_type == TK_SYM ) &&
	    	 ( ( tsymP = (TNSYM *)symtab_lookup( STtname,
		 			&tokP->tk_bstrP->bs_bP[0] ) ) != NULL ) ) {
		/* Found a type symbol */
		status = parse_decl_stmt( inP, tsymP, linkP );
		if ( status == SCOK )
		    nodeP = *linkP;
	    }
	    else {
	    	/* Not a type symbol */
		mfl_tokback( inP, tokP );
		mfl_warning( inP, "syntax error." );
		status = SCNOTDONE;
	    }
	}

	if ( status == SCOK ) {
	    /* Aggregate all parsed. */
	    for( memC = 0, nodeP = treeP; nodeP != NULL;
	    		nodeP = nodeP->pt_rightP )
		for( snodeP = nodeP->pt_leftP; snodeP != NULL;
			snodeP = snodeP->pt_leftP, ++memC )
		    ;

	    /* Make the aggregate array */
	    agP = (TYPEAG *)typeP->ty_nextP;
	    agP->ag_memC = memC;
	    agP->ag_memP = memP = (AGGMEM *)emalloc(
				"parse_type_agg",
				"aggregate member array",
				memC * sizeof(AGGMEM) );

	    for( size = 0, nodeP = treeP; nodeP != NULL;
	    		nodeP = nodeP->pt_rightP )
	    	for( snodeP = nodeP->pt_leftP; snodeP != NULL;
			snodeP = snodeP->pt_leftP, ++memP ) {
		    /* Steal the symbol name pointer from sym node */
		    memP->am_nameP = (char *)snodeP->pt_info.od_P;
		    snodeP->pt_info.od_P = NULL;

		    memP->am_typeP = (TYPE *)(snodeP->pt_rightP->pt_info.od_P);

		    if ( tclass == TC_STRUCT ) {
		        /* Do non-char alignment -- first find base type */
			for( etypeP = memP->am_typeP; ; )
			    if ( etypeP->ty_class != TC_DIM )
			        break;
			    else
			        etypeP = (TYPE *)etypeP->ty_nextP;

			if ( etypeP->ty_size != 1 ) {
			    /* Round up current offset */
			    if ( etypeP->ty_size == 2 )
			    	align = Align2;
			    else
			    	align = Align4;
			    if ( ( size % align ) != 0 ) 
			        size += (align - (size % align));
			}

			memP->am_offs = size;
		    	size += (memP->am_typeP)->ty_size;
		    }
		    else {
		    	memP->am_offs = 0;
		    	if ( (memP->am_typeP)->ty_size > size )
			    size = memP->am_typeP->ty_size;
		    }
		}

	    /* Round up the aggregate size */
	    if ( ( size % Align4 ) != 0 )
	    	size += ( Align4 - ( size % Align4 ) );

	    typeP->ty_size = agP->ag_size = size;
	    typeP->ty_flags |= TF_DEFINED;	    
	}

	parsetree_free( treeP );

	/* Out of this nest */
	--Mfl_nestC;
    }

    *typePP = typeP;

    return( status );
}
/*

*//* parse_type_aggname( inP, typePP, newFP, aclass )

	Parse an aggregate's name or lead-in

Accepts :

	inP		Ptr to input stack handle
	typePP		Where to put the type ptr
	newFP		Where to put "new definition" flag
	aclass		Type class of aggregate

Returns :

	<value>		SCxx status code

Notes :

 	This routine will either parse a reference to an aggregate
	by name (whether or not it exists) or the lead-in to a
	definition of the aggregate.

*/

STSCODE
parse_type_aggname ARGLIST( (inP, typePP, newFP, aclass) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TYPE		**typePP )	/* Where to put type */
   NFARG( BOOL		*newFP )	/* Where to put "new" flag */
    FARG( TYPECLASS	aclass )	/* Type class for aggregate */
{
	STSCODE		status;
	TYPE		*typeP;
	TNSYM		*tnsymP;	/* aggregate table entry ptr */
	TYPSYM		*typsymP;	/* Ptr to type symbol */
	TOKEN		*tokP;		/* Ptr to token */
	TOKEN		*ntokP;		/* Ptr to token containing the name */

    status = SCOK;

    /* See what comes next.  Should be either a name or "{" */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP != NULL ) && ( tokP->tk_type == TK_SYM ) ) {
    	/* It's an aggregate name.  Remember it.  Don't need to reserve
	   the token since it's a short lifetime.
	*/
	ntokP = tokP;

	/* Now try for the "{" */
	tokP = mfl_sigtoken( inP );
    }
    else
	ntokP = NULL;

    /* Check non-definition (reference to existing or future def) */
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
        *newFP = FALSE;
	mfl_tokback( inP, tokP );
    }
    else
	*newFP = TRUE;

    /* If a name given, check it out */
    if ( ntokP != NULL ) {
	/* Find or make the symbol for this aggregate name */
	tnsymP = type_aggname( aclass, (char *)ntokP->tk_bstrP->bs_bP );
	if ( tnsymP == NULL )
	    status = SCERR;
	else {
	    /* Make sure type class matches... we may have looked up a
	       previously defined aggregate name.
	    */
	    typeP = tnsymP->tn_typeP;
	    if ( typeP->ty_class != aclass ) {
		mfl_warning( inP, "wrong aggregate type." );
		mfl_warning( inP, "wanted %d; is %d", aclass, typeP->ty_class );
		status = SCNOTDONE;
	    }
	    else {
		/* Make sure the name isn't being redefined */
		if ( *newFP && ( ( typeP->ty_flags & TF_DEFINED ) != 0 ) ) {
		    mfl_warning( inP, "\"%s\" already defined -- ignoring name.",
					ntokP->tk_bstrP->bs_bP );
		    ntokP = NULL;
		}
	    }
	}
    }

    if ( ( status == SCOK ) && ( ntokP == NULL ) ) {
	/* No name. */
	if ( !(*newFP ) ) {
	    mfl_warning( inP, "Invalid aggregate reference." );
	    status = SCNOTDONE;
	}
	else {
	    /* Make new aggregate */
	    typsymP = type_mkagg( aclass );
	    typeP = &(typsymP->ts_type);
	}
    }

    *typePP = typeP;

    return( status );
}
/*

*//* parse_type_base( inP, tsymP, typePP )

	Parse base type definition

Accepts :

	inP		Ptr to input stack
	tsymP		Symbol introducing the type
	typePP		Where to put the type
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_type_base ARGLIST( (inP, tsymP, typePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TNSYM		*tsymP )	/* Type name symbol */
    FARG( TYPE		**typePP )	/* Where to put the type */
{
	STSCODE		status;
	int		tmod;		/* Type modifier code */
	TYPE		*typeP;

    tmod = tsymP->tn_flags & TF_MOD;
    if ( tmod != 0 ) {
    	switch( tmod ) {
	    case TMSTRUCT:
	    	status = parse_type_agg( inP, &typeP, TC_STRUCT );
		break;

	    case TMUNION:
	    	status = parse_type_agg( inP, &typeP, TC_UNION );
		break;

	    case TMENUM:
	    	status = parse_type_enum( inP, &typeP );
		break;

	    case TMSHORT:
	    case TMLONG:
	    case TMUNSIGNED:
	    case TMSIGNED:
	    	/* Parse modified int declaration */
	    	status = parse_type_modint( inP, &typeP, tmod );
		break;

	    default:
	    	mfl_warning( inP, "invalid use of \"%s\".",
			tsymP->tn_sym.ss_nameP );
		status = SCNOTDONE;
		break;
	}
    }
    else {
	typeP = tsymP->tn_typeP;
    	status = SCOK;
    }

    if ( status == SCOK ) {
	*typePP = typeP;
    }

    return( status );
}
/*

*//* parse_type_enum( inP, typePP )

	Parse type: enum

Accepts :

	inP		Ptr to input stack
	typePP		Where to put the type
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_type_enum ARGLIST( (inP, typePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( TYPE		**typePP )	/* Where to put type */
{
	STSCODE		status;		/* Status code */
	int		enval;		/* Enum value */
	BOOL		newF;		/* Whether new definition */
	TYPE		*typeP;		/* Ptr to type */
	PTNODE		*treeP;		/* Header of enum tree */
	PTNODE		*nodeP;		/* Current node */
	PTNODE		*pnodeP;	/* Previous node */
	VSYM		*vsymP;		/* Var symbol ptr */
	TYPEAG		*agP;		/* Aggregate ptr */
	AGGMEM		*memP;		/* Aggregate member ptr */
	TOKEN		*tokP;

    /* parse the opening part */
    status = parse_type_aggname( inP, &typeP, &newF, TC_ENUM );
    if ( status == SCOK ) {
        /* See if defining new enum */
	if ( newF ) {
	    treeP = nodeP = pnodeP = NULL;
	    for( enval = 0; ; ) {
		tokP = mfl_sigtoken( inP );
		if ( ( tokP == NULL ) || ( tokP->tk_type != TK_SYM ) ) {
		    mfl_warning( inP, 
		    		"syntax error in enum -- symbol expected." );
		    status = SCNOTDONE;
		    break;
		}

		/* make next SYM node */
		parsetree_mknode( &nodeP, PN_SYM );
		if ( pnodeP == NULL )
		    treeP = nodeP;
		else
		    pnodeP->pt_leftP = nodeP;
		pnodeP = nodeP;
		nodeP->pt_info.od_P = (void *)newstr( "parse_type_enum",
					"enum element",
					(char *)&tokP->tk_bstrP->bs_bP[0] );

		/* Update the member count */
		++enval;

		/* Check for comma or final brace */
		tokP = mfl_sigtoken( inP );
		if ( ( tokP != NULL ) &&
		     ( tokP->tk_type == TK_CHAR ) &&
		     ( tokP->tk_bstrP->bs_bP[0] == '}' ) )
		    break;
		if ( ( tokP == NULL ) ||
		     ( tokP->tk_type != TK_OPR ) ||
		     ( tokP->tk_bstrP->bs_bP[0] != ',' ) ) {
		    mfl_warning( inP, 
		       "syntax error in enum -- comma or brace expected." );
		    mfl_tokback( inP, tokP );
		    status = SCNOTDONE;
		    break;
		}
	    }

	    if ( status == SCOK ) {
		/* Make the member list array */
		if ( enval == 0 )
		    enval = 1;
		agP = (TYPEAG *)typeP->ty_nextP;
		agP->ag_memC = enval;
		agP->ag_memP = memP =
		    (AGGMEM *)emalloc( "parse_type_enum",
					     "member list array",
					     enval * sizeof(AGGMEM) );

		/* Now define the enum members */
		for( enval = 0, nodeP = treeP; nodeP != NULL;
				++enval, ++memP, nodeP = nodeP->pt_leftP ) {
		    vsymP = (VSYM *)symtab_lookup( STglobal,
						  nodeP->pt_info.od_P );
		    if ( vsymP != NULL )
		    	mfl_warning( inP, "duplicate symbol \"%s\"",
				nodeP->pt_info.od_P );
		    else {
		        vsymP = (VSYM *)symtab_enter( STglobal, 
				nodeP->pt_info.od_P, sizeof(VSYM) );
			vsymP->vs_value.v_flags = VF_CONST|VF_RVALUE;

#ifdef	MFL_STRICT_ENUMS
			vsymP->vs_value.v_typeP = typeP;
#else
			vsymP->vs_value.v_typeP = Tnint2->tn_typeP;
#endif	/* MFL_STRICT_ENUMS */

			vsymP->vs_value.v_data.d_word = enval;
		    }

		    /* Fill in the member list entry */
		    memP->am_offs = 0;
		    memP->am_typeP = Tnint2->tn_typeP;
		    memP->am_nameP = (char *)nodeP->pt_info.od_P;
		    nodeP->pt_info.od_P = NULL;
		}
	    if ( status == SCOK )
		typeP->ty_flags |= TF_DEFINED;	    
	    }

	    parsetree_free( treeP );
	}
    }

    *typePP = typeP;

    return( status );
}
/*

*//* parse_type_modint( inP, typePP, tmod )

	Parse a modified int type (e.g., unsigned)

Accepts :

	inP		Ptr to input stack
	typePP		Where to put the type
	tmod		Type modifier code
	
Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_type_modint ARGLIST( (inP, typePP, tmod) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TYPE		**typePP )	/* Where to put the type */
    FARG( int		tmod )		/* Type modifier code */
{
	STSCODE		status;		/* Status code */
	TYPE		newtype;	/* New type holder */
	TYPE		*typeP;		/* Ptr to type made/found */
	TNSYM		*rtsymP;	/* Reference type sym ptr */
	TYPSYM		*tpsymP;	/* Ptr to type pool entry */
	TOKEN		*tokP;		/* Token pointer */

    /* Look for typeclass */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP != NULL ) &&
         ( tokP->tk_type == TK_SYM ) &&
         ( ( rtsymP = (TNSYM *)symtab_lookup( STtname,
	 			&tokP->tk_bstrP->bs_bP[0] ) ) != NULL ) ) {
	/* Call base-type parser to get base type */
	status = parse_type_base( inP, rtsymP, &typeP );
    }
    else {
    	/* Token is not a type specifier.  Push it back */
	mfl_tokback( inP, tokP );

	/* Default to int */
	typeP = Tnint2->tn_typeP;
	status = SCOK;
    }

    /* Apply the modifier if possible */
    if ( status == SCOK ) {
        if ( ( tokP == NULL ) ||
	     ( ( typeP->ty_class != TC_INT ) &&
               ( typeP->ty_class != TC_UINT ) ) ) {
	    mfl_warning( inP, "invalid use of \"%s\" -- ignored",
	    		tmod==TMSHORT?"short":
			 (tmod==TMLONG?"long":
			  (tmod==TMSIGNED?"signed":"unsigned"
			  ))   );
	}
	else {
	    /* Make a new type with the appropriate modifier */
	    memcpy( &newtype, typeP, sizeof(TYPE ) );
	    switch( tmod ) {
	    	case TMSHORT:
		    newtype.ty_size = SZSHORT;
		    break;

		case TMLONG:
		    if ( newtype.ty_size == 1 )
		    	mfl_warning( inP,
			  "warning -- long char works, but it's silly." );
		    newtype.ty_size = SZLONG;
		    break;

		case TMUNSIGNED:
		    newtype.ty_class = TC_UINT;
		    break;

		case TMSIGNED:
		    newtype.ty_class = TC_INT;
		    break;

		default:
		    mfl_warning( inP, "parse_type_modint: internal error!" );
		    status = SCERR;
	    }


	    if ( status == SCOK ) {
		/* Lookup the new type */
		tpsymP = (TYPSYM *)symtab_enter( STtpool, &newtype,
					sizeof(TYPSYM) );

		typeP = &tpsymP->ts_type;	/* New base type */
	    }
	}
    }

    *typePP = typeP;

    return( status );
}
/*

*//* parse_type_usage( inP, btypeP, typePP, vnamePP, infoP )

	Parses a type usage

Accepts :

	inP		Ptr to input stack
	btypeP		Ptr to the base type being used
	typePP		Where to store the resultant type for this use
	vnamePP		Where to put variable name pointer
	infoP		Where to return extended info

Returns :

	<value>		SCxx status code

Notes :

	This routine parses the usage of a base type, once that base
	type has been seen.  The base type might be something like
	"int" or "struct foo" or some typedef name, and the usage
	might be "*x" or "x[3]" or even "*" (e.g. in the case of a
	typecast or sizeof).  Thus the type usage consists of the
	completion of the type reference plus any variable name.

	It produces the reference type and any variable name.  Note
	that a reference type may be produced and later orphaned (e.g.
	in the case of a syntax error) but that's not a huge penalty.
	If we wanted to deal with that we could be more strict about
	the reference counts in the type symbols.

	This routine functions by doing some setup work, then calling
	parse_type_usage_tree to generate a parse tree representing the
	decoded type usage structure, and finally walking that parse
	tree to generate the resulting type.

	infoP is used to return any extended info.  This may be
	passed as NULL if that info isn't wanted.  However, if it's
	accepted, caller should free the info when done with it.

*/

STSCODE
parse_type_usage ARGLIST( (inP, btypeP, typePP, vnamePP, infoP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( TYPE		*btypeP )	/* Ptr to the base type */
   NFARG( TYPE		**typePP )	/* Where to put the type */
   NFARG( char		**vnamePP )	/* Where to put varname ptr */
    FARG( ODINFO	*infoP )	/* Where to put info */
{
	STSCODE		status;
	long		lval;		/* Some long value storage */
	PTNODE		*utreeP;	/* Ptr to parse tree for type usage */
	PTNODE		*unodeP;	/* Ptr to node in type usage tree */
	TYPSYM		*tpsymP;	/* Type pool symbol */
	TYPE		newtype;	/* To construct new type */

    /* Initialize before generating the type usage parse tree */
    *vnamePP = NULL;

    /* Call the recursive work function to generate that tree */
    status = parse_type_usage_tree( inP, &utreeP, vnamePP );

    /* Variable name "NIL" is special, clean that up before continuing */
    if ( *vnamePP == NIL )
	*vnamePP = NULL;

    /* Any info on the *first* node is returned to the caller.  This is
       a hacky way to bubble up extended information from parsing a
       fundamental type such as a function.
    */
    if ( infoP != NULL ) {
	if ( ( status == SCOK ) &&
             ( utreeP != NULL ) &&
	     ( utreeP->pt_info.od_P != NULL ) ) {
	    infoP->od_P = utreeP->pt_info.od_P;
	    infoP->od_del = utreeP->pt_info.od_del;
	    utreeP->pt_info.od_P = NULL;
	}
	else
	    infoP->od_P = NULL;
    }

    /* Walk the resulting parse tree to generate the type reference */
    for( unodeP = utreeP;
	    ( status == SCOK ) && ( unodeP != NULL);
		unodeP = unodeP->pt_leftP ) {
	/* Make new type referencing base type, and look it up */
	newtype.ty_flags = 0;
	newtype.ty_nextP = (void *)btypeP;

	switch( unodeP->pt_type ) {
	    case PN_PTR:
	    	newtype.ty_class = TC_PTR;
		newtype.ty_size = sizeof(VALPTR);
		break;

	    case PN_DIM:
		/* Evaluate the dimension.  Note the ugliness of
		   evaluating a parse tree at parse time.  We could
		   possibly avoid this if the size were not part of
		   the type.
		*/

		eval_tree_long( unodeP->pt_rightP, &lval );

		newtype.ty_class = TC_DIM;
		newtype.ty_dim = lval;
		newtype.ty_size = newtype.ty_dim * btypeP->ty_size;
		break;

	    case PN_FUNCDEF:
		newtype.ty_class = TC_FUNCT;
		break;

	    default:
		mfl_warning( inP,
		  "unknown node type %d in parse_type; name=\"%s\"",
				unodeP->pt_type,
				*vnamePP == NULL ? "<none>" : *vnamePP );
		status = SCNOTDONE;
		break;
	}

	if ( status == SCOK ) {
	    /* Create/lookup new type based on this reference */
	    tpsymP = (TYPSYM *)symtab_enter( STtpool, &newtype,
				sizeof(TYPSYM) );
	    if ( tpsymP == NULL )
		status = SCNOTDONE;
	    else
		btypeP = &tpsymP->ts_type;	/* New base type */
	}
    }

    parsetree_free( utreeP );		/* Clean up the tree */

    /* btypeP now has the resultant type. */
    *typePP = btypeP;

    return( status );
}
/*

*//* parse_type_usage_tree( inP, treePP, vnamePP )

	Parse a type usage, generating a parse tree

Accepts :

	inP		Ptr to input stack
	treePP		Where to store the resultant parse tree
	vnamePP		Where to put variable name pointer

Returns :

	<value>		SCxx status code

Notes :

	This routine parses the usage of some base type.  (The base
	type doesn't matter here, so it's not passed into this
	function).  What is "usage?"  The base type might be something
	like "int" or "struct foo" or some typedef name, and the usage
	might be "*x" or "x[3]" or even "*" (e.g. in the case of a
	typecast or sizeof).  Thus the type usage consists of the
	completion of the type reference plus any variable name.

	This function produces a parse tree that represents that usage.
	It also returns the variable name if any was given.

	This routine has a recursive nature, due to the requirements
	of processing some inside-out type usage constructs.  (That's
	also one reason why an intermediate parse tree is generated
	rather than generating a type directly.)

*/

STSCODE
parse_type_usage_tree ARGLIST( (inP, typePP, vnamePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to put the tree */
    FARG( char		**vnamePP )	/* Where to put varname ptr */
{
	STSCODE		status;
	char		*vnameP;	/* Variable name ptr */
	PTNODE		*leftP;
	PTNODE		*rightP;
	TOKEN		*tokP;

    *treePP = NULL;
    leftP = rightP = NULL;
    status = SCOK;

    /* Start out with a token */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL ) {
	mfl_warning( inP, "parse_type_usage_tree: premature end of input." );
	return ( SCNOTDONE );
    }

    /* We divide parsing into left-side and right-side.  Left-side is
       anything (typically '*') that can come on the left of a variable
       name, and right-side is naturally that which can come on the
       right of a variable.  Note that because of the recursive nature
       of this function, the left-side parsing at one level can end up
       getting the ride-side stuff at a lower level.
    */

    /* If we haven't yet seen (or disallowed) a variable name, we can
       look for left-side stuff.
    */
    if ( *vnamePP == NULL ) {
	/* Check for parenthetical grouping */
	if ( ( tokP->tk_type == TK_OPR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == '(' ) ) {
	    status = parse_type_usage_tree( inP, &leftP, vnamePP );

	    if ( *vnamePP == NULL )
		*vnamePP = NIL;		/* Name can no longer appear. */

	    if ( status == SCOK ) {	/* Get trailing paren */
		tokP = mfl_sigtoken( inP );
		if ( ( tokP == NULL ) ||
	             ( tokP->tk_type != TK_CHAR ) ||
		     ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
		    mfl_warning( inP, "unclosed parenthesis" );
		    status = SCNOTDONE;
		}
		else
		    /* Must have next token in hand. */
		    tokP = mfl_sigtoken( inP );
	    }
	}

	/* Check for pointer */
	else if ( ( tokP->tk_type == TK_OPR ) &&
		  ( tokP->tk_bstrP->bs_bP[0] == '*' ) ) {
	    /* Make a pointer node and recurse to get what's pointed to. */
	    status = parsetree_mknode( &leftP, PN_PTR );
	    if ( status == SCOK )
		status = parse_type_usage_tree( inP,
					      &leftP->pt_leftP, vnamePP );

	    if ( status == SCOK ) {
		/* The called parse_type_usage_tree may have returned an
		   info pointer in od_P; if so, bubble it up.  This is
		   mainly a hack to support function prototypes, and
		   could/should probably be done better.
		*/
		if ( ( leftP->pt_leftP != NULL ) &&
		     ( leftP->pt_leftP->pt_info.od_P != NULL ) ) {
		    if ( leftP->pt_info.od_P != NULL ) {
			mfl_warning( inP,
			  "parse_type_usage_tree: multiple od_P situation." );
		    }
		    else {
			/* Steal the returned info pointer. */
			leftP->pt_info = leftP->pt_leftP->pt_info;
			leftP->pt_leftP->pt_info.od_P = NULL;
			/* unecessary
			leftP->pt_leftP->pt_info.od_del = NULL;
			*/
		    }
		}

		/* We want to have the next token */
		tokP = mfl_sigtoken( inP );
	    }
	}
    }


    /* Get any variable name that comes before the right side. */
    if ( ( status == SCOK ) && ( *vnamePP == NULL ) ) {
	if ( ( tokP != NULL ) && ( tokP->tk_type == TK_SYM ) ) {
	    *vnamePP = newstr( "parse_type_usage_tree", "variable name",
			          (char *)tokP->tk_bstrP->bs_bP );
	    if ( *vnamePP == NULL )
		status = SCERR;
	    else
		tokP = mfl_sigtoken( inP );
	}
	else
	    *vnamePP = NIL;
    }

    /* Get right side element(s) */
    for ( ; ( status == SCOK ) && ( tokP != NULL ); ) {
	/* Check for function */
	if ( ( tokP->tk_type == TK_OPR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == '(' ) ) {
	    status = parse_func_def( inP, &rightP, vnameP );
	}
	else if ( ( tokP->tk_type == TK_OPR ) &&
		  ( tokP->tk_bstrP->bs_bP[0] == '[' ) ) {
    	    status = parsetree_mknode( &rightP, PN_DIM );
	    if ( status == SCOK ) {
		status = parse_expr_right( inP, 0, &rightP->pt_rightP );
		tokP = mfl_sigtoken( inP );
		if ( ( tokP == NULL ) ||
		     ( tokP->tk_type != TK_CHAR ) ||
		     ( tokP->tk_bstrP->bs_bP[0] != ']' ) ) {
		    mfl_warning( inP, "unclosed array definition" );
		    status = SCNOTDONE;
		}
	    }
	}
	else
	    break;

	if ( status == SCOK ) {
	    /* The right side has higher precedence here so is linked
	       at the top.  We'll attached the left side on the left,
	       and what we parsed on the right will be the new result
	       (i.e. the leftP tree).
	    */
	    rightP->pt_leftP = leftP;
	    leftP = rightP;

	    /* Need the next token.. */
	    tokP = mfl_sigtoken( inP );
	}
    }

    /* Return the result (the left tree pointer) */
    *treePP = leftP;

    if ( ( status == SCOK ) && ( tokP != NULL ) )
	mfl_tokback( inP, tokP );

    return( status );
}
/*

*//* parse_typecast(inP, treePP, tsymP)

	Parse a typecast

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the typecast node
	tsymP		Ptr to typecast type symbol

Returns :

	<value>		SCxx status code

Notes :

	Caller has already recognized the opening paren and the
	symbol following it that indicated a typecast.
	'tsymP' represents that type symbol.

	*treePP is returned NULL if status is not SCOK

	The returned tree looks like this:

		treePP--->     PN_CAST
			         |\
				 | \
				 |  \
				 |   \ 
			      TYPE    expr to cast
			(on pt_info)

*/

STSCODE
parse_typecast ARGLIST( (inP, treePP, tsymP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to put the typecast node */
    FARG( TNSYM		*tsymP )	/* type name symbol ptr */
{
	STSCODE		status;
	TOKEN		*tokP;		/* Token for input */
	TYPE		*btypeP;	/* Base type */
	PTNODE		*nodeP;
	char		*vnameP;	/* Variable name ptr */

    status = parsetree_mknode( &nodeP, PN_CAST );
    if ( status != SCOK ) {
	*treePP = NULL;
	return ( status );
    }
    nodeP->pt_info.od_del = NULL;

    status = parse_type_base( inP, tsymP, &btypeP );
    if ( status == SCOK ) {
	status = parse_type_usage( inP, btypeP,
		     (TYPE **)&nodeP->pt_info.od_P, &vnameP,
		     (ODINFO *)NULL);
	if ( status == SCOK ) {
	    if ( vnameP != NULL ) {
		mfl_warning( inP, 
			"ignoring variable name \"%s\" in typecast",
				vnameP );
		dealloc( vnameP );
	    }
	    /* Next thing must be closing paren */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
		 ( tokP->tk_type != TK_CHAR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
		mfl_warning( inP, "Unclosed typecast" );
		status = SCNOTDONE;
	    }
	    else {
		/* Cast applies to next single term. */
		status = parse_expr_right( inP, ORPREFIX, &nodeP->pt_rightP );
	    }
	}
    }

    if ( status != SCOK ) {
	parsetree_free( nodeP );
	nodeP = NULL;
    }

    *treePP = nodeP;

    return ( status );
}
/*

*//* parse_typedef( inP, treePP )

	Parse typedef statement

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the parse tree head
	
Returns :

	<value>		SCxx status code

Notes :

	NULL typedef, e.g. "typedef int;" is allowed .

*/

STSCODE
parse_typedef ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to put tree node */
{
	STSCODE		status;
	TNSYM		*tsymP;
	PTNODE		*tdnodeP;	/* For the PN_TYPEDEF node */
	PTNODE		*treeP;		/* For the type tree */
	TOKEN		*tokP;

    tdnodeP = NULL;			/* Initialize the return value */

    /* Next token must be type introducer */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP != NULL ) &&
         ( tokP->tk_type == TK_SYM ) &&
         ( ( tsymP = (TNSYM *)symtab_lookup( STtname,
	 			&tokP->tk_bstrP->bs_bP[0] ) ) != NULL ) ) {

	/* Raise the nesting level around the declaration parsing.
	   This is kind of a hack to prevent functions being defined
	   and typedef'd.
	*/
	++Mfl_nestC;
	status = parse_decl_stmt( inP, tsymP, &treeP );
	--Mfl_nestC;

	if ( status == SCOK ) {
	    status = parsetree_mknode( &tdnodeP, PN_TYPEDEF );
	    if ( status == SCOK )
		tdnodeP->pt_leftP = treeP;
	    else
		tdnodeP = treeP;	/* So it will get freed up */
	}
    }
    else {
	mfl_warning( inP, "typedef requires a type introducer." );
	mfl_tokback( inP, tokP );	/* What the heck, take it back */
	status = SCNOTDONE;
    }

    if ( ( status != SCOK ) && ( tdnodeP != NULL ) ) {
	parsetree_free( tdnodeP );
	tdnodeP = NULL;
    }

    *treePP = tdnodeP;
    return( status );
}
/*

*//* parsetree_free( treeP )

	Free storage associated with parse tree

Accepts :

	treeP		Ptr to tree to free

Returns :

	<value>		SCxx status code

Notes :

 	This routine (and the make-parse-node routines) store the
	released parse nodes on a local list for fast re-acquisition
	of parse nodes.

*/

STSCODE
parsetree_free ARGLIST( (treeP) )
    FARG( PTNODE	*treeP )	/* Ptr to parse tree */
{
    if ( treeP != NULL ) {
        /* Free left subtree */
	if ( treeP->pt_leftP != NULL )
            parsetree_free( treeP->pt_leftP );

	/* Free right subtree */
	if ( treeP->pt_rightP != NULL )
	    parsetree_free( treeP->pt_rightP );

	/* Delete any info attached */
	if ( treeP->pt_info.od_P != NULL )
	    odinfo_delete( &treeP->pt_info );
	if ( treeP->pt_xinfo.od_P != NULL )
	    odinfo_delete( &treeP->pt_xinfo );

	treeP->pt_leftP = PnodelistP;
	PnodelistP = treeP;
    }

    return ( SCOK );
}
/*

*//* parsetree_mknode( nodePP, type )

	Make a tree node of a given type

Accepts :

	nodePP		Where to store the node pointer
	type		Node type

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parsetree_mknode ARGLIST( (nodePP, type) )
   NFARG( PTNODE	**nodePP )	/* Where to put ptr to new node */
    FARG( PNTYPE	type )		/* Node type */
{
	STSCODE		status;
	PTNODE		*nodeP;		/* Ptr to node being made */

    status = SCOK;			/* Assume success */

    /* Try the free node list first */
    if ( ( nodeP = PnodelistP ) != NULL )
	PnodelistP = nodeP->pt_leftP;

    /* Nothing on free node list, allocate a new one */
    else {
	nodeP = (PTNODE *)emalloc( "parsetree_mknode",
				    "new node",
				    sizeof(PTNODE) );
	if ( nodeP == NULL ) {
	warning( "error allocating parse tree node!" );
	status = SCERR;
	}
    }

    if ( status == SCOK ) {
	*nodePP = nodeP;
	nodeP->pt_type = type;
	nodeP->pt_leftP = nodeP->pt_rightP = NULL;
	nodeP->pt_info.od_P = NULL;
	nodeP->pt_info.od_del = (BOOL (*)())dealloc;
	nodeP->pt_xinfo.od_P = NULL;
	nodeP->pt_xinfo.od_del = (BOOL (*)())dealloc;
    }

    return( status );
}
/*

*//* parsetree_mktknode( nodePP, tokP )

	Make a tree node out of a token

Accepts :

	nodePP		Where to store the node pointer
	tokP		Token from which to take the information

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parsetree_mktknode ARGLIST( (nodePP, tokP) )
   NFARG( PTNODE	**nodePP )	/* Where to put ptr to new node */
    FARG( TOKEN		*tokP )		/* Source for node */
{
	STSCODE		status;
	int		i;
	PTNODE		*nodeP;		/* Ptr to node being made */

    /* Make an empty node */
    if ( ( status = parsetree_mknode( &nodeP, 0 ) ) != SCOK )
	return( status );

    /* Link it where the caller wanted it */
    *nodePP = nodeP;

    /* Fill in the node based on token type */
    switch( tokP->tk_type ) {
    	case	TK_OPR:
	    nodeP->pt_type = PN_OPR;
	    nodeP->pt_data.pv_int = tokP->tk_val.tv_int;
	    break;

	case	TK_SYM:			/* Symbol */
	    /* It could be a formal function argument */
	    if ( ( FargsP != NULL ) &&
		 ( ( i = fargs_find( FargsP,
				(char *)tokP->tk_bstrP->bs_bP ) ) >= 0 ) ) {
		nodeP->pt_type = PN_ARG;
		nodeP->pt_data.pv_int = i;
	    }
	    else {
		/* Nope, just a regular symbol. */
		nodeP->pt_type = PN_SYM;
		nodeP->pt_info.od_P = (void *)newstr( "parsetree_mktknode",
					"symbol copy",
					(char *)tokP->tk_bstrP->bs_bP );
	    }
	    break;

	case	TK_INTEGER:		/* Integer */
	    nodeP->pt_type = PN_INTEGER;
	    nodeP->pt_data.pv_long = tokP->tk_val.tv_long;
	    break;

	case	TK_FLOAT:	     	/* Floating point */
	    nodeP->pt_type = PN_FLOAT;
	    nodeP->pt_data.pv_float = tokP->tk_val.tv_float;
	    break;

	case	TK_LITSTR:
	    nodeP->pt_type = PN_LITSTR;
	    nodeP->pt_info.od_P = (void *)newstr( "parsetree_mktknode",
					"string copy",
					(char *)tokP->tk_bstrP->bs_bP );
	    break;

	case	TK_ELLIPSIS:
	    warning( "bad use of ellipsis." );
	    status = SCNOTDONE;
	    break;

	default:
	    warning(
         "parsetree_mktknode not implemented for token type %d (%s)",
	    	     tokP->tk_type, mfl_toktype(tokP->tk_type) );
	    nodeP->pt_type = 100;
	    nodeP->pt_info.od_P = (void *)newstr( "parsetree_mktknode",
					"unknown copy",
					(char *)tokP->tk_bstrP->bs_bP );
	    break;
    }

    return( status );
}
/*

*//* parsetree_print( fP, level, treeP )

	Print a representation of the tree

Accepts :

	fP		File to print to
	level		Level number for this call
	treeP		Pointer to tree 

Returns :

	<nothing>

Notes :


*/

void
parsetree_print ARGLIST( (fP, level, treeP) )
   NFARG( FILE		*fP )		/* Output file */
   NFARG( int		level )		/* Level */
    FARG( PTNODE	*treeP )	/* Tree ptr */
{
	int		i;

    if ( treeP != NULL ) {
	/* Go down the right side first so that it looks like it's on
	   the right, kinda, if you twist your head.
	*/
	if ( treeP->pt_rightP != NULL )
            parsetree_print ( fP, level+1, treeP->pt_rightP );

	/* Print this level */
	for( i = 0; i < level; ++i )
	    fprintf( fP, "  " );
	fprintf( fP, "%d. ", level+1 );
	switch( treeP->pt_type ) {
	    case PN_EXPR:
	    	fprintf( fP, "PN_EXPR\n" );
		break;


	    case PN_OPR:
	        fprintf( fP, "PN_OPR: \"%s\"  prio=%d\n",
			Optbl[treeP->pt_data.pv_int].op_str,
			Optbl[treeP->pt_data.pv_int].op_prio );
		break;


	    case PN_INTEGER:
	    	fprintf( fP, "PN_INTEGER: %ld\n", treeP->pt_data.pv_long );
		break;


	    case PN_FLOAT:
	    	fprintf( fP, "PN_FLOAT: %f\n", treeP->pt_data.pv_float );
		break;


	    case PN_SYM:
	    	fprintf( fP, "PN_SYM: \"%s\"\n", treeP->pt_info.od_P );
		break;


	    case PN_LITSTR:
	    	fprintf( fP, "PN_LITSTR: \"%s\"\n", treeP->pt_info.od_P );
		break;


	    case PN_CAST:
	    	fprintf( fP, "PN_CAST:" );
		type_print( fP, treeP->pt_info.od_P );
		fprintf( fP, "\n" );
		break;


	    case PN_TYPEDEF:
	    	fprintf( fP, "PN_TYPEDEF\n" );
		break;


	    case PN_DECL:
	    	fprintf( fP, "PN_DECL" );
		if ( treeP->pt_info.od_P != NULL )
		    fprintf( fP, " name=\"%s\"", treeP->pt_info.od_P );
		fprintf( fP, "\n" );
		break;

	    case PN_TYPEREF:
	    	fprintf( fP, "PN_TYPEREF:" );
		type_print( fP, treeP->pt_info.od_P );
		fprintf( fP, "\n" );
		break;


	    case PN_PTR:
	        fprintf( fP, "PN_PTR\n" );
		break;


	    case PN_FUNCDEF:
	    	fprintf( fP, "PN_FUNCDEF\n" );
		break;


	    case PN_FUNCREF:
	    	fprintf( fP, "PN_FUNCREF\n" );
		break;


	    case PN_DIM:
	    	fprintf( fP, "PN_DIM\n" );
		break;


	    case PN_LSEQ:
	    	fprintf( fP, "PN_LSEQ\n" );
		break;


	    case PN_BIS:
	        fprintf( fP, "PN_BIS: \"%s\"\n",
			Bistbl[treeP->pt_data.pv_int].bis_symP );
		break;

#ifdef	SIEVE
	    case PN_SVBIS:
	        sieve_node_print_svbis( fP, treeP );
		break;
#endif	/* SIEVE */

	    case PN_ARG:
		fprintf( fP, "PN_ARG: %d\n", treeP->pt_data.pv_int );
		break;

	    default:
		fprintf( fP, "type=%d\n", treeP->pt_type );
		break;
	}

	/* Now go down the left */
	if ( treeP->pt_leftP != NULL )
	    parsetree_print( fP, level+1, treeP->pt_leftP );
    }
}





/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* xsizeof( inP, prio, nodeP )

	Recursive routine to do the sizeof work

Accepts :

	inP		Ptr to input stack
	prio		priority of expr
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :

	Handler for parse_bis_sizeof

*/

static STSCODE
xsizeof ARGLIST( (inP, prio, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( int		prio )		/* Expression priority */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TYPE		*btypeP;	/* Base type */
	TNSYM		*tsymP;		/* type name symbol ptr */
	char		*vnameP;	/* Variable name ptr */
	TOKEN		*tokP;

    /* Sizeof can take a type description or any expression.  This
       makes the parsing a little odd.
    */

    /* Get first token */
    tokP = mfl_sigtoken( inP );

    /* Check for paren */
    if ( ( tokP != NULL ) &&
         ( tokP->tk_type == TK_OPR ) &&
	 ( tokP->tk_bstrP->bs_bP[0] == '(' ) ) {
    	status = xsizeof( inP, 0, nodeP );
	if ( status == SCOK ) {
	    /* Must have trailing paren too */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_CHAR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ')' ) ) {
	    	mfl_warning( inP, "missing ')' in sizeof." );
		mfl_tokback( inP, tokP );
		status = SCNOTDONE;
	    }
	}
    }
    else {
        /* Not paren at this level.  Check for type */
	if ( ( tokP != NULL ) &&
	     ( tokP->tk_type == TK_SYM ) &&
	     ( ( tsymP = (TNSYM *)symtab_lookup( STtname,
	     			&tokP->tk_bstrP->bs_bP[0] ) ) != NULL ) ) {
		/* It's a type */
		parsetree_mknode( &nodeP->pt_leftP, PN_TYPEREF );
		status = parse_type_base( inP, tsymP, &btypeP );
		if ( status == SCOK ) {
		    status = parse_type_usage( inP, btypeP,
					 (TYPE **)&nodeP->pt_leftP->pt_info.od_P,
		    			 &vnameP, (ODINFO *)NULL );
		    nodeP->pt_leftP->pt_info.od_del = NULL;
		    if ( vnameP != NULL )
			dealloc( vnameP );
		}
	}
	else {
	    /* Not a type -- parse expression. */
	    mfl_tokback( inP, tokP );
	    status = parse_expr_right( inP, prio, &nodeP->pt_leftP );
	}
    }

    return( status );
}
