/* mfl-pp.c -- preprecessor functions

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff for the preprocessor.  

Included:

	pp_directive	Handle a preprocessor directive
	pp_incdir	Add an include directory to the search list
	pp_invoke	Invoke a macro

plus internal support routines

*/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-pp.h"
#include "mfl-token.h"
#include "mfl-func.h"
#include <mml/mml-dns.h>

#include "mvmfmisc.h"
#include "errwarn.h"

/* Local definitions */

    /* In a macro body, this marks an argument */
#define	PPARGMARK	0xff

// #define PREPROCESSOR_WARN_INVALID	/* handy here for debugging */

    /* Macro body stream handle, for mml-in type input stream */
typedef
  struct {
    MML_IN	*mb_inP;		/* ptr back to stack */
    int		mb_bufX;		/* Buffer index */
    int		mb_argC;		/* Number of args */
    PPSYM	*mb_symP;		/* Ptr to symbol entry */
    UBYTE	*mb_bodyP;		/* Ptr to macro body */
    char	*mb_argP[1];		/* Argument pointers */
  } MBSTR;


/* External data referenced */

extern	BOOL	MexpandF;		/* Whether expanding macros */
extern	SYMTAB	*STglobal;		/* Global variables */
extern	SYMTAB	*STpreproc;		/* Preprocessor symbol table */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Directives */
static	STSCODE	pp_define PROTO( (MML_IN *inP) );
static	STSCODE	pp_else PROTO( (MML_IN *inP) );
static	STSCODE	pp_endif PROTO( (MML_IN *inP) );
static	STSCODE	pp_help PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifdef PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifdef_func PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifdef_var PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifndef PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifndef_func PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifndef_var PROTO( (MML_IN *inP) );
static	STSCODE	pp_ifxdef PROTO( (MML_IN *inP, char *dirnameP,
				  BOOL defsense, BOOL (*checkrtc)()) );
static	STSCODE	pp_include PROTO( (MML_IN *inP) );
static	STSCODE	pp_include_noerr PROTO( (MML_IN *inP) );
static	STSCODE	pp_include_x PROTO( (MML_IN *inP, BOOL okerrF) );

    /* Support routines */
static	void	in_h_mb_finish PROTO( (MML_INS *instrP) );
static	int	in_h_mb_getc PROTO( (MML_INS *instrP) );
static	BOOL	ps_checkdef PROTO( (char *nameP) );
static	BOOL	ps_checkdef_func PROTO( (char *nameP) );
static	BOOL	ps_checkdef_var PROTO( (char *nameP) );
static	STSCODE	ps_endline PROTO( (MML_IN *inP) );
static	void	ps_mbfree PROTO( (MBSTR *mbstrP) );
static	STSCODE	ps_skipfalse PROTO( (MML_IN *inP) );

/* Private data */

    /* Include file search paths per type */
static	LL_HEAD	Ipath_system;		/* INC_SYSTEM, system-wide */
static	LL_HEAD	Ipath_user;		/* INC_USER, user-level */

    /* Number of TRUE levels we're nested */
static	int	PptrueC = 0;

    /* Number of FALSE levels we're nested */
static	int	PpfalseC = 0;


    /* Dispatch table for preprocessor directives.  I suppose we could
       stick these in a symtab but just chunking through this list is
       probably more efficient.
    */
static struct {
    char	*pp_nameP;		/* Directive name */
    STSCODE	(*pp_rtc)();		/* Routine to call */
    	} Pptbl[] = {

	{ "define",	pp_define	},
	{ "else",	pp_else		},
	{ "endif",	pp_endif	},
	{ "help",	pp_help		},
	{ "ifdef",	pp_ifdef	},
	{ "ifdef_func",	pp_ifdef_func	},
	{ "ifdef_var",	pp_ifdef_var	},
	{ "ifndef",	pp_ifndef	},
	{ "ifndef_func", pp_ifndef_func	},
	{ "ifndef_var",	pp_ifndef_var	},
	{ "include",	pp_include	},
	{ "include_noerr", pp_include_noerr	},

	{ NULL,	NULL			}
		    };

/*

*//* pp_directive( inP )

	Handle a preprocessor directive

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status

Notes :

	Should be called when the rest of the line is thought to be a
	preprocessor directive, e.g. when the preprocessor introducer
	character is seen in the input.

	If PREPROCESSOR_WARN_INVALID is defined, a warning is issued
	for any anything that isn't a valid preprocessor directive.
	Whether warned or not, the bad preprocessor line is skipped
	(as if it were a comment line.)

*/

STSCODE
pp_directive ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		ppX;
	int		status;
	char		*nameP;
	TOKEN		*tokP;		/* Ptr to token */

    /* Get token and try to look it up */
    tokP = mfl_token( inP );
    if ( tokP == NULL )
	return ( SCNOTDONE );

    for( ppX = 0; ( nameP = Pptbl[ppX].pp_nameP ) != NULL; ++ppX )
    	if ( strcmp( nameP, (char *)tokP->tk_bstrP->bs_bP ) == 0 )
	    break;

    /* Handle success or failure */
    if ( nameP == NULL ) {
#ifdef  PREPROCESSOR_WARN_INVALID
	mfl_warning( inP,
		    "bad preprocessor directive: \"%s\"", tokP->tk_bstrP->bs_bP );
#endif
	status = SCOK;

	/* Skip to EOL */
	while ( ( tokP != NULL ) &&  ( tokP->tk_type != TK_EOL ) )
	    tokP = mfl_token( inP );
    }
    else {
    	/* Dispatch to handler routine */
	status = (*Pptbl[ppX].pp_rtc)( inP );
    }

    return( status );
}
/*

*//* pp_incdir( inctype, dirnameP )

	Add a directory to the include directory search path

Accepts :

	inctype		The include file/path type
	dirnameP	Name of new directory to add

Returns :

	<value>		TRUE if OK
			FALSE if not.  May take error exit.

Notes :

	Order is important- dirnames will be saved in the order
	that this function is called.

*/

BOOL
pp_incdir ARGLIST( (inctype, dirnameP) )
   NFARG( MFLINCTYPE	inctype )	/* Include file/path type */
    FARG( ESTR		*dirnameP )	/* New directory name */
{
	LL_HEAD		*pathP;		/* Search path head */

    /* Get include file search path handle for the include type */
    switch ( inctype ) {
	case INC_SYSTEM:
	    pathP = &Ipath_system;
	    break;

	case INC_USER:
	    pathP = &Ipath_user;
	    break;

	default:
	    warning( "pp_incdir: invalid include file/path type %d", inctype );
	    return ( FALSE );
    }

    /* Hand off to path_dir_add() */
    return ( path_dir_add( pathP, dirnameP ) );
}
/*

*//* pp_init()

	Initialize the preprocessor module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
pp_init NOARGLIST
{
    ll_init( &Ipath_system );
    ll_init( &Ipath_user );

    return ( TRUE );
}
/*

*//* pp_invoke(inP, symP)

	Invoke a macro

Accepts :

	inP		Ptr to current input stack
	symP		Ptr to preprocessor symbol table entry

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
pp_invoke ARGLIST( (inP, symP) )
   NFARG( MML_IN	*inP )		/* Input stack pointer */
    FARG( PPSYM		*symP )		/* Ptr to macro symbol table entry */
{
	int		status;
	int		argN;		/* Argument number */
	int		ch;		/* Character */
	int		parenC;		/* Parenthesis count */
	int		argblen;	/* Arg buffer length */
	int		arglen;		/* Argument length */
	int		tklen;		/* Token length */
	char		*argP;		/* Argument ptr */
	char		*tkstrP;	/* Token string */
	MBSTR		*mbstrP;	/* Macro invocation handle */
	TOKEN		*tokP;		/* Ptr to token container */

    /* Turn expansions off while getting macro arguments */
    MexpandF = FALSE;

    /* Allocate handle */
    mbstrP = (MBSTR *)emalloc( "pp_invoke", "macro body stream handle",
		       sizeof( MBSTR ) + symP->pp_argC*sizeof(char *) );
    if ( mbstrP == NULL )
	return ( SCERR );

    /* Initialize the handle */
    mbstrP->mb_inP = inP;
    mbstrP->mb_bufX = 0;
    mbstrP->mb_argC = symP->pp_argC;
    mbstrP->mb_symP = symP;
    mbstrP->mb_bodyP = symP->pp_bodyP->bs_bP;

    /* Collect macro arguments, if any, as indicated by an opening
       paren (theoretically in the same input stream)
    */

    if ( symP->pp_argC >= 0 ) {

#ifndef PP_SAMN			/* If no space allowed after macro name */
	    ch = in_char( inP );
#else
	    ch = in_nonblank( inP, FALSE );
#endif	/* PP_SAMN */

	if ( ch != '(' ) {		/* No argument list */
	    in_charback( inP, ch );
	    mfl_warning( inP,
			  "Warning-- missing argument list for macro \"%s\"",
				symP->pp_sym.ss_nameP );
	}
	else {
	    for( argN =argblen =arglen =parenC =0, argP = NULL; ; ) {
	        tokP = mfl_token( inP );
		if ( tokP->tk_rawF ) {
		    tkstrP = (char *)tokP->tk_rawP->bs_bP;
		    tklen = tokP->tk_rawP->bs_bC;
		}
		else {
		    tkstrP = (char *)tokP->tk_bstrP->bs_bP;
		    tklen = tokP->tk_bstrP->bs_bC;
		}
		if ( ( strcmp( tkstrP, "," ) == 0 ) ||
		     ( ( parenC == 0 ) && 
		       ( strcmp( tkstrP, ")" ) == 0 ) ) ) {
		    /* End of argument */
		    if ( argN < mbstrP->mb_argC )
			mbstrP->mb_argP[argN] = argP;

		    if ( *tkstrP == ')' ) {
			/* Special check for empty argument list */
			if ( ( arglen > 0 ) || ( argN != 0 ) )
			    ++argN;
			break;
		    }
			    
		    argP = NULL;
		    argblen = arglen = 0;
		    ++argN;
		}
		else {
		    /* Part of argument */
		    if ( strcmp( tkstrP, "(" ) == 0 )
		        ++parenC;
		    else if ( strcmp( tkstrP, ")" ) == 0 )
		        --parenC;

		    if ( argN <= mbstrP->mb_argC ) {
			// tklen = tokP->tk_bstrP->bs_bL;
			if ( ( arglen + tklen + 2 ) > argblen ) {
			    /* Need another buffer */
			    argblen += (tklen + 20);
			    if ( argP == NULL )
				argP = emalloc( "pp_invoke", "arg buffer",
					           argblen );
				else
				    argP = erealloc( "pp_invoke", "arg buffer",
						    argP, argblen );
			    if ( argP == NULL )
			        break;
			}

			if ( arglen != 0 )
			    argP[arglen++] = ' ';
			strcpy( &argP[arglen], tkstrP );
			arglen += tklen;
		    }
		}
	    }
	    /* Check correct number of args */
	    if( argN != mbstrP->mb_argC )
		mfl_warning( inP,
			"Warning--wrong number of arguments in call to \"%s\"",
			symP->pp_sym.ss_nameP );
	}
    }

    /* To prevent recursive macro invocations, the macro is
       "painted blue" during its expansion. */
	symP->pp_blue = TRUE;

    /* Push the macro body into the input stream */
    if ( ! in_stream_push( inP, in_h_mb_getc, in_h_mb_finish,
			   mbstrP, "macro" ) ) {
	ps_mbfree( mbstrP );
	status = SCERR;
    }
    else
	status = SCOK;

    /* Restore the expansion flag */
    MexpandF = TRUE;			/* Must be TRUE, if we're
					   in this routine! */

    return( status );
}


/***********************************************************************
 *                                                                     *
 *           Internal support routines -- directives                   *
 *                                                                     *
 ***********************************************************************/


/*

*//* pp_define(inP)

	Define a macro

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_define ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		status;
	int		i;		/* Scratch */
	int		ch;		/* Character */
	int		argC;
	int 		curexpand;	/* Current expansion flag */
	int		tklen;		/* Token length */
	UBYTE		*tkP;		/* Token pointer */
	PPSYM		*symP;		/* Symbol table entry */
	char		macname[MFL_TOKSIZE +1]; /* Macro name */
	char		macargs[PP_ARGMAX][MFL_TOKSIZE +1];
	TOKEN		*tokP;		/* Token */
	BSTR		*bodyP;		/* Ptr to body */

    if ( PpfalseC > 0 )			/* Ignore if in false state */
	return ( SCOK );

    /* Presume not OK until proven otherwise */
    status = SCERR;

    /* Turn off macro substitution while creating a definition */
    curexpand = MexpandF;
    MexpandF = FALSE;

    /* Input the macro name */
    tokP = mfl_token( inP );

    if ( ( tokP != NULL ) &&
          ( ( tokP->tk_type == TK_SYM ) || ( tokP->tk_type == TK_BIS ) ) ) {
	/* Save a copy of the name */
	strncpy( &macname[0], (char *)tokP->tk_bstrP->bs_bP, MFL_TOKSIZE +1 );

	/* Check for macro arguments -- note that the paren must
	   immediately follow the macro name (no spaces!) */
	if ( ( ch = in_char( inP ) ) == '(' ) {
	    /* Input the macro arguments */
	    for( argC = 0; ;) {
	        tokP = mfl_token( inP );
		if ( tokP == NULL )
		    break;

		if ( ( tokP->tk_type != TK_SYM ) &&
		     ( tokP->tk_type != TK_BIS ) ) {
		    if ( ( argC == 0 ) &&
			 ( strcmp( (char *)tokP->tk_bstrP->bs_bP, ")" ) == 0 ) )
			 status = SCOK;
		    break;
		}

		/* Save a copy of the arg name */
		if ( argC < PP_ARGMAX ) {
		    strncpy( &macargs[ argC ][0],
			      (char *)tokP->tk_bstrP->bs_bP,
			      MFL_TOKSIZE +1 );
		    ++argC;
		}

		/* Look at next non-blank char */
		ch = in_nonblank( inP, FALSE );

		if ( ch != ',' ) {
		    if ( ch == ')' )
		    	status = SCOK;
		    break;
		}
	    }
	}
	else {
	    /* No argument list */
	    in_charback( inP, ch );
	    argC = -1;
	    status = SCOK;
	}
    }

    /* Check status and continue. */
    if ( status != SCOK ) {
	mfl_warning( inP, "Syntax error in #define \"%s\".", &macname[0] );
	ps_endline( inP );
    }
    else {
	/* Collect the body.  We'll collect the body as tokens so that
	   we can recognize arguments, but then stuff the token text
	   into a body string.
	*/
	bodyP = bstr_new( "macro body", 0, 0 );
	if ( bodyP == NULL )
	    return ( SCERR );

	while ( ( tokP = mfl_token( inP ) ) != NULL ) {
	    if ( tokP->tk_type == TK_EOL )
	    	break;

	    tkP = tokP->tk_bstrP->bs_bP;
	    tklen = tokP->tk_bstrP->bs_bC;
	    if ( tokP->tk_type != TK_LITSTR ) {
		/* Check for formal argument substitution */
	    
		for( i = 0; i < argC; ++i )
		    if ( strcmp( (char *)tkP, &macargs[i][0] ) == 0 ) {

			/* Change the token into an argument reference.
			   By definition this will fit..
			*/
		    tkP[0] = PPARGMARK;
		    tkP[1] = i+1;
		    tklen = 2;
		    break;
		    }
	    }
	    else {
		/* Check for raw buffer */
		if ( tokP->tk_rawF ) {
		    tkP = tokP->tk_rawP->bs_bP;
		    tklen = tokP->tk_rawP->bs_bC;
		}
	    }

	    /* Add to body */
	    if ( bodyP->bs_bC != 0 )
		bstr_putb( bodyP, ' ' );
	    bstr_catn( bodyP, (UBYTE *)tkP, tklen );
	}
    }

    if ( status == SCOK ) {
	/* Terminate the body */
	bstr_nulterm( bodyP );

    	/* Make symbol */
	symP = (PPSYM *)symtab_enter( STpreproc, &macname[0],
					sizeof( PPSYM ) );

	if ( symP == NULL )
	    status = SCERR;		/* Message already been given */
	else {
	    if ( symP->pp_bodyP != NULL ) {
	    	/* Macro is being redefined */
		mfl_warning( inP, "--redefining macro %s --", &macname[0] );
		bstr_free( symP->pp_bodyP );
	    }
	    symP->pp_bodyP = bodyP;

	    symP->pp_argC = argC;
	    symP->pp_blue = FALSE;	/* Not being expanded */
	}
    }

    /* Reset original macro expansion state */
    MexpandF = curexpand;

    return( SCOK );
}
/*

*//* pp_else(inP)

	Process an #else

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	If inside a true conditional, we can switch to false at any time.

	If inside a false conditional, we can only switch to true if it's
	not a nested false.  i.e., the outermost false level makes all
	inner levels false.

	Note that therefore a false block can be nested inside a true
	block, but a true block can't be nested inside a false block.

*/

static STSCODE
pp_else ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	STSCODE		status;

    if ( ( PpfalseC == 0 ) && ( PptrueC == 0 ) ) {
    	/* Not in a conditional */
	mfl_warning( inP, "@else outside of conditional." );
	status = SCNOTDONE;
    }
    else if ( PpfalseC > 0 ) {
	/* Inside a false conditional.  Only the outer one can
	   be else'd.
	*/
	if ( PpfalseC == 1 ) {
	    /* Outermost false block: switch to true. */
	    PpfalseC = 0;
	    ++PptrueC;

	    /* Skip to end of line */
	    ps_endline( inP );
	}
	status = SCOK;
    }
    else {
	/* Inside a true conditional.  Switch to false and skip to
	the end.
	*/
	PpfalseC = 1;
	--PptrueC;

	/* Skip to the end of the false part. */
	status = ps_skipfalse( inP );
    }

    return ( status );
}
/*

*//* pp_endif(inP)

	Process an #endif directive

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_endif ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	STSCODE		status;

    /* Eat up end of line */
    ps_endline( inP );

    status = SCOK;

    /* Decrement whatever level we're in. */
    if ( PpfalseC > 0 )
	--PpfalseC;
    else if ( PptrueC > 0 )
	--PptrueC;
    else {
	mfl_warning( inP, "@endif outside of conditional." );
	status = SCNOTDONE;
    }

    return( status );
}
/*

*//* pp_help(inP)

	Process "help" directive (give list of directives)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Intended for interactive use only

*/

static STSCODE
pp_help ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		i;

    if ( PpfalseC > 0 )			/* Ignore if in false state */
	return ( SCOK );

    /* Check junk on end of line */
    ps_endline( inP );

    /* Give list of supported preprocessor directives */
    printf( "Supported directives:\n" );
    for( i = 0; Pptbl[i].pp_nameP != NULL; ++i )
    	printf( "      #%s\n", Pptbl[i].pp_nameP );

    return( SCOK );
}
/*

*//* pp_ifdef(inP)

	Process "ifdef" directive (true if symbol defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifdef ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifdef", TRUE, ps_checkdef ) );
}
/*

*//* pp_ifdef_func(inP)

	Process "ifdef_func" directive (true if function defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifdef_func ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifdef_func", TRUE, ps_checkdef_func ) );
}
/*

*//* pp_ifdef_var(inP)

	Process "ifdef_var" directive (true if variable defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifdef_var ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifdef_var", TRUE, ps_checkdef_var ) );
}
/*

*//* pp_ifndef(inP)

	Process "ifndef" directive (TRUE if symbol not defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifndef ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifndef", FALSE, ps_checkdef ) );
}
/*

*//* pp_ifndef_func(inP)

	Process "ifndef_func" directive (TRUE if function not defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifndef_func ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifndef_func", FALSE, ps_checkdef_func ) );
}
/*

*//* pp_ifndef_var(inP)

	Process "ifndef_var" directive (TRUE if variable not defined)

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifndef_var ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to ifxdef common code */
    return ( pp_ifxdef( inP, "@ifndef_var", FALSE, ps_checkdef_var ) );
}
/*

*//* pp_ifxdef(inP, dirnameP, defsense, checkrtc)

	Common code for ifdef/ifndef 

Accepts :

	inP		Ptr to input stack
	dirnameP	Name of the directive we're processing
	defsense	sense of the "is defined" test
	checkrtc	Function to check to see if a symbol is defined

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_ifxdef ARGLIST( (inP, dirnameP, defsense, checkrtc) )
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
   NFARG( char		*dirnameP )	/* Directive name */
   NFARG( BOOL		defsense )	/* Sense of the "is defined" test */
    FARG( BOOL		(*checkrtc)() )	/* check function */
{
	int		truthF;		/* True/False */
	int		curexpand;	/* Current subst. mode */
	TOKEN		*tokP;		/* Token */

    /* Make sure macro substitution is off */
    curexpand = MexpandF;
    MexpandF = FALSE;

    /* Get the symbol to be checked */
    tokP = mfl_token( inP );

    /* Turn macro expansion back the way it was */
    MexpandF = curexpand;

    /* Skip rest of line */
    if ( ( tokP != NULL ) && ( tokP->tk_type != TK_EOL ) ) {
	mfl_tokreserve( tokP );
	ps_endline( inP );
	mfl_tokrelease( tokP );
    }

    /* If we're already false, just be more false. */
    if ( PpfalseC > 0 )
	++PpfalseC;
    else {
	/* Currently not false: process the directive. */
	if ( ( tokP == NULL ) || ( tokP->tk_type != TK_SYM ) ) {
	    mfl_warning( inP, "Expected symbol in %s", dirnameP );
	}
	else {
	    /* Check the symbol definition */
	    truthF = (*checkrtc)( tokP->tk_bstrP->bs_bP );
	    if ( !defsense )
		truthF = !truthF;

	    /* Process according to truth of conditional */
	    if ( truthF )
		++PptrueC;		/* Bump in-truth level */
	    else {
		PpfalseC = 1;		/* We're false; */
		ps_skipfalse( inP );	/* skip false conditional */
	    }
	}
    }

    return( SCOK );
}
/*

*//* pp_include(inP)

	Process the "include" directive -- include another source file

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_include ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to common code */
    return ( pp_include_x( inP, FALSE ) );
}
/*

*//* pp_include_noerr(inP)

	Process the "include_noerr" directive -- include another source file

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Like #include but no errors if file isn't available

*/

static STSCODE
pp_include_noerr ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
    /* Hand off to common code */
    return ( pp_include_x( inP, TRUE ) );
}
/*

*//* pp_include_x(inP, okerrF )

	Common code for #include directives

Accepts :

	inP		Ptr to input stack
	okerrF		TRUE if it's OK for the file to not be available

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
pp_include_x ARGLIST( (inP, okerrF) )
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( BOOL		okerrF )	/* If unavailable is OK */
{
	BOOL		trycwdF;	/* Try unadulterated path first */
	int		status;
	int		ch;		/* Character */
	int		termch;		/* Terminating char */
	int		fnX;		/* Index to filename */
	int		len;
	FILE		*ifP;		/* New file ptr */
	LL_NODE		*nodeP;		/* A node in the search path */
	char		filename[PATH_MAX +1];
	char		fullname[PATH_MAX +1];	/* For <> forms */

    if ( PpfalseC > 0 )			/* Ignore if in false state */
	return ( SCOK );

    ch = in_nonblank( inP, FALSE );	/* Get opening char */
    switch( ch ) {
	case	'<':			/* System path */
	    nodeP = Ipath_system.llh_firstP;
	    termch = '>';
	    trycwdF = FALSE;
	    status = SCOK;
	    break;

	case	'"':			/* User path */
	    nodeP = Ipath_user.llh_firstP;
	    termch = '"';
	    trycwdF = TRUE;
	    status = SCOK;
	    break;

	default:
	    status = SCERR;
	    break;
    }

    if ( status == SCOK ) {
	/* Collect the rest of the name */
	for( fnX = 0; ; ) {
	    ch = in_charesc( inP, NULL );
	    if ( ch == termch )
		break;

	    if ( ( ch == EOF ) || ( fnX >= PATH_MAX ) ) {
		status = SCNOTDONE;
		break;
	    }
	    filename[ fnX++ ] = ch;
	}

	filename[ fnX ] = NUL;
    }

    /* Do end-of-line skipping */
    if ( ch != '\n' )
	ps_endline( inP );

    if ( status != SCOK )
	mfl_warning( inP, "syntax error in include." );
    else {
	if ( trycwdF ) {
	    /* Try unadulterated path first. */
	    ifP = fopen( &filename[0], "r" );
	}
	else
	    ifP = NULL;

	/* Try all of the include directories for this type. */
	for ( ; nodeP != NULL; nodeP = nodeP->lln_nextP ) {
	    len = snprintf( &fullname[0], sizeof(fullname), "%s/%s",
			(char *)nodeP->lln_valP, &filename[0] );
	    if ( ( len < sizeof(fullname) ) &&
		 ( ( ifP = fopen( &fullname[0], "r" ) ) != NULL ) ) {
		strcpy( &filename[0], &fullname[0] );
		break;
	    }
	}

	if ( ( ifP == NULL ) && !okerrF ) {
	    mfl_warning( inP, "Can not open file %s", &filename[0] );
	    status = SCERR;
	}
    }

    /* Now if successful, do the include. */
    if ( ( status == SCOK ) && ( ifP != NULL ) )
	in_stream_push_file( inP, ifP, &filename[0] );

    return( SCOK );
}


/***********************************************************************
 *                                                                     *
 *             Internal support routines -- general                    *
 *                                                                     *
 ***********************************************************************/


/*

*//* in_h_mb_finish( instrP )

	Stream handler for "macro body": finish the stream

Accepts :

	instrP		Ptr to the input stream

Returns :

	<nothing>

Notes :

	Follows mml-in character input stream semantics

*/

static void
in_h_mb_finish ARGLIST( (instrP) )
    FARG( MML_INS	*instrP )	/* Ptr to macro body stream */
{
    /* Free the macro body handle */
    ps_mbfree( (MBSTR *)instrP->is_handleP );

    /* done */
}
/*

*//* in_h_mb_getc( instrP )

	Stream handler for "macro body": get char

Accepts :

	instrP		Ptr to the input stream

Returns :

	<value>		next char if any,
			EOF if not

Notes :

	Follows mml-in character input stream semantics

*/

static int
in_h_mb_getc ARGLIST( (instrP) )
    FARG( MML_INS	*instrP )	/* Ptr to macro body stream */
{
	UBYTE		ch;
	char		*strP;		/* Arg ptr */
	MBSTR		*mbstrP;	/* Ptr to macro body stream */

    mbstrP = (MBSTR *)instrP->is_handleP;

    /* Check for actual argument substitution */
    for( ; ; ) {
	ch = mbstrP->mb_bodyP[mbstrP->mb_bufX++];
	if ( ch != PPARGMARK )
	    break;

	/* Get marked argument # */
	ch = mbstrP->mb_bodyP[mbstrP->mb_bufX++];
	if ( ch == PPARGMARK )
	    break;			/* Literal mark */

	/* Get actual argument string */
	strP = mbstrP->mb_argP[ch-1];
	if ( strP != NULL ) {
	    in_stream_push_str( mbstrP->mb_inP, strP, strlen( strP ), NULL );
	    return( in_char( mbstrP->mb_inP ) );
	}
    }


    /* Check for end of string */
    if ( ch == NUL )
	return ( EOF );

    return( (int)ch );
}
/*

*//* ps_checkdef( nameP )

	Check to see if a preprocessor symbol is defined.

Accepts :

	nameP		Name of symbol to check.

Returns :

	<value>		TRUE if defined
			FALSE if not (or error)

Notes :

	Specified as a callback routine for ifxdef()

*/

static BOOL
ps_checkdef ARGLIST( (nameP) )
    FARG( char		*nameP )	/* Name to check */
{
	SYMSTR		*symP;		/* Symbol ptr */

    symP = (SYMSTR *)symtab_lookup( STpreproc, nameP );

    return ( symP != NULL );
}
/*

*//* ps_checkdef_func( nameP )

	Check to see if a function name is defined.

Accepts :

	nameP		Name of function to check.

Returns :

	<value>		TRUE if defined
			FALSE if not (or error)

Notes :

	Specified as a callback routine for ifxdef()

*/

static BOOL
ps_checkdef_func ARGLIST( (nameP) )
    FARG( char		*nameP )	/* Name to check */
{
    return ( func_lookup( nameP ) != NULL );
}
/*

*//* ps_checkdef_var( nameP )

	Check to see if a symbol is defined as a variable

Accepts :

	nameP		Name of symbol to check.

Returns :

	<value>		TRUE if defined
			FALSE if not (or error)

Notes :

	Specified as a callback routine for ifxdef()

*/

static BOOL
ps_checkdef_var ARGLIST( (nameP) )
    FARG( char		*nameP )	/* Name to check */
{
	VSYM		*vsymP;		/* Ptr to variable symtab entry */

    vsymP = (VSYM *)symtab_lookup( STglobal, nameP );

    if ( ( vsymP == NULL ) ||
         ( vsymP->vs_value.v_typeP == NULL ) ||
	 ( vsymP->vs_value.v_typeP->ty_class == TC_FUNCT ) )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* ps_endline(inP)

	Handle end of line after preprocessor directive completion

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
ps_endline ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	TOKEN		*tokP;		/* Token */

    tokP = mfl_token( inP );
    if ( ( tokP != NULL ) && ( tokP->tk_type != TK_EOL ) ) {
	mfl_warning( inP, "Ignoring junk on end of preprocessor line." );
	while( ( tokP != NUL ) && ( tokP->tk_type != TK_EOL ) )
	    tokP = mfl_token( inP );
    }

    return( SCOK );
}
/*

*//* ps_mbfree( mbstrP )

	Free a MBSTR block

Accepts :

	mbstrP		Block to free

Returns :

	<nothing>

Notes :


*/

static void
ps_mbfree ARGLIST( (mbstrP) )
    FARG( MBSTR		*mbstrP )	/* MBSTR thing to free */
{
	int		i;		/* Scratch */

    /* Free the argument substitution strings */
    for( i = 0; i < mbstrP->mb_argC; ++i )
        if ( mbstrP->mb_argP[i] != NULL )
	    dealloc( mbstrP->mb_argP[i] );

    /* The macro is no longer "painted blue", and is now again
       eligible for expansion. */
    mbstrP->mb_symP->pp_blue = FALSE;

    /* Free the handle */
    dealloc( mbstrP );
}
/*

*//* ps_skipfalse(inP)

	Skip a false conditional

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
ps_skipfalse ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
#if 0
	TOKEN		*tokP;

    /* All we do is eat up tokens until we are no longer in a false
       conditional state.  The falseness will be reset when an
       appropriate ending or transitional directive is processed by
       the preprocessor at a lower level.
    */

    while ( PpfalseC > 0 ) {
	if ( ( tokP = mfl_token( inP ) ) == NULL )
	    break;
    }

    if ( tokP != NULL )
	mfl_tokback( inP, tokP );
#else
    while ( PpfalseC > 0 ) {
	if ( ! mfl_ppftoken( inP ) )
	    break;
    }
#endif

    return( SCOK );
}
