/* mfl-sieve-vacation -- Parsing & execution of sieve 'vacation' stuff.

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Handles parsing, execution, and evaluation for the optional vacation
part of "sieve" mail-filtering elements.  See mfl-sieve for basic sieve
stuff.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"

    /* This module should only be compiled if SIEVE and VACATION are
       enabled, but we'll ensure that here.
    */

#ifdef	SIEVE
# ifdef	 VACATION

#ifdef HAVE_NDBM_H
# include <ndbm.h>
#else
# ifdef HAVE_GDBM_NDBM_H
#  include <gdbm/ndbm.h>
# else
#  ifdef HAVE_GDBM_H
#   include <gdbm.h>
#  endif /* HAVE_GDBM_H */
# endif /* HAVE_GDBM_NDBM_H */
#endif /* HAVE_NDBM_H */


#include "controls.h"
#include "mvmfmisc.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-parse.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-sieve.h"
#include "mfl-sieve-vacation.h"

#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */

extern	BOOL	Mfl_testF;		/* Test mode */
extern	OMM	*OmmP;			/* Open mail message */


/* External routines used */



/* Local data publicly available */

    /* Capabilities defined in here */
	SIEVE_CAP	Cap_vacation =	{ "vacation",	FALSE };

/* Local routines and forward references */

    /* See if one of the wanted addresses is listed in a recipient header */
static	BOOL	vacation_check_addresses PROTO( (char **addrPP) );

    /* Consult/update a vacation history database */
static	BOOL	vacation_check_history PROTO( (char *handleP,
					       char *addrP,
					       int days ) );

/* Private data */



    /* Definitions of Sieve statement options that are unique to
       the vacation facility.  See main sieve module for any more
       comments..
    */

/*	name		class			simple value
		capability		parse function  */

	SIEVE_OPTION	Sieve_opt_addresses = {
	"addresses",	SOC_ADDRESSES,		SOV_SET,
		NULL,			parse_sieve_targ_string_list	};

	SIEVE_OPTION	Sieve_opt_days = {
	"days",		SOC_DAYS,		SOV_SET,
		NULL,			parse_sieve_targ_int	};

	SIEVE_OPTION	Sieve_opt_from = {
	"from",		SOC_FROM,		SOV_SET,
		NULL,			parse_sieve_targ_string	};

	SIEVE_OPTION	Sieve_opt_handle = {
	"handle",	SOC_HANDLE,		SOV_SET,
		NULL,			parse_sieve_targ_string	};

	SIEVE_OPTION	Sieve_opt_mime = {
	"mime",		SOC_MIME,		SOV_SET,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_subject = {
	"subject",	SOC_SUBJECT,		SOV_SET,
		NULL,			parse_sieve_targ_string	};


    /* Options allowed by the "vacation" statement */
	SIEVE_OPTION	*Svstmt_opt_vacation[] = {
	&Sieve_opt_addresses,
	&Sieve_opt_days,
	&Sieve_opt_from,
	&Sieve_opt_handle,
	&Sieve_opt_mime,
	&Sieve_opt_subject,
	NULL };


/*

*//* exec_svbis_vacation( nodeP, rvalP )

	Execute sieve builtin statement: vacation

Accepts :

	nodeP		Ptr to PN_SVBIS/vacation parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_vacation ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "vacation" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	BOOL		okF;
	int		i;
	int		days;
	void		*hdrP;		/* Header object */
	char		*strP, *s1P;	/* Generic strings */
	char		*reply_toP;	/* Address we are replying to */
	char		*loginfoP;	/* A reason to log. */
	char		*handleP;	/* The given or presumed handle */
	char		*fromP;		/* Who from (e.g. :from) */
	FILE		*msgfP;		/* For message being sent */
	ESTR		*reasonP;	/* The reason/message given */
	char		**addrPP;	/* List of alternate addresses */
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    sieve_optvals_init( &optvals[0], &Svstmt_opt_vacation[0] );

    /* Requires vacation capability */
    if ( !sieve_cap_enabled( &Cap_vacation ) )
	status = SCNOTDONE;

    else {
	/* Get any arguments attached to this statement */
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );
    }

    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    reply_toP = OmmP->om_aesenderP;

    /* Silently inhibit vacation reply if message isn't autoreplyable */
    if ( ( status == SCOK ) && ( !msg_autoreplyable( OmmP ) ) ) {
	status = SCDONE;
	loginfoP = "non-autoreplyable";
    }

    /* Silently inhibit vacation reply if one of our addresses isn't
       specifically mentioned in one of the recipient address headers
    */
    if ( status == SCOK ) {
	/* Get the pointer to the list of alternate addresses, if any */
	if ( optvals[SOC_ADDRESSES].ov_oov == SOV_NONE )
	    addrPP = NULL;
	else
	    addrPP = (char **)optvals[SOC_ADDRESSES].ov_val.ovv_bPP;

	if ( !vacation_check_addresses( addrPP ) ) {
	    status = SCDONE;
	    loginfoP = "not-specifically-addressed";
	}
    }

    /* If we're operating in test mode, we report what action we would
       have taken, rather than actually taking the action.
    */
    if ( ( status == SCOK ) && Mfl_testF ) {
	printf( "test: vacation\n" );
	status = SCNOTDONE;
    }

    /* Consult the vacation database associated with this vacation
       profile (which is specified by its handle).  Silently inhibit
       the reply if we've replied to this within the indicated number
       of days.
    */
    if ( status == SCOK ) {
	if ( optvals[SOC_HANDLE].ov_oov == SOV_NONE )
	    handleP = "default";
	else
	    handleP = (char *)optvals[SOC_HANDLE].ov_val.ovv_bP;

	if ( optvals[SOC_DAYS].ov_oov == SOV_NONE )
	    days = 7;			/* The default if no :days */
	else
	    days = optvals[SOC_DAYS].ov_val.ovv_int;
	if ( days < 7 )			/* Enforce a minimum for :days */
	    days = 7;

TODO	/* The minimum for :days should probably be settable via runtime
	   configuration (e.g. a control).  Also, should we check against
	   a maximum here?
	*/

	if ( !vacation_check_history( handleP, reply_toP, days ) ) {
	    status = SCDONE;
	    loginfoP = "history";
	}
    }


    /* Evaluate the reason given */
    if ( status == SCOK )
	status = val_estr_eval( nodeP->pt_leftP, &reasonP, NULL );

    /* The ":from" option may be given to specify the email address
       to send from; otherwise the email_primary_address control is
       used.  It is an error if an address can't be determined.
    */
    if ( status == SCOK ) {
	if ( optvals[SOC_FROM].ov_oov != SOV_NONE )
	    fromP = (char *)optvals[SOC_FROM].ov_val.ovv_bP;
	else if ( ( fromP = (char *)cntrl_str_get(
				estr_tmp_str( "email_address_primary" ) ) )
		  == NULL ) {
	    status = SCDONE;
	    loginfoP = "need-:from";
	}
    }

    if ( status == SCOK ) {
	/* Send the reply */
	msgfP = msg_send_begin( OmmP, fromP, reply_toP );
	if ( msgfP == NULL )
	    okF = FALSE;
	else {
	    /* Start off by supplying some common header elements */
	    okF = TRUE;

	    /* Auto-Submitted is required */
	    okF = okF && ( fprintf( msgfP,
				   "Auto-Submitted: auto-replied\n",
				   fromP ) > 0 );

	    /* Write out the appropriate reference headers */
	    okF = okF && msgpart_write_references( OmmP->om_mmsgP->mm_topmpP,
						   msgfP );

	    /* Add a "Date: " header */
	    okF = okF && ( fprintf( msgfP,
				    "Date: %s\n",
				    msgpart_date_string_now() ) > 0 );
	}

	/* Some redundant ifs to reduce indenting.
	   Report me to the coding police.
	*/
	if ( ( msgfP != NULL ) &&
	     ( optvals[SOC_MIME].ov_oov == SOV_SET ) ) {
	    /* The reason string is a complete message, just send it. */
	    i = estr_len( reasonP );
	    okF = okF && ( fwrite( reasonP->es_strP, 1, i, msgfP ) == i );
	}
	else if ( msgfP != NULL ) {
	    /* Write headers for new message */
	    okF = okF && ( fprintf( msgfP, "From: %s\n", fromP ) > 0 );
	    okF = okF && ( fprintf( msgfP, "To: %s\n", reply_toP ) > 0 );
	    if ( optvals[SOC_SUBJECT].ov_oov == SOV_NONE ) {
		/* Generate a subject */
		okF = okF && ( fprintf( msgfP,
			  "Subject: Auto: out-of-office reply\n" ) > 0 );
	    }
	    else {
TODO		/* Brute force folding of subject; should really be
		   done by more generalized function with RFC2047 encoding
		   etc. 
		*/
		strP = (char *)optvals[SOC_SUBJECT].ov_val.ovv_bP;
		okF = okF && ( fprintf( msgfP, "Subject: " ) > 0 );
		while ( (s1P = strchr( strP, '\n' ) ) != NULL ) {
		    ++s1P;
		    i = s1P - strP;
		    okF = okF && ( fwrite( strP, 1, i, msgfP ) == i );
		    if ( ( *s1P != ' ' ) &&
		         ( *s1P != '\t' ) &&
			 ( *s1P != NUL ) )
			okF = okF && ( fputc( ' ', msgfP ) != EOF );
		    else if ( *s1P != NUL )
			++s1P;
		    strP = s1P;
		}
		if ( *strP != NUL )
		    okF = okF && ( fprintf( msgfP, "%s\n", strP ) > 0 );
	    }

	    /* End of header */
	    okF = okF && ( fputc( '\n', msgfP ) != EOF );

	    /* Add body */
	    i = estr_len( reasonP );
	    okF = okF && ( fwrite( reasonP->es_strP, 1, i, msgfP ) == i );
	}

	if ( msgfP != NULL ) {
	    /* Terminate with newline */
	    okF = okF && ( fputc( '\n', msgfP ) != EOF );

	    /* Close up */
	    if ( !msg_send_end( OmmP, msgfP ) )
		okF = FALSE;
	}

	if ( !okF ) {
	    status = SCNOTDONE;
	    log_both( "Action: vacation \"%s\" failed.", reply_toP );
	}
    }

    /* Clear up evaluated option values */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_vacation[0] );

    if ( ( status == SCNOTDONE ) && Mfl_testF ) {
	status = SCDONE;
	loginfoP = "test";
    }

    if ( status == SCOK ) {
	log_both( "Action: vacation \"%s\" succeeded.",
		   reply_toP );
    }

    if ( status == SCDONE ) {
	if ( Mfl_testF )
	    printf( "test: vacation inhibited -- %s\n", loginfoP );
	else
	    log_both( "Action: vacation \"%s\" inhibited -- %s.",
		  reply_toP,
		  loginfoP );
	status = SCOK;
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* parse_svbis_vacation( inP, nodeP )

	Parse sieve builtin statement: vacation

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Results in a PN_SVBIS parse node with the reason string
	value attached on the left as follows:

			PN_SVBIS
			   /
			  /
			 /
		the reason string.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_vacation ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_vacation[0] );

    /* Parse the reason string and attach it to the left. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    /* End with semi */
    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* vacation_check_addresses( addrPP )

	See if a wanted address is listed in a recipient header

Accepts :

	addrPP		Ptr to list of alternate addresses (or NULL if none)

Returns :

	<value>		TRUE if an address is listed,
			FALSE if not.

Notes :

	If addrPP is non-null, it points to a NULL-terminated list
	of pointers to address strings, each of which is an acceptable
	address.

	In addition to the specified list, addresses that are also
	acceptable are the recipient envelope address if known, and
	the primary email address (also if known).

*/

static BOOL
vacation_check_addresses ARGLIST( (addrPP) )
    FARG( char		**addrPP )	/* Ptr to addresses list */
{
	int		hL;		/* Length of header name */
	int		i;
	char		*primaddrP;	/* Primary address */
	UBYTE		*addrP;		/* Address we're looking at */
	void		*hdrP;		/* Header object */
	REFSTR		*rsP;
	BSTR		*bsP;

    /* If primary email address is known, it's also a valid match */
    primaddrP = cntrl_str_get( estr_tmp_str( "email_address_primary" ) );

    /* Walk the header fields; check recipient ones. */
    for ( hdrP = NULL;
	 ( hdrP = msgpart_hdr_next( OmmP->om_mmsgP->mm_topmpP,
				    hdrP, NULL ) ) != NULL;
		) {

	rsP = msgpart_hdr_str( OmmP->om_mmsgP->mm_topmpP, hdrP );
	bsP = refstr_bstr( rsP );
	hL = hdr_name_skip_estr( estr_tmp_str( bsP->bs_bP ), 0 );

	if ( ( strnicmp( (char *)bsP->bs_bP, "Bcc", hL ) == 0 ) ||
	     ( strnicmp( (char *)bsP->bs_bP, "Cc", hL ) == 0 ) ||
	     ( strnicmp( (char *)bsP->bs_bP, "To", hL ) == 0 ) ||
	     ( strnicmp( (char *)bsP->bs_bP, "Resent-Bcc", hL ) == 0 ) ||
	     ( strnicmp( (char *)bsP->bs_bP, "Resent-Cc", hL ) == 0 ) ||
	     ( strnicmp( (char *)bsP->bs_bP, "Resent-To", hL ) == 0 ) ) {

	    /* Check all the addresses in this header field */
	    while ( (addrP = (UBYTE *)msgpart_hdr_addr(
				       OmmP->om_mmsgP->mm_topmpP, hdrP,
				       AP_LOCAL|AP_DOMAIN ) ) != NULL ) {

		if ( E_RecipP != NULL ) {
		    /* Compare envelope recipient */
		    if ( stricmp( (char *)addrP,
				  (char *)(refstr_bstr( E_RecipP )->bs_bP)
					) == 0 )
			return ( TRUE );
		}

		if ( primaddrP != NULL ) {
		    /* Also compare with user's primary email address */
		    if ( stricmp( (char *)addrP, primaddrP ) == 0 )
			return ( TRUE );
		}

		/* Check the addresses list */
		for ( i = 0; addrPP[i] != NULL; ++i ) {
		    if ( stricmp( (char *)addrP, (char *)(addrPP[i]) ) == 0 )
			return ( TRUE );
		}
	    }
	}
    }

    /* No addresses matched */
    return ( FALSE );
}
/*

*//* vacation_check_history( handleP, addrP, days )

	Consult/update a vacation history database

Accepts :

	handleP		A handle for the vacation occurance
	addrP		The email address to check
	days		How many days to hold off replying once replied

Returns :

	<value>		TRUE if OK to send the message
			FALSE if not.

Notes :

	This function will return TRUE if the vacation database allows
	  a response to the given address, i.e. no prior reply has
	  been logged within the last "days" days.

	FALSE is returned otherwise-- including for inability to access
	  the vacation database.

	"handle" is used as a name for a vacation profile.  Each vacation
	  profile has its own record (perhaps its own database).

*/

static BOOL
vacation_check_history ARGLIST( (handleP, addrP, days) )
   NFARG( char		*handleP )	/* Vacation handle */
   NFARG( char		*addrP )	/* Address to check */
    FARG( int		days )		/* How many days between replies */
{
	BOOL		allowF;
	char		*ddirP;		/* Ptr to data directory */
	char		*htailP;	/* Tail (args) part of handle */
	DBM		*dbP;		/* Database handle */
	datum		keyd, vald;	/* Descriptors for key and value */
	time_t		timethen, timenow;
	char		dbpath[ PATH_MAX +1 ];
	char		handle[ PATH_MAX +1 ];

    if ( handleP == NULL )
	return ( FALSE );

    /* Need to copy the handle to manipulate it.  The handle string may
       not contain '/' characters, and anything after a '/' is used
       as some extra information for mvmf vacation extended capabilities.
    */
    if ( snprintf( &handle[0], sizeof(handle), "%s", handleP )
		>= sizeof(handle) )
	return ( FALSE );
    handleP = &handle[0];
    htailP = strchr( handleP, '/' );
    if ( htailP != NULL )
	*htailP++ = NUL;

    ddirP = data_directory( TRUE );	/* Make/get data directory path */
    if ( ddirP == NULL )
	return ( FALSE );
    if ( snprintf( &dbpath[0], sizeof(dbpath), "%s/%s-%s",
			ddirP, "vacation", handleP ) >= sizeof(dbpath) )
	return ( FALSE );

    dbP = dbm_open( &dbpath[0], O_RDWR|O_CREAT, 0660 );
    if ( dbP == NULL )
	return ( FALSE );

    time( &timenow );			/* We'll need this */

    keyd.dptr = addrP;
    keyd.dsize = strlen( addrP );
    vald = dbm_fetch( dbP, keyd );
    if ( ( vald.dptr != NULL ) && ( vald.dsize == sizeof(timethen) ) ) {
	/* Looks like a valid timestamp; check it out */
	memcpy( &timethen, vald.dptr, sizeof(timethen) );
	allowF = ((timenow - timethen) > (days * 3600L * 24L));
    }
    else
	/* No data for that address, allow the sending. */
	allowF = TRUE;

    if ( allowF ) {
	/* No history, or expired; [re]write the entry. */
	vald.dptr = (char *)&timenow;
	vald.dsize = sizeof(timenow);
	if ( dbm_store( dbP, keyd, vald, DBM_REPLACE ) != 0 ) {
	    /* Failure to store: don't allow the reply, 'cuz we want
	       to err on the side of NOT sending extra messages.
	    */
	    allowF = FALSE;
	}
    }

    dbm_close( dbP );

    return ( allowF );
}
/*

*//* end of vacation and sieve conditionals */

#endif		/* VACATION */
#endif		/* SIEVE */
