/* mfl-sieve -- Parsing, execution, evaluation for "sieve" elements

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Handles parsing, execution, and evaluation for "sieve" mail-filtering
elements.

Sieve elements are introduced by a keyword "sieve" and are completely
enclosed in curley brackets.  Even though the sieve language is
composed primarily of simple statements, we allow a sieve construct
to appear anywhere that an expression value may occur.  So for
example:

    c-statement;
    sieve { sieve-body; }
    c-statement;
    x = term + sieve { sieve-body; } + term;
    if ( sieve { sieve-body } )  c-statement;


*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>
#include "dnsbl.h"
#include "data.h"

#include "mvmf.h"

    /* This entire module should only be compiled at all if the
       SIEVE functions are enabled.  But we'll be sure..
    */
#ifdef	SIEVE

#include "controls.h"
#include "mvmfmisc.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-parse.h"
#include "mfl-type.h"
#include "mfl-eval.h"
#include "mfl-sieve.h"

#ifdef	VACATION
# include "mfl-sieve-vacation.h"
#endif	/* VACATION */

#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "errwarn.h"


/* Local definitions */

    /* A sieve built-in statement */
typedef struct {
    char	*sb_kwdP;		/* Keyword */
    STSCODE	(*sb_parse)();		/* Parse routine */
    STSCODE	(*sb_exec)();		/* Execution routine */
    UWORD	sb_type;		/* Statement type(s) */
}  SIEVE_BIS;


/* External data referenced */

extern	char	*MailboxP;		/* Default mailbox */
extern	BOOL	Mfl_testF;		/* If testing, not delivering */
extern	OMM	*OmmP;			/* Open mail message */
extern	TNSYM	*Tnint1;
extern	TNSYM	*Tnint2;
extern	TNSYM	*Tnint4;

    /* There may also be external things defined below to group them
       with related items.
    */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	STSCODE	do_exec_svbis_address PROTO( (SIEVE_OPTVAL *optvalsP,
					      ESTR *hnameP, ESTR *patP) );
static	STSCODE	do_exec_svbis_dnsbl PROTO( (SIEVE_OPTVAL *optvalsP,
					    ESTR *dnameP, ESTR *rnameP) );
static	STSCODE	do_exec_svbis_envelope PROTO( (SIEVE_OPTVAL *optvalsP,
					       ESTR *hnameP, ESTR *patP) );
static	STSCODE	do_exec_svbis_header PROTO( (SIEVE_OPTVAL *optvalsP,
					     ESTR *hnameP, ESTR *patP) );
static	STSCODE	exec_sieve_block PROTO( (PTNODE *treeP, VALUE *rvalP) );
static	BOOL	exec_sieve_xof PROTO( (PTNODE *nodeP, VALUE *valP,
					BOOL untilF) );
static	BOOL	sieve_eval_rel PROTO( (int cmpval, SIEVE_OPT_ONEOF reltype) );
static	BOOL	sieve_string_compare PROTO( (SIEVE_OPTVAL *optvalsP,
					     ESTR *es1P, ESTR *es2P) );
static	BOOL	sieve_test_result PROTO( (VALUE *valP) );

/* Private data */


    /* Statics used briefly for various things */

static	int		Countval;	/* Item count for :count */

    /* Some parsing context.. */
static	int	P_lastctlX;		/* Index of last control statement
					   parsed */

    /* Some execution context */
static	int	X_ifF;			/* Running flag for if/elsif/else
					   chains.  Set to TRUE if any
					   block in the chain got executed.
					*/

    /* Supported sieve capabilities.  We define them this way
       (rather than just have a table) so we can point to them
       individually.  It's tempting to throw all this
       stuff into symbol tables but it's not clear that would be
       an actual improvement.

       Some of these things may be defined as externals.
    */

static	SIEVE_CAP	Cap_copy =	{ "copy",	FALSE };
static	SIEVE_CAP	Cap_dnsbl =	{ "vnd.mvmf.dnsbl",  FALSE };
static	SIEVE_CAP	Cap_editheader = { "editheader", FALSE };
static	SIEVE_CAP	Cap_envelope =	{ "envelope",	FALSE };
static	SIEVE_CAP	Cap_fileinto =	{ "fileinto",	FALSE };
static	SIEVE_CAP	Cap_regex =	{ "regex",	FALSE };
static	SIEVE_CAP	Cap_reject =	{ "reject",	FALSE };
static	SIEVE_CAP	Cap_relational = { "relational", FALSE };

    /* SIEVE_CAPs defined externally */
#ifdef	VACATION		/* for "vacation" statement */
extern	SIEVE_CAP	Cap_vacation;
#endif	/* VACATION */


    /* Comparator capabilities here: */
static	SIEVE_CAP	Ccap_i_ascii_casemap =
		{ "comparator-i;ascii-casemap",		TRUE };

static	SIEVE_CAP	Ccap_i_ascii_numeric =
		{ "comparator-i;ascii-numeric",		FALSE };

static	SIEVE_CAP	Ccap_i_octet =
		{ "comparator-i;octet",			TRUE };


static	SIEVE_CAP	*Sieve_caps[] = {
	&Cap_copy,
	&Cap_dnsbl,
	&Cap_editheader,
	&Cap_envelope,
	&Cap_fileinto,
	&Cap_regex,
	&Cap_reject,
	&Cap_relational,

#ifdef	VACATION		/* for "vacation" */
	&Cap_vacation,
#endif	/* VACATION */

	/* Comparators.. */
	&Ccap_i_ascii_casemap,
	&Ccap_i_ascii_numeric,
	&Ccap_i_octet,

	NULL				};


    /* Aliases for standard capability names */
static struct {
    char	*ca_nameP;		/* Alias name */
    SIEVE_CAP	*ca_capP;		/* Capability ptr */
} Sieve_cap_aliases[] = {

#ifdef	NONVENDOR_DNSBL_CAP
{	"dnsbl",	&Cap_dnsbl		},
#endif	/* NONVENDOR_DNSBL_CAP */

{	NULL,		NULL			}
    };



    /* Sieve built-in statement keywords */
static	SIEVE_BIS	Sieve_bistbl[] = {
/*	name		parse func		exec func
		type		*/

{	"addheader",	parse_svbis_addheader,	exec_svbis_addheader,
		SBIS_ACTION	},

{	"address",	parse_svbis_address,	exec_svbis_address,
		SBIS_TEST	},

{	"allof",	parse_svbis_allof,	exec_svbis_allof,
		SBIS_TEST	},

{	"anyof",	parse_svbis_anyof,	exec_svbis_anyof,
		SBIS_TEST	},

{	"C",		parse_svbis_c,		exec_svbis_c,
		SBIS_ACTION|SBIS_TEST	},

{	"deleteheader",	parse_svbis_deleteheader, exec_svbis_deleteheader,
		SBIS_ACTION	},

{	"discard",	parse_svbis_discard,	exec_svbis_discard,
		SBIS_ACTION	},

{	"dnsbl",	parse_svbis_dnsbl,	exec_svbis_dnsbl,
		SBIS_TEST	},

{	"else",		parse_svbis_else,	exec_svbis_else,
		SBIS_CONTROL	},

{	"elsif",	parse_svbis_elsif,	exec_svbis_elsif,
		SBIS_CONTROL	},

{	"envelope",	parse_svbis_envelope,	exec_svbis_envelope,
		SBIS_TEST	},

{	"exists",	parse_svbis_exists,	exec_svbis_exists,
		SBIS_TEST	},

{	"false",	parse_svbis_false,	exec_svbis_false,
		SBIS_TEST	},

{	"fileinto",	parse_svbis_fileinto,	exec_svbis_fileinto,
		SBIS_ACTION	},

{	"header",	parse_svbis_header,	exec_svbis_header,
		SBIS_TEST	},

{	"if",		parse_svbis_if,		exec_svbis_if,
		SBIS_CONTROL	},

{	"keep",		parse_svbis_keep,	exec_svbis_keep,
		SBIS_ACTION	},

{	"not",		parse_svbis_not,	exec_svbis_not,
		SBIS_TEST	},

{	"redirect",	parse_svbis_redirect,	exec_svbis_redirect,
		SBIS_ACTION	},

{	"reject",	parse_svbis_reject,	exec_svbis_reject,
		SBIS_ACTION	},

{	"replaceheader",	parse_svbis_replaceheader, exec_svbis_replaceheader,
		SBIS_ACTION	},

{	"require",	parse_svbis_require,	exec_svbis_require,
		SBIS_ACTION	},

{	"sieve",	parse_svbis_sieve,	exec_svbis_sieve,
		SBIS_ACTION|SBIS_TEST	},

{	"size",		parse_svbis_size,	exec_svbis_size,
		SBIS_TEST	},

{	"stop",		parse_svbis_stop,	exec_svbis_stop,
		SBIS_ACTION	},

{	"true",		parse_svbis_true,	exec_svbis_true,
		SBIS_TEST	},

#ifdef	VACATION
{	"vacation", parse_svbis_vacation, exec_svbis_vacation,
		SBIS_ACTION	},
#endif	/* VACATION */

{	NULL,		NULL,			NULL,
		0		}
				  };



    /* These are the definitions of various Sieve statement
       options (tagged arguments).  Each Sieve statement that
       takes tagged arguments will have a control table referencing
       the relevant option definitions.
    */

/*	name		class			simple value
		capability		parse function  */


	SIEVE_OPTION	Sieve_opt_all = {
	"all",		SOC_ADDR_PART,		SOV_ADDR_ALL,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_comparator = {
	"comparator",	SOC_COMPARATOR,		SOV_SET,
		NULL,			parse_sieve_targ_comparator   };

	SIEVE_OPTION	Sieve_opt_contains = {
	"contains",	SOC_MATCH_TYPE,		SOV_MATCH_CONTAINS,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_copy = {
	"copy",		SOC_COPY,		SOV_SET,
		&Cap_copy,		NULL		};

	SIEVE_OPTION	Sieve_opt_count = {
	"count",	SOC_MATCH_TYPE,		SOV_MATCH_COUNT,
		&Cap_relational,	parse_sieve_targ_rel	};

	SIEVE_OPTION	Sieve_opt_domain = {
	"domain",	SOC_ADDR_PART,		SOV_ADDR_DOMAIN,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_index = {
	"index",	SOC_INDEX,		SOV_SET,
		&Cap_editheader,	parse_sieve_targ_int	};

	SIEVE_OPTION	Sieve_opt_ip = {
	"ip",		SOC_IP,			SOV_SET,
		&Cap_dnsbl,		parse_sieve_targ_string	};

	SIEVE_OPTION	Sieve_opt_is = {
	"is",		SOC_MATCH_TYPE,		SOV_MATCH_IS,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_last = {
	"last",		SOC_LAST,		SOV_SET,
		&Cap_editheader,	NULL		};

	SIEVE_OPTION	Sieve_opt_localpart = {
	"localpart",	SOC_ADDR_PART,		SOV_ADDR_LOCAL,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_matches = {
	"matches",	SOC_MATCH_TYPE,		SOV_MATCH_MATCHES,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_newname = {
	"newname",	SOC_NEWNAME,		SOV_SET,
		&Cap_editheader,	parse_sieve_targ_string	};

	SIEVE_OPTION	Sieve_opt_newvalue = {
	"newvalue",	SOC_NEWVALUE,		SOV_SET,
		&Cap_editheader,	parse_sieve_targ_string	};

	SIEVE_OPTION	Sieve_opt_over = {
	"over",		SOC_OVER_UNDER,		SOV_OVUN_OVER,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_regex = {
	"regex",	SOC_MATCH_TYPE,		SOV_MATCH_REGEX,
		&Cap_regex,		NULL		};

	SIEVE_OPTION	Sieve_opt_under = {
	"under",	SOC_OVER_UNDER,		SOV_OVUN_UNDER,
		NULL,			NULL		};

	SIEVE_OPTION	Sieve_opt_value = {
	"value",	SOC_MATCH_TYPE,		SOV_MATCH_VALUE,
		&Cap_relational,	parse_sieve_targ_rel	};


    /* Options allowed by the "address" statement */
	SIEVE_OPTION	*Svstmt_opt_address[] = {
	&Sieve_opt_all,
	&Sieve_opt_comparator,
	&Sieve_opt_contains,
	&Sieve_opt_count,
	&Sieve_opt_domain,
	&Sieve_opt_is,
	&Sieve_opt_localpart,
	&Sieve_opt_matches,
	&Sieve_opt_regex,
	&Sieve_opt_value,
	NULL };


    /* Options allowed by the "deleteheader" statement */
	SIEVE_OPTION	*Svstmt_opt_deleteheader[] = {
	&Sieve_opt_all,
	&Sieve_opt_comparator,
	&Sieve_opt_contains,
	&Sieve_opt_index,
	&Sieve_opt_is,
	&Sieve_opt_last,
	&Sieve_opt_matches,
	&Sieve_opt_regex,
	NULL };

    /* Options allowed by the "dnsbl" statement */
	SIEVE_OPTION	*Svstmt_opt_dnsbl[] = {
	&Sieve_opt_ip,
	NULL };

    /* Options allowed by the "envelope" statement */
	SIEVE_OPTION	*Svstmt_opt_envelope[] = {
	&Sieve_opt_all,
	&Sieve_opt_comparator,
	&Sieve_opt_contains,
	&Sieve_opt_count,
	&Sieve_opt_domain,
	&Sieve_opt_is,
	&Sieve_opt_localpart,
	&Sieve_opt_matches,
	&Sieve_opt_regex,
	&Sieve_opt_value,
	NULL };


    /* Options allowed by the "fileinto" statement */
	SIEVE_OPTION	*Svstmt_opt_fileinto[] = {
	&Sieve_opt_copy,
	NULL };


    /* Options allowed by the "header" statement */
	SIEVE_OPTION	*Svstmt_opt_header[] = {
	&Sieve_opt_comparator,
	&Sieve_opt_contains,
	&Sieve_opt_count,
	&Sieve_opt_is,
	&Sieve_opt_matches,
	&Sieve_opt_regex,
	&Sieve_opt_value,
	NULL };


    /* Options allowed by the "redirect" statement */
	SIEVE_OPTION	*Svstmt_opt_redirect[] = {
	&Sieve_opt_copy,
	NULL };


    /* Options allowed by the "replaceheader" statement */
	SIEVE_OPTION	*Svstmt_opt_replaceheader[] = {
	&Sieve_opt_all,
	&Sieve_opt_comparator,
	&Sieve_opt_contains,
	&Sieve_opt_index,
	&Sieve_opt_is,
	&Sieve_opt_last,
	&Sieve_opt_matches,
	&Sieve_opt_regex,
	&Sieve_opt_newname,
	&Sieve_opt_newvalue,
	NULL };


    /* Options allowed by the "size" statement */
    /* NOTE-- IF YOU ADD OPTIONS HERE, CHANGE THE SIMPLE TEST
       FOR ANY OPTION PRESENT IN THE PARSE ROUTINE.
    */
	SIEVE_OPTION	*Svstmt_opt_size[] = {
	&Sieve_opt_over,
	&Sieve_opt_under,
	NULL };


/*

*//* exec__sieve( treeP, rvalP )

	Execute a "sieve" tree

Accepts :

	treeP		Ptr to parse tree node at head of the tree
			 (assumed to be a "sieve" anchor)
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	A "sieve" tree is simply an introducer for zero or more
	sieve language elements.  The top of the tree is assumed to
	be a PN_BIS node indicating a statement type "sieve".  The
	body of the sieve tree hangs off of the left node (see
	parse_sieve).

*/

STSCODE
exec__sieve ARGLIST( (treeP, rvalP ) )
   NFARG( PTNODE	*treeP )	/* Address of tree ("sieve" node) */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Hand off the body to the exec_sieve function. */
    return ( exec_sieve_block( treeP->pt_leftP, rvalP ) );
}
/*

*//* exec_sieve( treeP, rvalP )

	Execute a "sieve" statement

Accepts :

	treeP		Ptr to parse tree node for sieve statement
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	A sieve tree pointer is a node represents a sieve statement or
	a sequencer block connecting sieve statements.  This function
	executes the statement associated with such a node.  May be
	recursive.

*/

STSCODE
exec_sieve ARGLIST( (treeP, rvalP ) )
   NFARG( PTNODE	*treeP )	/* Address of statement tree node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* The tree pointer can be null if there is an empty sieve body
       at the top level (such as passed by exec__sieve).
    */
    if ( treeP == NULL ) {
	val_set_int( rvalP, 4, 0 );	/* Make constant zero */
	return( SCOK );
    }

    status = SCOK;				/* Presume good status */

    switch( treeP->pt_type ) {
	case	PN_LSEQ:			/* Left-wise sequencing */
	    /* Exec the body of this node */
	    status = exec_sieve( treeP->pt_rightP, rvalP );

	    if ( status == SCOK )
		/* Now sequence to the left */
	    	status = exec_sieve( treeP->pt_leftP, rvalP );
	    break;


	case	PN_SVBIS:		/* Built-in sieve */
	    status = (*Sieve_bistbl[treeP->pt_data.pv_int].sb_exec)(
	    			treeP, rvalP );
	    break;


	default:			/* Something bad */
	    mfl_warning( NULL,
"internal error at exec_sieve - invalid node type %d found",
	    		treeP->pt_type );
	    status = SCERR;
	    break;
    }

    return( status );
}
/*

*//* exec_sieve_get_bisargs( nodeP, optvalsP )

	Fetch options attached to a SVBIS node

Accepts :

	nodeP		Ptr to parse tree node for sieve statement
	optvalsP	Ptr to array of SIEVE_OPTVAL elements

Returns :

	<value>		SCxx status code

Notes :

	A helper routine for various statements, this function
	looks at the options attached to an SVBIS node and stores
	appropriate values in the option values array pointed
	to by optvalsP.

	The array @ optvalsP should be at least as large as required
	by the largest option class index used.  It's prudent just to
	always make the array the maximum size (SOC_MAX).

	This function does NOT preinitialize the option values
	array; that should be done by the caller prior to this call.

	Does not check appropriateness of options for a particular
	statement type -- that should be handled in the parser.

*/

STSCODE
exec_sieve_get_bisargs ARGLIST( (nodeP, optvalsP ) )
   NFARG( PTNODE	*nodeP )	/* Address of statement tree node */
    FARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
{
	STSCODE		status;
	int		p;		/* pass over strings */
	int		oX;		/* Option index */
	int		sX;		/* String index */
	SIEVE_OPT_CLASS	oc;		/* Copy of option class */
	SIEVE_BISARG	*argsP;		/* Args ptr */
	SIEVE_OPTREF	*optrefP;	/* Option reference */
	SIEVE_OPTION	*optP;		/* Option definition */
	ESTR		*es1P;
	PTNODE		*awalkP, *anodeP; /* parse tree nodes ptrs for args */

    status = SCOK;

    /* Get any actual arguments */
    if ( ( argsP = (SIEVE_BISARG *)nodeP->pt_info.od_P ) != NULL ) {
	for ( oX = 0; ( status == SCOK ) && ( oX < argsP->ba_optC ); ++oX ) {
	    optrefP = &argsP->ba_optP[oX];
	    optP = optrefP->or_optP;

	    /* If capability is required, check it */
	    if ( ( optP->od_capP != NULL ) &&
	         ( !optP->od_capP->cap_enabledF ) ) {
		mfl_warning( NULL, "\":%s\" requires capability \"%s\".",
			    optP->od_nameP, optP->od_capP->cap_nameP );
		status = SCNOTDONE;
		break;
	    }

	    switch ( ( oc = optP->od_class ) ) {
		case	SOC_ADDR_PART:
		    switch( optP->od_value ) {
			case	SOV_ADDR_ALL:
			    optvalsP[oc].ov_val.ovv_uint = (AP_LOCAL|AP_DOMAIN);
			    optvalsP[oc].ov_oov = SOV_SET;
			    break;

			case	SOV_ADDR_DOMAIN:
			    optvalsP[oc].ov_val.ovv_uint = (AP_DOMAIN);
			    optvalsP[oc].ov_oov = SOV_SET;
			    break;

			case	SOV_ADDR_LOCAL:
			    optvalsP[oc].ov_val.ovv_uint = (AP_LOCAL);
			    optvalsP[oc].ov_oov = SOV_SET;
			    break;

			default:
			    mfl_warning( NULL,
		"exec_sieve_get_bisargs got unknown addr-parts code %d",
					optP->od_value );
			    status = SCERR;
			    break;
		    }
		    break;

		case	SOC_COMPARATOR:
		    optvalsP[oc].ov_oov = optrefP->or_int;

		    /* Hack to check capability */
		    if ( ( optvalsP[oc].ov_oov == SOV_COMP_ASC_NUMERIC ) &&
		          !Ccap_i_ascii_numeric.cap_enabledF ) {
			mfl_warning( NULL, 
		":comparator \"%s\" requires capability \"%s\".",
				    "i;ascii-numeric",
				    "comparator-i;ascii-numeric" );
			status = SCNOTDONE;
			optvalsP[oc].ov_oov = SOV_NONE;
		    }
		    break;



			/* Things that set the one-of value from the
			    referenced option definition.
			*/
		case	SOC_COPY:
		case	SOC_LAST:
		case	SOC_OVER_UNDER:
		    optvalsP[oc].ov_oov = optP->od_value;
		    break;



			/* Match-type is like setting the one-of value
			   from the option definition, but with a twist.
			*/
		case	SOC_MATCH_TYPE:
		    optvalsP[oc].ov_oov = optP->od_value;

		    /* For :count and :value match types, the relational
		       type is also stored in the 'int' value of the
		       parsed reference.  We'll just copy it even if
		       it's junk.  We could test this way:

		          if ( optP->od_parse == parse_sieve_targ_rel )

		       but why bother, it just adds a test.
		    */
		    optvalsP[oc].ov_val.ovv_int = optrefP->or_int;
		    break;



			/* Things that set the integer value from the
			   parsed or evaluated reference value
			*/
		case	SOC_INDEX:

#ifdef	VACATION
		case	SOC_DAYS:
#endif	/* VACATION */

		    optvalsP[oc].ov_val.ovv_int = optrefP->or_int;
		    optvalsP[oc].ov_oov = SOV_SET;
		    break;


			/* Things that set a string value.  The ovv_bP
			   value will point to an allocated string.
			*/
		case	SOC_IP:
		case	SOC_NEWNAME:
		case	SOC_NEWVALUE:

#ifdef	VACATION
		case	SOC_FROM:
		case	SOC_HANDLE:
		case	SOC_MIME:
		case	SOC_SUBJECT:
#endif	/* VACATION */

		    status = val_estr_eval( optrefP->or_valP, &es1P, NULL );
		    if ( status == SCOK ) {
			optvalsP[oc].ov_val.ovv_bP =
			        (UBYTE *)newstr_estr( "exec_sieve_get_bisargs",
					     "string option value",
					     es1P );
			optvalsP[oc].ov_oov = SOV_SET;
			val_estr_finish( es1P );
		    }
		    break;



			/* Things that have a list of strings.  The
			   ovv_bPP value element will point to an allocated
			   array of pointers to allocated strings, ending
			   with a NULL pointer.
			*/
			/* NOTE: only vacation uses this at present, so
			   it's all in a vacation conditional.  Change
			   if needed..
			*/

#ifdef	VACATION
		case	SOC_ADDRESSES:
		    /* Make two passes over the list.  The first will
		       count the strings, the second will evaluate
		       them.
		    */
		    status = SCOK;	/* in case of no strings */
		    for ( p = 0; p < 2; ++p ) {
		        for ( sX = 0, awalkP = optrefP->or_valP;
				awalkP != NULL; ++sX ) {
			    /* Get string/expr node and step to next */
			    if ( awalkP->pt_type == PN_LSEQ ) {
				anodeP = awalkP->pt_rightP;
				awalkP = awalkP->pt_leftP;
			    }
			    else {
				anodeP = awalkP;
				awalkP = NULL;
			    }
			    if ( p == 1 ) {
				/* Evaluation pass. */
				status = val_estr_eval( anodeP, &es1P, NULL );
				if ( status != SCOK )
				    break;
				optvalsP[oc].ov_val.ovv_bPP[sX] =
				    (UBYTE *)newstr_estr(
					"exec_sieve_get_bisargs",
					"string list option value",
					es1P );
				val_estr_finish( es1P );
			    }
			}
			if ( p == 0 ) {
			    /* We counted the strings, make an array */
			    optvalsP[oc].ov_val.ovv_bPP =
				(UBYTE **)emalloc(
				    "exec_sieve_get_bisargs",
				    "string list array",
				    (sX +1) * sizeof(char *) );
			}
		    }
		    if ( status == SCOK ) {
		        optvalsP[oc].ov_val.ovv_bPP[sX] = NULL;
			optvalsP[oc].ov_oov = SOV_SET;
		    }
		    else {
			/* Some error, must undo what we did. */
			while ( sX-- > 0 )
			    dealloc( optvalsP[oc].ov_val.ovv_bPP[sX] );
			dealloc( optvalsP[oc].ov_val.ovv_bPP );
			optvalsP[oc].ov_val.ovv_bPP = NULL;
		    }
		    break;
#endif	/* VACATION */


		default:
		    mfl_warning( NULL, "exec_sieve_get_bisargs got unknown option \"%s\".",
			    optP->od_nameP );
		    status = SCERR;
		    break;
	    }
	}
    }

    return ( status );
}
/*

*//* exec_sieve_stringx2( optvalsP, list1P, list2P, rtc )

	Execute a cross product between two string lists

Accepts :

	optvalsP	Ptr to option values array
	list1P		Ptr to node starting the first string list
	list2P		Ptr to node starting the second string list
	rtc		Routine to call for each cross product

Returns :

	<value>		SCxx status code

Notes :

	The optvalsP option values array is passed along to the 'rtc'
	routine-to-call.

	If optvalsP is non-NULL, this function checks the specified
	matchtype (in SOC_MATCH_TYPE), and operates differently if:

	 - matchtype is SOV_MATCH_COUNT ("count" mode)
	 - matchtype is anything else   ("normal" mode)

    "Normal" mode operates this way:

	Walks through all combinations of strings, one from the
	first list, one from the second list, and calls the
	desired routine-to-call with each string.  The routine is
	called:

	    rtc( optvalsP, es1P, es2P )

	where es1P and es2P are pointers to ESTRs.

	The routine is expected to return SCDONE when it's OK to
	stop with a valid status, SCOK when the walk should
	continue, and any other code to indicate a problem.
	SCDONE will be returned to the caller if it's returned here.

	This allows for some sort of closing processing.

	This routine returns SCOK if it processes all strings without
	receiving an SCDONE result.


    "Count" mode calls exec_sieve_stringx2count -- see that
    description for more.  Note that "rtc" must be able to
    handle a call by either of these routines (that is,
    by either exec_sieve_stringx2 or exec_sieve_stringx2count)


*/

STSCODE
exec_sieve_stringx2 ARGLIST( (optvalsP, list1P, list2P, rtc) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( PTNODE	*list1P )	/* Ptr to first list */
   NFARG( PTNODE	*list2P )	/* Ptr to second list */
    FARG( STSCODE	(*rtc)())	/* Routine to call */
{
	STSCODE		status;
	SIEVE_OPT_ONEOF	matchtype;	/* Match type to use */
	PTNODE		*walk1P, *walk2P; /* To walk each list with */
	PTNODE		*node1P;	/* First string node */
	PTNODE		*node2P;	/* Second string node */
	ESTR		*es1P, *es2P;	/* ESTR pointers */

    /* Intercept "count" mode */
    if ( ( optvalsP != NULL ) &&
         ( optvalsP[ SOC_MATCH_TYPE ].ov_oov == SOV_MATCH_COUNT ) )
	return ( exec_sieve_stringx2count( optvalsP, list1P, list2P, rtc) );

    status = SCOK;			/* In case no strings.. */

    /* Walk through first list */
    for ( walk1P = list1P; ( walk1P != NULL ) && ( status == SCOK ); ) {
	/* Get first string/expr node.  */
	if ( walk1P->pt_type == PN_LSEQ ) {
	    node1P = walk1P->pt_rightP;
	    walk1P = walk1P->pt_leftP;
	}
	else {
	    node1P = walk1P;
	    walk1P = NULL;
	}

	/* Access the node's string */
	status = val_estr_eval( node1P, &es1P, NULL );
	if ( status != SCOK )
	    break;

	/* Walk through second string list */
	for ( walk2P = list2P; ( walk2P != NULL ) && ( status == SCOK ); ) {
	    /* Get second string/expr node */
	    if ( walk2P->pt_type == PN_LSEQ ) {
		node2P = walk2P->pt_rightP;
		walk2P = walk2P->pt_leftP;
	    }
	    else {
		node2P = walk2P;
		walk2P = NULL;
	    }

	    /* Access the string */
	    status = val_estr_eval( node2P, &es2P, NULL );
	    if ( status != SCOK )
		break;

	    /* Execute the desired routine */
	    status = (*rtc)( optvalsP, es1P, es2P );

	    /* End access to second string */
	    val_estr_finish( es2P );

	    if ( status != SCOK )
		break;			/* Need to stop */
	}

	/* End access to first string */
	val_estr_finish( es1P );
    }

    return ( status );
}
/*

*//* exec_sieve_stringx2count( optvalsP, list1P, list2P, rtc )

	Execute a counting cross product between two string lists
	 

Accepts :

	optvalsP	Ptr to option values array.
	list1P		Ptr to node starting the first string list
	list2P		Ptr to node starting the second string list
	rtc		Routine to call to accumulate the count

Returns :

	<value>		SCxx status code

Notes :

	Is called by exec_sieve_stringx2 when it detects
	a "count" mode.  See notes there for how that is
	detected.

	This routine first sets Countval to zero.
	Then for each string in list1P, calls the routine-to-call:

	    rtc( optvalsP, es1P, NULL )

	where es1P is a pointer to a ESTR.

	rtc is expected to update Countval according to however
	it does the counting.  rtc is expected to return SCOK;
	any other value terminates processing with the returned
	status code.  When the list1P list is exhausted, this
	routine then tests each string value in list2P against
	Countval, returning when any result is TRUE or when
	the list is finished.

	It bears repeating that any "rtc" that can have a counting
	mode must be capable of being called by either this routine
	or by exec_sieve_stringx2count().

	Returns the same as exec_sieve_stringx2, i.e.:
	   SCDONE means there is a TRUE result;
	   SCOK means there is FALSE result, but no error;
	   anything else is an abnormal return.

*/

STSCODE
exec_sieve_stringx2count ARGLIST( (optvalsP, list1P, list2P, rtc) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( PTNODE	*list1P )	/* Ptr to first list */
   NFARG( PTNODE	*list2P )	/* Ptr to second list */
    FARG( STSCODE	(*rtc)())	/* Routine to call */
{
	STSCODE		status;
	BOOL		resultF;
	PTNODE		*walk1P, *walk2P; /* To walk each list with */
	PTNODE		*node1P, *node2P; /* string arg nodes */
	ESTR		*es1P, *es2P;	  /* String values */
	char		count[20];	/* To contain the count string */

    status = SCOK;			/* In case no strings.. */
    Countval = 0;

    /* Walk through first list to accumulate the count */
    for ( walk1P = list1P; ( walk1P != NULL ) && ( status == SCOK ); ) {
	if ( walk1P->pt_type == PN_LSEQ ) {
	    node1P = walk1P->pt_rightP;
	    walk1P = walk1P->pt_leftP;
	}
	else {
	    node1P = walk1P;
	    walk1P = NULL;
	}

	status = val_estr_eval( node1P, &es1P, NULL );
	if ( status == SCOK ) {
	    /* Call the rtc to accumulate the count */
	    status = (*rtc)( optvalsP, es1P, NULL );
	    val_estr_finish( es1P );
	}
    }

    if ( status != SCOK )
	resultF = FALSE;
    else {

	/* We're going to use standard string comparisons to check the
           count, so turn the count into a string.
	   Note that es1P does not have to be "finished" since it
	   uses an estr_tmp
	*/
	snprintf( &count[0], sizeof(count)-1, "%d", Countval );
	es1P = estr_tmp_str( (UBYTE *)&count[0] );
	resultF = FALSE;

	/* Walk through second string list to test the count value */
	for ( walk2P = list2P; ( walk2P != NULL ) && ( status == SCOK ); ) {
	    if ( walk2P->pt_type == PN_LSEQ ) {
		node2P = walk2P->pt_rightP;
		walk2P = walk2P->pt_leftP;
	    }
	    else {
		node2P = walk2P;
		walk2P = NULL;
	    }

	    /* Perform string compare.  The node2P string is the
	       pattern, the count string is the string to test against.
	    */
	    status = val_estr_eval( node2P, &es2P, NULL );
	    if ( status == SCOK ) {
		resultF = sieve_string_compare( optvalsP, es2P, es1P );
		val_estr_finish( es2P );
	    }

	    /* Return whenever we get a TRUE */
	    if ( resultF )
		break;
	}
    }

    if ( status == SCOK ) {
	/* Finished-- now return SCDONE if true.  */
	if ( resultF )
	    status = SCDONE;
    }

    return ( status );
}
/*

*//* exec_svbis_addheader( nodeP, rvalP )

	Execute sieve builtin statement: addheader

Accepts :

	nodeP		Ptr to PN_SVBIS/addheader parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_addheader ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "addheader" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	ESTR		*hdrnameP;
	ESTR		*hdrvalP;

    /* Requires editheader capability */
    if ( !sieve_cap_enabled( &Cap_editheader ) )
	status = SCNOTDONE;

    else if ( OmmP == NULL ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    else {
	/* Extract the string values we will need. */
	status = val_estr_eval( nodeP->pt_leftP, &hdrnameP, NULL );
	if ( status == SCOK ) {
	    status = val_estr_eval( nodeP->pt_rightP, &hdrvalP, NULL );
	    if ( status == SCOK ) {
		mml_mutex_lock( &OmmP->om_curmpP->mp_hdrs.llh_lock );
		if ( msgpart_hdr_add2( OmmP->om_curmpP,
				       hdrnameP, hdrvalP, NULL ) )
		    status = SCOK;
		else
		    status = SCNOTDONE;
		mml_mutex_unlock( &OmmP->om_curmpP->mp_hdrs.llh_lock );
		val_estr_finish( hdrvalP );
	    }
	    val_estr_finish( hdrnameP );
	}
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* exec_svbis_address( nodeP, rvalP )

	Execute sieve builtin statement: address

Accepts :

	nodeP		Ptr to PN_SVBIS/address parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_address ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "address" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    /* Get any arguments attached to this statement */
    sieve_optvals_init( &optvals[0], &Svstmt_opt_address[0] );
    status = exec_sieve_get_bisargs( nodeP, &optvals[0] );
    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	val_set_int( rvalP, 4, 0 );
	return ( SCNOTDONE );
    }

    if ( status == SCOK ) {
	/* Execute our handler routine for each cross product of the
	   list of headers and the list of addresses
	*/
	status = exec_sieve_stringx2( &optvals[0],
				      nodeP->pt_leftP, nodeP->pt_rightP,
				      do_exec_svbis_address );
    }

    /* Clear up option value data as needed */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_address[0] );

#if 0
    /* SCDONE will indicate a match; return the result value from that */
    val_set_int( rvalP, 4, ( status == SCDONE ) ? 1 : 0 );

    /* And turn SCDONE into an OK result */
    if ( status == SCDONE )
	status = SCOK;
#else
    if ( status == SCDONE ) {
	/* SCDONE indicates success; as an added bonus, if the count
	   is non-zero, give that as the test result.
	*/
	val_set_int( rvalP, 4, ( Countval == 0 ) ? 1 : Countval );
	status = SCOK;
    }
    else
	/* No match, always return zero */
	val_set_int( rvalP, 4, 0 );
#endif

    return ( status );
}
/*

*//* exec_svbis_allof( nodeP, rvalP )

	Execute sieve builtin statement: allof

Accepts :

	nodeP		Ptr to PN_SVBIS/allof parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_allof ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "allof" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Hand off to exec_sieve_xof */
    return ( exec_sieve_xof( nodeP, rvalP, FALSE ) );
}
/*

*//* exec_svbis_anyof( nodeP, rvalP )

	Execute sieve builtin statement: anyof

Accepts :

	nodeP		Ptr to PN_SVBIS/anyof parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_anyof ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "anyof" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Hand off to exec_sieve_xof */
    return ( exec_sieve_xof( nodeP, rvalP, TRUE ) );
}
/*

*//* exec_svbis_c( nodeP, rvalP )

	Execute sieve builtin statement: C

Accepts :

	nodeP		Ptr to PN_SVBIS/c parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_c ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "c" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Just evaluate the tree */
    return ( eval_tree( nodeP->pt_leftP, rvalP ) );
}
/*

*//* exec_svbis_deleteheader( nodeP, rvalP )

	Execute sieve builtin statement: deleteheader

Accepts :

	nodeP		Ptr to PN_SVBIS/deleteheader parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_deleteheader ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "deleteheader" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	BOOL		matchF;
	STSCODE		status;
	int		hdrC, hdrN;	/* Header count, index */
	int		hdrnum;		/* Header number needed */
	int		index;		/* Any :index number given */
	void		*hdrP;		/* Ptr to header handle */
	PTNODE		*vnodeP, *vwalkP;  /* Value node & walk node */
	ESTR		*hbstrP;	/* Header stringval */
	ESTR		*hnameP;	/* Header name wanted */
	ESTR		*vcompP;	/* Value to compare */
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

TODO	/* this function should be cleaned up to have fewer
	   return points (like, one)
	*/

    sieve_optvals_init( &optvals[0], &Svstmt_opt_deleteheader[0] );

    /* Requires editheader capability */
    if ( !sieve_cap_enabled( &Cap_editheader ) )
	status = SCNOTDONE;

    /* Get any arguments attached to this statement */
    else
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );

    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK )
	status = val_estr_eval( nodeP->pt_leftP, &hnameP, NULL );
    if ( status != SCOK ) {
	sieve_optvals_finit( &optvals[0], &Svstmt_opt_deleteheader[0] );
	val_set_int( rvalP, 4, 0 );
	return ( status );
    }

    /* Pick up any :index value given */
    if ( optvals[ SOC_INDEX ].ov_oov != SOV_NONE )
	index = optvals[ SOC_INDEX ].ov_val.ovv_int;
    else
	index = 0;

    /* If relative to last header, must count occurences */
    if ( optvals[ SOC_LAST ].ov_oov == SOV_SET ) {
	hdrP = NULL;
	for ( hdrC = 0;
	    ( hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP ) ) != NULL;
	        ++hdrC )
	    ;

	/* Set desired number.. */
	if ( index == 0 )
	    index = 1;
	hdrnum = hdrC + 1 - index;

	/* Have to make sure it's in range here. */
	if ( ( hdrnum < 1 ) || ( hdrnum > hdrC ) ) {
	    val_estr_finish( hnameP );
	    sieve_optvals_finit( &optvals[0], &Svstmt_opt_deleteheader[0] );
	    val_set_int( rvalP, 4, 1 );
	    return ( SCOK );
	}
    }
    else {
	/* Relative to first header, set desired number */
	hdrnum = index;
    }

    /* Walk through the headers.  If hdrnum is set, pay attention
       only to that header number.  Otherwise pay attention to
       all of them.
    */
    hdrP = NULL;
    for ( hdrN = 1; 
	 ( hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP ) ) != NULL;
	        ++hdrN ) {

	/* Pay attention to this if we must */
	if ( ( hdrnum == 0 ) || ( hdrN == hdrnum ) ) {
	    /* See if this test matches */
	    if ( ( vnodeP = nodeP->pt_rightP ) == NULL )
		/* No value pattern, apply this by default */
		matchF = TRUE;
	    else {
		hbstrP = msgpart_hdr_tail( OmmP->om_curmpP, hdrP );

		/* Walk through the pattern(s) */
		for ( vwalkP = vnodeP; vwalkP != NULL; ) {
		    if ( vwalkP->pt_type == PN_LSEQ ) {
			vnodeP = vwalkP->pt_rightP;
			vwalkP = vwalkP->pt_leftP;
		    }
		    else {
			vnodeP = vwalkP;
			vwalkP = NULL;
		    }

		    status = val_estr_eval( vnodeP, &vcompP, NULL );
		    if ( status != SCOK )
			break;

		    matchF = sieve_string_compare( &optvals[0],
						   vcompP, hbstrP );
		    val_estr_finish( vcompP );
		    if ( matchF )
			break;
		}

		estr_finish( hbstrP );
	    }

	    /* If we match, we can delete this. */
	    if ( status == SCOK ) {
		if ( matchF )
		    hdrP = msgpart_hdr_del( OmmP->om_curmpP, hdrP );
	    }

	    /* If a specific header number was given, we're done */
	    if ( ( status != SCOK ) || ( hdrnum > 0 ) )
		break;
	}
    }

    val_estr_finish( hnameP );

    sieve_optvals_finit( &optvals[0], &Svstmt_opt_deleteheader[0] );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( SCOK );
}
/*

*//* exec_svbis_discard( nodeP, rvalP )

	Execute sieve builtin statement: discard

Accepts :

	nodeP		Ptr to PN_SVBIS/discard parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_discard ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "discard" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Simply turn off implicit keep in the mail message */
    if ( OmmP != NULL ) {
	OmmP->om_ikeepF = FALSE;
	++OmmP->om_deliveryC;		/* Count a "delivery" */

	if ( Mfl_testF )
	    printf( "test: discard\n" );
	else
	    log_both( "Action: discard succeeded." );
    }

    val_set_int( rvalP, 4, 1 );		/* Return truth constant */

    return ( SCOK );
}
/*

*//* exec_svbis_dnsbl( nodeP, rvalP )

	Execute sieve builtin statement: dnsbl

Accepts :

	nodeP		Ptr to PN_SVBIS/dnsbl parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_dnsbl ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "dnsbl" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    sieve_optvals_init( &optvals[0], &Svstmt_opt_dnsbl[0] );

    /* Requires dnsbl capability */
    if ( !sieve_cap_enabled( &Cap_dnsbl ) )
	status = SCNOTDONE;

    /* Get any arguments attached to this statement */
    else
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );

    if ( ( status == SCOK ) &&
         ( optvals[ SOC_IP ].ov_oov == SOV_NONE ) &&
	 ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    else {
	/* Execute our handler routine for each cross product of the
	   list of dnsbl names and the list of result names
	*/
	status = exec_sieve_stringx2( &optvals[0],
				      nodeP->pt_leftP, nodeP->pt_rightP,
				      do_exec_svbis_dnsbl );
    }

    /* Clear up any option value data as needed */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_dnsbl[0] );

    /* SCDONE will indicate a match; return the result value from that */
    val_set_int( rvalP, 4, ( status == SCDONE ) ? 1 : 0 );

    /* And turn SCDONE into an OK result */
    if ( status == SCDONE )
	status = SCOK;

    return ( status );
}
/*

*//* exec_svbis_else( nodeP, rvalP )

	Execute sieve builtin statement: else

Accepts :

	nodeP		Ptr to PN_SVBIS/else parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	See description of "if" processing for notes about
	the X_ifF flag.

*/

STSCODE
exec_svbis_else ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "else" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* else only gets invoked if no code block in this if/elsif/else
       chain has yet been invoked.
    */
    if ( !X_ifF )
	status = exec_sieve_block( nodeP->pt_rightP, rvalP );
    else
	status = SCOK;			/* else skipped */

    return ( status );
}
/*

*//* exec_svbis_elsif( nodeP, rvalP )

	Execute sieve builtin statement: elsif

Accepts :

	nodeP		Ptr to PN_SVBIS/elsif parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	See description of "if" processing for notes about
	the X_ifF flag.

*/

STSCODE
exec_svbis_elsif ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "elsif" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* elsif only gets invoked if no code block in this if/elsif/else
       chain has yet been invoked.
    */
    if ( !X_ifF ) {
	/* Evaluate the condition attached to the left */
	status = exec_sieve( nodeP->pt_leftP, rvalP );

	if ( status == SCOK ) {
	    /* Execute the code block if the test was true */
	    if ( sieve_test_result( rvalP ) ) {
		status = exec_sieve_block( nodeP->pt_rightP, rvalP );

		/* Set the "if block was executed" flag */
		X_ifF = TRUE;
	    }
	}
    }
    else  {
	/* elsif skipped */
	status = SCOK;
	val_set_int( rvalP, 4, 1 );
    }

    return ( status );
}
/*

*//* exec_svbis_envelope( nodeP, rvalP )

	Execute sieve builtin statement: envelope

Accepts :

	nodeP		Ptr to PN_SVBIS/envelope parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_envelope ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "envelope" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    sieve_optvals_init( &optvals[0], &Svstmt_opt_envelope[0] );

    /* Requires envelope capability */
    if ( !sieve_cap_enabled( &Cap_envelope ) )
	status = SCNOTDONE;

    else if ( OmmP == NULL ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    else
	/* Get any arguments attached to this statement */
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );

    if ( status == SCOK ) {
	/* Execute our handler routine for each cross product of the
	   list of headers and the list of addresses
	*/
	status = exec_sieve_stringx2( &optvals[0],
				      nodeP->pt_leftP, nodeP->pt_rightP,
				      do_exec_svbis_envelope );
    }

    /* Do any needed cleanup of option values data */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_envelope[0] );

    /* SCDONE will indicate a match; return the result value from that */
    val_set_int( rvalP, 4, ( status == SCDONE ) ? 1 : 0 );

    /* And turn SCDONE into an OK result */
    if ( status == SCDONE )
	status = SCOK;

    return ( status );
}
/*

*//* exec_svbis_exists( nodeP, rvalP )

	Execute sieve builtin statement: exists

Accepts :

	nodeP		Ptr to PN_SVBIS/exists parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_exists ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "exists" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	PTNODE		*hnodeP;	/* Ptr to header string node */
	void		*hdrP;		/* Ptr to header handle */
	ESTR		*hnameP;	/* Header name */

    status = SCOK;
    hdrP = NULL;

    if ( OmmP == NULL ) {
	mfl_warning( NULL, "There is no mail file open." );
	val_set_int( rvalP, 4, 0 );
	return ( SCNOTDONE );
    }

    /* Walk through the strings attached to left side, making
       sure each of them exists in the mail header.
    */
    for ( nodeP = nodeP->pt_leftP; ( nodeP != NULL ) && ( status == SCOK ); ) {

	if ( nodeP->pt_type == PN_LSEQ ) {
	    hnodeP = nodeP->pt_rightP;
	    nodeP = nodeP->pt_leftP;
	}
	else {
	    hnodeP = nodeP;
	    nodeP = NULL;
	}

	/* Access the string value */
	status = val_estr_eval( hnodeP, &hnameP, NULL );
	if ( status != SCOK )
	    break;

	/* Attempt to look up this header */
	hdrP = NULL;
	if ( OmmP != NULL )
	    hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP );

	val_estr_finish( hnameP );

	/* Stop if header not found. */
	if ( hdrP == NULL )
	    break;
    }

    /* set result */
    val_set_int( rvalP, 4, ( hdrP == NULL ? 0 : 1 ) );

    return ( status );
}
/*

*//* exec_svbis_false( nodeP, rvalP )

	Execute sieve builtin statement: false

Accepts :

	nodeP		Ptr to PN_SVBIS/false parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_false ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "false" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Set the return value to zero */
    val_set_int( rvalP, 4, 0 );

    /* All set */
    return ( SCOK );
}
/*

*//* exec_svbis_fileinto( nodeP, rvalP )

	Execute sieve builtin statement: fileinto

Accepts :

	nodeP		Ptr to PN_SVBIS/fileinto parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	Requires "fileinto" capability.

*/

STSCODE
exec_svbis_fileinto ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "fileinto" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	BOOL		okF;
	STSCODE		status;
	ESTR		*folderP;	/* Folder name */
	char		*fstrP;		/* Folder name string */
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    sieve_optvals_init( &optvals[0], &Svstmt_opt_fileinto[0] );

    /* Requires fileinto capability */
    if ( !sieve_cap_enabled( &Cap_fileinto ) )
	status = SCNOTDONE;

    /* Get any arguments attached to this statement */
    else
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );

    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	status = val_estr_eval( nodeP->pt_leftP, &folderP, NULL );
	if ( status == SCOK ) {
	    /* Only do this see if we've not attempted this before */
	    if ( !omm_action_check( OmmP,
				   estr_tmp_str( (UBYTE *)"fileinto" ),
				   folderP ) ) {

		/* Register the action attempt regardless 
		   of success or failure.
		*/
		omm_action_add( OmmP,
			       estr_tmp_str( (UBYTE *)"fileinto" ),
			       folderP );

		fstrP = (char *)tmpstr_estr( folderP );
		if ( *fstrP == '|' ) {	/* Check pipe request */
		    if ( ! cntrl_int_get( estr_tmp_str( "pipe_allow" ), 0 ) ) {
			mfl_warning( NULL, "pipe not allowed." );
			okF = FALSE;
		    }
		    else {
			if ( Mfl_testF ) {
			    okF = TRUE;
			    printf( "test: fileinto pipe to \"%s\"\n",
				       &fstrP[1] );
			}
			else
			    okF = msg_pipe( OmmP, &fstrP[1] );
		    }
		}
		else {
		    if ( Mfl_testF ) {
			okF = TRUE;
		        printf( "test: fileinto \"%s\"\n", fstrP );
		    }
		    else
			okF = msg_write( OmmP, fstrP );
		}
		if ( okF ) {
		    if ( !Mfl_testF )
			log_both( "Action: fileinto \"%s\" succeeded.",
				  tmpstr_estr( folderP ) );

		    /* Turn off implicit keep unless overriden
		       by :copy option
		    */
		    if ( optvals[ SOC_COPY ].ov_oov != SOV_SET )
			OmmP->om_ikeepF = FALSE;
		}
		else {
		    status = SCNOTDONE;
		    log_both( "Action: fileinto \"%s\" failed.",
				  tmpstr_estr( folderP ) );
		}
	    }
	    else {
		/* Duplicate action- flag OK */
		status = SCOK;
		if ( Mfl_testF )
		    printf( "test: duplicate fileinto \"%s\"\n",
			     (char *)tmpstr_estr( folderP ) );
	    }
	    val_estr_finish( folderP );
	}
    }

    /* Do any needed cleanup of option values data */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_fileinto[0] );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    return ( status );
}
/*

*//* exec_svbis_header( nodeP, rvalP )

	Execute sieve builtin statement: header

Accepts :

	nodeP		Ptr to PN_SVBIS/header parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_header ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "header" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    /* Get any arguments attached to this statement */
    sieve_optvals_init( &optvals[0], &Svstmt_opt_header[0] );
    status = exec_sieve_get_bisargs( nodeP, &optvals[0] );
    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	/* Execute our handler routine for each cross product of the
	    list of headers and the list of patterns 
	*/
	status = exec_sieve_stringx2( &optvals[0],
				      nodeP->pt_leftP, nodeP->pt_rightP,
				      do_exec_svbis_header );
    }

    /* Finalize any option values data */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_header[0] );

#if 0
    /* SCDONE will indicate a match; return the result value from that */
    val_set_int( rvalP, 4, ( status == SCDONE ) ? 1 : 0 );
#else
    if ( status == SCDONE ) {
	/* SCDONE indicates success; as an added bonus, if the count
	   is non-zero, give that as the test result.
	*/
	val_set_int( rvalP, 4, ( Countval == 0 ) ? 1 : Countval );
	status = SCOK;
    }
    else
	/* No match, always return zero */
	val_set_int( rvalP, 4, 0 );
#endif

    return ( status );
}
/*

*//* exec_svbis_if( nodeP, rvalP )

	Execute sieve builtin statement: if

Accepts :

	nodeP		Ptr to PN_SVBIS/if parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	X_ifF maintains a running state of whether any block
	in this if/elsif/else chain has been invoked.  The parser
	ensures that if/elsif/else statements are assembled
	only where they are allowed, so the X_ifF flag only
	has to be initialized by the "if" parser.  It's preserved
	in the execution context when a sieve block is entered
	(and restored when left).

*/

STSCODE
exec_svbis_if ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "if node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* Clear the if-executed flag for this if/elsif/else chain.
       This will be set to TRUE if any block in if/elsif is
       executed, preventing subsequent blocks from being
       attempted or invoked.
    */
    X_ifF = FALSE;

    /* Evaluate the condition attached to the left */
    status = exec_sieve( nodeP->pt_leftP, rvalP );

    if ( status == SCOK ) {
	/* Execute the code block if the test was true */
	if ( sieve_test_result( rvalP ) ) {
	    status = exec_sieve_block( nodeP->pt_rightP, rvalP );

	    /* Set the "if block was executed" flag */
	    X_ifF = TRUE;
	}
    }

    return ( status );
}
/*

*//* exec_svbis_keep( nodeP, rvalP )

	Execute sieve builtin statement: keep

Accepts :

	nodeP		Ptr to PN_SVBIS/keep parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	This writes the message to the default mailbox, and
	in that regard is just like "fileinto" with the default
	mailbox as its argument-- except that the "fileinto"
	capability is not required.

	We could combine this with the fileinto code, however
	for now we'll keep them separate.

*/

STSCODE
exec_svbis_keep ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "keep node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    if ( OmmP == NULL ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }
    else {
	/* Only do this see if we've not attempted this before */
	if ( !omm_action_check( OmmP,
				estr_tmp_str( (UBYTE *)"fileinto" ),
				estr_tmp_str( (UBYTE *)MailboxP ) ) ) {
	    /* Register the action attempt regardless 
		of success or failure.
	    */
	    omm_action_add( OmmP,
			   estr_tmp_str( (UBYTE *)"fileinto" ),
			   estr_tmp_str( (UBYTE *)MailboxP ) );

	    if ( Mfl_testF ) {
		printf( "test: keep\n" );
		OmmP->om_ikeepF = FALSE;
	    }
	    else if ( msg_write( OmmP, MailboxP ) ) {
		status = SCOK;

		log_both( "Action: keep succeeded." );

		/* Turn off implicit keep */
		OmmP->om_ikeepF = FALSE;
	    }
	    else {
		status = SCNOTDONE;
		log_both( "Action: keep failed." );
	    }
	}
	else {
	    /* Duplicate action- flag OK */
	    status = SCOK;
	    if ( Mfl_testF )
		printf( "test: duplicate keep\n" );
	}
    }
	
    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    return ( status );
}
/*

*//* exec_svbis_not( nodeP, rvalP )

	Execute sieve builtin statement: not

Accepts :

	nodeP		Ptr to PN_SVBIS/not parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_not ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "not" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* Evaluate the test block hanging off the left */
    status = exec_sieve( nodeP->pt_leftP, rvalP );

    if ( status == SCOK ) {
	/* Invert the result */
	val_set_int( rvalP, 4, !sieve_test_result( rvalP ) );
    }

    return ( status );
}
/*

*//* exec_svbis_redirect( nodeP, rvalP )

	Execute sieve builtin statement: redirect

Accepts :

	nodeP		Ptr to PN_SVBIS/redirect parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_redirect ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "redirect" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	BOOL		okF;
	STSCODE		status;
	ESTR		*recipP;	/* Recipient address */
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    sieve_optvals_init( &optvals[0], &Svstmt_opt_redirect[0] );
    status = exec_sieve_get_bisargs( nodeP, &optvals[0] );
    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	status = val_estr_eval( nodeP->pt_leftP, &recipP, NULL );
	if ( status == SCOK ) {
	    /* Only do this see if we've not attempted this before */
	    if ( !omm_action_check( OmmP,
				   estr_tmp_str( (UBYTE *)"redirect" ),
				   recipP ) ) {

		/* Register the action attempt regardless 
		   of success or failure.
		*/
		omm_action_add( OmmP,
			        estr_tmp_str( (UBYTE *)"redirect" ),
				recipP );

		if ( Mfl_testF ) {
		    printf( "test: redirect \"%s\"\n",
			     tmpstr_estr( recipP ) );
		    okF = TRUE;
		}
		else {
		    okF = msg_forward( OmmP,
				 (char *)tmpstr_estr( recipP ), TRUE );
		    if ( okF )
			log_both( "Action: redirect \"%s\" succeeded.",
			           tmpstr_estr( recipP ) );
		    else {
			status = SCNOTDONE;
			log_both( "Action: redirect \"%s\" failed.",
			           tmpstr_estr( recipP ) );
		    }
		}
		if ( okF ) {
		    /* Turn off implicit keep unless overriden
		       by :copy option
		    */
		    if ( optvals[ SOC_COPY ].ov_oov != SOV_SET )
			OmmP->om_ikeepF = FALSE;
		}
	    }
	    else {
		/* Duplicate action- flag OK */
		status = SCOK;
		if ( Mfl_testF )
		    printf( "test: duplicate redirect \"%s\"\n",
			     tmpstr_estr( recipP ) );
	    }
	    val_estr_finish( recipP );
	}
    }

    /* Clear up any option values data as needed */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_redirect[0] );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    return ( status );
}
/*

*//* exec_svbis_reject( nodeP, rvalP )

	Execute sieve builtin statement: reject

Accepts :

	nodeP		Ptr to PN_SVBIS/reject parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_reject ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "reject" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	BOOL		okF;
	STSCODE		status;
	ESTR		*reasonP;

    /* Requires reject capability */
    if ( !sieve_cap_enabled( &Cap_reject ) )
	status = SCNOTDONE;

    else if ( OmmP == NULL ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }
    else if ( OmmP->om_rejectF ) {
	/* RFC3028 says multiple rejects MUST be prohibited */
	mfl_warning( NULL, "\"reject\" only allowed once." );

	/* However it doesn't say we can't continue.. so let's. */
	status = SCOK;
    }
    else {
	status = val_estr_eval( nodeP->pt_leftP, &reasonP, NULL );
	if ( status == SCOK ) {
	    /* Only do this see if we've not attempted this before */
	    if ( !omm_action_check( OmmP,
				    estr_tmp_str( (UBYTE *)"reject" ),
				    estr_tmp_str( (UBYTE *)"any" ) ) ) {

		/* Register the action attempt regardless 
		   of success or failure.
		*/
		omm_action_add( OmmP,
			        estr_tmp_str( (UBYTE *)"reject" ),
				estr_tmp_str( (UBYTE *)"any" ) );

		if ( Mfl_testF ) {
		    printf( "test: reject \"%s\"\n",
			     tmpstr_estr( reasonP ) );
		    okF = TRUE;
		}
		else {
		    okF = msg_bounce( OmmP, (char *)tmpstr_estr( reasonP ) );
		    if ( okF )
			log_both( "Action: reject \"%s\" succeeded.",
				    tmpstr_estr( reasonP ) );
		    else {
			status = SCNOTDONE;
			log_both( "Action: reject \"%s\" failed.",
					tmpstr_estr( reasonP ) );
		    }
		}
		if ( okF ) {
		    /* Turn off implicit keep */
		    OmmP->om_ikeepF = FALSE;

		    /* Remember we did a reject */
		    OmmP->om_rejectF = TRUE;
		}
	    }
	    else {
		/* Duplicate action- flag OK */
		status = SCOK;
		if ( Mfl_testF )
		    printf( "test: duplicate reject \"%s\"\n",
			     tmpstr_estr( reasonP ) );
	    }
	}
	val_estr_finish( reasonP );
    }
	
    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    return ( status );
}
/*

*//* exec_svbis_replaceheader( nodeP, rvalP )

	Execute sieve builtin statement: replaceheader

Accepts :

	nodeP		Ptr to PN_SVBIS/relaceheader parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_replaceheader ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "replaceheader" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	BOOL		matchF;
	STSCODE		status;
	int		hdrC, hdrN;	/* Header count, index */
	int		hdrnum;		/* Header number needed */
	int		index;		/* Any :index number given */
	char		*newnameP;	/* Any :newname value specified */
	char		*newvalueP;	/* Any :newvalue value specified */
	void		*hdrP;		/* Ptr to header handle */
	PTNODE		*vnodeP, *vwalkP;  /* Value node & walk node */
	ESTR		*hbstrP;	/* Header string */
	ESTR		*hnameP;	/* Header name */
	ESTR		*vcompP;	/* Value to compare */
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

TODO	/* This function needs fewer return points (e.g. one) */

    sieve_optvals_init( &optvals[0], &Svstmt_opt_replaceheader[0] );

    /* Requires fileinto capability */
    if ( !sieve_cap_enabled( &Cap_editheader ) )
	status = SCNOTDONE;

    /* Get any arguments attached to this statement */
    else
	status = exec_sieve_get_bisargs( nodeP, &optvals[0] );

    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	/* Pick up the new string values given, if any */
	if ( optvals[ SOC_NEWNAME ].ov_oov == SOV_SET )
	    newnameP = (char *)optvals[ SOC_NEWNAME ].ov_val.ovv_bP;
	else
	    newnameP = NULL;
	if ( optvals[ SOC_NEWVALUE ].ov_oov == SOV_SET )
	    newvalueP = (char *)optvals[ SOC_NEWVALUE ].ov_val.ovv_bP;
	else
	    newvalueP = NULL;

	/* Pick up any :index value given */
	if ( optvals[ SOC_INDEX ].ov_oov != SOV_NONE )
	    index = optvals[ SOC_INDEX ].ov_val.ovv_int;
	else
	    index = 0;

	/* Do idiot check to make sure some new value has been supplied */
	if ( ( newnameP == NULL ) && ( newvalueP == NULL ) ) {
	    mfl_warning( NULL, "replaceheader with no new strings!" );
	    status = SCNOTDONE;
	}
    }

    if ( status == SCOK )
	/* Access the desired header name */
	status = val_estr_eval( nodeP->pt_leftP, &hnameP, NULL );

    if ( status != SCOK ) {
	sieve_optvals_finit( &optvals[0], &Svstmt_opt_replaceheader[0] );
	val_set_int( rvalP, 4, 0 );
	return ( status );
    }

    /* If relative to last header, must count occurences */
    if ( optvals[ SOC_LAST ].ov_oov == SOV_SET ) {
	hdrP = NULL;
	for ( hdrC = 0;
	    ( hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP ) ) != NULL;
	        ++hdrC )
	    ;

	/* Set desired number.. */
	if ( index == 0 )
	    index = 1;
	hdrnum = hdrC + 1 - index;

	/* Have to make sure it's in range here. */
	if ( ( hdrnum < 1 ) || ( hdrnum > hdrC ) ) {
	    val_estr_finish( hnameP );
	    sieve_optvals_finit( &optvals[0], &Svstmt_opt_replaceheader[0] );
	    val_set_int( rvalP, 4, 1 );
	    return ( SCOK );
	}
    }
    else {
	/* Relative to first header, set desired number */
	hdrnum = index;
    }

    /* Walk through the headers.  If hdrnum is set, pay attention
       only to that header number.  Otherwise pay attention to
       all of them.
    */
    hdrP = NULL;
    for ( hdrN = 1; 
	 ( hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP ) ) != NULL;
	        ++hdrN ) {

	/* Pay attention to this if we must */
	if ( ( hdrnum == 0 ) || ( hdrN == hdrnum ) ) {
	    /* See if this test matches */
	    if ( ( vnodeP = nodeP->pt_rightP ) == NULL )
		/* No value pattern, apply this by default */
		matchF = TRUE;
	    else {
		hbstrP = msgpart_hdr_tail( OmmP->om_curmpP, hdrP );

		/* Walk through the pattern(s) */
		for ( vwalkP = vnodeP; vwalkP != NULL; ) {
		    if ( vwalkP->pt_type == PN_LSEQ ) {
			vnodeP = vwalkP->pt_rightP;
			vwalkP = vwalkP->pt_leftP;
		    }
		    else {
			vnodeP = vwalkP;
			vwalkP = NULL;
		    }

		    status = val_estr_eval( vnodeP, &vcompP, NULL );
		    if ( status != SCOK )
			break;

		    matchF = sieve_string_compare( &optvals[0],
						   vcompP, hbstrP );
		    val_estr_finish( vcompP );
		    if ( matchF )
			break;
		}
	    }

	    if ( status == SCOK ) {
		/* If we match, we can do the replacement on this. */
		if ( matchF )
		    if ( !msgpart_hdr_change( OmmP->om_curmpP, hdrP,
				 estr_tmp_str( (UBYTE *)newnameP ),
				 estr_tmp_str( (UBYTE *)newvalueP ) ) )
			status = SCNOTDONE;
	    }

	    /* If a specific header number was given, we're done */
	    if ( ( status != SCOK ) || ( hdrnum > 0 ) )
		break;
	}
    }

    val_estr_finish( hnameP );

    /* Clean up any option values data that needs cleaning up */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_replaceheader[0] );

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    return ( status );
}
/*

*//* exec_svbis_require( nodeP, rvalP )

	Execute sieve builtin statement: require

Accepts :

	nodeP		Ptr to PN_SVBIS/require parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_require ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "require" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	int		capX;		/* Capabilities table index */
	STSCODE		status;
	SIEVE_CAP	*capP;		/* Ptr to capability */
	PTNODE		*cnodeP;	/* Ptr to capability string node */
	ESTR		*cnameP;	/* Capability name */

    status = SCOK;

    /* Walk through the arguments attached to left side */
    for ( nodeP = nodeP->pt_leftP; ( nodeP != NULL ) && ( status == SCOK ); ) {
	if ( nodeP->pt_type == PN_LSEQ ) {
	    cnodeP = nodeP->pt_rightP;
	    nodeP = nodeP->pt_leftP;
	}
	else {
	    cnodeP = nodeP;
	    nodeP = NULL;
	}

	/* Access and test the specified capability string */
	status = val_estr_eval( cnodeP, &cnameP, NULL );
	if ( status == SCOK ) {
	    /* Find the capability if we support it */
	    for ( capX = 0; ( capP = Sieve_caps[ capX ] ) != NULL; ++capX )
		if ( estr_cmp_str( cnameP, (UBYTE *)capP->cap_nameP ) == 0 )
		    break;

	    /* Check the aliases table if necessary */
	    if ( capP == NULL ) {
		for ( capX = 0;
		       Sieve_cap_aliases[capX].ca_nameP != NULL;
		          ++capX ) {
		    if ( estr_cmp_str( cnameP,
			  (UBYTE *)Sieve_cap_aliases[capX].ca_nameP ) == 0 ) {
			capP = Sieve_cap_aliases[capX].ca_capP;
			break;
		    }
		}
	    }

	    if ( capP != NULL ) {
		/* Found the capability- enable it */
		capP->cap_enabledF = TRUE;
	    }
	    else {
		mfl_warning( NULL,
			      "Desired capability \"%s\" not supported.",
			      tmpstr_estr( cnameP ) );
		status = SCNOTDONE;
	    }

	    /* Done with string value */
	    val_estr_finish( cnameP );
	}
    }

    val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );
    return ( status );
}
/*

*//* exec_svbis_sieve( nodeP, rvalP )

	Execute sieve builtin statement: sieve

Accepts :

	nodeP		Ptr to PN_SVBIS/sieve parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_sieve ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "sieve" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Just evaluate the tree */
    return ( eval_tree( nodeP->pt_leftP, rvalP ) );
}
/*

*//* exec_svbis_size( nodeP, rvalP )

	Execute sieve builtin statement: size

Accepts :

	nodeP		Ptr to PN_SVBIS/size parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	The size of the message will necessarily be approximate, as it
	is a mix of internal storage size and size of the input message;
	the actual message stored on disk can be different from these
	due to header folding, header canonicalization, line-termination
	storage (or not), and so forth.

*/

STSCODE
exec_svbis_size ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "size" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;
	BOOL		resultF;
	ULONG		bodysize;	/* Approximate body size */
	PTNODE		*valnodeP;
	SIEVE_OPTVAL	optvals[ SOC_MAX ];

    /* Get the arguments attached to this statement */
    sieve_optvals_init( &optvals[0], &Svstmt_opt_size[0] );
    status = exec_sieve_get_bisargs( nodeP, &optvals[0] );
    if ( ( status == SCOK ) && ( OmmP == NULL ) ) {
	mfl_warning( NULL, "There is no mail file open." );
	status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	/* Get the approximate message size, including child parts */
	bodysize = msgpart_size_body( OmmP->om_curmpP, TRUE );

	/* Process now according to the over/under value */
	valnodeP = nodeP->pt_rightP;

	switch ( optvals[ SOC_OVER_UNDER ].ov_oov ) {
	    case	SOV_OVUN_OVER:
		resultF = ( bodysize > valnodeP->pt_data.pv_long );
		break;

	    case	SOV_OVUN_UNDER:
		resultF = ( bodysize < valnodeP->pt_data.pv_long );
		break;

	    default:
		mfl_warning( NULL,
			  "Bad :over/:under value (%d) in exec_svbis_size!",
			   optvals[ SOC_OVER_UNDER ].ov_oov );
		status = SCERR;
		break;
	}
    }

    /* Clean up any evaluated option values data */
    sieve_optvals_finit( &optvals[0], &Svstmt_opt_size[0] );

    if ( status == SCOK )
	val_set_int( rvalP, 4, resultF ? 1 : 0 );
    else
	val_set_int( rvalP, 4, 0 );

    return ( status );
}
/*

*//* exec_svbis_stop( nodeP, rvalP )

	Execute sieve builtin statement: stop

Accepts :

	nodeP		Ptr to PN_SVBIS/stop parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
exec_svbis_stop ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "stop" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* All we do is return DONE.  This should stop processing
       without fuss.
    */
    val_set_int( rvalP, 4, 1 );
    return ( SCDONE );
}
/*

*//* exec_svbis_true( nodeP, rvalP )

	Execute sieve builtin statement: true

Accepts :

	nodeP		Ptr to PN_SVBIS/true parse tree node
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :



*/

STSCODE
exec_svbis_true ARGLIST( (nodeP, rvalP ) )
   NFARG( PTNODE	*nodeP )	/* Address of "true" node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
    /* Set the return value to one */
    val_set_int( rvalP, 4, 1 );

    /* All set */
    return ( SCOK );
}
/*

*//* parse_bis_sieve( inP, nodeP )

	Parse a sieve construct.

Accepts :

	inP		Ptr to input stack
	nodeP		Built-in-statement node
	
Returns :

	<value>		SCxx status code

Notes :

	Simply calls parse_sieve_compound to parse the body and
	hang it off of the node we receive here.  The input node
	is a PN_BIS node, this this will result in:

			PN_BIS
			  /
			 /
			/
		first statement or term


	See parse_sieve_compound for more.
	


*/

STSCODE
parse_bis_sieve ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* BIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Next token must be opening brace */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	mfl_warning( inP, "'{' required for sieve block, got \"%s\".",
		     tokP == NULL ? "<NUL>" :
			(char *)&tokP->tk_bstrP->bs_bP[0] );
	status = SCNOTDONE;
    }
    else {
	/* parse_sieve_compound will handle the body. */
	status = parse_sieve_compound( inP, &nodeP->pt_leftP );
    }

    return ( status );
}
/*

*//* parse_sieve( inP, treePP, oktypes )

	Parse sieve element (top level sieve parse)

Accepts :

	inP		Ptr to input stack
	treePP		Where to put the top tree node
	oktypes		A bitmask specifying the types of statements
			that are OK here.

Returns :

	<value>		SCxx status code

Notes :

	oktypes is a bitmask composed of SBIS_xxx types, indicating
	which of the statement types (test, control, action) are
	OK to accept here.  We are fairly loose with SIEVE statements,
	allowing all kinds of statements to be used in many
	places.  However, in some places (such as in the condition
	of an IF statement), some types are restricted.

*/

STSCODE
parse_sieve ARGLIST( (inP, treePP, oktypes) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to store the tree */
    FARG( UWORD		oktypes )	/* OK statement types */
{
	int		bisX;		/* Index into statement table */
	STSCODE		status;
	PTNODE		*nodeP;
	TOKEN		*tokP;

    /* First significant thing has to be a keyword. */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( ( tokP->tk_type != TK_SYM ) &&
	   ( tokP->tk_type != TK_BIS ) ) ) {
	mfl_warning( inP, "Keyword expected." );
	return ( SCNOTDONE );
	 }

    /* Attempt to look up the token as a statement keyword */
    for ( bisX = 0; Sieve_bistbl[ bisX ].sb_kwdP != NULL; ++bisX )
	if ( stricmp( (char *)tokP->tk_bstrP->bs_bP,
		     Sieve_bistbl[ bisX ].sb_kwdP ) == 0 )
	    break;

    if ( Sieve_bistbl[ bisX ].sb_kwdP == NULL ) {
	mfl_warning( inP, "Unknown sieve statement: \"%s\" .",
		tokP->tk_bstrP->bs_bP );
	status = SCNOTDONE;
    }
    else if ( ( Sieve_bistbl[ bisX ].sb_type & oktypes ) == 0 ) {
	mfl_warning( inP, "Sieve statement \"%s\" not allowed here.",
		tokP->tk_bstrP->bs_bP );
	status = SCNOTDONE;
    }
    else {
	/* Make node to hold this statement. */
	status = parsetree_mknode( &nodeP, PN_SVBIS );
	if ( status == SCOK ) {
	    nodeP->pt_data.pv_int = bisX;	/* Type the statement */

	    /* Call the parse routine.. */
	    status = (*Sieve_bistbl[ bisX ].sb_parse)( inP, nodeP );

	    if ( status != SCOK ) {
		/* Failure- get rid of node. */
		parsetree_free( nodeP );
	    }
	    else {
		/* Remember any context about the last thing
		   successfully parsed.  Perhaps we should remember
		   this about failed parses too..
		*/
		if ( Sieve_bistbl[ bisX ].sb_type == SBIS_CONTROL )
		    P_lastctlX = bisX;
	    }
	}
    }

    *treePP = ( status == SCOK ? nodeP : NULL );

    return ( status );
}
/*

*//* parse_sieve_compound( inP, treePP )

	Parse a compound sieve element


Accepts :

	inP		Ptr to input stack
	treePP		Location of pointer to attach parse tree to

Returns :

	<value>		SCxx status code

Notes :

	A compound sieve element is simply any sieve language block
	enclosed in curly braces like { ... }.  The block can
	contain zero or more sieve statements.

	It's assumed that the caller has already parsed the '{' .

	"treePP" provides a pointer to hang the first statement on.
	If there's only one statement in the compound list, we
	just return a pointer to that statement:

		treePP -> statement

	If there are multiple nodes, we return a pointer to an LSEQ
	list, like this:

		treePP->  LSEQ
			   /\
			  /  \
			 /   first statement  
		        /
		      LSEQ
		       /\
		      /  \
		     /   second statement
		    /
		  ...
		  /  
		last statement

	   (etc).. i.e. just like any other statement list.

*/

STSCODE
parse_sieve_compound ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to hang the result */
{
	STSCODE		status;
	PTNODE		*headP;		/* Ptr to the entire list, we'll
					    return this. */
	PTNODE		*nodeP;		/* New node being made */
	PTNODE		*seqP;		/* Ptr to an LSEQ node */
	PTNODE		**tailPP;	/* Any pointer to the end node */
	TOKEN		*tokP;

    /* Parsing context that needs saving between blocks */
	int		p_lastctlX;	/* Previous P_lastctlX */


    headP = NULL;			/* No list yet */
    tailPP = NULL;			/* Nothing points to the end */

    /* Push previous parsing context (bleh) */
    p_lastctlX = P_lastctlX;

    /* Init new parsing context */
    P_lastctlX = -1;

    /* Loop gathering statements until we reach an end brace */
    for ( status = SCOK; status == SCOK; ) {
	tokP = mfl_sigtoken( inP );

	if ( tokP == NULL )
	    break;

	/* If it's a naked semicolon, just ignore it. */
	if ( ( tokP->tk_type == TK_CHAR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == ';' ) )
	    continue;

	/* If it's end brace, we are done */
	if ( ( tokP->tk_type == TK_CHAR ) &&
	     ( tokP->tk_bstrP->bs_bP[0] == '}' ) )
	    break;

	/* Take back this token and parse a sieve statement */
	mfl_tokback( inP, tokP );
	status = parse_sieve( inP, &nodeP,
			       SBIS_ACTION | SBIS_CONTROL | SBIS_TEST );
	if ( status != SCOK )
	    break;

	/* Avoid making an LSEQ for the last node in this block.
	   We do this by remembering the linkage around the last
	   parsed node and relinking if there turn out to be more
	   statements after that one.
	*/
	if ( headP == NULL ) {
	    /* This is the first statement, just remember it */
	    headP = nodeP;
	    tailPP = &headP;		/* Pointer to last node ptr */
	}
	else {
	    /* This is not the first statement, which means that the
	       last one we parsed is hanging at the end and needs to
	       be replaced with an LSEQ.
	    */
	    status = parsetree_mknode( &seqP, PN_LSEQ );
	    if ( status == SCOK ) {
		/* Unlink the tail node from the end and replace
		   it with the lseq sequence
		*/
		seqP->pt_rightP = *tailPP;
		*tailPP = seqP;

		/* Point the LSEQ to the new tail, and remember the
		   tail pointer.
		*/
		seqP->pt_leftP = nodeP;
		tailPP = &(seqP->pt_leftP);
	    }
	}
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( headP != NULL ) ) {
    	parsetree_free( headP );
	headP = NULL;
    }

    /* Return the tree if any */
    *treePP = headP;

    /* Pop saved parsing context */
    P_lastctlX = p_lastctlX;

    return( status );
}
/*

*//* parse_sieve_compound_test( inP, treePP )

	Parse a sieve compound test group


Accepts :

	inP		Ptr to input stack
	treePP		Location of pointer to attach parse tree to

Returns :

	<value>		SCxx status code

Notes :

	A compound test group is one or more test statements separated
	by commas, all enclosed in parens.  It's used as the argument
	for grouping keywords such as "allof" and "anyof".

	"treePP" provides a pointer to hang the first test statement
	on.  If there's only one statement in the compound list, we
	just return a pointer to that statement:

		treePP -> statement

	If there are multiple nodes, we return a pointer to an LSEQ
	list, like this:

		treePP->  LSEQ
			   /\
			  /  \
			 /   first statement  
		        /
		      LSEQ
		       /\
		      /  \
		     /   second statement
		    /
		  ...
		  /  
		last statement

	   (etc).. i.e. just like any other statement list.

*/

STSCODE
parse_sieve_compound_test ARGLIST( (inP, treePP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	**treePP )	/* Where to hang the result */
{
	STSCODE		status;
	PTNODE		*headP;		/* Ptr to the entire list, we'll
					    return this. */
	PTNODE		*nodeP;		/* New node being made */
	PTNODE		*seqP;		/* Ptr to an LSEQ node */
	PTNODE		**tailPP;	/* Any pointer to the end node */
	TOKEN		*tokP;

    headP = NULL;			/* No list yet */
    tailPP = NULL;			/* Nothing points to the end */

    /* Opening paren must be first */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
	 ( tokP->tk_type != TK_OPR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '(' ) ) {
	mfl_warning( inP, "test-list must be enclosed in parentheses." );
	return ( SCNOTDONE );
    }

    /* Loop gathering statements until one is not followed
       by a comma.
    */
    for ( status = SCOK; status == SCOK; ) {
	status = parse_sieve( inP, &nodeP,
			        SBIS_TEST );
	if ( status != SCOK )
	    break;

	/* Avoid making an LSEQ for the last node in this block.
	   We do this by remembering the linkage around the last
	   parsed node and relinking if there turn out to be more
	   statements after that one.
	*/
	if ( headP == NULL ) {
	    /* This is the first statement, just remember it */
	    headP = nodeP;
	    tailPP = &headP;		/* Pointer to last node ptr */
	}
	else {
	    /* This is not the first statement, which means that the
	       last one we parsed is hanging at the end and needs to
	       be replaced with an LSEQ.
	    */
	    status = parsetree_mknode( &seqP, PN_LSEQ );
	    if ( status == SCOK ) {
		/* Unlink the tail node from the end and replace
		   it with the lseq sequence
		*/
		seqP->pt_rightP = *tailPP;
		*tailPP = seqP;

		/* Point the LSEQ to the new tail, and remember the
		   tail pointer.
		*/
		seqP->pt_leftP = nodeP;
		tailPP = &(seqP->pt_leftP);
	    }
	}

	/* Get next token- it should either be a comma for more
	   to come, or a close paren to end the list.
	*/
	if ( status == SCOK ) {
	    tokP = mfl_sigtoken( inP );

	    if ( tokP != NULL ) {
	        if ( ( tokP->tk_type == TK_CHAR ) &&
		     ( tokP->tk_bstrP->bs_bP[0] == ')' ) )
		    break;

		if ( ( tokP->tk_type == TK_OPR ) &&
		     ( tokP->tk_bstrP->bs_bP[0] == ',' ) )
		    continue;
	    }

	    /* Oops- something isn't right. */
	    mfl_warning( inP, "',' or ')' expected." );
	    status = SCNOTDONE;
	}
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( headP != NULL ) ) {
    	parsetree_free( headP );
	headP = NULL;
    }

    /* Return the tree if any */
    *treePP = headP;

    return( status );
}
/*

*//* parse_sieve_string( inP, treePP, multiF )

	Parse a SIEVE string or string list

Accepts :

	inP		Ptr to input stack
	treePP		Location of pointer to attach parse tree to
	multiF		TRUE if string list is OK
			FALSE if only a single string is allowed

Returns :

	<value>		SCxx status code

Notes :

	A SIEVE string can either be a single string element,
	  like "blah", or it can be a list of strings enclosed
	  in square brackets, like [ "blah", "blah" ].  A string
	  list containing only one string returns the same thing
	  as a single string.

	In MFL, sieve "strings" can be any expression.

	"multiF" is passed as TRUE if the string list format is
	  allowed.

	"treePP" provides a pointer to a pointer variable: where we
	  place the pointer to the resulting expression or expression
	  list.  The parse tree returned is the same as returned
	  by parse_expr_list() .. see that.

	The syntax and parsing might be rather twisted because of sieve's
	  insistence of using a space as a separator in some cases.
	  In MFL we want to allow arbitrary expressions-- but
	  in a sequence like this:

	      "foo" [ 3 ]

	  is this two terms (string "foo" and string "3") or an array
	  reference?  So we say that any string outside of a
	  square-bracketed list must be a simple string reference; any
	  single string expression must be given in string list format
	  (enclosed in square brackets).  This makes the example above
	  unambiguous.
	  
*/

STSCODE
parse_sieve_string ARGLIST( (inP, treePP, multiF) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	**treePP )	/* Where to hang the result */
    FARG( BOOL		multiF )	/* If multiple strings are allowed. */
{
	STSCODE		status;
	PTNODE		*listP;		/* Ptr to list or term (we'll
					   return this) */
	TOKEN		*tokP;

    /* How we proceed depends on the first token */
    tokP = mfl_sigtoken( inP );
    if ( tokP == NULL )
	return ( SCNOTDONE );

    listP = NULL;			/* No result yet */

    /* A string list is signalled by an opening square bracket. */
    if ( ( tokP->tk_type == TK_OPR ) &&
	 ( tokP->tk_bstrP->bs_bP[0] == '[' ) ) {
	/* Get the expression list */
	status = parse_expr_list( inP, &listP );
	if ( status == SCOK ) {
	    /* Get trailing square bracket */
	    tokP = mfl_sigtoken( inP );
	    if ( ( tokP == NULL ) ||
	         ( tokP->tk_type != TK_CHAR ) ||
		 ( tokP->tk_bstrP->bs_bP[0] != ']' ) ) {
		mfl_warning( inP, "Expected ']' ." );
		status = SCNOTDONE;
	    }
	}
    }
    else {
	/* No opening square bracket; must be a simple string. */
	if ( tokP->tk_type == TK_LITSTR ) {
	    /* Turn it into a node. */
	    status = parsetree_mktknode( &listP, tokP );
	}
	else {
	    mfl_warning( inP, "string expected." );
	    status = SCNOTDONE;
	}
    }

    /* If we got a list of strings, make sure that's allowed. */
    if ( ( status == SCOK ) &&
         ( listP != NULL ) &&
	 ( listP->pt_type == PN_LSEQ ) &&
	 ( !multiF ) ) {
	mfl_warning( inP, "string list not allowed here." );
	status = SCNOTDONE;
    }

    /* If there was an error, free any tree we created. */
    if ( ( status != SCOK ) && ( listP != NULL ) ) {
    	parsetree_free( listP );
	listP = NULL;
    }

    /* Return the list if any */
    *treePP = listP;

    return( status );
}
/*

*//* parse_sieve_tagged_args( inP, nodeP, optsPP )

	Parse tagged arguments (aka options) for a statement

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS node
	optsPP		Ptr to a list of SIEVE_OPTION ptrs

Returns :

	<value>		SCXX status code


Notes :

	optsPP is an array of pointers to the SIEVE_OPTION definitions
	that are valid for this statement.  Table ends with a NULL.

	Options are added to an extended argument block attached
	to the pt_info.od_P element of the PN_SVBIS node.

*/

STSCODE
parse_sieve_tagged_args ARGLIST( (inP, nodeP, okoptsP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
    FARG( SIEVE_OPTION	**optsPP )	/* List of valid options */
{
	int		i;
	STSCODE		status;
	TOKEN		*tokP;
	SIEVE_OPTION	*optP;		/* Option definition */
	SIEVE_OPTREF	*optrefP;	/* Option reference */
	SIEVE_BISARG	*argsP;		/* Args ptr */

    /* Loop parsing options */
    for ( status = SCOK; status == SCOK; ) {
	/* Tagged options are introduced by a colon */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( tokP->tk_type != TK_CHAR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != ':' ) ) {
	    /* Take back the token and stop here. */
	    mfl_tokback( inP, tokP );
	    break;
	}

	/* Next thing must be a symbol of some sort */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( ( tokP->tk_type != TK_SYM ) &&
	       ( tokP->tk_type != TK_BIS ) ) ) {
	    mfl_warning( inP, "Keyword expected after ':'." );
	    status = SCNOTDONE;
	    break;
	}

	/* Find this option in the options table */
	for ( i = 0; ( optP = optsPP[i] ) != NULL; ++i )
	    if ( stricmp( (char *)tokP->tk_bstrP->bs_bP,
			       optP->od_nameP ) == 0 )
		break;
	if ( optP == NULL ) {
	    mfl_warning( inP, "option \"%s\" not allowed here.",
		    tokP->tk_bstrP->bs_bP );
	    status = SCNOTDONE;
	    break;
	}

	/* Attach the option to the extended info block */
	if ( ( argsP = sieve_bis_args_get( nodeP ) ) == NULL ) {
	    status = SCERR;
	    break;
	}

	/* See if we have this option class registered already */
	for ( i = 0; i < argsP->ba_optC; ++i )
	    if ( argsP->ba_optP[i].or_optP->od_class == optP->od_class ) {
		mfl_warning( inP, "Conflicting option: \"%s\".",
			   tokP->tk_bstrP->bs_bP );
		status = SCNOTDONE;
		break;
	    }

	if ( status != SCOK )
	    break;

	/* Make room for this option in the options references */
	if ( argsP->ba_optC == argsP->ba_optL ) {
	    i = argsP->ba_optL + 3;	/* Make room for 3 more */
	    if ( argsP->ba_optL == 0 )
		argsP->ba_optP = (SIEVE_OPTREF *)emalloc(
				     "parse_sieve_tagged_args",
				     "options reference block",
				     i * sizeof(SIEVE_OPTREF) );
	    else
		argsP->ba_optP = (SIEVE_OPTREF *)erealloc(
				     "parse_sieve_tagged_args",
				     "options reference block",
				     argsP->ba_optP,
				     i * sizeof(SIEVE_OPTREF) );
	    if ( argsP->ba_optP == NULL ) {
		argsP->ba_optC = argsP->ba_optL = 0;
		status = SCERR;
		break;
	    }

	    argsP->ba_optL = i;
	}

	/* Add this option reference */
	optrefP = &argsP->ba_optP[ argsP->ba_optC ];
	optrefP->or_optP = optP;
        optrefP->or_int = 0;
	optrefP->or_valP = NULL;

	/* If any extended parsing required, do it now */
	if ( optP->od_parse != NULL )
	    status = (*optP->od_parse)( inP, optrefP );

	if ( status == SCOK )
	    ++argsP->ba_optC;
    }

    return ( status );
}
/*

*//* parse_sieve_targ_comparator( inP, optrefP )

	Extended tagged option parsing: comparator value

Accepts :

	inP		Ptr to input stack
	optrefP		Ptr to option reference block

Returns :

	<value>		SCXX status code

Notes :

	Stores the comparator type as a simple value in the
	options reference.

*/

STSCODE
parse_sieve_targ_comparator ARGLIST( (inP, optrefP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( SIEVE_OPTREF	*optrefP )	/* Ptr to option reference */
{
	TOKEN		*tokP;

    /* Get string token */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
	 ( tokP->tk_type != TK_LITSTR ) ) {
	mfl_warning( inP, "string (comparator name) expected." );
	return ( SCNOTDONE );
    }

    if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "i;ascii-casemap" ) == 0 )
	optrefP->or_int = SOV_COMP_ASC_CASEMAP;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "i;ascii-numeric" ) == 0 )
	optrefP->or_int = SOV_COMP_ASC_NUMERIC;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "i;octet" ) == 0 )
	optrefP->or_int = SOV_COMP_OCTET;
    else {
	mfl_warning( inP, "Unknown comparator \"%s\"", 
		      tokP->tk_bstrP->bs_bP );
	return ( SCNOTDONE );
    }

    return ( SCOK );
}
/*

*//* parse_sieve_targ_int( inP, optrefP )

	Extended tagged option parsing: integer value

Accepts :

	inP		Ptr to input stack
	optrefP		Ptr to option reference block

Returns :

	<value>		SCXX status code

Notes :


*/

STSCODE
parse_sieve_targ_int ARGLIST( (inP, optrefP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( SIEVE_OPTREF	*optrefP )	/* Ptr to option reference */
{
	TOKEN		*tokP;

    /* Get integer token */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
	 ( tokP->tk_type != TK_INTEGER ) ) {
	mfl_warning( inP, "integer expected." );
	return ( SCNOTDONE );
    }

    /* Grab the integer from the token */
    optrefP->or_int = tokP->tk_val.tv_long;

    return ( SCOK );
}
/*

*//* parse_sieve_targ_rel( inP, optrefP )

	Extended tagged option parsing: relational value

Accepts :

	inP		Ptr to input stack
	optrefP		Ptr to option reference block

Returns :

	<value>		SCXX status code

Notes :

	Relational value as for :count and :value.  The SOV_REL_xxx
	code is stored in the optref integer value.

*/

STSCODE
parse_sieve_targ_rel ARGLIST( (inP, optrefP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( SIEVE_OPTREF	*optrefP )	/* Ptr to option reference */
{
	TOKEN		*tokP;

    /* Get string token */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
	 ( tokP->tk_type != TK_LITSTR ) ) {
	mfl_warning( inP, "string expected." );
	return ( SCNOTDONE );
    }

    /* Brute force the comparison */
    if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "lt" ) == 0 )
	optrefP->or_int = SOV_REL_LT;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "le" ) == 0 )
	optrefP->or_int = SOV_REL_LE;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "eq" ) == 0 )
	optrefP->or_int = SOV_REL_EQ;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "ne" ) == 0 )
	optrefP->or_int = SOV_REL_NE;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "ge" ) == 0 )
	optrefP->or_int = SOV_REL_GE;
    else if ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "gt" ) == 0 )
	optrefP->or_int = SOV_REL_GT;
    else {
	mfl_warning( inP, "Unknown relational code \"%s\"", 
			tokP->tk_bstrP->bs_bP );
	return ( SCNOTDONE );
    }

    return ( SCOK );
}
/*

*//* parse_sieve_targ_string( inP, optrefP )

	Extended tagged option parsing: string value

Accepts :

	inP		Ptr to input stack
	optrefP		Ptr to option reference block

Returns :

	<value>		SCXX status code


Notes :


*/

STSCODE
parse_sieve_targ_string ARGLIST( (inP, optrefP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( SIEVE_OPTREF	*optrefP )	/* Ptr to option reference */
{
    /* We're just going to use parse_sieve_string to get a single
       string.
    */
    return ( parse_sieve_string( inP, &optrefP->or_valP, FALSE ) );
}
/*

*//* parse_sieve_targ_string_list( inP, optrefP )

	Extended tagged option parsing: string list value

Accepts :

	inP		Ptr to input stack
	optrefP		Ptr to option reference block

Returns :

	<value>		SCXX status code


Notes :


*/

STSCODE
parse_sieve_targ_string_list ARGLIST( (inP, optrefP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( SIEVE_OPTREF	*optrefP )	/* Ptr to option reference */
{
    /* Just hand off to parse_sieve_string */
    return ( parse_sieve_string( inP, &optrefP->or_valP, TRUE ) );
}
/*

*//* parse_svbis_addheader( inP, nodeP )

	Parse sieve builtin statement: addheader

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the newheader name to the left side,
	and the new value name to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	     (header name)     (value)

	Each branch is a string parsed by
	parse_sieve_string, so see that.

*/

STSCODE
parse_svbis_addheader ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse header name and attach it to the left. */
    status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    /* Parse value and attach it to the right */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_rightP, FALSE );

    /* End with semi */
    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_address( inP, nodeP )

	Parse sieve builtin statement: address

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the header list to the left side,
	and the test list to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	         (headers)     (test list)

	Each branch is a string or string list parsed by
	parse_sieve_string, so see that.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_address ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_address[0] );

    /* Parse header list and attach it to the left. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    /* Parse test list and attach it to the right */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

    return ( status );
}
/*

*//* parse_svbis_allof( inP, nodeP )

	Parse sieve builtin statement: allof

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Calls parse_sieve_compound_test to parse a group of
	test statements, and attaches them to the left side
	of the PN_SVBIS node like this:

			PN_SVBIS
			   /
			  /
			 /
		test statement(s)

	See parse_sieve_compound_test for more details.

*/

STSCODE
parse_svbis_allof ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse the compound test */
    status = parse_sieve_compound_test( inP, &nodeP->pt_leftP );

    return ( status );
}
/*

*//* parse_svbis_anyof( inP, nodeP )

	Parse sieve builtin statement: anyof

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Calls parse_sieve_compound_test to parse a group of
	test statements, and attaches them to the left side
	of the PN_SVBIS node like this:

			PN_SVBIS
			   /
			  /
			 /
		test statement(s)

	See parse_sieve_compound_test for more details.

*/

STSCODE
parse_svbis_anyof ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse the compound test */
    status = parse_sieve_compound_test( inP, &nodeP->pt_leftP );

    return ( status );
}
/*

*//* parse_svbis_c( inP, nodeP )

	Parse sieve builtin statement: C

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Parses a C-like statement group that must be enclosed
	in curly braces.    Attaches the parse tree to the
	left, like this:

			PN_SVBIS
			   /
			  /
			 /
		    parse tree

*/

STSCODE
parse_svbis_c ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Next token must be opening brace */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	mfl_warning( inP, "'{' required for C block, got \"%s\".",
		     tokP == NULL ? "<NUL>" :
			(char *)&tokP->tk_bstrP->bs_bP[0] );
	status = SCNOTDONE;
    }
    else {
	/* Parse the compound statement and attach it to the left. */
	status = parse_compound( inP, &nodeP->pt_leftP );
    }

    return ( status );
}
/*

*//* parse_svbis_deleteheader( inP, nodeP )

	Parse sieve builtin statement: deleteheader

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the header name to the left side,
	and any pattern value to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	     (header name)     (header value)

	Or if no values are specified:

			PN_SVBIS
			   /
			  /
			 /
	     (header name)

	The header name is a single string value.  The value (if
	supplied) is a string or string list.  Both are parsed by
	parse_sieve_string.  See comments there for more.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_deleteheader ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_deleteheader[0] );

    /* Parse header name and attach it to the left */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    if ( status == SCOK ) {
	/* The next thing is either a string or string list, or
	   the semicolon that ends the statement.
	*/

	tokP = mfl_sigtoken( inP );
	if ( tokP == NULL ) {
	    mfl_warning( inP, "semicolon or string list expected." );
	    status = SCNOTDONE;
	}
	else if ( ( tokP->tk_type == TK_LITSTR ) ||
	          ( tokP->tk_bstrP->bs_bP[0] == '[' ) ) {

	    /* This is the string value- parse it. */
	    mfl_tokback( inP, tokP );
	    status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

	    if ( status == SCOK )
		status = parse_semi( inP );
	}
	else if ( ( tokP->tk_type != TK_CHAR ) ||
	          ( tokP->tk_bstrP->bs_bP[0] != ';' ) ) {
	    mfl_warning( inP, "semicolon expected." );
	    status = SCNOTDONE;
	}
    }

    return ( status );
}
/*

*//* parse_svbis_discard( inP, nodeP )

	Parse sieve builtin statement: discard

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_svbis_discard ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Need semi */
    status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_dnsbl( inP, nodeP )

	Parse sieve builtin statement: dnsbl

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the dnsbl name or name list to the left side, and
	the result name or name list to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	      (dnsbl name)     (result name list)

	Each branch is a string or string list parsed by
	parse_sieve_string, so see that.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_dnsbl ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_dnsbl[0] );

    /* Parse dnsbl name list and attach it to the left. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    /* Parse result list and attach it to the right */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

    return ( status );
}
/*

*//* parse_svbis_else( inP, nodeP )

	Parse sieve builtin statement: else

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	The code block for "else" is attached to the right side
	of the PN_SVBIS node (just like "if" or "elsif" only
	without the condition).  Thus:

			PN_SVBIS
			    \
			     \
			      \
			      (statement or block)

	The code block is parsed using parse_sieve_compound, so
	see that routine for more.

*/

STSCODE
parse_svbis_else ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Else is only allowed immediately following "if" or "elsif" .
       This is a kludge.
    */
    if ( ( P_lastctlX == -1 ) ||
         ( ( stricmp( Sieve_bistbl[ P_lastctlX ].sb_kwdP, "if" ) != 0 ) &&
	   ( stricmp( Sieve_bistbl[ P_lastctlX ].sb_kwdP, "elsif" ) != 0 ) ) ) {
	mfl_warning( inP, "\"else\" must follow only \"if\" or \"elsif\" ." );
	status = SCNOTDONE;
    }
    else {
	/* RFC3028 wants the code block to be enclosed in curly braces,
           so we'll enforce that.
	*/
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
             ( tokP->tk_type != TK_CHAR ) ||
	     ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	    mfl_warning( inP, "'{' required for conditional block." );
	    status = SCNOTDONE;
	}
	else {
	    /* Use parse_sieve_compound to collect the body. */
	    status = parse_sieve_compound( inP, &nodeP->pt_rightP );
	}
    }

    return ( status );
}
/*

*//* parse_svbis_elsif( inP, nodeP )

	Parse sieve builtin statement: elsif

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	This uses the "if" parser and so constructs a node
	exactly like an "if" node.  See that.

*/

STSCODE
parse_svbis_elsif ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Elsif is only allowed immediately following "if" or "elsif" .
       This is a kludge.
    */
    if ( ( P_lastctlX == -1 ) ||
         ( ( stricmp( Sieve_bistbl[ P_lastctlX ].sb_kwdP, "if" ) != 0 ) &&
	   ( stricmp( Sieve_bistbl[ P_lastctlX ].sb_kwdP, "elsif" ) != 0 ) ) ) {
	mfl_warning( inP, "\"elsif\" must follow only \"if\" or \"elsif\" ." );
	status = SCNOTDONE;
    }
    else 
	status = parse_svbis_if( inP, nodeP );

    return ( status );
}
/*

*//* parse_svbis_envelope( inP, nodeP )

	Parse sieve builtin statement: envelope

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the header list to the left side,
	and the test list to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	         (headers)     (test list)

	Each branch is a string or string list parsed by
	parse_sieve_string, so see that.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_envelope ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_envelope[0] );

    /* Parse header list and attach it to the left. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    /* Parse test list and attach it to the right */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

    return ( status );
}
/*

*//* parse_svbis_exists( inP, nodeP )

	Parse sieve builtin statement: exists

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Results in a PN_SVBIS parse node with string value
	attached on the left as follows:

			PN_SVBIS
			   /
			  /
			 /
		    string argument or list, as per
		      parse_sieve_string (q.v.)

*/

STSCODE
parse_svbis_exists ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse string list and attach to left side */
    status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    return ( status );
}
/*

*//* parse_svbis_false( inP, nodeP )

	Parse sieve builtin statement: false

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_svbis_false ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
    /* There is nothing else here.. just return OK */
    return ( SCOK );
}
/*

*//* parse_svbis_fileinto( inP, nodeP )

	Parse sieve builtin statement: fileinto

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the folder name to the SVBIS node as follows:

			PN_SVBIS
			   /
			  /
			 /
		folder name (string)

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_fileinto ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_fileinto[0] );

    /* Parse a string argument.  Only one string allowed. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    /* End with semi */
    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_header( inP, nodeP )

	Parse sieve builtin statement: header

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the header list to the left side,
	and the test list to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	         (headers)     (test list)

	Each branch is a string or string list parsed by
	parse_sieve_string, so see that.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_header ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_header[0] );

    /* Parse header list and attach it to the left. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    /* Parse test list and attach it to the right */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

    return ( status );
}
/*

*//* parse_svbis_if( inP, nodeP )

	Parse sieve builtin statement: if

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	The test condition is attached to the left side of
	the PN_SVBIS node, and the block to conditionally
	execute is attached to the right.  Any "else" or
	"elsif" statements are considered separate, and are
	not attached here.

	Thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	      (condition)     (statement or block)

	Note that parse_sieve_compound is used to attach the block
	to the right side, so see that routine for more.

*/

STSCODE
parse_svbis_if ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Get the condition and attach it to the left branch */
    status = parse_sieve( inP, &nodeP->pt_leftP,
					SBIS_TEST );
    if ( status != SCOK )
	return ( status );

    /* RFC3028 wants the code block to be enclosed in curly braces,
       so we'll enforce that.
    */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	mfl_warning( inP, "'{' required for conditional block." );
	status = SCNOTDONE;
    }
    else {
	/* Use parse_sieve_compound to collect the body. */
	status = parse_sieve_compound( inP, &nodeP->pt_rightP );
    }

    return ( status );
}
/*

*//* parse_svbis_keep( inP, nodeP )

	Parse sieve builtin statement: keep

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_svbis_keep ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Just need statement end */
    status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_not( inP, nodeP )

	Parse sieve builtin statement: not

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the test to "not" on the left side of the PN_SVBIS
	node, like this:

			PN_SVBIS
			   /
			  /
			 /
		test statement

	Uses parse_sieve to parse the test.

*/

STSCODE
parse_svbis_not ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
    return ( parse_sieve( inP, &nodeP->pt_leftP,
				SBIS_TEST ) );
}
/*

*//* parse_svbis_redirect( inP, nodeP )

	Parse sieve builtin statement: redirect

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the redirect address to the SVBIS node as follows:

			PN_SVBIS
			   /
			  /
			 /
		redirection address (string)

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_redirect ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_redirect[0] );

    /* Parse a string argument.  Only one string allowed. */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    /* End with semi */
    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_reject( inP, nodeP )

	Parse sieve builtin statement: reject

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the reject reason string to the SVBIS node:

			PN_SVBIS
			   /
			  /
			 /
		   reason (string)

*/

STSCODE
parse_svbis_reject ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse a string argument (the reason).  Only one string allowed. */
    status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    /* End with semi */
    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_replaceheader( inP, nodeP )

	Parse sieve builtin statement: replaceheader

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Attaches the header name to the left side,
	and any pattern value to the right side, thus:

			PN_SVBIS
			   /\
			  /  \
			 /    \
	     (header name)     (header value)

	Or if no values are specified:

			PN_SVBIS
			   /
			  /
			 /
	     (header name)

	The header name is a single string value.  The value (if
	supplied) is a string or string list.  Both are parsed by
	parse_sieve_string.  See comments there for more.

	Also, any extended statement information may be attached
	to pt_info.od_P (via parse_sieve_tagged_args).

*/

STSCODE
parse_svbis_replaceheader ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse any tagged options */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_replaceheader[0] );

    /* Parse header name and attach it to the left */
    if ( status == SCOK )
	status = parse_sieve_string( inP, &nodeP->pt_leftP, FALSE );

    if ( status == SCOK ) {
	/* The next thing is either a string or string list, or
	   the semicolon that ends the statement.
	*/

	tokP = mfl_sigtoken( inP );
	if ( tokP == NULL ) {
	    mfl_warning( inP, "semicolon or string list expected." );
	    status = SCNOTDONE;
	}
	else if ( ( tokP->tk_type == TK_LITSTR ) ||
	          ( tokP->tk_bstrP->bs_bP[0] == '[' ) ) {

	    /* This is the string value- parse it. */
	    mfl_tokback( inP, tokP );
	    status = parse_sieve_string( inP, &nodeP->pt_rightP, TRUE );

	    if ( status == SCOK )
		status = parse_semi( inP );
	}
	else if ( ( tokP->tk_type != TK_CHAR ) ||
	          ( tokP->tk_bstrP->bs_bP[0] != ';' ) ) {
	    mfl_warning( inP, "semicolon expected." );
	    status = SCNOTDONE;
	}
    }

    return ( status );
}
/*

*//* parse_svbis_require( inP, nodeP )

	Parse sieve builtin statement: require

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Results in a PN_SVBIS parse node with string value
	attached on the left as follows:

			PN_SVBIS
			   /
			  /
			 /
		    string argument or list, as per
		      parse_sieve_string (q.v.)


*/

STSCODE
parse_svbis_require ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Parse string list and attach to left side */
    status = parse_sieve_string( inP, &nodeP->pt_leftP, TRUE );

    if ( status == SCOK )
	status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_sieve( inP, nodeP )

	Parse sieve builtin statement: sieve

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	Parses a sieve statement group that must be enclosed
	in curly braces.    Attaches the parse tree to the
	left, like this:

			PN_SVBIS
			   /
			  /
			 /
		    parse tree

*/

STSCODE
parse_svbis_sieve ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Next token must be opening brace */
    tokP = mfl_sigtoken( inP );
    if ( ( tokP == NULL ) ||
         ( tokP->tk_type != TK_CHAR ) ||
	 ( tokP->tk_bstrP->bs_bP[0] != '{' ) ) {
	mfl_warning( inP, "'{' required for sieve block, got \"%s\".",
		     tokP == NULL ? "<NUL>" :
			(char *)&tokP->tk_bstrP->bs_bP[0] );
	status = SCNOTDONE;
    }
    else {
	/* Parse the compound statement and attach it to the left. */
	status = parse_sieve_compound( inP, &nodeP->pt_leftP );
    }

    return ( status );
}
/*

*//* parse_svbis_size( inP, nodeP )

	Parse sieve builtin statement: size

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :

	The test value is stored as an integer type parse tree
	node on the right side of the PN_SVBIS node.  Thus:

			PN_SVBIS
			    \
			     \
			      \
			      (PN_INTEGER node)

*/

STSCODE
parse_svbis_size ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;
	TOKEN		*tokP;

    /* Parse any tagged options. */
    status = parse_sieve_tagged_args( inP, nodeP, &Svstmt_opt_size[0] );

    if ( status == SCOK ) {
	/* One of the over/under values is required.  Since we only
	   accept over or under, we can simply test the presence
	   of an argument pointer.  This test will not work if
	   any options are added due to extensions etc.
	*/
	if ( nodeP->pt_info.od_P == NULL ) {
	    mfl_warning( inP, "Must specify :over or :under ." );
	    status = SCNOTDONE;
	}
    }

    if ( status == SCOK ) {
	/* Test value is required, it must be a number. */
	tokP = mfl_sigtoken( inP );
	if ( ( tokP == NULL ) ||
	     ( tokP->tk_type != TK_INTEGER ) ) {
	    mfl_warning( inP, "\"size\" requires an integer test value." );
	    status = SCNOTDONE;
	}
	else {
	    /* Attach value node to the right */
	    status = parsetree_mktknode( &nodeP->pt_rightP, tokP );
	}
    }

    return ( status );
}
/*

*//* parse_svbis_stop( inP, nodeP )

	Parse sieve builtin statement: stop

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_svbis_stop ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	STSCODE		status;

    /* Get the trailing semi */
    status = parse_semi( inP );

    return ( status );
}
/*

*//* parse_svbis_true( inP, nodeP )

	Parse sieve builtin statement: true

Accepts :

	inP		Ptr to input stack
	nodeP		Ptr to PN_SVBIS for this statement

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
parse_svbis_true ARGLIST( (inP, nodeP) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
    /* Nothing else required */
    return ( SCOK );
}
/*

*//* sieve_bis_args_free(argsP)

	Free any info attached to a PN_SVBIS node

Accepts :

	argsP		Ptr to extended args block

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	This is specified as the pt_info.od_del "routine to call"
	in the PTNODE structure, as the function that is called
	to clean up the info pointer (pt_info.od_P) when a node is freed.

*/

BOOL
sieve_bis_args_free ARGLIST( (argsP) )
    FARG( SIEVE_BISARG	*argsP )	/* Extended arguments */
{
	int		i;
	SIEVE_OPTREF	*optrefP;	/* Option reference */

    if ( argsP != NULL ) {
	/* Handle any options list */
	if ( argsP->ba_optP != NULL ) {
	    for ( i = 0; i < argsP->ba_optC; ++i ) {
		optrefP = &argsP->ba_optP[i];
		if ( optrefP->or_valP != NULL ) {
		    parsetree_free( optrefP->or_valP );
		    optrefP->or_valP = NULL;
		}
	    }
	    dealloc( argsP->ba_optP );
	}

	/* Free the block */
	dealloc( argsP );
    }

    return ( TRUE );
}
/*

*//* sieve_bis_args_get(nodeP)

	Get (optionally make) the extended args block for a statement

Accepts :

	nodeP		Ptr to PN_SVBIS node for statement

Returns :

	<value>		Ptr to SIEVE_BISARG block
			NULL if error

Notes :

*/

SIEVE_BISARG *
sieve_bis_args_get ARGLIST( (nodeP) )
    FARG( PTNODE	*nodeP )	/* Ptr to PN_SVBIS node */
{
	SIEVE_BISARG	*argsP;		/* Args ptr */

    if ( ( argsP = (SIEVE_BISARG *)nodeP->pt_info.od_P ) == NULL ) {
	/* Args block needs creating */
	argsP = (SIEVE_BISARG *)emalloc( "sieve_bis_args_get", "args block",
			sizeof( SIEVE_BISARG ) );
	if ( argsP != NULL ) {
	    /* Initialize the block */
	    argsP->ba_optC = 0;
	    argsP->ba_optL = 0;
	    argsP->ba_optP = NULL;

	    /* Attach it to the parse node */
	    nodeP->pt_info.od_P = argsP;

	    /* Set the deletion routine for the parsetree freeing */
	    nodeP->pt_info.od_del = sieve_bis_args_free;
	}
    }

    return ( argsP );
}
/*

*//* sieve_cap_enabled( capP )

	Check for capability enabled

Accepts :

	capP		Ptr to capability block

Returns :

	<value>		TRUE if enabled
			FALSE if not.

Notes :

	Issues a warning message if not enabled

*/

BOOL
sieve_cap_enabled ARGLIST( (capP) )
    FARG( SIEVE_CAP	*capP )		/* Ptr to capability */
{
    if ( capP->cap_enabledF )
	return ( TRUE );

    warning ( "Capability \"%s\" required but not enabled.",
		capP->cap_nameP );
    return ( FALSE );
}
/*

*//* sieve_node_print_svbis( nodeP )

	Print a description of a sieve PN_SVBIS node (a la parsetree_print)

Accepts :

	nodeP		Address of PN_SVBIS node to print

Returns :

	<nothing>

Notes :

	Called e.g. by parsetree_print to print a node that's a
	type PN_SVBIS.  Only prints a single node, does not
	descend into the tree.

*/

void
sieve_node_print_svbis ARGLIST( (fP, nodeP) )
   NFARG( FILE		*fP )		/* Output file */
    FARG( PTNODE	*nodeP )	/* Node to print */
{
	int		i;
	SIEVE_BISARG	*argsP;		/* Args ptr */
	SIEVE_OPTREF	*optrefP;	/* Option reference */
	SIEVE_OPTION	*optP;		/* Option definition */

    fprintf( fP, "PN_SVBIS: \"%s\"",
		    Sieve_bistbl[ nodeP->pt_data.pv_int ].sb_kwdP );

    /* Print any args */
    if ( ( argsP = (SIEVE_BISARG *)nodeP->pt_info.od_P ) != NULL ) {
	fprintf( fP, "(");
	for ( i = 0; i < argsP->ba_optC; ++i ) {
	    if ( i != 0 )
		fprintf( fP, " " );
	    optrefP = &argsP->ba_optP[i];
	    optP = optrefP->or_optP;

	    fprintf( fP, ":%s", optP->od_nameP );

	    /* Do any special printing for options with arguments */
	    switch ( optrefP->or_optP->od_class ) {
		/* Integer values */
		case	SOC_INDEX:
		    fprintf( fP, " %d", optrefP->or_int );
		    break;

		/* String values */
		case	SOC_IP:
		case	SOC_NEWNAME:
		case	SOC_NEWVALUE:
		    fprintf( fP, " [expr]" );
		    break;

		default:
		    break;
	    }
	}
	fprintf( fP, ")" );
    }

    fprintf( fP, "\n" );
}
/*

*//* sieve_optval_finit( optvalP )

	Finalize (i.e. clear, deinitialize) a sieve statement option value

Accepts :

	optvalP		Ptr to an evaluated option argument value

Returns :

	<nothing>

Notes :

	Clears up whatever needs cleaning up from a prior evaluation
	of an option value.

	Note: the option value itself is not re-initialized; it will
	likely still contain the same bits that it came in with, even
	those any related data is finalized.  This is because most callers
	are not expected to reuse the data, and will pre-initialize any
	new option values before using them.
*/

void
sieve_optval_finit ARGLIST( (optvalP) )
    FARG( SIEVE_OPTVAL	*optvalP )	/* Option value */
{
	int		i;

    if ( optvalP->ov_oov == SOV_NONE )
	return;				/* Not initialized. */

    /* Process depending on the type of value */
    switch ( optvalP->ov_class ) {
		/* Classes that don't need any data removal */
	case	SOC_ADDR_PART:
	case	SOC_COMPARATOR:
	case	SOC_COPY:
	case	SOC_INDEX:
	case	SOC_LAST:
	case	SOC_MATCH_TYPE:
	case	SOC_OVER_UNDER:

#ifdef	VACATION
	case	SOC_DAYS:
#endif	/* VACATION */
	    break;


		/* Classes that have a single string */
	case	SOC_IP:
	case	SOC_NEWNAME:
	case	SOC_NEWVALUE:

#ifdef	VACATION
	case	SOC_FROM:
	case	SOC_HANDLE:
	case	SOC_MIME:
	case	SOC_SUBJECT:
#endif	/* VACATION */
	    dealloc( optvalP->ov_val.ovv_bP );
	    break;



	    /* Values that have a string list */
	    /* NOTE: only vacation uses this at present, so it's all in
	       that conditional.  Adjust conditional if something else
	       needs it.
	    */

#ifdef	VACATION

	case	SOC_ADDRESSES:
	    /* Strings are in a NULL-terminated array of pointers
	       referenced by the ovv_bPP value element.
	    */
	    for ( i = 0; optvalP->ov_val.ovv_bPP[i] != NULL; ++i ) {
		dealloc( optvalP->ov_val.ovv_bPP[i] );
	    }
	    dealloc( optvalP->ov_val.ovv_bPP );
	    break;

#endif	/* VACATION */


	default:
	    mfl_warning( NULL,
			"sieve_optval_finit: bad class %d",
			optvalP->ov_class );
	    break;
    }
}
/*

*//* sieve_optvals_finit( optvalsP, optsPP )

	Finalize sieve statement option values

Accepts :

	optvalsP	Ptr to array of option argument values
	optsPP		Ptr to list of SIEVE_OPTION pointers

Returns :

	<nothing>

Notes :

	This is a complement to sieve_optvals_init; this routine
	clears/finalizes/removes any data created by the init
	routine.

	As with sieve_optvals_init, optvalsP is the array of option
	values to finalize, and it SHOULD be at least SOC_MAX members
	large, but it must be at least as large as the largest option
	class (index into the array) used.

	optsPP, if non-NULL, is a NULL-terminated array of pointers
	to options definitions (just like the ones used in options
	parsing).  If given, this acts as a control for which option
	values should be finalized.

*/

void
sieve_optvals_finit ARGLIST( (optvalsP, optsPP) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
    FARG( SIEVE_OPTION	**optsPP )	/* List of valid options */
{
	int		i;
	SIEVE_OPTION	*optP;

    if ( optsPP == NULL ) {
	/* No guide array; finalize them all */
	for ( i = 0; i < SOC_MAX; ++i ) {
	    sieve_optval_finit( &optvalsP[i] );
	}
    }
    else {
	/* There is a guide array; set only the associated values */
	for ( ; ( optP = *optsPP ) != NULL; ++optsPP ) {
	    sieve_optval_finit( &optvalsP[ optP->od_class ] );
	}
    }
}
/*

*//* sieve_optvals_init( optvalsP, optsPP )

	Initialize sieve statement option values

Accepts :

	optvalsP	Ptr to array of option argument values
	optsPP		Ptr to list of SIEVE_OPTION pointers

Returns :

	<nothing>

Notes :

	optvalsP is the array of option values to initialize.
	It SHOULD be at least SOC_MAX members large, but it
	must be at least as large as the largest option class
	(index into the array) used.  The largest possible value
	is SOC_MAX -1, and this is the actual value used if
	optsPP is passed as NULL.

	optsPP, if non-NULL, is a NULL-terminated array of pointers
	to options definitions (just like the ones used in options
	parsing).  If given, this acts as a control for which option
	values should be initialized.

	Values are initialized by setting their "one-of" data
	to SOV_NONE.  The class of the value is also initialized so
	that the individual value can be used without knowing
	its array index.

*/

void
sieve_optvals_init ARGLIST( (optvalsP, optsPP) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
    FARG( SIEVE_OPTION	**optsPP )	/* List of valid options */
{
	int		i;
	SIEVE_OPTION	*optP;

    if ( optsPP == NULL ) {
	/* No guide array; initialize them all. */
	for ( i = 0; i < SOC_MAX; ++i ) {
	    optvalsP[i].ov_oov = SOV_NONE;
	    optvalsP[i].ov_class = (SIEVE_OPT_CLASS)i;
	}
    }
    else {
	/* There is a guide array; set only the associated values */
	for ( ; ( optP = *optsPP ) != NULL; ++optsPP ) {
	    i = optP->od_class;
	    optvalsP[i].ov_oov = SOV_NONE;
	    optvalsP[i].ov_class = i;
	}
    }
}



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* do_exec_svbis_address( optvalsP, hnameP, patP )

	Does the work of exec_svbis_address

Accepts :

	optvalsP	Ptr to option values array
	hnameP		Name of header to check
	patP		Name of pattern to look for

Returns :

	<value>		SCxx status code

Notes :

	This is a handler routine for exec_svbis_address.

	Called by exec_sieve_stringx2 -- see that function
	for interface details.

*/

STSCODE
do_exec_svbis_address ARGLIST( (optvalsP, hnameP, patP ) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( ESTR		*hnameP )	/* Ptr to header name */
    FARG( ESTR		*patP )		/* Ptr to pattern to look for */
{
	BOOL		resultF;	/* Result of test */
	UWORD		addr_parts;	/* Which address parts wanted */
	UBYTE		*addrP;		/* Ptr to address */
	void		*hdrP;

    /* Pick up the desired address parts mask */
    if ( optvalsP[ SOC_ADDR_PART ].ov_oov == SOV_SET )
	addr_parts = optvalsP[ SOC_ADDR_PART ].ov_val.ovv_uint;
    else
	addr_parts = (AP_LOCAL|AP_DOMAIN);

    /* Test pattern against each address in each header */
    hdrP = NULL;
    for ( ; ; ) {
	hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP );
	if ( hdrP == NULL )
	    break;

	/* Check or count all the addresses contained in this header */
	while ( ( addrP = (UBYTE *)msgpart_hdr_addr( OmmP->om_curmpP, hdrP,
					addr_parts ) ) != NULL ) {
	    if ( patP == NULL )
		++Countval;		/* "count" mode */
	    else {
		resultF = sieve_string_compare( optvalsP,
					        patP, estr_tmp_str( addrP ) );

		/* If match, we can return now */
		if ( resultF )
		    return ( SCDONE );
	    }
	}
    }

    /* Nothing matched -- just return OK */
    return ( SCOK );
}
/*

*//* do_exec_svbis_dnsbl( optvalsP, blnameP, rnameP )

	Does the work of exec_svbis_dnsbl

Accepts :

	optvalsP	Ptr to option values array
	blnameP		Name of dnsbl blacklist to check
	rnameP		Name of result code to test

Returns :

	<value>		SCxx status code

Notes :

	This is a handler routine for exec_svbis_dnsbl.

	Called by exec_sieve_stringx2 -- see that function
	for interface details.

*/

STSCODE
do_exec_svbis_dnsbl ARGLIST( (optvalsP, blnameP, rnameP ) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( ESTR		*blnameP )	/* Ptr to dnsbl name */
    FARG( ESTR		*rnameP )	/* Ptr to result name */
{
	BOOL		result;

    /* Check the dnsbl */
    if ( optvalsP[ SOC_IP ].ov_oov == SOV_SET )
	result = dnsbl_check_ip( (char *)optvalsP[ SOC_IP ].ov_val.ovv_bP,
				 (char *)tmpstr_estr( blnameP ),
				 (char *)tmpstr_estr( rnameP ) );
    else
	result = dnsbl_check_omm( OmmP,
				  (char *)tmpstr_estr( blnameP ),
				  (char *)tmpstr_estr( rnameP ) );

    if ( result )
	return ( SCDONE );
    else
	return ( SCOK );
}
/*

*//* do_exec_svbis_envelope( optvalsP, hnameP, patP )

	Does the work of exec_svbis_envelope

Accepts :

	optvalsP	Ptr to option values array
	hnameP		Name of envelope thing to check
	patP		Name of pattern to look for

Returns :

	<value>		SCxx status code

Notes :

	This is a handler routine for exec_svbis_envelope.

	Called by exec_sieve_stringx2 -- see that function
	for interface details.

*/

STSCODE
do_exec_svbis_envelope ARGLIST( (optvalsP, hnameP, patP ) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( ESTR		*hnameP )	/* Ptr to header name */
    FARG( ESTR		*patP )		/* Ptr to pattern to look for */
{
	BOOL		resultF;	/* Result of test */
	UWORD		addr_parts;	/* Which address parts wanted */
	UBYTE		*addrP;		/* Ptr to address */
	REFSTR		*hbstrP;	/* Header string */

    /* Pick up the desired address parts mask */
    if ( optvalsP[ SOC_ADDR_PART ].ov_oov == SOV_SET )
	addr_parts = optvalsP[ SOC_ADDR_PART ].ov_val.ovv_uint;
    else
	addr_parts = (AP_LOCAL|AP_DOMAIN);


    /* The "header name" indicates the name of an envelope field
       that isn't really in the header.  Choose the string based on
       the name.
    */
    if ( estr_icmp_str( hnameP, (UBYTE *)"From" ) == 0 )
	hbstrP = E_SenderP;
    else if ( estr_icmp_str( hnameP, (UBYTE *)"To" ) == 0 )
	hbstrP = E_RecipP;
    else {
TODO	/* Not really clear if this is an error- but for now we'll
	   just treat it as a non-match or a zero count.
	*/
	return ( SCOK );
    }

    /* There should be only a single address in this "header."  Get it. */
    hbstrP->rs_bX = 0;
    addrP = (UBYTE *)msgpart_refstr_addr( OmmP->om_mmsgP->mm_topmpP,
					 hbstrP, addr_parts );
    if ( addrP == NULL ) {
TODO	/* This seems like a problem, but we'll just return nomatch */
	return ( SCOK );
    }

    /* If count mode, count it and return. */
    if ( patP == NULL ) {
	++Countval;
	return ( SCOK );
    }

    /* Attempt the match. */
    resultF = sieve_string_compare( optvalsP, patP, estr_tmp_str( addrP ) );

    /* If match, return it.. */
    if ( resultF )
	return ( SCDONE );

    /* Not matched -- just return OK */
    return ( SCOK );
}
/*

*//* do_exec_svbis_header( optvalsP, hnameP, patP )

	Does the work of exec_svbis_header

Accepts :

	optvalsP	Ptr to option values array
	hnameP		Name of header to check
	patP		Name of pattern to look for

Returns :

	<value>		SCxx status code

Notes :

	This is a handler routine for exec_svbis_header.

	Called by exec_sieve_stringx2 -- see that function
	for interface details.

*/

STSCODE
do_exec_svbis_header ARGLIST( (optvalsP, hnameP, patP ) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( ESTR		*hnameP )	/* Ptr to header name */
    FARG( ESTR		*patP )		/* Ptr to pattern to look for */
{
	BOOL		resultF;	/* Result of test */
	void		*hdrP;
	REFSTR		*hsP;		/* Header REFSTR */
	ESTR		*hesP;		/* Header ESTR */

    /* Quick hack for body test */
    if ( estr_icmp_str( hnameP, (UBYTE *)"-body-" ) == 0 ) {
	/* Use the cached body if any */
	hsP = OmmP->om_curmpP->mp_body.ps_partP;
	if ( hsP == NULL )
	    return ( SCOK );		/* No body-- no match. */

	if ( patP == NULL ) {
	    /* "count" mode-- should probably not happen, but who cares */
	    ++Countval;
	}
	else {
	    hesP = estr_refstr( hsP );
	    resultF = sieve_string_compare( optvalsP, patP, hesP );
	    estr_finish( hesP );

	    /* If match, we can return now */
	    if ( resultF )
		return ( SCDONE );
	}

	/* No match */
	return ( SCOK );
    }


    /* Find each matching header */
    hdrP = NULL;
    for ( ; ; ) {
	hdrP = msgpart_hdr_next( OmmP->om_curmpP, hdrP, hnameP );
	if ( hdrP == NULL )
	    break;

	if ( patP == NULL )
	    /* "count" mode, just count the header occurence. */
	    ++Countval;
	else {
	    /* Do the comparison */
	    hesP = msgpart_hdr_tail( OmmP->om_curmpP, hdrP );
	    resultF = sieve_string_compare( optvalsP, patP, hesP );

	    /* If match, we can return now */
	    if ( resultF )
		return ( SCDONE );
	}
    }

    /* Nothing matched -- just return OK */
    return ( SCOK );
}
/*

*//* exec_sieve_block( treeP, rvalP )

	Execute a "sieve" block

Accepts :

	treeP		Ptr to parse tree node for sieve statement
	rvalP		Where to put the result

Returns :

	<value>		SCxx status code

Notes :

	Called to exec a new sieve context (i.e. a code block
	as a body of an if).  

*/

static STSCODE
exec_sieve_block ARGLIST( (treeP, rvalP ) )
   NFARG( PTNODE	*treeP )	/* Address of statement tree node */
    FARG( VALUE		*rvalP )	/* Where to put the result */
{
	STSCODE		status;

    /* Execution context that needs saving around blocks */
	int		x_ifF;		/* Running flag for if/elsif/else */

    /* Push previous execution context */
    x_ifF = X_ifF;

    /* Init new execution context */
    X_ifF = 0;				/* This doesn't really need
					   setting but what the heck */

    /* Execute the block */
    status = exec_sieve( treeP, rvalP );

    /* Restore previous execution context */
    X_ifF = x_ifF;

    return ( status );
}
/*

*//* exec_sieve_xof( nodeP, rvalP, untilF )

	Common code for x-of (anyof, allof) execution

Accepts :

	nodeP		Ptr to "sieve" parse tree node
	rvalP		Where to put the result
	untilF		Sense of "continue testing until.."

Returns :

	<value>		SCxx status code

Notes :

	untilF lets us know when to exit the evaluation.  If untilF
	is true, we continue testing until we get a TRUE.  If it's
	false, we continue testing until we get a FALSE.

*/

static BOOL
exec_sieve_xof ARGLIST( (nodeP, rvalP, untilF ) )
   NFARG( PTNODE	*nodeP )	/* Address of group node */
   NFARG( VALUE		*rvalP )	/* Where to put the result */
    FARG( BOOL		untilF )	/* Continue until.. */
{
	STSCODE		status;
	BOOL		testF;		/* Result of test */
	PTNODE		*tnodeP;	/* Ptr to test statement */

    status = SCOK;

    /* Walk through the arguments attached to left side, executing
       the test statements until until we can finish.
    */
    for ( nodeP = nodeP->pt_leftP;
	 ( nodeP != NULL ) && ( status == SCOK ); ) {

	if ( nodeP->pt_type == PN_LSEQ ) {
	    tnodeP = nodeP->pt_rightP;
	    nodeP = nodeP->pt_leftP;	/* Setup next node */
	}
	else {
	    tnodeP = nodeP;
	    nodeP = NULL;		/* No next node */
	}

	status = exec_sieve( tnodeP, rvalP );
	if ( status != SCOK )
	    break;			/* Discontinue if some problem */

	/* Check the result and see if we can stop */
	testF = sieve_test_result( rvalP );
	if ( untilF ) {
	    if ( testF )
		break;
	}
	else if ( !testF )
	    break;
    }

    /* Last result is the appropriate thing to return. */
    return( status );
}
/*

*//* sieve_eval_rel( cmpval, reltype )

	Evaluates a relational result with regard to a test

Accepts :

	cmpval		A comparison value e.g. as returned by strcmp
	reltype		The relational match type to test against

Returns :

	<value>		TRUE if cmpval satisfies the reltype
			FALSE if not.

Notes :

	cmpval is the result of a comparison between two terms:
	  < 0 if term1 < term2
	    0 if term1 = term2
	  > 0 if term1 > term2

	reltype is a SOV_REL_xxx relational type, e.g. SOV_REL_LE
	  if we are testing the "less than or equals" condition.

*/

static BOOL
sieve_eval_rel ARGLIST( (cmpval, reltype) )
   NFARG( int		cmpval )	/* Comparison value */
    FARG( SIEVE_OPT_ONEOF reltype )	/* The relational test */
{
    switch ( reltype ) {
	case SOV_REL_LT:		/* relational "lt" */
	    return ( cmpval < 0 );

	case SOV_REL_LE:		/* relational "le" */
	    return ( cmpval <= 0 );

	case SOV_REL_EQ:		/* relational "eq" */
	    return ( cmpval == 0 );

	case SOV_REL_NE:		/* relational "ne" */
	    return ( cmpval != 0 );

	case SOV_REL_GE:		/* relational "ge" */
	    return ( cmpval >= 0 );

	case SOV_REL_GT:		/* relational "gt" */
	    return ( cmpval > 0 );

	default:
	    mfl_warning( NULL, "Invalid reltype %d in sieve_eval_rel.", reltype );
	    return ( FALSE );
    }
}
/*

*//* sieve_string_compare (optvalsP, es1P, es2P)

	Does a sieve string comparison

Accepts :

	optvalsP	Ptr to option values array
	es1P		Ptr to first string (ESTR)
	es2P		Ptr to second string (ESTR)

Returns :

	<value>		TRUE if match
			FALSE if no match

Notes :

	Useful only on behalf of sieve execution, this routine does
	a string comparison paying attention to various tagged option
	settings optionally passed in optvalsP.  Typically es1P is
	a pattern and es2P is the string to find the pattern in,
	but that's not always the case.

	optvalsP may be passed as NULL, in which case defaults apply
	for any options.

*/

static BOOL
sieve_string_compare ARGLIST( (optvalsP, es1P, es2P) )
   NFARG( SIEVE_OPTVAL	*optvalsP )	/* Array of option values */
   NFARG( ESTR		*es1P )		/* First string, or pattern */
    FARG( ESTR		*es2P )		/* Second string, or value */
{
	BOOL		resultF;
	BOOL		processed;
	SIEVE_OPT_ONEOF	matchtype;	/* The match type specified */
	SIEVE_OPT_ONEOF	comparator;	/* The comparator specified */
	int		cmpval;
	char		*str1P, *str2P;

    /* Get any match type and comparator specified */
    if ( optvalsP == NULL ) {
	matchtype = SOV_MATCH_IS;
	comparator = SOV_COMP_ASC_CASEMAP;
    }
    else {
	if ( ( matchtype = optvalsP[ SOC_MATCH_TYPE ].ov_oov ) == SOV_NONE )
	    matchtype = SOV_MATCH_IS;
	if ( ( comparator = optvalsP[ SOC_COMPARATOR ].ov_oov ) == SOV_NONE )
	    comparator = SOV_COMP_ASC_CASEMAP;
    }

    /* Process according to the match type and the comparator type */
    processed = FALSE;
    switch ( matchtype ) {
	case	SOV_MATCH_IS:
	    switch ( comparator ) {
		case	SOV_COMP_OCTET:
		    /* pattern, string, substring OK, ignore case */
		    resultF = estr_match_exact( NULL, es1P, es2P,
					        0 );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_CASEMAP:
		    /* pattern, string, substring OK, ignore case */
		    resultF = estr_match_exact( NULL, es1P, es2P,
					        MF_IGNCASE );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_NUMERIC:
		    /* Strip leading whitespace for numeric comparison */
		    str1P = solbat( (char *)tmpstr_estr( es1P ), TRUE );
		    str2P = solbat( (char *)tmpstr_estr( es2P ), TRUE );

TODO		    /* Might want to verify numeric strings here. */
		    resultF = ( atod( str1P ) == atod( str2P ) );
		    processed = TRUE;
		    break;

		default:		/* error is given later */
		    break;
	    }
	    break;

	case	SOV_MATCH_CONTAINS:
	    switch ( comparator ) {
		case	SOV_COMP_OCTET:
		    /* pattern, string, substring OK, don't ignore case */
		    resultF = estr_match_exact( NULL, es1P, es2P,
					        MF_SUBSTR );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_CASEMAP:
		    /* pattern, string, substring OK, ignore case */
		    resultF = estr_match_exact( NULL, es1P, es2P,
				        MF_SUBSTR|MF_IGNCASE );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_NUMERIC:
		    mfl_warning( NULL,
	":comparator \"i;ascii-numeric\" not applicable to \":contains\"" );
		    processed = TRUE;
		    resultF = FALSE;
		    break;

		default:		/* error is given later */
		    break;
	    }
	    break;


	case  SOV_MATCH_MATCHES:
	    switch ( comparator ) {
		case	SOV_COMP_OCTET:
		    /* pattern, string, don't ignore case */
		    resultF = estr_match_wild( NULL, es1P, es2P,
					        0 );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_CASEMAP:
		    /* pattern, string, ignore case */
		    resultF = estr_match_wild( NULL, es1P, es2P,
					        MF_IGNCASE );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_NUMERIC:
		    mfl_warning( NULL,
	":comparator \"i;ascii-numeric\" not applicable to \":matches\"" );
		    processed = TRUE;
		    resultF = FALSE;
		    break;

		default:		/* error is given later */
		    break;
	    }
	    break;


	case  SOV_MATCH_REGEX:
	    switch ( comparator ) {
		case	SOV_COMP_OCTET:
		    /* pattern, string, no submatch, don't ignore case */
		    resultF = estr_match_regex( NULL, es1P, es2P,
					         0 );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_CASEMAP:
		    /* pattern, string, no submatch, ignore case */
		    resultF = estr_match_regex( NULL, es1P, es2P,
					        MF_IGNCASE );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_NUMERIC:
		    mfl_warning( NULL,
	":comparator \"i;ascii-numeric\" not applicable to \":regex\"" );
		    processed = TRUE;
		    resultF = FALSE;
		    break;

		default:		/* error is given later */
		    break;
	    }
	    break;


	case	SOV_MATCH_VALUE:		/* :value relational */
	case	SOV_MATCH_COUNT:		/* :count handled the same */

	    /* es1P is the value we test against,
	       es2P is the value to be tested.  i.e.
	       we test es2P against es1P.
	    */

	    /* Get string pointers, skiping leading blanks/tabs */
	    str1P = solbat( (char *)tmpstr_estr( es1P ), TRUE );
	    str2P = solbat( (char *)tmpstr_estr( es2P ), TRUE );

	    switch ( comparator ) {
		case	SOV_COMP_OCTET:
		    cmpval = strcmp( str2P, str1P );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_CASEMAP:
		    cmpval = stricmp( str2P, str1P );
		    processed = TRUE;
		    break;

		case	SOV_COMP_ASC_NUMERIC:
		    cmpval = (atod(str2P) - atod(str1P));
		    processed = TRUE;
		    break;

		default:		/* error is given later */
		    break;
	    }

	    if ( processed ) {
		/* The relational type is stored in the int val */
		resultF = sieve_eval_rel( cmpval,
				  optvalsP[ SOC_MATCH_TYPE ].ov_val.ovv_int );
	    }
	    break;


	default:
	    break;
    }

    if ( !processed ) {
	mfl_warning( NULL,
 "unknown matchtype/comparator combination (%d/%d) in sieve_string_compare",
		matchtype, comparator );
	resultF = FALSE;
    }

    return ( resultF );
}
/*

*//* sieve_test_result( valP )

	Return boolean from the result of an evaluation.

Accepts :

	valP		A value produced by executing a sieve statement
	
Returns :

	<value>		TRUE if the value was non-zero
			FALSE if the value was zero

Notes :

	Also coerces the value at valP to be an int type,
	so that it can be properly returned from a test.

*/

BOOL
sieve_test_result ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* Value to interpret */
{
    /* Force the value to be word */
    (void)val_coerce( valP, Tnint4->tn_typeP, FALSE );

    /* Test non-zero */
    return ( valP->v_data.d_long != 0 );
}


/*

*//* end of sieve conditional */

#endif		/* SIEVE */
