/* mfl-token.c -- Tokenizer for this language

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Tokenizer routines for language input.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>
#include <mml/mml-fmt.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-token.h"
#include "mfl-pp.h"
#include "mfl-func.h"

#include "errwarn.h"

/* Local definitions */


/* External data referenced */

extern	BOOL	MexpandF;		/* Whether expanding macros */
extern	OPINFO	Optbl[];		/* Operators table */
extern	SYMTAB	*STbis;			/* Built-in symbol table */
extern	SYMTAB	*STpreproc;		/* Preprocessor symbol table */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	int	eol_comment PROTO( (MML_IN *inP) );
static	int	non_comment_char PROTO( (MML_IN *inP, int nextch) );
static	TOKEN	*tok_grab PROTO( (void) );
static	BOOL	tok_seed PROTO( (void) );
static	TOKEN	*tok_text PROTO( (MML_IN *inP, TOKEN *tokP) );


/* Private data */

static	int	ToklistC;	/* Number of tokens in ToklistP */
static	TOKEN	*ToklistP;	/* A list of tokens */
static	TOKEN	*ToknextP;	/* The next token available */
static	TOKEN	*TokpbP;	/* Pushed back token(s) */

/*

*//* mfl_tokback( inP, tokP )

	Push a token back onto the input stream

Accepts :

	inP		Ptr to input stack
	tokP		Token to push back
			can be NULL (no action taken if so)

Returns :

	<value>		TRUE if OK
			FALSE if not (may take error exit)

Notes :

	This function reserves the token (removes it from the free
	pool).  Thus it MUST NOT already be reserved.

*/

BOOL
mfl_tokback ARGLIST( (inP, tokP ) )
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( TOKEN		*tokP )		/* Token to re-use */
{
    if ( tokP != NULL ) {
	/* Reserve the token (don't let it be allocated while we are using it */
	mfl_tokreserve( tokP );

	/* Push the token onto the pushed-back list, singly linked via
	   the next pointer.
	*/
	tokP->tk_nextP = TokpbP;
	TokpbP = tokP;
    }

    /* All set */
    return ( TRUE );
}
/*

*//* mfl_tokback_restore( pbtokP )

	Restores a previously saved tokback context

Accepts :

	pbtokP		The previously saved tokback head
			(likely from mfl_tokback_save())

Returns :

	<nothing>

Notes :


*/

void
mfl_tokback_restore ARGLIST( (pbtokP) )
    FARG( TOKEN		*pbtokP )	/* Saved tokback context (head) */
{
	TOKEN		*tokP;

    /* If there's anything on the tokback list, get rid of it */
    while ( ( tokP = TokpbP ) != NULL ) {
	TokpbP = tokP->tk_nextP;	/* Unlink node from list */
	mfl_tokrelease( tokP );		/* It's free again */
    }

    /* Restore the saved tokback head */
    TokpbP = pbtokP;

    /* done */
}
/*

*//* mfl_tokback_save()

	Save current tokback state (to caller), clear the list

Accepts :



Returns :

	<value>		The head of the tokback list prior to clearing

Notes :

	When a caller wants to start a new parsing context, the currently
	pushed-back tokens can't be allowed to interfere with it, yet they
	will needed again in the old context once that new parsing context
	is finished.  This function just clears the list, but returns the
	head of the current tokback list to the caller so that the caller
	may later restore it.

*/

TOKEN *
mfl_tokback_save NOARGLIST
{
	TOKEN		*pbtokP;

    pbtokP = TokpbP;			/* Remember list head */
    TokpbP = NULL;			/* Clear list */
    return ( pbtokP );			/* Return old head to caller */
}
/*

*//* mfl_tokrelease( tokP )

	Release a token back to the pool

Accepts :

	tokP		Ptr to token to be released

Returns :

	<nothing>

Notes :

	Tokens may be removed from the free pool via mfl_tokreserve().
	This function puts back a token (or perhaps puts a new one
	back into the pool).

*/

void
mfl_tokrelease ARGLIST( (tokP) )
    FARG( TOKEN		*tokP )		/* Token to release. */
{
#if 0
    /* This just goes to the head of the list. */
    tokP->tk_nextP = ToklistP;
    if ( ToklistP != NULL )
	ToklistP->tk_prevP = tokP;
    ToklistP = tokP;
    tokP->tk_prevP = NULL;
#else
    /* Release the token such that it's the last one on the fetch list */
    if ( ToknextP == NULL )
	ToknextP = ToklistP;
    if ( ( tokP->tk_nextP = ToknextP ) != NULL ) {
	tokP->tk_prevP = ToknextP->tk_prevP;
	ToknextP->tk_prevP = tokP;
    }
    else
	tokP->tk_prevP = NULL;
    if ( tokP->tk_prevP != NULL )
	tokP->tk_prevP->tk_nextP = tokP;
    else
	ToklistP = tokP;
#endif

    ++ToklistC;			/* Update free list count */
}
/*

*//* mfl_tokreserve( tokP )

	Reserve a token out of the free pool

Accepts :

	tokP		Ptr to token to be reserved

Returns :

	<nothing>

Notes :

	tokP indicates a token that has previously been gotten
	from the free pool (but is still in the pool, ready for
	re-use in typical round-robin fashion).  This routine
	takes the token out of the pool so it won't be given
	to some other caller until it's safe to use again.

	Note that this only works if all users of tokens treat them
	as if they have very limited lifetimes.

*/

void
mfl_tokreserve ARGLIST( (tokP) )
    FARG( TOKEN		*tokP )		/* Token to release. */
{
	TOKEN		*ptokP;		/* Previous adjacent token */
	TOKEN		*ntokP;		/* Next adjacent token */

    /* Must unlink from the doubly linked list. */
    ptokP = tokP->tk_prevP;
    ntokP = tokP->tk_nextP;

    if ( ptokP == NULL )
	ToklistP = ntokP;		/* No previous.. repoint head. */
    else
	ptokP->tk_nextP = ntokP;

    if ( ntokP != NULL )
	/* Change the next adjacent token to point to previous.. Note
	   that we don't maintain a tail pointer so we don't have to keep
	   that up to date.
	*/
	ntokP->tk_prevP = ptokP;

    /* Make sure this token isn't the one about to be taken next. */
    if ( ToknextP == tokP )
	ToknextP = ntokP;

    /* Update the free list count; if the list is too small, add more. */
    for ( --ToklistC; ToklistC < 10; )
	tok_seed();
}
/*

*//* mfl_ppftoken( inP )

	Consume input to the next directive when preprocessor is false.

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		TRUE, normally
			FALSE if some error or EOF seen.

Notes :

	This is called by the preprocessor when in a false condition.  It
	ignores input up to the next potential preprocessor line, (which
	it processes), paying attention to comments -- i.e., a
	preprocessor line inside a comment is ignored.

	This function does not actually return a token, but it has
	"token" in the name because it has to coordinate with the
	tokenizer state.

*/

BOOL
mfl_ppftoken ARGLIST( (inP ) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		ch;
	TOKEN		*tokP;

    /* Must empty the pushed-back token list first.  There should never
       be anything we need here.
    */
    while ( ( tokP = TokpbP ) != NULL ) {
	TokpbP = tokP->tk_nextP;	/* Unlink it */
	mfl_tokrelease( tokP );		/* Put it back in the pool */
    }

    /* Loop skipping comments until we have a reason to stop */
    for ( ; ; ) {
	/* Get next non-blank and check for beginning of line */
	ch = in_nonblank( inP, FALSE );
	if ( ch == EOF )
	    return ( FALSE );

	/* Check for preprocessor introducer at beginning of line */
	if ( inP->in_bolF && (
#ifdef	PREPROCESSOR_VIA_HASHMARK
	     ( ch == '#' ) ||
#endif

#ifdef	PREPROCESSOR_VIA_ATSIGN
	     ( ch == '@' ) ||
#endif

	      FALSE ) ) {

	    pp_directive( inP );	/* Ignore errors */
	    break;			/* Be done */
	}

	/* Can now do comment skipping without fear of missing something
	   at the beginning of a line.
	*/
	ch = non_comment_char( inP, ch );
	if ( ch == EOF )
	    return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* mfl_sigtoken( inP )

	Input the next significant token from an input stack

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		Ptr to TOKEN struct
			NULL if error

Notes :

	A "significant" token is one that has some non-whitespace
	meaning in the language.

	The return is the same as from mfl_token()

*/

TOKEN *
mfl_sigtoken ARGLIST( (inP ) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	TOKEN		*tokP;		/* Ptr to returned token */

    /* Loop until we get a significant token */
    for ( ; ; ) {
	tokP = mfl_token( inP );
	if ( ( tokP == NULL ) || ( tokP->tk_type != TK_EOL ) )
	    break;

	/* We are going to cheat and reuse this same token.. */
	ToknextP = tokP;
    }

    return ( tokP );
}
/*

*//* mfl_token( inP )

	Input the next token from an input stack

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		Pointer to a token, or
			NULL if error.

Notes :

	The returned token is one of a small number of tokens kept in a
	pool and that are used in a round-robin fashion.  The caller can
	use the token info for a short time, and may even consume a small
	number of additional tokens, but remember that the token remains
	in the pool and can be used again.  If the returned token needs
	to be kept for any significant length of time during which other
	tokens may be consumed, there then being a danger than the same
	token will be re-used, the caller should reserve the token via
	mfl_tokreserve() and release it via mfl_tokrelease() when done
	with it.

	This routine also looks for any preprocessor introducer
	characters at beginning the line, and calls the preprocessor
	when any such introducer is found.

*/

TOKEN *
mfl_token ARGLIST( (inP ) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		i;		/* Scratch */
	int		oplen;		/* Operator string length */
	int		opX;		/* Operator table index */
	int		opmat;		/* Matched operator */
	int		ch;		/* Character */
	long		intscale;	/* Integer scale factor */
	OPINFO		*opP;		/* Ptr to operator table */
	PPSYM		*ppsymP;	/* Preprocessor symbol ptr */
	BISSYM		*bissymP;	/* Built-in statement symbol ptr */
	TOKEN		*tokP;		/* Token being returned */
	char		opbuf[10];	/* Buffer for operator */
	char		rawbuf[10];	/* Buffer for raw char */

    /* Use any token that was previously pushed back. */
    if ( ( tokP = TokpbP ) != NULL ) {
	TokpbP = tokP->tk_nextP;	/* Unlink it */
	mfl_tokrelease( tokP );		/* Put it back in the pool */
	return ( tokP );
    }

    /* Get next non-blank char, ignoring escaped blanks and comments */
    for ( ; ; ) {
	/* Next non-blank */
	ch = in_nonblank( inP, FALSE );

	/* Check for escaped char */
	if ( ch == '\\' ) {
	    ch = in_char( inP );

	    /* Ignore escaped whitespace and end-of-line chars */
	    if ( ( ch == ' ' ) ||
	         ( ch == '\t' ) ||
	         ( ch == '\n' ) ||
	         ( ch == '\r' ) )
		continue;
	}

	/* Check for preprocessor introducer at beginning of line.
	   This has to be done inside the comment-skipping loop since
	   the preprocessor introducer might be the same character
	   as one that signals an end-of-line comment.  ugh, but ok.
	*/
	if ( inP->in_bolF && (
#ifdef	PREPROCESSOR_VIA_HASHMARK
	     ( ch == '#' ) ||
#endif

#ifdef	PREPROCESSOR_VIA_ATSIGN
	     ( ch == '@' ) ||
#endif

	      FALSE ) ) {

	    pp_directive( inP );	/* Ignore errors */

	    /* Recurse to get next token.  Note- can not just continue
	       here, since pp_directive() may have caused some token to
	       be pushed back and if so we have to deal with it as an
	       entire token.
	    */
	    return ( mfl_token( inP ) );
	}

	/* Skip any comment(s) and use the character that follows. */
	ch = non_comment_char( inP, ch );
	break;
    }

    if ( ch == EOF )
	return ( NULL );

    /* Now we'll need a token */
    if ( ( tokP = tok_grab() ) == NULL )
	return ( NULL );

    /* Check for string */
    if ( ch == '"' ) {
	/* For a string we also maintain the raw input. */
	tokP->tk_rawF = TRUE;
	tokP->tk_rawP->bs_bC = 0;
	bstr_putb( tokP->tk_rawP, ch );

        for( ; (ch = in_char(inP)) != EOF; ) {	/* Collect the string */
	    if ( ch == '"' ) {
		bstr_putb( tokP->tk_rawP, ch );

#ifdef	ADJACENT_LITSTR_CONCAT
		/* Adjacent strings function as string concatenation,
		   so we have to look for another opening quote
		*/
#if 0
  /* string concatenation across lines has problems */
		ch = in_nonblank( inP, TRUE );
#else
		ch = in_nonblank( inP, FALSE );
#endif
		if ( ch == '"' ) {
		    /* Yes, adjacent string. */
		    /* Add new opening quote to raw string */
		    bstr_putb( tokP->tk_rawP, ch );

		    continue;		/* Keep taking string chars */
		}

		/* Really done.  Take the char back */
		if ( ch != EOF )
		    in_charback( inP, ch );
#endif	/* ADJACENT_LITSTR_CONCAT */

		break;
	    }

	    if ( ch == '\\' ) {
		/* This is an escape char.. get the translation */
		in_charback( inP, ch );
		ch = in_charesc( inP, &rawbuf[0] );

		/* Store the raw input */
		bstr_cat( tokP->tk_rawP, &rawbuf[0] );
	    }
	    else
		/* Just take it as the raw char */
		bstr_putb( tokP->tk_rawP, ch );

	    /* Store the char */
	    bstr_putb( tokP->tk_bstrP, ch );
	}

	bstr_nulterm( tokP->tk_bstrP );
	bstr_nulterm( tokP->tk_rawP );

	tokP->tk_type = TK_LITSTR;
    }

    /* Quoted character constant */
    else if ( ch == '\'' ) {
	/* We maintain the raw input for this. */
	tokP->tk_rawF = TRUE;
	tokP->tk_rawP->bs_bC = 0;
	bstr_putb( tokP->tk_rawP, ch );

	/* Get the char, which can be escaped */
	ch = in_charesc( inP, &rawbuf[0] );

	tokP->tk_val.tv_long = ch & 0xff;

	/* Append the raw input to the raw bstr */
	bstr_cat( tokP->tk_rawP, &rawbuf[0] );

	/* Next char is supposed to be "'" */
	ch = in_char( inP );
	if ( ch != '\'' ) {
	    warning( "badly terminated character constant" );
	    in_charback( inP, ch );
	}
	else {
	    /* Got the terminator- store in raw buf */
	    bstr_putb( tokP->tk_rawP, ch );
	}

	bstr_nulterm( tokP->tk_bstrP );
	bstr_nulterm( tokP->tk_rawP );

	tokP->tk_type = TK_INTEGER;
    }

    /* See if it's a symbol */
    else if ( ( ch == '_' ) || ( ch == '$' ) || isalpha( ch ) ) {
	for( ; ; ) {			/* Collect the symbol */
	    bstr_putb( tokP->tk_bstrP, ch );
	    ch = in_char( inP );	/* Get next char */
	    if ( ( ch != '_' ) && ( ch != '$' ) &&
	         ( !isalpha( ch ) ) && ( !isdigit( ch ) ) )
		 break;
	}

	bstr_nulterm( tokP->tk_bstrP );

	/* Check for special "text:" reserved word */
	if ( ( ch == ':' ) &&
	     ( stricmp( (char *)tokP->tk_bstrP->bs_bP, "text" ) == 0 ) )
	    return( tok_text( inP, tokP ) );

	in_charback( inP, ch );		/* Take terminator back */

	/* Do macro substitution if possible and desirable */
	if ( MexpandF &&
	   ( ( ppsymP = (PPSYM *)symtab_lookup( STpreproc,
				tokP->tk_bstrP->bs_bP ) ) != NULL ) &&
	   ( !ppsymP->pp_blue ) ) {

	    /* Found macro */
	    pp_invoke( inP, ppsymP );
	    return ( mfl_token( inP ) );
	}

	else {
	    /* Check out the token found */
	    bissymP = (BISSYM *)symtab_lookup( STbis, tokP->tk_bstrP->bs_bP );
	    if ( bissymP != NULL ) {
		/* It's a built-in symbol */
		tokP->tk_type = TK_BIS;
		tokP->tk_val.tv_int = bissymP->bss_num;
	    }
	    else {
	    	/* General symbol */
		tokP->tk_type = TK_SYM;
	    }
	}
    }


    /* Check for integer or floating point */
    else if ( isdigit( ch ) ) {
    	UBYTE   fltF = 0;		/* Floating point state */
	BOOL	hexF = FALSE;		/* Hex flag */
	int	eptX;			/* Index to E */

	/*                          nnnnnnn[.[nnnnnnn]][E[+]nnnn]
	      Floating point state: 0000000 0 1111111   1 2 3444
	*/
	for( i = 0; ; ch = in_char( inP ), ++i ) {
	    if ( isdigit(ch) ||
	    	 ( ( ( ch == 'x' ) || ( ch == 'X' ) ) && 
		 	( i == 1 ) && ( tokP->tk_bstrP->bs_bP[0] == '0' ) ) ||
		 ( hexF && ( ( ( ch >= 'a' ) && ( ch <= 'f' ) ) ||
		 	     ( ( ch >= 'A' ) && ( ch <= 'F' ) ) ) ) ||
		 ( !hexF &&
		   ( ( ch == '.' ) && ( fltF == 0 ) ) ||
		   ( ( ( ch == 'e' ) || ( ch == 'E' ) ) && ( fltF < 2 ) ) ||
		   ( ( ( ch == '+' ) || ( ch == '-' ) ) && ( fltF == 2 ) )
		 )
	       ) {

		if ( ch == 'X' )
		    ch = 'x';
		if ( ch == 'x' )
		    hexF = TRUE;
		else if ( ch == '.' )
		    fltF = 1;
		else if ( ( ( ch == 'e' ) || ( ch == 'E' ) ) && !hexF ) {
		    fltF = 2;
		    eptX = tokP->tk_bstrP->bs_bC;
		}
		else if ( ( ch == '+' ) || ( ch == '-' ) )
		    fltF = 3;
		else if ( ( fltF > 1 ) && isdigit( ch ) )
		    fltF = 4;

		bstr_putb( tokP->tk_bstrP, ch );
	    }
	    else
	        break;
	}

	bstr_nulterm( tokP->tk_bstrP );

	/* Return the appropriate type */
	if ( fltF != 0 ) {
	    /* Floating point value */
	    if ( ( fltF == 2 ) || ( fltF == 3 ) ) {
	    	/* Ends with "E" or "E+" -- give message and take it */
		warning( "warning--badly formed floating point number" );
		tokP->tk_bstrP->bs_bP[ eptX ] = NUL;
	    }

	    tokP->tk_val.tv_float = atof( (char *)tokP->tk_bstrP->bs_bP );
	    tokP->tk_type = TK_FLOAT;
	}
	else {
	    /* Integer */

	    tokP->tk_type = TK_INTEGER;
	    if ( hexF )
		tokP->tk_val.tv_long = atoh( (char *)&tokP->tk_bstrP->bs_bP[2] );
	    else if ( *tokP->tk_bstrP->bs_bP == '0' )
		tokP->tk_val.tv_long = atoo( (char *)tokP->tk_bstrP->bs_bP );
	    else
	    	tokP->tk_val.tv_long = atod( (char *)tokP->tk_bstrP->bs_bP );

	    /* Process any multiplier char:
		k = kibi (*1024)
		m = mebi (*1024*1024)
		g = tebi (*1024*1024*1024)
	    */
	    if ( ( ch == 'k' ) || ( ch == 'K' ) )
		intscale = 1024;
	    else if ( ( ch == 'm' ) || ( ch == 'M' ) )
		intscale = 1024L * 1024;
	    else if ( (ch == 'g' ) || ( ch == 'G' ) )
		intscale = 1024L * 1024 * 1024;
	    else
		intscale = 0;
	    if ( intscale != 0 ) {
		tokP->tk_val.tv_long *= intscale;
		bstr_putb( tokP->tk_bstrP, ch );
		bstr_nulterm( tokP->tk_bstrP );
		ch = in_char( inP );
	    }
	}

	in_charback( inP, ch );
    }

    else if ( ( ch == '\n' ) || ( ch == '\r' ) || ( ch == '\014' ) ) {
	/* end-of-line */
	bstr_putb( tokP->tk_bstrP, ch );
	bstr_nulterm( tokP->tk_bstrP );
	tokP->tk_val.tv_char = ch;
	tokP->tk_type = TK_EOL;
    }

    else {
    	/* Try for operator */
	opmat = -1;			/* No operator found */
	oplen = 1;			/* Current operator length */
	opbuf[0] = ch;			/* Init operator string */

	for( opP = &Optbl[0], opX = 0; opP->op_str != NULL;
    		++opP, ++opX ) {

	    /* Check current operator */
	    if ( memcmp( &opbuf[0], opP->op_str, oplen ) == 0 ) {
		opmat = opX;		/* Is now the matched opr */

		/* Keep gathering characters until no match */
		for( ; (ch = in_char(inP)) != EOF; ) {
		    opbuf[oplen] = ch;	    /* Store in buffer */
		    if ( opP->op_str[oplen++] != ch )
		    	break;		    /* Reach non-match */
		}
		if ( ch == EOF )
		    break;
	    }
	}

	/* See if we found an operator */
	if ( opmat != -1 ) {
	    /* Hack in check for ellipsis here */
	    if ( ( ch == '.' ) && ( opbuf[0] == '.' ) ) {
	    	tokP->tk_type = TK_ELLIPSIS;	/* Note loose definition of ellipsis */
		opbuf[oplen++] = ch;
		ch = in_char( inP );
		if ( ch != '.' )
		    in_charback( inP, ch );
		else
		    opbuf[oplen++] = ch;
	    }
	    else {
		if ( ch != NUL ) {
		    in_charback( inP, ch );	/* Take back char not used */
		    --oplen;
		}
        	tokP->tk_val.tv_int = opmat;	/* Store operator number */
		tokP->tk_type = TK_OPR;	/* Type is operator */
	    }

	    bstr_catn( tokP->tk_bstrP, (UBYTE *)&opbuf[0], oplen );
	    bstr_nulterm( tokP->tk_bstrP );
	}
    }

    if ( tokP->tk_type == TK_UNK ) {
	/* Can't figure out what it is -- just return the character. */
	bstr_putb( tokP->tk_bstrP, ch );
	bstr_nulterm( tokP->tk_bstrP );
	tokP->tk_val.tv_char = ch;
	tokP->tk_type = TK_CHAR;
    }

    /* Warn if token overflow */
    if ( ( tokP->tk_bstrP->bs_bC >= tokP->tk_bstrP->bs_bM ) ||
         ( tokP->tk_rawF &&
	   ( ( tokP->tk_rawP->bs_bC >= tokP->tk_rawP->bs_bM ) ) ) )
	mfl_warning( inP, "token overflow." );

    return( tokP );
}
/*

*//* mfl_toktype( type )

	Return token type string

Accepts :

	type		The token type

Returns :

	<value>		String corresponding to token type

Notes :


*/

char *
mfl_toktype ARGLIST( (type) )
    FARG( TOKTYPE	type )		/* token type */
{
    switch (type) {
	case TK_UNK:
	    return( "TK_UNK" );

	case TK_BIS:
	    return( "TK_BIS" );

	case TK_SYM:
	    return( "TK_SYM" );

	case TK_OPR:
	    return( "TK_OPR" );

	case TK_LITSTR:
	    return( "TK_LITSTR" );

	case TK_SEMI:
	    return( "TK_SEMI" );

	case TK_INTEGER:
	    return( "TK_INTEGER" );

	case TK_FLOAT:
	    return( "TK_FLOAT" );

	case TK_CHAR:
	    return( "TK_CHAR" );

	case TK_EOL:
	    return( "TK_EOL" );

	case TK_ELLIPSIS:
	    return( "TK_ELLIPSIS" );

	default:
	    return( "<invalid>" );
    }
}



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* eol_comment( inP )

	Discard everything to the end of the line as a comment.

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		The end-of-line character that we end up with

Notes :

	Comment can be continued across lines if ended with '\'

*/

static int
eol_comment ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Ptr to input stack */
{
	int		ch;

    for ( ; ; ) {
	ch = in_char( inP );
	if ( ch == '\\' )
	    /* Next char is quoted, it stays in the comment. */
	    ch = in_char( inP );
	else if ( ( ch == '\n' ) || ( ch == EOF ) )
	    /* End of line, we are done. */
	    break;
    }

    return ( ch );
}
/*

*//* non_comment_char( inP, nextch )

	Input the next non-comment character from an input stream

Accepts :

	inP		Ptr to input stack
	nextch		The next character in the stream, or EOF

Returns :

	<value>		The next non-comment character

Notes :

	Skips comments, returns the next non-blank character

	Caller may have already inputted the next character, in which
	case it's passed in "nextch".  Otherwise pass EOF to indicate
	that the next character has not yet been gotten.

	end-of-line is considered a non-blank character.

*/

static int
non_comment_char ARGLIST( (inP, nextch) )
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( int		nextch )	/* Next character */
{
	int		ch;

    ch = ( nextch != EOF ) ? nextch : in_nonblank( inP, FALSE );

    /* This may iterate through multiple comments */
    for ( ; ; ch = in_nonblank( inP, FALSE ) ) {

	/* '#' opens a comment through the end of the line */
	if ( ch == '#' ) {
	    ch = eol_comment( inP );
	    break;
	}

	/* Check for '/'-style comments */
	if ( ch == '/' ) {
    	    ch = in_char( inP );
	    if ( ch == '*' ) {
		/* Found comment block start -- skip to end */
		for( ch = in_char( inP ); ; ) {
		    if ( ch == '*' ) {
			if ( (ch = in_char( inP ) ) == '/' )
			    break;
		    }
		    else
			ch = in_char( inP );
		}
		continue;		/* Skip more */
	    }
	    else if ( ch == '/' ) {
		ch = eol_comment( inP ); /* Found end-of-line comment */
		break;
	    }
	    else {
		/* Not comment */
		in_charback( inP, ch );
		ch = '/';
		break;
	    }
	}

	/* No comment.. */
	break;
    }

    return ( ch );
}
/*

*//* tok_grab()

	Allocates a token for temporary use

Accepts :


Returns :

	<value>		Address of token structure to use
			NULL if error (may take error exit)

Notes :

	Users of tokens returned by public routines in this module
	are expected to treat a returned token as if it has a very
	short lifetime, however we do want to allow a few tokens
	to be outstanding without conflict.  A short doubly linked
	list of token structures is maintained in this routine.
	Tokens may be temporarily removed from the list, mainly when a
	token is pushed back on the input as that is the most likely
	area of conflict.

*/

static TOKEN *
tok_grab NOARGLIST
{
	int		i;
	TOKEN		*tokP;		/* Ptr to token */

    /* ToknextP points to the token following the one most recently
        taken.  That may have been a NULL pointer to wrap back to
	the beginning.
    */
    if ( ( tokP = ToknextP ) == NULL )
	tokP = ToklistP;
    if ( tokP != NULL ) {
	/* Got one.  Initialize it for use. */
	tokP->tk_bstrP->bs_bC = 0;
	tokP->tk_type = TK_UNK;
	tokP->tk_rawF = FALSE;

	/* Set up the next pointer for next time. */
	ToknextP = tokP->tk_nextP;

	/* All set. */
	return ( tokP );
    }

    /* There are no tokens in the list.  Seed the list and try again.
       (This should really only happen the first time through)
    */
    for ( i = 0; i < 5; ++i ) {
	tok_seed();
    }
    return( tok_grab() );
}
/*

*//* tok_seed()

	Seeds a new token into the free token pool

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not (may take error exit)

Notes :


*/

static BOOL
tok_seed NOARGLIST
{
	TOKEN		*tokP;		/* A token */

    /* Allocate token structure */
    tokP = emalloc( "tok_seed", "new token", sizeof(TOKEN) );
    if ( tokP == NULL )
	return ( FALSE );

    /* Create the BSTRs */
    tokP->tk_bstrP = bstr_new( "token string in tok_seed",
			         MFL_TOKSIZE, MFL_TOKMAX );
    if ( tokP->tk_bstrP == NULL ) {
	dealloc( tokP );
	return ( FALSE );
    }
    tokP->tk_bstrP->bs_chunk = MFL_TOKSIZE;
    tokP->tk_rawP = bstr_new( "raw token string in tok_seed",
			         MFL_TOKSIZE, MFL_TOKMAX );
    if ( tokP->tk_rawP == NULL ) {
	bstr_free( tokP->tk_bstrP );
	dealloc( tokP );
	return ( FALSE );
    }
    tokP->tk_rawP->bs_chunk = MFL_TOKSIZE;

    /* Release it for use */
    mfl_tokrelease( tokP );

    return ( TRUE );
}
/*

*//* tok_text( inP, tokP )

	Process a "text:" input string

Accepts :

	inP		Ptr to input stack
	tokP		Where to put the token info

Returns :

	<value>		The token, or NULL if error.

Notes :

	This is a slave to mfl_token() and so makes some assumptions.

	tokP is assumed to have been grabbed by mfl_token() and is
	available to be filled in in the ways that mfl_token() does.

	If there is a syntax error, the token is released an this
	routine returns NULL.  Thus, mfl_token() can simply return
	the value that this function returns.

*/

static TOKEN *
tok_text ARGLIST( (inP, tokP) )
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( TOKEN		*tokP )		/* The token to fill in */
{
	BOOL		okF;
	int		ch;
	char		rawbuf[10];	/* Buffer for raw char */

    /* Reset the token to accept string input */
    tokP->tk_bstrP->bs_bC = 0;

    /* For a string we also maintain the raw input.  This includes
       the "text:" keyword, or at least a replacement of it..
    */
    tokP->tk_rawF = TRUE;
    tokP->tk_rawP->bs_bC = 0;
    bstr_cat( tokP->tk_rawP, "text:\n" );

    /* Skip to the end of line.  We'll complain if non-whitespace follows
       but we'll still collect the string.
    */
    for ( okF = TRUE; ( ch = non_comment_char( inP, EOF ) ) != EOF;  ) {
	if ( ch == '\n' )
	    break;
	if ( ( ch != ' ' ) &&
	     ( ch != '\t' ) &&
	     ( ch != '\r' ) )
	    okF = FALSE;
    }
    if ( !okF )
	mfl_warning( inP, "ignoring junk after \"text:\"" );

    /* Collect the string */
    while ( ( ch = in_char( inP ) ) != EOF ) {
	if ( inP->in_bolF && ( ch == '.' ) ) {
	    /* Dot on beginning of line could end the input.
	       This stuff about keeping the raw input makes things uglier.
	    */
	    bstr_putb( tokP->tk_rawP, '.' );
	    ch = in_char( inP );
	    if ( ch == '\r' ) {
		/* Hack to allow CRLF */
		ch = in_char( inP );
		if ( ch != '\n' ) {
		    in_charback( inP, ch );
		    ch = '\r';
		}
		else
		    bstr_putb( tokP->tk_rawP, '\r' );
	    }
	    if ( ( ch == '\n' ) || ( ch == EOF ) ) {
		bstr_putb( tokP->tk_rawP, ch );
		break;
	    }

	    /* At this point we'll keep the dot and swallow any
	       adjacent dots (two dots turn into one)
	    */
	    bstr_putb( tokP->tk_bstrP, '.' );
	    if ( ch == '.' )
		bstr_putb( tokP->tk_rawP, '.' );
	    else
		in_charback( inP, ch );
	    continue;
	}

	if ( ch == '\\' ) {
	    /* This is an escape char.. get the translation */
	    in_charback( inP, ch );
	    ch = in_charesc( inP, &rawbuf[0] );

	    /* Store the raw input */
	    bstr_cat( tokP->tk_rawP, &rawbuf[0] );
	}
	else 
	    /* Just take it as the raw char */
	    bstr_putb( tokP->tk_rawP, ch );

	/* Store the char */
	bstr_putb( tokP->tk_bstrP, ch );
    }

    bstr_nulterm( tokP->tk_bstrP );
    bstr_nulterm( tokP->tk_rawP );
    tokP->tk_type = TK_LITSTR;

    return ( tokP );
}
