/* mfl-type.c -- data type stuff for mfl

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains routines dealing with data types and values

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mfl.h"
#include "mfl-type.h"
#include "mfl-eval.h"

#include "errwarn.h"

/* Local definitions */



/* External data referenced */

extern	SYMTAB	*STstruct;		/* Structure names */
extern	SYMTAB	*STtpool;		/* Type pool symbol table */
extern	TNSYM	*Tnint1;
extern	TNSYM	*Tnint2;
extern	TNSYM	*Tnint4;
extern	TNSYM	*Tnstring;
extern	TYPSYM	*TystrpP;

/* External routines used */


/* Local data publicly available */



/* Local routines and forward references */



/* Private data */

static	FSV	*FsvpoolP;		/* Used, free FSVs */

/*

*//* dvs_access( dvsP )

	Claim an access to a DVS object

Accepts :

	dvsP		Ptr to dvs object

Returns :

	<value>		same as dvsP, for convenience

Notes :


*/

DVS *
dvs_access ARGLIST( (dvsP) )
    FARG( DVS		*dvsP )		/* DVS object */
{
    /* Claim access & check for potential badness */
    if ( ++dvsP->dv_refC <= 1 ) {
	error( "dvs_access: bad reference count." );
	dvsP = NULL;
    }

    return ( dvsP );			/* Return the object reference */
}
/*

*//* dvs_new( size, max )

	Create a new dvs object

Accepts :

	size		Number of bytes to allocate in object
	max		If non-zero, a maximum size that the object
			can grow to.

Returns :

	<value>		Ptr to dvs object if successful
			NULL if error (may take error exit)

Notes :

	Also makes reference (i.e., sets reference count to 1)
	  to the new object.

*/

DVS *
dvs_new ARGLIST( (size, max) )
   NFARG( int		size )		/* Initial dvs size  */
    FARG( int		max )		/* Any max size */
{
	DVS		*dvsP;		/* New dvs object */

    /* Allocate dvs object */
    dvsP = emalloc( "dvs_new", "dvs object", sizeof(DVS) );
    if ( dvsP == NULL )
	return ( NULL );

    /* Create the storage */
    dvsP->dv_dataP = emalloc( "dvs_new", "new DVS data", size );
    if ( dvsP->dv_dataP == NULL ) {
	dealloc( dvsP );
	return ( NULL );
    }
    memset( dvsP->dv_dataP, 0, size );

    dvsP->dv_size = size;
    dvsP->dv_max = max;

    /* Claim first reference */
    dvsP->dv_refC = 1;

    /* Return it. */
    return ( dvsP );
}
/*

*//* dvs_release(dvsP)

	Release a reference to a DVS object

Accepts :

	dvsP		Ptr to DVS object to release

Returns :

	<nothing>

Notes :

	On a final release to a dvs object, the object is deleted.
	If you have a sequence of releases/accesses, it would be best
	to do the accesses first, so as not to accidentally delete
	an about-to-be-needed object.

*/

void
dvs_release ARGLIST( (dvsP) )
    FARG( DVS		*dvsP )		/* Ptr to DVS object */
{
    if ( ( --dvsP->dv_refC ) == 0 ) {
	/* Final release-- delete it. */
	dealloc( dvsP->dv_dataP );
	dealloc( dvsP );
    }
    else if ( dvsP->dv_refC < 0 ) {
	error( "dvs_release: negative reference count." );
    }
}
/*

*//* fsv_access( fsvP )

	Add an access to a free-standing VALPTR 

Accepts :

	fsvP		FSV to access

Returns :

	<nothing>

Notes :


*/

void
fsv_access ARGLIST( (fsvP) )
    FARG( FSV		*fsvP )		/* FSV to access */
{
    /* Claim access & check for bad reference */
    if ( ++fsvP->fv_refC <= 1 ) {
	error( "fsv_access: bad reference count." );
    }
}
/*

*//* fsv_new()

	Allocate a free-standing VALPTR

Accepts :

	<nothing>

Returns :

	<value>		Address of FSV allocated,
			NULL if failure

Notes :

	It's assume that the caller will want to initialize the VALPTR
	immediately, so the VALPTR contents are in an undetermined state.
	

*/

FSV *
fsv_new NOARGLIST
{
	FSV		*fsvP;

    /* Previously released FSVs are kept on a pool, linked together
       via their VALPTR genP.  Try that pool first.
    */
    if ( ( fsvP = FsvpoolP ) != NULL ) {
	FsvpoolP = (FSV *)fsvP->fv_data.d_ptr.p_ptr.p_genP;
    }
    else {
	fsvP = (FSV *)emalloc( "fsv_new", "fsv", sizeof( FSV ) );
	if ( fsvP == NULL )
	    return ( NULL );
    }

    fsvP->fv_refC = 1;			/* Count first reference */

    return ( fsvP );
}
/*

*//* fsv_release( fsvP )

	Release access to a free-standing VALPTR 

Accepts :

	fsvP		FSV to release

Returns :

	<nothing>

Notes :

	Upon last release of FSV, the FSV is deleted.

*/

void
fsv_release ARGLIST( (fsvP) )
    FARG( FSV		*fsvP )		/* FSV to access */
{
	VALPTR		*vptrP;

    if ( --fsvP->fv_refC == 0 ) {
	/* If the FSV contains a pointer, release the pointed-to
	   thing.  This is a real hack in lieu of properly recursing
	   through released objects.
	*/
	if ( fsvP->fv_ptrF ) {
	    vptrP = (VALPTR *)&fsvP->fv_data.d_ptr;
	    if ( vptrP->p_ptr.p_genP != NULL ) {
		if ( vptrP->p_byteX >= 0 )
		    dvs_release( vptrP->p_ptr.p_dvsP );
		else if ( vptrP->p_byteX == PO_REFSTR )
		    refstr_free( (REFSTR *)vptrP->p_ptr.p_genP );
		/* Don't care about anything else.  I told you this was
		   a hack.
		*/
	    }
	}

	/* Time to free it.  Add it to the used FSV pool. */
	fsvP->fv_data.d_ptr.p_ptr.p_genP = FsvpoolP;
	FsvpoolP = fsvP;
    }
    else if ( fsvP->fv_refC < 0 ) {
	warning( "fsv_release: negative access." );
    }
}
/*

*//* odinfo_delete( infoP )

	Deletes any info from an opaque dynamic info block.

Accepts :

	infoP		Ptr to ODINFO opaque dynamic info block

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Does not delete the block itself.

*/

BOOL
odinfo_delete ARGLIST( (infoP) )
    FARG( ODINFO	*infoP )	/* The info ptr */
{
    if ( ( infoP != NULL ) && ( infoP->od_P != NULL ) ) {
	/* Only delete if there's a deletion routine */
	if ( infoP->od_del != NULL )
	    (*infoP->od_del)( infoP->od_P );

	infoP->od_P = NULL;
    }

    return ( TRUE );
}
/*

*//* tmpval()

	Access a temporary VALUE structure

Accepts :



Returns :

	<value>		Pointer to a VALUE structure
			NULL if error (warning given)

Notes :

	Returns a pointer to one of several scratch VALUE
	structures.  In the same vein as tmpstr(), these are
	available as *very* temporary registers.  Each call to
	this routine will return the next available buffer in
	round-robin fashion.  Caller must make sure the use is
	indeed temporary.

*/

VALUE *
tmpval NOARGLIST
{
#define	TMPVALNUM	5
static	int		tmpvalX = -1;
static	VALUE		tmpvals[ TMPVALNUM ];

    if ( ++tmpvalX == TMPVALNUM )
	tmpvalX = 0;

    return ( &tmpvals[ tmpvalX ] );
}
/*

*//* type_agg_member(typeP, mnameP)

	Find an aggregate member definition for an aggregate type

Accepts :

	typeP		The type to find the aggregate member for
	mnameP		The member name

Returns :

	<value>		Pointer to aggregate member struct if found
			NULL if not

Notes :

	Caller must make sure that the typeP is an aggregate type.

*/

AGGMEM *
type_agg_member ARGLIST( (typeP, mnameP) )
   NFARG( TYPE		*typeP )	/* Ptr to type */
    FARG( char		*mnameP )	/* name of member to look up */
{
	int		i;
	TYPEAG		*agP;		/* Ptr to aggregate */
	AGGMEM		*agmemP;	/* Ptr to aggregate member */

    /* Lookup the member */
    agP = (TYPEAG *)typeP->ty_nextP;
    for( i = 0, agmemP = &agP->ag_memP[0];
		i < agP->ag_memC;
		++i, ++agmemP ) {
	if ( strcmp( mnameP, agmemP->am_nameP ) == 0 )
	    break;
    }

    if ( i == agP->ag_memC )
	return ( NULL );		/* Member not found. */

    return ( agmemP );
}
/*

*//* type_aggp_member(typeP, mnameP)

	Find an aggregate member definition for a ptr-to-aggregate type

Accepts :

	typeP		The type to find the aggregate member for
	mnameP		The member name

Returns :

	<value>		Pointer to aggregate member struct if found
			NULL if not

Notes :

	Caller must ensure that typeP must be a ptr-to-aggregate type.

*/

AGGMEM *
type_aggp_member ARGLIST( (typeP, mnameP) )
   NFARG( TYPE		*typeP )	/* Ptr to type */
    FARG( char		*mnameP )	/* name of member to look up */
{
    return ( type_agg_member( (TYPE *)(typeP->ty_nextP), mnameP ) );
}
/*

*//* type_aggname(tclass, anameP)

	Make or look up an aggregate type by class and name

Accepts :

	tclass		Type class (TC_ENUM, TC_UNION, TC_STRUCT)
	tnameP		Name of aggregate

Returns :

	<value>		Pointer to TNSYM symbol table entry if success
			NULL if not

Notes :

	If the aggregate name already exists, no check is done on
	the type class.  Type class is only passed to instantiate
	a new name.  Caller should verify that the returned type
	class matches what's expected.

*/

TNSYM *
type_aggname ARGLIST( (tclass, anameP) )
   NFARG( int		tclass )	/* Type class */
    FARG( char		*anameP )	/* Aggregate name */
{
	TYPE		*typeP;
	TNSYM		*tnsymP;	/* aggregate table entry ptr */
	TYPSYM		*typsymP;	/* Ptr to type symbol */

    /* Try to find or make an entry for this name */
    tnsymP = (TNSYM *)symtab_enter( STstruct, anameP, sizeof( TNSYM ) );

    /* If new name, create the aggregate entry */
    if ( ( tnsymP != NULL ) && ( tnsymP->tn_flags & TF_DEFINED ) == 0 ) {
	typsymP = type_mkagg( tclass );
	if ( typsymP == NULL )
	    tnsymP = NULL;
	else {
	    tnsymP->tn_typeP = typeP = &(typsymP->ts_type);
	    tnsymP->tn_flags = TF_DEFINED|( tclass==TC_ENUM?TMENUM:
				     ( tclass==TC_STRUCT?TMSTRUCT:TMUNION));
	    ((TYPEAG *)(typeP->ty_nextP))->ag_tnsymP = tnsymP;
	}
    }

    return( tnsymP );
}
/*

*//* type_is_string( typeP )

	See if a type is a string

Accepts :

	typeP		PTR to a TYPE

Returns :

	<value>		TRUE if type is a string type
			FALSE if not.

Notes :

	A type is a string type if it is a string or if it
	is a pointer to a string.

*/

BOOL
type_is_string ARGLIST( (typeP) )
    FARG( TYPE		*typeP )	/* Ptr to the type */
{
#ifdef	PTR_TO_STRING_IS_STRING
    if ( ( typeP != NULL ) &&
         ( ( typeP->ty_class == TC_STRING ) ||
         ( ( typeP->ty_class == TC_PTR ) &&
	   ( typeP->ty_nextP != NULL ) &&
	   ( ((TYPE *)typeP->ty_nextP)->ty_class == TC_STRING ) ) ) )
	return ( TRUE );
#else
    if ( ( typeP != NULL ) &&
         ( ( typeP->ty_class == TC_STRING ) ) )
	return ( TRUE );
#endif
    return ( FALSE );
}
/*

*//* type_mkagg( tclass )

	Make an aggregate type

Accepts :

	tclass		The type class

Returns :

	<value>		Pointer to TYPSYM for type

Notes :


*/

TYPSYM *
type_mkagg ARGLIST( (tclass) )
    FARG( int		tclass )	/* Type class */
{
	TYPSYM		*tpsymP;	/* Ptr to type pool entry */
	TYPEAG		*agP;		/* Ptr to aggregate block */
	TYPE		dtype;		/* New type */

    /* Allocate an aggregate block */
    agP = (TYPEAG *)emalloc( "mkagtype", "aggregate block",
			       sizeof(TYPEAG) );
    agP->ag_refC = 1;
    agP->ag_memC = 0;
    agP->ag_size = 0;
    agP->ag_tnsymP = NULL;

    /* Make type entry */
    dtype.ty_class = tclass;
    dtype.ty_flags = TF_AGGR | ( tclass == TC_ENUM ? TMENUM :
    			( tclass == TC_STRUCT ? TMSTRUCT : TMUNION ) );
    dtype.ty_size = 1;
    dtype.ty_nextP = (void *)agP;

    tpsymP = (TYPSYM *)symtab_enter( STtpool, &dtype, sizeof(TYPSYM) );

    return( tpsymP );
}
/*

*//* type_print(fP, typeP)

	Print a description of a type

Accepts :

	fP		FILE descriptor to print to
	typeP		Ptr to type tp print

Returns :

	<value>		SCxx code

Notes :


*/

STSCODE
type_print ARGLIST( (fP, typeP) )
   NFARG( FILE		*fP )		/* file to print to */
    FARG( TYPE		*typeP )	/* type to print */
{
	BOOL		firstF;		/* If first */
	TYPEAG		*agP;		/* Ptr to aggregate */

    for( firstF = TRUE; typeP != NULL; firstF = FALSE ) {
        if ( !firstF )
	    fprintf( fP, " " );

	switch( typeP->ty_class ) {
	    case TC_INT:
	    	fprintf( fP, "int(%d)", typeP->ty_size );
		typeP = (TYPE *)typeP->ty_nextP;
		break;


	    case TC_UINT:
	    	fprintf( fP, "unsigned int(%d)", typeP->ty_size );
		typeP = (TYPE *)typeP->ty_nextP;
		break;


	    case TC_FLOAT:
	    	fprintf( fP, "float(%d)", typeP->ty_size );
		typeP = (TYPE *)typeP->ty_nextP;
		break;


	    case TC_PTR:
	    	fprintf( fP, "ptr-to" );
		typeP = (TYPE *)typeP->ty_nextP;
		break;

	    case TC_STRING:
	    	fprintf( fP, "string" );
		typeP = (TYPE *)typeP->ty_nextP;
		break;

	    case TC_DIM:
	    	fprintf( fP, "dim=%d/size=%d", typeP->ty_dim, typeP->ty_size );
		typeP = (TYPE *)typeP->ty_nextP;
		break;

	    case TC_FUNCT:
	    	fprintf( fP, "funct-returning" );
		typeP = (TYPE *)typeP->ty_nextP;
		break;

	    case TC_ENUM:
	        agP = (TYPEAG *)typeP->ty_nextP;
	    	fprintf( fP, "enum" );
		if ( agP->ag_tnsymP != NULL )
		    fprintf( fP, " %s", agP->ag_tnsymP->tn_sym.ss_nameP );
		typeP = NULL;
		break;

	    case TC_STRUCT:
	    	agP = (TYPEAG *)typeP->ty_nextP;
		fprintf( fP, "struct" );
		if ( agP->ag_tnsymP != NULL )
		    fprintf( fP, " %s", agP->ag_tnsymP->tn_sym.ss_nameP );
	    	fprintf( fP, "/size=%d/memC=%d", agP->ag_size,
					agP->ag_memC );
		typeP = NULL;
		break;

	    case TC_UNION:
	    	agP = (TYPEAG *)typeP->ty_nextP;
		fprintf( fP, "union" );
		if ( agP->ag_tnsymP != NULL )
		    fprintf( fP, " %s", agP->ag_tnsymP->tn_sym.ss_nameP );
	    	fprintf( fP, "/size=%d/memC=%d", agP->ag_size,
					agP->ag_memC );
		typeP = NULL;
		break;

	    default:
	    	fprintf( fP, "unknown(%d)!", typeP->ty_class );
		typeP = NULL;
		break;
	}
    }

    return ( SCOK );
}
/*

*//* type_ptr( stypeP )

	Make or find a type that's a pointer to an existing type

Accepts :

	stypeP		Ptr to TYPE for source type

Returns :

	<value>		Ptr to TYPSYM for ptr-to type
			NULL if problem

Notes :

	e.g., given TYPE for type "char" this returns
	TNSYM pointer for type "char *"

*/

TYPSYM *
type_ptr ARGLIST( (stypeP) )
    FARG( TYPE		*stypeP )	/* Source type */
{
	TYPSYM		*tpsymP;	/* Type pool symbol */
	TYPE		newtype;	/* For type of lvalue expr. */

    /* Set up the ptr-to type to look up or enter */
    newtype.ty_flags = 0;
    newtype.ty_nextP = (void *)stypeP;
    newtype.ty_class = TC_PTR;
    newtype.ty_size = sizeof(VALPTR);

    /* Create or lookup the ptr-to type */
    tpsymP = (TYPSYM *)symtab_enter( STtpool, &newtype, sizeof(TYPSYM) );

    return ( tpsymP );
}
/*

*//* type_symtab_comp(symP, keyP)

	Symbol table comparison routine for type

Accepts :

	symP		Ptr to symbol table entry
	keyP		Ptr to candidate key to compare with

Returns :

	<value>		<0 if symP is less than keyP
			0 if comparison is equal
			1 if symP is greater than keyP

Notes :

	Handler routine for mml-symtab symbol table.
*/

int
type_symtab_comp ARGLIST( (symP, keyP) )
   NFARG( TYPSYM	*symP )		/* Ptr to symbol table entry */
    FARG( TYPE		*keyP )		/* Address of candidate key */
{
    if ( ( symP->ts_type.ty_class == keyP->ty_class ) &&
         ( (symP->ts_type.ty_flags|TF_DEFINED) ==
	   (keyP->ty_flags|TF_DEFINED) ) &&
	 ( ( ( keyP->ty_flags & TF_AGGR ) != 0 ) ||
	   ( symP->ts_type.ty_size == keyP->ty_size ) ) &&
	 ( symP->ts_type.ty_nextP == keyP->ty_nextP ) )
	return( 0 );
    else
        return( -1 );
}
/*

*//* type_symtab_hash(symP, keyP)

	Make a hash value for a type symtab from a key

Accepts :

	keyP		Ptr to candidate key

Returns :

	<value>		16-bit non-normalized hash value

Notes :

	Handler routine for mml-symtab symbol table.
*/

UWORD
type_symtab_hash ARGLIST( (symP, keyP) )
    FARG( TYPE		*keyP )		/* Address of candidate key */
{


	int		hashval;	/* Hash value */


    hashval = keyP->ty_class;

    /* If aggregate, don't use size in hash value */
    if ( ( keyP->ty_flags & TF_AGGR ) == 0 )
    	hashval *= keyP->ty_size;

    /* If any further type reference, use that reference address
       as part of the hash. */
    if ( keyP->ty_nextP != NULL )
        hashval *= ((int)keyP->ty_nextP) & 0x7ff;

    return( hashval );
}
/*

*//* type_symtab_keyassoc(symP, keyP)

	Symbol table key association routine for type

Accepts :

	symP		Ptr to symbol table entry
	keyP		Ptr to key to associate with it

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Handler routine for mml-symtab symbol table.
*/

BOOL
type_symtab_keyassoc ARGLIST( (symP, keyP) )
   NFARG( TYPSYM	*symP )		/* Ptr to symbol table entry */
    FARG( TYPE		*keyP )		/* Address of candidate key */
{
    memcpy( &symP->ts_type, keyP, sizeof(TYPE) );

    return ( TRUE );
}
/*

*//* type_symtab_keydis(symP)

	Symbol table key disassociation for type

Accepts :

	symP		Ptr to symbol table entry

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Handler routine for mml-symtab symbol table.
*/

BOOL
type_symtab_keydis ARGLIST( (symP) )
    FARG( TYPSYM	*symP )		/* Ptr to symbol table entry */
{
    /* Nothing to do */
    return ( TRUE );
}
/*

*//* type_uint( size )

	Find an unsigned integer type

Accepts :

	size		Desired size

Returns :

	<value>		Address of type

Notes :


*/

TYPE *
type_uint	ARGLIST( (size) )
    FARG( int		size )		/* The desired size */
{
	TYPE		*typeP;		/* Ptr to type */
	TYPSYM		*tpsymP;	/* Ptr to type pool entry */
	TYPE		newtype;	/* Type being referenced */

    /* Construct the type */
    newtype.ty_class = TC_UINT;
    newtype.ty_flags = TF_BUILTIN;
    newtype.ty_size = size;
    newtype.ty_nextP = NULL;

    /* Lookup resulting type */
    tpsymP = (TYPSYM *)symtab_enter( STtpool, &newtype, sizeof(TYPSYM) );

    if ( tpsymP != NULL )
    	typeP = &tpsymP->ts_type;
    else
    	typeP = NULL;

    return( typeP );
}
/*

*//* typeclass_po(tc)

	Suggest a PO_xxx value for a type class

Accepts :

	tc		The TC_xxx typeclass

Returns :

	<value>		PO_xxx 

Notes :

	Returns the PO_xxx pointer offset value for an lvalue which
	would reference a variable's symtab entry of this type.

	Does not mean that all values of this type would be referenced
	using the returned PO_xxx type

*/

int
typeclass_po ARGLIST( (tc) )
    FARG( TYPECLASS	tc )		/* The type class */
{
    switch ( tc ) {
	case TC_VOID:			/* Simple types */
	case TC_UINT:
	case TC_INT:
	case TC_FLOAT:
	case TC_ENUM:
	case TC_PTR:
	    return ( PO_SIMPLE );

	case TC_UNION:
	case TC_STRUCT:
	case TC_DIM:
	    return ( 0 );

	case TC_FUNCT:
	    return ( PO_PRIVATE );

	case TC_STRING:
	    return ( PO_REFSTR );

	default:
	    warning( "typeclass_po: unknown type class %d", tc );
	    return ( 0 );
    }
}
/*

*//* val_access( valP )

	Access a value

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	Gains access to a VALUE structure.  Doesn't actually
	obtain the information, but makes sure any references to
	this value are noted.  i.e., this is the complement
	to val_release().

*/

void
val_access ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	TYPECLASS	tc;		/* Type class */

    if ( valP->v_typeP == NULL )
	return;				/* No type attached. */

    /* If this is an lvalue, the access is against the lvalue
       object (and it's strictly a pointer reference to that
       object.
    */
    if ( ( valP->v_flags & VF_LVALUE ) != 0 )
	tc = TC_PTR;
    else
	tc = valP->v_typeP->ty_class;

    /* Process by type. */
    switch( tc ) {
	case TC_FLOAT:			/* Simple values */
	case TC_ENUM:
	case TC_INT:
	case TC_UINT:
	case TC_VOID:
	    break;


	case TC_DIM:			/* Values with complex storage */
	case TC_STRUCT:
	case TC_UNION:
	    /* If there's a DVS reference, obtain our access */
	    if ( ( valP->v_data.d_ptr.p_byteX >= 0 ) &&
		 ( valP->v_data.d_ptr.p_ptr.p_dvsP != NULL ) )
		(void)dvs_access( valP->v_data.d_ptr.p_ptr.p_dvsP );

	    break;


	case TC_PTR:
	    val_access_ptr( valP );
	    break;


	case TC_STRING:
	    /* Access any refstr */
	    if ( ( valP->v_data.d_ptr.p_byteX == PO_REFSTR ) &&
		 ( valP->v_data.d_ptr.p_ptr.p_genP != NULL ) )
		refstr_access( (REFSTR *)valP->v_data.d_ptr.p_ptr.p_genP );
	    break;


	default:
	    warning( "val_finish: unknown type %d", tc );
	    break;
    }
}
/*

*//* val_access_ptr( valP )

	Access a pointer value

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	val_access for PTR type

*/

void
val_access_ptr ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	VALPTR		*vptrP;		/* The pointer thing */

    /* Access according to pointer type */
    vptrP = &valP->v_data.d_ptr;
    if ( vptrP->p_ptr.p_genP == NULL )
	;				/* do nothing with a null ptr */
    else if ( vptrP->p_byteX >= 0 )
	(void)dvs_access( vptrP->p_ptr.p_dvsP );
    else {
	switch ( vptrP->p_byteX ) {
	    case  PO_UNDEF:
	    case  PO_SIMPLE:
	    case  PO_PRIVATE:
		break;

	    case  PO_REFSTR:
		/* A REFSTR isn't a pointer, this shouldn't happen */
		warning( "val_access_ptr: type says pointer, data says string." );
		break;

	    case  PO_FSV:
		fsv_access( (FSV *)vptrP->p_ptr.p_genP );
		break;

	    default:
		warning( "val_access_ptr: unknown PO type %d",
				vptrP->p_byteX );
		break;
	}
    }
}
/*

*//* val_assign(rvalP, svalP)

	Assign from one value into another

Accepts :

	rvalP		Result value
	svalP		Source value

Returns :

	<value>		SCxx status code

Notes :

	rvalP is a VALUE that represents the lvalue of the
	storage object.  i.e. rvalP is not itself the destination;
	it is an lvalue container (identifying the destination).

*/

STSCODE
val_assign ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
	STSCODE		status;
	UBYTE		*dataP;		/* Data pointer */

    /* Result must be assignable */
    if ( ( rvalP->v_flags & VF_LVALUE ) == 0 ) {
	warning( "lvalue required." );
	return ( SCNOTDONE );
    }

    /* Source value must be forced into the destination type */
    status = val_coerce( svalP, rvalP->v_typeP, FALSE );

    if ( status == SCOK ) {
	/* Do the assignment */
	switch( rvalP->v_typeP->ty_class ) {
	    case TC_ENUM:
	    case TC_INT:
	    case TC_UINT:
	    case TC_FLOAT:
		/* Resolve the lvalue reference */
		dataP = valptr_resolve( &rvalP->v_data.d_ptr );
		if ( dataP == NULL ) {
		    warning( "Illegal reference." );
		    status = SCNOTDONE;
		}
		else {
		    /* Copy the data */
		    memcpy( dataP, 
			          &svalP->v_data,
			          rvalP->v_typeP->ty_size );
		}
		break;

	    case TC_PTR:
		status = val_assign_ptr( rvalP, svalP );
		break;

	    case TC_STRING:
		status = val_assign_string( rvalP, svalP );
		break;

	    default:
		warning( "Illegal assignment." );
		status = SCNOTDONE;
		break;
	}
    }

    return( status );
}
/*

*//* val_assign_ptr(rvalP, svalP)

	val_assign for ptr type

Accepts :

	rvalP		Result value
	svalP		Source value

Returns :

	<value>		SCxx status code

Notes :

	Handles val_assign() function for TC_PTR type.

*/

STSCODE
val_assign_ptr ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
	STSCODE		status;
	UBYTE		*dataP;		/* Data pointer */
	VALUE		ptrval;		/* The resolved pointer value */

    /* Resolve the lvalue reference */
    dataP = valptr_resolve( &rvalP->v_data.d_ptr );
    if ( dataP == NULL ) {
	warning( "Illegal reference to pointer." );
	status = SCNOTDONE;
    }
    else {
	/* Since we are changing the pointer value, we will have to
	   gain an access to the new pointer value and release
	   an access to the old one.  You might think that we'd have
	   to be careful to do the release after the gain (in the
	   case where they are the same, so we don't accidentally
	   free up the pointed-to thing).  However, the source value
	   must already have an access on it, so it shouldn't be
	   possible for us to completely release the source value here.
	*/
	ptrval = *rvalP;		/* Need to resolve this */

	/* Get the actual pointer value (the rvalue).  This will
	   result in releasing the lvalue, but that should be a no-op.
	*/
	val_rvalue(&ptrval);
	val_release(&ptrval);		/* val_rvalue added one access */
	val_release(&ptrval);		/* plus the original lvalue's hold. */

	/* Copy the data */
	memcpy( dataP, &svalP->v_data, rvalP->v_typeP->ty_size );

	/* Now we need to gain the access to the new pointer value, on
	   behalf of the lvalue object.
	*/
	ptrval = *rvalP;
	val_access( &ptrval );
	val_rvalue( &ptrval );

	status = SCOK;
    }

    return ( status );
}
/*

*//* val_assign_string(rvalP, svalP)

	val_assign for string type

Accepts :

	rvalP		Result value
	svalP		Source value

Returns :

	<value>		SCxx status code

Notes :

	Handles val_assign() function for TC_STRING type.

*/

STSCODE
val_assign_string ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
	STSCODE		status;
	VALPTR		*vptrP;
	REFSTR		*rsP;		/* REFSTR */
	ESTR		*esP;

    /* The destination is an lvalue that identifies the VALPTR
       containing the string pointer.  Resolve the valptr.
    */
    vptrP = (VALPTR *)valptr_resolve( &rvalP->v_data.d_ptr );
    if ( vptrP == NULL ) {
	warning( "Bad string destination." );
	status = SCNOTDONE;
    }
    else {
	if ( vptrP->p_ptr.p_genP == NULL ) {
	    vptrP->p_ptr.p_genP = refstr_new( "val_string_assign",
						20, 0 );
	    if ( vptrP->p_ptr.p_genP == NULL )
		return ( SCNOTDONE );
	    vptrP->p_byteX = PO_REFSTR;
	}

	/* Make sure we have a valid string pointer */
	if ( ( vptrP->p_byteX != PO_REFSTR ) ||
             ( ( rsP = (REFSTR *)vptrP->p_ptr.p_genP ) == NULL ) ) {
	    warning( "Invalid destination in string assignment." );
	    status = SCNOTDONE;
	}
	else if ( ( esP = val_estr_tmp( svalP ) ) == NULL )
	    status = SCERR;
	else if ( ! refstr_replacee( rsP, 0, refstr_len( rsP ), esP ) ) {
	    warning( "val_assign_string: unknown failure." );
	    status = SCERR;
	}
	else {
	    rsP->rs_bX = -1;	    /* Set the access index to "unintialized" */
	    status = SCOK;
	}
    }

    return( status );
}
/*

*//* val_attrprop( rvalP, val1P, val2P )

	Propagate type attributes

Accepts :

	rvalP		Result value
	val1P		First value
	val2P		Second value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Makes ancillary type attributes of result compatible with
	those of two components.

*/

void
val_attrprop ARGLIST( (rvalP, val1P, val2P) )
   NFARG( VALUE		*rvalP )	/* Result value */
   NFARG( VALUE		*val1P )	/* First component value */
    FARG( VALUE		*val2P )	/* Second component value */
{
	UWORD		rflags;		/* Result's flags */
	UWORD		flags1;		/* First's flags */
	UWORD		flags2;		/* Second's flags */

    /* Get the flags */
    rflags = rvalP->v_flags;
    flags1 = val1P->v_flags;
    flags2 = val2P->v_flags;

    /* Determine constant/noconstant */
    if ( ( ( flags1 & VF_CONST ) != 0 ) && ( ( flags2 & VF_CONST ) != 0 ) )
    	rflags |= VF_CONST;
    else
    	rflags &= ~VF_CONST;

    /* Put the flags in the result */
    rvalP->v_flags = rflags;
}
/*

*//* val_coerce( valP, typeP, forceF )

	Coerce a value into a particular type

Accepts :

	valP		Ptr to value to coerce
	typeP		Ptr to the type for the value to become
	forceF		Whether to force an unnatural conversion

Returns :

	<value>		SCxx status code

Notes :

	forceF applies in only a few cases.  Most possible coercions
	will be done implicitly but some have to be explicit.
	forceF presses the explicit nature.
	

*/

STSCODE
val_coerce ARGLIST( (valP, typeP, forceF) )
   NFARG( VALUE		*valP )		/* Value to be coerced */
   NFARG( TYPE		*typeP )	/* Type to coerce to */
    FARG( BOOL		forceF )	/* Force unnatural implicit conversion */
{
	STSCODE		status;
	int		curtclass;	/* Current type class */
	int		newtclass;	/* New type class */
	long		dint;
	double		dval;

    /* Get current and new types */
    curtclass = valP->v_typeP->ty_class;
    newtclass = typeP->ty_class;

    /* Make sure we have the rvalue */
    val_rvalue( valP );

    /* Quick check to see if we're already the right type */
    if ( valP->v_typeP == typeP )
	return ( SCOK );

    /* Process according to desired type */
    status = SCOK;			/* Assume success */
    switch( newtclass ) {
	case	TC_ENUM:		/* enum <-> int conversion */
#ifdef	MFL_STRICT_ENUMS
	    if ( !forceF ) {
		status = SCNOTDONE;	/* Unnatural conversion not forced */
		break;
	    }
	    /* fallthrough */
#endif	/* MFL_STRICT_ENUMS */

        case	TC_INT:			/* Integer value */
	case	TC_UINT:
	    switch( curtclass ) {

		case TC_ENUM:		/* enum <-> int conversion */
#ifndef	MFL_STRICT_ENUMS
		    if ( !forceF ) {
			/* Unnatural coercion not forced */
			status = SCNOTDONE;
			break;
		    }
		    /* fallthrough */
#endif	/* MFL_STRICT_ENUMS */

	    	case TC_INT:
		case TC_UINT:
		case TC_FLOAT:
		case TC_PTR:		/* maybe for comparison? */
		    /* Fetch int value */
		    dint = val_get_int( valP );
		    break;


		default:
		    /* If it's a string we can convert from, do so */
		    if ( type_is_string( valP->v_typeP ) )
TODO			/* Should do numeric validation? */
			dint = estr_to_l( val_estr_tmp( valP ) );
		    else
			status = SCNOTDONE;
		    break;
	    }

	    if ( status == SCOK ) {
		/* Store the new value */
		valP->v_typeP = typeP;
		val_put_int( valP, dint );
	    }
	    break;


        case	TC_FLOAT:		/* Floating point value */
	    switch( curtclass ) {
		case TC_ENUM:		/* Enum promoted to float */
#ifndef	MFL_STRICT_ENUMS
		    if ( !forceF ) {
			status = SCNOTDONE;
			break;
		    }
		    /* fallthrough... */
#endif	/* MFL_STRICT_ENUMS */

	    	case TC_INT:
		case TC_UINT:
		case TC_FLOAT:
	    	    dval = (double)val_get_float( valP );
		    break;
			    
		default:
		    /* If it's a string we can convert from, do that */
		    if ( type_is_string( valP->v_typeP ) )
TODO			/* Should validate the format? */
			dval = estr_to_f( val_estr_tmp( valP ) );
		    else
			status = SCNOTDONE;
		    break;
	    }

	    if ( status == SCOK ) {
		/* Set new type and store the value */
		valP->v_typeP = typeP;
		val_put_float( valP, dval );
	    }
	    break;


	case	TC_STRUCT:
	case	TC_UNION:

	    /* I suppose this is possible, for whatever silly reason.
	       but vet the source types.
	    */
	    switch( curtclass ) {
		case TC_STRUCT:		/* Allow these */
		case TC_UNION:
		    break;

		default:		/* Don't allow anything else */
		    status = SCNOTDONE;
		    break;
	    }

	    /* If OK, do it. */
	    if ( status == SCOK ) {
	    	/* Make it the correct type */
	    	valP->v_typeP = typeP;
		val_rvalue( valP );
	    }
	    break;


	case	TC_DIM:			/* A base address */
	case	TC_PTR:			/*  or a pointer */
	    status = val_coerce_ptr( valP, typeP );
	    break;


	case	TC_STRING:
	    switch( curtclass ) {
		case TC_STRING:		/* Already a string? */
		    break;		/*  yay. */

		default:		/* not a string */
		    /* Use val_string_val to make a string. */
		    status = val_string_val( valP, valP, FALSE );
		    break;
	    }
	    break;


	default:
	    status = SCNOTDONE;
	    break;
    }

    if ( status != SCOK )
        warning( "Invalid type conversion." );

    return( status );
}
/*

*//* val_coerce_ptr( valP, typeP )

	Coerce a value into a particular pointer type

Accepts :

	valP		Ptr to value to coerce
	typeP		Ptr to the type for the value to become

Returns :

	<value>		SCxx status code

Notes :

	Subservient to val_coerce when the new type is a pointer type.

*/

STSCODE
val_coerce_ptr ARGLIST( (valP, typeP) )
   NFARG( VALUE		*valP )		/* Value to be coerced */
    FARG( TYPE		*typeP )	/* Type to coerce to */
{
	STSCODE		status;
	FSV		*fsvP;
	VALPTR		*vptrP;

    /* Depends on the type of thing being pointed to. */
    switch ( ((TYPE *)typeP->ty_nextP)->ty_class ) {
	case	TC_VOID:
	case	TC_UINT:
	case	TC_INT:
	case	TC_FLOAT:
	case	TC_ENUM:
	    if ( ( valP->v_typeP->ty_class == TC_PTR ) ||
		 ( valP->v_typeP->ty_class == TC_DIM ) ) {
		/* Already a pointer/array -- ok if compatible. */
		switch ( ((TYPE *)valP->v_typeP->ty_nextP)->ty_class ) {
		    case	TC_VOID:
		    case	TC_UINT:
		    case	TC_INT:
		    case	TC_FLOAT:
		    case	TC_ENUM:
			status = SCOK;
			break;

		    default:
			status = SCNOTDONE;
			break;
		}
	    }
	    else {
		/* Non-pointer, can't convert. */
		status = SCNOTDONE;
	    }
	    break;


	case	TC_UNION:
	case	TC_STRUCT:
	    /* OK if a compatible pointer type */
	    if ( ( valP->v_typeP->ty_class == TC_PTR ) ||
		 ( valP->v_typeP->ty_class == TC_DIM ) ) {
		switch ( ((TYPE *)valP->v_typeP->ty_nextP)->ty_class ) {
		    case	TC_UNION:
		    case	TC_STRUCT:
			status = SCOK;
			break;

		    default:
			status = SCNOTDONE;
			break;
		}
	    }
	    else {
		/* Can't convert. */
		status = SCNOTDONE;
	    }
	    break;


	case	TC_DIM:
	case	TC_PTR:
TODO	    /* Should probably recurse to resolve this.  But do we
		really want to support that ?
	    */
	    warning( "val_coerce_ptr: please don't do that." );
	    status = SCNOTDONE;
	    break;


	case	TC_FUNCT:		/* no way */
	    status = SCNOTDONE;
	    break;


	case	TC_STRING:
	    /* If source is a pointer, see if we can make a string */
	    if ( ( valP->v_typeP->ty_class == TC_PTR ) ||
		 ( valP->v_typeP->ty_class == TC_DIM ) ) {
		status = val_string_val( valP, valP, TRUE );
	    }
	    else
		status = SCOK;

	    if ( ( status == SCOK ) &&
		 ( valP->v_typeP->ty_class != TC_STRING ) ) {
		status = SCNOTDONE;
	    }
	    if ( status == SCOK ) {
		/* We have a string; now we need to turn it into a
		   free-standing valptr so we can point to it.
		*/
		fsvP = fsv_new();
		if ( fsvP == NULL ) {
		    warning( "val_coerce_ptr: fsv_new() failed." );
		    status = SCERR;
		}
		else {
		    vptrP = (VALPTR *)&fsvP->fv_data;
		    vptrP->p_byteX = PO_REFSTR;
		    vptrP->p_ptr.p_genP = valP->v_data.d_ptr.p_ptr.p_genP;

		    fsvP->fv_ptrF = TRUE;	/* HACK: see fsv_release() */

		    valP->v_flags = VF_RVALUE;
		    valP->v_data.d_ptr.p_ptr.p_genP = fsvP;
		    valP->v_data.d_ptr.p_byteX = PO_FSV;
		}
	    }
	    break;


	default:
	    warning( "Invalid destination pointer type %d", 
		    ((TYPE *)typeP->ty_nextP)->ty_class );
	    status = SCERR;
	    break;
    }

    /* If failed, see if it's because of an attempt to convert
       constant 0 to a pointer.  That's legit and I'd rather
       check here than check every case before the fact.
    */
    if ( ( status == SCNOTDONE ) &&
         ( ( valP->v_typeP->ty_class == TC_INT ) ||
	   ( valP->v_typeP->ty_class == TC_UINT ) ) &&
	 ( val_get_int( valP ) == 0 ) ) {
	valP->v_flags = VF_RVALUE;
	valP->v_data.d_ptr.p_byteX = PO_UNDEF;
	valP->v_data.d_ptr.p_ptr.p_genP = 0;
	status = SCOK;
    }

    if ( status == SCOK )
	valP->v_typeP = typeP;		/* OK to go */

    return ( status );
}
/*

*//* val_common_type(val1P, val2P)

	Find a type common to two values

Accepts :

	val1P		First value
	val2P		Second value

Returns :

	<value>		A common type

Notes :


*/

TYPE *
val_common_type ARGLIST( (val1P, val2P) )
   NFARG( VALUE		*val1P )	/* Result value */
    FARG( VALUE		*val2P )	/* Source value */
{
	TYPE		*type1P;	/* First type */
	TYPE		*type2P;	/* Second type */
	TYPE		*rtypeP;	/* Result type */

    type1P = val1P->v_typeP;
    type2P = val2P->v_typeP;

    /* Brute force.  An optimization would be to precompile all of
       these combinations and just look up the combo at runtime.
       That exercise left for future busy-work.
    */
    switch( type1P->ty_class ) {
        case TC_INT:
	    switch( type2P->ty_class ) {
		case TC_UINT:		/* int/uint: recurse to uint/int */
		    return( val_common_type( val2P, val1P ) );

	    	case TC_INT:		/* int/int: use larger size */
		    rtypeP = ( type1P->ty_size >= type2P->ty_size ) ?
					type1P : type2P;
		    break;

		case TC_FLOAT:		/* int/float: promote to float */
		    rtypeP = type2P;
		    break;

		case TC_PTR :		/* int/ptr: use ptr */
		    rtypeP = type2P;
		    break;

		default:		/* int/unhandled: no result */
		    rtypeP = NULL;
		    break;
	    }
	    break;


        case TC_UINT:
	    switch( type2P->ty_class ) {
	    	case TC_UINT:		/* uint/uint: use larger size */
		    rtypeP = ( type1P->ty_size >= type2P->ty_size ) ?
					type1P : type2P;
		    break;

		case TC_INT:
		    /* Combination of int and unsigned int must promote
		       to unsigned int.  This is a degenerate case (it
		       should be arrested) in that the resulting type is
		       not necessily either of the original types; it is
		       unsigned type of the larger storage size. */
		    rtypeP = type_uint( 
		    		type1P->ty_size > type2P->ty_size ?
					type1P->ty_size : type2P->ty_size );
		    break;
		    

		case TC_FLOAT:		/* uint/float: promote to float */
		    rtypeP = type2P;
		    break;

		case TC_PTR:		/* uint/ptr: use ptr */
		    rtypeP = type2P;
		    break;

		default:		/* uint/unhandled: no result */
		    rtypeP = NULL;
		    break;
	    }
	    break;


	case TC_FLOAT:
	    switch( type2P->ty_class ) {
	    	case TC_INT:		/* float/int: use float */
		case TC_UINT:		/* float/uint: use float */
		    rtypeP = type1P;
		    break;


		case TC_FLOAT:		/* float/float: use larger size */
		    rtypeP = ( type1P->ty_size >= type2P->ty_size ) ?
					type1P : type2P;
		    break;

		default:		/* float/unhandled: no result */
		    rtypeP = NULL;
		    break;
	    }
	    break;


	case TC_PTR:
	    switch( type2P->ty_class ) {
	    	case TC_INT:		/* ptr/int: use ptr */
		case TC_UINT:		/* ptr/uint: use ptr */
		case TC_PTR:		/* ptr/ptr: use first ptr type */
		    rtypeP = type1P;
		    break;

		default:		/* ptr/unhandled: no result */
		    rtypeP = NULL;
		    break;
	    }
	    break;


	case TC_ENUM:
	    switch( type2P->ty_class ) {
	    	case TC_ENUM:		/* enum/enum: only ok if same enum */
		    if ( type1P == type2P )
			rtypeP = type1P;
		    else
			rtypeP = NULL;
		    break;

		default:		/* enum/unhandled: no result */
		    rtypeP = NULL;
		    break;
	    }
	    break;


	default:			/* Unhandled types */
	    rtypeP = NULL;
	    break;
    }

    return( rtypeP );
}
/*

*//* val_deref(rvalP, svalP)

	Dereference a pointer 

Accepts :

	rvalP		Result value
	svalP		Source value (pointer)

Returns :

	<nothing>

Notes :


*/

STSCODE
val_deref ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
	STSCODE		status;
	TYPE		*typeP;		/* Type... */

    /* Get type and check valid operation */
    typeP = svalP->v_typeP;
    if ( ( typeP->ty_class != TC_PTR ) && ( typeP->ty_class != TC_DIM ) ) {
    	warning( "Indirection through non-pointer." );
	status = SCNOTDONE;
    }
    else {
        /* Construct destination */
	*rvalP = *svalP;

	/* We'll need access to the eventual rvalue.  Do it now;
	   the result of val_rvalue will give us the correct
	   access.
	*/
	val_access( rvalP );

	/* Need the rvalue: the actual pointer value. */
	val_rvalue( rvalP );

	/* Re-type into reference type */
	rvalP->v_typeP = (TYPE *)typeP->ty_nextP;

	/* Lvalue is now valid */
	rvalP->v_flags |= VF_LVALUE;

	/* rvalue is not yet present */
	rvalP->v_flags &= ~VF_RVALUE;

	status = SCOK;
    }

    return( status );
}
/*

*//* val_estr_eval( treeP, esPP, valP )

	Evaluate a node or tree, return val_estr ESTR struct

Accepts :

	treeP		Ptr to the node or tree to eval
	esPP		Ptr to ESTR to return result into
	valP		Ptr to any value object to use

Returns :

	<value>		STxxx status code

Notes :

	This routine gives access to a string result via an ESTR
	 encapsulated string object.  (See val_finish() for what
	 we mean by this.)

	The caller may pass the address of a VALUE object to use
	 in the evaluation.  If so, the returned ESTR may be used
	 for as long as the VALUE object exists.

	Otherwise (if caller passes valP as NULL), a temporary value will
	 be used.  In this case, caller should assume that the ESTR
	 returned is extremely temporary and must release it as soon as
	 possible, without recursion or other activity that might allocate
	 very many other temporary values.

	Caller must release the ESTR via val_estr_finish()

*/

STSCODE
val_estr_eval ARGLIST( (treeP, esPP, valP) )
   NFARG( PTNODE	*treeP )	/* Node or tree to evaluate */
   NFARG( ESTR		**esPP )	/* Where to put the ESTR result */
    FARG( VALUE		*valP )		/* Value to use, if not NULL */
{
	STSCODE		status;
	ESTR		*esP;		/* ESTR object */
	REFSTR		*rsP;

    /* If it's just a single node, see if we can reference the string
       without any costly evaluations.
    */
    if ( ( treeP->pt_leftP == NULL ) &&
	 ( treeP->pt_rightP == NULL ) ) {
	switch ( treeP->pt_type ) {
	    case  PN_LITSTR:		/* Literal string */
		*esPP = estr_str( (UBYTE *)treeP->pt_info.od_P );
		return ( ( *esPP == NULL ) ? SCERR : SCOK );

	    default:			/* Anything else */
		break;
	}
    }

    /* Nothing simple, must eval. */
    if ( valP == NULL ) {
	if ( ( valP = tmpval() ) == NULL )
	    return ( SCERR );
    }

    status = eval_tree( treeP, valP );
    if ( status == SCOK )
	val_coerce( valP, Tnstring->tn_typeP, FALSE );
    if ( status == SCOK ) {
	if ( ( rsP = val_get_string( valP ) ) == NULL ) {
	    status = SCERR;
	}
	else {
	    /* Get an ESTR and attach a value pointer to it */
	    esP = estr_refstr( rsP );
	    if ( esP == NULL ) {
		refstr_free( rsP );
		status = SCERR;
	    }
	    else
		esP->es_xP = valP;
	}
    }

    if ( status != SCOK ) {
	esP = NULL;
	val_finish( valP );
    }

    *esPP = esP;

    return ( status );
}
/*

*//* val_estr_finish( esP )

	Finish access to val_estr ESTR

Accepts :

	esP		Ptr to val_estr previously obtained

Returns :

	<nothing>

Notes :

	A "val_estr" is simply an ESTR that was obtained via
	some val_estr_xxx function.  There may be a value or other
	information attached that needs to be dealt with before the
	ESTR is released.  This function does the necessary releases.

*/

void
val_estr_finish ARGLIST( (esP) )
    FARG( ESTR		*esP )		/* The ESTR to release */
{
    /* Finish any attached info */
    if ( esP->es_xP != NULL ) {
	/* The attached info is a VALUE obtained via tmpval().  We must
	   finish the value, but we don't have to free it.
	*/
	val_finish( esP->es_xP );
    }

    /* OK to finish the ESTR now. */
    estr_finish( esP );
}
/*

*//* val_estr_tmp( valP )

	Access a string buffer representing a value

Accepts :

	valP		Ptr to value

Returns :

	<value>		Pointer to a ESTR encaspulating the string
			NULL if error (warning given)

Notes :

	Returns a pointer to a temporary ESTR that represents a
	value's contents.   This uses a temporary ESTR (an "estr_tmp")
	that encapsulates a string that may itself be temporary (i.e.
	a "tmpstr").

	Caller must assume that the returned information is transient and
	must be used before any other call to val_estr_tmp.  Refer to
	"estr_tmp"s and "tmpstrs" for more.

*/

ESTR *
val_estr_tmp ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	int		len;
	int		intval;
	char		*strP;
	TYPE		*ntypeP;	/* Next type */
	ESTR		*esP;		/* Pointer to temp ESTR */
	REFSTR		*rsP;		/* REFSTR */
	TYPEAG		*agP;		/* Ptr to aggregate */

    /* Must have rvalue */
    val_rvalue( valP );

    /* Get string depending on type. */
    switch ( valP->v_typeP->ty_class ) {
	case TC_INT:			/* String from integer value */
	case TC_UINT:
	    strP = (char *)tmpstr( 20 ); /* Grab a temporary string */
	    len = snprintf( strP, 20, "%d", val_get_int( valP ) );
	    break;

	case TC_FLOAT:			/* String from floating point number */
	    strP = (char *)tmpstr( 20 );
	    len = snprintf( strP, 20, "%g", val_get_float( valP ) );
	    break;


	case TC_ENUM:			/* String from enum */
	    /* Give enum value name if in range, value digit string if not */
	    agP = (TYPEAG *)valP->v_typeP->ty_nextP;
	    intval = val_get_int( valP );
	    if ( ( intval < 0 ) || ( intval >= agP->ag_memC ) ) {
		strP = (char *)tmpstr(20);
		len = snprintf( strP, 20, "%d", intval );
	    }
	    else {
		strP = agP->ag_memP[intval].am_nameP;
		len = strlen( strP );
	    }
	    break;


	case TC_PTR:			/* Some pointer */
	    /* See what it points to */
	    ntypeP = (TYPE *)valP->v_typeP->ty_nextP;
	    if ( ntypeP == NULL ) {
		strP = NULL;
	    }
	    else if ( ( ntypeP->ty_class == TC_INT ) &&
		 ( ntypeP->ty_size == 1 ) ) {
		/* It's a pointer to char.. we'll take it. */
		strP = valptr_resolve( &valP->v_data.d_ptr );
		len = strlen( strP );
	    }
	    else if ( ntypeP->ty_class == TC_STRING ) {
		/* Pointer to string: while this would be easy to resolve,
		   it's equally easy (and correct) for the MFL coder to
		   dereference the pointer.  Doing it implicitly would just
		   create extra fuzziness.  This case is called out
		   specially mainly so I can write this comment...
		*/
		warning( "attempt to use string pointer as string." );
		strP = NULL;
	    }
	    else {
		warning( "attempt to use pointer as string." );
		strP = NULL;
	    }
	    break;


	case TC_STRING:			/* A string */
	    /* Make sure we have a valid string pointer */
	    if ( ( rsP = val_get_string( valP ) ) == NULL ) {
		    strP = NULL;
		}
	    else {
		strP = (char *)refstr_bstr( rsP )->bs_bP + refstr_startx( rsP );
		len = refstr_len( rsP );
	    }
	    break;


	default:
	    warning( "val_estr_tmp: can't yield string from type %d",
				valP->v_typeP->ty_class );
	    strP = NULL;
	    break;
    }

    /* Construct the ESTR */
    if ( strP != NULL )
	esP = estr_tmp_nstr( (UBYTE *)strP, len );
    else
	esP = NULL;

    return ( esP );
}
/*

*//* val_finish( valP )

	Finish accessing a VALUE (release and de-initialize it)

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	Finishes access to a value storage by releasing it and
	wiping it out.  i.e., a combination of val_release()
	and val_nill().

*/

void
val_finish ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
    val_release( valP );
    val_nill( valP );
}
/*

*//* val_get_float( valP )

	Get a floating point number from a value

Accepts :

	valP		Ptr to value to get the value from 

Returns :

	<value>		The value

Notes :

	Will attempt to get a value in reasonable situations.

*/

double
val_get_float ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value pointer */
{
	double		dval;

    switch( valP->v_typeP->ty_class ) {
        case TC_INT:
	case TC_UINT:
	case TC_ENUM:
	    dval = (double)val_get_int( valP );
	    break;


	case TC_FLOAT:
	    if ( valP->v_typeP->ty_size == 4 )
	    	dval = (double)valP->v_data.d_float;
	    else
	    	dval = valP->v_data.d_double;
	    break;


	default:
	    warning( "Internal error in val_get_float." );
	    break;
    }

    return( dval );
}
/*

*//* val_get_int( valP )

	Get an int from a value

Accepts :

	valP		Ptr to value to get int from 

Returns :

	<value>		The value

Notes :

	Will attempt to get an int value in reasonable situations.

*/

long
val_get_int ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value pointer */
{
	long		dint;

    switch( valP->v_typeP->ty_class ) {
        case TC_INT:
	    switch( valP->v_typeP->ty_size ) {
	    	case 1:
		    dint = valP->v_data.d_byte;
		    break;

		case 2:
		    dint = valP->v_data.d_word;
		    break;

		case 4:
		    dint = valP->v_data.d_long;
		    break;

	    }
	    break;


	case TC_UINT:
	case TC_ENUM:
	    switch( valP->v_typeP->ty_size ) {
	    	case 1:
		    dint = (unsigned char)valP->v_data.d_byte;
		    break;

		case 2:
		    dint = (UWORD)valP->v_data.d_word;
		    break;

		case 4:
		    dint = valP->v_data.d_long;
		    break;

	    }
	    break;

	case TC_FLOAT:
	    dint = (long)val_get_float( valP );
	    break;


	case TC_PTR:		/* Fetching ptr value as int is
				   generally useful when comparing
				   a pointer -- to 0, say.
				*/
	    if ( valP->v_data.d_ptr.p_byteX == PO_UNDEF )
		dint = 0;
	    else
		dint = (long)valP->v_data.d_ptr.p_ptr.p_genP;
	    break;


	default:
	    warning( "Internal error in val_get_int." );
	    break;
    }

    return( dint );
}
/*

*//* val_get_string( valP )

	Get an string reference from a value

Accepts :

	valP		Ptr to value to get string from 

Returns :

	<value>		Pointer to the REFSTR
			NULL if not a string.

Notes :

	The passed value must be a string value with a REFSTR
	attached.  This function merely gets the pointer to the
	REFSTR -- NO attempt is made to coerce to a string.

	If the value is not a REFSTR value, NULL is returned.
	No warning message is issued.

*/

REFSTR *
val_get_string ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value pointer */
{
	VALPTR		*vptrP;
	REFSTR		*rsP;

    vptrP = &valP->v_data.d_ptr;
    if ( ( vptrP == NULL ) ||
	 ( vptrP->p_byteX != PO_REFSTR ) ||
	 ( ( rsP = (REFSTR *)vptrP->p_ptr.p_genP ) == NULL ) ) {
	return ( NULL );
    }

    return ( rsP );
}
/*

*//* val_init( valP, typeP )

	Initialize/instantiate a variable

Accepts :

	valP		Ptr to value
	typeP		Ptr to TYPE for new value

Returns :

	<value>		Status code

Notes :

	Instantiates a value with a new type and null value.
	No prior contents are preserved.

*/

STSCODE
val_init ARGLIST( (valP, typeP) )
   NFARG( VALUE		*valP )		/* The value */
    FARG( TYPE		*typeP )	/* New type */
{
	STSCODE		status;

    status = SCOK;			/* Assume success */

    valP->v_typeP = typeP;		/* Set the type */

    /* Init the common flags */
    valP->v_flags = VF_RVALUE;

    /* Instantiate according to type */
    switch( typeP->ty_class ) {
	case TC_INT:			/* Various simple scalar types */
	case TC_UINT:
	case TC_ENUM:
	case TC_FLOAT:
	    val_put_int( valP, 0L );	/* Set to zero */
	    break;

	case TC_PTR:			/* Pointer */
	    /* Must set the pointer to undef,
	       and clear the pointer for good measure.
	    */
	    valP->v_data.d_ptr.p_byteX = PO_UNDEF;
	    valP->v_data.d_ptr.p_ptr.p_genP = NULL;
	    break;


	case TC_DIM:
	case TC_UNION:
	case TC_STRUCT:
	    /* Allocate dynamic storage for this using DVS */
	    valP->v_data.d_ptr.p_ptr.p_dvsP = dvs_new( typeP->ty_size,
							 typeP->ty_size );

	    if ( valP->v_data.d_ptr.p_ptr.p_dvsP == NULL ) {
		valP->v_data.d_ptr.p_byteX = PO_UNDEF;
		warning( "val_init: dvs_new failed." );
		status = SCERR;
	    }
	    else {
		valP->v_data.d_ptr.p_byteX = 0;
	    }
	    break;


	case TC_STRING:
	    /* Allocate a new refstr */
	    valP->v_data.d_ptr.p_ptr.p_genP = refstr_new( "val_init",
							   20,
							   0 );

	    if ( valP->v_data.d_ptr.p_ptr.p_genP == NULL ) {
		valP->v_data.d_ptr.p_byteX = PO_UNDEF;
		warning( "val_init: refstr_new failed." );
		status = SCERR;
	    }
	    else {
		valP->v_data.d_ptr.p_byteX = PO_REFSTR;
	    }
	    break;


	case TC_FUNCT:			/* Function  */
	    /* Just clear the value (ptr to function info) */
	    valP->v_data.d_ptr.p_byteX = PO_UNDEF;
	    valP->v_data.d_ptr.p_ptr.p_genP = NULL;
	    break;


	default:			/* unknown */
	    warning( "val_init: unimplemented for type %d", typeP->ty_class );
	    status = SCERR;
	    break;
    }

    return ( status );
}
/*

*//* val_isint(valP, mustF)

	See if a value is an integer

Accepts :

	valP		Ptr to value
	mustF		If it must be an integer (vs any numeric value)

Returns :

	<value>		TRUE if it's an integer,
			FALSE if not.

Notes :

	If mustF is false, any number is acceptable.

*/

BOOL
val_isint ARGLIST( (valP, mustF) )
   NFARG( VALUE		*valP )		/* Value to check */
    FARG( BOOL		mustF )		/* If strictly integer required */
{
	BOOL		isintF;

    switch( valP->v_typeP->ty_class ) {
	case	TC_INT:
	case	TC_UINT:
	    isintF = TRUE;
	    break;


	case	TC_PTR:
	case	TC_ENUM:
	case	TC_FLOAT:
	    if ( mustF ) {
		warning( "integer value required." );
		isintF = FALSE;
	    }
	    else
	    	isintF = TRUE;
	    break;


	default:
	    if ( mustF )
	        warning( "integer value required." );
	    else
	        warning( "integer-like value required." );
	    isintF = FALSE;
	    break;
    }

    return( isintF );
}
/*

*//* val_lvalue(rvalP, svalP)

	Get the lvalue of a value

Accepts :

	rvalP		Result value
	svalP		Source value

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
val_lvalue ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
	STSCODE		status;		/* Return code */
	TYPSYM		*tpsymP;	/* Type pool symbol */

    if ( ( svalP->v_flags & VF_LVALUE ) == 0 ) {
        warning( "l-value required." );
	status = SCNOTDONE;
    }
    else {
	/* The new type is ptr-to the existing type */
	tpsymP = type_ptr( svalP->v_typeP );
	if ( tpsymP == NULL )
	    status = SCERR;
	else {
	    /* Create the new value */
	    memcpy( rvalP, svalP, sizeof( VALUE ) );
	    rvalP->v_typeP = &(tpsymP->ts_type);

	    /* There's no lvalue in the result */
	    rvalP->v_flags &= ~VF_LVALUE;

	    /* However, value's rvalue IS now present. */
	    rvalP->v_flags |= VF_RVALUE;

	    /* And we need the access. */
	    val_access( rvalP );

	    status = SCOK;
	}
    }

    return( status );
}
/*

*//* val_member(rvalP, svalP, mnameP)

	Reference a member of an aggregate

Accepts :

	rvalP		Result value
	svalP		Source value
	mnameP		Aggregate name

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
val_member ARGLIST( (rvalP, svalP, mnameP) )
   NFARG( VALUE		*rvalP )	/* Result value */
   NFARG( VALUE		*svalP )	/* Source value */
    FARG( char		*mnameP )	/* Member name */
{
	STSCODE		status;
	TYPE		*typeP;		/* Type of thing */
	AGGMEM		*agmemP;	/* Ptr to aggr member */

    /* Input value must be an aggregate */
    typeP = svalP->v_typeP;		/* Get type */
    if ( ( typeP->ty_flags & TF_AGGR ) == 0 ) {
    	warning( "Aggregate required for \"->\" or \".\"" );
	status = SCNOTDONE;
    }
    else if ( svalP->v_data.d_ptr.p_byteX < 0 ) {
	warning( "Aggregate not DVS." );
	status = SCNOTDONE;
    }
    else {
	agmemP = type_agg_member( typeP, mnameP );
	if ( agmemP == NULL ) {
	    /* Member not found. */
	    warning( "Invalid member reference: \"%s\"", mnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Member found -- construct resultant value. */
	    rvalP->v_flags = VF_LVALUE;
	    rvalP->v_typeP = agmemP->am_typeP;
	    rvalP->v_data.d_ptr = svalP->v_data.d_ptr;
	    rvalP->v_data.d_ptr.p_byteX += agmemP->am_offs;

	    val_access( rvalP );

	    status = SCOK;
	}
    }

    return( status );
}
/*

*//* val_nill( valP )

	Make a value unused (turn it to nill)

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	Makes a value unusable/unused.  Absolutely no thought is
	given to any prior contents.  If there were such
	prior contents, they should have been dealt with before
	calling this.

*/

void
val_nill ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
    /* We just clear the type reference.  A NULL type is sufficient
       indication that the value is not in use.
    */
    valP->v_typeP = NULL;
}
/*

*//* val_pmember(rvalP, svalP, mnameP)

	Reference a member of a pointer to an aggregate

Accepts :

	rvalP		Result value
	svalP		Source value
	mnameP		Aggregate name

Returns :

	<value>		SCxx status code

Notes :


*/

STSCODE
val_pmember ARGLIST( (rvalP, svalP, mnameP) )
   NFARG( VALUE		*rvalP )	/* Result value */
   NFARG( VALUE		*svalP )	/* Source value */
    FARG( char		*mnameP )	/* Member name */
{
	STSCODE		status;
	VALUE		avalue;		/* Value of aggregate */

    /* Dereference the pointer to get the aggregate */
    status = val_deref( &avalue, svalP );
    if ( status == SCOK )
	/* Perform membership reference on the structure. */
	status = val_member( rvalP, &avalue, mnameP );

    return( status );
}
/*

*//* val_print(fP, valP, level, tagP)

	Print a value

Accepts :

	fP		FILE descriptor to print to
	valP		Ptr to value to print
	level		Nesting level of recursion
	tagP		Any tag to print after the indent level

Returns :

	<value>		SCxx code

Notes :


*/

STSCODE
val_print ARGLIST( (fP, valP, level, tagP) )
   NFARG( FILE		*fP )		/* file to print to */
   NFARG( VALUE		*valP )		/* Value to print */
   NFARG( int		level )		/* Recursion level */
    FARG( char		*tagP )		/* Tag */
{
	int		i;
	long		intval;		/* Integer value */
	UBYTE		*dataP;		/* Ptr to any data value */
	TYPEAG		*agP;		/* Ptr to aggregate */
	REFSTR		*rsP;
	VALUE		val;		/* Scratch value */
	char		tag[10];	/* Tag string */

    /* Indent according to level */
    for ( i = 0; i < level; ++i )
	fprintf( fP, "  " );

    /* Print the tag */
    fprintf( fP, "%s ", tagP );

    /* If no value supplied, say omitted */
    if ( valP == NULL ) {
	fprintf( fP, ": <value omitted>\n" );
	return ( SCOK );
    }

    /* Before printing, the value must be an r-value */
    val_rvalue( valP );

    /* Print the type */
    type_print( fP, valP->v_typeP );
    fprintf( fP, ": " );

    if ( ( valP->v_flags & VF_CONST ) != 0 )
        fprintf( fP, "(const) " );

    switch( valP->v_typeP->ty_class ) {
	case  TC_INT:
	    /* Get the val and print it as if long */
	    intval = val_get_int( valP );
	    fprintf( fP, "%ld 0x%lx\n", intval, intval );
	    break;


	case  TC_UINT:			/* Unsigned integer */
	    /* Get the value, print as unsigned. */
	    intval = val_get_int( valP );
	    fprintf( fP, "%lu 0x%lx\n", intval, intval );
	    break;


	case  TC_ENUM:			/* Enumerated constant */
	    /* Get the value, print as unsigned. */
	    intval = val_get_int( valP );
	    fprintf( fP, "%lu 0x%lx", intval, intval );
	    
	    /* Decode the enumeration name */
	    agP = (TYPEAG *)valP->v_typeP->ty_nextP;
	    if ( ( intval < 0 ) || ( intval >= agP->ag_memC ) )
	    	fprintf( fP, " (illegal value)\n" );
	    else
	        fprintf( fP, " (%s)\n", agP->ag_memP[intval].am_nameP );
	    break;


	case  TC_FLOAT:
	    switch( valP->v_typeP->ty_size ) {
		case 4:
		    fprintf( fP, "%f\n", valP->v_data.d_float );
		    break;

		case 8:
		    fprintf( fP, "%f\n", valP->v_data.d_double );
		    break;

		default:
		    warning( "unknown size!" );
	    }
	    break;



	case TC_PTR:
	    /* Follow if not a NULL reference */
	    if ( ( valP->v_data.d_ptr.p_byteX != PO_UNDEF ) &&
		 ( valP->v_data.d_ptr.p_ptr.p_genP != NULL ) ) {
		val_deref( &val, valP );
		fprintf( fP, " 0x%08lx/%d\n",
			val.v_data.d_ptr.p_ptr.p_genP,
			val.v_data.d_ptr.p_byteX );
		val_print( fP, &val, level + 1, "" );
		val_release( &val );
	    }
	    else
		fprintf( fP, "<null>\n" );
	    break;

	case TC_STRING:
	    rsP = val_get_string( valP );
	    if ( rsP == NULL )
		fprintf( fP, "<null>" );
	    else {
		fprintf( fP, "\"" );
		fwrite( refstr_bstr( rsP )->bs_bP + refstr_startx( rsP ),
		          1, refstr_len( rsP ), fP );
		fprintf( fP, "\"" );
	    }
	    fprintf( fP, "\n" );
	    break;

	case TC_STRUCT:
	case TC_UNION:
	    /* This is big enough that we made it a seperate function. */
	    val_print_agg( fP, valP, level +1, "" );
	    break;

	case TC_DIM:
	    fprintf( fP, "\n" );
	    dataP = valptr_resolve( &valP->v_data.d_ptr );

	    for ( i = 0; i < valP->v_typeP->ty_dim; ++i ) {
		/* Setup value -- pretend it's non-dvs. */
		val.v_flags = VF_LVALUE;
		val.v_typeP = (TYPE *)valP->v_typeP->ty_nextP;
		val.v_data.d_ptr.p_ptr.p_genP =
			(void *)(dataP + i * val.v_typeP->ty_size);
		val.v_data.d_ptr.p_byteX = PO_SIMPLE;

		sprintf( &tag[0], "[%d]:", i );
		val_access( &val );
		val_print( fP, &val, level + 1, tag );
		val_release( &val );
	    }

	    break;

	default:
	    warning( "can't print this type (class=%d)!",
	    			valP->v_typeP->ty_class );
	    break;
    }

    return( SCOK );
}
/*

*//* val_print_agg(fP, valP, level, tagP)

	Print a value that's an aggregate

Accepts :

	fP		FILE descriptor to print to
	valP		Ptr to value to print
	level		Recursion level
	tagP		Any tag to print after indenting

Returns :

	<value>		SCxx code

Notes :

	This is just a helper for val_print.

*/

STSCODE
val_print_agg ARGLIST( (fP, valP, level, tagP) )
   NFARG( FILE		*fP )		/* file to print to */
   NFARG( VALUE		*valP )		/* Value to print */
   NFARG( int		level )		/* Recursion level */
    FARG( char		*tagP )		/* Any tag */
{
	int		memX;		/* Member index */
	TYPE		*typeP;		/* Type */
	TYPEAG		*agP;		/* Ptr to the aggregate */
	AGGMEM		*agmemP;	/* Ptr to each member */
	VALUE		memval;		/* Value for member */
	char		*dataP;		/* Ptr to base data */

    typeP = valP->v_typeP;

    /* Print address info */
    dataP = valptr_resolve( &valP->v_data.d_ptr );
    if ( dataP == NULL ) {
	fprintf( fP, "<badaggref>\n" );
	return ( SCNOTDONE );
    }
    
    fprintf( fP, "0x%08lx\n", dataP );

    /* Go through the members and print them. */
    agP = (TYPEAG *)typeP->ty_nextP;
    for( memX = 0, agmemP = &agP->ag_memP[0]; memX < agP->ag_memC;
    		++memX, ++agmemP ) {

	/* Construct value for this member.  Cheat and pretend it's
	   a non-DVS value regardless of what it is.
	*/
	memval.v_flags = VF_LVALUE;
	memval.v_typeP = agmemP->am_typeP;
	memval.v_data.d_ptr.p_ptr.p_genP = (void *)(dataP + agmemP->am_offs);
	memval.v_data.d_ptr.p_byteX = PO_SIMPLE;
	val_rvalue( &memval );

	/* Print via examining member type.  When dumping an aggregate,
	   we want to limit the automatic output, so we'll output
	   all simple types, and decide further about more complicated
	   ones.  User can always get the individual member output anyway.
	*/
	switch( agmemP->am_typeP->ty_class ) {
	    case TC_INT:
	    case TC_UINT:
	    case TC_FLOAT:
	    case TC_PTR:
	    case TC_ENUM:
	    case TC_STRING:
	    	val_print( fP, &memval, level, agmemP->am_nameP );
		break;

	    case TC_STRUCT:
	    case TC_UNION:
		/* I guess we'll just print these.  Keep them a separate
		   case, though, just for grins.
		*/
	    	val_print( fP, &memval, level, agmemP->am_nameP );
		break;

	    case TC_DIM:
		/* Inside of an aggregate, only print small arrays. */
		if ( memval.v_typeP->ty_dim <= 5 )
		    val_print( fP, &memval, level, agmemP->am_nameP );
		else
		    val_print( fP, NULL, level, agmemP->am_nameP );
		break;


	    default:
		/* Just let the val_print function catch this. */
	    	val_print( fP, &memval, level, agmemP->am_nameP );
		break;
	}
    }

    return( SCOK );
}
/*

*//* val_put_float( valP, dval )

	Store a floating point number into a value

Accepts :

	valP		Ptr to value to put int into
	dval		The floating point number to store

Returns :

	<nothing>

Notes :


*/

void
val_put_float ARGLIST( (valP, dval) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( double	dval )		/* The value to store */
{
    switch( valP->v_typeP->ty_class ) {
        case TC_INT:
	case TC_UINT:
	case TC_ENUM:
	    val_put_int( valP, (long)dval );
	    return;


	case TC_FLOAT:
	    if ( valP->v_typeP->ty_size == 4 )
		valP->v_data.d_float = (float)dval;
	    else
		valP->v_data.d_double = dval;
	    break;


	default:
	    warning( "Internal error in val_put_float." );
	    break;
    }

    valP->v_flags |= VF_RVALUE;
}
/*

*//* val_put_int( valP, dint )

	Store an int into a value

Accepts :

	valP		Ptr to value to put int into
	dint		The value

Returns :

	<nothing>

Notes :


*/

void
val_put_int ARGLIST( (valP, dint) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( long		dint )		/* The int value */
{
    switch( valP->v_typeP->ty_class ) {
        case TC_INT:
	case TC_UINT:
	case TC_ENUM:
	    switch( valP->v_typeP->ty_size ) {
	    	case 1:
		    valP->v_data.d_byte = dint;
		    break;

		case 2:
		    valP->v_data.d_word = dint;
		    break;

		case 4:
		    valP->v_data.d_long = dint;
		    break;

	    }
	    break;


	case TC_FLOAT:
	    val_put_float( valP, (double)dint );
	    return;


	default:
	    warning( "Internal error in val_put_int." );
	    break;
    }

    valP->v_flags |= VF_RVALUE;
}
/*

*//* val_ref( rvalP, svalP )

	Make a new reference to an existing value

Accepts :

	rvalP		Result value
	svalP		Source value

Returns :

	<nothing>

Notes :

	Instantiates a new value that's a reference to another
	one.  This will make an lvalue that either points to
	the same data as the source or points directly to the
	"simple" data in the source data.  Also calls
	val_access where needed to obtain the correct
	access.
*/

void
val_ref ARGLIST( (rvalP, svalP) )
   NFARG( VALUE		*rvalP )	/* Result value */
    FARG( VALUE		*svalP )	/* Source value */
{
    /* Get the flags, masked by what we want to keep */
    rvalP->v_flags = svalP->v_flags & (0);
    rvalP->v_flags |= VF_LVALUE;

    /* Get the type */
    rvalP->v_typeP = svalP->v_typeP;

    /* Set up the data info depending on the type */
    switch ( rvalP->v_typeP->ty_class ) {
	case TC_VOID:		/* Data is in value */
	case TC_UINT:
	case TC_INT:
	case TC_FLOAT:
	case TC_ENUM:
	case TC_PTR:
	case TC_STRING:
	    rvalP->v_data.d_ptr.p_byteX = PO_SIMPLE;
	    rvalP->v_data.d_ptr.p_ptr.p_genP = (void *)(&svalP->v_data);
	    break;


	case TC_UNION:		/* Data is dynamically allocated using DVS */
	case TC_STRUCT:
	case TC_DIM:
	    /* Copy the DVS reference and take an access */
	    rvalP->v_data.d_ptr = svalP->v_data.d_ptr;
	    val_access( rvalP );
	    break;

	case TC_FUNCT:		/* Private data */
TODO  /* this is out */
#if 0
	    rvalP->v_data.d_ptr = svalP->v_data.d_ptr;
#else
	    warning( "val_ref: reference to function unimplemented." );
	    val_set_int( rvalP, 4, 0 );
#endif
	    break;


	default:
	    warning( "val_ref: unknown type class %d", 
			rvalP->v_typeP->ty_class );
	    val_set_int( rvalP, 4, 0 );
	    break;
    }
}
/*

*//* val_release( valP )

	Finish using a value

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	Finishes access to a value storage.  Must be called after
	a VALUE type has been initialized and is done with.

*/

void
val_release ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	TYPECLASS	tc;		/* Type class */

    if ( valP->v_typeP == NULL )
	return;				/* No type attached. */

    /* Just like val_access, if this is an lvalue, we access the
       lvalue target via the pointer value.
    */
    if ( ( valP->v_flags & VF_LVALUE ) != 0 )
	tc = TC_PTR;
    else
	tc = valP->v_typeP->ty_class;

    /* Process by type. */
    switch( tc ) {
	case TC_FLOAT:			/* Simple values */
	case TC_ENUM:
	case TC_INT:
	case TC_UINT:
	case TC_VOID:
	    break;


	case TC_DIM:			/* Values with complex storage */
	case TC_STRUCT:
	case TC_UNION:
	    /* If there's a DVS reference, release our access */
	    if ( ( valP->v_data.d_ptr.p_byteX >= 0 ) &&
		 ( valP->v_data.d_ptr.p_ptr.p_dvsP != NULL ) )
		(void)dvs_release( valP->v_data.d_ptr.p_ptr.p_dvsP );

	    break;


	case TC_PTR:
	    val_release_ptr( valP );
	    break;


	case TC_STRING:
	    /* Release any refstr */
	    if ( ( valP->v_data.d_ptr.p_byteX == PO_REFSTR ) &&
		 ( valP->v_data.d_ptr.p_ptr.p_genP != NULL ) )
		(void)refstr_free( (REFSTR *)valP->v_data.d_ptr.p_ptr.p_genP );
	    break;


	default:
	    warning( "val_release: unknown type %d", tc );
	    break;
    }
}
/*

*//* val_release_ptr( valP )

	Release a pointer value

Accepts :

	valP		Ptr to value

Returns :

	<nothing>

Notes :

	val_release() for PTR types

*/

void
val_release_ptr ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	VALPTR		*vptrP;		/* The pointer thing */

    /* Release according to pointer type */
    vptrP = &valP->v_data.d_ptr;
    if ( vptrP->p_ptr.p_genP == NULL )
	;				/* do nothing with a null ptr */
    else if ( vptrP->p_byteX >= 0 )
	(void)dvs_release( vptrP->p_ptr.p_dvsP );
    else {
	switch ( vptrP->p_byteX ) {
	    case  PO_UNDEF:
	    case  PO_SIMPLE:
	    case  PO_PRIVATE:
		break;

	    case  PO_REFSTR:
		/* A REFSTR isn't a pointer, this shouldn't happen */
		warning( "val_release_ptr: type says pointer, data says string." );
		break;

	    case  PO_FSV:
		fsv_release( (FSV *)vptrP->p_ptr.p_genP );
		break;

	    default:
		warning( "val_release_ptr: unknown PO type %d",
				vptrP->p_byteX );
		break;
	}
    }
}
/*

*//* val_rvalue(valP)

	Make sure we have the rvalue of a value

Accepts :

	valP		The value

Returns :

	<value>		SCxx status code

Notes :

	Releases the lvalue access, gains the rvalue access

*/

STSCODE
val_rvalue ARGLIST( (valP) )
    FARG( VALUE		*valP )		/* The value */
{
	STSCODE		status;		/* the ubiquitous status value */
	UBYTE		*dataP;		/* Ptr to data */
	VALPTR		*vptrP;
	VALUE		lvalue;		/* Original lvalue */

#ifdef	RESOLVE_NULL_STRING_POINTERS
	FSV		*fsvP;		/* freestanding value */
#endif

    if ( ( valP->v_flags & VF_RVALUE ) == 0 ) {
    	/* rvalue not present, need to fetch it. */
	if ( ( valP->v_flags & VF_CONST ) != 0 ) {
	    /* simple variable -- just rvalue it. */
	    valP->v_flags |= VF_RVALUE;
	    status = SCOK;
	}
	else if ( ( valP->v_flags & VF_LVALUE ) != 0 ) {
	    /* Need a copy of the lvalue for later release (can't
	       release it until we gain the rvalue access later
	       in case of cascading release of the value that
	       we actually want).  That's a very unlikely case
	       but we have to take it into account.
	    */

TODO	    /* All this copying for an extreme case slows down every
	       dereference/rvalue.  We need a better way.
	    */
	    lvalue = *valP;

	    /* Depending on the value type, fetch the value */
	    switch ( valP->v_typeP->ty_class ) {
		case TC_INT:		/* Values that can be fetched */
		case TC_UINT:
		case TC_FLOAT:
		case TC_ENUM:
		case TC_PTR:
		    /* Obtain from lvalue. */
		    dataP = valptr_resolve( &valP->v_data.d_ptr );
		    if ( dataP == NULL ) {
			warning( "Bad rvalue reference." );
			status = SCNOTDONE;
		    }
		    else {
			/* Fetch new rvalue data */
			memcpy( &valP->v_data, 
			         dataP,
			         valP->v_typeP->ty_size );
		    }
		    status = SCOK;
		    break;


		case TC_STRING:		/* String type */
		    /* Similar to the simple types, but we instantiate
		       the string if need be.
		    */
		    vptrP = (VALPTR *)valptr_resolve( &valP->v_data.d_ptr );

		    if ( vptrP == NULL ) {

TODO	/* RESOLVE_NULL_STRING_POINTERS is an experimental hack that
	   isn't settable anywhere nor documented anywhere.  Define it
	   to turn null string pointer dereferences into "" strings.
	*/
#ifdef	RESOLVE_NULL_STRING_POINTERS
			/* Create a new free-standing string */
			if ( ( fsvP = fsv_new() ) == NULL ) {
			    warning( "val_rvalue: fsv_new failed." );
			    status = SCERR;
			    break;
			}
			vptrP = (VALPTR *)&fsvP->fv_data;
			vptrP->p_byteX = PO_REFSTR;
			vptrP->p_ptr.p_genP = NULL;

			fsvP->fv_ptrF = TRUE;	/* HACK: see fsv_release() */
#else	/* RESOLVE_NULL_STRING_POINTERS */
			warning( "Bad rvalue reference." );
			status = SCNOTDONE;
			break;
		    }
#endif	/* RESOLVE_NULL_STRING_POINTERS */

		    if ( vptrP->p_ptr.p_genP == NULL ) {
			vptrP->p_byteX = PO_REFSTR;
			vptrP->p_ptr.p_genP = refstr_new( "val_rvalue",
							 20, 0 );
			if ( vptrP->p_ptr.p_genP == NULL ) {
			    warning( "val_rvalue: string initialization failed." );
			    status = SCNOTDONE;
			    break;
			}
		    }
		    if ( ( vptrP->p_byteX != PO_REFSTR ) ||
		         ( vptrP->p_ptr.p_genP == NULL ) ) {
			warning( "val_rvalue: bad string reference." );
			status = SCNOTDONE;
		    }
		    else {
			valP->v_data.d_ptr = *vptrP;
			status = SCOK;
		    }
		    break;

		

		case TC_UNION:		/* Values that are references */
		case TC_STRUCT:
		case TC_DIM:
		    status = SCOK;
		    break;

		default:
		    warning( "Unknown type %d for val_rvalue.", 
			       valP->v_typeP->ty_class );
		    status = SCNOTDONE;
		    break;
	    }

	    if ( status == SCOK ) {
		valP->v_flags |= VF_RVALUE;
		valP->v_flags &= ~VF_LVALUE;

		/* Gain access to the result */
		val_access( valP );

		/* Relese access to the original lvalue */
		val_release( &lvalue );
	    }
	}
	else {
	    /* No value */
	    warning( "val_rvalue can't get rvalue." );
	    status = SCNOTDONE;
	}

	if ( ( status == SCOK ) && 
             ( ( valP->v_flags & VF_RVALUE ) == 0 ) )
	    warning( "val_rvalue failed" );
    }
    	
    return( status );
}
/*

*//* val_set_int( valP, isize, ival )

	Set a value to an integer constant

Accepts :

	valP		Ptr to value to put int into
	isize		Size of value in bytes
	ival		The value

Returns :

	<nothing>

Notes :


*/

void
val_set_int ARGLIST( (valP, isize, ival) )
   NFARG( VALUE		*valP )		/* The value pointer */
   NFARG( int		isize )		/* The byte size */
    FARG( long		ival )		/* The int value */
{
    /* Set flags */
    valP->v_flags = VF_CONST|VF_RVALUE;

    /* Take the type from the appropriate sized value prototype
       and set the appropriate component value
    */
    switch ( isize ) {
	case 1:
	    valP->v_typeP = Tnint1->tn_typeP;
	    valP->v_data.d_byte = ival;
	    break;

	case 2:
	    valP->v_typeP = Tnint2->tn_typeP;
	    valP->v_data.d_word = ival;
	    break;

	default:
	    warning( "Bad isize %d in val_set_int()!", isize );
	    /* fall through */
	case 4:
	    valP->v_typeP = Tnint4->tn_typeP;
	    valP->v_data.d_long = ival;
	    break;
    }
}
/*

*//* val_set_string( valP, esP )

	Set a value to a string

Accepts :

	valP		Ptr to value to put string
	esP		Source string to copy

Returns :

	<value>		Status code

Notes :


*/

STSCODE
val_set_string ARGLIST( (valP, esP) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( ESTR		*esP )		/* Source string */
{
	STSCODE		status;
	int		len;
	REFSTR		*rsP;		/* New REFSTR */

    len = estr_len( esP );
    rsP = refstr_new( "val_set_string", len, 0 );
    if ( rsP == NULL ) {
	warning( "val_set_string failed to make string." );
	return ( SCERR );
    }

    if ( !refstr_replacen( rsP, 0, 0, esP->es_strP, len ) ) {
	status = SCNOTDONE;
	refstr_free( rsP );
    }
    else
	status = val_set_string_refstr( valP, rsP );

    return ( status );
}
/*

*//* val_set_string_refstr( valP, rsP )

	Set a value to a string from a refstr

Accepts :

	valP		Ptr to value to put string
	rsP		refstr to use as string value

Returns :

	<value>		Status code

Notes :


*/

STSCODE
val_set_string_refstr ARGLIST( (valP, rsP) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( REFSTR	*rsP )		/* New string */
{
    /* Set up the value */
    valP->v_typeP = Tnstring->tn_typeP;
    valP->v_data.d_ptr.p_ptr.p_genP = rsP;
    valP->v_data.d_ptr.p_byteX = PO_REFSTR;
    valP->v_flags = VF_RVALUE;

    return ( SCOK );
}
/*

*//* val_set_string_refstr_ptr( valP, rsP )

	Set a value to a string pointer to an existing refstr

Accepts :

	valP		Ptr to value to put string pointer
	rsP		REFSTR to set the pointer to.

Returns :

	<value>		Status code

Notes :

	Creates a string pointer using an FSV (free-standing value)

*/

STSCODE
val_set_string_refstr_ptr ARGLIST( (valP, rsP) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( REFSTR	*rsP )		/* The REFSTR to point to */
{
	STSCODE		status;
	int		len;
	FSV		*fsvP;		/* free-standing value (to point to) */
	VALPTR		*vptrP;		/* Access to the valptr in the FSV */

    /* Make free-standing value for the pointer to point to */
    fsvP = fsv_new();
    if ( fsvP == NULL ) {
	warning( "val_set_string_ptr: fsv_new failed for string" );
	status = SCERR;
    }
    else {
	vptrP = (VALPTR *)&fsvP->fv_data;
	vptrP->p_byteX = PO_REFSTR;
	vptrP->p_ptr.p_genP = rsP;

	fsvP->fv_ptrF = TRUE;		/* HACK: see fsv_release() */

	valP->v_flags = VF_RVALUE;
	valP->v_typeP = &TystrpP->ts_type;
	valP->v_data.d_ptr.p_ptr.p_genP = fsvP;
	valP->v_data.d_ptr.p_byteX = PO_FSV;

	status = SCOK;
    }

    return ( status );
}
/*

*//* val_set_string_ptr( valP, esP )

	Set a value to a string pointer

Accepts :

	valP		Ptr to value to put string pointer
	esP		Source string to copy, or NULL

Returns :

	<value>		Status code

Notes :

	Creates a string pointer using an FSV (free-standing value)

	If esP is passed as NULL, this creates a NULL string pointer.

*/

STSCODE
val_set_string_ptr ARGLIST( (valP, esP) )
   NFARG( VALUE		*valP )		/* The value pointer */
    FARG( ESTR		*esP )		/* Source string */
{
	STSCODE		status;
	int		len;
	REFSTR		*rsP;		/* New REFSTR */

    if ( esP == NULL ) {
	/* NULL source is special case */
	valP->v_flags = VF_RVALUE;
	valP->v_typeP = &TystrpP->ts_type;
	valP->v_data.d_ptr.p_ptr.p_genP = NULL;
	valP->v_data.d_ptr.p_byteX = PO_UNDEF;
	return ( SCOK );
    }

    len = estr_len( esP );
    rsP = refstr_new( "val_set_string", len, 0 );
    if ( rsP == NULL ) {
	warning( "val_set_string failed to make string." );
	status = SCERR;
    }
    else if ( !refstr_replacen( rsP, 0, 0, esP->es_strP, len ) ) {
	refstr_free( rsP );
	status = SCNOTDONE;
    }
    else if ( ( status = val_set_string_refstr_ptr( valP, rsP ) ) != SCOK ) {
	refstr_free( rsP );
    }
    else {
	/* Successful, initialize anything special here */
	;
    }

    return ( status );
}
/*

*//* val_string_cmp( resultP, patvalP, strvalP, mtype, mflags )

	Compare two string values

Accepts :

	resultP		Where to put result
	patvalP		Second string (pattern string)
	strvalP		First string (test string)
	mtype		Match type (MAT_xxxx)
	mflags		Match flags (from mml)

Returns :

	<value>		Status code

Notes :

	Function value (status code) is SCOK unless there is some
	error accessing the string values.

	If value is SCOK, we return the test result in *resultP.

*/

STSCODE
val_string_cmp ARGLIST( (resultP, patvalP, strvalP, mtype, mflags) )
   NFARG( int		*resultP )	/* Where to store comparison result */
   NFARG( VALUE		*patvalP )	/* Pattern string */
   NFARG( VALUE		*strvalP )	/* String to test */
   NFARG( STRMATTYPE	mtype )		/* String match type */
    FARG( UWORD		mflags )	/* Match flags */
{
	ESTR		*targetP;	/* String to test (target) */
	ESTR		*patP;		/* Pattern */

    /* Get underlying strings */
    targetP = val_estr_tmp( strvalP );
    if ( targetP == NULL )
	return ( SCNOTDONE );
    patP = val_estr_tmp( patvalP );
    if ( patP == NULL )
	return ( SCNOTDONE );

    /* Hand off to strmatch */
    *resultP = estr_match( patP, targetP, mtype, mflags );

    return( SCOK );
}
/*

*//* val_string_val(rvalP, svalP, stealF)

	Create/steal a string value from another value

Accepts :

	rvalP		Result value
	svalP		Source value
	stealF		If it's OK to steal the string rather than dup it

Returns :

	<value>		SCxx status code

Notes :

	Creates or copies a string from one value to another

	If stealF is set (TRUE), this routine will steal the string
	instead of making another copy, if:
	  - the source is a string type
	  - the source string has only one reference
	Otherwise, an attempt is made to create a new copy.

	It's assumed that rvalP is finished (i.e. we don't have
	to val_finish it before we initialize it).  Caller must
	make sure of this.

	rvalP and svalP *may* point to the same thing, in which
	case stealF is forced to be TRUE.

*/

STSCODE
val_string_val ARGLIST( (rvalP, svalP, stealF) )
   NFARG( VALUE		*rvalP )	/* Result value */
   NFARG( VALUE		*svalP )	/* Source value */
    FARG( BOOL		stealF )	/* If it's OK to steal the string */
{
	STSCODE		status;
	REFSTR		*srsP, *rrsP;	/* Source/result RSP */
	ESTR		*esP;
	VALPTR		*vptrP;

    /* Assume good status */
    status = SCOK;

    val_rvalue( svalP );		/* Need the rvalue */

    /* Gotta depend on the source type. */
    switch ( svalP->v_typeP->ty_class ) {
	/* Types that we can generate a string from easily. */
	case TC_INT:
	case TC_UINT:
	case TC_ENUM:
	case TC_FLOAT:
	case TC_PTR:
	    /* Just use val_estr_tmp */
	    esP = val_estr_tmp( svalP );
	    if ( esP == NULL ) {
		/* Can't make string; warning already given. */
		status = SCNOTDONE;
	    }
	    else {
		/* Create a new REFSTR from this */
		rrsP = refstr_new( "val_string_val", estr_len( esP ), 0 );
		if ( rrsP == NULL )
		    status = SCERR;
		else if ( !refstr_replacee( rrsP, 0, 0, esP ) )
		    status = SCNOTDONE;
		else {
		    /* Success; any special initialization here. */
		    ;
		}
	    }
	    break;


	case TC_STRING:			/* String from string */
	    /* Validate source */
	    vptrP = (VALPTR *)&svalP->v_data.d_ptr;
	    if ( ( vptrP == NULL ) ||
	         ( vptrP->p_byteX != PO_REFSTR ) ) {
		warning( "val_string_val: source string not string." );
		status = SCNOTDONE;
	    }
	    else {
		srsP = (REFSTR *)vptrP->p_ptr.p_genP;
		if ( srsP != NULL ) {
		    /* Just steal it if we can */
		    if ( ( stealF && ( srsP->rs_refC == 1 ) ) ||
		         ( svalP == rvalP ) ) {
#if 0
warning( "val_string_val stealing string." );
#endif
			val_nill( svalP );
			rrsP = srsP;
			esP = NULL;
		    }
		    else {
			/* Get an ESTR to feed the new REFSTR */
			esP = estr_tmp_refstr( srsP );
			if ( esP == NULL )
			    status = SCERR;
		    }
		}
		else
		    /* NULL source string */
		    esP = estr_tmp_str( (UBYTE *)"" );

		/* Create new REFSTR */
		if ( ( esP != NULL ) && ( status == SCOK ) ) {
		    rrsP = refstr_new( "val_string_val", estr_len( esP ), 0 );
		    if ( rrsP == NULL )
			status = SCERR;
		    else if ( !refstr_replacee( rrsP, 0, 0, esP ) )
			status = SCNOTDONE;
		    else {
			/* Success; any special initialization here. */
			;
		    }
		}
	    }
	    break;

	default:
	    warning( "val_string_val: can't make string from type %d",
		       svalP->v_typeP->ty_class );
	    status = SCNOTDONE;
	    break;
    }

    if ( status == SCOK )
	/* rrsP points to the [new] REFSTR.  Complete the result. */
	status = val_set_string_refstr( rvalP, rrsP );

    return ( status );
}
/*

*//* valptr_resolve(vptrP)

	Resolve a VALPTR reference to an actual reference

Accepts :

	vptrP		Ptr to VALPTR

Returns :

	<value>		The actual address pointed to

Notes :

	Returned value should be considered very short-lived, i.e.
	for immediate use

*/

void *
valptr_resolve ARGLIST( (vptrP) )
    FARG( VALPTR	*vptrP )	/* comment */
{
	int		byteX;		/* For fast access to offset */
	UBYTE		*dataP;		/* Resolved data pointer */
	DVS		*dvsP;

    byteX = vptrP->p_byteX;

    /* The kind of byte offset determines how we resolve this */
    if ( byteX >= 0 ) {
	/* DVS reference */
	if ( ( dvsP = vptrP->p_ptr.p_dvsP ) == NULL )
	    dataP = NULL;
	else {
	    if ( byteX >= dvsP->dv_size )
		dataP = NULL;
	    else
		dataP = dvsP->dv_dataP + byteX;
	}
    }

    /* Negative- the pointer depends on what p_byteX is */
    else switch ( byteX ) {
	case PO_SIMPLE:			/* Ptr is to value storage */
	case PO_REFSTR:
	    dataP = vptrP->p_ptr.p_genP;
	    break;

	case PO_FSV:			/* Ptr to single VALPTR @ FSV */
	    dataP = (UBYTE *)(&((FSV *)vptrP->p_ptr.p_genP)->fv_data);
	    break;

	default:			/* Unknown! */
	    dataP = NULL;
	    break;
    }

    return ( (void *)dataP );
}
