/* authmisc.c -- General functions related to authentication

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some general functions related to authentication

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>

#include "mvmf.h"

#include <mml/mml-ll.h>
#include <mml/mml-in.h>
#include <mml/mml-str.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "mvmfmisc.h"
#include "authmisc.h"

/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */



/*

*//* auth_checkpassword( progP, unameP, passwdP )

	An interface to a djb-style "checkpassword" program

Accepts :

	progP		Ptr to the path to the checkpassword program
	unameP		Username to check
	passwdP		Password to check

Returns :

	<value>		STSCODE, one of:
			  SCOK:  authenticated
			  SCREJECT: not authenticated
			  SCNOTDONE: not attempted (some syntax or other issue)
			  SCERR:  some error

Notes :

	Uses the interface described at

	    http://cr.yp.to/checkpwd/interface.html

	which is, roughly:

	    send username<NUL>password<NUL>timestamp<NUL>
	      (and optionally more data)
	    to the specified checkpassword program on fd 3.
	    The checkpassword program will, on success, exec another
	      program (the second optional part of progP),
	      and returns the result code from that program.
	      Expected result codes are;
	        0 -- ok
		1 -- not authenticated
		2 -- error in the interface
		MEX_FORKFAIL -- error in setting up child process 

	The checkpassword program will theoretically invoke a second
	program, which we default to /usr/bin/true or some such.  The
	passed argument progP may include that second program path, by
	including a space and that path.  e.g. progP could contain

	    "/usr/local/bin/checkpassword bin/true"

	Some simple checks are done on progP to try to reject any
	funny stuff.

*/

STSCODE
auth_checkpassword ARGLIST( (progP, unameP, passwdP) )
   NFARG( char		*progP )	/* Path to the checkpassword program */
   NFARG( UBYTE		*unameP )	/* Username */
    FARG( UBYTE		*passwdP )	/* Password */
{
	STSCODE		status;
	BOOL		okF;
	char		ch;
	int		i;
	int		prog2X;
	int		waitstat, exitcode;
	pid_t		pid;		/* child pid */
	pid_t		epid;		/* Pid of child that exited */
	int		cppipe[2];	/* Pipe to talk to checkpassword with */
	char		cppath[ PATH_MAX +1 ];
	char		cp2path[ PATH_MAX +1 ];
	char		*argv[3];	/* Arguments for execv */

    /* Scan the program name for:
         - suspicious characters
	 - any optional second program for checkpassword to invoke
    */
    prog2X = 0;
    for ( i = 0; ( ch = progP[i] ) != NUL; ++i ) {
	/* Check for a few bad chars */
	if ( ( ch == ';' ) ||
	     iscntrl( ch ) )
	return ( SCNOTDONE );

	/* Check for start of second prog name */
	if ( ( i > 0 ) && ( prog2X == 0 ) && ( ch == ' ' ) )
	    prog2X = i;
    }

    /* Looks OK.. get the paths to the two programs. */
    if ( prog2X == 0 ) {
	if ( i >= ( sizeof( cppath ) -1 ) )
	    return ( SCNOTDONE );
	strcpy( &cppath[0], progP );
	if ( snprintf( &cp2path[0], sizeof(cp2path),
		       "/usr/bin/true" ) >= sizeof(cp2path) )
	    return ( SCNOTDONE );
    }
    else {
	if ( prog2X >= ( sizeof( cppath ) -1 ) )
	    return ( SCNOTDONE );
	memcpy( &cppath[0], progP, prog2X );
	cppath[prog2X] = NUL;
	for ( ++prog2X; ( ch = progP[prog2X] ) != NUL; ++prog2X )
	    if ( ch == ' ' )
		continue;
	if ( ch == NUL ) {
	    /* A second name wasn't really there, just spaces. */
	    if ( snprintf( &cp2path[0], sizeof(cp2path),
			   "/usr/bin/true" ) >= sizeof(cp2path) )
		return ( SCNOTDONE );
	}
	else {
	    if ( snprintf( &cp2path[0], sizeof(cp2path),
			   "%s",
			   &progP[prog2X] ) >= sizeof(cp2path) )
		return ( SCNOTDONE );
	}
    }

    /* Make pipe & child */
    if ( pipe( &cppipe[0] ) < 0 )
	 return ( SCERR );
    pid = fork();
    if ( pid < 0 ) {
	/* Error.. close pipe, be done */
	close( cppipe[0] );
	close( cppipe[1] );
	return ( SCERR );
    }

    if ( pid == 0 ) {
	/* Child process.  First sanitize.. */
	if ( !sanitize_process() )
	    _exit( MEX_FORKFAIL );

	/* Close the write side of the pipe, and make
	   sure the read side is on fd 3.
	*/
	close( cppipe[1] );
	if ( cppipe[0] != 3 ) {
	    if ( dup2( cppipe[0], 3 ) < 0 )
		_exit(MEX_FORKFAIL);
	    close( cppipe[0] );
	}

	/* Become the checkpassword program */
	argv[0] = &cppath[0];
	argv[1] = &cp2path[0];
	argv[2] = NULL;
	execv( &cppath[0], &argv[0] );

	_exit( MEX_FORKFAIL );			/* Some failure */
    }

    /* Parent here.  We're only writing to the pipe, so close the
       read end.
    */
    close( cppipe[0] );

    /* Write the username including its trailing NUL */
    i = strlen( (char *)unameP ) +1;
    okF = write( cppipe[1], unameP, i ) == i;
    if ( okF ) {
	/* Write the passwd and its NUL */
	i = strlen( (char *)passwdP ) +1;
	okF = write( cppipe[1], passwdP, i ) == i;
    }
    if ( okF ) {
	/* Write the timestamp (just null string for now, maybe ever) */
	okF = write( cppipe[1], "", 1 ) == 1;
    }

    close( cppipe[1] );

    epid = waitpid( pid, &waitstat, 0 );
    if ( epid != pid )
	 return ( SCERR );

    exitcode = WEXITSTATUS( waitstat );
    switch ( exitcode ) {
	case	0:			/* Auth OK */
	    status = SCOK;
	    break;

	case	1:			/* Some non-auth */
	case	MEX_FORKFAIL:
	    status = SCREJECT;
	    break;

	default:			/* Anything else */
	    status = SCERR;
	    break;
    }

    return ( status );
}
