/* controls -- Stuff about mvmf program controls

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This module collects some of the mvmf controls, and has some
functions to access them.

*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-in.h>
#include <mml/mml-str.h>
#include <mml/mml-avl.h>

#include "mvmf.h"

#include "controls.h"

#include "errwarn.h"

/* Local definitions */

    /* Structure of a node in the controls tree */
typedef struct {
    AVLNODE	cn_avl;		/* AVL node header */
    CNTRL	cn_cntrl;	/* The control info */
}  CNTRLNODE;


/* External data referenced */



/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* AVL support functions for AVL tree */

static	int	cntrl_avl_cmpkey();
static	AVLNODE	*cntrl_avl_mknode();
static	int	cntrl_avl_rmnode();

    /* Called for each node by avlwalk on behalf of cntrl_walk */
static	int	cntrl_walk_node();

    /* Add a new control to the table. */
static	CNTRL	*define_cntrl PROTO( (ESTR *nameP,
				       CNTRLTYPE type, void *infoP) );

/* Private data */

static	BOOL	Cntrl_initF = FALSE;	/* Whether inited */
static	AVLTREE	*Cntrl_treeP;		/* Tree for controls */

/*

*//* cntrl_define_int( nameP, value, abslow, abshigh, userlow, userhigh)

	Define an integer control

Accepts :

	nameP		Name of the control
	value		Initial value for the control
	abslow		Absolute low value that the control can have
	abshigh		Absolute high value that the control can have
	userlow		Low value that can be set in user mode
	userhigh	High value that can be set in user mode

Returns :

	<value>		Ptr to CNTRL struct if successful,
			NULL if not (duplicate name or error)

Notes :

	Uses define_cntrl, which makes a copy of the name.

*/

CNTRL *
cntrl_define_int ARGLIST( (nameP, value, abslow, abshigh, userlow, userhigh) )
   NFARG( ESTR		*nameP )	/* Control name */
   NFARG( long		value )		/* Initial value */
   NFARG( long		abslow )	/* Low limit (admin) */
   NFARG( long		abshigh )	/* High limit (admin) */
   NFARG( long		userlow )	/* Low limit (user) */
    FARG( long		userhigh )	/* High limit (user) */
{
	CNTRLINT	*ciP;		/* Integer info for control */
	CNTRL		*ctlP;

    /* Make sure it doesn't already exist */
    if ( cntrl_find( nameP ) != NULL )
	return ( NULL );

    /* Allocate and fill in the integer info */
    ciP = (CNTRLINT *)emalloc( "cntrl_define_int", "int info",
				    sizeof(CNTRLINT) );
    if ( ciP == NULL )
	return ( NULL );

    ciP->cv_int = value;
    ciP->cv_abslow = abslow;
    ciP->cv_abshigh = abshigh;
    ciP->cv_userlow = userlow;
    ciP->cv_userhigh = userhigh;

    ctlP = define_cntrl( nameP, CTL_INT, (void *)ciP );
    if ( ctlP == NULL )
	dealloc( ciP );

    return ( ctlP );
}
/*

*//* cntrl_define_str( nameP, valueP, adminF )

	Define a new string control

Accepts :

	nameP		Name of the control
	valueP		Initial value for the control, or NULL
	adminF		Admin-lock flag (set this if value can only be
			  changed in admin mode)

Returns :

	<value>		Ptr to CNTRL struct if successful,
			NULL if not (duplicate name or error)

Notes :

	Uses define_cntrl, which makes a copy of the name.

	valueP may be passed as NULL to leave the value unset.
	Otherwise, this function makes a copy of the initial value.

*/

CNTRL *
cntrl_define_str ARGLIST( (nameP, valueP, adminF) )
   NFARG( ESTR		*nameP )	/* Control name */
   NFARG( ESTR		*valueP )	/* Initial value */
    FARG( BOOL		adminF )	/* Admin-lock flag */
{
	CNTRLSTR	*csP;		/* String info for control */
	CNTRL		*ctlP;

    /* Make sure it doesn't already exist */
    if ( cntrl_find( nameP ) != NULL )
	return ( NULL );

    /* Allocate and fill in the string info */
    csP = (CNTRLSTR *)emalloc( "cntrl_define_str", "string info",
				    sizeof(CNTRLSTR) );
    if ( csP == NULL )
	return ( NULL );

    csP->cs_adminF = adminF;
    if ( valueP != NULL ) {
	csP->cs_strP = (UBYTE *)newstr_n( "cntrl_define_str", "string value",
				          (char *)valueP->es_strP,
					  estr_len( valueP ) );
	if ( csP->cs_strP == NULL ) {
	    dealloc( csP );
	    return ( NULL );
	}
    }
    else {
	/* No initial value: set to NULL */
	csP->cs_strP = NULL;
    }

    /* Add/define it */
    ctlP = define_cntrl( nameP, CTL_STR, (void *)csP );
    if ( ctlP == NULL ) {
	if ( csP->cs_strP != NULL )
	    dealloc( csP->cs_strP );
	dealloc( csP );
    }

    return ( ctlP );
}
/*

*//* cntrl_find( nameP )

	Find a control given its name

Accepts :

	nameP		Name of control to find

Returns :

	<value>		Address of CNTRL struct for the control named.

Notes :


*/

CNTRL *
cntrl_find ARGLIST( (nameP) )
    FARG( ESTR		*nameP )	/* The name of the control */
{
	CNTRLNODE	*cnodeP;

    /* Make sure things are initialized */
    if ( !Cntrl_initF )
	if ( !cntrl_init() )
	    return ( NULL );

    /* Look up the node and return what we found */
    cnodeP = (CNTRLNODE *)avlfind( Cntrl_treeP, nameP );
    if ( cnodeP == NULL )
	return ( NULL );

    return ( &cnodeP->cn_cntrl );
}
/*

*//* cntrl_init()

	Initialize the controls interface

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	It is not required that an application call this function;
	cntrl_xxx interface routines will call it as needed.

*/

BOOL
cntrl_init NOARGLIST
{
    if ( !Cntrl_initF ) {
	Cntrl_treeP = avlnew( NULL,
			     cntrl_avl_cmpkey,
			     cntrl_avl_mknode,
			     cntrl_avl_rmnode );
	if ( Cntrl_treeP != NULL )
	    Cntrl_initF = TRUE;
    }

    return ( Cntrl_initF );
}
/*

*//* cntrl_int_get(nameP, notfound)

	Get the value of an integer control

Accepts :

	nameP		Name of the control
	notfound	Value to return if control is not found.

Returns :

	<value>		the value, or the notfound value

Notes :

	This is a convenient routine to get the value of an integer
	control.  If the control doesn't exist, the passed "notfound"
	value is returned.  "notfound" should be chosen to be out of
	the normal range of values for the control being queried,
	or a value which will be appropriate for a default if the
	control is not found.

	If there is no out-of-range value possible, this function should
	probably not be used.

*/

long
cntrl_int_get ARGLIST( (nameP, notfound) )
   NFARG( ESTR		*nameP )	/* Control name */
    FARG( long		notfound )	/* Value if not found */
{
	CNTRL		*ctlP;

    ctlP = cntrl_find( nameP );
    if ( ( ctlP != NULL ) &&
	 ( ctlP->ctl_type == CTL_INT ) )
	return ( ((CNTRLINT *)(ctlP->ctl_infoP))->cv_int );

    return ( notfound );
}
/*

*//* cntrl_int_set( nameP, cat, val )

	Set an integer control, by name

Accepts :

	nameP		Name of the control
	cat		Access type
	val		New value to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Mainly a wrapper around cntrl_intc_set() -- see that for notes

*/

BOOL
cntrl_int_set ARGLIST( (nameP, cat, val) )
   NFARG( ESTR		*nameP )	/* Name of control */
   NFARG( CNTRLACC	cat )		/* Access type */
    FARG( long		val )		/* Value to set */
{
	CNTRL		*ctlP;

    ctlP = cntrl_find( nameP );
    if ( ( ctlP == NULL ) ||
	 ( ctlP->ctl_type != CTL_INT ) )
	return ( FALSE );

    return ( cntrl_intc_set( (CNTRLINT *)(ctlP->ctl_infoP), cat, val ) );
}
/*

*//* cntrl_int_setlim( nameP, cat, low, high )

	Set integer control limits

Accepts :

	nameP		Name of the control
	cat		Access type
	val		New low limit to set
	high		New high limit to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Mainly a wrapper around cntrl_intc_setlim() -- set that for notes

*/

BOOL
cntrl_int_setlim ARGLIST( (nameP, cat, low, high) )
   NFARG( ESTR		*nameP )	/* Name of control */
   NFARG( CNTRLACC	cat )		/* Access type */
   NFARG( long		low )		/* New low limit */
    FARG( long		high )		/* New high limit */
{
	CNTRL		*ctlP;

    ctlP = cntrl_find( nameP );
    if ( ( ctlP == NULL ) ||
	 ( ctlP->ctl_type != CTL_INT ) )
	return ( FALSE );

    return ( cntrl_intc_setlim( (CNTRLINT *)(ctlP->ctl_infoP),
			         cat, low, high ) );
}
/*

*//* cntrl_intc_set( ciP, cat, val )

	Set an integer control, by control ptr

Accepts :

	ciP		Ptr to integer control struct to adjust
	cat		Access type
	val		New value to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
cntrl_intc_set ARGLIST( (ciP, cat, val) )
   NFARG( CNTRLINT	*ciP )		/* Ptr to int control info */
   NFARG( CNTRLACC	cat )		/* Access type */
    FARG( long		val )		/* Value to set */
{
	BOOL		okF;

    /* Validate according to access type */
    switch ( cat ) {
	case	CA_USER:
	    okF = ( ( val >= ciP->cv_userlow ) &&
		    ( val <= ciP->cv_userhigh ) );
	    break;

	case	CA_ADMIN:
	    okF = ( ( val >= ciP->cv_abslow ) &&
		    ( val <= ciP->cv_abshigh ) );
	    break;


	default:
	    error( "cntrl_int_set: unknown access type %d", cat );
	    okF = FALSE;
	    break;
    }

    if ( okF )
	ciP->cv_int = val;

    return ( okF );
}
/*

*//* cntrl_intc_setlim( ciP, cat, low, high )

	Set integer control limits, by control pointer

Accepts :

	ciP		Ptr to integer control struct to adjust
	cat		Access type
	val		New low limit to set
	high		New high limit to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Adjusts user limits only.  Absolute limits are not changed.

	This function is intended to be used for an admin access to
	adjust the user limits.  However the potential for a user
	to downgrade their limits is also provided.  Not sure that
	part will ever be used..

*/

BOOL
cntrl_intc_setlim ARGLIST( (ciP, cat, low, high) )
   NFARG( CNTRLINT	*ciP )		/* Ptr to int control info */
   NFARG( CNTRLACC	cat )		/* Access type */
   NFARG( long		low )		/* New low limit */
    FARG( long		high )		/* New high limit */
{
	BOOL		okF;

    /* Validate according to access type */
    switch ( cat ) {
	case	CA_USER:
	    okF = ( ( low >= ciP->cv_userlow ) &&
		    ( low <= ciP->cv_userhigh ) &&
		    ( high >= ciP->cv_userlow ) &&
		    ( high <= ciP->cv_userhigh ) );
	    break;

	case	CA_ADMIN:
	    okF = ( ( low >= ciP->cv_abslow ) &&
		    ( low <= ciP->cv_abshigh ) &&
		    ( high >= ciP->cv_abslow ) &&
		    ( high <= ciP->cv_abshigh ) );
	    break;


	default:
	    error( "cntrl_int_set: unknown access type %d", cat );
	    okF = FALSE;
	    break;
    }

    if ( okF && ( high < low ) )
	okF = FALSE;

    if ( okF ) {
	ciP->cv_userlow = low;
	ciP->cv_userhigh = high;
    }

    return ( okF );
}
/*

*//* cntrl_str_get(nameP)

	Get the value of a string control

Accepts :

	nameP		Name of the control

Returns :

	<value>		the value
			NULL if no such string control.

Notes :

	This is a convenient routine to get the value of a string
	control.  If the control doesn't exist or is not a string
	control, NULL is returned.

*/

UBYTE *
cntrl_str_get ARGLIST( (nameP) )
    FARG( ESTR		*nameP )	/* Control name */
{
	CNTRL		*ctlP;

    ctlP = cntrl_find( nameP );
    if ( ( ctlP != NULL ) &&
	 ( ctlP->ctl_type == CTL_STR ) )
	return ( ((CNTRLSTR *)(ctlP->ctl_infoP))->cs_strP );

    return ( NULL );
}
/*

*//* cntrl_str_set( nameP, cat, valP )

	Set a string control, by name

Accepts :

	nameP		Name of the control
	cat		Access type
	valP		New string to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Mainly a wrapper around cntrl_strc_set() -- see that for notes

*/

BOOL
cntrl_str_set ARGLIST( (nameP, cat, valP) )
   NFARG( ESTR		*nameP )	/* Name of control */
   NFARG( CNTRLACC	cat )		/* Access type */
    FARG( ESTR		*valP )		/* New string value */
{
	CNTRL		*ctlP;

    ctlP = cntrl_find( nameP );
    if ( ( ctlP == NULL ) ||
	 ( ctlP->ctl_type != CTL_STR ) )
	return ( FALSE );

    return ( cntrl_strc_set( (CNTRLSTR *)(ctlP->ctl_infoP), cat, valP ) );
}
/*

*//* cntrl_strc_set( csP, cat, valP )

	Set a string control, by control pointer

Accepts :

	csP		Ptr to string control struct to adjust
	cat		Access type
	valP		New string to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	New values are always installed in freshly allocated memory.

*/

BOOL
cntrl_strc_set ARGLIST( (csP, cat, valP) )
   NFARG( CNTRLSTR	*csP )		/* Ptr to string control info */
   NFARG( CNTRLACC	cat )		/* Access type */
    FARG( ESTR		*valP )		/* New string value */
{
	BOOL		okF;

    /* Validate according to access type */
    switch ( cat ) {
	case	CA_USER:
	    okF = !csP->cs_adminF;
	    break;

	case	CA_ADMIN:
	    okF = TRUE;
	    break;


	default:
	    error( "cntrl_string_set: unknown access type %d", cat );
	    okF = FALSE;
	    break;
    }

    if ( okF ) {
	if ( csP->cs_strP != NULL )
	    dealloc( csP->cs_strP );

	csP->cs_strP = (UBYTE *)newstr_n( "cntrl_str_set", "new string value",
				          (char *)valP->es_strP,
					  estr_len( valP ) );
	if ( csP->cs_strP == NULL )
	    okF = FALSE;
    }

    return ( okF );
}
/*

*//* cntrl_table_def_int( defP )

	Define integer controls out of a table

Accepts :

	defP		Ptr to a table of definitions

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	defP points to a table of CNTRLDEF_INT structs, ending with
	  an entry with a NULL name pointer.

*/

BOOL
cntrl_table_def_int ARGLIST( (defP) )
    FARG( CNTRLDEF_INT	*defP )		/* Ptr to definitions table */
{

    /* Define from the table */
    for ( ; defP->cti_nameP != NULL; ++defP )
	if ( !cntrl_define_int( estr_tmp_str( defP->cti_nameP ),
			        defP->cti_int,
			        defP->cti_abslow, defP->cti_abshigh,
			        defP->cti_userlow, defP->cti_userhigh ) )
	    return ( FALSE );

    return ( TRUE );
}
/*

*//* cntrl_table_def_str( defP )

	Define string controls out of a table

Accepts :

	defP		Ptr to a table of definitions

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	defP points to a table of CNTRLDEF_STR structs, ending with
	  an entry with a NULL name pointer.

*/

BOOL
cntrl_table_def_str ARGLIST( (defP) )
    FARG( CNTRLDEF_STR	*defP )		/* Ptr to definitions table */
{
    /* Define from the table */
    for ( ; defP->cts_nameP != NULL; ++defP )
	if ( !cntrl_define_str( estr_tmp_str( defP->cts_nameP ),
			        estr_tmp_str( defP->cts_strP ),
				defP->cts_adminF ) )
	    return ( FALSE );

    return ( TRUE );
}
/*

*//* cntrl_walk( rtc )

	Walk the list of controls

Accepts :

	rtc		routine to call, see Notes

Returns :

	<nothing>

Notes :

	Calls the callback routine 'rtc' for each control defined.
	rtc should be defined as:

		void rtc( CNTRL *ctlP )

*/

void
cntrl_walk ARGLIST( (rtc) )
    FARG( void		(*rtc)() )	/* Routine to call */
{
    /* Pass the rtc to our local walk routine, via the tree struct */
    Cntrl_treeP->t_walkP = (void *)rtc;
    avlwalk( Cntrl_treeP, cntrl_walk_node );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* cntrl_avl_cmpkey( keyP, nodeP )

	Compare keys for avl tree routines, for controls tree

Accepts :

	keyP		Ptr to the subject key (a string)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:

	Note that the key is a counted string, whereas the string
	  in the node is a NUL-terminated string.

*/

static int
cntrl_avl_cmpkey( keyP, nodeP )
	ESTR		*keyP;		/* Key to look for */
	CNTRLNODE	*nodeP;		/* Node to compare against */
{
	int		cmpval;
	int		keylen;

    /* Do case-insensitive compare on the strings. */
    keylen = estr_len( keyP );
    cmpval = strnicmp( (char *)keyP->es_strP, nodeP->cn_cntrl.ctl_nameP,
		       keylen );

    if ( cmpval == 0 )
	/* Might be a match, but not if the string in the node is
	   longer than the string in the key.
	*/
	if ( keylen < strlen( nodeP->cn_cntrl.ctl_nameP ) )
	    cmpval = -1;

    return ( cmpval );
}
/*

*//* cntrl_avl_mknode( treeP, keyP, dataP, enodeP )

	Make an avl node on behalf of AVL tree routines,
	  for the controls tree.

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (the control name)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
cntrl_avl_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	ESTR		*keyP;		/* Key for the new node */
	void		*dataP;		/* Addr of new data (n/a) */
	CNTRLNODE	*enodeP;	/* Existing node... */
{
	int		keylen;		/* Length of new key */
	int		nodesize;	/* Size of node required */
	CNTRLNODE	*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the name/key) */
    keylen = estr_len( keyP );
    nodesize = sizeof(CNTRLNODE) + keylen + 1;
    nodeP = (CNTRLNODE *)emalloc( "cntrl_avl_mknode", "avl node",
					 nodesize );

    /* Setup the node struct */
    nodeP->cn_cntrl.ctl_nameP = (void *)&nodeP[1];
    memcpy( nodeP->cn_cntrl.ctl_nameP, keyP->es_strP, keylen );
    nodeP->cn_cntrl.ctl_nameP[ keylen ] = NUL;
    nodeP->cn_cntrl.ctl_type = 0;
    nodeP->cn_cntrl.ctl_infoP = NULL;

    return( (AVLNODE *)nodeP );
}
/*

*//* cntrl_avl_rmnode( treeP, nodeP )

	Delete an avl node on behalf of AVL tree routines,
	  for controls tree

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :


*/

static int
cntrl_avl_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	CNTRLNODE	*nodeP;
{
    /* Deallocate the info */
    if ( nodeP->cn_cntrl.ctl_infoP != NULL ) {
	dealloc( nodeP->cn_cntrl.ctl_infoP );
	nodeP->cn_cntrl.ctl_infoP = NULL;
    }

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* cntrl_walk_node( treeP, nodeP, depth )

	Handle each tree node on behalf of cntrl_walk()

Accepts :

	treeP		Ptr to AVL tree
	nodeP		Ptr to the node to give to application's rtc
	depth		depth of the node

Returns :

	<value>		0 to continue
			non-zero to stop

Notes :

	Per avlwalk() interface; see that.

*/

static int
cntrl_walk_node ARGLIST( (treeP, nodeP, depth) )
   NFARG( AVLTREE	*treeP )	/* Ptr to tree */
   NFARG( CNTRLNODE	*nodeP )	/* Ptr to node */
    FARG( int		depth )		/* Depth of node */
{
	void		(*rtc)();	/* Application's rtc */

    /* NULL node signals end of walk */
    if ( nodeP == NULL )
	return ( 0 );			/* no special processing */

    /* Pass the control struct to the application's function.
       Use intermediate rtc variable to avoid funky casts.
    */
    rtc = treeP->t_walkP;
    (*rtc)(&nodeP->cn_cntrl);

    return ( 0 );
}
/*

*//* define_cntrl( nameP, type, infoP )

	Define a new control

Accepts :

	nameP		Name of the control
	type		Type of control (CNTRLTYPE)
	infoP		Specific control info to attach

Returns :

	<value>		Ptr to CNTRL struct if successful,
			NULL if not (duplicate name or error)

Notes :

	Assumes caller has already checked for duplicates.

	Makes a copy of the name for the new control struct.

*/

static CNTRL *
define_cntrl ARGLIST( (nameP, type, infoP) )
   NFARG( ESTR		*nameP )	/* Control name */
   NFARG( CNTRLTYPE	type )		/* Control type */
    FARG( void		*infoP )	/* Control info */
{
	int		i;
	CNTRLNODE	*cnodeP;	/* AVL node */
	CNTRL		*ctlP;		/* New control */

    /* Make sure things are initialized */
    if ( !Cntrl_initF )
	if ( !cntrl_init() )
	    return ( NULL );

    /* Insert the new node and set it up */
    avlinsert( Cntrl_treeP, nameP, NULL );
    cnodeP = (CNTRLNODE *)avlfind( Cntrl_treeP, nameP );
    if ( cnodeP == NULL )
	return ( NULL );

    ctlP = &cnodeP->cn_cntrl;
    
    ctlP->ctl_type = type;
    ctlP->ctl_infoP = infoP;

    return ( ctlP );
}
