/* dnsbl.c -- functions dealing with DNS blacklist processing

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Functions dealing with DNS blacklist processing.

*/

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "mvmf.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "data.h"

#include "mvmfmisc.h"
#include "msgpart.h"
#include "omm.h"

/* Local definitions */


    /* Information about each result type from a DNSBL */
typedef struct {
    char	*br_nameP;	/* Name for this result */
    ns_type	br_type;	/* Type of result */
    char	*br_codeP;	/* Code returned for this result */
    BOOL	br_queriedF;	/* Whether previously queried */
    BOOL	br_matchedF;	/* If resolved, whether it matched */
}  BLRESULT;


    /* Information about each blacklist */
typedef struct dnsbl {
    char	*bl_nameP;	/* The name of the DNSBL */
    char	*bl_hostP;	/* Hostname of the DNSBL */
    LL_HEAD	bl_results;	/* Kinds of results */
    BOOL	bl_wildqueryF;	/* If any wildcard query has been done */
    BOOL	bl_wildmatchF;	/* If any wildcard query has matched */
}  DNSBL;



/* External data referenced */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	DNSBL	*dnsbl_find PROTO( (char *blnameP) );
static	BLRESULT  *result_find PROTO( (DNSBL *dnsblP, char *rnameP) );

/* Private data */

	LL_HEAD	Dnsbls;		/* List of DNSbls */


/*

*//* dnsbl_check_ip( ipP, blnameP, rnameP )

	Check a specific IP address against a dnsbl

Accepts :

	ipP		String representing the IP address to check
	blnameP		Name of blacklist to check
	rnameP		Name of the result to check

Returns :

	<value>		TRUE if matched
			FALSE if not.

Notes :


*/

BOOL
dnsbl_check_ip ARGLIST( (ipP, blnameP, rnameP) )
   NFARG( char		*ipP )		/* IP address string */
   NFARG( char		*blnameP )	/* Name of DNSBL */
    FARG( char		*rnameP )	/* Name of result */
{
	DNSBL		*dnsblP;	/* Ptr to dnsbl entry */
	BLRESULT	*resultP;	/* Ptr to result block */
	MML_DNS		*dnsP;		/* mml-dns handle */
	in_addr_t	testaddr;	/* An address to test */
	in_addr_t	ipaddr;		/* an IP address */
	UBYTE		ip_octets[4];	/* IP octets */
	char		qname[ NS_MAXDNAME ];

    /* Check the ip pointer for kicks */
    if ( ipP == NULL )
	return ( FALSE );

    if ( ( dnsblP = dnsbl_find( blnameP ) ) == NULL )
	return ( FALSE );		/* No such DNSBL name */

    /* Find this result code */
    if ( ( rnameP == NULL ) || ( strcmp( rnameP, "*" ) == 0 ) ) {
	/* No specific result test- remember, this is OK. */
	resultP = NULL;

TODO	/* This result-caching stuff was when we only had a single IP
	   ... it's broken now with the multiple IP om_rip stuff.
	*/
#if 0
	/* If there has been any wildcard query against this dnsbl
	   previously, return that result now.
	*/
	if ( dnsblP->bl_wildqueryF )
	    return ( dnsblP->bl_wildmatchF );

	/* Remember that we made a wildcard query. */
	dnsblP->bl_wildqueryF = TRUE;
#endif
    }
    else {
	resultP = result_find( dnsblP, rnameP );
	if ( resultP == NULL )
	    /* Specific result wanted but not found: return no match */
	    return ( FALSE );

TODO	/* result caching broken, see above. */
#if 0
	/* If there has already been a query against this result block,
	   return the result now rather than doing another DNS lookup.
	*/
	if ( resultP->br_queriedF )
	    return ( resultP->br_matchedF );
	else
	    /* Remember that we have already tested for this result */
	    resultP->br_queriedF = TRUE;
#endif
    }

    
    /* Gotta do a DNS query */
    if ( ( dnsP = dns_handle() ) == NULL )
	return ( FALSE );

    testaddr = inet_addr( ipP );	/* Note: returns in network order */

    /* Formulate a query and check the answer. */
    mdns_ip4_iptoba( ntohl( testaddr ), &ip_octets[0] );
    snprintf( &qname[0], sizeof(qname), "%d.%d.%d.%d.%s",
	      ip_octets[3],
	      ip_octets[2],
	      ip_octets[1],
	      ip_octets[0],
	      dnsblP->bl_hostP );

    if ( !mdns_query( dnsP, &qname[0], ns_c_in,
		      (resultP != NULL) ? resultP->br_type : ns_t_a ) )
	return ( FALSE );

    /* There is a result of some sort- if no specific result wanted
       (code of "*") then we pass.
    */
    if ( resultP == NULL ) {
	/* Remember that there was some match against this dnsbl */
	dnsblP->bl_wildmatchF = TRUE;
	return ( TRUE );
    }

    /* Do any per-type setup */
    switch ( resultP->br_type ) {
	case	ns_t_a:
	    ipaddr = inet_addr( resultP->br_codeP );
	    break;

	default:
	    break;
    }

    /* Check the answer(s) against the result */
    do {
	/* Check for desired answer type */
	if ( dnsP->dn_ansX != resultP->br_type )
	    continue;

	/* Process according to type */
	switch ( resultP->br_type ) {
	    case	ns_t_a:		/* Address */
		if ( ipaddr == dnsP->dn_ipaddr ) {
		    /* a match!  remember for later */
		    resultP->br_matchedF = TRUE;

		    /* Also classify this as a wildcard match. */
		    dnsblP->bl_wildqueryF = TRUE;
		    dnsblP->bl_wildmatchF = TRUE;
		    
		    return ( TRUE );
		}
		break;

	    default:			/* Anything else */
		break;
	}

    } while ( mdns_ans_next( dnsP ) );

    return ( FALSE );
}
/*

*//* dnsbl_check_omm( ommP, blnameP, rnameP )

	Check omm info for particular DNSBL result

Accepts :

	ommP		Ptr to open mail message struct
	blnameP		Name of blacklist to check
	rnameP		Name of the result to check

Returns :

	<value>		TRUE if matched
			FALSE if not.

Notes :

	Checks all of the "responsible IP addresses" in the OMM
	against the specified dnsbl.

	rnameP may be passed as NULL or "*" for any standard match

*/

BOOL
dnsbl_check_omm ARGLIST( (ommP, blnameP, rnameP) )
   NFARG( OMM		*ommP )		/* Open mail message thing */
   NFARG( char		*blnameP )	/* Name of DNSBL */
    FARG( char		*rnameP )	/* Name of result */
{
	int		ipX;

    if ( ommP == NULL )
	return ( FALSE );

    /* Loop through the testable IP address(es) and check
       each one via dnsbl_check_ip()
    */
    for ( ipX = 0; ipX < ommP->om_ripC; ++ipX ) {
	if ( dnsbl_check_ip( ommP->om_ripPP[ipX], blnameP, rnameP ) )
	    return ( TRUE );
    }

    return ( FALSE );
}
/*

*//* dnsbl_init()

	Initialize DNSBL stuff

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
dnsbl_init NOARGLIST
{
	int		ch;
	int		tokX;		/* Token index */
	char		*strP;
	char		*typeP;		/* Result type */
	FILE		*ifP;		/* File ptr */
	DNSBL		*dnsblP;	/* DNSBL entry */
	BLRESULT	*resultP;	/* Result code block */

    /* Init the DNSBL linked list */
    ll_init( &Dnsbls );

    /* Process the DNSBL definitions file */
    ifP = fopen( LIBDNSBL, "r" );
    if ( ifP == NULL )
	return TRUE;			/* Can't access, just ignore */

    /* Set up to use the generic buffer */
    Gbs1P->bs_bM = 10000;
    Gbs1P->bs_chunk = 100;

    for ( ; ; ) {
	/* Get next line */
	if ( fgetbline( ifP, Gbs1P ) < 0 )
	    break;

	/* Tokenize */
	tkline( Gbs1P, GtknsP );
	if ( ( GtknsP->tp_tC == 0 ) ||
	     ( GtkbsP->bs_bP[ GtknsP->tp_tbP[0].tb_startX ] == '#' ) )
	    continue;

	/* Initialize DNSBL block */
	dnsblP = (DNSBL *)emalloc( "dnsbl_init", "DNSBL block", sizeof(DNSBL) );
	dnsblP->bl_nameP = newstr( "dnsbl_init", "DNSBL name",
		(char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ 0 ].tb_startX ] );
	dnsblP->bl_hostP = NULL;
	dnsblP->bl_wildqueryF = FALSE;
	dnsblP->bl_wildmatchF = FALSE;
	ll_init( &dnsblP->bl_results );

	/* Process tokens in this definition */
	for ( tokX = 1; ; ) {
	    /* If we've run out of tokens, check for a continuation line */
	    if ( tokX >= GtknsP->tp_tC ) {
		ch = fgetc( ifP );
		if ( ( ch != '#' ) &&
		     ( ch != ' ' ) &&
		     ( ch != '	' ) ) {
		    /* Not continuation line, take back character */
		    if ( ch != EOF )
			ungetc( ch, ifP );

		    /* Exit.. this entry all processed */
		    break;
		}

		/* Get and tokenize line */
		fgettkbline( ifP, Gbs1P, GtknsP );

		tokX = 0;
		continue;		/* Try again */
	    }

	    /* Process next token.. */
	    strP = (char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ tokX ].tb_startX ];
	    if ( stricmp( strP, "host" ) == 0 ) {
		if ( (tokX +2) > GtknsP->tp_tC )
		    continue;		/* No host token.. ignore */
		dnsblP->bl_hostP = newstr( "dnsbl_init", "host name",
					  (char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ tokX +1 ].tb_startX ] );
		tokX += 2;
	    }

	    else if ( stricmp( strP, "Result" ) == 0 ) {
		/* Requires 3 arguments:  name, type, code to match */
		if ( ( tokX + 4 ) > GtknsP->tp_tC )
		    continue;

		/* Create a result block */
		resultP = (BLRESULT *)emalloc( "dnsbl_init", "result block",
						sizeof( BLRESULT ) );

		resultP->br_nameP = newstr( "dnsbl_init", "result name",
			(char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ tokX+1 ].tb_startX ] );

		resultP->br_codeP = newstr( "dnsbl_init", "result code",
			(char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ tokX+3 ].tb_startX ] );

		resultP->br_queriedF = FALSE;
		resultP->br_matchedF = FALSE;

		/* Fill in the result type */
		typeP = (char *)&GtkbsP->bs_bP[ GtknsP->tp_tbP[ tokX+2 ].tb_startX ];
		if ( stricmp( typeP, "A" ) == 0 )
		    resultP->br_type = ns_t_a;
		else if ( stricmp( typeP, "TXT" ) == 0 )
		    resultP->br_type = ns_t_txt;
		else
		    /* Dunno- default to "A" without giving error */
		    resultP->br_type = ns_t_a;

		/* Link in this result code to the dnsbl entry */
		ll_addvt( &dnsblP->bl_results, (void *)resultP );

		tokX += 4;
	    }
	    else if ( strcmp( strP, "//" ) == 0 ) {
		/* Comment/remark -- just ignore rest of line */
		tokX = GtknsP->tp_tC;
	    }
	    else {
#if 0
		fprintf( stderr, "Unknown keyword \"%s\" in dnsbl file", strP );
#endif
		++tokX;
	    }
	}

	/* Done with definition; add it if it is reasonably complete */
	if ( dnsblP->bl_hostP != NULL ) {
	    ll_addvt( &Dnsbls, (void *)dnsblP );
	}
    }

    return ( TRUE );
}
/*

*//* dnsbl_walk( rtc )

	Walks through the DNSBL result definitions, calling a function for each
	 
Accepts :

	rtc		Ptr to routine to call

Returns :

	<nothing>

Notes :

	rtc is passed (nameP, rnameP)
	   nameP is the name of the blacklist
	   rnameP is the name of the result definition

	a final call with (NULL, NULL) signals the end.

	rtc can do whatever it wants with the info.  It returns TRUE
	   to continue, FALSE to stop.

*/

void
dnsbl_walk ARGLIST( (rtc) )
    FARG( BOOL		(*rtc)() )	/* Routine to call */
{
	DNSBL		*dnsblP;	/* Ptr to dnsbl entry */
	BLRESULT	*resultP;	/* Ptr to result block */
	LL_NODE		*dnodeP;	/* Ptr to node for DNSBL */
	LL_NODE		*rnodeP;	/* Ptr to node for result */

    /* Go through the list */
    for ( dnodeP = Dnsbls.llh_firstP;
	    dnodeP != NULL;
	       dnodeP = dnodeP->lln_nextP ) {

	dnsblP = (DNSBL *)dnodeP->lln_valP;

	/* Go through the result list */
	for ( rnodeP = dnsblP->bl_results.llh_firstP;
	       rnodeP != NULL;
		  rnodeP = rnodeP->lln_nextP ) {

	    resultP = (BLRESULT *)rnodeP->lln_valP;

	    /* Call the function */
	    if ( ! (*rtc)( dnsblP->bl_nameP, resultP->br_nameP ) )
		return;
	}
    }

    /* Send a "done" indication */
    (*rtc)( (char *)NULL, (char *)NULL );
}



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/



/*

*//* dnsbl_find( blnameP )

	Find a DNSBL entry by name

Accepts :

	blnameP		Name of DNSBL to look up

Returns :

	<value>		ptr to dnsbl entry if found
			NULL if not

Notes :


*/

static DNSBL *
dnsbl_find ARGLIST( (blnameP) )
    FARG( char		*blnameP )	/* BL name */
{
	DNSBL		*dnsblP;	/* Ptr to entry */
	LL_NODE		*nodeP;		/* Ptr to list node */

    /* Go through the list.. */
    for ( nodeP = Dnsbls.llh_firstP;
	    nodeP != NULL;
	       nodeP = nodeP->lln_nextP ) {

	dnsblP = (DNSBL *)nodeP->lln_valP;
	if ( stricmp( dnsblP->bl_nameP, blnameP ) == 0 )
	    return ( dnsblP );
    }

    /* No go */
    return ( NULL );
}
/*

*//* result_find( dnsblP, rnameP )

	Find a DNSBL result code entry by name

Accepts :

	dnsblP		Ptr to DNSBL entry
	rnameP		Name of result code look up

Returns :

	<value>		ptr to result block if found
			NULL if not

Notes :


*/

static BLRESULT *
result_find ARGLIST( (dnsblP, rnameP) )
   NFARG( DNSBL		*dnsblP )	/* DNSBL entry to look in */
    FARG( char		*rnameP )	/* result name */
{
	BLRESULT	*resultP;	/* Result ptr */
	LL_NODE		*nodeP;		/* Ptr to list node */

    /* Go through the list.. */
    for ( nodeP = dnsblP->bl_results.llh_firstP;
	    nodeP != NULL;
	       nodeP = nodeP->lln_nextP ) {

	resultP = (BLRESULT *)nodeP->lln_valP;
	if ( stricmp( resultP->br_nameP, rnameP ) == 0 )
	    return ( resultP );
    }

    /* Not found. */
    return ( NULL );
}
