/* errwarn.c -- handle error and warning messages for mvmf

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Various mvmf/mfl functions call error routines with printf-style
arguments.  These are those routinges.

*/

#define	D_MVMF_ERRWARN		/* To include defines local to this module */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <syslog.h>
#include <pwd.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include <mml/mml.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-fmt.h>

#include "mvmf.h"
#include "controls.h"

#include "msgpart.h"
#include "omm.h"
#include "errwarn.h"
#include "data.h"

/* Local definitions */



/* External data referenced */

extern	OMM	*OmmP;		/* Any open mail message */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	void	get_log_priv_name PROTO( (void) );
static	void	log_xpriv PROTO( ( char *fmtP, va_list ap ) );
static	void	log_xsys PROTO( ( char *fmtP, va_list ap ) );
static	void	mfl_xwarning PROTO( ( MML_IN *inP, char *fmtP, va_list ap) );


/* Private data */

static	BOOL	Log_enabledF = FALSE;	/* If log_xxx() are enabled yet */
static	FILE	*Log_privfP = NULL;	/* FP for log_priv() */
static	char	Log_priv_fname[ PATH_MAX +1 ];
static	BOOL	Log_sys_initF = FALSE;	/* If log_sys() is inited yet */
static	char	Errbuf[200];		/* Buffer for formatted error text */


/*

*//* log_begin()

	Signify that log_xxx functions can operate

Accepts :

	<nothing>

Returns :

	<nothing>


Notes :

	log_xxx() operations are inhibited until log_begin is called.

*/

void
log_begin NOARGLIST
{
    Log_enabledF = TRUE;
}
/*

*//* log_both( fmtP, args... )

	Log a message to both log_priv and log_sys

Accepts :

	fmtP		printf-style format string
	args		printf arguments.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

*/

void
log_both ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    /* Hand off to the underlying log routines */
    log_xpriv( fmtP, ap );
    log_xsys( fmtP, ap );

    va_end( ap );
}
/*

*//* log_end()

	Terminate log_xxx functions

Accepts :

	<nothing>

Returns :

	<nothing>


Notes :

	Finishes up all log_xxx() operations and disables
	further ones.

*/

void
log_end NOARGLIST
{
    Log_enabledF = FALSE;
}
/*

*//* log_priv( fmtP, args... )

	Log a message to the private log

Accepts :

	fmtP		printf-style format string
	args		printf arguments.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

*/

void
log_priv ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    /* Hand off to the underlying log routine */
    log_xpriv( fmtP, ap );

    va_end( ap );
}
/*

*//* log_sys( fmtP, args... )

	Log a message to the system log

Accepts :

	fmtP		printf-style format string
	args		printf arguments.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

*/

void
log_sys ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    /* Hand off to the underlying log routine */
    log_xsys( fmtP, ap );

    va_end( ap );
}
/*

*//* error( fmtP, args... )

	Print an error and don't return.

Accepts :

	fmtP		printf-style format string
	args		printf arguments.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

*/

void
error ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    /* Format the basic error text */
    vsnprintf( &Errbuf[0], sizeof(Errbuf), fmtP, ap );

    /* Give it to stderr */
    fprintf( stderr, "Error: %s\n", &Errbuf[0] );

    va_end( ap );

    /* Exit with bad status */
    exit( 1 );
}
/*

*//* mfl_warning( inP, fmtP, args... )

	Issue a warning message for mfl functions

Accepts :

	inP		Pointer to input stream, or NULL
	fmtP		printf-style format string
	args		printf arguments.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

	Issues its warning text to stderr

	If inP is non-NULL, attempts to give some information about the
	input location where the error occured.

	If (global var) OmmP is non-NULL, attaches the warning text to the
	omm's error message area (which presumably will be written to the
	mail message body)

*/

void
mfl_warning ARGLIST( ( inP, fmtP, va_decl ) )
   NFARG( MML_IN	*inP )		/* Input stream */
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    /* Hand off to mfl_xwarning */
    mfl_xwarning( inP, fmtP, ap );

    va_end( ap );
}
/*

*//* warning( fmtP, args... )

	Issue a warning message.

Accepts :

	fmtP		printf-style format string
	args		printf arguments.  Quantity is limited.

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

*/

void
warning ARGLIST( ( fmtP, va_decl ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs arg pointer */

    /* Start varargs */
    va_start( ap, fmtP );

    mfl_xwarning( NULL, fmtP, ap );

    va_end( ap );
}



/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* get_log_priv_name()

	Get the actual private log filename

Accepts :


Returns :

	<nothing>

Notes :

	An MFL user may specify the private log file name by
	setting the "log_private_name" string control.  This
	function copies that name into a private buffer with
	any required translations and/or sanitizations.

	Specifically:

	  - leading ~/ is translated to homedir/

*/

void
get_log_priv_name NOARGLIST
{
	int		len;
	char		*fnameP;	/* The user-directed name */
	char		*userP;
	uid_t		myuid;
	struct passwd	*pwdP;

    /* Get a local reference to the user-specified name */
    fnameP = (char *)cntrl_str_get( estr_tmp_str( "log_private_name" ) );
    if ( fnameP == NULL )
	fnameP = "";

    /* Since there's only one translation we do (the "~/") we just
       do it in hackish fashion.
    */
    if ( ( fnameP[0] == '~' ) && ( fnameP[1] == '/' ) ) {
	myuid = getuid();

	/* If there's a USER environment variable that matches my uid,
	   we'll allow that user to override the default uid lookup.
	*/
	userP = getenv( "USER" );
	if ( ( userP != NULL ) && ( *userP != NUL ) ) {
	    pwdP = getpwnam( userP );
	    if ( ( pwdP == NULL ) || ( pwdP->pw_uid != myuid ) )
		pwdP = getpwuid( myuid );
	}
	else
	    pwdP = getpwuid( myuid );
	if ( pwdP == NULL ) {
	    warning( "log_priv_fname: can't expand \"~/\"." );
	    return;			/* Sorry. */
	}

	len = snprintf( &Log_priv_fname[0], sizeof(Log_priv_fname),
		  "%s/%s", pwdP->pw_dir,
		  ( fnameP[2] != NUL ) ? &fnameP[2] : "mvmf.log" );
    }
    else
	len = snprintf( &Log_priv_fname[0], sizeof(Log_priv_fname),
		       "%s", fnameP );

    if ( len >= sizeof(Log_priv_fname) ) {
	warning( "log_priv_fname: name too long." );
	Log_priv_fname[0] = NUL;
    }
}
/*

*//* log_xpriv( fmtP, ap )

	Handler for log_priv

Accepts :

	fmtP		printf-style format string
	ap		va_list argument stack pointer

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

	See log_priv for basic description

	This function exists as common code for log_priv and other
	functions that want to pass their varags to that function.

*/

static void
log_xpriv ARGLIST( ( fmtP, ap ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( va_list	ap )		/* Varargs argument list */
{
	int		i;
	int		size;
	time_t		timenow;
	char		*errP;		/* Ptr to error buffer */
	va_list		lap;		/* Local copy of ap */

    /* If logging not enabled, don't do anything */
    if ( !Log_enabledF )
	return;

    /* Copy the argument list handle for use here */
    va_copy( lap, ap );

    get_log_priv_name();		/* Copy/sanitize log file name */
    if ( Log_priv_fname[0] == NUL )
	return;

    /* Briefly open the private file for append. */
    Log_privfP = fopen( &Log_priv_fname[0], "a" );
    if ( Log_privfP == NULL ) {
TODO		/* Should probably log this? */
	return;
    }

    errP = &Errbuf[0];
    size = sizeof(Errbuf);

    /* Prefix with timestamp */
    time( &timenow );
    i = snprintf( errP, size, "%s ", time_to_fullstr( timenow ) );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;
    

    /* Might as well put in my uid/username */
    i = snprintf( errP, size, "User: %d/%s ",
		  Myuid,
		  (MyusernameP != NULL) ? MyusernameP : "<unknown>" );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;


    /* Append/format the error text */
    i = vsnprintf( errP, size, fmtP, lap );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;

    /* Log it */
    fprintf( Log_privfP, "%s\n", &Errbuf[0] );

    /* Done with file */
    fclose( Log_privfP );
}
/*

*//* log_xsys( fmtP, ap )

	Handler for log_sys

Accepts :

	fmtP		printf-style format string
	ap		va_list argument stack pointer

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

	See log_sys for basic description

	This function exists as common code for log_sys and other
	functions that want to pass their varags to that function.

*/

static void
log_xsys ARGLIST( ( fmtP, ap ) )
   NFARG( char		*fmtP )		/* message format string */
    FARG( va_list	ap )		/* stdargs argument list handle */
{
	int		i;
	int		size;
	char		*prognameP;	/* Program name */
	char		*errP;		/* Ptr to error buffer */
	va_list		lap;		/* Local copy of ap */

    /* If logging not enabled, don't do anything */
    if ( !Log_enabledF )
	return;

    /* Copy the argument list handle for use here */
    va_copy( lap, ap );

    /* If system logging not yet initialized, initialize it */
    if ( !Log_sys_initF ) {
	prognameP = (char *)cntrl_str_get( estr_tmp_str( "log_program_name" ) );
	if ( prognameP == NULL )
	    prognameP = "mvmf";
	openlog( prognameP, LOG_PID, MVMF_LOGFACILITY );
	Log_sys_initF = TRUE;
    }

    errP = &Errbuf[0];
    size = sizeof(Errbuf);

    /* Start off with uid/username */
    i = snprintf( errP, size, "User: %d/%s ",
		  Myuid,
		  (MyusernameP != NULL) ? MyusernameP : "<unknown>" );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;


    /* Append/format the error text */
    i = vsnprintf( errP, size, fmtP, lap );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;

    /* Log it */
    syslog( LOG_INFO, "%s", &Errbuf[0] );
}
/*

*//* mfl_xwarning( inP, fmtP, ap )

	Handler for mfl_warning

Accepts :

	inP		Pointer to input stream, or NULL
	fmtP		printf-style format string
	ap		va_list argument stack pointer

Returns :

	Nothing.

Notes :

	Assumes stdc with function prototypes.

	See mfl_warning for basic description

	This function exists as common code for mfl_warning and other
	functions that want to pass their varags to that function.

*/

static void
mfl_xwarning ARGLIST( ( inP, fmtP, ap ) )
   NFARG( MML_IN	*inP )		/* Input stream */
   NFARG( char		*fmtP )		/* message format string */
    FARG( va_list	ap )		/* stdargs arglist handle */
{
	int		i;
	int		size;
	char		*errP;		/* Ptr to error buffer */
	char		*prognameP;
	va_list		lap;		/* Local copy of ap */

    /* Copy the argument list handle for use here */
    va_copy( lap, ap );

    errP = &Errbuf[0];
    size = sizeof(Errbuf);

    prognameP = (char *)cntrl_str_get( estr_tmp_str( "log_program_name" ) );
    if ( prognameP == NULL )
	prognameP = "mvmf";

    if ( inP != NULL ) {
	/* Try to give some scope/location */
	i = snprintf( errP, size, "%s at %s: ",
		       prognameP,
		       in_stream_location( inP ));
    }
    else {
	i = snprintf( errP, size, "%s: ", prognameP );
    }
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;
    

    /* Append/format the error text */
    i = vsnprintf( errP, size, fmtP, lap );
    if ( i > size )
	i = size;
    if ( ( i > 0 ) && ( errP[i -1] == NUL ) )
	--i;
    errP += i;
    size -= i;

    /* Tell stderr */
    fprintf( stderr, "%s\n", &Errbuf[0] );

    /* If any open mail message, save the message there */
    if ( OmmP != NULL ) {
	bstr_cat( OmmP->om_errsP, &Errbuf[0] );
	bstr_putb( OmmP->om_errsP, '\n' );
    }
}
