/* msgpart.c -- Stuff dealing with message parts

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Provides functions to deal with message parts (see msgpart.h).
A message part is a header/body pair that represents, well,
a part of a message.  The body may be recursively divided into
other message parts as well.  A complete mail message would be
represented by a top-level message part -- its header and its
subparts (if any).

*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <regex.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "mvmf.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "msgpart.h"
#include "omm.h"

#include "data.h"

#include "errwarn.h"

/* Local definitions */

    /* How header lines are terminated (in the context of getbtrec().
       Header lines can end with CRLF, CR, or LF.
    */
#define	HDR_TXR		((TXR_END_CRLF)|(TXR_END_CR)|(TXR_END_LF))

    /* How body lines are terminated (in the context of getbtrec().
       Body lines can end with CRLF, CR, or LF,
       and we stop reading when the buffer is full.
    */
#define	BODY_TXR	((TXR_END_CRLF)|(TXR_END_CR)|(TXR_END_LF)|TXR_END_FULL)



    /* Flags for writing message.  Easier than passing lots of BOOLs. */
#define	MWF_BLBEGIN	0x0001		/* Start with a blank line */
#define	MWF_NLEND	0x0002		/* Make sure it ends with newline */
#define	MWF_BLEND	0x0004		/* Make sure it ends with a blank
					   line (two consecutive newlines)
					*/



/* External data referenced */

extern	char	*Daynames[];
extern	char	*Monthnames[];

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* See if a string is a boundary for a message part */
static	int	isboundary PROTO( (MSGPART *mpP, BSTR *bsP) );

    /* Hdrlearn_disp[] dispatch routines */
static	BOOL	hdrlearn_content_type PROTO( (MSGPART *mpP, REFSTR *rsP) );

    /* Free MIMEINFO data */
static	BOOL	mimeinfo_free PROTO( (MIMEINFO *miP, BOOL freeF) );

    /* Get a temporary BSTR */
static	BSTR	*msgpart_temp_bstr PROTO( (void) );

    /* Write from an input stack to a file */
static	BOOL	msgpart_write_stack PROTO( (MSGPART *mpP, MML_IN *inP,
					    FILE *fP, UWORD flags) );

    /* Add bytes from counted buffer to a MSGPSTR storage */
static	BOOL	msgpstr_add_n PROTO( (MSGPSTR *psP,
				      UBYTE *bufP, int bufL) );

    /* Free MSGPSTR data */
static	BOOL	msgpstr_free PROTO( (MSGPSTR *psP, BOOL freeF) );

    /* Read a byte-string part via MSGPSTR */
static	BOOL	msgpstr_read PROTO( (MSGPART *mpP, MSGPSTR *psP,
				     BOOL toendF) );

    /* Initialize a byte string part */
static	void	msgpstr_init PROTO( (MSGPSTR *psP) );

    /* Get the size of a byte string part */
static	ULONG	msgpstr_size PROTO( (MSGPART *mpP, MSGPSTR *psP, BOOL ncF) );

    /* Write any byte string part */
static	BOOL	msgpstr_write PROTO( (MSGPART *mpP, MSGPSTR *psP,
				      FILE *ofP, UWORD flags) );


/* Private data */

static	BSTR	*BsP = NULL;		/* A BSTR used for scratch */

    /* Dispatch table for msgpart_learn_headers -- header names
       and routines to call to learn about them.  Must be in
       alphabetical order by header name (case insensitive!)
    */
static	struct {
    char	*hl_hnameP;		/* Header name */
    BOOL	(*hl_rtc)();		/* Routine to call */
}  Hdrlearn_disp[] = {

{	"Content-Type",		hdrlearn_content_type	},
{	NULL,			NULL			}

};


    /* Mail character table.  Each index/entry corresponds to a byte
       value; the table value is a bitmask of attributes for that
       byte value.  This is generated by a quick&dirty program that
       may be still laying around here.

       Note this is actually public data.

    */
UWORD	Mchartbl[256] = {
    0|MC_CTL,					/* 0x00   (0)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x01   (1)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x02   (2)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x03   (3)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x04   (4)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x05   (5)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x06   (6)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x07   (7)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x08   (8)  */
    0|MC_CTL,					/* 0x09   (9)  */
    0|MC_CTL,					/* 0x0a  (10)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x0b  (11)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x0c  (12)  */
    0|MC_CTL,					/* 0x0d  (13)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x0e  (14)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x0f  (15)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x10  (16)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x11  (17)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x12  (18)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x13  (19)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x14  (20)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x15  (21)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x16  (22)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x17  (23)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x18  (24)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x19  (25)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1a  (26)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1b  (27)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1c  (28)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1d  (29)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1e  (30)  */
    0|MC_CTL|MC_NWSCTL,				/* 0x1f  (31)  */
    0|MC_SPC,					/* 0x20  (32)  ' '  */
    0|MC_ATOM,					/* 0x21  (33)  '!'  */
    0|MC_TSPL,					/* 0x22  (34)  '"'  */
    0|MC_ATOM,					/* 0x23  (35)  '#'  */
    0|MC_ATOM,					/* 0x24  (36)  '$'  */
    0|MC_ATOM,					/* 0x25  (37)  '%'  */
    0|MC_ATOM,					/* 0x26  (38)  '&'  */
    0|MC_ATOM,					/* 0x27  (39)  '''  */
    0|MC_TSPL,					/* 0x28  (40)  '('  */
    0|MC_TSPL,					/* 0x29  (41)  ')'  */
    0|MC_ATOM,					/* 0x2a  (42)  '*'  */
    0|MC_ATOM,					/* 0x2b  (43)  '+'  */
    0|MC_TSPL,					/* 0x2c  (44)  ','  */
    0|MC_ATOM,					/* 0x2d  (45)  '-'  */
    0,						/* 0x2e  (46)  '.'  */
    0|MC_ATOM|MC_TSPL,				/* 0x2f  (47)  '/'  */
    0|MC_DIGIT,					/* 0x30  (48)  '0'  */
    0|MC_DIGIT,					/* 0x31  (49)  '1'  */
    0|MC_DIGIT,					/* 0x32  (50)  '2'  */
    0|MC_DIGIT,					/* 0x33  (51)  '3'  */
    0|MC_DIGIT,					/* 0x34  (52)  '4'  */
    0|MC_DIGIT,					/* 0x35  (53)  '5'  */
    0|MC_DIGIT,					/* 0x36  (54)  '6'  */
    0|MC_DIGIT,					/* 0x37  (55)  '7'  */
    0|MC_DIGIT,					/* 0x38  (56)  '8'  */
    0|MC_DIGIT,					/* 0x39  (57)  '9'  */
    0|MC_TSPL,					/* 0x3a  (58)  ':'  */
    0|MC_TSPL,					/* 0x3b  (59)  ';'  */
    0|MC_TSPL,					/* 0x3c  (60)  '<'  */
    0|MC_ATOM|MC_TSPL,				/* 0x3d  (61)  '='  */
    0|MC_TSPL,					/* 0x3e  (62)  '>'  */
    0|MC_ATOM|MC_TSPL,				/* 0x3f  (63)  '?'  */
    0|MC_TSPL,					/* 0x40  (64)  '@'  */
    0|MC_ALPHA,					/* 0x41  (65)  'A'  */
    0|MC_ALPHA,					/* 0x42  (66)  'B'  */
    0|MC_ALPHA,					/* 0x43  (67)  'C'  */
    0|MC_ALPHA,					/* 0x44  (68)  'D'  */
    0|MC_ALPHA,					/* 0x45  (69)  'E'  */
    0|MC_ALPHA,					/* 0x46  (70)  'F'  */
    0|MC_ALPHA,					/* 0x47  (71)  'G'  */
    0|MC_ALPHA,					/* 0x48  (72)  'H'  */
    0|MC_ALPHA,					/* 0x49  (73)  'I'  */
    0|MC_ALPHA,					/* 0x4a  (74)  'J'  */
    0|MC_ALPHA,					/* 0x4b  (75)  'K'  */
    0|MC_ALPHA,					/* 0x4c  (76)  'L'  */
    0|MC_ALPHA,					/* 0x4d  (77)  'M'  */
    0|MC_ALPHA,					/* 0x4e  (78)  'N'  */
    0|MC_ALPHA,					/* 0x4f  (79)  'O'  */
    0|MC_ALPHA,					/* 0x50  (80)  'P'  */
    0|MC_ALPHA,					/* 0x51  (81)  'Q'  */
    0|MC_ALPHA,					/* 0x52  (82)  'R'  */
    0|MC_ALPHA,					/* 0x53  (83)  'S'  */
    0|MC_ALPHA,					/* 0x54  (84)  'T'  */
    0|MC_ALPHA,					/* 0x55  (85)  'U'  */
    0|MC_ALPHA,					/* 0x56  (86)  'V'  */
    0|MC_ALPHA,					/* 0x57  (87)  'W'  */
    0|MC_ALPHA,					/* 0x58  (88)  'X'  */
    0|MC_ALPHA,					/* 0x59  (89)  'Y'  */
    0|MC_ALPHA,					/* 0x5a  (90)  'Z'  */
    0|MC_TSPL,					/* 0x5b  (91)  '['  */
    0|MC_TSPL,					/* 0x5c  (92)  '\'  */
    0|MC_TSPL,					/* 0x5d  (93)  ']'  */
    0|MC_ATOM,					/* 0x5e  (94)  '^'  */
    0|MC_ATOM,					/* 0x5f  (95)  '_'  */
    0|MC_ATOM,					/* 0x60  (96)  '`'  */
    0|MC_ALPHA,					/* 0x61  (97)  'a'  */
    0|MC_ALPHA,					/* 0x62  (98)  'b'  */
    0|MC_ALPHA,					/* 0x63  (99)  'c'  */
    0|MC_ALPHA,					/* 0x64 (100)  'd'  */
    0|MC_ALPHA,					/* 0x65 (101)  'e'  */
    0|MC_ALPHA,					/* 0x66 (102)  'f'  */
    0|MC_ALPHA,					/* 0x67 (103)  'g'  */
    0|MC_ALPHA,					/* 0x68 (104)  'h'  */
    0|MC_ALPHA,					/* 0x69 (105)  'i'  */
    0|MC_ALPHA,					/* 0x6a (106)  'j'  */
    0|MC_ALPHA,					/* 0x6b (107)  'k'  */
    0|MC_ALPHA,					/* 0x6c (108)  'l'  */
    0|MC_ALPHA,					/* 0x6d (109)  'm'  */
    0|MC_ALPHA,					/* 0x6e (110)  'n'  */
    0|MC_ALPHA,					/* 0x6f (111)  'o'  */
    0|MC_ALPHA,					/* 0x70 (112)  'p'  */
    0|MC_ALPHA,					/* 0x71 (113)  'q'  */
    0|MC_ALPHA,					/* 0x72 (114)  'r'  */
    0|MC_ALPHA,					/* 0x73 (115)  's'  */
    0|MC_ALPHA,					/* 0x74 (116)  't'  */
    0|MC_ALPHA,					/* 0x75 (117)  'u'  */
    0|MC_ALPHA,					/* 0x76 (118)  'v'  */
    0|MC_ALPHA,					/* 0x77 (119)  'w'  */
    0|MC_ALPHA,					/* 0x78 (120)  'x'  */
    0|MC_ALPHA,					/* 0x79 (121)  'y'  */
    0|MC_ALPHA,					/* 0x7a (122)  'z'  */
    0|MC_ATOM,					/* 0x7b (123)  '{'  */
    0|MC_ATOM,					/* 0x7c (124)  '|'  */
    0|MC_ATOM,					/* 0x7d (125)  '}'  */
    0|MC_ATOM,					/* 0x7e (126)  '~'  */
    0|MC_CTL|MC_NWSCTL,				/* 0x7f (127)  DEL  */
    0|MC_NASC,					/* 0x80 (128)  */
    0|MC_NASC,					/* 0x81 (129)  */
    0|MC_NASC,					/* 0x82 (130)  */
    0|MC_NASC,					/* 0x83 (131)  */
    0|MC_NASC,					/* 0x84 (132)  */
    0|MC_NASC,					/* 0x85 (133)  */
    0|MC_NASC,					/* 0x86 (134)  */
    0|MC_NASC,					/* 0x87 (135)  */
    0|MC_NASC,					/* 0x88 (136)  */
    0|MC_NASC,					/* 0x89 (137)  */
    0|MC_NASC,					/* 0x8a (138)  */
    0|MC_NASC,					/* 0x8b (139)  */
    0|MC_NASC,					/* 0x8c (140)  */
    0|MC_NASC,					/* 0x8d (141)  */
    0|MC_NASC,					/* 0x8e (142)  */
    0|MC_NASC,					/* 0x8f (143)  */
    0|MC_NASC,					/* 0x90 (144)  */
    0|MC_NASC,					/* 0x91 (145)  */
    0|MC_NASC,					/* 0x92 (146)  */
    0|MC_NASC,					/* 0x93 (147)  */
    0|MC_NASC,					/* 0x94 (148)  */
    0|MC_NASC,					/* 0x95 (149)  */
    0|MC_NASC,					/* 0x96 (150)  */
    0|MC_NASC,					/* 0x97 (151)  */
    0|MC_NASC,					/* 0x98 (152)  */
    0|MC_NASC,					/* 0x99 (153)  */
    0|MC_NASC,					/* 0x9a (154)  */
    0|MC_NASC,					/* 0x9b (155)  */
    0|MC_NASC,					/* 0x9c (156)  */
    0|MC_NASC,					/* 0x9d (157)  */
    0|MC_NASC,					/* 0x9e (158)  */
    0|MC_NASC,					/* 0x9f (159)  */
    0|MC_NASC,					/* 0xa0 (160)  */
    0|MC_NASC,					/* 0xa1 (161)  */
    0|MC_NASC,					/* 0xa2 (162)  */
    0|MC_NASC,					/* 0xa3 (163)  */
    0|MC_NASC,					/* 0xa4 (164)  */
    0|MC_NASC,					/* 0xa5 (165)  */
    0|MC_NASC,					/* 0xa6 (166)  */
    0|MC_NASC,					/* 0xa7 (167)  */
    0|MC_NASC,					/* 0xa8 (168)  */
    0|MC_NASC,					/* 0xa9 (169)  */
    0|MC_NASC,					/* 0xaa (170)  */
    0|MC_NASC,					/* 0xab (171)  */
    0|MC_NASC,					/* 0xac (172)  */
    0|MC_NASC,					/* 0xad (173)  */
    0|MC_NASC,					/* 0xae (174)  */
    0|MC_NASC,					/* 0xaf (175)  */
    0|MC_NASC,					/* 0xb0 (176)  */
    0|MC_NASC,					/* 0xb1 (177)  */
    0|MC_NASC,					/* 0xb2 (178)  */
    0|MC_NASC,					/* 0xb3 (179)  */
    0|MC_NASC,					/* 0xb4 (180)  */
    0|MC_NASC,					/* 0xb5 (181)  */
    0|MC_NASC,					/* 0xb6 (182)  */
    0|MC_NASC,					/* 0xb7 (183)  */
    0|MC_NASC,					/* 0xb8 (184)  */
    0|MC_NASC,					/* 0xb9 (185)  */
    0|MC_NASC,					/* 0xba (186)  */
    0|MC_NASC,					/* 0xbb (187)  */
    0|MC_NASC,					/* 0xbc (188)  */
    0|MC_NASC,					/* 0xbd (189)  */
    0|MC_NASC,					/* 0xbe (190)  */
    0|MC_NASC,					/* 0xbf (191)  */
    0|MC_NASC,					/* 0xc0 (192)  */
    0|MC_NASC,					/* 0xc1 (193)  */
    0|MC_NASC,					/* 0xc2 (194)  */
    0|MC_NASC,					/* 0xc3 (195)  */
    0|MC_NASC,					/* 0xc4 (196)  */
    0|MC_NASC,					/* 0xc5 (197)  */
    0|MC_NASC,					/* 0xc6 (198)  */
    0|MC_NASC,					/* 0xc7 (199)  */
    0|MC_NASC,					/* 0xc8 (200)  */
    0|MC_NASC,					/* 0xc9 (201)  */
    0|MC_NASC,					/* 0xca (202)  */
    0|MC_NASC,					/* 0xcb (203)  */
    0|MC_NASC,					/* 0xcc (204)  */
    0|MC_NASC,					/* 0xcd (205)  */
    0|MC_NASC,					/* 0xce (206)  */
    0|MC_NASC,					/* 0xcf (207)  */
    0|MC_NASC,					/* 0xd0 (208)  */
    0|MC_NASC,					/* 0xd1 (209)  */
    0|MC_NASC,					/* 0xd2 (210)  */
    0|MC_NASC,					/* 0xd3 (211)  */
    0|MC_NASC,					/* 0xd4 (212)  */
    0|MC_NASC,					/* 0xd5 (213)  */
    0|MC_NASC,					/* 0xd6 (214)  */
    0|MC_NASC,					/* 0xd7 (215)  */
    0|MC_NASC,					/* 0xd8 (216)  */
    0|MC_NASC,					/* 0xd9 (217)  */
    0|MC_NASC,					/* 0xda (218)  */
    0|MC_NASC,					/* 0xdb (219)  */
    0|MC_NASC,					/* 0xdc (220)  */
    0|MC_NASC,					/* 0xdd (221)  */
    0|MC_NASC,					/* 0xde (222)  */
    0|MC_NASC,					/* 0xdf (223)  */
    0|MC_NASC,					/* 0xe0 (224)  */
    0|MC_NASC,					/* 0xe1 (225)  */
    0|MC_NASC,					/* 0xe2 (226)  */
    0|MC_NASC,					/* 0xe3 (227)  */
    0|MC_NASC,					/* 0xe4 (228)  */
    0|MC_NASC,					/* 0xe5 (229)  */
    0|MC_NASC,					/* 0xe6 (230)  */
    0|MC_NASC,					/* 0xe7 (231)  */
    0|MC_NASC,					/* 0xe8 (232)  */
    0|MC_NASC,					/* 0xe9 (233)  */
    0|MC_NASC,					/* 0xea (234)  */
    0|MC_NASC,					/* 0xeb (235)  */
    0|MC_NASC,					/* 0xec (236)  */
    0|MC_NASC,					/* 0xed (237)  */
    0|MC_NASC,					/* 0xee (238)  */
    0|MC_NASC,					/* 0xef (239)  */
    0|MC_NASC,					/* 0xf0 (240)  */
    0|MC_NASC,					/* 0xf1 (241)  */
    0|MC_NASC,					/* 0xf2 (242)  */
    0|MC_NASC,					/* 0xf3 (243)  */
    0|MC_NASC,					/* 0xf4 (244)  */
    0|MC_NASC,					/* 0xf5 (245)  */
    0|MC_NASC,					/* 0xf6 (246)  */
    0|MC_NASC,					/* 0xf7 (247)  */
    0|MC_NASC,					/* 0xf8 (248)  */
    0|MC_NASC,					/* 0xf9 (249)  */
    0|MC_NASC,					/* 0xfa (250)  */
    0|MC_NASC,					/* 0xfb (251)  */
    0|MC_NASC,					/* 0xfc (252)  */
    0|MC_NASC,					/* 0xfd (253)  */
    0|MC_NASC,					/* 0xfe (254)  */
    0|MC_NASC					/* 0xff (255)  */
};



/*

*//* hdr_name_skip_estr(hstrP, bX)

	Skip over the header name in an ESTR

Accepts :

	hstrP		Ptr to ESTR string containing header
	bX		Start point within ESTR
	

Returns :

	<value>		Index of first char after valid header name

Notes :

	Skips over a string that can be a header name, and returns
	the index to the next character.  This can be used to validate
	a header name syntax or to find the colon after the name, etc.

*/

int
hdr_name_skip_estr ARGLIST( (hstrP, bX) )
   NFARG( ESTR		*hstrP )	/* Ptr to header string */
    FARG( int		bX )		/* Index to start */
{
	int		ch;

    /* Per RFC2822 a header name is any sequence of printable
       ASCII characters except space and colon.  The RFC says
       nothing about a header name beginning with an alpha or
       about disallowing weird characters.  So all we do is check
       for the disallowed characters.
    */
    for ( ; ; ++bX ) {
	ch = estr_bytex( hstrP, bX );	/* Get current char */

	if ( ch == EOF )		/* Check string end */
	    break;

	/* Disallow control characters, space, and non-ascii */
	if ( ( Mchartbl[ch] & (MC_CTL|MC_SPC|MC_NASC) ) != 0 )
	    break;

	/* Colon is also disallowed */
	if ( ch == ':' )
	    break;

	/* Everything else is ok. */
    }

    return ( bX );
}
/*

*//* hdr_name_skip_str(hstrP)

	Skip over the header name in a string

Accepts :

	hstrP		Ptr to string containing header

Returns :

	<value>		Address of first char after valid header name

Notes :

	Uses hdr_name_skip_estr, so see that

*/

UBYTE *
hdr_name_skip_str ARGLIST( (hstrP) )
    FARG( UBYTE		*hstrP )	/* Ptr to header string */
{
	int		bX;

    bX = hdr_name_skip_estr( estr_tmp_str( hstrP ), 0 );
    return ( &hstrP[bX] );
}
/*

*//* mmsg_free( mmsgP )

	Delete a previously-created MMSG object

Accepts :

	mmsgP		PTR to mmsg struct

Returns :

	<value>		TRUE if OK
			FALSE if any part failed

Notes :

	Removes the mmsg and all of its info including messageparts.

	Also clears any pointers just for the sake of it.

	Returns FALSE if freeing any of its parts failed, but completes
	  the process regardless.

*/

BOOL
mmsg_free ARGLIST( (mmsgP) )
    FARG( MMSG		*mmsgP )	/* Ptr to the message */
{
	BOOL		okF;

    okF = TRUE;

    /* Free up the message container. */
TODO    /* Note... what we really want here is to disassociate the mmsg
	   from the container, and have other container controls.  This
	   would generalize the container stuff.  Some day if needed.
	*/
    if ( mmsgP->mm_conP != NULL ) {
	if ( !msgcon_free( mmsgP->mm_conP ) )
	    okF = FALSE;
	mmsgP->mm_conP = NULL;
    }

    /* Free any msgpart(s) */
    if ( mmsgP->mm_topmpP != NULL ) {
	if ( !msgpart_free( mmsgP->mm_topmpP ) )
	    okF = FALSE;
	mmsgP->mm_topmpP = NULL;
    }

    /* Free the thing itself */
    dealloc( mmsgP );

    return ( okF );
}
/*

*//* mmsg_new( conP )

	Allocate and set up a new MMSG message object

Accepts :

	conP		Ptr to the message container object

Returns :

	<value>		Ptr to new MMSG
			NULL if error

Notes :

	The conP is passed for convenience, to point the MMSG to
	  its container.  It may be passed as NULL as it is not
	  otherwise used.

*/

MMSG *
mmsg_new ARGLIST( (conP) )
    FARG( MSGCON	*conP )		/* Ptr to container */
{
	MMSG		*mmsgP;		/* Ptr to what we're making */

    /* Allocate it */
    mmsgP = (MMSG *)emalloc( "msgcon_new", "mmsgP", sizeof(MMSG) );
    if ( mmsgP == NULL )
	return ( FALSE );

    /* Init it */
    mmsgP->mm_conP = conP;		/* Point to the container */
    mmsgP->mm_topmpP = NULL;		/* Dunno top message part yet */
    mmsgP->mm_cachemax = OMM_BODYCACHE_MAX; /* Reader will probably set this */
    mmsgP->mm_cacheC = 0L;		/* How many body bytes are cached */
    mmsgP->mm_oqfromF = FALSE;		/* Writer should set this */
    mmsgP->mm_cnlC = 0;			/* What the hey */

    return ( mmsgP );
}
/*

*//* mmsg_openfp( fP )

	Given a file pointer, open/create a message object

Accepts :

	fP		File pointer

Returns :

	<value>		Ptr to MMSG if successfull
			NULL if not

Notes :

	Given an open file, creates a message container, a message
	  object, and the top message part for that message.

*/

MMSG *
mmsg_openfp ARGLIST( (fP) )
    FARG( FILE		*fP )		/* The file pointer */
{
	MSGCON		*conP;		/* The new container */
	MMSG		*mmsgP;		/* The new message */

    /* Create the container */
    conP = msgcon_new();
    if ( conP == NULL )
	return ( NULL );

    /* Associate the file pointer with the container */
    conP->mc_ifP = fP;

    /* Create the message object */
    mmsgP = mmsg_new( conP );
    if ( mmsgP == NULL ) {
	msgcon_free( conP );
	return ( NULL );
    }

    /* Create the top message part for the message object */
    mmsgP->mm_topmpP = msgpart_new( mmsgP, NULL );
    if ( mmsgP == NULL ) {
	mmsg_free( mmsgP );
	msgcon_free( conP );
    }

    return ( mmsgP );
}
/*

*//* mmsg_openip( inP )

	Given an input stream pointer, open/create a message object

Accepts :

	inP		MML_IN input stack pointer

Returns :

	<value>		Ptr to MMSG if successfull
			NULL if not

Notes :

	Given an open MML_IN input stack, creates a message container,
	  a message object, and the top message part for that message.

*/

MMSG *
mmsg_openip ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* The input stream pointer */
{
	MSGCON		*conP;		/* The new container */
	MMSG		*mmsgP;		/* The new message */

    /* Create the container */
    conP = msgcon_new();
    if ( conP == NULL )
	return ( NULL );

    /* Associate the stream pointer with the container */
    conP->mc_inP = inP;

    /* Create the message object */
    mmsgP = mmsg_new( conP );
    if ( mmsgP == NULL ) {
	msgcon_free( conP );
	return ( NULL );
    }

    /* Create the top message part for the message object */
    mmsgP->mm_topmpP = msgpart_new( mmsgP, NULL );
    if ( mmsgP == NULL ) {
	mmsg_free( mmsgP );
	msgcon_free( conP );
    }

    return ( mmsgP );
}
/*

*//* msgcon_free( conP )

	Free a message container object

Accepts :

	conP		Ptr to the message container

Returns :

	<value>		TRUE if OK
			FALSE if any part fails

Notes :

	Frees a message container object. 

	Also clears any pointers just for the sake of it.

	Returns FALSE if any step fails, but completes them all.

*/

BOOL
msgcon_free ARGLIST( (conP) )
    FARG( MSGCON	*conP )		/* The message container */
{
	BOOL		okF;

    okF = TRUE;

    /* If any input file, close it */
    if ( conP->mc_ifP != NULL ) {
	fclose( conP->mc_ifP );
	conP->mc_ifP = NULL;
    }

    /* If any input stack, remove it */
    if ( conP->mc_inP != NULL ) {
	in_destroy( conP->mc_inP );
	conP->mc_inP = NULL;
    }

    /* If any tmpfP, close it.  Note that it was already unlinked when
       opened.
    */
    if ( conP->mc_tmpfP != NULL ) {
	fclose( conP->mc_tmpfP );
	conP->mc_tmpfP = NULL;
    }

    /* Clear this; if non, NULL, it's a copy of either mc_ifP or mc_tmpfP */
    conP->mc_usefP = NULL;

    if ( conP->mc_tmpfnameP != NULL ) {
	dealloc( conP->mc_tmpfnameP );
	conP->mc_tmpfnameP = NULL;
    }

    /* Clear the thing itself */
    dealloc( conP );

    return ( okF );
}
/*

*//* msgcon_new()

	Allocate and set up a new MSGCON message container

Accepts :

	<none>

Returns :

	<value>		Ptr to new MSGCON
			NULL if error

Notes :

	Does not open any files; that's left to the caller.

*/

MSGCON *
msgcon_new NOARGLIST
{
	MSGCON		*conP;		/* New MSGCON we're building */

    /* Allocate it */
    conP = (MSGCON *)emalloc( "msgcon_new", "conP", sizeof(MSGCON) );
    if ( conP == NULL )
	return ( FALSE );

    /* Init it */
    conP->mc_ifP = NULL;		/* No input file */
    conP->mc_inP = NULL;		/* No input stream */
    conP->mc_tmpfP = NULL;		/* No temp file */
    conP->mc_usefP = NULL;		/* Not using either file */
    conP->mc_tmpfnameP = NULL;		/* No tempname */
    conP->mc_uncachedF = FALSE;		/* Not set up for non-cached access */
    conP->mc_prereadC = -1;		/* No pre-read bytes available */

    return ( conP );
}
/*

*//* msgcon_readline (conP, rctl)

	Read a source line from the message container

Accepts :

	conP		Ptr to message container
	rctl		Control flagword to pass to getbtrec()

Returns :

	<value>		>= 0: Number of characters read,
			  -1: EOF

Notes :

	This is a go-between function that mainly calls the appropriate
	<x>getbtrec() on behalf of the caller, depending on whether the
	input is from a file or from an MML_IN stream, reading into the
	Gbs1P buffer.  However, it also pays attention to the
	mc_prereadC; if that value is non-negative, it's assumed that
	the Gbsp buffer has a record that should be used instead of
	reading from the container file, and mc_prereadC is the count
	that should be returned.

*/

int
msgcon_readline ARGLIST( (conP, rctl) )
   NFARG( MSGCON	*conP )		/* Ptr to message container */
    FARG( UWORD		rctl )		/* flags for getbtrec() routines */
{
	int		len;

    /* If anything was preread, don't read, but re-use the buffer */
    if ( conP->mc_prereadC >= 0 ) {
	len = conP->mc_prereadC;
	conP->mc_prereadC = -1;
    }
    else {
	/* Read from either file or stream */
	if ( conP->mc_ifP != NULL )
	    len = fgetbtrec( conP->mc_ifP, Gbs1P, rctl, NULL );
	else
	    len = igetbtrec( conP->mc_inP, Gbs1P, rctl, NULL );
    }

    return ( len );
}
/*

*//* msgcon_readlinea (conP, rctl)

	Append a source line from the message container

Accepts :

	conP		Ptr to message container
	rctl		Control flagword to pass to getbtrec()

Returns :

	<value>		>= 0: Number of characters read,
			  -1: EOF

Notes :

	Appends a line from the input to the Gbs1P buffer, sort of
	like msgcon_readline().
	Mainly just decides whether to read from file or MML_IN stream.

	Unlike msgcon_readline, this function doesn't pay attention
	to the preread flag.

*/

int
msgcon_readlinea ARGLIST( (conP, rctl) )
   NFARG( MSGCON	*conP )		/* Ptr to message container */
    FARG( UWORD		rctl )		/* flags for getbtrec() routines */
{
	int		len;

    /* Read from either file or stream */
    if ( conP->mc_ifP != NULL )
	len = fgetbtreca( conP->mc_ifP, Gbs1P, rctl, NULL );
    else
	len = igetbtreca( conP->mc_inP, Gbs1P, rctl, NULL );

    return ( len );
}
/*

*//* msgcon_uncache( conP )

	Setup a message container for uncached content

Accepts :

	conP		Ptr to the message container

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	msgpart code will attempt to cache as much of a message as
	possible in memory.  If there are portions that can't be
	cached (e.g. a limit has been reached), then the non-cached
	portions must be kept in a file.  If the container file is
	seekable, the cached stuff is just kept in the container.
	Otherwise, we cache to a tempfile.

	It's assumed that SOME content has been cached, and that the
	container file pointer has been advanced far enough that a
	seek back to zero will tell us whether or not we can seek
	into the file.  We check for this explicitly, and will
	use a tempfile if that's not true.

*/

BOOL
msgcon_uncache ARGLIST( (conP) )
    FARG( MSGCON	*conP )		/* The message container */
{
	BOOL		seekableF;	/* If seekable */
	int		tmpfd;		/* Tempfd */
	off_t		curpos;		/* seek position on entry */
	char		tmpfname[ PATH_MAX ];  /* Tempfile name */
	
    if ( conP->mc_uncachedF )
	return ( TRUE );		/* Already prepped... */

    /* Seek only works if the input is a file.  If the input is
       an MML_IN stream, we have to use a tempfile.
    */
    if ( conP->mc_ifP != NULL ) {
	/* Let's assume we can seek */
	seekableF = TRUE;

	/* Note the file position before we muck with it */
	curpos = ftell( conP->mc_ifP );

	/* Now let's find out if we can seek.   First try going back to
           the zero point.
	*/
	if ( ( fseek( conP->mc_ifP, 0L, SEEK_SET ) != 0 ) ||
             ( ftell( conP->mc_ifP ) != 0 ) )
	    seekableF = FALSE;
	else {
	    /* We were able to rewind; now go back to where we were.
	       If this fails, we've lost the input...
	    */
	    if ( ( fseek( conP->mc_ifP, curpos, SEEK_SET ) != 0 ) ||
                 ( ftell( conP->mc_ifP ) != curpos ) ) {
		warning( "msgcon_uncache: file position lost." );
		return ( FALSE );
	    }
	}

	/* If the entry (current) position is too small (by some
	   arbitrary choice) we'll declare the file non-seekable.  This
	   tries to catch the case where the caller hasn't really
	   attempted to cache a lot before calling this routine, and
	   where our previous test(s) might be illusory because of
	   stdio or kernel buffering.
	*/
	if ( curpos < ( 50L * 1024) )
	    seekableF = FALSE;

	/* If it's seekable, we're OK */
	if ( seekableF ) {
	    conP->mc_usefP = conP->mc_ifP;
	    conP->mc_uncachedF = TRUE;	    /* Remember we did this. */
	    return ( TRUE );
	}
    }

    /* A temp file is needed for cacheing. */
    if ( snprintf( &tmpfname[0], PATH_MAX, "%s/mvmf.uncache.XXXXXXXX",
		        TmpdirP ) >= PATH_MAX ) {
	warning( "msgcon_uncache: tmpdir name \"%s\" is too long.",
			TmpdirP );
	return ( FALSE );
    }

    tmpfd = mkstemp( &tmpfname[0] );
    if ( tmpfd < 0 ) {
	warning( "msgcon_uncache: error (%d) making tempfile.", errno );
	return ( FALSE );
    }

    /* Remember the name */
    conP->mc_tmpfnameP = newstr( "msgcon_uncache", "temp name", 
					&tmpfname[0] );

    /* Unlink the file now so it goes away when we are done with it */
    unlink( &tmpfname[0] );

    if ( ( conP->mc_tmpfP = fdopen( tmpfd, "w+" ) ) == NULL ) {
	warning( "msgcon_uncache: fdopen failure, errno %d", errno );
	close( tmpfd );
	return ( FALSE );
    }

    conP->mc_usefP = conP->mc_tmpfP;

    conP->mc_uncachedF = TRUE;	/* All prepped */

    return ( TRUE );
}
/*

*//* msgcon_uncache_data( conP, byteP, byteC, newlineF )

	Make data available as uncached data

Accepts :

	conP		Message container
	byteP		Ptr to buffer of data
	byteC		Number of bytes of data
	newlineF	If a newline should be added to the data

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	This makes sure that some data that has been read from a message
	container is still available if it isn't cached.  This means
	either:

	  - the message container input is seekable, so the data is
	    implicitly available;

	  - the message container has a temporary file for uncached
	    data, so we store the data there.

*/

BOOL
msgcon_uncache_data ARGLIST( (conP, byteP, byteC, newlineF) )
   NFARG( MSGCON	*conP )		/* Message container */
   NFARG( UBYTE		*byteP )	/* Ptr to data */
   NFARG( int		byteC )		/* Number of bytes */
    FARG( BOOL		newlineF )	/* Whether to add a newline */
{
	BOOL		okF;

    /* Only add to the file if there's a tempfile. */
    if ( conP->mc_tmpfP == NULL )
	return ( TRUE );

    okF = ( ( byteC == 0 ) ||
	       ( fwrite( byteP, 1, byteC, conP->mc_tmpfP ) == byteC ) );

    if ( okF && newlineF )
	okF = ( fputc( '\n', conP->mc_tmpfP ) != EOF );

    if ( !okF )
	warning( "msgcon_uncache_data: error (%d) writing to tmpfile.",
		    errno );

    return ( okF );
}
/*

*//* msgpart_add( mpP, childP, endF )

	Add a new child messagepart to a parent

Accepts :

	mpP		Messagepart to add a child to
	childP		The child messagepart
	endF		TRUE to add the messagepart to the end (after any
			  other existing child messageparts);
			  otherwise add to the beginning (before any
			  other child messageparts).

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgpart_add ARGLIST( (mpP, childP, endF) )
   NFARG( MSGPART	*mpP )		/* Messagepart to add to */
   NFARG( MSGPART	*childP )	/* Messagepart to add */
    FARG( BOOL		endF )		/* TRUE to add to the end */
{
	BOOL		okF;

    /* Add the child into the parent */
    if ( endF ) {
	okF = ( ll_addvt( &mpP->mp_parts, (void *)childP ) != NULL );
	if ( okF )
	    childP->mp_slnodeP = mpP->mp_parts.llh_lastP;
    }
    else {
	okF = ( ll_addvh( &mpP->mp_parts, (void *)childP ) != NULL );
	if ( okF )
	    childP->mp_slnodeP = mpP->mp_parts.llh_firstP;
    }

    return ( okF );
}
/*

*//* msgpart_date_string(timeP)

	Format a date/time string for use in message parts

Accepts :

	timeP		Ptr to the date/time to print, a time_t value

Returns :

	<value>		Ptr to a formatted date string

Notes :

	The returned string is a temporary buffer; caller should make
	quick use of the result.

	At present the format is fixed to something like this:

	    Sun, 2 Oct 2005 13:03:29 -0400

	The addition of timezone name is inhibited, e.g. not this:
	    Sun, 2 Oct 2005 13:03:29 -0400 (EDT)

*/

char *
msgpart_date_string ARGLIST( (timeP) )
    FARG( time_t	*timeP )	/* Ptr to the time value */
{
	int		i;
	int		tzoff, tzsign;
	char		*strP;
	struct tm	*tmP;

    strP = tmpstr( 50 );		/* Should only need 31 */

    /* Decode in local time */
    tmP = localtime( timeP );

    tzoff = tmP->tm_gmtoff / 3600;
    if ( tzoff < 0 ) {
	tzsign = '-';
	tzoff = -tzoff;
    }
    else
	tzsign = '+';

    i = snprintf( strP, 50,
		 "%3.3s, %d %3.3s %4d %02d:%02d:%02d %c%04d",
		 Daynames[ tmP->tm_wday ],
		 tmP->tm_mday,
		 Monthnames[ tmP->tm_mon],
		 tmP->tm_year + 1900,
		 tmP->tm_hour, tmP->tm_min, tmP->tm_sec,
		 tzsign, tzoff
		 );

    if ( i >= 50 ) {
	strP[50] = NUL;
	i = 50;
    }

#if 0
    /* If we wanted to append a timezone name we'd do it here. */
    if ( ( tmP->tm_zone != NULL ) &&
         ( ( (50 - i) - strlen( tmP->tm_zone ) ) > 4 ) ) {
	sprintf( &strP[i], " (%s)", tmP->tm_zone );
	i += strlen( &strP[i] );
    }
#endif

    return ( strP );
}
/*

*//* msgpart_date_string_now()

	Return msgpart_date_string() for the current date/time

Accepts :

	<nothing>

Returns :

	<value>		date string corresponding to the current date/time

Notes :

	Just hands the time now to msgpart_date_string.  See that.

*/

char *
msgpart_date_string_now NOARGLIST
{
	time_t		timenow;

    time( &timenow );
    return ( msgpart_date_string( &timenow ) );
}
/*

*//* msgpart_free( mpP )

	Free a messagepart and all its owned data

Accepts :

	mpP		Ptr to messagepart

Returns :

	<value>		TRUE if OK
			FALSE if any part failed

Notes :

	Removes the msgpart and any of its child messageparts

	Also clears any pointers just for kicks.

	Returns FALSE if freeing any of its parts failed, but completes
	  the process regardless.

*/

BOOL
msgpart_free ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Ptr to messagepart */
{
	BOOL		okF;
	LL_NODE		*nodeP;		/* Linked list node */

    okF = TRUE;

    /* Free headers */
    while ( ( nodeP = mpP->mp_hdrs.llh_firstP ) != NULL )
	msgpart_hdr_del( mpP, nodeP );

    /* Free MIME info */
    if ( !mimeinfo_free( &mpP->mp_mime, FALSE ) )
	okF = FALSE;

    /* Free body part strings */
    if ( !msgpstr_free( &mpP->mp_body, FALSE ) )
	okF = FALSE;
    if ( !msgpstr_free( &mpP->mp_epilog, FALSE ) )
	okF = FALSE;

    /* Free any children */
    while ( ( nodeP = mpP->mp_parts.llh_firstP ) != NULL ) {
	msgpart_free( (MSGPART *)nodeP->lln_valP );
	nodeP->lln_valP = NULL;
	ll_rem( nodeP );
    }

    /* Free the messagepart object itself */
    dealloc( mpP );

    return ( okF );
}
/*

*//* msgpart_hdr_add( mpP, hstrP, beforeP )

	Add a new header string

Accepts :

	mpP		Ptr to MSGPART
	hstrP		Ptr to new header string ESTR
	beforeP		Linked list node before which to add the header

Returns :

	<value>		TRUE if success
			FALSE if not

Notes :

	If beforeP is NULL, the node will be added at the end.

	The caller must have locked the header's linked list before
	  calling this function, so that the "before" node remains
	  valid and accessible.  (Only if threaded, but do it anyway.)

*/

BOOL
msgpart_hdr_add ARGLIST( (mpP, hstrP, beforeP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( ESTR		*hstrP )	/* New header string */
    FARG( LL_NODE	*beforeP )	/* Node to add before */
{
	int		len;
	int		bX;
	MPHL		*hlP;		/* Header line */
	BSTR		*bsP;		/* Header's underlying BSTR */

    /* Find point after header name (and validate the name) */
    bX = hdr_name_skip_estr( hstrP, 0 );
    if ( bX == 0 ) {
	warning( "msgpart_hdr_add: invalid header \"%s\"", tmpstr_estr( hstrP ) );
	return ( FALSE );
    }

    /* Supposedly whitespace can occur after the keyword but feh,
       don't allow it.  Colon must occur here.
    */
    if ( estr_bytex( hstrP, bX ) != ':' ) {
	warning( "msgpart_hdr_add: no colon after header name." );
	return ( FALSE );
    }

    /* Allocate header structure */
    hlP = (MPHL *)emalloc( "msgpart_hdr_add", "hlP", sizeof(MPHL) );
    if ( hlP == NULL )
	return ( FALSE );

    /* Create the refstr for the string */
    len = estr_len( hstrP );
    hlP->mh_hP = refstr_new( "msgpart_hdr_add header string", len+1, 0 );
    if ( hlP->mh_hP == NULL ) {
	dealloc( hlP );
	return ( FALSE );
    }

    /* Copy string and suppress any trailing newline(s) */
    bsP = refstr_bstr( hlP->mh_hP );
    bstr_catn( bsP, hstrP->es_strP, len );
    while ( ( bsP->bs_bC > 0 ) && ( bsP->bs_bP[ bsP->bs_bC -1 ] == '\n' ) )
	--bsP->bs_bC;
    bstr_nulterm( bsP );

    /* Add to the list in the proper spot */
    if ( ll_addvb( &mpP->mp_hdrs, beforeP, (void *)hlP ) == NULL ) {
	refstr_free( hlP->mh_hP );
	dealloc( hlP );
	return ( FALSE );
    }

    /* Update the accumulated approximate size, add 1 for newline */
    mpP->mp_hdrsize += ( bsP->bs_bC + 1 );

    return ( TRUE );
}
/*

*//* msgpart_hdr_add2( mpP, hnameP, hvalP, beforeP )

	Add a new header string from components

Accepts :

	mpP		Ptr to MSGPART header
	hnameP		Ptr to new header name
	hvalP		Ptr to new header value
	beforeP		Linked list node before which to add the header

Returns :

	<value>		TRUE if success
			FALSE if not

Notes :

	hnameP should NOT have the trailing colon, we add it here

	If beforeP is NULL, the node will be added at the end.

	The caller must have locked the header's linked list before
	  calling this function, so that the "before" node remains
	  valid and accessible.  (Only if threaded, but do it anyway.)

*/

BOOL
msgpart_hdr_add2 ARGLIST( (mpP, hnameP, hvalP, beforeP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( ESTR		*hnameP )	/* New header name */
   NFARG( ESTR		*hvalP )	/* New header value */
    FARG( LL_NODE	*beforeP )	/* Node to add before */
{
	int		len;
	int		bX;
	MPHL		*hlP;		/* Header line */
	BSTR		*bsP;		/* Header's underlying BSTR */

    /* Validate the name */
    bX = hdr_name_skip_estr( hnameP, 0 );
    if ( ( bX == 0 ) || ( estr_bytex( hnameP, bX ) != EOF ) ) {
	warning( "msgpart_hdr_add2: bad header name" );
	return ( FALSE );
    }

    /* Allocate header structure */
    hlP = (MPHL *)emalloc( "msgpart_hdr_add2", "hlP", sizeof(MPHL) );
    if ( hlP == NULL )
	return ( FALSE );

    /* Create the refstr for the string */
    len = estr_len( hnameP ) + estr_len( hvalP );
    hlP->mh_hP = refstr_new( "msgpart_hdr_add header string", len+2, 0 );
    if ( hlP->mh_hP == NULL ) {
	dealloc( hlP );
	return ( FALSE );
    }

    /* Insert header name and value, suppress any trailing newline(s) */
    bsP = refstr_bstr( hlP->mh_hP );
    bstr_cate( bsP, hnameP, 0 );
    bstr_cat( bsP, ": " );
    bstr_cate( bsP, hvalP, 0 );
    while ( ( bsP->bs_bC > 0 ) && ( bsP->bs_bP[ bsP->bs_bC -1 ] == '\n' ) )
	--bsP->bs_bC;
    bstr_nulterm( bsP );

    /* Add to the list in the proper spot */
    if ( ll_addvb( &mpP->mp_hdrs, beforeP, (void *)hlP ) == NULL ) {
	refstr_free( hlP->mh_hP );
	dealloc( hlP );
	return ( FALSE );
    }

    /* Update the accumulated approximate size (add 1 for newline) */
    mpP->mp_hdrsize += ( bsP->bs_bC +1 );

    return ( TRUE );
}
/*

*//* msgpart_hdr_addr( mpP, hdrP, parts )

	Get next address part(s) from header string

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		Address of a header object previously returned
			  e.g. from msgpart_hdr_next
	parts		bitmask specifying which header part(s) to return

Returns :

	<value>		Pointer to an internal string buffer containing
			  the selected address parts
			NULL if no more addresses in this header

Notes :

	Simply calls msgpart_refstr_addr, so see the notes there.

*/

char *
msgpart_hdr_addr ARGLIST( (mpP, hdrP, parts) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node if any */
    FARG( UWORD		parts )		/* Selected parts */
{
	REFSTR		*rsP;		/* REFSTR for header */

    if ( ( rsP = msgpart_hdr_str( mpP, hdrP ) ) == NULL )
	return ( NULL );

    return ( msgpart_refstr_addr( mpP, rsP, parts ) );
}
/*

*//* msgpart_hdr_change( mpP, hdrP, newnameP, newvalP )

	Change parts of a header

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		Address of a header object previously returned
			  e.g. from msgpart_hdr_next
	newnameP	New header name (tag, as in "From:" but
			  without the colon)
	newvalP		New header value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Either newnameP or newvalP may be passed as NULL.

*/

BOOL
msgpart_hdr_change ARGLIST( (mpP, hdrP, newnameP, newvalP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node */
   NFARG( ESTR		*newnameP )	/* New name (or NULL) */
    FARG( ESTR		*newvalP )	/* New value */
{
	int		bodyX;		/* Where header body starts */
	int		nnX;		/* Index after new name name */
	int		oldbC;		/* Header byte count before change */
	BOOL		okF;
	REFSTR		*rsP;		/* Ptr to header REFSTR */
	BSTR		*bsP;		/* Ptr to header BSTR */
	UBYTE		*hstrP;		/* Header string itself */
	UBYTE		*hbodyP;	/* Header body(after name) */

    /* Get the header object string */
    if ( ( rsP = msgpart_hdr_str( mpP, hdrP ) ) == NULL )
	return ( FALSE );
    bsP = refstr_bstr( rsP );
    hstrP = bsP->bs_bP;
    oldbC = bsP->bs_bC;

    if ( newnameP != NULL ) {
	/* Find the length of the new name.  Caller isn't supposed to supply
           a colon but we're going to make sure.
	*/
	nnX = hdr_name_skip_estr( newnameP, 0 );
	if ( ( nnX == 0 ) || ( estr_bytex( newnameP, nnX ) != EOF ) ) {
	    warning( "msgpart_hdr_change: bad header name" );
	    return ( FALSE );
	}

	/* If we're replacing both parts, just store them without the
           replacement hassle.
	*/
	if ( newvalP != NULL ) {
	    bsP->bs_bC = 0;

	    /* Store name plus colon and space */
	    okF = bstr_cate( bsP, newnameP, 0 );
	    okF = okF && bstr_catn( bsP, (UBYTE *)": ", 2 );
	    okF = okF && bstr_cate( bsP, newvalP, 0 );
	    bstr_nulterm( bsP );

	    /* Update the approximate header count */
	    mpP->mp_hdrsize += (bsP->bs_bC - oldbC);

	    return ( okF );
	}
    }

    /* Find the point after the old name */
    hbodyP = hdr_name_skip_str( hstrP );
    if ( *hbodyP++ != ':' ) {
	warning( "no colon in original header in msgpart_hdr_change." );
	return ( FALSE );
    }
    bodyX = hbodyP - hstrP;

    /* Make sure there is whitespace following the colon. */
    if ( ( hstrP[ bodyX ] != ' ' ) &&
         ( hstrP[ bodyX ] != '\t' ) ) {
	if ( refstr_replacen( rsP, bodyX, bodyX, (UBYTE *)" ", 1 ) )
	    return ( FALSE );
	hstrP = bsP->bs_bP;		/* In case it changed */
    }

    rsP->rs_bX = bodyX;			/* Side effect */

    /* Replace name if desired */
    if ( newnameP != NULL )
	okF = refstr_replacee( rsP, 0, bodyX - 1, newnameP );
    else if ( newvalP != NULL )
	okF = refstr_replacee( rsP, bodyX +1, bsP->bs_bC, newvalP );
    else
	okF = TRUE;

    /* Update the header size, regardless of status */
    mpP->mp_hdrsize += (bsP->bs_bC - oldbC);

    return ( okF );
}
/*

*//* msgpart_hdr_del( mpP, hdrP )

	Delete a specific header

Accepts :

	mpP		Ptr to messagepart
	hdrP		Ptr to the header to delete

Returns :

	<value>		address of prior node if success
			(NULL indicates no previous node)

Notes :


*/

LL_NODE *
msgpart_hdr_del ARGLIST( (mpP, hdrP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
    FARG( LL_NODE	*hdrP )		/* Ptr to header node to delete */
{
	int		bC;		/* Number of bytes removing */
	LL_NODE		*prevP;		/* Ptr to previous node */
	MPHL		*hlP;		/* Header line */

    /* Get rid of the stuff in this node */
    hlP = (MPHL *)hdrP->lln_valP;
    bC = refstr_bstr( hlP->mh_hP )->bs_bC;
    refstr_free( hlP->mh_hP );
    hlP->mh_hP = NULL;			/* why not. */
    hdrP->lln_valP = NULL;
    dealloc( hlP );

    /* Find prior node to return */
    prevP = hdrP->lln_prevP;

    /* Remove this from the linked list */
    ll_rem( hdrP );

    /* Update the header size (include newline) */
    mpP->mp_hdrsize -= ( bC +1 );

    /* Return previous header */
    return ( prevP );
}
/*

*//* msgpart_hdr_next( mpP, hdrP, hnameP )

	Find the [next] [matching] header

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		If NULL, start at the beginning;
			otherwise start at the header following this one.
	hnameP		If non-NULL, only consider headers with this
			  name (case-insensitive match).

Returns :

	<value>		ptr to next opaque header object if success,
			NULL if no more headers matching.

Notes :

	Sets the BSTR fetch ptr (bs_bX) to the first non-blank
	character of the header argument (i.e., text of header after
	the colon).

*/

LL_NODE *
msgpart_hdr_next ARGLIST( (mpP, hdrP, hnameP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node if any */
    FARG( ESTR		*hnameP )	/* Any initial matching string */
{
    /* Just hand off to msgpart_hdr_step() */
    return ( msgpart_hdr_step( mpP, hdrP, hnameP, TRUE ) );
}
/*

*//* msgpart_hdr_parts( mpP, hdrP, namePP, tailPP )

	Constructs REFSTRs for header name and/or body

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		Ptr to the header
	namePP		Where to put REFSTR ptr for header name
	tailPP		Where to put REFSTR ptr for header tail

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Creates new REFSTRs so should be used only where needed.

	namePP and/or tailPP may be passed as NULL for no-op on
	that part.

*/

BOOL
msgpart_hdr_parts ARGLIST( (mpP, hdrP, namePP, tailPP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node */
   NFARG( REFSTR	**namePP )	/* Where to put REFSTR for name */
    FARG( REFSTR	**tailPP )	/* Where to put REFSTR for tail */
{
	UBYTE		ch;
	int		cX;
	UBYTE		*strP;
	REFSTR		*rsP;		/* REFSTR for header string */
	BSTR		*bsP;		/* BSTR corresponding to rsP */

    rsP = msgpart_hdr_str( mpP, hdrP );
    bsP = refstr_bstr( rsP );

    strP = bsP->bs_bP;
    cX = (hdr_name_skip_str( strP ) - strP ) +1;

    if ( namePP != NULL ) {
	*namePP = refstr_ref( rsP, 0, cX );
    }
    if ( tailPP != NULL ) {
	/* Skip whitespace after name */
	for ( ; ( ch = strP[cX] ) != NUL; ++cX ) {
	    if ( ( ch != ' ' ) &&
	         ( ch != '\t' ) &&
		 ( ch != '\n' ) )
		break;
	}
	*tailPP = refstr_ref( rsP, cX, refstr_len( rsP ) );
    }

    return ( TRUE );
}
/*

*//* msgpart_hdr_prev( mpP, hdrP, hnameP )

	Find the [previous] [matching] header

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		If NULL, start at the end;
			otherwise start at the header preceding this one.
	hnameP		If non-NULL, only consider headers with this
			  name (case-insensitive match).

Returns :

	<value>		ptr to next previous header object if success,
			NULL if no prior headers matching.

Notes :

	Sets the BSTR fetch ptr (bs_bX) to the first non-blank
	character of the header argument (i.e., text of header after
	the colon).

*/

LL_NODE *
msgpart_hdr_prev ARGLIST( (mpP, hdrP, hnameP) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node if any */
    FARG( ESTR		*hnameP )	/* Any initial matching string */
{
    /* Just hand off to msgpart_hdr_step() */
    return ( msgpart_hdr_step( mpP, hdrP, hnameP, FALSE ) );
}
/*

*//* msgpart_hdr_step( mpP, hdrP, hnameP, forwardF )

	Find the [adjacent] [matching] header

Accepts :

	mpP		Ptr to MSGPART header
	hdrP		If NULL, start at the beginning (or end, if
			  forwardF is false)
			otherwise start at the header adjacent to this one.
	hnameP		If non-NULL, only consider headers with this
			  name (case-insensitive match).
	forwardF	TRUE to step forward, non-TRUE to step backwards.

Returns :

	<value>		ptr to opaque header object if success,
			NULL if no more headers matching.

Notes :

	Sets the BSTR fetch ptr (bs_bX) to the first non-blank
	character of the header argument (i.e., text of header after
	the colon).

*/

LL_NODE *
msgpart_hdr_step ARGLIST( (mpP, hdrP, hnameP, forwardF) )
   NFARG( MSGPART	*mpP )		/* MSGPART header ptr */
   NFARG( LL_NODE	*hdrP )		/* Ptr to header node if any */
   NFARG( ESTR		*hnameP )	/* Any initial matching string */
    FARG( BOOL		forwardF )	/* TRUE to step forward */
{
	UBYTE		ch;
	int		hnameL;		/* Length of header name if given */
	UBYTE		*hbodyP;	/* Ptr to header body */
	LL_NODE		*nodeP;		/* Ptr to list node */
	LL_NODE		**nodePP;	/* Ptr to node pointer */
	REFSTR		*rsP;		/* Reference string for header */
	BSTR		*bsP;		/* Corresponding bstr */

    /* Get start point */
    if ( hdrP != NULL )
	nodePP = forwardF ? &(hdrP->lln_nextP) :
			    &(hdrP->lln_prevP);
    else
	nodePP = forwardF ? &(mpP->mp_hdrs.llh_firstP) :
			    &(mpP->mp_hdrs.llh_lastP);

    if ( hnameP != NULL )
	hnameL = estr_len( hnameP );
    else
	hnameL = 0;			/* We'll want this to be zero later */

    /* Go through the rest of the list */
    for ( ; ( nodeP = *nodePP ) != NULL;
	     nodePP = forwardF ? &(nodeP->lln_nextP) : &(nodeP->lln_prevP) ) {

	/* Access the header REFSTR and underlying string */
	rsP = msgpart_hdr_str( mpP, nodeP );
	bsP = refstr_bstr( rsP );

	/* See if this header is one that the caller wants (desired
	   header name must be NULL or must match the actual header)
	*/
	if ( ( hnameP == NULL ) ||
	     ( ( bsP->bs_bC >= hnameL ) &&
	       ( strnicmp( (char *)hnameP->es_strP,
	               (char *)bsP->bs_bP, hnameL ) == 0 ) ) ) {

	    /* Yep, skip to the start of the header body */
	    hbodyP = bsP->bs_bP + hnameL;
	    if ( hnameP == NULL ) {
		/* No header name specified... skip to the colon */
		while ( ( ch = *hbodyP ) != NUL ) {
		    ++hbodyP;
		    if ( ch == ':' )
			break;
		}
	    }
	    else {
		/* Header name was given, Verify that we are at the
		   end of the header name, and skip over the colon.
		*/
		if ( *hbodyP == ':' )	/* Should be true if exact match */
		    ++hbodyP;
		else
		    continue;		/* Not exact match, try next header. */
	    }

	    /* Skip whitespace following header name */
	    for ( ; ( ch = *hbodyP ) != NUL; ++hbodyP ) {
		if ( ( ch != ' ' ) &&
		     ( ch != '\t' ) &&
		     ( ch != '\n' ) )
		    break;
	    }

	    rsP->rs_bX = hbodyP - bsP->bs_bP;
	    break;
	}
    }

    return ( nodeP );
}
/*

*//* msgpart_hdr_str( mpP, hdrP )

	Get raw string from header

Accepts:

	mpP		Ptr to message part
	hdrP		Ptr to header element

Returns :

	<value>		Ptr to the REFSTR for the header node


Notes :

	mpP isn't currently used; it's specified in case we ever
	want to make use of it.

*/


REFSTR *
msgpart_hdr_str ARGLIST( (mpP, hdrP) )
   NFARG( MSGPART	*mpP )		/* Ptr to MSGPART header */
    FARG( LL_NODE	*hdrP )		/* Ptr to node for header line */
{
    return ( ((MPHL *)hdrP->lln_valP)->mh_hP );
}
/*

*//* msgpart_hdr_tail( mpP, hdrP )

	Get header tail string, encapsulated as an ESTR

Accepts:

	mpP		Ptr to message part
	hdrP		Ptr to header node

Returns :

	<value>		Ptr to the ESTR for the header tail

Notes:

	The tail of the header is the part starting at the current
	access point, though the end of the buffer.  e.g. if
	this is applied to the result of msgpart_hdr_next, then the
	tail is the part after the header name plus any whitespace.

	Note that as with any ESTR, the returned value is just a
	snapshot at the time of the call.  The ESTR must be released
	before any changes are made to the underlying REFSTR.
	(Or at least before making an references after such a change.)

*/


ESTR *
msgpart_hdr_tail ARGLIST( (mpP, hdrP) )
   NFARG( MSGPART	*mpP )		/* Ptr to MSGPART header */
    FARG( LL_NODE	*hdrP )		/* Ptr to node for header line */
{
	int		rbX;		/* Normalized copy of rs_bX */
	REFSTR		*rsP;		/* The header */
	ESTR		*esP;		/* Returned value */

    rsP = msgpart_hdr_str( mpP, hdrP );	/* Get header */
    rbX = rsP->rs_bX;			/* Get current access point */
    if ( rbX < 0 )
	rbX = 0;

    /* Formulate ESTR from the tail */
    esP = estr_nstr( refstr_bstr( rsP )->bs_bP + rbX,
		     refstr_len( rsP ) - rbX );

    return ( esP );
}
/*

*//* msgpart_hdrchar(rsP, flags)

	Get the next character from a header string

Accepts :

	rsP		Ptr to REFSTR containing the header
	flags		flag bits to apply to the process (HC_XXXX)

Returns :

	<value>		next byte, or
			EOF if no more string

Notes :

	The rsP has been set up for character fetching..


*/

int
msgpart_hdrchar ARGLIST( (rsP, flags) )
   NFARG( REFSTR	*rsP )		/* REFSTR for header */
    FARG( UWORD		flags )		/* flags for the operation */
{
	int		ch;
	int		parenC;		/* Paren nesting level */

    for ( ; ( ch = refstr_getb( rsP ) ) != EOF; ) {

	/* Skip any parenthesized comments if the flags say to.  Note
	    that this may catch comments where they shouldn't be,
	    syntactically, but we won't worry about that here, this is
	    just a dumb routine.
	*/
	if ( ( ch == '(' ) && ( ( flags & HC_PAREN ) != 0 ) ) {
	    /* Eat up characters until the closing paren */
	    parenC = 1;
	    while ( ( ch = refstr_getb( rsP ) ) != EOF ) {
		if ( ch == '\\' ) {	/* Some literal char (even paren..) */
		    ch = refstr_getb( rsP );
		    continue;
		}

		if ( ch == '(' )
		    ++parenC;
		else if ( ch == ')' ) {
		    if ( --parenC == 0 )
			break;
		}
	    }
	    continue;			/* Done with parenthesized area */
	}

	/* Skip any quoted string.  Ditto on the dumbness. */
	if ( ( ch == '"' ) && ( ( flags & HC_QUOTED ) != 0 ) ) {
	    /* Eat up characters until the closing quote */
	    while ( ( ch = refstr_getb( rsP ) ) != EOF ) {
		if ( ch == '\\' ) {	/* Some literal char, take it */
		    ch = refstr_getb( rsP );
		    continue;
		}

		if ( ch == '"' )
		    break;
	    }
	    continue;			/* Done with quoted string */
	}

	/* Must be OK. */
	break;
    }

    return ( ch );
}
/*

*//* msgpart_hdroptpair( rsP, optP, optL, valP, valL, semireqF )

	Collect any option=value pair from a header string

Accepts :

	rsP		REFSTR containing the header string
	optP		Pointer to buffer for the option name
	optL		Size of the optP buffer
	valP		Pointer to buffer for the option value
	valL		Size of the valP buffer
	semireqF	TRUE if an opening semicolon is required.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Many headers are defined to end with any number of option=value
	pairs, each preceded with a semicolon.  So for example, in:

	    Content-type:  multipart/alternative; boundary="blat"

	the sequence
		; boundary="blat"

	is the option pair sequence (semicolon, option, =, value)

	This function extracts the option and the value and returns
	it to the caller.  Caller specifies semireqF as TRUE if the
	leading semicolon is required.  If passed as FALSE, the semi
	is optional.  (doubtful that it's ever optional, but what
	the heck..)

	We return FALSE if any syntax element is missing.  The token
	state will be restored to its input position in this case.

*/

BOOL
msgpart_hdroptpair ARGLIST( (rsP, optP, optL, valP, valL, semireqF) )
   NFARG( REFSTR	*rsP )		/* REFSTR for header */
   NFARG( char		*optP )		/* Where to put the option name */
   NFARG( int		optL )		/* Size of optP buffer */
   NFARG( char		*valP )		/* Where to put the option value */
   NFARG( int		valL )		/* Size of valP buffer */
    FARG( BOOL		semireqF )	/* if semicolon is required */
{
	BOOL		okF;
	int		orgbX;		/* Original fetch point */
	int		optC, valC;	/* Char counts */

    orgbX = rsP->rs_bX;

    /* Start by getting the first token */
    optC = msgpart_hdrtoken( rsP, optP, optL );

    /* See if first is the semicolon and get another token if so */
    if ( ( optC == 1 ) && ( optP[0] == ';' ) ) {
	optC = msgpart_hdrtoken( rsP, optP, optL );
	okF = TRUE;
    }
    else {
	okF = !semireqF;		/* Bad if semi required. */
    }

    if ( okF ) {
	/* Get the "=" token */
	valC = msgpart_hdrtoken( rsP, valP, valL );
	if ( ( valC == 1 ) && ( valP[0] == '=' ) )
	    valC = msgpart_hdrtoken( rsP, valP, valL );
	else
	    okF = FALSE;
    }

    if ( !okF ) {
	rsP->rs_bX = orgbX;		/* Restore fetch point */
    }

    return ( okF );
}
/*

*//* msgpart_hdrtoken( rsP, tokP, tokL )

	Extract the next token from a header string

Accepts :

	rsP		Ptr to REFSTR containing the header
	tokP		Where to put the token string
	tokP		Length of token buffer

Returns :

	<value>		Number of characters in the token (see note)

Notes :

	This is a specialized function to extract a "token" from a
	header string.  It's specialized in that it's written only
	for a particular need at the time of its writing, as an
	alternative to doing something more generic a la gettoken()
	from the mml stuff.

	A token is recognized by this function after skipping whitespace
	(spaces, tabs, newlines), and will be one of:

	  - any sequence of characters enclosed in double quotes;

	  - Any sequence of characters constituting a token in the
	    RFC2045 sense, which is to say:
		token := 1*<any (US-ASCII) CHAR except SPACE, CTLs,
		         or tspecials>;

	  - any single character not in the previous set.

	With the exception of a quoted string tokens, all header
	comments in parentheses are ignored.

	Example:  a header string of:
	    text/plain; (user specified) charset=iso-8859-1; format="flowed"
	would yield the hdrtokens:
		text
		/
		plain
		;
		charset
		=
		iso-8859-1
		;
		format
		=
		flowed


	The returned token will be NUL-terminated.

	Returned value:  the returned value is the number of characters
	that are in the actual token, whether or not that many characters
	were returned.  If the returned value is greater than tokL, this
	means that the some bytes of the token were not returned (since
	they could not all fit in the buffer provided).

*/

int
msgpart_hdrtoken ARGLIST( (rsP, tokP, tokL) )
   NFARG( REFSTR	*rsP )		/* REFSTR for header */
   NFARG( char		*tokP )		/* Where to put the token */
    FARG( int		tokL )		/* Size of token buffer */
{
	char		quote;		/* Any quoted char */
	UWORD		flags;
	int		cC;		/* Character count */
	int		ch;		/* Character */

    /* Reserve space for terminating NUL */
    --tokL;

    /* Get chars */
    flags = HC_PAREN;
    cC = 0;
    quote = NUL;
    for ( ; ; ) {
	ch = msgpart_hdrchar( rsP, flags );
	if ( ch == EOF )
	    break;

	/* Special stuff if first character */
	if ( cC == 0 ) {
	    /* Skip leading whitespace */
	    if ( isspace( ch ) )
		continue;

	    /* Check for opening quote */
	    if ( ( quote == NUL ) && ( ch == '"' ) ) {
		quote = ch;
		flags = 0;
		continue;
	    }
	}

	/* Special if quoted string */
	if ( quote != NUL ) {
	    if ( ch == quote )
		break;

	    /* Count it and take it if it fits */
	    if ( cC++ < tokL )
		*tokP++ = ch;
	}

	else {
	    /* Count/take this character */
	    if ( cC++ < tokL )
		*tokP++ = ch;

	    /* If it's a space, CTL, or tspecial, be done */
#if 0
	    /* If it's not a token-constituent character, we're done */
	    if ( !(
	            isalpha(ch) ||
		    isdigit(ch) ||
		    ( ch == '-' )
		   ) ) {
#endif
	    if ( ( Mchartbl[ch] & (MC_SPC|MC_CTL|MC_TSPL|MC_NASC) ) != 0 ) {
		/* If there were others already collected, we want to
		   get this one next time
		*/
		if ( cC > 1 ) {
		    if ( --cC < tokL )
			--tokP;
		    --rsP->rs_bX;
		}
		break;
	    }
	}
    }

    /* Terminate the token and return the character count */
    *tokP = NUL;
    return ( cC );
}
/*

*//* msgpart_learn_headers( mpP )

	Glean information from messagepart headers

Accepts :

	mpP		Messagepart to get info from

Returns :

	<value>		TRUE if OK
			FALSE if something bad happened

Notes :

	Looks at this messagepart's headers and collects some
	info from them.

*/

BOOL
msgpart_learn_headers ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Messagepart ptr */
{
	int		i;
	int		len;
	int		cmp;
	char		*deftypeP;	/* Default type */
	char		*defsubtypeP;	/* Default subtype */
	void		*hdrP;		/* Ptr to header handle */
	REFSTR		*rsP;		/* REFSTR of header */
	BSTR		*bsP;		/* base BSTR of header's REFSTR */

    /* Walk through the headers */
    for ( hdrP = NULL;
	 ( hdrP = msgpart_hdr_next( mpP, hdrP, NULL ) ) != NULL;
		) {

	rsP = msgpart_hdr_str( mpP, hdrP );
	bsP = refstr_bstr( rsP );

	/* Find this in the dispatch table */
	for ( i = 0; Hdrlearn_disp[i].hl_hnameP != NULL; ++i ) {
	    len = strlen( Hdrlearn_disp[i].hl_hnameP );
	    cmp = strnicmp( Hdrlearn_disp[i].hl_hnameP,
			   (char *)bsP->bs_bP, len );

	    /* Table is alphabetical order- exit now if we've passed
	       where this entry might be.
	    */
	    if ( cmp < 0 )
		break;

	    if ( cmp > 0 )
		continue;		/* No match, try next */

	    /* Verify that this is a real match */
	    if ( bsP->bs_bP[len] != ':' )
		continue;

	    /* Call the processing routine */
	    if ( !(*Hdrlearn_disp[i].hl_rtc)(mpP, rsP) )
		return ( FALSE );

	    break;			/* Done with this header */
	}
    }

    /* Default the type and subtype if not given. */
    if ( mpP->mp_mime.mi_typeP == NULL ) {
	/* If the parent is multipart/digest, this type is supposed to
	   default to message/rfc822.
	*/
	if ( ( mpP->mp_parentP != NULL ) &&
	     ( stricmp( mpP->mp_parentP->mp_mime.mi_typeP, "multipart" ) == 0 ) &&
	     ( stricmp( mpP->mp_parentP->mp_mime.mi_subtypeP, "digest" ) == 0 ) ) {
	    deftypeP = "message";
	    defsubtypeP = "rfc822";
	}
	else {
	    /* In all other cases, default is text/plain */
	    deftypeP = "text";
	    defsubtypeP = "plain";
	}

	mpP->mp_mime.mi_typeP = newstr( "msgpart_learn_headers",
				        "default MIME type",
					deftypeP );
	mpP->mp_mime.mi_subtypeP = newstr( "msgpart_learn_headers",
				        "default MIME subtype",
					defsubtypeP );
    }

    return ( TRUE );
}
/*

*//* msgpart_new( mmsgP, parentP )

	Allocate and set up a new msgpart struct

Accepts :

	mmsgP		Ptr to the message that contains it
	parentP		Ptr to parent if any

Returns :

	<value>		Ptr to new MSGPART
			NULL if error

Notes :

	Likewise, if parentP is non-NULL, the new MSGPART is set up
	to point to that parent.  This is a convenience-- the child
	is NOT linked into the parent at this point.

*/

MSGPART *
msgpart_new ARGLIST( (mmsgP, parentP) )
   NFARG( MMSG		*mmsgP )	/* The message that it's in */
    FARG( MSGPART	*parentP )	/* Parent, if any */
{
	MSGPART		*mpP;		/* The messagepart we're making */

    /* Allocate it */
    mpP = (MSGPART *)emalloc( "msgpart_new", "mpP", sizeof(MSGPART) );
    if ( mpP == NULL )
	return ( NULL );

    /* Initialize the new struct */
    mpP->mp_mmsgP = mmsgP;	/* The message that this is in */
    mpP->mp_parentP = parentP;	/* The parent, if any known */
    mpP->mp_slnodeP = NULL;	/* Sibling list node */
    mpP->mp_multiF = FALSE;	/* Not multipart (so far) */
    mpP->mp_deleteF = FALSE;	/* Don't delete this part */
    mpP->mp_hdrsize = 0L;	/* Length of headers */

    ll_init( &mpP->mp_hdrs );	/* Init the headers list */

    msgpstr_init( &mpP->mp_body );   /* Init the body/prolog string part */
    msgpstr_init( &mpP->mp_epilog ); /* Init the multipart epilog string part */
    
    ll_init( &mpP->mp_parts );	/* Init the child message parts list */

    /* Init the mime info area */
    mpP->mp_mime.mi_versionP = NULL;
    mpP->mp_mime.mi_typeP = NULL;
    mpP->mp_mime.mi_subtypeP = NULL;
    mpP->mp_mime.mi_boundaryP = NULL;
    mpP->mp_mime.mi_boundaryL = 0;

    return ( mpP );
}
/*

*//* msgpart_read_headers( mpP )

	Collect message part headers into memory

Accepts :

	mpP		Ptr to messagepart struct

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgpart_read_headers ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Message part thing */
{
	BOOL		okF;
	BOOL		overflowF;	/* If any overflow */
	int		ch;		/* Character */
	MSGCON		*conP;		/* Message container */
	UBYTE		*hbodyP;	/* Ptr to header body */
	MPHL		*hlP;		/* Header line */

    conP = mpP->mp_mmsgP->mm_conP;

    /* Read header lines using the generic buffer */
    Gbs1P->bs_bM = MAX_HEADER_LENGTH;
    Gbs1P->bs_chunk = 10000;
    overflowF = FALSE;
    while ( msgcon_readline( conP, HDR_TXR ) > 0 ) {
	/* We have a line.  Validate it as a potential header.
	   If it's the first line it might be a "From " line
	   (aka a "From_" line).
	*/
	hbodyP = hdr_name_skip_str( Gbs1P->bs_bP );
	if ( ( ( hbodyP == Gbs1P->bs_bP ) ||
	       ( *hbodyP != ':' ) ) &&
	     ( ( mpP->mp_hdrs.llh_firstP != NULL ) ||
	       ( strncmp( (char *)Gbs1P->bs_bP, "From ", 5 ) != 0 ) ) ) {

	    /* No header name or no colon or not initial "From " line,
	       so this is not part of header.  Will need to re-use
	       this text later.
	    */
	    conP->mc_prereadC = Gbs1P->bs_bC;
	    break;
	}

	/* This indeed looks like a header line.  Check for continuations */
	for ( ; ; ) {
	    /* Peek at the next character and put it back */
	    if ( conP->mc_ifP != NULL ) {
		/* Peek from file */
		if ( ( ch = fgetc( conP->mc_ifP ) ) != EOF )
		    ungetc( ch, conP->mc_ifP );
		else
		    break;
	    }
	    else {
		/* Peek from stream */
		if ( ( ch = in_char( conP->mc_inP ) ) != EOF )
		    in_charback( conP->mc_inP, ch );
		else
		    break;
	    }

	    if ( ( ch != ' ' ) && ( ch != '\t' ) )
		break;

	    /* Next line is a continuation-- add a newline,
	       append the continuation line, etc.
	    */
	    bstr_putb( Gbs1P, '\n' );
	    msgcon_readlinea( conP, HDR_TXR );
	}

	/* The buffer at Gbs1P contains a complete header line.
	   Make a header line thing for it and collect it.
	*/

	hlP = (MPHL *)emalloc( "msgpart_read_headers", "hlP", sizeof(MPHL) );
	if ( hlP == NULL )
	    return ( FALSE );

	hlP->mh_hP = refstr_new_bstr_copy( "omm_read_header", Gbs1P );
	if ( hlP->mh_hP == NULL )
	    return ( FALSE );

	/* Add to tail of header list */
	if ( ll_addvt( &mpP->mp_hdrs, (void *)hlP ) == NULL )
	    return ( FALSE );

	/* Update the header size (add one for newline) */
	mpP->mp_hdrsize += ( Gbs1P->bs_bC +1 );

	/* Note any overflow */
	if ( Gbs1P->bs_bC >= Gbs1P->bs_bM )
	    overflowF = TRUE;
    }

    if ( overflowF ) {
	warning( "Header line(s) overflowed." );
	return ( FALSE );
    }

    /* Learn something about the headers */
    okF = msgpart_learn_headers( mpP );

    return ( okF );
}
/*

*//* msgpart_read_msgpart( mpP )

	Read an entire messagepart (header and body or subparts)

Accepts :

	mpP		Ptr to messagepart struct

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgpart_read_msgpart ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Message part */
{
	BOOL		okF;

    /* Read the headers */
    if ( !msgpart_read_headers( mpP ) )
	return( FALSE );

    /* Read body and/or subparts */
    if ( mpP->mp_multiF )
	okF = msgpart_read_multipart( mpP );
    else if ( ( stricmp( mpP->mp_mime.mi_typeP, "message" ) == 0 ) &&
              ( stricmp( mpP->mp_mime.mi_subtypeP, "rfc822" ) == 0 ) )
	okF = msgpart_read_multipart_implied( mpP );
    else {
	/* Just read the body string.  If there's a parent, need to
	   pay attention to boundaries; if not, need to read to the
	   end.  This test is necessary to handle malformed nested
	   MIME.
	*/
	if ( mpP->mp_parentP != NULL )
	    okF = msgpstr_read( mpP->mp_parentP, &mpP->mp_body, FALSE );
	else
	    okF = msgpstr_read( mpP, &mpP->mp_body, TRUE );
    }

    return ( okF );
}
/*

*//* msgpart_read_multipart( mpP )

	Read a messagepart body that's a MIME multipart set

Accepts :

	mpP		Ptr to messagepart struct

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	See msgpstr_read for notes about cacheing.

*/

BOOL
msgpart_read_multipart ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Message part */
{
	BOOL		okF;
	int		i;
	int		len;
	MSGCON		*conP;		/* Container */
	MSGPART		*childP;	/* New child messagepart */

    conP = mpP->mp_mmsgP->mm_conP;

    /* Read any "body" (text prior to the first boundary string) */
    okF = msgpstr_read( mpP, &mpP->mp_body, FALSE );

    /* Now read the parts; read one part for every boundary string
       encountered (except that the final boundary string is a
       terminator).
    */
    while ( okF && ( len = msgcon_readline( conP, HDR_TXR ) ) >= 0 ) {
	/* This isn't strictly legit, but skip blank lines. */
	if ( len == 0 )
	    continue;

	/* Non-blank line is supposed to be our boundary string. */
	i = isboundary( mpP, Gbs1P );
	if ( i <= 0 ) {
	    /* Not our boundary string.  This is a structure error, but
	       just keep the line for the next read, and be done with
	       this part.
	    */
	    conP->mc_prereadC = len;
	    break;
	}

	/* Be done if it's the end boundary */
	if ( i == 2 )
	    break;

	/* Not end boundary.. make new child */
	childP = msgpart_new( mpP->mp_mmsgP, mpP );
	if ( childP == NULL ) {
	    okF = FALSE;
	    break;
	}

	/* Read the child */
	okF = msgpart_read_msgpart( childP );
	if ( !okF )
	    break;

	/* Add child */
	okF = msgpart_add( mpP, childP, TRUE );
	if ( !okF )
	    break;
    }

    /* Now read any epilog part.  No bytes left behind. 
       If there's a parent, pay attention to boundaries; otherwise
       read to the end.  If we didn't have to worry about
       handling malformed MIME, we wouldn't need this test.
    */
    if ( mpP->mp_parentP != NULL )
	okF = okF && msgpstr_read( mpP->mp_parentP, &mpP->mp_epilog, FALSE );
    else
	okF = okF && msgpstr_read( mpP, &mpP->mp_epilog, TRUE );

    return ( okF );
}
/*

*//* msgpart_read_multipart_implied( mpP )

	Read a messagepart body that's an implied MIME child

Accepts :

	mpP		Ptr to messagepart struct

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	An implied multipart has, as its body, another message part that
	is to be considered as a child of the part being read (for
	example, a message/rfc822 part).  The "body" thus has a set
	of message headers and its own body.

*/

BOOL
msgpart_read_multipart_implied ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Message part */
{
	BOOL		okF;
	int		i;
	int		len;
	MSGCON		*conP;		/* Container */
	MSGPART		*childP;	/* New child messagepart */

    conP = mpP->mp_mmsgP->mm_conP;
    okF = TRUE;

    /* Skip any blank lines prior to the child header. */
TODO  /* should extra (more than 1) blank lines be put in the prolog? */

    while ( okF && ( len = msgcon_readline( conP, HDR_TXR ) ) <= 0 )
	;

    /* Re-use the non-blank line */
    if ( len > 0 )
	conP->mc_prereadC = len;

    /* Make and read new child */
    childP = msgpart_new( mpP->mp_mmsgP, mpP );
    if ( childP == NULL )
	okF = FALSE;
    else {
	okF = msgpart_read_msgpart( childP );
	if ( okF )
	    okF = msgpart_add( mpP, childP, TRUE );
    }

    return ( okF );
}
/*

*//* msgpart_refstr_addr( mpP, rsP, parts )

	Get next address part(s) from a REFSTR

Accepts :

	mpP		Ptr to message part
	rsP		Ptr to REFSTR set up to point to the next
			  address in the string.
	parts		bitmask specifying which header part(s) to return

Returns :

	<value>		Pointer to an internal string buffer containing
			  the selected address parts
			NULL if no more addresses in this header

Notes :

	Uses an internal string buffer; caller should assume that the
	returned string is temporary.

	If both AP_LOCAL and AP_DOMAIN are selected, the "@" is
	also included.

	In current implementation, selecting AP_COMMENTS and
	AP_PERSNAME does nothing.  We only return local and/or domain
	parts.

	mpP isn't currently used; it's specified in case we ever
	want to make use of it.

	Address groups (group:mailbox-list;) are not [yet] handled.

*/

char *
msgpart_refstr_addr ARGLIST( (mpP, rsP, parts) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
   NFARG( REFSTR	*rsP )		/* REFSTR for header */
    FARG( UWORD		parts )		/* Selected parts */
{
	int		ch;
	int		i;
	char		*addrP;
	BSTR		*dbsP;		/* destination BSTR for return value */

    if ( ( dbsP = msgpart_temp_bstr() ) == NULL )
	return ( NULL );

    /* First copy the entire next address field into our destination,
       skipping parenthesized comments, skipping leading whitespace.
    */

    for ( ; ; ) {
	ch = msgpart_hdrchar( rsP, HC_PAREN|HC_QUOTED );

	/* If end of this address, be done */
	if ( ( ch == EOF ) || ( ch == ',' ) )
	    break;

	/* If leading whitespace, ignore it */
	if ( ( dbsP->bs_bC == 0 ) && 
	     ( ( ch == ' ' ) || ( ch == '\t' ) ||
	      ( ch == '\n' ) || ( ch == '\r' ) ) )
	    continue;

	/* Store the char into destination buffer */
	bstr_putb( dbsP, ch );
    }

    /* Back up over any trailing whitespace */
    while ( ( dbsP->bs_bC > 0 ) &&
	    ( ( ch = dbsP->bs_bP[ dbsP->bs_bC -1 ] == ' ' ) ||
	      ( ch == '\t' ) ||
	      ( ch == '\n' ) || ( ch == '\r' ) ) )
	--dbsP->bs_bC;

    bstr_nulterm( dbsP );

    /* If null string, give nothing */
    if ( dbsP->bs_bC == 0 )
	return ( NULL );

    /* Now isolate the address part.  This is a really stupid check, but
       other than failing on address group constructs, it ought to be
       correct.  If the string ends with '>' the address is enclosed in
       anglebrackets.  If it doesn't, the address is the entire string.
    */
    if ( dbsP->bs_bP[ dbsP->bs_bC -1] != '>' )
	addrP = (char *)dbsP->bs_bP;
    else {
	/* remove the anglebrackets to get the raw address */
	addrP = (char *)&dbsP->bs_bP[ dbsP->bs_bC -1 ];
	*addrP-- = NUL;

	/* Find opening anglebracket.  Note how cleverly we ignore the
	   error if it is missing.
	*/
	do {
	    if ( *addrP == '<' ) {
		++addrP;
		break;
	    }
	} while ( addrP-- != (char *)dbsP->bs_bP );
    }

    /* Now isolate either the localpart or the domainpart if required.
    */
    if ( ( parts & (AP_LOCAL|AP_DOMAIN) ) == (AP_LOCAL) ) {
	/* Find any @ sign and terminate the address there. */
	for ( i = 0; addrP[i] != NUL; ++i )
	    if ( addrP[i] == '@' ) {
		addrP[i] = NUL;
		break;
	    }
    }
    else if ( ( parts & (AP_LOCAL|AP_DOMAIN) ) == (AP_DOMAIN) ) {
	/* Find any @ sign and begin the address there */
	for ( ; *addrP != NUL; ++addrP )
	    if ( *addrP == '@' ) {
		++addrP;
		break;
	    }
    }

    if ( *addrP == NUL )
	return ( NULL );
    return ( addrP );
}
/*

*//* msgpart_size_body( mpP, childrenF )

	Return the size of a messagepart's body

Accepts :

	mpP		Ptr to messagepart
	childrenF	Whether to include child part sizes

Returns :

	<value>		The size

Notes :

	Returns the size of the body, excluding the header.  The body
	  is whatever is in the prolog/body and the epilog partstrings,
	  as a raw, non-decoded count.

	If childrenF is TRUE, the size of all MIME children (including
	  their headers and their bodies and their children) is added to
	  the value.

*/

ULONG
msgpart_size_body ARGLIST( (mpP, childrenF) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
    FARG( BOOL		childrenF )	/* Whether to count child parts */
{
	ULONG		count;
	LL_NODE		*nodeP;
	MSGPART		*submpP;

    /* The "body" size is the size of the body string parts */
    count = msgpstr_size( mpP, &mpP->mp_body, TRUE ) +
	    msgpstr_size( mpP, &mpP->mp_epilog, TRUE );

    /* Add the children if we are supposed to */
    if ( childrenF )
	for ( nodeP = mpP->mp_parts.llh_firstP;
	         nodeP != NULL;  nodeP = nodeP->lln_nextP ) {
	    submpP = (MSGPART *)nodeP->lln_valP;
	    count += msgpart_size_hdr( submpP );
	    count += msgpart_size_body( submpP, TRUE );
	}

    /* That should do it */
    return ( count );
}
/*

*//* msgpart_size_hdr( mpP )

	Return the size of a messagepart's header

Accepts :

	mpP		Ptr to messagepart

Returns :

	<value>		The size

Notes :

	Returns the size that we keep track of.

	This is a function rather than a member access just in case we
	  might want to do some kind of interpretation in the future.

*/

ULONG
msgpart_size_hdr ARGLIST( (mpP) )
    FARG( MSGPART	*mpP )		/* Ptr to message part */
{
    return ( mpP->mp_hdrsize );
}
/*

*//* msgpart_write_body(mpP, ofP, forceF)

	Write message body to an open file ptr

Accepts :

	mpP		Ptr to the message part
	ofP		File ptr to write body to
	forceF		Force writing subparts even if delete flag is on

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Assuming that the body follows the header, a blank line is
	written first, if not already present.

	If the body is a multipart MIME structure, this recursively
	writes the headers/bodies of all the included parts that are
	not marked for deletion (the passed force flag, forceF, will
	bypass the deletion check). Caller can have more discrete
	control over this recursion by taking care not to call this
	function for multiparts.

	If the delete flag is set on any part (and the forceF argument
	is not set), writing of any subparts is also inhibited.  i.e.
	deleting any part deletes all of its subparts.

	The force flag does NOT apply to this message part, only its
	subparts (if any).  It's assumed that if this routine is called
	at all, the caller wants the body written.

*/


BOOL
msgpart_write_body ARGLIST( (mpP, ofP, forceF) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
   NFARG( FILE		*ofP )		/* Where to write body */
    FARG( BOOL		forceF )	/* Force writing deleted subparts */
{
	BOOL		okF;
	int		partC;		/* Parts counter */
	LL_NODE		*nodeP;		/* Linked list node */
	MMSG		*mmsgP;		/* Message control */
	MSGPART		*submpP;	/* Subpart */

    okF = TRUE;
    mmsgP = mpP->mp_mmsgP;

    /* Make sure that there is at least one blank line.  This has to be
       done before writing any prolog, to avoid suppressing other blank
       lines that might be added by the prolog.
    */
    while ( okF && ( mmsgP->mm_cnlC < 2 ) ) {
	okF = ( fputc( '\n', ofP ) != EOF );
	++mmsgP->mm_cnlC;		/* Regardless of error */
    }

    /* Write any body/prolog partstring. */
    if ( mpP->mp_body.ps_partP != NULL )
	okF = okF && msgpstr_write( mpP, &mpP->mp_body, ofP,
				    MWF_BLBEGIN | MWF_NLEND );

    /* If this has children (e.g. multipart or implied multipart),
       loop to write them.
    */
    if ( okF && ( mpP->mp_parts.llh_firstP != NULL ) ) {
	partC = 0;
	for ( nodeP = mpP->mp_parts.llh_firstP;
	         nodeP != NULL;  nodeP = nodeP->lln_nextP ) {
	    submpP = (MSGPART *)nodeP->lln_valP;

	    /* Skip if deleted and not forced */
	    if ( submpP->mp_deleteF && !forceF )
		continue;

	    /* If real multipart, write the MIME boundary string.
	       Do not write boundary for implied multipart, e.g.
	       message/rfc822 type, even if one was present in the
	       msgpart header.
	    */
	    if ( mpP->mp_multiF ) {
		okF = okF &&
		  ( fprintf( ofP, "--%s\n", mpP->mp_mime.mi_boundaryP ) > 0 );
		mmsgP->mm_cnlC = 1;
	    }

	    if ( okF )
		++partC;

	    /* Recurse to write the headers and body */
	    okF = okF && msgpart_write_hdrs( submpP, ofP, TRUE );
	    okF = okF && msgpart_write_body( submpP, ofP, forceF );
	}

	/* If any subparts written, write the final boundary string
	   (followed by "--" to indicate a closing boundary).
	   Again, if the multipart is implied, the boundary string
	   is not written.
	*/
	if ( okF && ( partC > 0 ) && mpP->mp_multiF ) {
	    okF = ( fprintf( ofP, "--%s--\n",
			     mpP->mp_mime.mi_boundaryP ) > 0 );
	    mmsgP->mm_cnlC = 1;
	}
    }

    /* Write any epilog partstring, without leading blank line (if
       it's wanted or present, it'll be in the epilog string.)
    */
    okF = okF && msgpstr_write( mpP, &mpP->mp_epilog, ofP,
			        MWF_NLEND );

    return ( okF );
}
/*

*//* msgpart_write_hdrs(mpP, ofP, embedF)

	Write headers optionally matching initial string

Accepts :

	mpP		Ptr to message part
	ofP		File ptr to write header(s) to
	embedF		TRUE if we're embedding these headers inside
			  another message.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	embedF is used to signal the fact that these headers are
	being included inside another message.  Mainly this will
	simply inhibit the writing of any "From " header or
	any other problematic headers.

*/

BOOL
msgpart_write_hdrs ARGLIST( (mpP, ofP, embedF) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
   NFARG( FILE		*ofP )		/* Where to write body */
    FARG( BOOL		embedF )	/* TRUE if embedded in another msg */
{
	BOOL		okF;
	MMSG		*mmsgP;		/* Message object */
	LL_NODE		*nodeP;		/* Ptr to list node */
	BSTR		*bsP;		/* Byte string ptr */

    mmsgP = mpP->mp_mmsgP;

    /* Check that there is at least one header */
    nodeP = mpP->mp_hdrs.llh_firstP;
    if ( nodeP == NULL ) {
	/* This could conceivably be OK, e.g. for a child part.
	   However some things need at least some header elements.
	     - Anything without a parent;
	     - any message/rfc822 part;
	     - any outermost part being written (e.g. if some child is
	       being written as a standalone message).
	*/
	if ( ( mpP->mp_parentP == NULL ) ||
	     ( !embedF ) ||
	     ( ( stricmp( mpP->mp_mime.mi_typeP, "message" ) == 0 ) &&
               ( stricmp( mpP->mp_mime.mi_subtypeP, "rfc822" ) == 0 ) )
	   ) {
	    /* Must supply at least one header here. */
TODO	    /* This is dumb; could do something better here. */
	    okF = ( fprintf( ofP,
		    "X-There-Was-No-Header: so I supplied one.\n" ) > 0 );
	    mmsgP->mm_cnlC = 1;
	}
	else
	    okF = TRUE;
    }
    else {
	/* Go through the list */
	for ( okF = TRUE;
	        okF && ( nodeP != NULL );
		   nodeP = nodeP->lln_nextP ) {

	    bsP = refstr_bstr( msgpart_hdr_str( mpP, nodeP ) );

	    if ( embedF ) {
		/* May need to inhibit this header field */
		if ( strnicmp( "From ", (char *)bsP->bs_bP, 5 ) == 0 )
		    continue;
	    }

	    /* Write this header */
	    if ( fwrite( bsP->bs_bP, 1, bsP->bs_bC, ofP ) != bsP->bs_bC )
		okF = FALSE;
	    else if ( fputc( '\n', ofP ) == EOF )
		okF = FALSE;
	    mmsgP->mm_cnlC = 1;		/* Regardless of error */
	}
    }
	
    return ( okF );
}
/*

*//* msgpart_write_references(mpP, ofP)

	Write reference headers for a new message

Accepts :

	mpP		Ptr to message part which supplies the source info
	ofP		File ptr to write header(s) to

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Outputs the headers that are references to another messages; the
	message being referred to is headed by the passed messagepart.
	RFC2822 section 3.6.4 supplies the model for creating reference
	headers.

*/

BOOL
msgpart_write_references ARGLIST( (mpP, ofP) )
   NFARG( MSGPART	*mpP )		/* Ptr to messagepart */
    FARG( FILE		*ofP )		/* Where to write references */
{
	BOOL		okF;
	int		i;
	int		midL;		/* Message ID length */
	int		refL;		/* Old references header length */
	void		*hdrP;		/* Header object */
	UBYTE		*bP;
	ESTR		*midP;		/* Ptr to referenced message-id */
	ESTR		*refP;		/* Ptr to original references */

    okF = TRUE;

    /* We'll need the referenced message's message-id */
    hdrP = msgpart_hdr_next( mpP, NULL,
			     estr_tmp_str( (UBYTE *)"Message-ID" ) );
    if ( hdrP != NULL ) {
	midP = msgpart_hdr_tail( mpP, hdrP );
	midL = estr_len( midP );

	/* Write "In-Reply-To" header */
	okF = okF && ( fprintf( ofP, "In-Reply-To: " ) > 0 );
	okF = okF && ( fwrite( midP->es_strP, 1, midL, ofP ) == midL );
	okF = okF && ( fputc( '\n', ofP ) != EOF );

	/* Now for a References header.  This should be based on the
	   reference messages's Reference header if it has one, its
	   In-Reply-To header failing that, or created from the
	   referenced message-id.
	*/
	hdrP = msgpart_hdr_next( mpP, NULL,
				 estr_tmp_str( (UBYTE *)"References" ) );
	if ( hdrP == NULL )
	    hdrP = msgpart_hdr_next( mpP, NULL,
				     estr_tmp_str( (UBYTE *)"In-Reply-To" ) );

	okF = okF && ( fprintf( ofP, "References: " ) > 0 );
	if ( hdrP != NULL ) {
	    refP = msgpart_hdr_tail( mpP, hdrP );
	    refL = estr_len( refP );
	    okF = okF && ( fwrite( refP->es_strP, 1, refL, ofP ) == refL );

	    /* Add a continuation if the last partial line is longer than
	       some swagged value here
	    */
	    i = refL;
	    bP = (UBYTE *)strrchr( (char *)refP->es_strP, '\n' );
	    if ( bP != NULL )
		i -= (bP - refP->es_strP);
	    if ( i > 78 )
		okF = okF && (fprintf( ofP, "\n" ) > 0 );

	    /* Add space between the referenced message-id and the others */
	    okF = okF && ( fputc( ' ', ofP ) != EOF );
	}
	okF = okF && ( fwrite( midP->es_strP, 1, midL, ofP ) == midL );
	okF = okF && ( fputc( '\n', ofP ) != EOF );

	mpP->mp_mmsgP->mm_cnlC = 1;	/* Even if error, maintain this */
    }

    return ( okF );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/***********
 *
 *  hdrlearn_xxx functions
 *    called out of the Hdrlearn_disp[] table by msgpart_learn_headers()
 *
 *    Each is called with (mpP, rsP):
 *         mpP is the pointer to the messagepart structure
 *         rsP is the pointer to the REFSTR containing the header,
 *            presumed to be set up to point to the first byte after
 *            the header name (i.e., after the colon)
 *
 *    Each function returns TRUE as a normal thing, FALSE if something
 *      bad happened.
 *
 ***********/


    /* hdrlearn_content_type:  "Content-Type"	*/
static BOOL
hdrlearn_content_type ARGLIST( (mpP, rsP) )
   NFARG( MSGPART	*mpP )		/* Message part */
    FARG( REFSTR	*rsP )		/* Header string REFSTR */
{
	int		tok1L, tok2L, tok3L;
	char		tok1[100], tok2[100], tok3[100];

    /* Should have mediatype/subtype next */
    tok1L = msgpart_hdrtoken( rsP, &tok1[0], sizeof(tok1) );
    tok2L = msgpart_hdrtoken( rsP, &tok2[0], sizeof(tok2) );
    tok3L = msgpart_hdrtoken( rsP, &tok3[0], sizeof(tok3) );

    if ( ( tok1L == 0 ) ||
         ( tok3L == 0 ) ||
	 ( tok2L != 1 ) ||
	 ( tok2[0] != '/' ) )
	/* Badly formatted line, just ignore it. */
	return ( TRUE );

    /* Store the media type and subtype */
    mpP->mp_mime.mi_typeP = newstr( "hdrlearn_content_type", "mime type",
					&tok1[0] );
    if ( mpP->mp_mime.mi_typeP == NULL )
	return ( FALSE );

    mpP->mp_mime.mi_subtypeP = newstr( "hdrlearn_content_type", "mime subtype",
					&tok3[0] );
    if ( mpP->mp_mime.mi_subtypeP == NULL )
	return ( FALSE );

    /* Get any options */
    for ( ; ; ) {
	if ( !msgpart_hdroptpair( rsP, &tok1[0], sizeof(tok1),
					&tok2[0], sizeof(tok2),
					TRUE ) )
	    break;

	if ( stricmp( &tok1[0], "boundary" ) == 0 ) {
	    mpP->mp_mime.mi_boundaryP = newstr( "hdrlearn_content_type",
					         "boundary string",
					         &tok2[0] );
	    mpP->mp_mime.mi_boundaryL = strlen( mpP->mp_mime.mi_boundaryP );
	}
    }

    /* Postprocess what we learned from this header */

    mpP->mp_multiF = ( ( stricmp( mpP->mp_mime.mi_typeP, "multipart" ) == 0 ) &&
		       ( mpP->mp_mime.mi_boundaryP != NULL ) );

    return ( TRUE );
}



/***********
 *
 *    End of hdrlearn_xxxx section
 *
 ***********/


/*

*//* isboundary( mpP, bsP )

	See if a string is a boundary string

Accepts :

	mpP		Ptr to MSGPART that specifies the boundary string.
	bsP		BSTR containing the candidate line to look at

Returns :

	<value>		0 if bsP does not contain a boundary string, or
			 1: non-final boundary string (no trailing "--")
			 2: final boundary string (trailing "--")
			-1: like 1, only matched a parent, not current part
			-2: like 2, only matched a parent, not current part

Notes :

	mpP will typically either be the messagepart being worked on
	  (in the case where it's a multipart container and caller
	  is looking for each boundary in order to add children),
	  or the parent (in the case where it's not a multipart,
	  but a child of one).

	No judgement is made as to whether or not we SHOULD check for
	  a boundary string (i.e., whether the messagepart is of a type
	  that is appropriate for the check).  Caller should do that.

	RFC2046 is somewhat contradictory on the subject of boundary
	  strings.  It says that boundary strings in the message body may
	  only be followed by zero or more spaces and a CRLF.  However it
	  also says to choose boundary strings such that that no initial
	  substrings of message lines contain them, and it also says
	  (5.1.1):

	     NOTE TO IMPLEMENTORS:  Boundary string comparisons must
	     compare the boundary value with the beginning of each
	     candidate line.  An exact match of the entire candidate
	     line is not required; it is sufficient that the boundary
	     appear in its entirety following the CRLF.

	We'll follow that advice and do non-strict checks (only
	checking the initial substring).  The check has been isolated
	to this function so that it can be changed if desired.

	The boundary string search must include all ancestors.  This seems
	  obvious, but it's backed up by RFC2046 (5.1.2):

	    It is essential that such entities be handled correctly
	    when they are themselves imbedded inside of another
	    "multipart" structure.  MIME implementations are therefore
	    required to recognize outer level boundary markers at ANY
	    level of inner nesting.  It is not sufficient to only
	    check for the next expected marker or other terminating
	    condition.

*/

static int
isboundary ARGLIST( (mpP, bsP) )
   NFARG( MSGPART	*mpP )		/* Message part */
    FARG( BSTR		*bsP )		/* The candidate string (to test) */
{
	int		result;
	UBYTE		*bP;		/* Underlying bsP buffer */
	MSGPART		*ompP;		/* Another messagepart */

    if ( bsP->bs_bC < 3 )
	return ( 0 );			/* Impossible */

    bP = Gbs1P->bs_bP;

    /* Boundary strings are always prefixed by a double hyphen, so
       that's a quick precheck.
    */
    if ( ( bP[0] != '-' ) || ( bP[1] != '-' ) )
	return ( 0 );

    /* Check against boundary strings of this part and all parents. */
    for ( ompP = mpP; ompP != NULL; ompP = ompP->mp_parentP ) {
	if ( ( ompP->mp_mime.mi_boundaryP != NULL ) &&
	     ( bsP->bs_bC >= ( ompP->mp_mime.mi_boundaryL + 2 ) ) &&
	     ( strncmp( (char *)&bP[2],
		      ompP->mp_mime.mi_boundaryP,
		      ompP->mp_mime.mi_boundaryL ) == 0 ) )
	    break;
    }

    if ( ompP == NULL )
	return ( 0 );			/* No match */

    /* Got a match, set result depending on whether it's an end marker */
    if ( ( bsP->bs_bC >= ( ompP->mp_mime.mi_boundaryL + 4 ) ) &&
	 ( bP[ ompP->mp_mime.mi_boundaryL + 2 ] == '-' ) &&
	 ( bP[ ompP->mp_mime.mi_boundaryL + 3 ] == '-' ) )
	result = 2;			/* End marker */
    else
	result = 1;			/* Not end marker */

    /* Negate the result if it was a parent match */
    if ( ompP != mpP )
	result = -result;

    return ( result );
}
/*

*//* mimeinfo_free( miP, freeF )

	Free a MIMEINFO struct's data

Accepts :

	miP		Ptr to the MIMEINFO struct
	freeF		TRUE to also free the MIMEINFO struct.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Will return FALSE if any of its component freeing fails, but
	  will complete regardless.

	Also clears any pointers that are freed.

*/

static BOOL
mimeinfo_free ARGLIST( (miP, freeF) )
   NFARG( MIMEINFO	*miP )		/* The MIMEINFO to free */
    FARG( BOOL		freeF )		/* Whether to free the thing itself */
{
	BOOL		okF;

    okF = TRUE;

    /* Free up the components, noting any error we can */
    if ( miP->mi_versionP != NULL ) {
	dealloc( miP->mi_versionP );
	miP->mi_versionP = NULL;
    }
    if ( miP->mi_typeP != NULL ) {
	dealloc( miP->mi_typeP );
	miP->mi_typeP = NULL;
    }
    if ( miP->mi_subtypeP != NULL ) {
	dealloc( miP->mi_subtypeP );
	miP->mi_subtypeP = NULL;
    }
    if ( miP->mi_boundaryP != NULL ) {
	dealloc( miP->mi_boundaryP );
	miP->mi_boundaryP = NULL;
    }

    /* If it's wanted, free the MIMEINFO thing */
    if ( freeF )
	dealloc( miP );

    return ( okF );
}
/*

*//* msgpart_temp_bstr()

	Return ptr to scratch BSTR ready for use

Accepts :

	<nothing>

Returns :

	<value>		Ptr to BSTR 
			or NULL if problem.

Notes :

	BSTR is prepared for storing a new string

*/

static BSTR *
msgpart_temp_bstr NOARGLIST
{
	BSTR		*bsP;		/* Ptr to bstr */

    if ( ( bsP = BsP ) == NULL ) {
	/* No scratch BSTR -- make one */
	bsP = bstr_new( "msgpart_temp_bstr", 1000, 0 );
	if ( bsP == NULL )
	    return ( NULL );

	BsP = bsP;
    }

    bsP->bs_bC = 0;
    bsP->bs_bX = 0;

    return ( bsP );
}
/*

*//* msgpart_write_stack( mpP, inP, ofP, flags )

	Write from an input stack

Accepts :

	mpP		The relevant messagepart
	inP		Ptr to input stack
	ofP		File ptr to write to
	flags		MWF message write flags

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Mainly exists to do any special quoting required, e.g. "From "
	line quoting.

*/


static BOOL
msgpart_write_stack ARGLIST( (mpP, inP, ofP, flags) )
   NFARG( MSGPART	*mpP )		/* The relevant messagepart */
   NFARG( MML_IN	*inP )		/* Input stack to write from */
   NFARG( FILE		*ofP )		/* File to write to */
    FARG( UWORD		flags )		/* Message writing flags */
{
	BOOL		okF;
	int		ch;
	int		flen;		/* Index into fromP "From " string */
	MMSG		*mmsgP;		/* Message object */

	/* We look for this string in a strange way. */
static	char		*fromP = "From ";

    /* This is complicated by these things:

       - If MWF_BLBEGIN is set, we have to make sure there is a blank
	 line at the beginning;
       - we have to translate naked CRs into newlines;
       - we have to translate CRLFs into newlines;
       - If MWF_BLEND is set, we have to make sure there is a blank
	 line after the end of the body.

    */

    okF = TRUE;
    mmsgP = mpP->mp_mmsgP;		/* Ptr to message object */

    /* If an initial blank line is needed, check to see that either
       a blank line has already been output (check the consecutive
       newline count in the message object), or that there is one
       present in the input.
    */
    if ( ( ( flags & MWF_BLBEGIN ) != 0 ) &&
         ( mmsgP->mm_cnlC < 2 ) ) {

TODO	/* If nlC is 0, we really need TWO conecutive line-ends.  But
	   at present, we can assume that every output part starts on a
	   line beginning.  If that changes, we'll have to investigate
	   further.
	*/
	ch = in_char( inP );
	if ( ch != EOF )
	    in_charback( inP, ch );
	if ( ( ch != '\n' ) && ( ch != '\r' ) ) {
	    if ( fputc( '\n', ofP ) == EOF )
		okF = FALSE;
	    else
		++mmsgP->mm_cnlC;
	}
    }

    /* Output the body, doing translations per above */
    while ( okF ) {
	if ( ( ch = in_char( inP ) ) == EOF )
	    break;

	/* Check for "From" at beginning of line following any
	   blank line (not just at the beginning of any line.
	   This is pretty hideous.
	*/
	if ( ( mmsgP->mm_cnlC > 1 ) && ( ch == 'F' ) && mmsgP->mm_oqfromF ) {
	    for ( flen = 1; ; ++flen ) {
		ch = in_char( inP );
		if ( ch != fromP[flen] )
		    break;
	    }
	    if ( fromP[flen] == NUL ) {
		/* Matched-- output the escape char */
		if ( fputc( '>', ofP ) == EOF )
		    okF = FALSE;
	    }

	    /* Output the [partial] "From " string */
	    okF = okF && ( fwrite( fromP, 1, flen, ofP ) == flen );

	    mmsgP->mm_cnlC = 0;		/* No newlines.. */
	}

	/* Check for naked CR or CRLF sequence */
	if ( ch == '\r' ) {
	    ch = in_char( inP );
	    if ( ch != '\n' ) {
		/* CR not followed by LF -- put back the following
		   char and assume the LF anyway.
		*/
		if ( ch != EOF )
		    in_charback( inP, ch );
		ch = '\n';
	    }
	}

	/* Ok to output the character */
	okF = okF && ( fputc( ch, ofP ) != EOF );

	if ( ch == '\n' ) {
	    /* We count consecutive newlines, sort of, both as a
	       beginning-of-line indicator, and so we can potentially
	       add a blank line at the end of the body if one wasn't
	       there.  The counter only has to go up to 2.
	    */
	    if ( mmsgP->mm_cnlC < 10 )
		++mmsgP->mm_cnlC;
	}
	else
	    mmsgP->mm_cnlC = 0;
    }

    /* If wanted, make sure the output ends with a newline */
    if ( okF && ( ( flags & ( MWF_NLEND | MWF_BLEND ) ) != 0 ) )
	if ( mmsgP->mm_cnlC < 1 ) {
	    okF = ( fputc( '\n', ofP ) != EOF );
	    ++mmsgP->mm_cnlC;
	}


    /* If wanted, make sure the output ends with a blank line
       (two consecutive end-of-line markers).  The first one
       will already have been handled above.
    */
    if ( okF && ( ( flags & MWF_BLEND ) != 0 ) )
	if ( mmsgP->mm_cnlC < 2 ) {
	    okF = ( fputc( '\n', ofP ) != EOF );
	    ++mmsgP->mm_cnlC;
	}

    return ( okF );
}
/*

*//* msgpstr_add_n( psP, bufP, bufC )

	Add bytes from a counted string to a MSGPSTR's string cache

Accepts :

	psP		Ptr to the partstr struct
	bufP		Ptr to byte string
	bufC		Number of bytes to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Will initialize the cache buffer as needed

*/

static BOOL
msgpstr_add_n ARGLIST( (psP, bufP, bufC) )
   NFARG( MSGPSTR	*psP )		/* Struct for the byte string */
   NFARG( UBYTE		*bufP )		/* Ptr to bytes to add */
    FARG( int		bufC )		/* Number of bytes to add */
{
    /* Initialize cache if needed */
    if ( psP->ps_partP == NULL )  {
	/* Create a cache for the string */
	psP->ps_partP = refstr_new( "msgpart_initbody",
				     5000,
				     10000L + OMM_BODYCACHE_MAX );
	if ( psP->ps_partP == NULL )
	    return ( FALSE );

	psP->ps_partbsP = refstr_bstr( psP->ps_partP );
	psP->ps_partbsP->bs_chunk = 25000;
    }

    /* Add to the cache */
    return ( bstr_catn( psP->ps_partbsP, bufP, bufC ) );
}
/*

*//* msgpstr_free( psP, freeF )

	Free a MSGPSTR struct's data

Accepts :

	psP		Ptr to the MSGPSTR struct
	freeF		Whether to also free the MSGPSTR struct.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Will return FALSE if any of its component freeing fails, but
	  will complete regardless.

	Also clears any pointers.

*/

static BOOL
msgpstr_free ARGLIST( (psP, freeF) )
   NFARG( MSGPSTR	*psP )		/* The MSGPSTR to free */
    FARG( BOOL		freeF )		/* Whether to free the thing itself */
{
	BOOL		okF;

    okF = TRUE;

    /* Free the components, noting any failures that we can note */
    if ( psP->ps_partP != NULL ) {
	if ( !refstr_free( psP->ps_partP ) )
	    okF = FALSE;
	psP->ps_partP = NULL;
    }
    psP->ps_partbsP = NULL;

    return ( okF );
}
/*

*//* msgpstr_init( psP )

	Initialize a MSGPSTR struct

Accepts :

	psP		Ptr to MSGPSTR struct

Returns :

	<void>

Notes :

	Simply clears the MSGPSTR.

*/

static void
msgpstr_init ARGLIST( psP )
    FARG( MSGPSTR	*psP )		/* The partstr to init */
{
    psP->ps_partP = NULL;		/* Cached part */
    psP->ps_partbsP = NULL;		/* BSTR for ps_partP */
    psP->ps_nc_start = 0;		/* Start of non-cached portion */
    psP->ps_nc_end = 0;			/* End of non-cached portion */
}
/*

*//* msgpstr_read( mpP, psP, toendF )

	Read a part of a body that's a simple byte string

Accepts :

	mpP		Ptr to messagepart
	psP		Ptr to partstr struct
	toendF		TRUE to read to the end of input, ignoring
			  MIME boundaries.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	This is called to read a byte stream part only.  It will
	  cache some of the byte stream in the cache area, and
	  buffer any further part in the file area specified by
	  the message container.

	Some amount of any body string (body/prolog or epilog) read by
	  this function will be cached in memory, according to the limit
	  set in the message struct and the current count recorded
	  there.  A minimum cache size will be used for each body part
	  even if the cache limit has been met.

	The passed mpP should be the MSGPART that is most relevant for
	  finding the end of the byte string.  e.g., if we're reading
	  the prolog of a multipart type, it should be the multipart
	  msgpart itself.  If we're reading the epilog of a multipart
	  msgpart, it should be the parent part.  If we're reading
	  just a simple body, it should be that simple part.

	It's assumed the file pointer is at the first byte
	  of the byte string, and that any optional preceding blank
	  lines have been skipped.

*/

static BOOL
msgpstr_read ARGLIST( (mpP, psP, toendF) )
   NFARG( MSGPART	*mpP )		/* Message part */
   NFARG( MSGPSTR	*psP )		/* Struct for the byte string */
    FARG( BOOL		toendF )	/* To read to the end */
{
	BOOL		okF;
	BOOL		newlineF;	/* if next bufferfull starts a line */
	BOOL		boundaryF;	/* Set if we hit end MIME boundary */
	int		i;
	int		len;
	long		cacheC;		/* How much to cache, more or less */
	MSGCON		*conP;		/* Message container */
	MMSG		*mmsgP;		/* Message object */
	MSGPART		*ompP;		/* Another messagepart */

    conP = mpP->mp_mmsgP->mm_conP;
    mmsgP = mpP->mp_mmsgP;

    /* How much we can cache.  Guarantee a minimum regardless of current
       usage.
    */
    cacheC = mmsgP->mm_cachemax - mmsgP->mm_cacheC;
    if ( cacheC < OMM_BODYCACHE_GTY )
	cacheC = OMM_BODYCACHE_GTY;

    /* Loop reading lines, using the generic buffer.  This is
       potentially slower than direct reads but we have to look for
       special things like MIME boundaries.
    */
    Gbs1P->bs_bM = 1000;			/* Don't buffer too much at a time
					   to simplify cacheing decision.
					*/
    Gbs1P->bs_chunk = 1000;
    newlineF = TRUE;			/* Next bufferfull is a new line */
    boundaryF = FALSE;			/* Haven't hit end MIME boundary */

    /* Force the toendF flag if there can be no boundaries to check. */
    if ( ( !mpP->mp_multiF ) && ( mpP->mp_parentP == NULL ) )
	toendF = TRUE;

    okF = TRUE;

    while ( ( len = msgcon_readline( conP, BODY_TXR ) ) >= 0 ) {

	/* Check for a MIME boundary.  There are these cases:
	   -  the caller doesn't want us to check (toendF is TRUE).
	      This is necessary e.g. when we are reading an epilog
	      that could have our MIME boundary in it.  That would be
	      broken MIME, but we still have to handle it;

	   -  the passed message part is a multipart, and we should stop
	      when we see any of its boundary strings;

	   - the part we're reading has a parent, then that parent must be a
	      multipart (explicit or implied) and should stop when we see
	      any of the ancestor's boundary strings;

	   -  otherwise there are no boundary tests and we just read to
	      the end of input.

	   Note that some of those conditions have been pre-tested above,
	   via forcing the toendF flag.
	   
	*/
	if ( newlineF ) {
	    if ( toendF )
		i = 0;
	    else if ( mpP->mp_multiF )
		i = isboundary( mpP, Gbs1P );
	    else if ( mpP->mp_parentP != NULL )
		i = isboundary( mpP->mp_parentP, Gbs1P );
	    else
		i = 0;			/* Should not get here */

	    if ( i != 0 )
		boundaryF = TRUE;

	    if ( boundaryF ) {
		conP->mc_prereadC = len; /* Data in Gbsp is to be used again */
		break;			 /* and we're done */
	    }
	}

	/* Next bufferfull is a new line if this read didn't overflow */
	newlineF = ( Gbs1P->bs_bC < Gbs1P->bs_bM );

	/* We'll put this into the body cache if we haven't yet exceeded
	   our quota.  Note this means we might cache slightly more than
	   specified as the max (up to the buffering size we use)
	*/
	if ( ( psP->ps_partP == NULL ) ||
	     ( psP->ps_partbsP->bs_bC < cacheC ) ) {
	    /* Store the new stuff into the body cache */
	    if ( !msgpstr_add_n( psP, &Gbs1P->bs_bP[0], Gbs1P->bs_bC ) ) {
		okF = FALSE;
		break;
	    }
	    if ( newlineF && ( !bstr_putb( psP->ps_partbsP, '\n' ) ) ) {
		okF = FALSE;
		break;
	    }

	    /* Now if the cache is full, switch to non-caching mode */
	    if ( psP->ps_partbsP->bs_bC >= cacheC ) {
		if ( ! msgcon_uncache( conP ) ) {
		    okF = FALSE;
		    break;
		}

		/* Remember the start of the non-cached area */
		psP->ps_nc_start = ftell( conP->mc_usefP );

		/* Optimize here to seek to the end of the body *if*
		   we know we are reading to the end of input and not
		   looking for a boundary.
		*/
		if ( toendF ) {
		    if ( fseek( conP->mc_usefP, 0L, SEEK_END ) != 0 ) {
			warning( "msgpstr_read: seek to end failed (%d)",
					errno );
			okF = FALSE;
			break;
		    }
		}

		/* Regardless, remember the current non-cached end */
		psP->ps_nc_end = ftell( conP->mc_usefP );
	    }
	}
	else {
	    /* Not adding to the cache-- make sure this data is
	       available as uncached data.
	    */
	    if ( !msgcon_uncache_data( conP,
				      &Gbs1P->bs_bP[0], Gbs1P->bs_bC,
				      newlineF ) ) {
		okF = FALSE;
		break;
	    }

	    /* Remember the current non-cached end.  It would be
	       better to do this at the end, but since we have a
	       lookahead buffer, the file position at the end will
	       be one line ahead of where we want, so we keep track
	       of the last known end point after each line.  This
	       would be a great thing to fix.
	    */
	    psP->ps_nc_end = ftell( conP->mc_usefP );
	}
    }

    /* Update the cached body bytes total */
    if ( psP->ps_partP != NULL )
	mmsgP->mm_cacheC += psP->ps_partbsP->bs_bC;

    /* Check for error */
    if ( okF ) {
	if ( conP->mc_ifP != NULL )
	    if ( ferror( conP->mc_ifP ) ) {
		warning( "Error (%d) reading message body.", errno );
		okF = FALSE;
	}
	else {
	    /* For input stream, best we can do is return the end of data;
	       let caller analyze whether there was an error.
	    */
	    ;
	}
    }

#if 0
    /* If appropriate, remember the end of this body */
    if ( ( conP->mc_usefP != NULL ) && ( psP->ps_nc_start != 0 ) )
	psP->ps_nc_end = ftell( conP->mc_usefP );
#endif

    return ( okF );
}
/*

*//* msgpstr_size(mpP, psP, ncF)

	Get the size of a messagepart string

Accepts :

	mpP		Ptr to the message part
	psP		Ptr to the byte string part
	ncF		Whether to include non-cached data

Returns :

	<value>		The size of the part string

Notes :

	If ncF is FALSE, only the size of the cached part (the part
	  in BSTR storage) is returned.  If ncF is TRUE, the size
	  also includes any non-cached part.

	mpP is passed in case it might ever be needed.

*/

static ULONG
msgpstr_size ARGLIST( (mpP, psP, ncF) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
   NFARG( MSGPSTR	*psP )		/* Ptr to partstring struct */
    FARG( BOOL		ncF )		/* Include non-cached total */
{
	ULONG		count;

    /* Get size of cached portion */
    count = ( psP->ps_partbsP == NULL ) ? 0 : psP->ps_partbsP->bs_bC;

    /* If wanted, add size of non-cached portion */
    if ( ncF )
	count += ( psP->ps_nc_end - psP->ps_nc_start );

    return ( count );
}
/*

*//* msgpstr_write(mpP, psP, ofP, flags);

	Write a byte-string part

Accepts :

	mpP		Ptr to the message part
	psP		Ptr to the byte string part to write
	ofP		File ptr to write body to
	flags		Message writing flags

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msgpstr_write ARGLIST( (mpP, psP, ofP, flags) )
   NFARG( MSGPART	*mpP )		/* Ptr to message part */
   NFARG( MSGPSTR	*psP )		/* Struct for the byte string */
   NFARG( FILE		*ofP )		/* Where to write body */
    FARG( UWORD		flags )		/* Message-writing flags */
{
	BOOL		okF;
	MML_IN		*inP;		/* Input stack */
	MSGCON		*conP;		/* Message container */

    okF = TRUE;

    /* Don't do anything if there's nothing to do.  There will always
       be something buffered if there is anything in the partstr at all.
    */
    if ( ( psP->ps_partbsP == NULL ) ||
         ( psP->ps_partbsP->bs_bC == 0 ) )
	return ( TRUE );

    /* Create an input stack for the writing. */
    if ( ( inP = in_new() ) == NULL )
	return ( FALSE );

    /* If the entire body isn't cached and we've previously read to
       the end of the message, we have to do special stuff. 
    */
    if ( ( psP->ps_nc_end - psP->ps_nc_start ) > 0 ) {
	/* Seek to the non-cached start */
	conP = mpP->mp_mmsgP->mm_conP;
	if ( ( fseek( conP->mc_usefP, psP->ps_nc_start, SEEK_SET ) != 0 ) ||
	     ( ftell( conP->mc_usefP ) != psP->ps_nc_start ) ) {
	    warning( "msgpstr_write: seek failed to non-cached start." );
	    okF = FALSE;
	}

	/* Push the non-cached part onto the stack */
	else if ( !in_stream_push_cfile( inP,
				    conP->mc_usefP,
				    psP->ps_nc_end - psP->ps_nc_start,
				    "non-cached body" ) ) {
	    warning( "msgpstr_write: can't push non-cached part." );
	    okF = FALSE;
	}
    }

    /* Push any cached body onto the stack */
    if ( okF && ( psP->ps_partP != NULL ) &&
         ( psP->ps_partbsP->bs_bC > 0 ) )
	if ( !in_stream_push_str( inP,
				 (char *)psP->ps_partbsP->bs_bP,
				 psP->ps_partbsP->bs_bC,
				 "cached body" ) ) {
	    warning( "msgpstr_write: can't push cached part" );
	    okF = FALSE;
	}

    /* Use msgpart_write_stack() to output the byte stream. */
    okF = okF && msgpart_write_stack( mpP, inP, ofP, flags );

    /* Get rid of the input stack */
    in_destroy( inP );

    return ( okF );
}
