/* msgstore.c  -- stuff dealing with mail message storage

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains routines to deal with message storage for mvmf

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "msgstore.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "errwarn.h"

/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */

    /* List of msgstore namespaces */
	LL_HEAD	Msnslist;

    /* List of specific message store types. */
	LL_HEAD	Msstlist;


/* Local routines and forward references */



/* Private data */

    /* Text that introduces a bounce reason */
static	char	*Bounce_pre[] = {
"",
"-- This message is being returned with the following note:",
"",
NULL };

    /* Text to explain and identify the probable recipient */
static	char	*Bounce_recip[] = {
"",
"-- The [possibly-permuted] recipient address was:",
NULL };


    /* Text that follows the bounce reason */
static	char	*Bounce_post[] = {
"",
"-- This bounce is generated by the 'mvmf' mail processing agent.",
"-- If you feel that this agent is behaving improperly, please",
"-- feel free to report a bug to mvmf-bug@mv.com .  Abuse of the",
"-- agent can be reported to the intended recipient or to abuse@mv.com .",
"",
"== Original message follows. ==",
"",
NULL };


/*

*//* msfolder_fname(fnameP)

	Derive MSFOLDER info from a raw folder name

Accepts :

	fnameP		Raw folder name, e.g. as supplied by a user

Returns :

	<value>		ptr to MSFOLDER if successful,
			NULL if not (with appropriate warning(s))

Notes :

	returned value should be released via msfolder_release()

*/

MSFOLDER *
msfolder_fname ARGLIST( (fnameP) )
    FARG( UBYTE		*fnameP )	/* Raw folder name */
{
	MSNS		*msnsP;		/* Namespace ptr */
	MSFOLDER	*msfP;		/* Folder ptr */

    /* Find the namespace that this is a part of. */
    msnsP = msns_find_path( fnameP );
    if ( msnsP == NULL ) {
	warning( "No namespace contains \"%s\"", (char *)fnameP );
	return ( NULL );
    }

    /* Build and return a folder struct */
    msfP = (MSFOLDER *)emalloc( "msfolder_fname", "folder info",
					sizeof(MSFOLDER) );
    if ( msfP == NULL )
	return ( NULL );

    msfP->mf_msnsP = msnsP;		/* Ptr to namespace */
    msfP->mf_fnameP = fnameP;		/* Handy reference to raw name */
    msfP->mf_ftailP = fnameP + msnsP->ns_prefixlen;
					/* Tail (after namespace) */

    return ( msfP );
}
/*

*//* msfolder_release( msfP )

	Release a previously "gotten" msfolder

Accepts :

	msfP		Ptr to MSFOLDER previously gotten

Returns :

	<nothing>

Notes :


*/

void
msfolder_release ARGLIST( (msfP) )
    FARG( MSFOLDER	*msfP )		/* Ptr to folder info */
{
    dealloc( msfP );
}
/*

*//* msg_autoreplyable( ommP )

	See if a message is autoreplyable

Accepts :

	ommP		Open message pointer

Returns :

	<status>	TRUE if message is autoreplyable
			FALSE if not

Notes :

	An autoreply should not be attempted for a number of reasons;
	this routine checks for those reasons (or at least the ones
	we know about and agree with..)

	Many of these checks were derived from common practice; internet
	draft "draft-moore-auto-email-response" (now RFC3834) codified
	some of them and added some others.

*/

BOOL
msg_autoreplyable ARGLIST( (ommP ) )
    FARG( OMM		*ommP )		/* Open message pointer */
{
	int		i;
	int		hL;		/* Length of header name */
	void		*hdrP;		/* Header object */
	MSGPART		*mpP;
	REFSTR		*rsP;
	BSTR		*bsP;

    /* See if the ascertained envelope sender is not returnable-to.  We
       don't autoreply to the null sender, nor do we send to a number of
       known strings.
    */
    if ( ( ommP->om_aesenderP == NULL ) ||
         ( *ommP->om_aesenderP == NUL ) ||
	 ( strnicmp( ommP->om_aesenderP, "listserv@", 9 ) == 0 ) ||
	 ( strnicmp( ommP->om_aesenderP, "mailer-daemon@", 13 ) == 0 ) ||
	 ( strnicmp( ommP->om_aesenderP, "majordomo@", 10 ) == 0 ) ||
	 ( strnicmp( ommP->om_aesenderP, "postmaster@", 11 ) == 0 ) ||
	 ( strnicmp( ommP->om_aesenderP, "owner-", 6 ) == 0 ) )
	return ( FALSE );
    i = strlen( ommP->om_aesenderP );
    if ( ( i > 8 ) &&
         ( stricmp( &ommP->om_aesenderP[i-8], "-request" ) == 0 ) )
	return ( FALSE );

TODO    /* We might want to inhibit reply if there is no '@' in the
	   sender address. */


    /* Walk the top part's headers and see if we spot some header field
       that makes this non-autoreplyable.
    */
    mpP = ommP->om_mmsgP->mm_topmpP;
    for ( hdrP = NULL;
	 ( hdrP = msgpart_hdr_next( mpP, hdrP, NULL ) ) != NULL;
		) {

	rsP = msgpart_hdr_str( mpP, hdrP );
	bsP = refstr_bstr( rsP );
	hL = hdr_name_skip_estr( estr_tmp_str( bsP->bs_bP ), 0 );

	if ( strnicmp( (char *)bsP->bs_bP, "Auto-Submitted", hL ) == 0 ) {
	    /* Do not send if there is an Auto-Submitted header with
	       value other than "no".
	    */
	    if ( !strmatch_exact( NULL,
				 (UBYTE *)"no", &bsP->bs_bP[ bsP->bs_bX ],
				 MF_SUBSTR|MF_IGNCASE ) )
		return ( FALSE );
	}

	/* Check RFC2919 List-ID header */
	else if ( strnicmp( (char *)bsP->bs_bP, "List-ID", hL ) == 0 )
	    return ( FALSE );

	/* Check RFC2369 list headers */
	else if ( strnicmp( (char *)bsP->bs_bP, "List-Archive", hL ) == 0 ||
	          strnicmp( (char *)bsP->bs_bP, "List-Help", hL ) == 0 ||
	          strnicmp( (char *)bsP->bs_bP, "List-Owner", hL ) == 0 ||
	          strnicmp( (char *)bsP->bs_bP, "List-Post", hL ) == 0 ||
	          strnicmp( (char *)bsP->bs_bP, "List-Subscribe", hL ) == 0 ||
	          strnicmp( (char *)bsP->bs_bP, "List-Unsubscribe", hL ) == 0 )
	    return ( FALSE );

	/* Check Precedence (mentioned in RFC2076 but commonly used
	   to inhibit autoreplies)
	*/
	else if ( strnicmp( (char *)bsP->bs_bP, "Precedence", hL ) == 0 ) {
	    if ( strmatch_exact( NULL,
				 (UBYTE *)"bulk", &bsP->bs_bP[ bsP->bs_bX ],
				 MF_SUBSTR|MF_IGNCASE ) )
		return ( FALSE );
	    if ( strmatch_exact( NULL,
				 (UBYTE *)"list", &bsP->bs_bP[ bsP->bs_bX ],
				 MF_SUBSTR|MF_IGNCASE ) )
		return ( FALSE );
	}
    }

    /* Can't find a reason why this isn't autoreplyable. */
    return ( TRUE );
}
/*

*//* msg_bounce( ommP, reasonP )

	Bounce the message, with explanatory text

Accepts :

	ommP		Open message pointer
	reasonP		Prologue text with a reason for the bounce

Returns :

	<status>	TRUE if successfully handled,
			FALSE if not

Notes :

	Checks bounceability via msg_autoreplyable().

	Should probably do MDN encoding.  (Certainly RFC3028
	wants us to)

*/

BOOL
msg_bounce ARGLIST( (ommP, reasonP ) )
   NFARG( OMM		*ommP )		/* Open message pointer */
    FARG( char		*reasonP )	/* Text about the bounce */
{
	BOOL		okF;
	int		err;		/* Errno copy */
	int		len;
	MSGPART		*mpP;
	void		*hdrP;		/* Header object */
	REFSTR		*rsP;		/* REFSTR ptr */
	BSTR		*bsP;		/* BSTR ptr */
	FILE		*msgfP;		/* FILE handle for message */
	char		**msgPP;	/* Message text. */

    /* Don't bounce twice, and check autoreplyability */
    if ( ommP->om_bouncedF || ( !msg_autoreplyable( ommP ) ) ) {
	++ommP->om_deliveryC;		/* Pretend we delivered */
	ommP->om_bouncedF = TRUE;	/* And that we bounced */
	return ( TRUE );
    }

    mpP = ommP->om_mmsgP->mm_topmpP;	/* Top message part */
    mpP->mp_mmsgP->mm_oqfromF = FALSE;	/* Don't quote "From " lines */
    mpP->mp_mmsgP->mm_cnlC = 1;		/* Beginning; pretend newline */

    msgfP = msg_send_begin( ommP, NULL, ommP->om_aesenderP );
    if ( msgfP == NULL ) {
	omm_delivery_error( ommP, errno, "bounce" );
	return ( FALSE );
    }
    okF = TRUE;

    /* Write headers for new message */
    okF = okF && ( fprintf( msgfP, "Return-Path: <>\n" ) > 0 );
    okF = okF && ( fprintf( msgfP, "From: MAILER-DAEMON\n" ) > 0 );
    okF = okF && ( fprintf( msgfP, "To: <%s>\n", ommP->om_aesenderP ) > 0 );

    if ( okF ) {
	/* Construct a subject. */
	hdrP = msgpart_hdr_next( mpP, NULL, estr_tmp_str((UBYTE *)"Subject") );
	if ( hdrP != NULL ) {
	    rsP = msgpart_hdr_str( mpP, hdrP );
	    bsP = refstr_bstr( rsP );
	    okF = ( fprintf( msgfP, "Subject: Auto: %s\n",
			    &bsP->bs_bP[rsP->rs_bX] ) > 0 );
	}
	else {
	    okF = ( fprintf( msgfP, "Subject: Auto: Rejected mail\n" ) > 0 );
	}
    }

    /* Write the appropriate references headers */
    okF = okF && msgpart_write_references( mpP, msgfP );

    /* Add a "Date: " header */
    okF = okF && ( fprintf( msgfP,
			    "Date: %s\n",
			    msgpart_date_string_now() ) > 0 );

    /* Add "Auto-submitted" header per RFC3834 */
    okF = okF && ( fprintf( msgfP, "Auto-Submitted: auto-replied\n" ) > 0 );
    
    /* Write bounce prolog */
    for ( msgPP = &Bounce_pre[0]; okF && ( *msgPP != NULL ); ++msgPP )
	okF = okF && ( fprintf( msgfP, "%s\n", *msgPP ) > 0 );

    /* Write bounce reason */
    okF = okF && ( fprintf( msgfP, "%s\n\n", reasonP ) > 0 );

    /* Write any cached error text */
    okF = okF && omm_write_errs( ommP, msgfP, FALSE );

    /* Try to identify the recipient if possible */
    if ( okF && ( E_RecipP != NULL ) &&
	        ( ( len = refstr_len( E_RecipP ) ) > 0 ) ) {
	for ( msgPP = &Bounce_recip[0]; okF && ( *msgPP != NULL ); ++msgPP )
	    okF = okF && ( fprintf( msgfP, "%s\n", *msgPP ) > 0 );
	okF = okF && ( fprintf( msgfP, "    " ) > 0 );
	okF = okF && fwrite( refstr_bstr( E_RecipP )->bs_bP,
			    1, len, msgfP ) == len;
	okF = okF && ( fprintf( msgfP, "\n" ) > 0 );
    }

    /* Write rest of bounce prolog */
    for ( msgPP = &Bounce_post[0]; okF && ( *msgPP != NULL ); ++msgPP )
	okF = okF && ( fprintf( msgfP, "%s\n", *msgPP ) > 0 );

    /* Copy in the headers from the old message */
    okF = okF && msgpart_write_hdrs( mpP, msgfP, TRUE );

    /* Write the body.  Note- for a bounce, we don't really care if
       there's a failure writing the body.
    */
    if ( okF )
        msgpart_write_body( mpP, msgfP, TRUE );
     
    /* Be done */
    if ( !okF )
	err = errno;			/* Save copy of errno */

    if ( !msg_send_end( ommP, msgfP ) ) {
	if ( okF ) {
	    err = errno;		/* This is first error, save errno */
	    okF = FALSE;
	}
    }

    if ( okF ) {
	++ommP->om_deliveryC;		/* Count the delivery */
	ommP->om_bouncedF = TRUE;	/* Remember that we bounced */
    }
    else {
	warning( "Failure sending bounce." );
	omm_delivery_error( ommP, err, "bounce" );
    }

    return ( okF );
}
/*

*//* msg_forward( ommP, rnameP, keepfromF )

	Forward the message to a new recipient

Accepts :

	ommP		Open message pointer
	rnameP		Name of the recipient
	keepfromF	Whether to keep the original envelope From

Returns :

	<status>	TRUE if successfully forwarded,
			FALSE if not

Notes :

	Does not change the header.

	If keepfromF is set, be more explicit about preserving the
	original envelope "From" (i.e., use the "ascertained from"
	address when invoking the mailer)

*/

BOOL
msg_forward ARGLIST( (ommP, rnameP, keepfromF ) )
   NFARG( OMM		*ommP )		/* Open message pointer */
   NFARG( char		*rnameP )	/* Recipient address */
    FARG( BOOL		keepfromF )	/* "keep from" flag */
{
	BOOL		okF;
	int		err;		/* Copy of errno */
	MSGPART		*mpP;
	FILE		*msgfP;		/* File PTR for message */

    /* Don't forward if we're already forwarding: e.g. to prevent
       recursion inside of any hooks that we might use.
    */
    if ( ommP->om_fwdingF ) {
	warning( "recursive forwarding: don't forward inside forwarding hooks!" );
	return ( FALSE );
    }

    okF = TRUE;
    mpP = ommP->om_mmsgP->mm_topmpP;	/* Top message part */

    ommP->om_fwdingF = TRUE;

    mpP->mp_mmsgP->mm_oqfromF = FALSE;	/* Don't quote "From " lines */
    mpP->mp_mmsgP->mm_cnlC = 1;		/* Beginning; pretend newline */

    if ( keepfromF && ( ommP->om_aesenderP != NULL ) )
	msgfP = msg_send_begin( ommP, ommP->om_aesenderP, rnameP );
    else
	msgfP = msg_send_begin( ommP, NULL, rnameP );

    if ( msgfP == NULL ) {
	omm_delivery_error( ommP, errno, "pipe" );
	ommP->om_fwdingF = FALSE;
	return ( FALSE );
    }

    /* Write the headers */
    okF = okF && msgpart_write_hdrs( mpP, msgfP, FALSE );

    /* Write any cached errors */
    okF = okF && omm_write_errs( ommP, msgfP, TRUE );

    /* Write the body */
    okF = okF && msgpart_write_body( mpP, msgfP, FALSE );

    /* Be done */
    if ( !okF )
	err = errno;			/* Save copy of errno */

    if ( !msg_send_end( ommP, msgfP ) ) {
	if ( okF ) {
	    err = errno;		/* This is first error, save errno */
	    okF = FALSE;
	}
    }

    if ( okF ) {
	++ommP->om_deliveryC;		/* Count the delivery */
    }
    else {
	warning( "Failure forwarding message to \"%s\"", rnameP );
	omm_delivery_error( ommP, err, "forward" );
    }

    ommP->om_fwdingF = FALSE;		/* No longer forwarding */

    return ( okF );
}
/*

*//* msg_pipe( ommP, cmdP )

	Write the current message to a pipe

Accepts :

	ommP		Open message pointer
	cmdP		Command string for pipe

Returns :

	<status>	TRUE if successfully written,
			FALSE if not

Notes :

	This is an initial and quick stab at a pipe implementation,
	something to hold me while some other things are being
	thought about.  This simply uses popen and doesn't do
	much if anything with the pipe output.  I expect to
	supplant this with real pipe creation, plus a way to
	access system-defined plug-ins instead of user-defined pipes.
	
*/

BOOL
msg_pipe ARGLIST( (ommP, cmdP) )
   NFARG( OMM		*ommP )		/* Open message pointer */
    FARG( char		*cmdP )		/* Pipe command */
{
	BOOL		status;
	int		err;		/* Copy of errno */
	int		coutfd;		/* fd for output to child */
	int		cinfd;		/* fd for input from child */
	int		cerrfd;		/* fd for error in from child */
	pid_t		pid;		/* Child pid */
	char		*argv[5];	/* Argument vector */
	MSGPART		*mpP;
	FILE		*msgfP;		/* Pipe handle */

    /* Don't pipe if we're already piping (e.g. to prevent recursion
       inside of any hooks that might be called here).
    */
    if ( ommP->om_pipingF ) {
	warning( "recursive piping: don't pipe inside piping hooks!" );
	return ( FALSE );
    }
    ommP->om_pipingF = TRUE;

    /* Exec & get pipes to child.  We are going to cheat here and
       use the shell to decode the command buffer.  Maybe someday
       we can do better.
    */
    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = cmdP;
    argv[3] = NULL;
    pid = pexec_std( "/bin/sh", &argv[0], &coutfd, &cinfd, &cerrfd );
    if ( pid < 0 ) {
	err = errno;
	warning( "msg_pipe can not open \"%s\"", cmdP );
	omm_delivery_error( ommP, err, "pipe" );
	ommP->om_pipingF = FALSE;
	return ( FALSE );
    }
    msgfP = fdopen( coutfd, "w" );

    /* Write the headers */
    mpP = ommP->om_mmsgP->mm_topmpP;
    mpP->mp_mmsgP->mm_oqfromF = FALSE;	/* Don't quote "From " lines */
    mpP->mp_mmsgP->mm_cnlC = 1;		/* Beginning; pretend newline */

    if ( status = msgpart_write_hdrs( mpP, msgfP, FALSE ) ) {
	/* Write any cached errors */
	if ( status = omm_write_errs( ommP, msgfP, TRUE ) ) {
	    /* Write the body */
	    status = msgpart_write_body( mpP, msgfP, FALSE );
	}
    }

    if ( status )
	status = (fclose(msgfP) == 0);
    else {
	err = errno;
	omm_delivery_error( ommP, err, "pipe" );
	fclose(msgfP);
    }

    /* Be done with child.  Note we already closed the output. */
    pexec_std_end( pid, -1, cinfd, cerrfd );

    if ( status ) {
	++ommP->om_deliveryC;		/* Count the delivery */
    }

    ommP->om_pipingF = FALSE;

    return ( status );
}
/*

*//* msg_send_begin( ommP, fromP, toP )

	Begin sending a message

Accepts :

	ommP		Open message pointer
	fromP		"From" address, or NULL
	toP		Who to send the message to ("To" address)

Returns :

	<value>		FILE ptr for the started message, 
			NULL if failure

Notes :

	The "fromP" may be passed as non-NULL to override supplying an
	envelope sender.

	Caller should continue sending by writing the headers, a
	blank line, and the body, and then calling msg_send_end().

	Logs an error if failure, preserving errno for the caller

*/

FILE *
msg_send_begin ARGLIST( (ommP, fromP, toP) )
   NFARG( OMM		*ommP )		/* Open message pointer */
   NFARG( char		*fromP )	/* Envelope "from" */
    FARG( char		*toP )		/* Destination */
{
	int		err;		/* Copy of errno */
	FILE		*msgfP;		/* File PTR for message */
	char		cmd[PATH_MAX +1];  /* Command to execute */

#ifdef	QMAIL			/* If using qmail */

    /* Build a command for qmail */
    if ( fromP != NULL )
	snprintf( &cmd[0], PATH_MAX, "%s/bin/qmail-inject -f \"%s\" \"%s\"", 
			QMAILDIR, fromP, toP );
    else
	snprintf( &cmd[0], PATH_MAX, "%s/bin/qmail-inject \"%s\"", 
			QMAILDIR, toP );

#else				/* Not using qmail */

    /* Build command for sendmail */
    if ( fromP != NULL )
	snprintf( &cmd[0], PATH_MAX, "%s -i -f \"%s\" \"%s\"",
			SENDMAIL, fromP, toP );
    else
	snprintf( &cmd[0], PATH_MAX, "%s -i \"%s\"",
			SENDMAIL, toP );

#endif	/* QMAIL */

    /* Open it up */
    msgfP = popen( &cmd[0], "w" );
    if ( msgfP == NULL ) {
	err = errno;
	warning( "msg_send_begin can not open for command:  %s", &cmd[0] );
	errno = err;
    }

    return ( msgfP );
}
/*

*//* msg_send_end(ommP, msgfP)

	Finish sending a message

Accepts :

	ommP		Open message pointer
	msgfP		Open mail message ptr e.g. previously obtained
			  via msg_send_begin()

Returns :

	<value>		TRUE if OK
			FALSE if some error.

Notes :


*/

BOOL
msg_send_end ARGLIST( (ommP, msgfP) )
   NFARG( OMM		*ommP )		/* Open message pointer */
    FARG( FILE		*msgfP )	/* FILE ptr for message being sent */
{
    if ( fclose( msgfP ) != 0 )
	return ( FALSE );
    return ( TRUE );
}
/*

*//* msg_write( ommP, fnameP )

	Write the current message to a folder

Accepts :

	ommP		Open message pointer
	fnameP		Name of the folder (namespace + path suffix)

Returns :

	<status>	TRUE if successfully written,
			FALSE if not

Notes :


*/

BOOL
msg_write ARGLIST( (ommP, fnameP) )
   NFARG( OMM		*ommP )		/* Open message pointer */
    FARG( char		*fnameP )	/* Folder name */
{
	BOOL		okF;
	MSFOLDER	*msfP;		/* Folder handle */
	MSGT		*msgtP;		/* General type handler */
	MSST		*msstP;		/* Specific type handler */
	MSMSG		*msmsgP;	/* Mailstore message handle */

    /* Don't write if we're already storing: e.g. to prevent
       recursion inside of any hooks that we might use.
    */
    if ( ommP->om_storingF ) {
	warning( "recursive storing: don't store inside storing hooks!" );
	return ( FALSE );
    }

    /* Get first stab at message info for this name */
    if ( ( msmsgP = msmsg_fname( fnameP ) ) == NULL )
	return ( FALSE );

    msfP = msmsgP->mm_folderP;		/* Folder ptr, for convenience */

    /* General folder type is known at this point, even if we don't yet
       know the specific type.
    */
    msgtP = msfP->mf_msnsP->ns_msstP->st_msgtP;

    /* Use the general type to do pre-access stuff.  If successful this
       will narrow down a specific type as well.
    */
    if ( !(*msgtP->gt_folder_preacc)( msmsgP ) )
	return ( FALSE );

    /* Now set to go. */
    ommP->om_storingF = TRUE;
    msstP = msmsgP->mm_msstP;

    /* Hand off to the specific type handler */
    okF = (*msstP->st_msg_write)( ommP, msmsgP );

TODO	/* Get error/status info here */

    /* Get rid of message/folder info */
    msmsg_release( msmsgP );

    ommP->om_storingF = FALSE;

    return ( okF );
}
/*

*//* msgstore_init()

	Initialize the msgstore module

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgstore_init NOARGLIST
{
    /* Initialize the namespace list */
    ll_init( &Msnslist );

    /* Initialize the specific msgstore types list */
    ll_init( &Msstlist );

    /* Initialize general storage types handlers */
    if ( !msgt_file_init() )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* msmsg_fname(fnameP)

	Get an MSMSG from a raw folder name

Accepts :

	fnameP		Raw folder name, e.g. as supplied by a user

Returns :

	<value>		ptr to MSMSG if successful,
			NULL if not (with appropriate warning(s))

Notes :

	Also determines other info such as the attached MSFOLDER.

	returned value should be released via msmsg_release()

*/

MSMSG *
msmsg_fname ARGLIST( (fnameP) )
    FARG( UBYTE		*fnameP )	/* Raw folder name */
{
	MSFOLDER	*msfP;		/* Folder ptr */
	MSMSG		*msmsgP;	/* MSMSG ptr */

    /* Get folder info */
    if ( ( msfP = msfolder_fname( fnameP ) ) == NULL )
	return ( FALSE );

    /* Build MSMSG info */
    msmsgP = (MSMSG *)emalloc( "msmsg_fname", "msmsg info",
					sizeof(MSMSG) );
    if ( msmsgP == NULL ) {
	msmsg_release( msmsgP );
	return ( NULL );
    }

    msmsgP->mm_folderP = msfP;		/* Ptr to folder info */
    msmsgP->mm_msstP = NULL;		/* No specific type override yet */
    msmsgP->mm_gtdataP = NULL;		/* No general type data yet */
    msmsgP->mm_stdataP = NULL;		/* No specific type data yet */

    return ( msmsgP );
}
/*

*//* msmsg_folder_tail( msmsgP )

	Get the translated tail part of the file folder

Accepts :

	msmsgP		Ptr to msgstore message handle

Returns :

	<value>		Ptr to the tail string

Notes :

	This function is responsible for doing any user-desired
	translations on the tail portion of the folder.  If
	this translation fails, the unaltered tail is returned instead.

	Note that the returned string is a tmpstr(), so it should be
	consumed ASAP.

*/

UBYTE *
msmsg_folder_tail ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
{
	STSCODE		status;
	int		len;
	UBYTE		*stailP;	/* Source tail ptr */
	UBYTE		*ttailP;	/* Translated tail ptr */
	ESTR		*retP;		/* Returned string value */
	MSNS		*msnsP;		/* Namespace ptr */
	VALUE		tailval;
	char		mflcmd[200];	/* Buffer for MFL command */

    msnsP = msmsgP->mm_folderP->mf_msnsP; /* Namespace ptr */

    /* Get the source tail ptr */
    stailP = msmsgP->mm_folderP->mf_ftailP;

    /* Assume no translated tail */
    ttailP = NULL;

    /* If there's MFL code to translate, apply it */
    if ( msnsP->ns_transhook.as_strP != NULL ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		       "%s( \"%s\" );",
		       msnsP->ns_transhook.as_strP,
		       stailP ) >= sizeof(mflcmd) ) {
	    warning( "msns_folder_tail: buffer not large enough for MFL command." );
	}
	else {
	    status = mfl_run_str( &tailval, estr_tmp_str( &mflcmd[0] ),
				     estr_tmp_str( "Mc" ) );
	    if ( status == SCOK ) {
		/* Get the ESTR from the returned value */
		retP = val_estr_tmp( &tailval );

		/* Get NUL-terminated string to work with. */
		if ( retP != NULL ) {
		    len = estr_len( retP );
		    ttailP = tmpstr( len +1 );
		    if ( ttailP != NULL ) {
			memcpy( ttailP, retP->es_strP, len );
			ttailP[len] = NUL;
		    }
		}

		val_release( &tailval );
	    }
	}
    }

    /* If we got a translated tail string, use that. */
    if ( ttailP != NULL )
	stailP = ttailP;

    /* Return the tail we have */
    return ( stailP );
}
/*

*//* msmsg_release( msmsgP )

	Release a previously "gotten" MSMSG

Accepts :

	msmsgP			Ptr to MSMSG previously gotten

Returns :

	<nothing>

Notes :

	Releases any contained info such as the associated MSFOLDER.

*/

void
msmsg_release ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Ptr to MSMSG info */
{
    msfolder_release( msmsgP->mm_folderP );

    /* Get rid of general and specific type data */
    if ( msmsgP->mm_msstP != NULL ) {
	(*msmsgP->mm_msstP->st_stdata_release)( msmsgP->mm_stdataP );
	(*msmsgP->mm_msstP->st_gtdata_release)( msmsgP->mm_gtdataP );
    }

    dealloc( msmsgP );
}
/*

*//* msns_find( prefixP )

	Look up a namespace by specific prefix

Accepts :

	prefixP		Ptr to the prefix to look up

Returns :

	<value>		Ptr to MSNS if found,
			NULL if not.

Notes :


*/

MSNS *
msns_find ARGLIST( (prefixP) )
    FARG( UBYTE		*prefixP )
{
	LL_NODE		*nodeP;
	MSNS		*msnsP;

    /* Walk the list looking for this prefix */
    mml_mutex_lock( &Msnslist.llh_lock );
    for ( nodeP = Msnslist.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	msnsP = (MSNS *)nodeP->lln_valP;
	if ( strcmp( (char *)msnsP->ns_prefixP, (char *)prefixP ) == 0 )
	    break;
    }

    mml_mutex_unlock( &Msnslist.llh_lock );

    if ( nodeP == NULL )
	msnsP = NULL;

    return ( msnsP );
}
/*

*//* msns_find_path( pathP )

	Look up a namespace corresponding to a path

Accepts :

	pathP		Ptr to the path to find a namespace for.

Returns :

	<value>		Ptr to MSNS if found,
			NULL if not.

Notes :


*/

MSNS *
msns_find_path ARGLIST( (pathP) )
    FARG( UBYTE		*pathP )
{
	int		pathlen;
	LL_NODE		*nodeP;
	MSNS		*msnsP;

    pathlen = strlen( (char *)pathP );

    /* Walk the list to find a prefix within this path.
       The list is ordered by prefix length.
    */
    mml_mutex_lock( &Msnslist.llh_lock );
    for ( nodeP = Msnslist.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	msnsP = (MSNS *)nodeP->lln_valP;

	if ( ( msnsP->ns_prefixlen <= pathlen ) &&
	     ( memcmp( msnsP->ns_prefixP, pathP, msnsP->ns_prefixlen ) == 0 ) )
	    break;
    }

    mml_mutex_unlock( &Msnslist.llh_lock );

    if ( nodeP == NULL )
	msnsP = NULL;

    return ( msnsP );
}
/*

*//* msns_register( prefixP )

	Register a new namespace by a prefix.

Accepts :

	prefixP		Ptr to the prefix to register.

Returns :

	<value>		Ptr to MSNS if found,
			NULL if not.

Notes :

	It is an error to register two namespaces with the same prefix.

	However, one namespace may be a substring of another.  When
	looking up namespaces, the longest one wins.

*/

MSNS *
msns_register ARGLIST( (prefixP) )
    FARG( UBYTE		*prefixP )	/* Prefix to register */
{
	int		prefixlen;
	LL_NODE		*inodeP;	/* Node to insert before */
	LL_NODE		*nodeP;		/* New node */
	MSNS		*msnsP;

    prefixlen = strlen( (char *)prefixP );

    /* Find the spot in the list where the new prefix will go.  The
       list is ordered by descending prefix length; this facilitates
       the "longest prefix wins" aspect (and allows a null prefix
       as the namespace of last resort).
    */
    mml_mutex_lock( &Msnslist.llh_lock );
    for ( inodeP = Msnslist.llh_firstP; inodeP != NULL;
		inodeP = inodeP->lln_nextP ) {
	msnsP = (MSNS *)inodeP->lln_valP;

	/* Done looping if we found the insertion point. */
	if ( prefixlen > msnsP->ns_prefixlen )
	    break;

	/* If same length, check for duplicate. */
	if ( ( prefixlen == msnsP->ns_prefixlen ) &&
	     ( memcmp( prefixP, msnsP->ns_prefixP, prefixlen ) == 0 ) ) {
	    warning( "msns_register: duplicate prefix \"%s\".",
		      (char *)prefixP );
	    mml_mutex_unlock( &Msnslist.llh_lock );
	    return ( NULL );
	}
    }

    /* Make new block */
    msnsP = (MSNS *)emalloc( "msns_register", "MSNS block", sizeof(MSNS) );
    if ( msnsP == NULL )
	return ( NULL );

    msnsP->ns_prefixP = (UBYTE *)newstr( "msns_register", "prefix", prefixP );
    if ( msnsP->ns_prefixP == NULL )
	return ( NULL );

    msnsP->ns_prefixlen = prefixlen;
    attr_str_unset( &msnsP->ns_transhook );
    msnsP->ns_msstP = NULL;
    msnsP->ns_inboxF = FALSE;
    msnsP->ns_mustsstF = FALSE;
    msnsP->ns_gtdataP = NULL;
    msnsP->ns_stdataP = NULL;
    

    /* Insert prior to the insert-before node that we remembered */
    nodeP = ll_addvb( &Msnslist, inodeP, msnsP );
    if ( nodeP == NULL ) {
	msnsP = NULL;			/* Some error, just bail */
    }

    mml_mutex_unlock( &Msnslist.llh_lock );

    return ( msnsP );
}
/*

*//* msst_find( nameP )

	Look up a specific msgstore type, by name

Accepts :

	Name to look up

Returns :

	<value>		Ptr to MSST definition if found,
			NULL if not.

Notes :

	Lookup is case insensitive

*/

MSST *
msst_find ARGLIST( (nameP) )
    FARG( UBYTE		*nameP )	/* Name to look up */
{
	int		cmpval;
	LL_NODE		*nodeP;
	MSST		*msstP;

    mml_mutex_lock( &Msstlist.llh_lock );

    /* Walk the list looking for this name.  List is sorted by name,
       so we can stop early if we pass where it should be.
    */
    for ( nodeP = Msstlist.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	msstP = (MSST *)nodeP->lln_valP;

	cmpval = stricmp( nameP, msstP->st_nameP );
	if ( cmpval <= 0 )
	    break;
    }

    mml_mutex_unlock( &Msstlist.llh_lock );

    if ( ( cmpval != 0 ) || ( nodeP == NULL ) )
	msstP = NULL;

    return ( msstP );
}
/*

*//* msst_register( newmsstP )

	Register a new specific msgstore type, by name

Accepts :

	newmsstP	Ptr to MSST definition to register

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Warning is given for any error.

	MSST is register by name, case-insensitive.

*/

BOOL
msst_register ARGLIST( (newmsstP) )
    FARG( MSST		*newmsstP )	/* New MSST to register */
{
	BOOL		okF;
	int		cmpval;
	LL_NODE		*nodeP;
	MSST		*msstP;

    mml_mutex_lock( &Msnslist.llh_lock );

    /* Walk the list to find out where to insert the new one (also
       looking for already registered name).  The list is sorted by
       name, case-insensitive.
    */
    for ( nodeP = Msstlist.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	msstP = (MSST *)nodeP->lln_valP;

	cmpval = stricmp( newmsstP->st_nameP, msstP->st_nameP );
	if ( cmpval <= 0 )
	    break;
    }

    if ( ( nodeP != NULL ) && ( cmpval == 0 ) ) {
	warning( "msst_register: duplicate name \"%s\"", newmsstP->st_nameP );
	okF = FALSE;
    }
    else {
	/* Add new block prior to the one we located. */
	okF = ( ll_addvb( &Msstlist, nodeP, newmsstP ) != NULL );
    }

    mml_mutex_unlock( &Msnslist.llh_lock );

    return ( okF );
}
