/* msgt-file.c -- msgstore generic type "file"

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff to handle the msgstore generic type "file"

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "mvmf-locking.h"
#include "msgstore.h"

#include "msgt-file.h"
#include "msst-file-maildir.h"
#include "msst-file-mbox.h"

/* Local definitions */



/* External data referenced */

extern	LL_HEAD	Msstlist;		/* List of specific types */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Handlers for msgt */
static	BOOL	msgt_file_folder_preacc PROTO( (MSMSG *msmsgP) );

    /* Various support functions */
static	BOOL	attr_str_set_layout();
static	BOOL	attr_str_set_locking();
static	BOOL	attr_str_set_path();
static	MSGT_FILE_DATA *file_data_get PROTO( (MSGT_FILE_DATA *mfdataP) );
static	void	file_data_release PROTO( (MSGT_FILE_DATA *mfdataP) );
static	BOOL	file_folder_check PROTO( (MSMSG *msmsgP, MSST **msstPP) );
static	BOOL	file_folder_path PROTO( (MSMSG *msmsgP) );
static	BOOL	file_folder_path_fs PROTO( (MSMSG *msmsgP) );
static	BOOL	file_folder_path_mdpp PROTO( (MSMSG *msmsgP) );

/* Private data */

    /* Template general data */
static	MSGT_FILE_DATA	*MfdataP = NULL;

    /* The MSGT definition for the msgt_file type */
/* static */	MSGT	Msgt_file = {
	"file",				/* name of type */

		/* handlers */
	NULL,				/* Set integer attribute */
	NULL,				/* Set string attribute */
	msgt_file_folder_preacc		/* Any pre-access folder processing */
};


    /* integer attributes dispatch table */
static struct {
    char	*ia_nameP;		/* Name */
    BOOL	(*ia_set)();		/* Function for setting */
}  Msgt_file_attrs_int[] = {

{	NULL,		NULL		}
};


    /* string attributes dispatch table */
static struct {
    char	*sa_nameP;		/* Name */
    BOOL	(*sa_set)();		/* Function for setting */
}  Msgt_file_attrs_str[] = {

{	"layout",	attr_str_set_layout	},
{	"locking",	attr_str_set_locking	},
{	"path",		attr_str_set_path	},

{	NULL,		NULL		}
};

/*

*//* msgt_file_init()

	Initialize the msgt_file stuff

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgt_file_init NOARGLIST
{
    /* Create template general type data.  I suppose it could be
       statically defined but it's easier to initialize this way.
    */
    MfdataP = emalloc( "msgt_file_init", "MfdataP",
		        sizeof( MSGT_FILE_DATA ) );
    if ( MfdataP == NULL )
	return ( FALSE );
    MfdataP->gf_fP = NULL;		/* File handle */
    MfdataP->gf_mltype = ML_UNK;	/* Folder layout type */
    MfdataP->gf_oqfromF = FALSE;	/* Quote "From " lines? */
    MfdataP->gf_blendF = FALSE;		/* Blank line at end required? */
    MfdataP->gf_afproot[0] = NUL;	/* Any folder path root */
    MfdataP->gf_afp[0] = NUL;		/* Absolute folder path */
    MfdataP->gf_aflkp[0] = NUL;		/* Absolute path to folder lock */
    MfdataP->gf_locks.ls_lockC = 0;	/* Locks */

    attr_str_unset( &MfdataP->gf_locking );  /* How to do locking */


    /* Initialize the specific types.  There should be a better way
       to get this done without hardcoding it here, but this is a
       start.
    */
    if ( !msst_file_maildir_init() )
	return ( FALSE );
    if ( !msst_file_mbox_init() )
	return ( FALSE );

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                          MSGT HANDLERS                              *
 *                                                                     *
 *   These are handler functions which support msgt functions for      *
 *   this type.                                                        *
 *                                                                     *
 ***********************************************************************/

/*

*//* msgt_file_folder_preacc( msmsgP )

	Called prior to accessing a folder.

Accepts :

	msmsgP		Ptr to msgstore message handle to pre-access

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msgt_file_folder_preacc ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
{
	MSGT_FILE_DATA	*mfdataP;
	MSFOLDER	*msfP;		/* copy of MSFOLDER handle */
	MSNS		*msnsP;		/* copy pf namespace handle */
	MSST		*msstP;		/* specific type handle */

    msfP = msmsgP->mm_folderP;
    msnsP = msfP->mf_msnsP;

    /* If the folder exists, we use whatever specific type it is.  Note
       there may be some timing windows here, between checking the folder
       type and using it later.
    */
    if ( !file_folder_check( msmsgP, &msstP ) )
	/* Error, e.g. file of unknown folder type, or bad access */
	return ( FALSE );
    if ( msstP == NULL ) {
	/* No existing folder or no specific type; use the namespace default */
	if ( ( msstP = msnsP->ns_msstP ) == NULL ) {
	    /* Bug here. */
	    warning( "msgt_file_folder_preacc: can't default the specific type." );
	    return ( FALSE );
	}
    }
    msmsgP->mm_msstP = msstP;

    /* Instantiate the general and specific data; use namespace
       data if present.
    */
    msmsgP->mm_gtdataP = (*msstP->st_gtdata_get)( msnsP->ns_gtdataP );
    if ( msmsgP->mm_gtdataP == NULL ) {
	warning( "msgt_file_folder_preacc: can't instantiate general type data" );
	return ( FALSE );
    }
    msmsgP->mm_stdataP = (*msstP->st_stdata_get)( msnsP->ns_stdataP );
    if ( msmsgP->mm_stdataP == NULL ) {
	warning( "msgt_file_folder_preacc: can't instantiate specific type data" );
	return ( FALSE );
    }

    /* Calculate the full absolute path to the folder, based on layout */
    if ( !file_folder_path( msmsgP ) ) {
	/* Should only happen if there's a bug, eh? */
	return ( FALSE );
    }

    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                FOR USE BY MSST OF THIS TYPE                         *
 *                                                                     *
 *   These are functions that may be used by specific types that are   *
 *   of this general type.                                             *
 *                                                                     *
 ***********************************************************************/


/*

*//* msgt_file_attr_int_set( mfdataP, nameP, val )

	Set an integer attribute value for this general type

Accepts :

	mfdataP		Ptr to general file data
	nameP		The attribute name
	val		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgt_file_attr_int_set ARGLIST( (mfdataP, nameP, val) )
   NFARG( MSGT_FILE_DATA *mfdataP )	/* General data for file types */
   NFARG( char		*nameP )	/* Attribute */
    FARG( long		val )		/* The value to set */
{
	int		i;

    /* Search the table for the name */
    for ( i = 0; Msgt_file_attrs_int[i].ia_nameP != NULL; ++i ) {
	if ( stricmp( nameP, Msgt_file_attrs_int[i].ia_nameP ) == 0 ) {
	    /* Found it-- hand off from here. */
	    return ( (*Msgt_file_attrs_int[i].ia_set)( mfdataP, val ) );
	}
    }

    /* Not found. */
    warning( "unknown integer attribute name: \"%s\"", nameP );
    return ( FALSE );
}
/*

*//* msgt_file_attr_str_set( mfdataP, nameP, valP )

	Set a string attribute value for this specific type

Accepts :

	mfdataP		Ptr to general file data
	nameP		The attribute name
	valP		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgt_file_attr_str_set ARGLIST( (mfdataP, nameP, valP) )
   NFARG( MSGT_FILE_DATA *mfdataP )	/* General data for file types */
   NFARG( char		*nameP )	/* Attribute */
    FARG( char		*valP )		/* The value to set */
{
	int		i;

    /* Search the table for the name */
    for ( i = 0; Msgt_file_attrs_str[i].sa_nameP != NULL; ++i ) {
	if ( stricmp( nameP, Msgt_file_attrs_str[i].sa_nameP ) == 0 ) {
	    /* Found it-- hand off from here. */
	    return ( (*Msgt_file_attrs_str[i].sa_set)( mfdataP, valP ) );
	}
    }

    /* Not found */
    warning( "unknown string attribute name: \"%s\"", nameP );
    return ( FALSE );
}
/*

*//* msgt_file_gtdata_get( srcP )

	Instantiate a new copy of general data for msgt file type

Accepts :

	srcP		Ptr to source/template data to make a copy of

Returns :

	<value>		Pointer to MSGT_FILE_DATA copy
			NULL if failed

Notes :


*/

MSGT_FILE_DATA *
msgt_file_gtdata_get ARGLIST( (srcP) )
    FARG( MSGT_FILE_DATA *srcP )	/* Ptr to source to copy from */
{
	MSGT_FILE_DATA	*mfdataP;	/* New data being made */

    /* If no source, use our internal template */
    if ( srcP == NULL )
	srcP = MfdataP;

    mfdataP = (MSGT_FILE_DATA *)emalloc( "file_data_get", "new data",
					   sizeof(MSGT_FILE_DATA) );
    if ( mfdataP == NULL )
	return ( NULL );

    /* Set up the data, element by element */
    mfdataP->gf_fP = NULL;		    /* File handle .. clear it */
    mfdataP->gf_mltype = srcP->gf_mltype;   /* Folder layout type */
    mfdataP->gf_oqfromF = srcP->gf_oqfromF; /* Quote "From " lines? */
    mfdataP->gf_blendF = srcP->gf_blendF;   /* Blank line at end required? */

    strcpy( (char *)&mfdataP->gf_afproot[0], /* Any folder path root */
            (char *)&srcP->gf_afproot[0] );

    strcpy( (char *)&mfdataP->gf_afp[0],    /* Absolute folder path */
            (char *)&srcP->gf_afp[0] );

    strcpy( (char *)&mfdataP->gf_aflkp[0],  /* Absolute path to folder lock */
            (char *)&srcP->gf_aflkp[0] );

    /* Locking parameters; copy */
    attr_str_unset( &mfdataP->gf_locking );
    if ( srcP->gf_locking.as_strP != NULL )
	attr_str_set( &mfdataP->gf_locking, srcP->gf_locking.as_strP, TRUE );

    return ( mfdataP );
}
/*

*//* msgt_file_gtdata_release( mfdataP )

	Get rid of a copy of general data for msgt file type

Accepts :

	mfdataP		Ptr to general data to release

Returns :

	<nothing>

Notes :


*/

void
msgt_file_gtdata_release ARGLIST( (mfdataP) )
    FARG( MSGT_FILE_DATA *mfdataP )	/* Ptr to data to release */
{
    attr_str_clear( &MfdataP->gf_locking );

    dealloc( mfdataP );
}
/*

*//* msgt_file_msg_write( ommP, msmsgP )

	Write a message

Accepts :

	ommP		Ptr to the open message to be written
	msmsgP		Ptr to msgstore message handle for writing

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
msgt_file_msg_write ARGLIST( (ommP, msmsgP) )
   NFARG( OMM		*ommP )		/* Open message handle */
    FARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
{
	BOOL		okF;
	int		err;		/* Copy of errno */
	MSFOLDER	*msfP;		/* copy of MSFOLDER handle */
	MSNS		*msnsP;		/* copy pf namespace handle */
	MSGT_FILE_DATA	*mfdataP;	/* General data for file type */
	MSSTIF_FILE	*stifP;		/* Specific type interface */
	MSGPART		*mpP;
	MMSG		*msgP;
	void		*hdrP;		/* A header pointer */
	REFSTR		*rsP;
	BSTR		*bsP;

    msfP = msmsgP->mm_folderP;
    msnsP = msfP->mf_msnsP;
    stifP = msmsgP->mm_msstP->st_stifP;	/* Specific type interface */
    mfdataP = msmsgP->mm_gtdataP;
    mpP = ommP->om_mmsgP->mm_topmpP;	/* Top messagepart */
    msgP = mpP->mp_mmsgP;		/* Message object */
    msgP->mm_cnlC = 1;			/* Beginning; pretend newline */
    msgP->mm_oqfromF = mfdataP->gf_oqfromF;
					/* Whether to quote "From " lines */

    /* Set the From_F flag depending on whether the first header
       line is "From " .  This flag is a remnant of an old way of
       scanning the header but we'll use it.
    */
    hdrP = msgpart_hdr_next( mpP, NULL, NULL );
    if ( hdrP != NULL ) {
	rsP = msgpart_hdr_str( mpP, hdrP );
	bsP = refstr_bstr( rsP );
	From_F = ( strncmp( (char *)bsP->bs_bP, "From ", 5 ) == 0 );
    }
    else
	From_F = FALSE;

    /* Create new message in the folder. */
    if ( ! (*stifP->fi_msg_new)( ommP, msmsgP ) ) {
	err = errno;
	warning( "msgt_file_write can not open \"%s\"",
		 (char *)&mfdataP->gf_afp[0] );
	omm_delivery_error( ommP, err, "store" );
	return ( FALSE );
    }

    /* Write the headers */
    if ( okF = msgpart_write_hdrs( mpP, mfdataP->gf_fP, FALSE ) ) {
	/* Write any cached errors */
	if ( okF = omm_write_errs( ommP, mfdataP->gf_fP, TRUE ) ) {
	    /* Write the body */
	    okF = msgpart_write_body( mpP, mfdataP->gf_fP, FALSE );
	}
    }



    /* If it must end with a blank line, make it */
    if ( mfdataP->gf_blendF ) {
#if 1				/* This forces writing a blank line, even
				   if there was one already there.  This
				   is probably more proper than just
				   checking, but it could end up adding an
				   extra blank line to the body.
				*/
	if ( msgP->mm_cnlC > 0 )
	    msgP->mm_cnlC = 1;	/* Only need 1 if there is already some */
#endif

	while ( okF && ( msgP->mm_cnlC++ < 2 ) )
	    okF = ( fputc( '\n', mfdataP->gf_fP ) != EOF );
    }

    /* Wrap up and get status */
    okF = okF && (*stifP->fi_msg_end)( msmsgP );

    if ( okF ) {
	++ommP->om_deliveryC;		/* Count the delivery */
    }
    else {
	err = errno;
	omm_delivery_error( ommP, err, "store" );
	(*stifP->fi_msg_abort)( msmsgP );
    }

    return ( okF );
}
/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* attr_str_set_layout( mfdataP, valP )

	String attribute set: "layout"

Accepts :

	mfdataP		General type data to modify
	valP		Value to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Subservient to string attribute setting; called via dispatch table

*/

static BOOL
attr_str_set_layout ARGLIST( (mfdataP, valP) )
   NFARG( MSGT_FILE_DATA *mfdataP )	/* General data for file types */
    FARG( char		*valP )		/* The value to set */
{
    if ( ( stricmp( valP, "fs" ) == 0 ) ||
         ( stricmp( valP, "filesystem" ) == 0 ) ) {
	mfdataP->gf_mltype = ML_FS;
    }
    else if ( ( stricmp( valP, "maildir" ) == 0 ) ||
              ( stricmp( valP, "maildir++" ) == 0 ) ) {
	mfdataP->gf_mltype = ML_MDPP;
    }
    else {
	warning( "Unknown file layout type \"%s\"", valP );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* attr_str_set_locking( mfdataP, valP )

	String attribute set: "locking"

Accepts :

	mfdataP		General type data to modify
	valP		Value to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Subservient to string attribute setting; called via dispatch table

*/

static BOOL
attr_str_set_locking ARGLIST( (mfdataP, valP) )
   NFARG( MSGT_FILE_DATA *mfdataP )	/* General data for file types */
    FARG( char		*valP )		/* The value to set */
{
    return ( attr_str_set( &mfdataP->gf_locking, (UBYTE *)valP, TRUE ) );
}
/*

*//* attr_str_set_path( mfdataP, valP )

	String attribute set: "path"

Accepts :

	mfdataP		General type data to modify
	valP		Value to set

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Subservient to string attribute setting; called via dispatch table

	This sets the folder path root.

*/

static BOOL
attr_str_set_path ARGLIST( (mfdataP, valP) )
   NFARG( MSGT_FILE_DATA *mfdataP )	/* General data for file types */
    FARG( char		*valP )		/* The value to set */
{
    if ( strlen( valP ) >= sizeof( mfdataP->gf_afproot ) ) {
	warning( "file path too long: \"%s\"", valP );
	mfdataP->gf_afproot[0] = NUL;
	return ( FALSE );
    }

    strcpy( (char *)&mfdataP->gf_afproot[0], valP );
    return ( TRUE );
}
/*

*//* file_folder_check( msmsgP, msstPP )

	Checks for an existing folder

Accepts :

	msmsgP		Ptr to msgstore message handle to check
	msstPP		Where to put a pointer to the specific type handle

Returns :

	<value>		TRUE if OK
			FALSE if error

Notes :

	The primary return (TRUE/FALSE) indicates function success only.
	Whether a folder is found is determined by the msstPP return
	value.

	If a file exists at the absolute folder path, and its type can
	be determined (and used), its specific type is returned via
	the msstPP pointer.

	If a file exists but can't be accessed, the return is as if
	the file did not exist.  The access issue will be discovered
	later.  (This behaviour might change..)

	If a file exists and its type is unknown or is invalid for this
	namespace, FALSE is returned.

*/

static BOOL
file_folder_check ARGLIST( (msmsgP, msstPP) )
   NFARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
    FARG( MSST		**msstPP )	/* Where to return the specific type */
{
	LL_NODE		*nodeP;		/* Linked list node */
	MSST		*msstP;		/* Candidate MSST */
	MSSTIF_FILE	*stifP;		/* Specific-type interface */
	MSNS		*msnsP;		/* Namespace ptr */
	MSGT_FILE_DATA	*mfdataP;	/* General data for file type */
	struct stat	statblk;	/* For stat of a file */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data pointer */
    msnsP = msmsgP->mm_folderP->mf_msnsP; /* Namespace ptr */

    /* Check for existence of this file. */
    if ( lstat( (char *)&mfdataP->gf_afp[0], &statblk ) < 0 ) {
	/* Can't stat, assume file doesn't exist */
	*msstPP = NULL;
	return ( TRUE );
    }

    /* There's some file there; let's ask all of the specific types if
       they recognize it.  However, if the namespace requires a specific
       type, only ask that type.
    */
    mml_mutex_lock( &Msstlist.llh_lock );

    /* Walk the list looking for specific types of this same general type. */
    for ( nodeP = Msstlist.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	msstP = (MSST *)nodeP->lln_valP;
	if ( msstP->st_msgtP != &Msgt_file )
	    continue;
	if ( ( !msnsP->ns_mustsstF ) ||
	     ( msstP == msnsP->ns_msstP ) ) {

	    /* Use intermediate variable for readability.  You wouldn't
	       like the expression that was here before.
	    */
	    stifP = (MSSTIF_FILE *)msstP->st_stifP;
	    if ( (*stifP->fi_folder_check)( msmsgP ) )
		break;			/* This type says "yes" */
	}
    }

    mml_mutex_unlock( &Msstlist.llh_lock );

    /* If we found a type, return it. */
    if ( msstP != NULL ) {
	*msstPP = msstP;
	return ( TRUE );
    }

    /* Can't determine the type, give failure return. */
    *msstPP = NULL;			/* What the hey */
    return ( FALSE );
}
/*

*//* file_folder_path( msmsgP )

	Determine the full filename path for a particular layout type

Accepts :

	msmsgP		Ptr to message store handle

Returns :

	<value>		TRUE if path created OK
			FALSE if some error.

Notes :

	Sets up the absolute folder paths within the general file data
	structure.

	The message handle's general data pointer is presumed to have
	been instantiated at this point.

*/

static BOOL
file_folder_path ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Ptr to message handle */
{
    /* The general data pointer specifies a layout type.  Proceed based
       on that.
    */
    switch ( ((MSGT_FILE_DATA *)msmsgP->mm_gtdataP)->gf_mltype ) {
	case	ML_UNK:			/* As yet unknown */
	    warning( "file_folder_path: layout type is \"unknown\" -- incomplete namespace?" );
	    return ( FALSE );

	case	ML_FS:			/* General filesystem */
	    return ( file_folder_path_fs( msmsgP ) );

	case	ML_MDPP:		/* Maildir++ */
	    return ( file_folder_path_mdpp( msmsgP ) );

	default:
	    warning( "file_folder_path: unknown layout type %d",
		      ((MSGT_FILE_DATA *)msmsgP->mm_gtdataP)->gf_mltype );
	    return ( FALSE );
    }
}
/*

*//* file_folder_path_fs( msmsgP )

	Determine the path for a particular folder type using FS layout

Accepts :

	msmsgP		Ptr to message store handle

Returns :

	<value>		TRUE if path created OK
			FALSE if some error.

Notes :

	Subservient to file_folder_path for this specific layout type.

*/

static BOOL
file_folder_path_fs ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Ptr to message handle */
{
	UBYTE		ch;
	int		plen;		/* Current path length */
	int		pmax;		/* Maximum path length */
	UBYTE		*srcP;		/* String ptr */
	UBYTE		*fpathP;	/* Ptr to path being built */
	MSGT_FILE_DATA	*mfdataP;	/* General data for file type */

    mfdataP = msmsgP->mm_gtdataP;
    fpathP = &mfdataP->gf_afp[0];	/* Where path is being built */
    pmax = sizeof( mfdataP->gf_afp );	/* Maximum path length */
    --pmax;				/* Allow for trailing NUL */

    /* Start off with any filesystem constraint prefix (root) */
    strncpy( (char *)fpathP, (char *)&mfdataP->gf_afproot[0], pmax );
    plen = strlen( (char *)fpathP );
    if ( ( plen > 0 ) && ( fpathP[plen-1] != '/' ) )
	fpathP[ plen++ ] = '/';

    /* Append foldername (the tail after the namespace prefix),
       ignoring any possible /../ sequences
    */
    for ( srcP = msmsg_folder_tail( msmsgP ); (ch = *srcP++) != NUL; ) {
	if ( plen < pmax ) {
	    /* If constrained, check for '/../' sequence */
	    if ( ( ch == '.' ) &&
		 ( mfdataP->gf_afproot[0] != NUL ) &&
	         ( plen > 0 ) &&
		 ( fpathP[plen -1] == '/' ) &&
		 ( *srcP == '.' ) &&
		 ( srcP[1] == '/' ) ) {
		/* We simply don't allow it. */
		return ( FALSE );
	    }

	    /* OK, take the char. */
	    fpathP[ plen++ ] = ch; 
	}
    }

    if ( plen >= pmax )
	return ( FALSE );

    /* Strip any trailing '/' char(s) */
    while ( ( plen > 0 ) && ( fpathP[ plen -1] == '/' ) )
	--plen;

    fpathP[ plen++ ] = NUL;

    return ( TRUE );
}
/*

*//* file_folder_path_mdpp( msmsgP )

	Determine the path for a particular folder type using Maildir++ layout

Accepts :

	msmsgP		Ptr to message store handle

Returns :

	<value>		TRUE if path created OK
			FALSE if some error.

Notes :

	Subservient to file_folder_path for this specific layout type.

	Within a maildir, we have the following rules:

	    The underlying mailbox name is the tail (after the
	    namespace prefix).

	    The null name (if only the namespace itself is given,
	    no tail) refers to the maildir itself.

	    A non-null name is a subfolder within the maildir.  These are
	    stored in a flat hierarchy with "/" translated to "." (at
	    least for now).  The maildir directory name for a subdir
	    begins with an additional "." .

*/

static BOOL
file_folder_path_mdpp ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Ptr to message handle */
{
	UBYTE		ch;		/* Character */
	int		plen;		/* Current path length */
	int		pmax;		/* Path max */
	UBYTE		*srcP;		/* Source pointer */
	UBYTE		*fpathP;	/* Ptr to path being built */
	MSGT_FILE_DATA	*mfdataP;	/* General data for file type */

    mfdataP = msmsgP->mm_gtdataP;
    fpathP = &mfdataP->gf_afp[0];	/* Where path is being built */
    pmax = sizeof( mfdataP->gf_afp );	/* Maximum path length */
    --pmax;				/* Allow for trailing NUL */

    /* The folder root is the fundamental path to the Maildir.  It
       must have been set.
    */
    if ( mfdataP->gf_afproot[0] == NUL )
	return ( FALSE );

TODO
    /* It's unclear as to whether we should make sure that the root
       maildir actually exists, and is a maildir.  At present, we don't.
    */

    /* The tail (after the namespace prefix) is either empty, which
       indicates the primary maildir, or it's non-empty indicating
       a subfolder.
    */
    srcP = msmsg_folder_tail( msmsgP );

    /* Construct full path, with appropriate translations. */
    strncpy( (char *)fpathP, (char *)&mfdataP->gf_afproot[0], pmax );
    plen = strlen( fpathP );

    /* Remove any trailing '/'.  We'll probably add one back in if
       there's any non-null foldername- but we have to strip the
       trailing '/' first because of a mkdir issue in BSD/OS (it
       barfs if you give it a trailing '/').
    */
    while ( ( plen > 0 ) && ( fpathP[ plen-1 ] == '/' ) )
	--plen;

    /* Now if any subfolder, tack on "/." plus the subfolder name */
    if ( *srcP != NUL ) {
	if ( plen < ( pmax -2) ) {
	    fpathP[ plen++ ] = '/';
	    fpathP[ plen++ ] = '.';
	}

	while( ( ch = *srcP++ ) != NUL ) {
	    if ( plen < pmax ) {
		if ( ch == '/' )
		    ch = '.';
		fpathP[ plen++ ] = ch;
	    }
	}
    }

    if ( plen >= pmax )
	return ( FALSE );

    /* Strip any trailing '/' char(s) */
    while ( ( plen > 0 ) && ( fpathP[ plen -1] == '/' ) )
	--plen;

    fpathP[ plen++ ] = NUL;

    return ( TRUE );
}


