/* msst-file-maildir.c -- msgstore specific file type "maildir"

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff to handle the msgstore specific file type "maildir"

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/time.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "mvmf-locking.h"
#include "msgstore.h"
#include "msgt-file.h"
#include "msst-file-maildir.h"

/* Local definitions */



/* External data referenced */

extern	MSGT	Msgt_file;		/* Control for "file" general type */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* MSST handler functions: */
static	BOOL	msst_maildir_attr_int_set PROTO( (void *gtdataP, void *stdataP,
					          char *nameP, long val) );
static	BOOL	msst_maildir_attr_str_set PROTO( (void *gtdataP, void *stdataP,
					          char *nameP, char *valP) );
static	void	*msst_maildir_gtdata_get PROTO( (void *srcP) );
static	void	msst_maildir_gtdata_release PROTO( (void *gtdataP) );
static	void	*msst_maildir_stdata_get PROTO( (void *srcP) );
static	void	msst_maildir_stdata_release PROTO( (void *stdataP) );
static	BOOL	msst_maildir_msg_write PROTO( (OMM *ommP, MSMSG *msmsgP) );


    /* MSSTIF_FILE handler functions (generic/specific interface): */
static	BOOL	msstif_maildir_folder_check PROTO( (MSMSG *msmsgP) );
static	void	msstif_maildir_msg_abort PROTO( (MSMSG *msmsgP) );
static	BOOL	msstif_maildir_msg_end PROTO( (MSMSG *msmsgP) );
static	BOOL	msstif_maildir_msg_new PROTO( (OMM *ommP, MSMSG *msmsgP) );

    /* Other support functions */
static	BOOL	maildir_check PROTO( (char *fnameP) );
static	BOOL	maildir_make PROTO( (char *fnameP) );


/* Private data */

    /* Maildir sequence number */
static	int	Maildir_seq = 1001;

    /* List of subdirs that a Maildir must have to be a Maildir */
static	char	*Maildir_subdirs[] = {
	"cur", "new", "tmp", NULL
};


    /* The MSSTIF_FILE definition, providing specific type info
       used by general type functions of this type.
    */
static	MSSTIF_FILE Msstif_file_maildir = {
	msstif_maildir_folder_check,	/* Check folder for this type */
	msstif_maildir_msg_new,		/* Begin new maildir message */
	msstif_maildir_msg_end,		/* End new maildir message */
	msstif_maildir_msg_abort	/* Abort new maildir message */
};


    /* The MSST definition for the msst_file_maildir type. */
static	MSST	Msst_file_maildir = {
	"maildir",			/* The name */

	&Msgt_file,			/* Ptr to general type */
	&Msstif_file_maildir,		/* Specific type interface */

	NULL,				/* General data */
	NULL,				/* Specific data */

		/* Handlers */
	msst_maildir_gtdata_get,	/* Get general data */
	msst_maildir_gtdata_release,	/* Release general data */
	msst_maildir_gtdata_get,	/* Get specific data */
	msst_maildir_gtdata_release,	/* Release specific data */
	msst_maildir_attr_int_set,	/* Set integer attribute */
	msst_maildir_attr_str_set,	/* Set string attribute */
	msst_maildir_msg_write		/* Write a message */

};
	

/*

*//* msst_file_maildir_init()

	Initialize the specific file type handler

Accepts :


Returns :

	<value>		STSCODE status code:
			  SCOK if successful
			  something else if not.

Notes :


*/

BOOL
msst_file_maildir_init NOARGLIST
{
	MSGT_FILE_DATA	*mfdataP;	/* general file data */

    /* Get initial (template) general type data */
    Msst_file_maildir.st_gtdataP = msst_maildir_gtdata_get( NULL );
    if ( ( mfdataP = Msst_file_maildir.st_gtdataP ) == NULL ) {
	warning( "msst_file_maildir_init: failed to initialize gtdata." );
	return ( FALSE );
    }
    /* Tweak for this specific type */
    mfdataP->gf_oqfromF = FALSE;	/* Don't quote "From " lines */
    mfdataP->gf_blendF = FALSE;		/* Don't need blank line at end */


    /* Get initial (template) specific type data */
    Msst_file_maildir.st_stdataP = msst_maildir_stdata_get( NULL );
    if ( Msst_file_maildir.st_stdataP == NULL ) {
	warning( "msst_file_maildir_init: failed to initialize stdata." );
	return ( FALSE );
    }

    /* Register this type */
    if ( !msst_register( &Msst_file_maildir ) )
	return ( FALSE );

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                          MSST HANDLERS                              *
 *                                                                     *
 *   These are handler functions which support msst functions for      *
 *   this type.                                                        *
 *                                                                     *
 ***********************************************************************/

/*

*//* msst_maildir_attr_int_set( gtdataP, stdataP, nameP, val )

	Set an integer attribute value for this specific type

Accepts :

	gtdataP		Ptr to general data
	stdataP		Ptr to specific data
	nameP		The attribute name
	val		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msst_maildir_attr_int_set ARGLIST( (gtdataP, stdataP, nameP, val) )
   NFARG( void		*gtdataP )	/* General data */
   NFARG( void		*stdataP )	/* Specific data */
   NFARG( char		*nameP )	/* Attribute */
    FARG( long		val )		/* The value to set */
{
    /* Hand off to the general type's value setting */
    return ( msgt_file_attr_int_set( (MSGT_FILE_DATA *)gtdataP,
				      nameP, val ) );
}
/*

*//* msst_maildir_attr_int_set( gtdataP, stdataP, nameP, valP )

	Set a string attribute value for this specific type

Accepts :

	gtdataP		Ptr to general data
	stdataP		Ptr to specific data
	nameP		The attribute name
	valP		The value to set for that attribute

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
msst_maildir_attr_str_set ARGLIST( (gtdataP, stdataP, nameP, valP) )
   NFARG( void		*gtdataP )	/* General data */
   NFARG( void		*stdataP )	/* Specific data */
   NFARG( char		*nameP )	/* Attribute name */
    FARG( char		*valP )		/* The value to set */
{
    /* Hand off to the general type's value setting */
    return ( msgt_file_attr_str_set( (MSGT_FILE_DATA *)gtdataP,
				      nameP, valP ) );
}
/*

*//* msst_maildir_gtdata_get( srcP )

	Instantiate a new copy of general data for this specific type

Accepts :

	srcP		Ptr to source/template data to make a copy of

Returns :

	<value>		Pointer to the general data
			NULL if failed

Notes :

	If the source is NULL, the template data associated with this
	specific type is used as a source.  (That may, in turn, be
	NULL, in which case the general type getter will fill it in.)

*/

void *
msst_maildir_gtdata_get ARGLIST( (srcP) )
    FARG( void		*srcP)		/* Ptr to source to copy from */
{
	MSGT_FILE_DATA	*mfdataP;	/* New data being made */

    /* Use this specific type's template data if no source was given */
    if ( srcP == NULL )
	srcP = Msst_file_maildir.st_gtdataP;

    /* Use general type function to get a copy */
    mfdataP = msgt_file_gtdata_get( (MSGT_FILE_DATA *)srcP );

    /* No further msst refinements needed. */

    return ( (void *)mfdataP );
}
/*

*//* msst_maildir_gtdata_release( gtdataP )

	Get rid of a copy of general data for msgt file type

Accepts :

	gtdataP		Ptr to general data to release

Returns :

	<nothing>

Notes :


*/

void
msst_maildir_gtdata_release ARGLIST( (gtdataP) )
    FARG( void		*gtdataP )	/* Data to release */
{
    /* We can safely use the general type release for all the work */
    msgt_file_gtdata_release( (MSGT_FILE_DATA *)gtdataP );
}
/*

*//* msst_maildir_stdata_get( srcP )

	Instantiate a new copy of general data for this specific type

Accepts :

	srcP		Ptr to source/template data to make a copy of

Returns :

	<value>		Pointer to the general data
			NULL if failed

Notes :

	This specific type has, as yet, no specific data, so this
	function just returns a dummy address, to pretend to succeed.

*/

void *
msst_maildir_stdata_get ARGLIST( (srcP) )
    FARG( void		*srcP)		/* Ptr to source to copy from */
{
    /* Return dummy address */
    return ( (void *)&Msst_file_maildir );
}
/*

*//* msst_maildir_stdata_release( stdataP )

	Get rid of a copy of general data for msgt file type

Accepts :

	stdataP		Ptr to general data to release

Returns :

	<nothing>

Notes :

	This specific type has no specific type data, so this function
	is simply a no-op.

*/

void
msst_maildir_stdata_release ARGLIST( (stdataP) )
    FARG( void		*stdataP )	/* Data to release */
{
}
/*

*//* msst_maildir_msg_write( ommP, msmsgP )

	Write a message for this specific type

Accepts :

	ommP		Ptr to the open message to be written
	msmsgP		Ptr to msgstore message handle for writing

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	The general type interface provides most (or all!) of the
	logic for writing a message to a file, as guided by the
	parameters associated with this message.

*/

static BOOL
msst_maildir_msg_write ARGLIST( (ommP, msmsgP) )
   NFARG( OMM		*ommP )		/* Open message handle */
    FARG( MSMSG		*msmsgP )	/* Mailstore message for writing */
{
	BOOL		okF;
	MSGT_FILE_DATA	*mfdataP;	/* General type's data */

    mfdataP = (MSGT_FILE_DATA *)msmsgP->mm_gtdataP;

    /* Set up the path to the base lockfile name, which is the same
       as the folder name.  
    */
    if ( snprintf( (char *)&mfdataP->gf_aflkp[0], sizeof(mfdataP->gf_aflkp),
		   "%s",
		     (char *)&mfdataP->gf_afp[0]
		 ) >= sizeof( mfdataP->gf_aflkp ) ) {
	warning( "Lock path too long for \"%s.lock\"",
			(char *)&mfdataP->gf_afp[0] );
	return ( FALSE );
    }

    /* Use the general type's function */
    okF = msgt_file_msg_write( ommP, msmsgP );

    return ( okF );
}

/***********************************************************************
 *                                                                     *
 *                       MSSTIF_FILE HANDLERS                          *
 *                                                                     *
 *   These are handler functions which support specific type interface *
 *   functions between the general type and this specific type         *
 *                                                                     *
 ***********************************************************************/

/*

*//* msstif_maildir_msg_abort( msmsgP )

	Abort a maildir message already started

Accepts :

	msmsgP		Ptr to mailstore message handle

Returns :

	<none>

Notes :


*/

static void
msstif_maildir_msg_abort ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* mailbox being written */
{
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MAILDIR_DATA *mddP;	/* type-specific (maildir) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mddP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Close the fP and delete the /tmp file */
    unlink( (char *)&mddP->md_tmpname[0] );
    fclose( mfdataP->gf_fP );
    mfdataP->gf_fP = NULL;

    /* Take care of locks */
    unlock_lockset( &mfdataP->gf_locks, -1, &mfdataP->gf_aflkp[0] );

    /* some higher level will deal with msmsg further. */
}
/*

*//* msstif_maildir_msg_end( msmsgP )

	End a maildir message being written

Accepts :

	msmsgP		ptr to handle for mailbox being written

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	On failure, the mailbox will still be in some opened state; caller
	 is expected to do further cleanup (such as calling msg_abort() )

*/

static BOOL
msstif_maildir_msg_end ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* mailbox being written handle */
{
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MAILDIR_DATA *mddP;	/* type-specific (maildir) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mddP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Close the file */
    if ( fclose( mfdataP->gf_fP ) != 0 ) {
	/* Got an error- abort and exit badly */
	return ( FALSE );
    }

    /* Rename the tmp file to the new directory */
    if ( link( &mddP->md_tmpname[0], &mddP->md_newname[0] ) < 0 ) {
	/* Some error. */
	return ( FALSE );
    }

    /* Get rid of tmpfile */
    unlink( &mddP->md_tmpname[0] );

    /* Unlock any locks */
    unlock_lockset( &mfdataP->gf_locks, -1, &mfdataP->gf_aflkp[0] );

    /* Higher level will deal with the MSMSG */

    return ( TRUE );
}
/*

*//* msstif_maildir_msg_new( ommP, msmsgP )

	Create a new message in a maildir folder

Accepts :

	ommP		Ptr to open message handle
	msmsgP		Ptr to working message handle

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

static BOOL
msstif_maildir_msg_new ARGLIST( (ommP, msmsgP) )
   NFARG( OMM		*ommP )		/* Ptr to open message */
    FARG( MSMSG		*msmsgP )	/* Ptr to MSMSG in progress */
{
	FILE		*fP;
	MSFOLDER	*msfP;		/* Folder */
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */
	MSST_MAILDIR_DATA *mddP;	/* type-specific (maildir) data */
	struct timeval	tvnow;
	char		hostname[50];

    msfP = msmsgP->mm_folderP;		/* Get folder ptr */
    mfdataP = msmsgP->mm_gtdataP;	/* Get general data */
    mddP = msmsgP->mm_stdataP;		/* Get type-specific data ptr */

    /* Maildir name follows guidelines at:
	http://cr.yp.to/proto/maildir.html

       Generally, the name is ttttt.stuff.hhhhh, where
         ttttt = seconds from time(),
	 stuff = some stuff to add uniqueness,
	 hhhhh = hostname.

       "stuff" can be a lot of things.  We use MnPnQn, where:
         in Mn, n is microsecond counter from gettimeofday,
	 in Pn, n is our process ID,
	 in Qn, n is a uniqueness counter within this process.
    */

    gettimeofday( &tvnow, (struct timezone *)NULL );
    gethostname( &hostname[0], sizeof(hostname)-1 );

    if ( snprintf( (char *)&mddP->md_basename[0], sizeof(mddP->md_basename),
	      "%ld.M%ldP%ldQ%d.%s",
	         (long)tvnow.tv_sec,
		 (long)tvnow.tv_usec,
	         getpid(),
	         Maildir_seq++,
	         &hostname[0] ) >= sizeof(mddP->md_basename) ) {
	warning( "msstif_maildir_msg_new: base name too long for \"%s\".",
			(char *)msfP->mf_fnameP );
	return ( FALSE );
    }

    /* Create new/ and /tmp names */
    if ( snprintf( (char *)&mddP->md_newname[0], sizeof(mddP->md_newname),
		     "%s/new/%s",
		       (char *)&mfdataP->gf_afp[0],
		       (char *)&mddP->md_basename[0]
		  ) >= sizeof(mddP->md_newname) ) {
	warning( "msstif_maildir_msg_new: new/ name too long for \"%s\".",
			(char *)msfP->mf_fnameP );
	return ( FALSE );
    }

    if ( snprintf( (char *)&mddP->md_tmpname[0], sizeof(mddP->md_tmpname),
		     "%s/tmp/%s",
		       (char *)&mfdataP->gf_afp[0],
		       (char *)&mddP->md_basename[0]
		  ) >= sizeof(mddP->md_tmpname) ) {
	warning( "msstif_maildir_msg_new: tmp/ name too long for \"%s\".",
			(char *)msfP->mf_fnameP );
	return ( FALSE );
    }

    /* Create the tmp file */
    mfdataP->gf_fP = fopen( (char *)&mddP->md_tmpname[0], "w" );
    if ( mfdataP->gf_fP == NULL ) {
	/* Check for directory error */
	if ( ( errno == ENOTDIR ) || ( errno == ENOENT ) ) {
	    /* Try making maildir and trying again */
	    if ( maildir_make( (char *)&mfdataP->gf_afp[0] ) )
		mfdataP->gf_fP = fopen( (char *)&mddP->md_tmpname[0], "w" );
	}
    }

    if ( mfdataP->gf_fP != NULL ) {
	/* Make any lock(s) required */
	if ( !lock_lockset( &mfdataP->gf_locks,
		       fileno(mfdataP->gf_fP), &mfdataP->gf_aflkp[0] ) ) {
	    msstif_maildir_msg_abort( msmsgP );
	    mfdataP->gf_fP = NULL;	/* Redundant, I'm sure */
	}
    }

    return ( mfdataP->gf_fP != NULL );
}
/*

*//* msstif_maildir_folder_check( msmsgP )

	Specific type check for a folder

Accepts :

	msmsgP		Ptr to the mailstore message handle

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	A pointer to some general type data is known to be present at this
	point.

	For this general type (file), the check is to see if an
	existing folder is of this specific type.

*/

static BOOL
msstif_maildir_folder_check ARGLIST( (msmsgP) )
    FARG( MSMSG		*msmsgP )	/* Mailstore message handle */
{
	MSGT_FILE_DATA	*mfdataP;	/* general type (file) data */

    mfdataP = msmsgP->mm_gtdataP;	/* Get general data pointer */

    /* Hand off to a check on the absolute path. */
    return ( maildir_check( (char *)&mfdataP->gf_afp[0] ) );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* maildir_check( fnameP )

	Check a file to see if it's a maildir

Accepts :

	fnameP		Name of file that might be a maildir

Returns :

	<value>		TRUE if it's a maildir
			FALSE if not.

Notes :


*/

static BOOL
maildir_check ARGLIST( (fnameP) )
    FARG( char		*fnameP )	/* Filename */
{
	int		i;
	struct stat	statblk;	/* For stat of a file */
	char		sdpath[ PATH_MAX +1 ];  /* Subdir path */

    /* Make sure it's a directory (or is a symbolic link to one) */
    if ( ( lstat( fnameP, &statblk ) < 0 ) ||
         ( ( statblk.st_mode & S_IFMT ) != S_IFDIR ) )
	return ( FALSE );

    /* It's a directory, Check for presence of maildir subdirs */
    for ( i = 0; Maildir_subdirs[i] != NULL; ++i ) {
	if ( ( strlen( fnameP ) + strlen( Maildir_subdirs[i] ) +1 )
		> PATH_MAX )
	    return ( FALSE );

	sprintf( &sdpath[0], "%s/%s", fnameP, Maildir_subdirs[i] );
	if ( ( lstat( &sdpath[0], &statblk ) < 0 ) ||
	     ( ( statblk.st_mode & S_IFMT ) != S_IFDIR ) )
	    return ( FALSE );
    }

    /* Looks like it's a maildir */
    return ( TRUE );
}
/*

*//* maildir_make( fpathP )

	Attempt to make a maildir

Accepts :

	fpathP		Path to maildir file to make

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
maildir_make ARGLIST( (fpathP) )
    FARG( char		*fpathP )	/* Path to maildir to make */
{
	int		i;
	struct stat	statblk;	/* For stat of a file */
	char		sdpath[ PATH_MAX +1 ];  /* Subdir path */

    /* Check for directory */
    if ( ( lstat( fpathP, &statblk ) < 0 ) ||
         ( ( statblk.st_mode & S_IFMT ) != S_IFDIR ) ) {
	if ( mkdir( fpathP, 0750 ) < 0 )
	    return ( FALSE );
    }

    /* Presumably have a directory now- make the subdirs */
    for ( i = 0; Maildir_subdirs[i] != NULL; ++i ) {
	if ( ( strlen( fpathP ) + strlen( Maildir_subdirs[i] ) +1 )
		> PATH_MAX )
	    return ( FALSE );

	sprintf( &sdpath[0], "%s/%s", fpathP, Maildir_subdirs[i] );
	if ( ( lstat( &sdpath[0], &statblk ) < 0 ) ||
	     ( ( statblk.st_mode & S_IFMT ) != S_IFDIR ) ) {
	    if ( mkdir( &sdpath[0], 0750 ) < 0 ) {
		/* I suppose we could undo anything we made, but
		   nahhh.
		*/
		return ( FALSE );
	    }
	}
    }

    /* If we get here we are all set. */
    return ( TRUE );
}
