/* mvmf-locking.c -- stuff dealing with file locking for mvmf.

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains routines to deal with message storage for mvmf

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "controls.h"
#include "msgpart.h"
#include "omm.h"
#include "data.h"
#include "mvmf-locking.h"

#include "errwarn.h"

/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */



/*

*//* lock_dotlock( lfnameP )

	Establish a lock via dotlocking

Accepts :

	lfnameP		Full path of file to lock (the lockname base)

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
lock_dotlock ARGLIST( (lfnameP) )
    FARG( UBYTE		*lfnameP )	/* Path to base lockfile name */
{
	BOOL		ageF;
	int		i;
	int		retryC;		/* Retry counter */
	int		fd;		/* For dotlock file */
	time_t		timenow;
	struct stat	statblk;
	char		lfname[ PATH_MAX +1 ];

    /* Create the dotlock name */
    i = snprintf( &lfname[0], sizeof(lfname), "%s.lock", (char *)lfnameP );
    if ( i >= sizeof(lfname) ) {
	return ( FALSE );		/* Name is too long, can't lock it,
					   but then again, nobody else
					   will be able to either.
					*/
    }

    /* We'll try for a parameterized number of times, then attempt to
       age the lock, then try again without the aging.
    */
    for ( ageF = TRUE; ; ageF = FALSE ) {
	/* Loop trying to make the lockfile */
	for ( retryC = 0; retryC < DOTLOCK_MAXTRIES; ++retryC ) {
	    /* Sleep if not first time */
	    sleep( DOTLOCK_TRYEVERY );

	    /* Attempt to make the file */
	    fd = open( &lfname[0], O_WRONLY|O_EXCL|O_CREAT );
	    if ( fd >= 0 ) {
		/* Made the file.  Put an invalid pid in it just in case. */
		write( fd, "0", 1 );
		close( fd );
		return ( TRUE );
	    }
	}

	/* At this point all the retries have failed.  Maybe age it out. */
	if ( !ageF )
	    break;

	if ( stat( &lfname[0], &statblk ) < 0 )
	    break;			/* No luck.. fail. */

	time( &timenow );		/* Get time to compare to */
	if ( ( timenow - statblk.st_mtime ) > DOTLOCK_MAXAGE ) {
	    /* Old enough- get rid of it. */
	    if ( unlink( &lfname[0] ) != 0 )
		break;			/* Error, fail. */
	}

	/* Next loop will try again, without the aging. */
    }

    /* We failed. */
    return ( FALSE );
}
/*

*//* lock_fcntl( lfd )

	Establish an fcntl lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

#ifdef	HAVE_FCNTL

BOOL
lock_fcntl ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
	struct flock	fl;

    fl.l_type = F_WRLCK;		/* Write lock */
    fl.l_whence = SEEK_SET;		/* Absolute offset range */
    fl.l_start = 0L;			/* Start at beginning */
    fl.l_len = 0L;			/* Lock whole file */
    while ( fcntl( lfd, F_SETLKW, &fl ) < 0 ) {
	if ( ( errno != EAGAIN ) &&
	     ( errno != EINTR ) ) {

	    return ( FALSE );
	}
    }

    return ( TRUE );
}

#endif	/* HAVE_FCNTL */
/*

*//* lock_flock( lfd )

	Establish an flock lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

#ifdef	HAVE_FLOCK

BOOL
lock_flock ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
    while( flock( lfd, LOCK_EX ) < 0 ) {
	if ( errno != EINTR )
	    return ( FALSE );
    }

    return ( TRUE );
}

#endif	/* HAVE_FLOCK */
/*

*//* lock_lockf( lfd )

	Establish a lockf lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

#ifdef	HAVE_LOCKF

BOOL
lock_lockf ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
    while ( lockf( lfd, F_LOCK, 0 ) < 0 ) {
	if ( ( errno != EAGAIN ) &&
	     ( errno != EINTR ) ) {

	    return ( FALSE );
	}
    }

    return ( TRUE );
}

#endif	/* HAVE_LOCKF */
/*

*//* lock_lockset( lsP, lfd, lfnameP )

	Establish lock(s) via lockset

Accepts :

	lsP		Pointer to LOCKSET controlling the locks
	lfd		fd for open file to lock, if relevant
	lfnameP		Full path of file to lock (the lockname base)

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	lfd and lockname may or may not be relevant.  They should
	be provided nonetheless.

*/

BOOL
lock_lockset ARGLIST( (lsP, lfd, lfnameP) )
   NFARG( LOCKSET	*lsP )		/* Ptr to lockset control */
   NFARG( int		lfd )		/* fd for open file to lock */
    FARG( UBYTE		*lfnameP )	/* Path to base lockfile name */
{
	int		i;
	LOCKINFO	*lockP;

    /* Apply desired locks */
    for ( i = 0; i < lsP->ls_lockC; ++i ) {
	lockP = &lsP->ls_locks[ i ];
	switch ( lockP->lk_type ) {
	    case LCK_DOTLOCK:
		if ( !lock_dotlock( lfnameP ) )
		    return ( FALSE );
		lockP->lk_lockedF = TRUE;
		break;

#ifdef	HAVE_FCNTL
	    case LCK_FCNTL:
		if ( !lock_fcntl( lfd ) )
		    return ( FALSE );
		lockP->lk_lockedF = TRUE;
		break;
#endif	/* HAVE_FCNTL */

#ifdef	HAVE_FLOCK
	    case LCK_FLOCK:
		if ( !lock_flock( lfd ) )
		    return ( FALSE );
		lockP->lk_lockedF = TRUE;
		break;
#endif	/* HAVE_FLOCK */

#ifdef	HAVE_LOCKF
	    case LCK_LOCKF:
		if ( !lock_lockf( lfd ) )
		    return ( FALSE );
		lockP->lk_lockedF = TRUE;
		break;
#endif	/* HAVE_LOCKF */

	    default:
		warning( "lock_lockset: bad lock type %d", lockP->lk_type );
		break;
	}
    }

    return ( TRUE );
}
/*

*//* lock_parse( lockspecP, lsP )

	Parse a lock specification string into a LOCKSET

Accepts :

	locksetP	String containing locking spec
	lsP		LOCKSET to fill in

Returns :

	<nothing>

Notes :

	Lack of a control string will result in no locking;

	Unknown things in the control value are silently ignored, on
	  the theory that this may help survive changes in the environment
	  or code versions etc.

	If the control value is completely unrecognized, this will also
	  result in no locking.

*/

void
lock_parse ARGLIST( (lockspecP, lsP) )
   NFARG( char		*lockspecP )	/* locking specification string */
    FARG( LOCKSET	*lsP )		/* LOCKSET to fill in */
{
	int		i;
	char		*strP;
	LOCKINFO	*lockP;

    lsP->ls_lockC = 0;			/* No locking specified */

    /* If NULL lockspec string, use a default */
    if ( lockspecP == NULL )
	lockspecP = "";

    /* Lock types are comma-separated strings.  Process each part. */
    for ( ; lsP->ls_lockC < LOCKMAX; ) {
	/* Skip blanks/tabs */
	lockspecP = solbat( lockspecP, TRUE );
	if ( *lockspecP == NUL )
	    break;

	/* Find end of this part */
	strP = strchr( lockspecP, ',' );
	if ( strP != NULL )
	    i = strP - lockspecP;
	else {
	    i = strlen( lockspecP );
	}

	/* Store this part in the Gbs2P scratch BSTR */
	Gbs2P->bs_bC = 0;
	bstr_catn( Gbs2P, lockspecP, i );

	/* Setup lockspecP to point to next one for next time through */
	lockspecP += i;
	if ( *lockspecP == ',' )
	    ++lockspecP;

	/* Remove any trailing blanks */
	while ( ( ( i = Gbs2P->bs_bC ) > 0 ) &&
	        ( ( Gbs2P->bs_bP[i] == ' ' ) || ( Gbs2P->bs_bP[i] == '\t' ) ) )
	    --Gbs2P->bs_bC;
	bstr_nulterm( Gbs2P );


	/* Set up for this lock type */
	lockP = &lsP->ls_locks[ lsP->ls_lockC ];

	if ( stricmp( (char *)Gbs2P->bs_bP, "dotlock" ) == 0 ) {
	    lockP->lk_type = LCK_DOTLOCK;
	    lockP->lk_lockedF = FALSE;
	    ++lsP->ls_lockC;
	}

#ifdef	HAVE_FCNTL
	else if ( stricmp( (char *)Gbs2P->bs_bP, "fcntl" ) == 0 ) {
	    lockP->lk_type = LCK_FCNTL;
	    lockP->lk_lockedF = FALSE;
	    ++lsP->ls_lockC;
	}
#endif	/* HAVE_FCNTL */

#ifdef	HAVE_FLOCK
	else if ( stricmp( (char *)Gbs2P->bs_bP, "flock" ) == 0 ) {
	    lockP->lk_type = LCK_FLOCK;
	    lockP->lk_lockedF = FALSE;
	    ++lsP->ls_lockC;
	}
#endif	/* HAVE_FLOCK */

#ifdef	HAVE_LOCKF
	else if ( stricmp( (char *)Gbs2P->bs_bP, "lockf" ) == 0 ) {
	    lockP->lk_type = LCK_LOCKF;
	    lockP->lk_lockedF = FALSE;
	    ++lsP->ls_lockC;
	}
#endif	/* HAVE_LOCKF */

	else {
	    /* Unrecognized, just ignore. */
	    ;
	}
    }
}
/*

*//* unlock_dotlock( lfnameP )

	Release a lock via dotlocking

Accepts :

	lfnameP		Full path of file to lock (the lockname base)

Returns :

	<nothing>

Notes :


*/

void
unlock_dotlock ARGLIST( (lfnameP) )
    FARG( UBYTE		*lfnameP )	/* Path to base lockfile name */
{
	int		i;
	char		lfname[ PATH_MAX +1 ];

    /* Create the dotlock name.  This shouldn't fail, since we should only
       get here if the dotlock file was created in the first place.  But
       check anyway.
    */
    i = snprintf( &lfname[0], sizeof(lfname), "%s.lock", lfnameP );
    if ( i >= sizeof(lfname) ) {
	warning( "unlock_dotlock: peculiar inability to create dotlock name." );
	return;
    }

    /* Remove. */
    unlink( &lfname[0] );
}
/*

*//* unlock_fcntl( lfd )

	Release an fcntl lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<nothing>

Notes :


*/

#ifdef	HAVE_FCNTL

void
unlock_fcntl ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
    /* This is a no-op -- lock is removed when file is closed. */
}

#endif	/* HAVE_FCNTL */
/*

*//* unlock_flock( lfd )

	Release an flock lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<nothing>

Notes :


*/

#ifdef	HAVE_FLOCK

void
unlock_flock ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
    /* This is a no-op -- lock is removed when file is closed. */
}

#endif	/* HAVE_FLOCK */
/*

*//* unlock_lockf( lfd )

	Release a lockf lock

Accepts :

	lfd		fd for the open file to lock

Returns :

	<nothing>

Notes :


*/

#ifdef	HAVE_LOCKF

void
unlock_lockf ARGLIST( (lfd) )
    FARG( int		lfd )		/* fd for open file to lock */
{
    /* This is a no-op -- lock is removed when file is closed. */
}

#endif	/* HAVE_LOCKF */
/*

*//* unlock_lockset( lsP, lfd, lfnameP )

	Release lock(s) via lockset

Accepts :

	lsP		Pointer to LOCKSET controlling the locks
	lfd		fd for open file to lock, if relevant
	lfnameP		Full path of file to lock (the lockname base)

Returns :

	<nothing>

Notes :

	lfd and lockname may or may not be relevant.  They should
	be provided nonetheless.

*/

void
unlock_lockset ARGLIST( (lsP, lfd, lfnameP) )
   NFARG( LOCKSET	*lsP )		/* Ptr to lockset control */
   NFARG( int		lfd )		/* fd for open file to lock */
    FARG( UBYTE		*lfnameP )	/* Path to base lockfile name */
{
	int		i;
	LOCKINFO	*lockP;

    /* Remove any locks that were set */
    for ( i = 0; i < lsP->ls_lockC; ++i ) {
	lockP = &lsP->ls_locks[ i ];
	switch ( lockP->lk_type ) {
	    case LCK_DOTLOCK:
		if ( lockP->lk_lockedF )
		    unlock_dotlock( lfnameP );
		lockP->lk_lockedF = FALSE;
		break;

#ifdef	HAVE_FCNTL
	    case LCK_FCNTL:
		if ( lockP->lk_lockedF )
		    unlock_fcntl( lfd );
		lockP->lk_lockedF = FALSE;
		break;
#endif	/* HAVE_FCNTL */

#ifdef	HAVE_FLOCK
	    case LCK_FLOCK:
		if ( lockP->lk_lockedF )
		    unlock_flock( lfd );
		lockP->lk_lockedF = FALSE;
		break;
#endif	/* HAVE_FLOCK */

#ifdef	HAVE_LOCKF
	    case LCK_LOCKF:
		if ( lockP->lk_lockedF )
		    unlock_lockf( lfd );
		lockP->lk_lockedF = FALSE;
		break;
#endif	/* HAVE_LOCKF */

	    default:
		warning( "unlock_lockset: bad lock type %d", lockP->lk_type );
		break;
	}
    }
}
