/* mvmfmisc.c -- General functions

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Some general functions

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <errno.h>

#include "mvmf.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-in.h>
#include <mml/mml-str.h>
#include <mml/mml-iopf.h>
#include <mml/mml-base64.h>
#include <mml/mml-dns.h>

#include "data.h"

#include "mvmfmisc.h"

/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */



/*

*//* attr_int_clear( aiP )

	Clear an integer attribute

Accepts :

	aiP		Ptr to integer attribute info

Returns :

	<nothing>

Notes :

	Marks the attribute as unset

*/

void
attr_int_clear ARGLIST( (aiP) )
    FARG( ATTR_INT	*aiP )		/* Integer attribute */
{
    aiP->ai_setF = FALSE;
}
/*

*//* attr_int_set( aiP, val )

	Set an integer attribute

Accepts :

	aiP		Ptr to integer attribute info
	val		The value to set

Returns :

	<nothing>

Notes :


*/

void
attr_int_set ARGLIST( (aiP, val) )
   NFARG( ATTR_INT	*aiP )		/* Integer attribute */
    FARG( long		val )		/* Value to give it */
{
    aiP->ai_val = val;
    aiP->ai_setF = TRUE;
}
/*

*//* attr_str_clear( asP )

	Clear a string attribute

Accepts :

	asP		Ptr to string attribute

Returns :

	<nothing>

Notes :

	Removes the string; deallocates it if it was flagged as
	  having been allocated.

*/

void
attr_str_clear ARGLIST( (asP) )
    FARG( ATTR_STR	*asP )		/* String attribute */
{
    /* Dealloc as required */
    if ( asP->as_allocF && ( asP->as_strP != NULL ) )
	dealloc( asP->as_strP );

    asP->as_strP = NULL;
}
/*

*//* attr_str_set( asP, strP, copyF )

	Set a string attribute

Accepts :

	asP		Ptr to string attribute
	strP		New string to set
	copyF		Whether to make a copy of the string

Returns :

	BOOL		TRUE if successful
			FALSE if not

Notes :

	Returns FALSE if allocation error, or if the passed new string
	value is NULL.

*/

BOOL
attr_str_set ARGLIST( (asP, strP, copyF) )
   NFARG( ATTR_STR	*asP )		/* String attribute */
   NFARG( UBYTE		*strP )		/* New string value */
    FARG( BOOL		copyF )		/* Whether to copy the string */
{
    /* If there's already a string value, clear it first */
    if ( asP->as_strP != NULL )
	attr_str_clear( asP );

    /* Make new value reference */
    if ( copyF )
	asP->as_strP = (UBYTE *)newstr( "attr_str_set", "string",
					(char *)strP );
    else
	asP->as_strP = strP;

    asP->as_allocF = copyF;

    return ( asP->as_strP != NULL );
}
/*

*//* attr_str_unset( asP )

	Mark a string attribute as unset

Accepts :

	asP		Ptr to string attribute

Returns :

	<nothing>

Notes :

	Simply marks the attribute as unset, without concern for any
	  previous value.  This is an initialization function.

*/

void
attr_str_unset ARGLIST( (asP) )
    FARG( ATTR_STR	*asP )		/* String attribute */
{
    asP->as_strP = NULL;
}
/*

*//* b64_decode_bstr_str_n( dstP, srcP, srcL )

	Decode base64 from counted string into BSTR

Accepts :

	dstP		Ptr to BSTR to decode into
	srcP		Ptr to buffer with base64-encoded string
	srcL		Number of bytes in source string

Returns :

	<value>		TRUE if successful
			FALSE if not

Notes :

	Destination string will be NUL-terminated.

*/

BOOL
b64_decode_bstr_str_n ARGLIST( (dstP, srcP, srcL) )
   NFARG( BSTR		*dstP )		/* Destination BSTR */
   NFARG( UBYTE		*srcP )		/* Source string */
    FARG( int		srcL )		/* Source length */
{
	int		len;
	MMLB64		b64;		/* encode/decode handle */

    /* Make sure there's room in the BSTR */
    len = b64_decode_size( srcL );
    if ( dstP->bs_bL < len )
	if ( !bstr_grow( dstP, (len - dstP->bs_bL) +1 ) )
	    return ( FALSE );

    /* Do the decode */
    len = b64_decode( srcP, srcL, dstP->bs_bP, dstP->bs_bL );
    if ( len < 0 )
	return ( FALSE );

    /* Set up the BSTR properly and NUL terminate it */
    dstP->bs_bC = len;
    bstr_nulterm( dstP );

    return ( TRUE );
}
/*

*//* b64_decode_tmpstr_str_n( srcP, srcL, dstLP )

	Decode base64 from counted string into a tmpstr

Accepts :

	srcP		Ptr to buffer with base64-encoded string
	srcL		Number of bytes in source string
	dstLP		Where to put the decoded length (NULL if not wanted)

Returns :

	<value>		Pointer to a tmpstr holding the decoded string,
			NULL if some error.

Notes :

	Destination string will be NUL-terminated.

*/

UBYTE *
b64_decode_tmpstr_str_n ARGLIST( (srcP, srcL, dstLP) )
   NFARG( UBYTE		*srcP )		/* Source string */
   NFARG( int		srcL )		/* Source length */
    FARG( int		*dstLP )	/* Where to store the decoded length */
{
	int		len;
	UBYTE		*dstP;		/* Ptr to result string */
	MMLB64		b64;		/* encode/decode handle */

    /* Get a tmpstr that's big enough */
    len = b64_decode_size( srcL );
    dstP = (UBYTE *)tmpstr( len + 1 );
    if ( dstP != NULL ) {
	/* Do the decode */
	len = b64_decode( srcP, srcL, dstP, len );
	if ( len < 0 )
	    dstP = NULL;
	else
	    dstP[len] = NUL;
    }

    if ( dstLP != NULL )
	*dstLP = len;

    return ( dstP );
}
/*

*//* dns_handle

	Accesses the DNS handle

Accepts :


Returns :

	<value>		Ptr to MDNS handle ready for use
			NULL if some problem.

Notes :

	Uses the global DNS handle.

	Initializes the handle if not already done.

*/

MML_DNS *
dns_handle NOARGLIST
{
    if ( DnsP == NULL )
	DnsP = mdns_new();

    return ( DnsP );
}
/*

*//* file_rotate(fnameP, newfnameP, verC)

	Rotate a file name through n names

Accepts :

	fnameP		The base filename
	newfnameP	Any new file to make the new base filename
	verC		Number of versions to maintain

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Given a base filename "foo" this function will make backup
	  rotated names "foo.0" , "foo.1" -- etc through foo.'verC'.
	  Note that this makes one more than verC suffixed copies
	  (.0 plus 1 thru verC).

	Passing verC as 0 will result in just the .0 copy;  it is an
	  error for verC to be less than zero.

	If newfnameP is non-NULL, it will be renamed to the base name.

	Any errors will produce a warning message.

*/

BOOL
file_rotate ARGLIST( (fnameP, newfnameP, verC) )
   NFARG( char		*fnameP )	/* Base filename */
   NFARG( char		*newfnameP )	/* Any new filename */
    FARG( int		verC )		/* # of versions */
{
	int		len;
	int		verN;		/* Version number */
	char		fn1[ PATH_MAX +1 ];  /* for renames */
	char		fn2[ PATH_MAX +2 ];

    /* Idjit check on verC */
    if ( verC < 0 ) {
	warning( "file_rotate: bad value %d for verC", verC );
	return ( FALSE );
    }

    /* Rotate .0 through .verC backup copies */
    for ( verN = verC; verN > 0; --verN ) {
	/* Rename file.n-1 to file.n */
	len = snprintf( &fn1[0], sizeof(fn1), "%s.%d", fnameP, verN );
	if ( len >= sizeof(fn1) ) {
	    warning( "file_rotate, name too long: \"%s.%d", fnameP, verN );
	    return ( FALSE );
	}

	/* By definition the n-1 name fits if the above succeeded. */
	snprintf( &fn2[0], sizeof(fn2), "%s.%d", fnameP, verN -1 );

	/* Remove target, but ignore error */
	unlink( &fn1[0] );

	/* Rotate */
	if ( link( &fn2[0], &fn1[0] ) < 0 ) {
	    /* Ignore probably-benign error(s) */
	    if ( errno != ENOENT ) {
		warning( "file_rotate, link error %d verN=%d \"%s\"",
			  errno, verN, fnameP );
		return ( FALSE );
	    }
	}
    }



    /* fn2 should have .0 -- rename the base name to this */
    unlink( &fn2[0] );
    if ( link( fnameP, &fn2[0] ) < 0 ) {
	if ( errno != ENOENT ) {
	    warning( "file_rotate, link error %d making .0 from \"%s\"",
			errno, fnameP );
	    return ( FALSE );
	}
    }

    unlink( fnameP );			/* Remove current base name */

    /* Install new name if passed */
    if ( newfnameP != NULL ) {
	if ( link( newfnameP, fnameP ) < 0 ) {
	    /* Should be no benign errors here */
	    warning( "file_rotate, error %d installing \"%s\" as \"%s\"",
		      errno, newfnameP, fnameP );
	}
	else {
	    unlink( newfnameP );
	}
    }

    return ( TRUE );
}
/*

*//* ip_decode_str( srcP, ipP )

	Decode an IP(v4) address from a string

Accepts :

	srcP		String containing the IP address
	ipP		Where to put the decoded IP address

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	IP address must be complete, containing all 4 byte values separated
	by periods.

	The IP address is returned in network order.

*/

BOOL
ip_decode_str ARGLIST( (srcP, ipP) )
   NFARG( UBYTE		*srcP )		/* Source string */
    FARG( in_addr_t	*ipP )		/* Where the result goes */
{
	BOOL		dF;		/* Digit flag */
	int		o;		/* Octet number */
	unsigned int	v;		/* Octet value */
	UBYTE		ch;
	
	int		oval;		/* Octet value */
	in_addr_t	ip;

    for ( ip = 0, o = 0; o < 4; ++o ) {
	for ( dF = FALSE, v = 0; ( ch = *srcP++ ) != NUL;  dF = TRUE ) {
	    if ( !isdigit( ch ) )
		break;
	    v = (v * 10) + (ch - '0');
	}
	if ( !dF )
	    return ( FALSE );		/* No digits this octet */

	if ( ( v < 0 ) || ( v > 255 ) )
	    return ( FALSE );

	ip = ( ip << 8 ) | v;

	/* First three octets must be terminated by '.' */
	if ( ( o < 3 ) && ( ch != '.' ) )
	    return ( FALSE );
    }

    /* Last octet must be terminated by NUL */
    if ( ch != NUL )
	return ( FALSE );

    *ipP = htonl( ip );

    return ( TRUE );
}
/*

*//* path_dir_add( pathP, dirnameP )

	Adds a directory name to a directory path

Accepts :

	pathP		The linked list header for the path
	dirnameP	Name of directory to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Directory path is a linked list of names.

	An attempt is made to de-duplicate the path names.  This attempt
	is not very sophisticated.

	The name added here will be a newly allocated copy; caller need
	  not preserve the string at dirnameP.

*/

BOOL
path_dir_add ARGLIST( (pathP, dirnameP) )
   NFARG( LL_HEAD	*pathP )	/* Path list header */
    FARG( ESTR		*dirnameP )	/* Directory name */
{
	int		dlen;		/* Length of directory name */
	UBYTE		*dnstrP;	/* Ptr to actual name string */
	char		*strP;		/* Generic string pointer */
	LL_NODE		*nodeP;		/* Linked list node */

    /* Find the length of the new directory name, minus any trailing '/' */
    dlen = estr_len( dirnameP );
    dnstrP = dirnameP->es_strP;
    while ( ( dlen > 0 ) && ( dnstrP[dlen -1] == '/' ) )
	--dlen;

    /* Search the existing list to see if the path is already there. */
    for ( nodeP = pathP->llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	strP = (char *)nodeP->lln_valP;
	if ( ( strlen( strP ) == dlen ) &&
	     ( strncmp( strP, (char *)dnstrP, dlen ) == 0 ) )
	    break;			/* Already present */
    }

    if ( nodeP == NULL ) {
	/* No match found; add the new one. */
	strP = newstr_n( "path_dir_add", "new path", (char *)dnstrP, dlen );
	if ( strP == NULL )
	    return ( FALSE );
	if ( ll_addvt( pathP, strP ) == NULL )
	    return ( FALSE );
    }

    /* Already there or successfully added. */
    return ( TRUE );
}
/*

*//* pexec_std( pathP, argPP, coutfdP, cinfdP, cerrfdP )

	Exec a program, opening up standard file descriptors.

Accepts :

	pathP		Path of the program to exec
	argPP		Argument array for the program
	coutfdP		Where to put fd for output to child's stdin
	cinfdP		Where to put fd for input from child's stdout
	cerrfdP		Where to put fd for input from child's stderr

Returns :

	<value>		pid of the child exec'd, or
			-1 if failed.

Notes :

	Forks subchild, opens standard file descriptors.

	Sanitizes child environment.

*/

pid_t
pexec_std ARGLIST( (pathP, argPP, coutfdP, cinfdP, cerrfdP) )
   NFARG( char		*pathP )	/* Path to program to exec */
   NFARG( char		**argPP )	/* Arguments to program */
   NFARG( int		*coutfdP )	/* Rtn fd: output to child's stdin */
   NFARG( int		*cinfdP )	/* Rtn fd: input from child's stdout */
    FARG( int		*cerrfdP )	/* Rtn fd: input from child's stderr */
{
	int		opipe[2];	/* pipe for output to child's stdin */
	int		ipipe[2];	/* pipe for input from child's stdout */
	int		epipe[2];	/* pipe for input from child's stderr */
	pid_t		pid;

    /* Make pipe(s) and child.  It might be nice if we could assume
       bidirectional pipes, but this will do fine.
       The order of pipe open is deliberate, so that we can assume the
       fds assigned are in the order opened.
    */
    if ( pipe( &opipe[0] ) < 0 )
	warning( "pexec_std can't create opipe" );
    else {
	if ( pipe( &ipipe[0] ) < 0 )
	    warning( "pexec_std can't create ipipe." );
	else {
	    if ( pipe( &epipe[0] ) < 0 )
		warning( "pexec_std can't create epipe." );
	    else {
		/* Pipes are open, we can fork. */
		pid = fork();
		if ( pid == 0 ) {
		    /* This is the child.  Child wants to read from
		       opipe on fd 0, write to ipipe on fd 1, and write
		       to epipe on fd 2.
		    */
		    close( opipe[1] );
		    close( ipipe[0] );
		    close( epipe[0] );

		    if ( opipe[0] != 0 ) {
			/* Make this be fd 0.  We don't (or shouldn't!)
			   have to worry about one of the other pipe fds
			   being fd 0 since we opened pipes in the order
			   that we wanted the fds assigned.  Could this
			   fail on some strange system?  Maybe... if/when
			   it does, we can jump through hoops to worry
			   about it.
			*/
			if ( dup2( opipe[0], 0 ) < 0 )
			    _exit( MEX_FORKFAIL );
			close( opipe[0] );
		    }
		    if ( ipipe[1] != 1 ) {  /* Likewise, make this be fd 1 */
			if ( dup2( ipipe[1], 1 ) < 0 )
			    _exit( MEX_FORKFAIL );
			close( ipipe[1] );
		    }
		    if ( epipe[1] != 2 ) {		/* and this be fd 2 */
			if ( dup2( epipe[1], 2 ) < 0 )
			    _exit( MEX_FORKFAIL );
			close( epipe[1] );
		    }

TODO		    /* Close other FDs */

		    /* Finish sanitizing the process */
		    if ( !sanitize_process() )
			_exit( MEX_FORKFAIL );

		    /* Become the new program */
		    execv( pathP, argPP );

		    /* Can't exec... */
		    _exit( MEX_FORKFAIL );
		}

		else if ( pid > 0 ) {
		    /* Parent here.  Close the pipe ends that we aren't using. */
		    close( opipe[0] );
		    close( ipipe[1] );
		    close( epipe[1] );

		    /* Return the fds */
		    *coutfdP = opipe[1];
		    *cinfdP = ipipe[0];
		    *cerrfdP = epipe[0];

		    /* And return successfully with the pid */
		    return ( pid );
		}

		/* Here there's been an error. */
		close( epipe[0] );
		close( epipe[1] );
	    }
	    close( ipipe[0] );
	    close( ipipe[1] );
	}
	close( opipe[0] );
	close( opipe[1] );
    }

    return ( -1 );
}
/*

*//* pexec_std_end( pid, coutfd, cinfd, cerrfd )

	Finishes access to child previously opened by pexec_std()

Accepts :

	pid		Child process ID
	coutfd		fd for output to child
	cinfd		fd for input from child
	cerrfd		fd for input from child's stderr

Returns :

	<value>		waitstatus result, or -1 if some error.

Notes :

	Any of the fds may be passed as -1 to suppress closing, e.g.
	  when caller has already closed it.

*/

int
pexec_std_end ARGLIST( (pid, coutfd, cinfd, cerrfd) )
   NFARG( pid_t		pid )		/* Child pid */
   NFARG( int		coutfd )	/* fd: output to child */
   NFARG( int		cinfd )		/* fd: input from child */
    FARG( int		cerrfd )	/* fd: input from child's stderr */
{
	int		waitstat;
	pid_t		epid;		/* Pid that exited */

    /* Close the fds first, so we don't get into a waitlock */
    if ( coutfd >= 0 )
	close( coutfd );
    if ( cinfd >= 0 )
	close( cinfd );
    if ( cerrfd >= 0 )
	close( cerrfd );

    epid = waitpid( pid, &waitstat, 0 );
    if ( epid != pid )
	waitstat = -1;

    return ( waitstat );
}
/*

*//* sanitize_process()

	Sanitize the process environment

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Prepares the process for unknown use, by dropping privileges
	  and doing any other safety-related measures.

	If root, changes to the user that was determined earlier.

*/

BOOL
sanitize_process NOARGLIST
{
	BOOL		flag;		/* Some generic flag */
	int		i,j;		/* fortran, eh? */
	int		grpC;		/* Number of groups */
	int		ugrpC;		/* # groups in ugroups */
	uid_t		effuid;
	gid_t		groups[ NGROUPS +1 ];
	gid_t		ugroups[ NGROUPS +1 ];

    effuid = geteuid();			/* effective UID */

    if ( ( effuid == 0 ) && ( Myuid != 0 ) ) {
	/* If we're root, we'll prepare the uid/gid environment for
	   the proper user.
	*/

	/* Set up the groups from scratch */
	grpC = NGROUPS;
	if ( getgrouplist( MyusernameP, Myuid, &groups[0], &grpC ) < 0 ) {
	    perror( "getgrouplist" );
	    return ( FALSE );
	}

	/* Establish the groups */
	if ( setgroups( grpC, &groups[0] ) < 0 ) {
	    perror( "setgroups" );
	    return( FALSE );
	}

	/* Now become the desired user. */
	if ( setuid( Myuid ) < 0 ) {
	    perror( "setuid" );
	    return ( FALSE );
	}

	/* We'll want to cd to that user's homedir.  If it fails, we'll
	   just go to /tmp.
	*/
	if ( chdir( MyhomedirP ) < 0 ) {
	    warning( "Can't cd to \"%s\" when becoming \"%s\" -- %d.\n",
		        MyhomedirP,
		        MyusernameP,
		        errno );
	    chdir( "/tmp" );
	}
    }
    else if ( ( effuid != 0 ) && ( effuid == Myuid ) ) {
	/* We're not running as root.  We'll want to sanitize the group
	   list just in case we added any (via setgid).
	*/
	if ( setgid( getgid() ) < 0 ) {
	    perror( "setgid" );
	    return ( FALSE );
	}
    }

#if 0
    else {
	warning( "impossible condition in sanitize_process\n" );
	return ( FALSE );
    }
#endif

    return ( TRUE );
}
/*

*//* ws_token( bufPP, tokP, tokL )

	Wrapper around strtoken_ws

Accepts :

	bufPP		Ptr to the buffer pointer
	tokP		Where to put the token
	tokL		Length of token buffer

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	More or less a wrapper around strtoken_ws() from mml;
	as with strtoken_ws,
	  - the pointer at *bufPP is updated to point after the extracted
	    token.
	  - the returned token is NUL-terminated.

	The FALSE return will happen if the buffer is too small for
	  the token, and a warning message will be given.

*/

BOOL
ws_token ARGLIST( (bufPP, tokP, tokL) )
   NFARG( UBYTE		**bufPP )	/* Ptr to buffer ptr */
   NFARG( UBYTE		*tokP )		/* Where to put the token */
    FARG( int		tokL )		/* Length of token buffer */
{
	int		i;
	UBYTE		*bufP;

    /* Get a copy of the buffer pointer, in case we need it for warning */
    bufP = *bufPP;

    /* Do token-getting */
    i = strtoken_ws( bufPP, tokP, tokL, TRUE );
    if ( i >= tokL ) {
	warning( "ws_token: token too large at \"%s\"", bufP );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* ws_token_int( bufPP, lvP )

	Fetch a token, interpret as int

Accepts :

	bufPP		Ptr to the buffer pointer
	lvP		Where to put the int result (as a long)

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	The FALSE return will happen if the buffer is too small for
	  the token, and a warning message will be given.

	False will also happen if the input doesn't look like an int.

	On FALSE, the value *lvP will not be touched.

*/

BOOL
ws_token_int ARGLIST( (bufPP, lvP) )
   NFARG( UBYTE		**bufPP )	/* Ptr to buffer ptr */
    FARG( long		*lvP )		/* Where to put the int */
{
	UBYTE		tok[50];	/* Token buffer */

    /* Get the token and check the first char */
    if ( ws_token( bufPP, &tok[0], sizeof(tok) ) &&
         ( isdigit( tok[0] ) || ( tok[0] == '+' ) || ( tok[0] == '-' ) ) ) {
	/* Looks OK; get value and return it. */
	*lvP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    return ( FALSE );
}
