/* omm.c -- Manage an open mail message

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Provides a semi-opaque interface to manage a mail message that's been
opened.  The general concept is that a mail message is open for
reading on a file descriptor.  Use this module to read the message
header, manipulate header elements, and perform other functions related
to the open message.

With the addition of MIME knowledge, much of what used to be here has been
migrated over to msgpart.c (and msgpart.h).

*/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <regex.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "mvmf.h"

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "msgpart.h"
#include "omm.h"

#include "data.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "errwarn.h"

/* Local definitions */


    /* Info about an action */
typedef struct {
    char	*ac_nameP;		/* Action name */
    char	*ac_valP;		/* Action value */
}  OMM_ACTION;

/* External data referenced */

extern	OMM	*OmmP;

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

static	void	omm_learn PROTO( (OMM *ommP) );
static	OMM	*omm_open_common PROTO( (MMSG *msgP, long bodycache ) );

/* Private data */


/*

*//* omm_action_add( ommP, acnameP, acvalP );

	Add an item to the actions list

Accepts :

	ommP		Ptr to OMM header
	acnameP		Name of the action (an ESTR)
	acvalP		Value of the action (an ESTR)

Returns :

	<value>		TRUE if success
			FALSE if not

Notes :

	Adds the action without checking duplicates-- it's assumed
	the caller has checked that via omm_action_check .

	The list of actions is for reference by callers.  omm routines
	do not add or delete any action notes as side-effects of
	performing actions.

	Action name is case-insensitive (case doesn't matter)
	Action value is case-sensive (case matters).  Caller should
	  ensure common case if case isn't supposed to matter here.

*/

BOOL
omm_action_add ARGLIST( (ommP, acnameP, acvalP) )
   NFARG( OMM		*ommP )		/* OMM header ptr */
   NFARG( ESTR		*acnameP )	/* Action name */
    FARG( ESTR		*acvalP )	/* Value of action */
{
	OMM_ACTION	*actP;		/* Action block */

    /* Create the block */
    actP = (OMM_ACTION *)emalloc( "omm_action_add", "new action block",
		   sizeof(OMM_ACTION) );
    if ( actP == NULL )
	return ( FALSE );

    /* Fill in the components */
    actP->ac_nameP = newstr_estr( "omm_action_add", "action name", acnameP );
    if ( actP->ac_nameP == NULL ) {
	dealloc( actP );		/* Probably pointless */
	return ( FALSE );
    }
    actP->ac_valP = newstr_estr( "omm_action_add", "action value", acvalP );
    if ( actP->ac_valP == NULL ) {
	dealloc( actP->ac_nameP );
	dealloc( actP );
	return ( FALSE );
    }

    /* Add it to the end of the list */
    if ( ll_addvt( &ommP->om_actions, (void *)actP ) == NULL ) {
	dealloc( actP->ac_valP );
	dealloc( actP->ac_nameP );
	dealloc( actP );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* omm_action_check( ommP, acnameP, acvalP );

	Check for an item on the actions list

Accepts :

	ommP		Ptr to OMM header
	acnameP		Name of the action (an ESTR)
	acvalP		Value of the action (an ESTR)

Returns :

	<value>		TRUE if action has been registered (is on list)
			FALSE if not

Notes :

	Action name is case-insensitive (case doesn't matter)
	Action value is case-sensive (case matters).  Caller should
	  ensure common case if case isn't supposed to matter here.

*/

BOOL
omm_action_check ARGLIST( (ommP, acnameP, acvalP) )
   NFARG( OMM		*ommP )		/* OMM header ptr */
   NFARG( ESTR		*acnameP )	/* Action name */
    FARG( ESTR		*acvalP )	/* Value of action */
{
	LL_NODE		*nodeP;		/* Ptr to list node */
	OMM_ACTION	*actP;		/* Action block */

    /* Walk the linked list looking for actions. */
    for ( nodeP = ommP->om_actions.llh_firstP; nodeP != NULL;
		nodeP = nodeP->lln_nextP ) {
	actP = (OMM_ACTION *)nodeP->lln_valP;
	if ( ( estr_icmp_str( acnameP, (UBYTE *)actP->ac_nameP ) == 0 ) &&
	     ( estr_cmp_str( acvalP, (UBYTE *)actP->ac_valP ) == 0 ) )
	    return ( TRUE );		/* Found our match */
    }

    /* No match */
    return ( FALSE );
}
/*

*//* omm_delivery_error( ommP, err, dtypeP )

	Deal with a delivery error

Accepts :

	ommP		Ptr to open message handle
	err		errno value
	dtypeP		the name of the delivery type, e.g. "forward"

Returns :

	<nothing>

Notes :

	This is called when a delivery error has happened.  This
	function is expected to increment the hard or soft failure
	counts and set the "bounce requested" flag if needed.

	Some of these decisions are passed off to an MFL hook, like
	this:

	    $hook_delivery_error( (string)type, (string)error );

	with "type" being the delivery type (dtypeP), and "error"
	being the name of the error encountered.  For simplicity only
	a few specific errors are classified, with the bulk of them
	merely being identified as "soft" or "hard" .  This of course
	can be expanded as needed.  See the code and/or external
	documentation for the list of errors.  The MFL hook function
	is expected to return:

	    0 (FALSE) : To do nothing (make no note, etc);
	 non-0 (TRUE) : To record an appropriate error message.

	The MFL hook may call other MFL functions to request a
	bounce or refusal.

	NB: This is an early pass at this functionality.  It might
	change.

*/

void
omm_delivery_error ARGLIST( (ommP, err, dtypeP) )
   NFARG( OMM		*ommP )		/* Open mail message pointer */
   NFARG( int		err )		/* errno -like value */
    FARG( char		*dtypeP )	/* Delivery type */
{
	STSCODE		status;
	BOOL		retval;
	BOOL		hardF;		/* Set if hard error */
	char		*errnameP;	/* Name/class of error */
	char		*errstrP;	/* Any specific note to record.. */
	char		mflcmd[200];	/* MFL code string */
	VALUE		hookval;

    switch ( err ) {

	    /* Specific error codes that we recognize */
	case	EDQUOT:			/* Disc quota exceeded */
	    errnameP = "quota";
	    errstrP = "This mailbox/folder is over quota, i.e., it is full.";
	    hardF = TRUE;
	    break;

	    /* Known error codes that we treat as "soft" (i.e.,
	       maybe it will work if we try again later.)
	    */
	case	ECHILD:			/* No child processes */
	case	ENFILE:			/* Too many open files in system */
	case	ENOSPC:			/* No space left on device */
	case	EROFS:			/* Read-only file system */
	case	EUSERS:			/* Too many users */
#ifdef	EPROCLIM
	case	EPROCLIM:		/* Too many processes */
#endif	/* EPROCLIM */
	    errnameP = "soft";
	    errstrP = NULL;
	    hardF = FALSE;
	    break;


	    /* Known error codes that we treat as "hard" (i.e.,
	       it probably won't work again on a retry)
	    */
	case	EPERM:			/* Operation not permitted */
	case	ENOEXEC:		/* Exec format error */
	    errnameP = "hard";
	    errstrP = NULL;
	    hardF = TRUE;
	    break;

	default:			/* Anything else, we treat as soft */
	    errnameP = "soft";
	    errstrP = NULL;
	    hardF = FALSE;
	    break;
    }

    /* Do delivery error hook */
    if ( mfl_func_hook_find( "$hook_delivery_error" ) ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( \"%s\", \"%s\" );",
		      "$hook_delivery_error",
		      dtypeP,
		      errnameP ) >= sizeof(mflcmd) ) {
	    error( "omm_delivery_error: mflcmd buffer too small!" );
	    return;
	}

	status = mfl_run_str( &hookval, estr_tmp_str( &mflcmd[0] ),
			      estr_tmp_str( "Mc" ) );

	retval = val_get_int( &hookval );
	val_finish( &hookval );
    }
    else {
	/* MFL hook is not defined: do something standard.  In this
	   version, "something standard" is to record the error.
	*/
	retval = TRUE;
    }

    /* Now increment the appropriate counts. */
    if ( hardF )
	++ommP->om_hardfailC;
    else
	++ommP->om_softfailC;

    if ( retval ) {
	/* An error annotation is desired. */
	if ( errstrP )
	    warning( "%s", errstrP );
	else
	    warning( "Delivery error; errno %d: %s", err, strerror( err ) );
    }

    /* Log the delivery error. */
    log_both( "Error: delivery \"%s\" errno %d errname \"%s\" errclass \"%s\"",
		dtypeP, errno, errnameP,
		hardF ? "hard" : "soft" );
}
/*

*//* omm_free( ommP )

	Free /release open mail message

Accepts :

	ommP		Ptr to open mail message to free

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Frees the ommP and its parts, including its associated message,
	  with whatever implications freeing those parts have.

	Will return FALSE if any of its steps fail; however, will still
	  proceed to other steps.

*/

BOOL
omm_free ARGLIST( (ommP) )
    FARG( OMM		*ommP )		/* The OMM to free */
{
	BOOL		okF;
	int		i;

    okF = TRUE;

    /* Just go through and free up the pieces */

    if ( ommP->om_aesenderP != NULL )
	dealloc( ommP->om_aesenderP );

    for ( i = 0; i < ommP->om_ripC; ++i )
	dealloc( ommP->om_ripPP[i] );
    if ( ommP->om_ripPP != NULL )
	dealloc( ommP->om_ripPP );

    if ( ( ommP->om_errsP != NULL ) &&
	!bstr_free( ommP->om_errsP ) )
	okF = FALSE;

    /* Save the best for last */
    if ( ( ommP->om_mmsgP != NULL ) &&
         !mmsg_free( ommP->om_mmsgP ) )
	okF = FALSE;

    /* And delete the omm */
    dealloc( ommP );

    return ( okF );
}
/*

*//* omm_openfp( fP, bodycache )

	"Open" a mail message that's attached to a file pointer

Accepts :

	fP		FILE ptr on which a mail message is open.
	bodycache	Amount of body to be cached (approximate)


Returns :

	<value>		opaque handle for future omm references.
			NULL if failure.

Notes :


*/


void *
omm_openfp ARGLIST( (fP, bodycache) )
   NFARG( FILE		*fP )		/* File that the message is on. */
    FARG( long		bodycache )	/* Approximate amount of body to cache */
{
	OMM		*ommP;		/* OMM info structure */
	MMSG		*msgP;		/* Message */

    /* Make a new message from the fP */
    msgP = mmsg_openfp( fP );
    if ( msgP == NULL )
	return ( NULL );

    ommP = omm_open_common( msgP, bodycache );

    return ( (void *)ommP );
}
/*

*//* omm_openip( inP, bodycache )

	"Open" a mail message for input from an MML_IN input stack

Accepts :

	inP		MML_IN input stack from which the message
			is to be read.
	bodycache	Amount of body to be cached (approximate)


Returns :

	<value>		opaque handle for future omm references.
			NULL if failure.

Notes :


*/


void *
omm_openip ARGLIST( (inP, bodycache) )
   NFARG( MML_IN	*inP )		/* Input stack to take message from */
    FARG( long		bodycache )	/* Approximate amount of body to cache */
{
	OMM		*ommP;		/* OMM info structure */
	MMSG		*msgP;		/* Message */

    /* Make a new messagepart from the fP */
    msgP = mmsg_openip( inP );
    if ( msgP == NULL )
	return ( NULL );

    ommP = omm_open_common( msgP, bodycache );

    return ( (void *)ommP );
}
/*

*//* omm_rip_add( ommP, ripP )

	Add an entry to the list of responsible sender IPs

Accepts :

	ommP		Ptr to open mail message
	ripP		A responsible sender IP to add

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
omm_rip_add ARGLIST( (ommP, ripP) )
   NFARG( OMM		*ommP )		/* Ptr to open message */
    FARG( ESTR		*ripP )		/* rip to add */
{
	int		i;

    /* Make sure it isn't already in the table */
    for ( i = 0; i < ommP->om_ripC; ++i )
	if ( estr_icmp_str( ripP, (UBYTE *)ommP->om_ripPP[i] ) == 0 )
	    return ( TRUE );		/* Found, just be OK */

    /* Add new one. */
    if ( ommP->om_ripC == 0 )
	ommP->om_ripPP = (char **)emalloc( "omm_rip_add", "new rip table",
				      sizeof( char ** ) );
    else
	ommP->om_ripPP = (char **)erealloc( "omm_rip_add", "new rip table",
				      ommP->om_ripPP,
			      ( ommP->om_ripC +1 ) * sizeof( char ** ) );

    ommP->om_ripPP[ ommP->om_ripC++ ] = newstr_estr( "omm_rip_add",
					       "new rip entry",
					       ripP );

    return ( TRUE );
}
/*

*//* omm_write_errs(ommP, ofP, bracketF)

	Write any cached error text to an open file ptr

Accepts :

	ommP		Ptr to open message handle
	ofP		File ptr to write error text to
	bracketF	TRUE to enclose error text in bracketing text

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Assuming that the body follows the header, a blank line is
	written first.

*/

BOOL
omm_write_errs ARGLIST( (ommP, ofP, bracketF) )
   NFARG( OMM		*ommP )		/* Ptr to open message */
   NFARG( FILE		*ofP )		/* Where to write body */
    FARG( BOOL		bracketF )	/* Whether to include bracketing text */
{
	BOOL		okF;
	int		ch;
	char		*strP;

    /* If there's any error/warning text attached, push it into the
       message body.
    */
    okF = TRUE;
    if ( ( ommP->om_errsP != NULL ) && ( ommP->om_errsP->bs_bC > 0 ) ) {
	bstr_nulterm( ommP->om_errsP );

	if ( bracketF )
	    okF = okF && fprintf( ofP, 
	        "\n\n---- text added by mvmf mail filter ----\n" );
	okF = okF && fprintf( ofP, "\n" );

	/* Walk the text, shifting to the right by some margin */
	strP = (char *)ommP->om_errsP->bs_bP;
	if ( *strP != '\n' )
	    okF = okF && fprintf( ofP, "%5c", ' ' );

	while ( ( ch = *strP++ ) != NUL ) {
	    okF = okF && ( fputc( ch, ofP ) != EOF );
	    if ( ( ch == '\n' ) && ( *strP != NUL ) && ( *strP != '\n' ) )
		okF = okF && fprintf( ofP, "%5c", ' ' );
	}

	okF = okF && fprintf( ofP, "\n" );
	if ( bracketF )
	    okF = okF && fprintf( ofP, 
	        "\n---- end of text added by mvmf mail filter ----\n\n" );
    }

    return ( okF );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* omm_learn( ommP )

	Learn some stuff about the open messages

Accepts :

	ommP		Ptr to open mail message object

Returns :

	<value>		nothing

Notes :


*/

static void
omm_learn ARGLIST( (ommP) )
    FARG( OMM		*ommP )		/* Open message object */
{
	void		*hdrP;		/* Ptr to header handle */
	char		*strP;		/* String */
	MSGPART		*mpP;
	REFSTR		*rsP;		/* REFSTR ptr */
	OMM		*ommsaveP;

    mpP = ommP->om_mmsgP->mm_topmpP;	/* Top message part */
    /* Attempt to ascertain the envelope sender or a reasonable equivalent.
       Use the SMTP envelope if available, otherwise use the value in the
       Return-Path header.  Note that we specifically do NOT use the info in
       the "From " line-- "Return-Path" is the correct place to find it.
    */
    rsP = NULL;
    if ( refstr_len( E_SenderP ) > 0 ) {
	/* Use the SMTP envelope sender. */
	rsP = E_SenderP;
	rsP->rs_bX = 0;
    }
    else {
	hdrP = msgpart_hdr_next( mpP, NULL,
				 estr_tmp_str( (UBYTE *)"Return-Path" ) );
	if ( hdrP != NULL )
	    rsP = msgpart_hdr_str( mpP, hdrP );
    }

    /* Extract next address from the BSTR containing the sender */
    if ( rsP != NULL ) {
	strP = msgpart_refstr_addr( mpP, rsP, AP_LOCAL|AP_DOMAIN );
	if ( strP != NULL )
	    ommP->om_aesenderP = newstr( "omm_learn",
					"ascertained envelope sender",
						strP );
    }

    /* Apply the read-message hook if one exists. */

TODO	/* Hack to set up global OmmP which is likely to be used by
	   this hook.  Probably need a better way.
	*/
    ommsaveP = OmmP;
    OmmP = ommP;
    mfl_func_hook( "$hook_msg_read_after" );
    OmmP = ommsaveP;
}
/*

*//* omm_open_common( msgP, bodycache )

	Common part of omm_openxx

Accepts :

	msgP		Ptr to message object that's been opened
	bodycache	Amount of body to be cached (approximate)

Returns :

	<value>		opaque handle for future omm references.
			NULL if failure.

Notes :

	Called by one of the omm_openxx functions after the
	top-level messagepart (and its input container) has
	been set up.

	On failure, will free up the msgP.

	bodycache is constrained to lie within OMM_BODYCACHE_MIN and
	  OMM_BODYCACHE_MAX, inclusive.

*/

static OMM *
omm_open_common ARGLIST( (msgP, bodycache) )
   NFARG( MMSG		*msgP )		/* "Opened" message */
    FARG( long		bodycache )	/* Approximate amount of body to cache */
{
	OMM		*ommP;		/* OMM info structure */

    /* Get a new OMM pointer */
    ommP = (OMM *)emalloc( "omm_openfp", "new OMM", sizeof(OMM) );
    if ( ommP == NULL ) {
TODO	/* Should clean up mpP parts here probably */
	dealloc( msgP );
	return ( NULL );
    }

    /* Set up body cache parameters */
    if ( bodycache < OMM_BODYCACHE_MIN )
	bodycache = OMM_BODYCACHE_MIN;
    if ( bodycache > OMM_BODYCACHE_MAX )
	bodycache = OMM_BODYCACHE_MAX;
	
    msgP->mm_cachemax = bodycache;
    msgP->mm_cacheC = 0;

    /* Init the omm elements */
    ommP->om_mmsgP = msgP;
    ommP->om_curmpP = msgP->mm_topmpP;

    ll_init( &ommP->om_actions );	/* Init the actions list */

    ommP->om_ripPP = NULL;		/* No list of responsible IPs */
    ommP->om_ripC = 0;

    /* Create an error message area */
    ommP->om_errsP = bstr_new( "omm_openfp", 5000, 50000 );
    if ( ommP->om_errsP == NULL ) {
	mmsg_free( msgP );
	dealloc( ommP );
	return ( NULL );
    }
    ommP->om_errsP->bs_chunk = 1000;

    ommP->om_aesenderP = NULL;		/* Ascertained envelope sender */

    ommP->om_storingF = FALSE;		/* Not currently storing */
    ommP->om_fwdingF = FALSE;		/* Not currently forwarding */
    ommP->om_pipingF = FALSE;		/* Not currently piping */

    ommP->om_ikeepF = TRUE;		/* Turn on implicit keep */
    ommP->om_wantbounceF = FALSE;	/* No bounce wanted */
    ommP->om_wantrejectF = FALSE;	/* No SMTP-time reject wanted. */

    ommP->om_rejectF = FALSE;		/* Reject hasn't happened */
    ommP->om_bouncedF = FALSE;		/* No bounce has happened */

    ommP->om_deliveryC = 0;		/* No successful deliveries */
    ommP->om_softfailC = 0;		/* No soft failures */
    ommP->om_hardfailC = 0;		/* No hard failures */


    /* Read the top-level messagepart and any subparts */
    if ( !msgpart_read_msgpart( msgP->mm_topmpP ) )
	return ( NULL );

    /* Learn some useful stuff about the mail message */
    omm_learn( ommP );

    return ( ommP );
}
