/* qmail-queue-if -- interface to qmail's qmail-queue program

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Functions to interface to qmail's qmail-queue program.

Much of the particulars of the qmail-queue interface code was derived
from 'man qmail-queue' and by inspection of some of djb's qmail code.
(ain't open source great?)

Use must follow a particular sequence.  It's tempting to enforce some
of these sequences via data elements (whether queueing things up and
releasing them in the proper order, or setting state flags to give errors
if the sequence is wrong), but I guess we'll just make the caller do
the right thing.  Sequence is:

    Open up qmail-queue via qmail_queue_open()
    Optionally prefix text to the message via qmail_queue_write_text()
      (e.g. for "Received" line)
    Send the message to qmail-queue via qmail_queue_write_message()
    Give the sender address via qmail_queue_write_sender()
    Give recipients via qmail_queue_write_recipient()
    Finish up with qmail_queue_close()

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <ctype.h>


#include "mvmf.h"

#ifdef	QMAIL		/* The rest is only enabled if qmail stuff is. */

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-dns.h>

#include "mvmfmisc.h"
#include "msgpart.h"
#include "qmail-queue-if.h"



/* Local definitions */



/* External data referenced */



/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */



/* Private data */



/*

*//* qmail_queue_close( qqP, keepF )

	Close a previously-opened connection to qmail-queue

Accepts :

	qqP		QMQIF ptr (from qmail_queue_open)
	keepF		TRUE to keep the message, FALSE to abort.

Returns :

	<value>		a status code, one of:
			  SCOK:  operation was completed
			  SCNOTDONE: some temporary error
			  SCERR: some permanent error.

Notes :


*/

STSCODE
qmail_queue_close ARGLIST( (qqP, keepF) )
   NFARG( QMQIF		*qqP )		/* qmail-queue handle */
    FARG( BOOL		keepF )		/* whether to keep the message. */
{
	BOOL		okF;
	STSCODE		status;
	int		waitstat;
	int		exitcode;
	pid_t		pid;

    /* Send final envelope nul */
    okF = ( fputc( NUL, qqP->qq_envfP ) != EOF );

    if ( fclose( qqP->qq_msgfP ) != 0 )
	okF = FALSE;
    if ( fclose( qqP->qq_envfP ) != 0 )
	okF = FALSE;

    /* Collect exit status */
    pid = waitpid( qqP->qq_pid, &waitstat, 0 );
    if ( pid != qqP->qq_pid )
	return ( SCERR );

TODO	;
    /* Note: we really need to bubble up these individual errors
       better.  But this utility is still in early formative mode.
       (how long will this comment be here?)
    */

    exitcode = WEXITSTATUS( waitstat );
    switch ( exitcode ) {
	case QQS_OK:			/* OK, message accepted */
	    status = SCOK;
	    break;

	    /* These are permanent errors. */
	case QQS_ELONGSTRING:		/* Envelope string too long */
	case QQS_REJECTED:		/* Recipients rejected */
	    status = SCERR;
	    break;


	    /* These are temporary errors */
	case QQS_MEMORY:		/* Out of memory */
	case QQS_TIMEOUT:		/* Timeout */
	case QQS_WRITE:			/* Write error */
	case QQS_ACCEPT:		/* Error reading data we're sending it */
	case QQS_READ:			/* Error reading config file */
	case QQS_NET_OUT:		/* Can't initiate some net connection */
	case QQS_QMAILDIR:		/* Some problem with the qmail directory */
	case QQS_QUEUE_DIR:		/* Problem with the queue directory */
	case QQS_QUEUE_PID:		/* Problem with queue/pid */
	case QQS_QUEUE_MESS:		/* Problem with queue/mess */
	case QQS_QUEUE_INTD:		/* Problem with queue/intd */
	case QQS_QUEUE_TODO:		/* Problem with queue/todo */
	case QQS_TMP_REJECT:		/* Recipients temporarily rejected */
	case QQS_MSCON_TIMEOUT:		/* Connection to mailserver timeout */
	case QQS_MSCON_REJECT:		/* Connection to mailserver rejected */
	case QQS_MSCON_ABORT:		/* Connection to mailserver aborted */
	case QQS_ERR:			/* Internal error */
	    status = SCNOTDONE;
	    break;


	    /* This is documented as a temporary error, but it sure
	       seems permanent to me..
	    */
	case QQS_ENVERR:		/* Envelope format error */
	    status = SCERR;
	    break;


	case QQS_EXEC:			/* Internal code:
					   Can't exec qmail-queue.
					*/
	    status = SCERR;		/* Treat as hard error. */
	    break;


	default:			/* Some unknown error */
	    status = SCERR;
	    break;
    }

    if ( ( status == SCOK ) && !okF )
	status = SCERR;

    return ( status );
}
/*

*//* qmail_queue_open()

	Open a connection to qmail-queue

Accepts :



Returns :

	<value>		Ptr to a QMQIF struct if successfull,
			NULL if not.

Notes :


*/

QMQIF *
qmail_queue_open NOARGLIST
{
	QMQIF		*qqP;		/* Ptr to control struct */
	pid_t		pid;
	int		msgpipe[2];	/* Pipe for the message */
	int		envpipe[2];	/* Pipe for the envelope */
	char		qqpath[ PATH_MAX +1 ];
	char		*argv[2];	/* Arguments for execv */

    /* Where qmail-queue is */
    if ( snprintf( &qqpath[0], PATH_MAX, "%s/bin/%s",
			QMAILDIR, "qmail-queue" ) >= PATH_MAX )
	return ( NULL );

    /* Allocate and init control structure. */
    qqP = (QMQIF *)emalloc( "qmail_queue_open", "QMQIF struct",
				sizeof( QMQIF ) );
    if ( qqP == NULL )
	return ( NULL );


    /* qmail-queue accepts a message on fd 0 and envelope info on fd 1.
       Make a pipe for each.
    */
    if ( pipe( &msgpipe[0] ) < 0 )
	return ( NULL );
    if ( pipe( &envpipe[0] ) < 0 ) {
	close( msgpipe[0] );
	close( msgpipe[1] );
	dealloc( qqP );
	return ( NULL );
    }

    /* Fork to invoke qmail-queue */
    pid = fork();
    if ( pid < 0 ) {
	/* Error.  Undo all of the above. */
	close( msgpipe[0] );
	close( msgpipe[1] );
	close( envpipe[0] );
	close( envpipe[1] );
	dealloc( qqP );
	return ( NULL );
    }

    if ( pid == 0 ) {
	/* Child process here.  Sanitize first. */
	if ( !sanitize_process() )
	    _exit( QQS_EXEC );

	/* we'll use the read ends of the pipes, so close the write
	   ends and arrange to have the file descriptors on 0 and 1 as
	   required by qmail-queue.

	   On failure to invoke qmail-queue for any reason we return
	   a special exit code QQS_EXEC that doesn't match any
	   valid qmail-queue exit code.  (Value chosen by mimic'ing
	   djb.)
	*/
	close( msgpipe[1] );
	close( envpipe[1] );

	if ( msgpipe[0] != 0 ) {
	    if ( dup2( msgpipe[0], 0 ) < 0 )
		_exit( QQS_EXEC );
	    close( msgpipe[0] );
	}
	if ( envpipe[0] != 1 ) {
	    if ( dup2( envpipe[0], 1 ) < 0 )
		_exit( QQS_EXEC );
	    close( envpipe[1] );
	}

	/* Turn into qmail-queue */
	argv[0] = "qmail-queue";
	argv[1] = NULL;
	execv( &qqpath[0], &argv[0] );

	_exit( QQS_EXEC );		/* Some failure. */
    }

    /* Parent here.  We'll use the write ends of the pipes.  Close the
       read ends and remember the fds.
    */
    close( msgpipe[0] );
    close( envpipe[0] );

    if ( ( qqP->qq_msgfP = fdopen( msgpipe[1], "w" ) ) == NULL ) {
	close( msgpipe[1] );
	close( envpipe[1] );
	dealloc( qqP );
	return ( NULL );
    }

    if ( ( qqP->qq_envfP = fdopen( envpipe[1], "w" ) ) == NULL ) {
	close( envpipe[1] );
	fclose( qqP->qq_msgfP );
	dealloc( qqP );
	return ( NULL );
    }

    qqP->qq_pid = pid;
    qqP->qq_errF = FALSE;

    /* Return the handle */
    return ( qqP );
}
/*

*//* qmail_queue_write_message( qqP, mpP )

	Write a message to qmail-queue

Accepts :

	qqP		QMQIF ptr (from qmail_queue_open)
	mpP		Ptr to top-level message part

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
qmail_queue_write_message ARGLIST( (qqP, mpP) )
   NFARG( QMQIF		*qqP )		/* qmail-queue handle */
    FARG( MSGPART	*mpP )		/* Message part */
{
	BOOL		okF;

    if ( !qqP->qq_errF ) {
	/* Write the headers */
	okF = msgpart_write_hdrs( mpP, qqP->qq_msgfP, FALSE );

	/* Write the body */
	okF = okF && msgpart_write_body( mpP, qqP->qq_msgfP, FALSE );

	if ( !okF )
	    qqP->qq_errF = TRUE;
    }

    return ( !qqP->qq_errF );
}
/*

*//* qmail_queue_write_recipient( qqP, addrP )

	Write the recipient address to qmail-queue

Accepts :

	qqP		QMQIF ptr (from qmail_queue_open)
	addrP		Recipient email address

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
qmail_queue_write_recipient ARGLIST( (qqP, addrP) )
   NFARG( QMQIF		*qqP )		/* qmail-queue handle */
    FARG( UBYTE		*addrP )	/* Recipient address */
{
	BOOL		okF;

    if ( !qqP->qq_errF ) {
	okF = ( fputc( 'T', qqP->qq_envfP ) != EOF );

	okF = okF && ( fputs( (char *)addrP, qqP->qq_envfP ) == 0 );

	okF = okF && ( fputc( NUL, qqP->qq_envfP ) != EOF );

	if ( !okF )
	    qqP->qq_errF = TRUE;
    }

    return ( !qqP->qq_errF );
}
/*

*//* qmail_queue_write_sender( qqP, addrP )

	Write the sender address to qmail-queue

Accepts :

	qqP		QMQIF ptr (from qmail_queue_open)
	addrP		Sender email address

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
qmail_queue_write_sender ARGLIST( (qqP, addrP) )
   NFARG( QMQIF		*qqP )		/* qmail-queue handle */
    FARG( UBYTE		*addrP )	/* Sender address */
{
	BOOL		okF;

    if ( !qqP->qq_errF ) {
	okF = ( fputc( 'F', qqP->qq_envfP ) != EOF );

	okF = okF && ( fputs( (char *)addrP, qqP->qq_envfP ) == 0 );

	okF = okF && ( fputc( NUL, qqP->qq_envfP ) != EOF );

	if ( !okF )
	    qqP->qq_errF = TRUE;
    }

    return ( !qqP->qq_errF );
}
/*

*//* qmail_queue_write_text( qqP, textP )

	Write some message text to qmail-queue

Accepts :

	qqP		QMQIF ptr (from qmail_queue_open)
	textP		Ptr to NUL-terminated text string to write.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Intended as a way to prefix a "Received" line to the message.

*/

BOOL
qmail_queue_write_text ARGLIST( (qqP, textP) )
   NFARG( QMQIF		*qqP )		/* qmail-queue handle */
    FARG( UBYTE		*textP )	/* Text to write */
{
    if ( !qqP->qq_errF ) {
	if ( fputs( (char *)textP, qqP->qq_msgfP ) != 0 )
	    qqP->qq_errF = TRUE;
    }

    return ( !qqP->qq_errF );
}
/*

*/

#endif	/* QMAIL */
