/* xmtp.c -- functions in support of xMTP dialog

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains some general functions to help an application engage in
an xMTP dialog (SMTP, etc).

*/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <ctype.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>

#include "mvmf.h"
#include "msgpart.h"
#include "omm.h"
#include "xmtp.h"

/* Local definitions */



/* External data referenced */

extern	UWORD	Mchartbl[];		/* Mail characters table */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Get a byte from a dot-stuffed input stream */
static	int	xmtpdss_getb PROTO( ( MML_IN *inP, BIO *bioP ) );


/* Private data */

static	ULONG	Msg_maxsize;		/* Maximum message size */
static	ULONG	Msg_byteC;		/* Current message size */
static	UBYTE	*ReplyP = NULL;		/* Reply buffer */
static	int	ReplyL = 0;		/* Size of reply buffer */

/*

*//* in_h_xmtp_dss_getc( instrP )

	mml-in style input handler for XMTPDSS dot-stuffed input stream

Accepts :

	instrP		Ptr to the input stream

Returns :

	<value>		As per mml-in streams,
			>=0: the byte input
			< 0: an ending code (here, EOF)

Notes :

	This takes input from another stream and processes dot-stuffed
	input.  It also pays attention to the size limit of the message.
	If the limit is exceeded, the input is simply ignored.

*/

int
in_h_xmtp_dss_getc ARGLIST( (instrP) )
    FARG( MML_INS	*instrP )	/* The input stream */
{
	int		ch;
	BIO		*bioP;
	MML_IN		*inP;

    /* Normal processing first */
    ch = xmtpdss_getb( instrP->is_inP, (BIO *)instrP->is_handleP );
    if ( ( ch == EOF ) ||
         ( Msg_maxsize == 0 ) ||
	 ( ++Msg_byteC < Msg_maxsize ) )
	return ( ch );

    /* We overflowed.  Flush to end; must maintain the end-of-line
       flag so the dot-unstuffing will work.
    */
    inP = instrP->is_inP;
    bioP = (BIO *)instrP->is_handleP;
    while ( ( ch = xmtpdss_getb( inP, bioP ) ) != EOF ) {
	inP->in_eolF = ( ch == '\n' );
    }

    return ( ch );
}
/*

*//* in_stream_push_xmtp_dss(inP, srcinP, nameP )

	Push a dotstuffed line handler onto an input stack

Accepts :

	inP		Input stack to push onto
	srcinP		Source input stream
	nameP		Name for this stream, or NULL if none
	maxsize		Maximum bytes to accept

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Pushes a dotstuffed character input stream onto inP;
	the input stream fetches from the srcinP stack.

*/

BOOL
in_stream_push_xmtp_dss ARGLIST( (inP, bioP, nameP) )
   NFARG( MML_IN	*inP )		/* Input stack */
   NFARG( BIO		*bioP )		/* Source buffered input */
    FARG( char		*nameP )	/* Name for the stream */
{
    /* Push onto the stack */
    return ( in_stream_push( inP, in_h_xmtp_dss_getc, NULL,
			     (void *)bioP, nameP ) );
}
/*

*//* xmtp_getcmd( xmtP )

	Input and tokenize a command from xMTP peer

Accepts :

	xmtP		Ptr to xMTP dialog control block

Returns :

	<value>		SCOK if OK;
			SCNOTDONE if input could not be handled
			  (too much data for buffer);
			SCBREAK if timeout (hey, I'm not proud);
			SCERR if some error.

Notes :

	On SCERR return, caller should check the BIO result.

	On SCOK return, this function sets up:
	  xmtP->mt_cmdP to point to the command word in the buffer;
	  xmtP->mt_cmdL is the number of bytes in the command word
	    (note that the command word is probably NOT NUL-terminated);
	  xmtP->mt_cmdtailP to point to the first non-blank after the
	    command word.

*/

STSCODE
xmtp_getcmd ARGLIST( (xmtP) )
    FARG( XMTP		*xmtP )		/* control block */
{
	STSCODE		status;
	UBYTE		ch;
	int		len;
	UBYTE		*bP;

    /* Input a line from client */
    status = xmtp_getline( xmtP, xmtP->mt_cmbsP );
    if ( status != SCOK )
	return ( status );

#ifdef BAD_IDEA
    /* Tokenize it */
    tkline( xmtP->mt_cmbsP, xmtP->mt_cmdtkP );

    /* Return the number of tokens */
    return ( xmtP->mt_cmdtkP->tp_tC );
#endif

    /* Isolate the command word and the command tail */
    xmtP->mt_cmdP = bP = (UBYTE *)solbat( (char *)xmtP->mt_cmbsP->bs_bP, FALSE );
    for ( ; (ch = *bP) != NUL; ++bP ) {
	if ( ( ch == ' ' ) || ( ch == '\t' ) )
	    break;
    }
    xmtP->mt_cmdL = bP - xmtP->mt_cmdP;
    xmtP->mt_cmdtailP = (UBYTE *)solbat( (char *)bP, FALSE );

    return ( SCOK );
}
/*

*//* xmtp_getline( xmtP, bsP )

	Input a single line of text from xMTP peer

Accepts :

	xmtP		Ptr to xMTP dialog control block
	bsP		Ptr to a BSTR to contain the line

Returns :

	<value>		SCOK if OK;
			SCNOTDONE if input could not be handled
			  (too much data for buffer);
			SCBREAK if timeout
			SCERR if some error.

Notes :

	On SCERR return, caller should check the BIO result.

*/

STSCODE
xmtp_getline ARGLIST( (xmtP, bsP) )
   NFARG( XMTP		*xmtP )		/* control block */
    FARG( BSTR		*bsP )		/* Destination BSTR */
{
	int		len;

    /* Input a record */
    len = igetbtrec( xmtP->mt_inP, bsP, XMTP_CMD_TXR, NULL );
		    
    if ( len < 0 ) {
	if ( xmtP->mt_in.bio_result == IOR_TIMEOUT )
	    return ( SCBREAK );
	return ( SCERR );
    }

    /* Check for overflow */
    if ( bsP->bs_bC == bsP->bs_bM )
	return ( SCNOTDONE );

    return ( SCOK );
}
/*

*//* xmtp_getmsg( xmtP, ommPP, maxsize )

	Get a message from the xMTP session

Accepts :

	xmtP		Ptr to XMTP control block
	ommPP		Where to put the OMM message pointer if successful
	maxsize		The maximum number of bytes to take

Returns :

	<value>		status code

Notes :

	Reads a message from the xMTP dialog, using dot-stuffed input,
	e.g. in response to a DATA command.

	The variable at ommPP is set non-NULL only if a valid message
	is gathered.

	Possible return values should be (are?):

	  SCOK -- message gathered OK.

	  SCERR -- failure to collect message at some underlying level.

	  SCBREAK -- message size limit exceeded.

	  SCNOTDONE -- some failure in input.  The BIO result information
	     should be consulted for more detail.

*/

STSCODE
xmtp_getmsg ARGLIST( (xmtP, ommPP, maxsize) )
   NFARG( XMTP		*xmtP )		/* XMTP object */
   NFARG( OMM		**ommPP )	/* Where to return the omm ptr */
    FARG( ULONG		maxsize )	/* Maximum number of bytes to take */
{
	STSCODE		status;
	OMM		*ommP;
	MML_IN		*inP;		/* An input stack */

    /* Message size limit is controlled through statics... this should
       be improved, eh?  Or, perhaps, to the message-reading function.
    */
    Msg_maxsize = maxsize;
    Msg_byteC = 0;

    /* Pre-clear the resulting OMM */
    *ommPP = NULL;

    /* Setup an input stack and attach the dot-stuffed input to it. */
    if ( ( inP = in_new() ) == NULL )
	return ( SCERR );

    if ( !in_stream_push_xmtp_dss( inP, &xmtP->mt_in, "message input" ) )
	return ( SCERR );

    /* Open the mail message */
    ommP = omm_openip( inP, 250000L );

    if ( ommP == NULL ) {
	status = SCERR;
    }
    else {
	/* Disassociate the container with the input stream, since we
	   are destroying the input stream in this function.
	*/
	if ( ommP->om_mmsgP != NULL )
	    ommP->om_mmsgP->mm_conP->mc_inP = NULL;

	/* If there's still an input stream, something failed. */
	if ( in_char( inP ) != EOF )
	    status = SCERR;

	/* Check for size overflow */
	else if ( ( Msg_maxsize != 0 ) &&
                  ( Msg_byteC >= Msg_maxsize ) )
	    status = SCBREAK;

	/* All OK */
	else
	    status = SCOK;

	/* If we're still good, make sure the input status is
	   IOR_NONE, which indicates a valid dot-terminated message.
	   Any other status is a sign of an input problem.
	*/
	if ( ( status == SCOK ) &&
	     ( xmtP->mt_in.bio_result != IOR_NONE ) )
	    status = SCNOTDONE;

	if ( status != SCOK ) {
	    /* Bad status, remove the message */
	    omm_free( ommP );
	    ommP = NULL;
	}
    }

    in_destroy( inP );
    *ommPP = ommP;

    return ( status );
}
/*

*//* xmtp_init( fdin, fdout )

	Initialize an xMTP control block

Accepts :

	fdin		Input file handle
	fdout		Output file handle

Returns :

	<value>		Pointer to new XMTP block if successful
			NULL if not

Notes :


*/

XMTP *
xmtp_init ARGLIST( (fdin, fdout) )
   NFARG( int		fdin )		/* Input file handle */
    FARG( int		fdout )		/* Output file handle */
{
	XMTP		*xmtP;		/* XMTP ptr */

    xmtP = emalloc( "xmtp_init", "xMTP block", sizeof( XMTP ) );
    if ( xmtP == NULL )
	return ( NULL );

    /* Set up command buffer */
    if ( ( xmtP->mt_cmbsP = bstr_new( "xmtp cmd", 5000, 0 ) ) == NULL )
	return ( NULL );

    /* Set up an input stream and attach xMTP input to it */
    if ( ( xmtP->mt_inP = in_new() ) == NULL )
	return ( NULL );
    if ( !in_stream_push_bio( xmtP->mt_inP, &xmtP->mt_in, "xMTP input" ) )
	return ( NULL );

    bio_init( &xmtP->mt_in, fdin );
    bio_init( &xmtP->mt_out, fdout );

    xmtP->mt_ecodesF = TRUE;	/* Extended codes are on by default */

    return ( xmtP );
}
/*

*//* xmtp_mbox_alloc()

	Allocate and initialize an XMTPMBX struct

Accepts :

	<nothing>

Returns :

	<value>		Ptr to XMTPMBX struct
			  NULL if failure

Notes :


*/

XMTPMBX *
xmtp_mbox_alloc NOARGLIST
{
	XMTPMBX		*mbxP;

    mbxP = (XMTPMBX *)emalloc( "xmtp_mbox_alloc", "XMTPMBX struct",
			      sizeof(XMTPMBX) );

    if ( mbxP != NULL ) {
	mbxP->mb_pathP = NULL;
	mbxP->mb_adlF = FALSE;
	mbxP->mb_atX = -1;
	mbxP->mb_len = 0;
    }

    return ( mbxP );
}
/*

*//* xmtp_mbox_dealloc( mbxP )

	Free up a previously allocated/used XMTPMBX

Accepts :

	mbxP		Ptr to XMTPMBX to free

Returns :

	<nothing>

Notes :


*/

void
xmtp_mbox_dealloc ARGLIST( (mbxP) )
    FARG( XMTPMBX	*mbxP )		/* Mailbox to free */
{
    if ( mbxP->mb_pathP != NULL )
	dealloc( mbxP->mb_pathP );
    dealloc( mbxP );
}
/*

*//* xmtp_parse_adl( cmdPP, bufP, bufL )

	Parse a source route (at-domain-list) from a [command] buffer.

Accepts :

	cmdPP		Ptr to variable that points to the command string
	bufP		Where to put the string
	bufL		How large the buffer is

Returns :

	<value>		XPRSLT code

Notes :

	The source route is extracted from the string at *cmdPP.  For
	an OK return, *cmdPP is updated to point to the next byte after
	the source route.

	Input must start with '@' to be considered a source route.
	Caller should probably have already checked this.

	Leading/trailing spaces are NOT skipped by this function.

	On OK return, destination buffer is NUL-terminated.

*/

XPRSLT
xmtp_parse_adl ARGLIST( (cmdPP, bufP, bufL) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
   NFARG( UBYTE		*bufP )		/* Where to put the source route */
    FARG( int		bufL )		/* Size of buffer */
{
	XPRSLT		result;
	int		len;
	UBYTE		*cmdP;		/* Ptr to input string */

    cmdP = *cmdPP;

    /* Source route from here is a list of domain names. */
    for ( ; ; ) {
	/* Each domain in the list must start with a '@' */
	if ( *cmdP++ != '@' )
	    return ( XPR_SYNTAX );

	/* Parse domain */
	result = xmtp_parse_domain( &cmdP, bufP, bufL );
	if ( result != XPR_OK )
	    return ( result );

	/* Update buffer pointer/length */
	len = strlen( (char *)bufP );
	bufP += len;
	bufL -= len;

	/* Check for another in the list */
	if ( *cmdP != ',' )
	    break;

	/* Iterate to get next domain */
	++cmdP;
    }

    /* All done.. */
    *cmdPP = cmdP;
    return ( XPR_OK );
}
/*

*//* xmtp_parse_atom( cmdPP, bufP, bufL )

	Parse RFC2821/2822 -style atom from a [command] buffer.

Accepts :

	cmdPP		Ptr to variable that points to the source string
	bufP		Where to put the atom
	bufL		How large the buffer is

Returns :

	<value>		XPRSLT code

Notes :

	The atom is extracted from the string at *cmdPP.  For an OK
	return, *cmdPP is updated to point to the next byte after the
	atom.

	Leading/trailing spaces are NOT skipped by this function.

	On OK return, destination buffer is NUL-terminated.

	If the atom is empty, XPR_SYNTAX error is returned.

*/

XPRSLT
xmtp_parse_atom ARGLIST( (cmdPP, bufP, bufL) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
   NFARG( UBYTE		*bufP )		/* Where to put the source route */
    FARG( int		bufL )		/* Size of buffer */
{
	UBYTE		ch;
	UBYTE		*cmdP;		/* Ptr to input string */

    cmdP = *cmdPP;

    for ( ; ( ch = *cmdP ) != NUL ; ++cmdP ) {

	/* Bytes may be quoted by backslash.  However, per RFC2821 4.1.2,
	   the quoted character still must not be a CTL or a non-ASCII
	   char.
	*/
	if ( ch == '\\' ) {
	    ch = cmdP[1];
	    if ( ( Mchartbl[ch] & ( MC_CTL|MC_NASC ) ) != 0 )
		/* Unacceptable; atom ends prior to backslash. */
		break;
	    ++cmdP;
	}

	/* Otherwise done if non-atom byte */
	else if ( ( Mchartbl[ch] & ( MC_DIGIT|MC_ALPHA|MC_ATOM ) ) == 0 )
	    break;

	/* Ok char, atom-wise. */
	if ( bufL-- <= 1 )
	    return ( XPR_OFL );

	*bufP++ = ch;
    }

    *bufP = NUL;

    /* Thwart empty atom return */
    if ( cmdP == *cmdPP )
	return ( XPR_SYNTAX );

    *cmdPP = cmdP;
    return ( XPR_OK );
}
/*

*//* xmtp_parse_domain( cmdPP, bufP, bufL )

	Parse a domain name from a [command] buffer.

Accepts :

	cmdPP		Ptr to variable that points to the source string
	bufP		Where to put the domain name
	bufL		How large the buffer is

Returns :

	<value>		XPRSLT code

Notes :

	The domain name is extracted from the string at *cmdPP.  For an
	OK return, *cmdPP is updated to point to the next byte after the
	domain name.

	Leading/trailing spaces are NOT skipped by this function.

	On OK return, destination buffer is NUL-terminated.

*/

XPRSLT
xmtp_parse_domain ARGLIST( (cmdPP, bufP, bufL) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
   NFARG( UBYTE		*bufP )		/* Where to put the source route */
    FARG( int		bufL )		/* Size of buffer */
{
	BOOL		firstF;		/* If first char in domain-part */
	UBYTE		ch;
	int		bufX;
	UBYTE		*cmdP;		/* Ptr to input string */

    cmdP = *cmdPP;

    if ( *cmdP == '[' ) {
	/* Collect bytes up to the closing ']' */
	do {
	    if ( ( ch = *cmdP++ ) == NUL )
		return ( XPR_SYNTAX );
	    if ( bufL-- < 1 )
		return ( XPR_OFL );
	    *bufP++ = ch;
	} while ( ch != ']' );
    }
    else {
	/* Loop getting domain parts separated by dots */
	for ( firstF = TRUE, bufX = 0; (ch = *cmdP) != NUL; ++cmdP ) {
	    if ( firstF ) {
TODO	/* probably want to have some control over accepting "bad" chars */
		if ( ! ( isalpha(ch) ||
#if 1
			 (ch == '_') ||
#endif
			 isdigit(ch) ) )
		    break;
		else
		    firstF = FALSE;
	    }
	    else {
		if ( ! ( isalpha(ch) ||
			 isdigit(ch) ||
#if 1
			 (ch == '_') ||
#endif
			 (ch == '-') ) ) {
		    /* End of domain part- make sure last char wasn't '-' */
		    if ( cmdP[-1] == '-' )
			return ( XPR_SYNTAX );

		    if ( ch != '.' )
			break;

		    /* '.' introduces a new subdomain part. */
		    firstF = TRUE;
		}
	    }

	    /* Take this char */
	    if ( --bufL < 1 )		/* Leave room for trailing NUL */
		return ( XPR_OFL );
	    *bufP++ = ch;
	}

	if ( firstF )
	    return ( XPR_SYNTAX );	/* Null domain part */
    }

    *bufP = NUL;
    *cmdPP = cmdP;

    return ( XPR_OK );
}
/*

*//* xmtp_parse_localpart( cmdPP, bufP, bufL )

	Parse local part from  a [command] buffer.

Accepts :

	cmdPP		Ptr to variable that points to the source string
	bufP		Where to put the localpart
	bufL		How large the buffer is

Returns :

	<value>		XPRSLT code

Notes :

	The localpart is extracted from the string at *cmdPP.  For an OK
	return, *cmdPP is updated to point to the next byte after the
	localpart.

	Leading/trailing spaces are NOT skipped by this function.

	On OK return, destination buffer is NUL-terminated.

*/

XPRSLT
xmtp_parse_localpart ARGLIST( (cmdPP, bufP, bufL) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
   NFARG( UBYTE		*bufP )		/* Where to put the source route */
    FARG( int		bufL )		/* Size of buffer */
{
	XPRSLT		result;
	int		len;
	UBYTE		*cmdP;

    cmdP = *cmdPP;

    /* May be a quoted string */
    if ( *cmdP == '"' )
	return ( xmtp_parse_quoted_string( cmdPP, bufP, bufL ) );

    /* Nope, it must be a dot-string (atom(s) separated by dots) */
    for ( ; ; ) {
	result = xmtp_parse_atom( &cmdP, bufP, bufL );
	if ( result != XPR_OK )
	    return ( result );
	if ( *cmdP != '.' )
	    break;

	/* Add the dot and another atom */
	len = strlen( (char *)bufP );
	bufP += len;
	bufL -= len;
	if ( --bufL < 1 )		/* Need room for .<NUL> */
	    return ( XPR_OFL );
	*bufP++ = *cmdP++;
    }

    /* bufP is already NUL-terminated */

    *cmdPP = cmdP;

    return ( XPR_OK );
}
/*

*//* xmtp_parse_mailbox( cmdPP, mbxPP )

	Parse a mailbox string into a mailbox struct

Accepts :

	cmdPP		Ptr to variable that points the source string
	mbxPP		Ptr to XMTPMBX pointer where an XMTPMBX struct
			  is returned.

Returns :

	<value>		XPRSLT code

Notes :

	The mailbox is extracted from the string at *cmdPP.  For an OK
	return, *cmdPP is updated to point to the next byte after the
	mailbox.

	Leading/trailing spaces are NOT skipped by this function.

	Any source-route (at-domain-list) in the mailbox is skipped and
	ignored.

	On OK return, destination buffer is NUL-terminated.

	*mbxPP will always be returned a valid XMTPMBX pointer or NULL.

*/

XPRSLT
xmtp_parse_mailbox ARGLIST( (cmdPP, mbxPP) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
    FARG( XMTPMBX	**mbxPP )	/* XMTP mailbox info pointer */
{
	BOOL		adlF;		/* If we have a source route */
	XPRSLT		result;
	int		len;
	int		bufL;
	int		atX;		/* Index of '@' in buffer */
	UBYTE		*cmdP;		/* Command buffer */
	UBYTE		*bufP;
	XMTPMBX		*mbxP;		/* an XMTPMBX that we make */
	UBYTE		adlbuf[XMTP_PATH_MAX];
	char		path[XMTP_PATH_MAX ];  /* A path buffer */

    cmdP = *cmdPP;
    *mbxPP = NULL;

    /* If it starts with '@', start out with any source-route. */
    if ( *cmdP == '@' ) {
	result = xmtp_parse_adl( &cmdP, &adlbuf[0], sizeof(adlbuf) );
	if ( result != XPR_OK )
	    return ( result );

	/* Source route must end with ':' */
	if ( *cmdP++ != ':' )
	    return ( XPR_SYNTAX );
	adlF = TRUE;
    }
    else
	adlF = FALSE;


    /* Set up to parse the actual mailbox */
    bufP = (UBYTE *)&path[0];
    bufL = sizeof( path );

    result = xmtp_parse_localpart( &cmdP, bufP, bufL );
    if ( result != XPR_OK )
	return ( result );

    if ( *cmdP++ != '@' )		/* There must be an @ */
	return ( XPR_SYNTAX );

    atX = len = strlen( (char *)bufP );
    bufP += len;
    bufL -= len;

    /* Store the '@' - leave room for at least one more char + NUL */
    if ( --bufL < 2 )
	return ( XPR_OFL );
    *bufP++ = '@';

    /* Collect the domain name */
    result = xmtp_parse_domain( &cmdP, bufP, bufL );
    if ( result != XPR_OK )
	return ( result );

    /* OK now. */
    if ( ( mbxP = xmtp_mbox_alloc() ) == NULL )
	return ( XPR_ERR );
    mbxP->mb_pathP = (UBYTE *)newstr( "xmtp_parse_mailbox", "mb_pathP", &path[0] );
    if ( mbxP->mb_pathP == NULL )
	return ( XPR_ERR );

    mbxP->mb_len = strlen( (char *)mbxP->mb_pathP );
    mbxP->mb_adlF = adlF;
    mbxP->mb_atX = atX;

    *mbxPP = mbxP;
    *cmdPP = cmdP;

    return ( XPR_OK );
}
/*

*//* xmtp_parse_path( cmdPP, mbxPP )

	Parse a path string into a mailbox struct

Accepts :

	cmdPP		Ptr to variable that points the source string
	mbxPP		Ptr to XMTPMBX pointer variable

Returns :

	<value>		XPRSLT code

Notes :

	For our purposes, a path string is just a mailbox enclosed
	in anglebrackets, except that the string may be one of
	these special things:

		<>		null string

	The return is as per xmtp_parse_mailbox.

	The anglebrackets are NOT returned.

*/

XPRSLT
xmtp_parse_path ARGLIST( (cmdPP, mbxPP) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
    FARG( XMTPMBX	**mbxPP )	/* XMTP mailbox ptr */
{
	XPRSLT		result;
	UBYTE		*cmdP;
	XMTPMBX		*mbxP;

    cmdP = *cmdPP;

    if ( *cmdP++ != '<' ) {
	*mbxPP = NULL;
	return ( XPR_SYNTAX );
    }

    /* Check for special cases */
    if ( *cmdP == '>' ) {
	if ( ( mbxP = xmtp_mbox_alloc() ) == NULL )
	    result = XPR_ERR;
	else {
	    /* Initialize a null buffer */
	    mbxP->mb_pathP = (UBYTE *)newstr( "xmtp_parse_path", "null path", "" );
	    if ( mbxP->mb_pathP == NULL ) {
		mbxP = NULL;
		result = XPR_ERR;
	    }
	    else {
		++cmdP;
		result = XPR_OK;
	    }
	}
    }
    else {
	/* Just get a mailbox */
	result = xmtp_parse_mailbox( &cmdP, &mbxP );
	if ( result == XPR_OK ) {
	    if ( *cmdP++ != '>' ) {	/* Path must end with '>' */
		xmtp_mbox_dealloc( mbxP );
		result = XPR_SYNTAX;
	    }
	}
    }

    *cmdPP = cmdP;
    *mbxPP = mbxP;

    return ( result );
}
/*

*//* xmtp_parse_quoted_string( cmdPP, bufP, bufL )

	Parse a quoted string from a [command] buffer.

Accepts :

	cmdPP		Ptr to variable that points to the source string
	bufP		Where to put the domain name
	bufL		How large the buffer is

Returns :

	<value>		XPRSLT code

Notes :

	The quoted string is extracted from the string at *cmdPP.  For an
	OK return, *cmdPP is updated to point to the next byte after the
	terminating quote.

	No translations are done; the quoted string is returned in
	its entirety (minus the quotes)

	Note that this is an RFC2821-style quoted string extraction,
	which is different from RFC2822 quoted-string in that
	*no* control characters are acceptable (RFC2822 says that
	NO_WS_CTL are OK).  This could probably be generalized to allow
	either, see the notes in code.

	Leading/trailing spaces are NOT skipped by this function.

	An empty quoted string will return XPR_SYNTAX error.

	On OK return, destination buffer is NUL-terminated.

*/

XPRSLT
xmtp_parse_quoted_string ARGLIST( (cmdPP, bufP, bufL) )
   NFARG( UBYTE		**cmdPP )	/* Ptr to command pointer */
   NFARG( UBYTE		*bufP )		/* Where to put the source route */
    FARG( int		bufL )		/* Size of buffer */
{
	UBYTE		ch;
	UBYTE		*orgbufP;	/* bufP on entry */
	UBYTE		*cmdP;

    orgbufP = bufP;
    cmdP = *cmdPP;

    /* Skip the first quote, which must be present */
    if ( *cmdP++ != '"' )
	return ( XPR_SYNTAX );

    /* Collect the string */
    for ( ; ; ) {
	ch = *cmdP++;
	if ( ch == NUL )		/* End of string without quote */
	    return ( XPR_SYNTAX );

	/* Byte may be quoted by backslash */
	if ( ch == '\\' )
	    ch = *cmdP++;

	/* Otherwise check for closing quote */
	else if ( ch == '"' )
	    break;

	/* Byte must still be valid for a quoted string.  RFC2821 4.1.2
	   and qcontent from RFC2822 3.2.5.

	   Note that double-quote and backslash will pass this test, while
	   they are technically illegal unless quoted with a backslash, and
	   that's the only way they will be here.
	*/
#if 0		/* This would be RFC2822 (3.2.5) qcontent */
	if ( ( ( Mchartbl[ch] & (MC_CTL|MC_NWSCTL) ) == MC_CTL ) ||
	     ( ( Mchartbl[ch] & MC_NASC ) != 0 ) )

#else		/* whereas this is RFC2821 (4.1.2) quoted string */
	if ( ( ( Mchartbl[ch] & (MC_CTL) ) != 0 ) ||
	     ( ( Mchartbl[ch] & MC_NASC ) != 0 ) )
#endif
	    return ( XPR_SYNTAX );
	
	if ( --bufL < 2 )		/* Leave room for byte and NUL */
	    return ( XPR_OFL );
	*bufP++ = ch;
    }

    /* If we get here, we have the closing quote and the complete string. */
    *bufP = NUL;

    /* Trap empty string */
    if ( bufP == orgbufP )
	return ( XPR_SYNTAX );

    *cmdPP = cmdP;
    return ( XPR_OK );
}
/*

*//* xmtp_reply(xmtP, nowF, rcode, ecodeP, fmtP, args...)

	Send an xMTP reply

Accepts :

	xmtP		xMTP dialog control block
	nowF		TRUE if output should be sent immediately (flushed)
	rcode		xMTP reply code
	ecodeP		extended xMTP reply code string, or NULL if none
	fmtP		Format string for output
	args		Any arguments to go with the format.

Returns :

	<value>		TRUE if OK
			FALSE if some error.

Notes :

	Handles multi-line formats, albiet in a weird way.

	Naked "\n" line endings are translated to "\r\n"

	rcode may be passed as a negative number to mean that more
	lines are coming (and we should end on a continuation line).
	The absolute value of rcode is used.

	Extended result codes are suppressed if the mt_ecodesF
	XMTP element is FALSE.

	In order to handle the variable arguments and the multi-line
	output, this function buffers the reply string into an allocated
	buffer, which is may grow as needed.  However, there's a hard
	limit on how far this buffer will be expanded.  If for any
	reason there's a failure getting the buffer, only the
	numeric result(s) will be emitted.

	Note also that this function uses malloc/realloc rather
	than mml's emalloc and realloc- we can't have those functions'
	error traps happening.

	The "nowF" controls whether the reply is flushed to the client,
	e.g. in order to support pipelineing.  Caller must take care
	to observe proper pipelineing when setting this FALSE.

*/

BOOL
xmtp_reply ARGLIST( (xmtP, nowF, rcode, ecodeP, fmtP, va_decl) )
   NFARG( XMTP		*xmtP )		/* control block */
   NFARG( BOOL		nowF )		/* Whether to send reply now */
   NFARG( int		rcode )		/* Reply code */
   NFARG( char		*ecodeP )	/* Extended reply code */
   NFARG( char		*fmtP )		/* Format string */
    FARG( ... )				/* stdc variable arguments */
{
	va_list		ap;		/* varargs argument pointer */

	BOOL		okF;
	BOOL		moreF;		/* If more coming later */
	int		len, plen;
	char		*strP, *estrP;
	char		prefix[20];

    /* Start varargs */
    va_start( ap, fmtP );

    /* If first pass, do first allocation of buffer */
    if ( ReplyL == 0 ) {
	if ( ( ReplyP = (UBYTE *)malloc( XMTP_REPCHUNK ) ) != NULL )
	    ReplyL = XMTP_REPCHUNK;
	else
	    ReplyL = -1;
    }

    /* Negative rcode means this should end with continuation line */
    if ( rcode < 0 ) {
	moreF = TRUE;
	nowF = FALSE;
	rcode = -rcode;
    }
    else
	moreF = FALSE;

    /* Format the string */
    for ( ; ; ) {
	if ( ReplyL < 0 ) {
	    /* Some prior allocation failure; don't give any string. */
	    strP = "";
	    break;
	}

	if ( ( len = vsnprintf( (char *)ReplyP, ReplyL, fmtP, ap ) ) < ReplyL ) {
	    /* Buffer is fine; use it. */
	    strP = (char *)ReplyP;
	    break;
	}

	len = ReplyL + XMTP_REPCHUNK;
	if ( len > XMTP_REPMAX ) {
	    /* Buffer is maxed out, don't give any string. */
	    strP = "";
	    break;
	}

	if ( ( ReplyP = (UBYTE *)realloc( ReplyP, len ) ) == NULL )
	    ReplyL = -1;		/* Remember the failure. */
	else
	    ReplyL = len;
    }

    /* Prepare the result code prefix */
    if ( xmtP->mt_ecodesF && ( ecodeP != NULL ) )
	plen = snprintf( &prefix[0], sizeof(prefix), "%d %s ",
			rcode, ecodeP );
    else
	plen = snprintf( &prefix[0], sizeof(prefix), "%d ", rcode );


    /* Emit potentially multi-line reply */
    for ( ; ; ) {
	if ( ( estrP = strchr( strP, '\n' ) ) == NULL )
	    estrP = strP + strlen( strP );

	/* Ignore any preceding CR */
	len = estrP - strP;
	if ( ( len > 0 ) && ( strP[ len -1 ] == '\r' ) )
	    --len;

	/* Step to new multiline beginning */
	if ( *estrP != NUL )
	    ++estrP;

	/* Make prefix end with either "-" if there's another line,
	   or " " if there isn't.
	*/
	if ( moreF || ( *estrP != NUL ) )
	    prefix[ plen -1 ] = '-';
	else
	    prefix[ plen -1 ] = ' ';

	/* Emit this line */
	okF = bio_putn( &xmtP->mt_out, (UBYTE *)&prefix[0], plen );
	okF = okF && bio_putn( &xmtP->mt_out, (UBYTE *)strP, len );
	okF = okF && bio_putn( &xmtP->mt_out, (UBYTE *)"\r\n", 2 );

	if ( ( *estrP == NUL ) || !okF )
	    break;			/* Done */
	strP = estrP;
    }

    /* Force reply if desired */
    if ( nowF )
	okF = okF && bio_drain( &xmtP->mt_out );

    va_end( ap );

    return ( okF );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* xmtpdss_getb( inP, bioP )

	Get a byte from a dot-stuffed input

Accepts :

	inP		Ptr to input stack
	bioP		Ptr to BIO for input

Returns :

	<value>		>=0: the byte input
			< 0: an ending code (here, EOF)

Notes :

	This takes input from a BIO input structure, and processes it
	  (doing dot-unstuffing).

	This function is intended to be called by an MML_IN
	  stream handler for character input.  Upon end of input,
	  it will set up the inP result information.  On a legitimate
	  dot-stuffed ending, the input result will be IOR_NONE.
	  For some other ending, the result information will be
	  set by underlying input handlers.

*/

static int
xmtpdss_getb ARGLIST( (inP, bioP) )
   NFARG( MML_IN	*inP )		/* The input stack */
    FARG( BIO		*bioP )		/* The BIO for input */
{
	int		ch;

    /* Get a byte from the input and check for period at beginning of line.
    */
    ch = bio_getb( bioP );
    if ( inP->in_eolF && ( ch == '.' ) ) {

	/* Gotta check the following character sequence.  If it's a line end,
	   we have to return EOF.  Otherwise we just return the next byte
	   (ignoring the dot).
	*/
	ch = bio_getb( bioP );
	if ( ch == '\r' ) {
	    /* CR should be followed by LF */
	    ch = bio_getb( bioP );
	    if ( ch != '\n' ) {
		/* CR without adjacent following LF.  Return the CR. */
		bio_ungetb( bioP, ch );
		ch = '\r';
	    }
	    else {
		/* .<CR><LF> -- be done. */
		inP->in_result = IOR_NONE;
		return ( EOF );
	    }
	}
	else if ( ch == '\n' ) {
	    /* Naked LF is grudgingly taken */
TODO	    /* This behaviour is frowned on in RFC2821 but is, unfortunately,
	       needed in the real world.  This choice should probably be
	       given to the admin though, either compile-time or run-time.
	       Ditto the treatment of naked CR (above).
	    */
	    inP->in_result = IOR_NONE;
	    return ( EOF );
	}
    }

    if ( ch < 0 ) {
	/* Here the input has simply stopped- this is an invalid
	   end to the input, and we must set up an appropriate
	   error condition in the input stack.  
	*/
	inP->in_result = bioP->bio_result;
	inP->in_errno = bioP->bio_errno;
	if ( inP->in_result == IOR_NONE ) {
	    warning( "xmtpdss_getb got end of data, but IOR_NONE.  Bug!" );
	    inP->in_result = IOR_FAIL;
	}
	ch = EOF;
    }

    return ( ch );
}
