/* mcdb-avl.c -- In-memory implementation of mcdb module, using AVL trees.

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This module provides the low-level access to data used by mvmcas.

This particular implementation is a first pass, quick & dirty module
that keeps its runtime info in memory while loading initial data from
text files and periodically checkpointing its data to those text files.

I hope to play with this module before moving along to something
more scalable.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <pthread.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-fmt.h>
#include <mml/mml-avl.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "mvmcas-misc.h"
#include "mcdb.h"
#include "timer.h"

/* Local definitions */

    /* How often to prune each mclient tree, in seconds */
#define	MCLIENT_PRUNE_INTERVAL	(2L * 3600)

    /* A short prune interval (e.g. when one pruning stopped short) */
#define	MCLIENT_PRUNE_INTERVAL_SHORT	(10 * 60)

    /* Max nodes to prune at any one pass */
#define	MCLIENT_PRUNE_MAX	10000


    /* Info for a mail client tree */
typedef struct {
    AVLTREE	*mct_treeP;	/* The tree */
    int		mct_writeC;	/* Number of writes scheduled */
    long	mct_purgeC;	/* Number of nodes purged */
    int		mct_pruneC;	/* Number of times pruned */
    time_t	mct_nptime;	/* Next prune/purge time */
    char	*mct_tnameP;	/* Name for tree type, e.g. "auth";
				   used to make filename for storage.
				*/
    pthread_mutex_t mct_lock;	/* General operational lock */
}  MCTREE;


    /* A node in the mail client tree */
typedef struct {
    AVLNODE	mcn_avl;	/* AVL node header */
    MCTREE	*mcn_mctreeP;	/* Ptr back to the owning MCTREE */
    pthread_mutex_t mcn_lock;	/* A lock for this client */
    MCLIENT	mcn_mclient;	/* Mail client info */
    BOOL	mcn_updateF;	/* If update needed */
    BOOL	mcn_accessF;	/* If somebody is accessing this */
}  MCNODE;


    /* Information used by tree walker when writing a mail client tree */
typedef struct {
    FILE	*mcw_fP;	/* The file to write to */
    BOOL	mcw_okF;	/* If OK */
}  MCWRITE;


/* External data referenced */

extern	int	Debug;		/* Debugging level */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* AVL support functions for mail client AVL trees */
    /* These are for trees with keys that are IPv4 addresses */
static	int	avl_mc_ip4_cmpkey();
static	AVLNODE	*avl_mc_ip4_mknode();
static	int	avl_mc_ip4_rmnode();

    /* These are for trees with keys that are simple strings */
static	int	avl_mc_str_cmpkey();
static	AVLNODE	*avl_mc_str_mknode();
static	int	avl_mc_str_rmnode();

    /* Read mail client data from a directory */
static	BOOL	mcdb_read_dir PROTO( (char *dnameP, BOOL newF) );

    /* Read mail client data from an open file */
static	BOOL	mcdb_read_fp PROTO( (FILE *fP, char *fnameP) );

    /* Read mail client data from a named file */
static	BOOL	mcdb_read_file PROTO( (char *fnameP) );

    /* Read mail client data from MML_IN stread */
static	BOOL	mcdb_read_inp PROTO( (MML_IN *inP) );

    /* Read from input until an "end" keyword is seen */
static	BOOL	mcdb_read_inp_end PROTO( (MML_IN *inP) );

    /* Read an integer from input */
static	BOOL	mcdb_read_inp_int PROTO( (MML_IN *inP, long *longP) );

    /* Read an unsigned integer from input */
static	BOOL	mcdb_read_inp_uint PROTO( (MML_IN *inP, ULONG *ulongP) );

    /* Read a single MCLIENT definition from mail client data input stream */
static	BOOL	mcdb_read_inp_mclient PROTO( (MML_IN *inP) );

    /* Read an MCNAME from mail client data input stream */
static	BOOL	mcdb_read_inp_mcname PROTO( (MML_IN *inP, MCNAME_S *mcnsP) );

    /* Read a token from input stream for mcdb_read stuff */
static	BOOL	mcdb_read_token PROTO( (MML_IN *inP, UBYTE *bufP, int bufL) );

    /* Inherit info from a mclient prototype */
static	void	mclient_proto_inherit PROTO( (MCLIENT *mcP, MCLIENT *protoP) );

    /* Prune an mclient tree */
static	void	mclient_prune PROTO( (void *argP) );

    /* Called for each node by avlwalk on behalf of mclient_prune */
static	int	mclient_prune_node();

    /* Write the mctrees to a file */
static	void	mctree_write PROTO( (MCTREE *mctreeP) );

    /* Schedule a write of an mctree */
static	void	mctree_write_sched PROTO( (MCTREE *mctreeP, int mdelay) );

    /* Write one node of the mctree */
static	int	mctree_write_node();


/* Private data */

    /* When reading mclient files, any mclient not heard from since this
       date may be ignored.
    */
static	time_t	Mclient_oldest_wanted;

    /* When pruning, any mclient with lctime older than this may be
       pruned.
    */
static	time_t	Mclient_prune_before;

    /* The current time when mail client pruning is being done */
static	time_t	Mclient_prune_time;

    /* How often to prune each mclient tree, in secondes */
static	int	Mclient_prune_interval = MCLIENT_PRUNE_INTERVAL;

    /* List to hold mclient nodes that are being pruned */
static	LL_HEAD	Mclient_prune_list;

    /* Number of entries on the prune list */
static	int	Mclient_prune_listC;

    /* IP masklen jump table, controlling how we look for IP parents.
       Each entry index corresponds to the masklen that we have; the
       entry value is the masklen to check for parent.
    */

static	BYTE	Ipmjtbl[] = {
	-1,			/*  0: no parent */
	 0,			/*  1: parent:  0 */
	 0,			/*  2: parent:  0 */
	 0,			/*  3: parent:  0 */
	 0,			/*  4: parent:  0 */
	 0,			/*  5: parent:  0 */
	 0,			/*  6: parent:  0 */
	 0,			/*  7: parent:  0 */
	 0,			/*  8: parent:  0 */
	 8,			/*  9: parent:  8 */
	 8,			/* 10: parent:  8 */
	 8,			/* 11: parent:  8 */
	 8,			/* 12: parent:  8 */
	 8,			/* 13: parent:  8 */
	 8,			/* 14: parent:  8 */
	 8,			/* 15: parent:  8 */
	 8,			/* 16: parent:  8 */
	16,			/* 17: parent: 16 */
	17,			/* 18: parent: 17 */
	18,			/* 19: parent: 18 */
	19,			/* 20: parent: 19 */
	20,			/* 21: parent: 20 */
	21,			/* 22: parent: 21 */
	22,			/* 23: parent: 22 */
	23,			/* 24: parent: 23 */
	24,			/* 25: parent: 24 */
	25,			/* 26: parent: 25 */
	25,			/* 27: parent: 25 */
	25,			/* 28: parent: 25 */
	25,			/* 29: parent: 25 */
	25,			/* 30: parent: 25 */
	25,			/* 31: parent: 25 */
	25			/* 32: parent: 25 */
};



    /* AVL trees for the various mail client types */
static	MCTREE	Mctree_auth;		/* Auth id tree */
static	MCTREE	Mctree_ip;		/* IPv4 tree */

   /* A list of the MCTREEs */
static	MCTREE	*Mctrees[] = {
	&Mctree_auth,
	&Mctree_ip,
	NULL
};

/*

*//* mcdb_init()

	Initialize the mcdb module.

Accepts :


Returns :

	<value>		TRUE if successful;
			FALSE if not.

Notes :


*/

BOOL
mcdb_init NOARGLIST
{
	int		i;
	time_t		timenow;
	char		fname[ PATH_MAX +1];  /* filename */

    /* Init the tree for auth IDs */
    Mctree_auth.mct_treeP = avlnew( NULL,
			   avl_mc_str_cmpkey,
			   avl_mc_str_mknode,
			   avl_mc_str_rmnode );
    if ( Mctree_auth.mct_treeP == NULL )
	return ( FALSE );
    Mctree_auth.mct_treeP->t_infoP = (void *)&Mctree_auth;
    Mctree_auth.mct_writeC = 0;
    Mctree_auth.mct_purgeC = 0;
    Mctree_auth.mct_pruneC = 0;
    Mctree_auth.mct_nptime = 0;
    Mctree_auth.mct_tnameP = "auth";
    if ( pthread_mutex_init( &Mctree_auth.mct_lock, NULL ) != 0 )
	warning( "mcdb_init: can't init auth tree lock" );

    /* Init the tree for IP addresses */
    Mctree_ip.mct_treeP = avlnew( NULL,
			   avl_mc_ip4_cmpkey,
			   avl_mc_ip4_mknode,
			   avl_mc_ip4_rmnode );
    if ( Mctree_ip.mct_treeP == NULL )
	return ( FALSE );
    Mctree_ip.mct_treeP->t_infoP = (void *)&Mctree_ip;
    Mctree_ip.mct_writeC = 0;
    Mctree_ip.mct_purgeC = 0;
    Mctree_ip.mct_pruneC = 0;
    Mctree_ip.mct_nptime = 0;
    Mctree_ip.mct_tnameP = "ip";
    if ( pthread_mutex_init( &Mctree_ip.mct_lock, NULL ) != 0 )
	warning( "mcdb_init: can't init auth tree lock" );

    /* Scan and read the user-supplied data directory */
    mcdb_read_dir( MVMCAS_UDBDIR, FALSE );
    mcdb_read_dir( MVMCAS_UDBDIR, TRUE );

    /* Read any cached files */
    for ( i = 0; Mctrees[i] != NULL; ++i ) {
	if ( snprintf( &fname[0], sizeof(fname), "%s/mctree_%s",
		  MVMCAS_RDBDIR,
		  Mctrees[i]->mct_tnameP ) >= sizeof(fname) ) {
	    warning( "mcdb_init: file name too large for type %s",
		      Mctrees[i]->mct_tnameP );
	}
	else {
	    mcdb_read_file( &fname[0] );
	}
    }


    /* Initialize the mclient prune list */
    ll_init( &Mclient_prune_list );

    /* Shorten the mclient prune interval if debugging */
    if ( Debug > 1 ) {
	Mclient_prune_interval = 60;
    }

    /* Schedule the first pruning */
    time( &timenow );
    if ( !timer_add( timenow + MCLIENT_PRUNE_INTERVAL_SHORT,
			mclient_prune, Mctrees[0] ) ) {
	warning( "mcdb_init: timer_add failed for mclient_prune." );
	return ( FALSE );
    }
    
    return ( TRUE );
}
/*

*//* mcdb_init_again()

	Initialize mcdb module again, e.g. run-time check.

Accepts :


Returns :

	<value>		TRUE if successful;
			FALSE if not.

Notes :

	Tells the mcdb module to do any processing of new information.
	What that might be is db-specific.

*/

BOOL
mcdb_init_again NOARGLIST
{
    return ( TRUE );
}
/*

*//* mcdb_mclient_access( mcnameP )

	Get client info for specified mail client type

Accepts :

	mcnameP		A MCNAME mail client name

Returns :

	<value>		Ptr to MCLIENT block if successful
			NULL if not (some error)

Notes :


*/

MCLIENT *
mcdb_mclient_access ARGLIST( (mcnameP) )
    FARG( MCNAME	*mcnameP )	/* Mail client name */
{
	MCLIENT		*mcP;

    switch ( mcnameP->mn_type ) {
	case MCT_AUTH:
	    mcP = mcdb_mclient_access_auth( (UBYTE *)mcnameP->mn_nameP );
	    break;

	case MCT_IP:
	    mcP = mcdb_mclient_access_ip( (MCNAME_IP4 *)mcnameP->mn_nameP );
	    break;

	case MCT_IPv6:
	    warning( "mcdb_mclient_access: IPv6 type not supported." );
	    mcP = NULL;
	    break;

	default:
	    warning( "mcdb_mclient_access: unknown mctype %d",
		     mcnameP->mn_type );
	    mcP = NULL;
	    break;
    }

    return ( mcP );
}
/*

*//* mcdb_mclient_access_auth( authnameP )

	Get mail client info for AUTH mail client type

Accepts :

	authnameP	Ptr to auth name string

Returns :

	<value>		Ptr to MCLIENT block if successful
			NULL if not (some error)

Notes :


*/

MCLIENT *
mcdb_mclient_access_auth ARGLIST( (authnameP) )
    FARG( UBYTE		*authnameP )	/* Auth name */
{
	MCNODE		*mcnodeP;	/* AVL node ptr */

    /* Try to find the entry, keeping the tree locked until done */
    mcnodeP = (MCNODE *)avlfind_nounlock( Mctree_auth.mct_treeP, (void *)authnameP );
    if ( mcnodeP == NULL ) {
	/* Can't find it, need to create it. */
	avlinsert_nolock( Mctree_auth.mct_treeP, (void *)authnameP, (void *)NULL );
	mcnodeP = (MCNODE *)avlfind_nolock( Mctree_auth.mct_treeP, (void *)authnameP );
	if ( mcnodeP == NULL ) {
	    /* Error.. */
	    if ( pthread_mutex_unlock( &Mctree_auth.mct_treeP->t_lock ) != 0 )
		warning( "mcdb_client_access_auth: can't unlock tree in failure case" );
	    return ( NULL );
	}

	/* This node will need an update.. */
	mcnodeP->mcn_updateF = TRUE;

	/* Finish setting any client info */
	mcnodeP->mcn_mclient.mc_mcname.mn_type = MCT_AUTH;
    }

    /* Lock & return */
    if ( pthread_mutex_lock( &mcnodeP->mcn_lock ) != 0 )
	warning( "mcdb_mclient_access_auth: can't access lock" );
    if ( mcnodeP->mcn_accessF )
	warning( "mcdb_mclient_access_auth: already accessed.  BUG!" );

    mcnodeP->mcn_accessF = TRUE;

    if ( pthread_mutex_unlock( &Mctree_auth.mct_treeP->t_lock ) != 0 )
	warning( "mcdb_client_access_auth: can't unlock tree" );

    return ( &mcnodeP->mcn_mclient );
}
/*

*//* mcdb_mclient_access_ip( mcnameP )

	Get mail client info for IP mail client type

Accepts :

	ipnameP		Name of client, MCNAME_IP4 format.

Returns :

	<value>		Ptr to MCLIENT block if successful
			NULL if not (some error)

Notes :


*/

MCLIENT *
mcdb_mclient_access_ip ARGLIST( (ipnameP) )
    FARG( MCNAME_IP4	*ipnameP )	/* Mail client name (IP address) */
{
	int		i;
	MCNODE		*mcnodeP;	/* AVL node ptr */
	MCNODE		*pnodeP;	/* Parent node if any */
	MCLIENT		*mcP;		/* Handle to mail client part */
	MCNAME_IP4	pip;		/* Parent IP to check */

    /* Try to find the entry, keeping the tree locked until done */
    mcnodeP = (MCNODE *)avlfind_nounlock( Mctree_ip.mct_treeP, (void *)ipnameP );
    if ( mcnodeP == NULL ) {
	/* Node not found; see if we can find a higher level */
	pip = *ipnameP;
	pnodeP = NULL;
	for ( ; ; ) {
	    if ( ( i = Ipmjtbl[pip.nip_masklen] ) < 0 )
		break;

	    pip.nip_masklen = i;
	    pnodeP = (MCNODE *)avlfind_nolock( Mctree_ip.mct_treeP,
					       (void *)&pip );
	    if ( pnodeP != NULL )
		break;
	}

	if ( pnodeP != NULL ) {
	    /* Found a parent; use it directly if it's an aggregate. */
	    if ( ( pnodeP->mcn_mclient.mc_flags & MCF_AGG ) != 0 ) {
		mcnodeP = pnodeP;
		pnodeP = NULL;
	    }

	    /* If parent is not a prototype, just ignore it. */
	    else if ( ( pnodeP->mcn_mclient.mc_flags & MCF_PROTO ) == 0 )
		pnodeP = NULL;
	}

	if ( mcnodeP == NULL ) {
	    /* Don't have it yet, need to create it. */
	    avlinsert_nolock( Mctree_ip.mct_treeP, (void *)ipnameP, (void *)NULL );
	    mcnodeP = (MCNODE *)avlfind_nolock( Mctree_ip.mct_treeP,
					        (void *)ipnameP );
	    if ( mcnodeP == NULL ) {
		/* Error.. */
		if ( pthread_mutex_unlock( &Mctree_ip.mct_treeP->t_lock ) != 0 )
		    warning( "mcdb_client_access_ip: can't unlock tree in failure case" );
		return ( NULL );
	    }

	    /* This node will need an update.. */
	    mcnodeP->mcn_updateF = TRUE;

	    /* Finish setting any client info */
	    mcP = &mcnodeP->mcn_mclient;
	    mcP->mc_mcname.mn_type = MCT_IP;

	    /* Instantiate other inheritable stuff from parent */
	    if ( pnodeP != NULL )
		mclient_proto_inherit( mcP, &pnodeP->mcn_mclient );
	}
    }

    /* Lock & return */
    if ( pthread_mutex_lock( &mcnodeP->mcn_lock ) != 0 )
	warning( "mcdb_mclient_access_ip: can't access lock" );
    if ( mcnodeP->mcn_accessF )
	warning( "mcdb_mclient_access_ip: already accessed.  BUG!" );
    mcnodeP->mcn_accessF = TRUE;

    if ( pthread_mutex_unlock( &Mctree_ip.mct_treeP->t_lock ) != 0 )
	warning( "mcdb_client_access_ip: can't unlock tree" );

    return ( &mcnodeP->mcn_mclient );
}
/*

*//* mcdb_mclient_delete( mcP )

	Delete a previously accessed mail client

Accepts :

	mcP		Ptr to MCLIENT block to delete

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

BOOL
mcdb_mclient_delete ARGLIST( (mcP) )
    FARG( MCLIENT	*mcP )		/* Mail client to delete */
{
	BOOL		okF;
	MCTREE		*mctreeP;

    /* Get the ptr to the tree so we can schedule a write after */
    mctreeP = ((MCNODE *)mcP->mc_dbP)->mcn_mctreeP;

    /* Perform a deletion.  This will remove locks etc. */
    okF = ( avldelete( mctreeP->mct_treeP,
		      (void *)mcP->mc_mcname.mn_nameP ) >= 0 );

    /* Hard to know how to recover from a failure here.  However,
       schedule a write regardless.
    */

    mctree_write_sched( mctreeP, -1 );

    return ( okF );
}
/*

*//* mcdb_mclient_release( mcP, updateF )

	Release access to mail client block

Accepts :

	mcP		Ptr to MCLIENT block to release
	updateF		Whether to hard-update the data 

Returns :

	<nothing>

Notes :

	updateF merely says that the data has been modified and should
	be updated in permanent storage.  Caller may have changed the
	cached data anyway.

*/

void
mcdb_mclient_release ARGLIST( (mcP, updateF) )
   NFARG( MCLIENT	*mcP )		/* Mail client to release */
    FARG( BOOL		updateF )	/* Whether to update */
{
	MCNODE		*mcnodeP;	/* The AVL node */

    mcnodeP = (MCNODE *)mcP->mc_dbP;

    /* If this node is flagged for update, force the update flag */
    if ( mcnodeP->mcn_updateF ) {
	updateF = TRUE;
	mcnodeP->mcn_updateF = FALSE;
    }

    /* Schedule a write if needed */
    if ( updateF && ( mcnodeP->mcn_mctreeP->mct_writeC == 0 ) ) {
	mctree_write_sched( mcnodeP->mcn_mctreeP, 0 );
    }

    /* Unlock & return */
    mcnodeP->mcn_accessF = FALSE;
    if ( pthread_mutex_unlock( &mcnodeP->mcn_lock ) != 0 )
	warning( "mcdb_mclient_release: can't release lock." );
}
/*

*//* mcdb_mclient_stats( bsP )

	Print some brief stats into a buffer

Accepts :

	bsP		BSTR to print into

Returns :

	<nothing>

Notes :

	bsP is already prepped by caller (this function will likely
	  be appending to it)

	Status should be reasonably short and sweet, and formatted
	  as part of a STATS messsage.

*/

void
mcdb_mclient_stats ARGLIST( (bsP) )
    FARG( BSTR		*bsP )		/* BSTR to print into */
{
    bstr_printf( bsP, "  mc(C P): auth %ld %ld/%d, ip %ld %ld%/%d ;\n",
		Mctree_auth.mct_treeP->t_nodeC,
		Mctree_auth.mct_purgeC, Mctree_auth.mct_pruneC,
		Mctree_ip.mct_treeP->t_nodeC,
		Mctree_ip.mct_purgeC, Mctree_ip.mct_pruneC );
}
/*

*//* mcdb_mclient_update( mcP )

	Mark a mail client as changed

Accepts :

	mcP		Ptr to MCLIENT block to release

Returns :

	<nothing>

Notes :

	A call to this function indicates says that the mail client data has
	been modified and should be updated in permanent storage.  The
	update will be forced when the mail client is released.

*/

void
mcdb_mclient_update ARGLIST( (mcP) )
    FARG( MCLIENT	*mcP )		/* Mail client to release */
{
	MCNODE		*mcnodeP;	/* The AVL node */

    mcnodeP = (MCNODE *)mcP->mc_dbP;

    mcnodeP->mcn_updateF = TRUE;
}
/*

*//* mcdb_sync()

	Synchronize any cached memory and disk store

Accepts :



Returns :

	<value>		<nothing>

Notes :


*/

void
mcdb_sync NOARGLIST
{
TODO    /* Should probably only do this if there has been a change */
    mctree_write_sched( &Mctree_auth, -1 );
    mctree_write_sched( &Mctree_ip, -1 );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* avl_mc_ip4_cmpkey( keyP, nodeP )

	Compare mail client keys for avl tree routines,
	  where the client identifier is an IP address

Accepts :

	keyP		Ptr to the subject key (an MCNAME_IP4 value)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:


*/


static int
avl_mc_ip4_cmpkey( keyP, nodeP )
	MCNAME_IP4	*keyP;		/* Key to look for */
	MCNODE		*nodeP;		/* Node to compare against */
{
	MCNAME_IP4	*nkeyP;		/* The key in the node */

    nkeyP = (MCNAME_IP4 *)nodeP->mcn_mclient.mc_mcname.mn_nameP;

    return ( mcname_ip4_cmp( keyP, nkeyP ) );
}
/*

*//* avl_mc_ip4_mknode( treeP, keyP, dataP, enodeP )

	Make a mail client avl node on behalf of AVL tree routines,
	  where the client identifier is an IP address

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (an MCNAME_IP4 value)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
avl_mc_ip4_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	MCNAME_IP4	*keyP;		/* Key for the new node (IP) */
	void		*dataP;		/* Addr of new data (n/a) */
	MCNODE		*enodeP;	/* Existing node... */
{
	int		nodesize;	/* Size of node required */
	MCNODE		*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the name/key) */
    nodesize = sizeof(MCNODE) + sizeof( MCNAME_IP4 );
    nodeP = (MCNODE *)emalloc( "avl_mc_ip4_mknode", "avl node",
					 nodesize );

    /* Reset/clear the client struct */
    mclient_init( &nodeP->mcn_mclient );

    /* Setup the node struct */
    nodeP->mcn_mclient.mc_mcname.mn_nameP = (void *)&nodeP[1];
    *(MCNAME_IP4 *)nodeP->mcn_mclient.mc_mcname.mn_nameP = *keyP;

    /* Init the node mutex */
    if ( pthread_mutex_init( &nodeP->mcn_lock, NULL ) != 0 )
	warning( "avl_mc_ip4_mknode: can't init lock" );

    /* Provide a hook back from the node to the MCTREE */
    nodeP->mcn_mctreeP = (MCTREE *)treeP->t_infoP;

    nodeP->mcn_updateF = FALSE;
    nodeP->mcn_accessF = FALSE;

    /* Provide a hook back from the mail client structure to the AVL node */
    nodeP->mcn_mclient.mc_dbP = (void *)nodeP;

    return( (AVLNODE *)nodeP );
}
/*

*//* avl_mc_ip4_rmnode( treeP, nodeP )

	Delete a mail client avl node on behalf of AVL tree routines,
	  where the client identifier is an IP address

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :

	The caller may or may not have this node locked.  We'll
	attempt to unlock it and ignore errors.

	Just a dumb deletion.

*/

static int
avl_mc_ip4_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	MCNODE		*nodeP;
{
    /* Dumb pthreads won't let you destroy a lock you own, so
       we have to try unlocking it just in case.  But don't
       give an error.
    */
    if ( pthread_mutex_unlock( &nodeP->mcn_lock ) != 0 )
	// warning( "avl_mc_ip4_rmnode: can't release lock." );
	;
    if ( pthread_mutex_destroy( &nodeP->mcn_lock ) != 0 )
	warning( "avl_mc_ip4_rmnode: can't destroy lock" );

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* avl_mc_str_cmpkey( keyP, nodeP )

	Compare mail client keys for avl tree routines,
	  where the client identifier is a simple string.

Accepts :

	keyP		Ptr to the subject key (a string)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:


*/


static int
avl_mc_str_cmpkey( keyP, nodeP )
	char		*keyP;		/* Key to look for */
	MCNODE		*nodeP;		/* Node to compare against */
{
	int		cmpval;

    /* Do case insensitive string compare */
    cmpval = stricmp( keyP, (char *)nodeP->mcn_mclient.mc_mcname.mn_nameP );

    return ( cmpval );
}
/*

*//* avl_mc_str_mknode( treeP, keyP, dataP, enodeP )

	Make a mail client avl node on behalf of AVL tree routines,
	  where the client identifier is a simple string.

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (a string)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
avl_mc_str_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	char		*keyP;		/* Key for the new node (string) */
	void		*dataP;		/* Addr of new data (n/a) */
	MCNODE		*enodeP;	/* Existing node... */
{
	int		nodesize;	/* Size of node required */
	MCNODE		*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the name/key) */
    nodesize = sizeof(MCNODE) + strlen( keyP ) + 1;
    nodeP = (MCNODE *)emalloc( "avl_mc_str_mknode", "avl node",
					 nodesize );

    /* Reset/clear the client struct */
    mclient_init( &nodeP->mcn_mclient );

    /* Setup the node struct */
    nodeP->mcn_mclient.mc_mcname.mn_nameP = (void *)&nodeP[1];
    strcpy( (char *)nodeP->mcn_mclient.mc_mcname.mn_nameP, keyP );

    /* Init the node mutex */
    if ( pthread_mutex_init( &nodeP->mcn_lock, NULL ) != 0 )
	warning( "avl_mc_str_mknode: can't init lock" );

    /* Provide a hook back from the node to the MCTREE */
    nodeP->mcn_mctreeP = (MCTREE *)treeP->t_infoP;

    nodeP->mcn_updateF = FALSE;
    nodeP->mcn_accessF = FALSE;

    /* Provide a hook back from the mail client structure to the AVL node */
    nodeP->mcn_mclient.mc_dbP = (void *)nodeP;

    return( (AVLNODE *)nodeP );
}
/*

*//* avl_mc_str_rmnode( treeP, nodeP )

	Delete a mail client avl node on behalf of AVL tree routines,
	  where the client identifier is a simple string

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :

	Caller may or may not have the node locked; we'll attempt to
	unlock it here first since idiot pthreads won't let you destroy
	a lock if it's locked, even if you own that lock.

	Just a dumb deletion.

*/

static int
avl_mc_str_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	MCNODE		*nodeP;
{
    /* Dumb pthreads */
    if ( pthread_mutex_unlock( &nodeP->mcn_lock ) != 0 )
	// warning( "avl_mc_str_rmnode: can't release lock." );
	;
    if ( pthread_mutex_destroy( &nodeP->mcn_lock ) != 0 )
	warning( "avl_mc_str_rmnode: can't destroy lock" );

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* mcdb_read_dir( dnameP, newF )

	Read mail client data from a directory

Accepts :

	dnameP		Directory name
	newF		TRUE if processing new files only

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	newF controls what kinds of files to look for.
	if FALSE, only the persistent files are read.
	if TRUE, only the new update file are read.

	Files ending in .new are "new" temporary files.
	All other files are persistent files.

	Note Well: when processing new/update files, those files are
	  deleted after being read.

	In all cases, files with names ending in ".err" or ".inf"
	are skipped.

*/

static BOOL
mcdb_read_dir ARGLIST( (dnameP, newF) )
   NFARG( char		*dnameP )	/* Directory name */
    FARG( BOOL		newF )		/* If in new-file mode */
{
	BOOL		okF;
	BOOL		isnewF;
	int		namlen;
	DIR		*dirP;		/* Open directory ptr */
	struct dirent	*deP;		/* Directory entry ptr */
	char		fname[ PATH_MAX +1];  /* filename */
	char		fn2[ PATH_MAX +1];    /* another filename */

    if ( ( dirP = opendir( dnameP ) ) == NULL ) {
	warning( "mcdb_read_dir can't open \"%s\"", dnameP );
	return ( FALSE );
    }

    /* Scan the directory */
    for ( okF = TRUE; okF && ( deP = readdir( dirP ) ) != NULL;  ) {
	/* Skip blank or .* or non-regular file entries */
	namlen = strlen( &deP->d_name[0] );
	if ( ( namlen == 0 ) ||
	     ( deP->d_name[0] == '.' ) ||
	     ( ( deP->d_type & DT_REG ) == 0 ) )
	    continue;

	/* Skip files with names ending in '.err' or ".inf" */
	if ( ( namlen > 4 ) &&
	     ( ( strcmp( &deP->d_name[namlen -4], ".err" ) == 0 ) ||
	       ( strcmp( &deP->d_name[namlen -4], ".inf" ) == 0 ) ) )
	    continue;

	/* Isolate new or non-new files, as required */
	isnewF = ( namlen > 4 ) &&
		 ( strcmp( &deP->d_name[namlen -4], ".new" ) == 0 );

	/* Need a boolean XOR here */
	if ( ( newF && !isnewF ) ||
	     ( isnewF && !newF ) )
	    continue;

	/* Process this file */
	if ( snprintf( &fname[0], sizeof(fname), "%s/%s",
		       dnameP, &deP->d_name[0] ) >= sizeof(fname) ) {
	    warning( "mcdb_read_dir: file name too long \"%s/%s\"",
		       dnameP, &deP->d_name[0] );
	    okF = FALSE;
	}
	else {
	    okF = mcdb_read_file( &fname[0] );

	    if ( newF ) {
		/* Remove the file if OK; rename if not. */
		if ( !okF ) {
		    /* Guaranteed to fit */
		    strcpy( &fn2[0], &fname[0] );
		    strcpy( &fn2[namlen -4], ".err" );
		    link( &fname[0], &fn2[0] );
		}
		unlink( &fname[0] );
	    }
	}
    }

    closedir( dirP );			/* Ignore error */

    return ( okF );
}
/*

*//* mcdb_read_file( fnameP )

	Read from a named file

Accepts :

	fnameP		Name of the file to read from

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
mcdb_read_file ARGLIST( (fnameP) )
    FARG( char		*fnameP )	/* Filename */
{
	BOOL		okF;
	FILE		*fP;

    if ( ( fP = fopen( fnameP, "r" ) ) == NULL ) {
	warning( "mcdb_read_file: can't open \"%s\"", fnameP );
	okF = FALSE;
    }
    else {
	/* Set up the oldest-wanted threshold */
	time( &Mclient_oldest_wanted );
	Mclient_oldest_wanted -= MCLIENT_PRUNE_AGE;

	okF = mcdb_read_fp( fP, fnameP );
	fclose( fP );
    }

    return ( okF );
}
/*

*//* mcdb_read_fp( fP, fnameP )

	Read data from an already-opened file

Accepts :

	fP		Ptr to open file handle
	fnameP		Name of the file for messages, or NULL

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
mcdb_read_fp ARGLIST( (fP, fnameP) )
   NFARG( FILE		*fP )		/* FILE handle */
    FARG( char		*fnameP )	/* Filename */
{
	BOOL		okF;
	MML_IN		*inP;		/* Input stack */

    if ( fnameP == NULL )
	fnameP = "<unnamed>";

    /* Create input stack for the file */
    if ( ( inP = in_new() ) == NULL ) {
	warning( "mcdb_read_fp: can not create input stack for \"%s\"",
			fnameP );
	okF = FALSE;
    }
    else {
	/* Push the input file onto the stack */
	if ( !in_stream_push_file_nc( inP, fP, fnameP ) ) {
	    warning( "mcdb_read_fp: can't push input for \"%s\"", fnameP );
	    okF = FALSE;
	}
	else {
	    /* Process the input */
	    okF = mcdb_read_inp( inP );
	}

	in_destroy( inP );
    }

    return ( okF );
}
/*

*//* mcdb_read_inp( inP )

	Read from an MML_IN input stream

Accepts :

	inP		Ptr to input stream

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
mcdb_read_inp ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* The input stream */
{
	BOOL		okF;
	UBYTE		tok[50];	/* Primary token */

    /* Loop getting stuff */
    for ( okF = TRUE; okF ; ) {
	okF = mcdb_read_token( inP, &tok[0], sizeof(tok) );
	if ( tok[0] == NUL )
	    break;

	if ( !okF ) {
	    /* Token overflow; skip to next 'end' and continue. */
	    okF = mcdb_read_inp_end( inP );
	    continue;
	}

	/* Process the main token that we got */
	if ( stricmp( (char *)&tok[0], "mclient" ) == 0 )
	    okF = mcdb_read_inp_mclient( inP );

	else {
	    /* Unknown main token */
	    warning(
"mcdb_read_inp: unrecognized token \"%s\" at top level, line %d in \"%s\"",
			&tok[0],
			in_lineno( inP ), in_name( inP ) );

	    if ( stricmp( (char *)&tok[0], "end" ) != 0 )
		okF = mcdb_read_inp_mclient( inP );
	}
    }

    return ( okF );
}
/*

*//* mcdb_read_inp_end( inP )

	Skip to an "end" token

Accepts :

	inP		Ptr to input stream

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Gets tokens until an 'end' is seen.  Returns FALSE if no
	'end' ever seen.

*/

static BOOL
mcdb_read_inp_end ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* The input stream */
{
	UBYTE		tok[50];	/* Primary token */

    /* Look for end */
    for ( ; ; ) {
	if ( mcdb_read_token( inP, &tok[0], sizeof(tok) ) ) {
	    if ( tok[0] == NUL )
		return ( FALSE );	/* End of input, no 'end' */

	    if ( stricmp( (char *)&tok[0], "end" ) == 0 )
		break;
	}
    }

    return ( TRUE );
}
/*

*//* mcdb_read_inp_int( inP, longP )

	Read an integer value from the input

Accepts :

	inP		Ptr to input stream
	longP		Where to store the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a sign or a digit.

	For a FALSE return, the return value *longP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

static BOOL
mcdb_read_inp_int ARGLIST( (inP, longP) )
   NFARG( MML_IN	*inP )		/* The input stream */
    FARG( long		*longP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mcdb_read_token( inP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*longP = 0;
	return ( FALSE );
    }

    if ( ( tok[0] == '-' ) || ( tok[0] == '+' ) ||
	 ( isdigit( tok[0] ) ) ) {
	/* Input looks OK; get value and return it. */
	*longP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NUL ) )
	*longP = 1;
    else
	*longP = 0;

    return ( FALSE );
}
/*

*//* mcdb_read_inp_mclient( inP )

	Read an MCLIENT definition

Accepts :

	inP		Ptr to input stream

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Called when beginning of 'mclient' is seen in the input.

	Returns FALSE if there is some significant error or some reason to
	  stop looking at the input.  An error in the mclient definition
	  will be handled here and still return TRUE.

*/

static BOOL
mcdb_read_inp_mclient ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* The input stream */
{
	BOOL		skipF;		/* Skipping to semicolon */
	BOOL		keepF;		/* Whether to keep this entry */
	long		lv1;
	ULONG		ulv1, ulv2, ulv3;
	MCSSTATE	state1, state2;
	MCNAME_S	mcnms;		/* A place to build a mcname */
	MCLIENT		*mcP;
	MCTREE		*mctreeP;	/* Which tree it belongs in */
	char		nb[50];		/* Name, for debugging */
	UBYTE		tok[ MCNAME_BSMAX + 20 ];
	UBYTE		tok1[ MCNAME_BSMAX + 20 ];
	UBYTE		tok2[ MCNAME_BSMAX + 20 ];
	UBYTE		tok3[ MCNAME_BSMAX + 20 ];

    if ( !mcdb_read_inp_mcname( inP, &mcnms ) )
	return ( TRUE );		/* Bad input, but don't stop */

#if 1	/* Probably DO want the update flag set */

    /* Have a valid mcname; now would like to insert it into the tree
       even if there are later errors.  The insertion should be done
       without the side-effect of setting the "update me" flag though.
       I suppose it would be good to have a way to tell the generic
       mcdb_mclient_access() function not to set that flag, but then
       it wouldn't be so generic.  So just tediously do the a preparatory
       insert here; ignore errors.
    */
    switch ( mcnms.mns_mcname.mn_type ) {
	case MCT_AUTH:
	    mctreeP = &Mctree_auth;
	    break;

	case MCT_IP:
	    mctreeP = &Mctree_ip;
	    break;

	default:
	    warning( "case error in mcdb_read_inp_mclient; type=%d",
		      mcnms.mns_mcname.mn_type );
	    return ( mcdb_read_inp_end( inP ) );
    }

    avlinsert( mctreeP->mct_treeP,
		       (void *)mcnms.mns_mcname.mn_nameP, (void *)NULL );

#endif

    /* And now access the DB entry so we can update it. */
    mcP = mcdb_mclient_access( &mcnms.mns_mcname );
    if ( mcP == NULL ) {
	warning( "mcdb_read_inp_client: can't access mclient" );
	return ( mcdb_read_inp_end( inP ) );
    }
    mcP->mc_mcname.mn_type = mcnms.mns_mcname.mn_type;

    if ( Debug > 1 ) {
	mcname_str( &mcP->mc_mcname, &nb[0], sizeof(nb) );
	fprintf( stderr, "adding entry: %s\n", nb );
    }

    /* Now get the data for this mail client.  This is probably pretty
       ugly.
    */
    skipF = TRUE;			/* Skip to first semi, first */
    for ( ; ; ) {
	if ( ! mcdb_read_token( inP, &tok[0], sizeof(tok) ) ) {
	    /* Token too large; skip to the next semi. */
	    skipF = TRUE;
	    continue;			/* too large, warning given */
	}
	if ( tok[0] == NUL )
	    break;			/* No more tokens */
	if ( skipF ) {
	    /* Skipping to semi... */
	    if ( ( tok[0] != ';' ) || ( tok[1] != NUL ) ) {
		warning( "mcdb_read_inp_mclient: expected semi, got \"%s\"",
			&tok[0] );
	    }
	    else
		skipF = FALSE;		/* Found it */
	    continue;
	}

	/* Toggle the skip flag back on; anything else handled here (valid
	   or not) will need to look for semicolon.
	*/
	skipF = TRUE;

	/* Process token.  Check end first.. */
	if ( stricmp( (char *)&tok[0], "END" ) == 0 )
	    break;

	else if ( stricmp( (char *)&tok[0], "AGG" ) == 0 )
	    /* Flag as an aggregate */
	    mcP->mc_flags |= MCF_AGG;

	else if ( stricmp( (char *)&tok[0], "AUTHREQ" ) == 0 )
	    /* AUTH is required for this source. */
	    mcP->mc_flags |= MCF_AUTHREQ;

	else if ( stricmp( (char *)&tok[0], "BLOCK" ) == 0 ) {
	    /* Blocked:  type how end code */
	    if ( !mcdb_read_token( inP, &tok1[0], sizeof(tok1) ) ) {
		warning( "mcdb_read_inp_mclient: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else if ( ( tok1[0] == ';' ) && ( tok1[1] == NUL ) )
		skipF = FALSE;
	    else if ( !mcdb_read_token( inP, &tok2[0], sizeof(tok2) ) ) {
		warning( "mcdb_read_inp_mclient: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else if ( ( tok2[0] == ';' ) && ( tok2[1] == NUL ) )
		skipF = FALSE;
	    else if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else if ( !mcdb_read_token( inP, &tok3[0], sizeof(tok3) ) )
		;
	    else if ( ( tok3[0] == ';' ) && ( tok3[1] == NUL ) ) {
		warning( "mcdb_read_inp_mclient: too few arguments to BLOCK" );
		skipF = FALSE;
	    }
	    else {
		if ( stricmp( (char *)&tok1[0], "conn" ) == 0 ) {
		    state1 = MCS_BLOCK_C_ADM;   state2 = MCS_BLOCK_C_AUTO;
		}
		else if ( stricmp( (char *)&tok1[0], "greet" ) == 0 ) {
		    state1 = MCS_BLOCK_G_ADM;   state2 = MCS_BLOCK_G_AUTO;
		}
		else if ( stricmp( (char *)&tok1[0], "smtp" ) == 0 ) {
		    state1 = MCS_BLOCK_S_ADM;   state2 = MCS_BLOCK_S_AUTO;
		}
		else {
		    warning( "mcdb_read_inp_mclient: invalid block type \"%s\"",
			      &tok1[0] );
		    state1 = MCS_NONE;
		}

		if ( state1 == MCS_NONE )
		    ;
		else if ( stricmp( (char *)&tok2[0], "admin" ) == 0 )
		    mcP->mc_sstate = state1;
		else if ( stricmp( (char *)&tok2[0], "auto" ) == 0 )
		    mcP->mc_sstate = state2;
		else {
		    warning( "mcdb_read_inp_mclient: invalid block type \"%s\"",
			      &tok2[0] );
		    mcP->mc_sstate = MCS_NONE;
		}

		if ( mcP->mc_sstate != MCS_NONE ) {
		    mcP->mc_sstime = ulv1;
TODO		    /* Fill in the code from tok3 */
		}
	    }
	}

	else if ( stricmp( (char *)&tok[0], "CONN" ) == 0 ) {
	    /* Connection count */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else
		mcP->mc_connC = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "DNSBL" ) == 0 )
	    /* DNSBL check should be done for this source */
	    mcP->mc_flags |= MCF_DNSBL;

	else if ( stricmp( (char *)&tok[0], "FCTIME" ) == 0 ) {
	    /* First connect time */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else
		mcP->mc_fctime = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "LCTIME" ) == 0 ) {
	    /* Last connect time */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else
		mcP->mc_lctime = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "NO-RGOV" ) == 0 )
	    /* Suppress rcpt-governing for this client */
	    mcP->mc_flags |= MCF_NO_RGOV;

	else if ( stricmp( (char *)&tok[0], "PASS" ) == 0 ) {
	    /* Gives client a "pass" on some checks.  The value
	       is, as with greylisting, either an end time (for specific
	       client) or an interval (for a template).
	    */
	    if ( mcdb_read_inp_uint( inP, &ulv1 ) ) {
		mcP->mc_ostate = MCO_PASS;
		mcP->mc_ostime = ulv1;
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );
	    }
	}

	else if ( stricmp( (char *)&tok[0], "PGD" ) == 0 ) {
	    /* Pre-greeting delay parameters:
	       PGD sec wcon wday
	    */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else if ( !mcdb_read_inp_uint( inP, &ulv2 ) )
		skipF = ( ulv2 != 1 );
	    else if ( !mcdb_read_inp_uint( inP, &ulv3 ) )
		skipF = ( ulv3 != 1 );
	    else {
		mcP->mc_pgd_sec = ulv1;
		mcP->mc_pgd_wcon = ulv2;
		mcP->mc_pgd_wday = ulv3;
	    }
	}
	else if ( stricmp( (char *)&tok[0], "PGDWIN" ) == 0 ) {
	    /* Pre-greeting delay window data:
	       PGDWIN conn wstart
	    */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else if ( !mcdb_read_inp_uint( inP, &ulv2 ) )
		skipF = ( ulv2 != 1 );
	    else {
		mcP->mc_pgd_wconC = ulv1;
		mcP->mc_pgd_wtime = ulv2;
	    }
	}
	else if ( stricmp( (char *)&tok[0], "PROTO" ) == 0 )
	    /* Flag as a prototype */
	    mcP->mc_flags |= MCF_PROTO;

	else if ( stricmp( (char *)&tok[0], "RCPT" ) == 0 ) {
	    /* RCPT-TO counters */
	    if ( !mcdb_read_inp_uint( inP, &ulv1 ) )
		skipF = ( ulv1 != 1 );
	    else if ( !mcdb_read_inp_uint( inP, &ulv2 ) )
		skipF = ( ulv2 != 1 );
	    else {
		mcP->mc_grcptC = ulv1;
		mcP->mc_brcptC = ulv2;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "RELAYOK" ) == 0 )
	    /* Relay is OK for this */
	    mcP->mc_flags |= MCF_RELAYOK;

	else if ( stricmp( (char *)&tok[0], "RGOV" ) == 0 ) {
	    /* RCPT-TO governor value and last decay time */
	    if ( !mcdb_read_inp_int( inP, &lv1 ) )
		skipF = ( lv1 != 1 );
	    else {
		mcP->mc_rcptgov = lv1;

		/* Last decay time is optional for now */
		if ( !mcdb_read_inp_uint( inP, &ulv2 ) ) {
		    mcP->mc_rgtime = 0;
		    skipF = (ulv2 != 1 );
		}
		else
		    mcP->mc_rgtime = ulv2;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "TRANSIENT" ) == 0 )
	    /* This (or any child created from it) is transient */
	    mcP->mc_flags |= MCF_TRANSIENT;

	else if ( stricmp( (char *)&tok[0], "T.GL" ) == 0 ) {
	    /* Specifies greylisting; value is either end time or 
		interval: at any rate, it's a number.
	    */
	    if ( mcdb_read_inp_uint( inP, &ulv1 ) ) {
		mcP->mc_sstate = MCS_GREY;
		mcP->mc_sstime = ulv1;
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		skipF = ( ulv1 != 1 );
	    }
	}

	else if ( stricmp( (char *)&tok[0], "T.GREET" ) == 0 ) {
	    /* Pre-greeting delay */
	    /* NOTE: OBSOLETE.  This is an obsolete element, replaced by
	       PGD.  Kept here to help with updating the databases.
	    */
	    if ( mcdb_read_inp_int( inP, &lv1 ) ) {
		if ( ( lv1 < 0 ) || ( lv1 > 255 ) )
		    warning( "mcdb_read_inp_mclient: invalid T.GREET %ld",
				lv1 );
		else {
		    mcP->mc_pgd_sec = lv1;

		    /* Don't overwrite/set any other pgd values; they
		       should mostly be zeroed already anyway.
		    */
		}
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		skipF = ( lv1 != 1 );
	    }
	}

	else {
	    /* Unrecognized; ignore by skipping to semi. */
	    warning( "mcdb_read_inp_mclient: unknown \"%s\"", &tok[0] );
	}
    }
	
    /* If we're here, we found the "END" or ran out of input.  At any
       rate, we're done reading this mail client.
    */

    /* See if we want to keep it */
    if ( ( mcP->mc_flags & (MCF_AGG|MCF_PROTO) ) != 0 )
	/* Always keep prototypes and aggregates */
	keepF = TRUE;
    else if ( ( mcP->mc_sstate == MCS_BLOCK_C_ADM ) ||
	      ( mcP->mc_sstate == MCS_BLOCK_G_ADM ) ||
	      ( mcP->mc_sstate == MCS_BLOCK_S_ADM ) )
	/* Always keep admin blocks */
	keepF = TRUE;
    else if ( mcP->mc_ostate != MCO_NONE )
	/* Always keep blocks with override states */
	keepF = TRUE;
    else {
	/* Don't have to keep; see if we want to drop it */
	if ( mcP->mc_lctime < Mclient_oldest_wanted )
	    /* Haven't seen in a long time, drop it. */
	    keepF = FALSE;
	else
	    /* No decision, keep it. */
	    keepF = TRUE;
    }

    mcdb_mclient_release( mcP, FALSE );

    if ( !keepF ) {
	if ( Debug > 1 )
	    fprintf( stderr, " ... dropping, no longer wanted.\n" );

	avldelete( mctreeP->mct_treeP, (void *)mcnms.mns_mcname.mn_nameP );
    }


    return ( TRUE );
}
/*

*//* mcdb_read_inp_mcname( inP, mcnmsP )

	Read a mail client name from input stream

Accepts :

	inP		Ptr to input stream
	mcnmsP		Where to put the MCNAME info

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	On FALSE return, this function will have already attempted to find
	the next 'end'

*/

static BOOL
mcdb_read_inp_mcname ARGLIST( (inP) )
   NFARG( MML_IN	*inP )		/* The input stream */
    FARG( MCNAME_S	*mcnmsP )	/* The name block to fill in */
{
	BOOL		okF;
	int		i;
	UBYTE		*strP;		/* Generic pointer */
	UBYTE		ttok[50];	/* type */
	UBYTE		ntok[MCNAME_BSMAX +1 ];

    /* Initialize the static pointer to point to the contained data */
    mcnmsP->mns_mcname.mn_nameP = (void *)&mcnmsP->mns_data;

    /* Get the type and the name. */
    okF = mcdb_read_token( inP, &ttok[0], sizeof(ttok) );
    okF = okF && mcdb_read_token( inP, &ntok[0], sizeof(ntok) );

    if ( okF ) {
	if ( stricmp( (char *)&ttok[0], "auth" ) == 0 ) {
	    mcnmsP->mns_mcname.mn_type = MCT_AUTH;
	    strncpy( (char *)&mcnmsP->mns_data.nm_bs[0],
		     (char *)&ntok[0], MCNAME_BSMAX );
	}

	else if ( stricmp( (char *)&ttok[0], "ip" ) == 0 ) {
	    strP = (UBYTE *)strchr( (char *)&ntok[0], '/' );
	    if ( strP != NULL )
		*strP++ = NUL;
	    else
		strP = (UBYTE *)"32";

	    if ( !ip_decode_str( &ntok[0],
			         &mcnmsP->mns_data.nm_ip4.nip_addr ) ) {
		warning( "mcdb_read_inp_mcname: bad IP string \"%s\"",
				&ntok[0] );
		okF = FALSE;
	    }
	    else if ( ( ( i = atod( strP ) ) < 0 ) || ( i > 32 ) ) {
		warning( "mcdb_read_inp_mcname: bad IP masklen \"%d\"",
				i );
		okF = FALSE;
	    }
	    else {
		mcnmsP->mns_data.nm_ip4.nip_masklen = i;
		mcnmsP->mns_mcname.mn_type = MCT_IP;
	    }
	}

	else {
	    warning( "mcdb_read_inp_mcname: unknown mcname type \"%s\"",
		      &ttok[0] );
	    okF = FALSE;
	}
    }

    if ( !okF )
	/* Skip to next 'end' */
	mcdb_read_inp_end( inP );

    return ( okF );
}
/*

*//* mcdb_read_inp_uint( inP, ulongP )

	Read an unsigned integer value from the input

Accepts :

	inP		Ptr to input stream
	ulongP		Where to store the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a digit or a plus.

	For a FALSE return, the return value *ulongP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

static BOOL
mcdb_read_inp_uint ARGLIST( (inP, ulongP) )
   NFARG( MML_IN	*inP )		/* The input stream */
    FARG( ULONG		*ulongP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mcdb_read_token( inP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*ulongP = 0;
	return ( FALSE );
    }

    if ( isdigit( tok[0] ) || ( tok[0] == '+' ) ) {
	/* Input looks OK; get value and return it. */
	*ulongP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NUL ) )
	*ulongP = 1;
    else
	*ulongP = 0;

    return ( FALSE );
}
/*

*//* mcdb_read_token( inP, bufP, bufL )

	Read a token on behalf of mcdb_read functions

Accepts :

	inP		Ptr to input stream
	bufP		Where to store the token
	bufL		Length of token buffer

Returns :

	<value>		TRUE if OK
			FALSE if not (token overflow).

Notes :

	This is just a wrapper around in_token_ws; the wrapper will
	  emit a warning message if token overflow.

	This wrapper also processes end-of-line comments.  If a token
	  starts an end-of-line comment, the rest of the line will be
	  skipped.

	An empty token indicates end of data, and is OK (yielding TRUE).

*/

static BOOL
mcdb_read_token ARGLIST( (inP, bufP, bufL) )
   NFARG( MML_IN	*inP )		/* The input stream */
   NFARG( UBYTE		*bufP )		/* Where to put token */
    FARG( int		bufL )		/* Token buffer size */
{
	int		cC;
	int		ch;

    /* Wrap in a comment-skipping loop. */
    for ( ; ; ) {
	cC = in_token_ws( inP, bufP, bufL, TRUE );
	if ( cC >= bufL ) {
	    warning( "mcdb_read: token too large in line %d of \"%s\"",
		      in_lineno( inP ), in_name( inP ) );
	    return ( FALSE );
	}

	/* Check comment */
	if ( ( ( bufP[0] == '#' ) && ( bufP[1] == NUL ) ) ||
	     ( ( bufP[0] == '/' ) && ( bufP[1] == '/' ) && ( bufP[2] == NUL ) )
	   ) {
	    while ( ( ch = in_nonblank( inP, FALSE ) ) != EOF )
		if ( ch == '\n' )
		    break;
	}
	else {
	    break;			/* Not comment... take it */
	}
    }

    return ( TRUE );
}
/*

*//* mclient_proto_inherit( mcP, protoP )

	Inherit info from an mclient prototype

Accepts :

	mcP		The mail client to fill in
	protoP		The mail client to use as a prototype

Returns :

	<nothing>

Notes :


*/

static void
mclient_proto_inherit ARGLIST( (mcP, protoP) )
   NFARG( MCLIENT	*mcP )		/* Mail client to fill in */
    FARG( MCLIENT	*protoP )	/* Prototype mail client */
{
    mcP->mc_flags = protoP->mc_flags & ~(MCF_AGG|MCF_PROTO);

    /* Inherit and set up any simple state */
    mcP->mc_sstate = protoP->mc_sstate;
    switch ( mcP->mc_sstate ) {
	case MCS_NONE:			/* No specific state */
	    break;

	case MCS_GREY:			/* In greylisting period */
	    /* The prototype's sstime contains the greylisting
	       interval.  Set up the new sstime as the end time.
	    */
	    time( &mcP->mc_sstime );
	    mcP->mc_sstime += protoP->mc_sstime;
	    break;

	case MCS_BLOCK_C_ADM:		/* Various block states */
	case MCS_BLOCK_C_AUTO:
	case MCS_BLOCK_G_ADM:
	case MCS_BLOCK_G_AUTO:
	case MCS_BLOCK_S_ADM:
	case MCS_BLOCK_S_AUTO:
	    break;

	default:
	    warning( "mclient_proto_inherit -- unknown mc_sstate %d", 
		      mcP->mc_sstate );
	    mcP->mc_sstate = MCS_NONE;
	    break;
    }

    /* Inherit and set up any override state */
    mcP->mc_ostate = protoP->mc_ostate;
    switch ( mcP->mc_ostate ) {
	case MCO_NONE:			/* No specific state */
	    break;

	case MCO_PASS:			/* Gets a pass */
	    /* The prototype's ostime contains an interval, not an end
	       time.  We'll use that to set up a new end time.
	    */
	    time( &mcP->mc_ostime );
	    mcP->mc_ostime += protoP->mc_ostime;
	    break;

	default:
	    warning( "mclient_proto_inherit -- unknown mc_ostate %d", 
		      mcP->mc_ostate );
	    mcP->mc_ostate = MCO_NONE;
	    break;
    }

    /* Other value(s) to inherit */
    mcP->mc_pgd_sec= protoP->mc_pgd_sec;
    mcP->mc_pgd_wcon= protoP->mc_pgd_wcon;
    mcP->mc_pgd_wday= protoP->mc_pgd_wday;

    /* done */
}
/*

*//* mclient_prune( argP )

	Prune an mclient tree

Accepts :

	argP		Ptr to the tree to prune

Returns :

	<nothing>

Notes :

	This is called via a timer function.  The timer function passes
	an opaque argument, so this function is defined to accept one.

	Reschedules itself.

*/

static void
mclient_prune ARGLIST( (argP) )
    FARG( void		*argP )		/* Dummy argument */
{
	int		n;
	time_t		timenow;
	MCTREE		*mctreeP;
	LL_NODE		*lnodeP;	/* Linked list node */
	MCNODE		*mcnodeP;	/* Mail client node */

    /* Get tree pointer we're working with */
    mctreeP = (MCTREE *)argP;

    if ( Debug > 1 )
	printf( "mclient_prune here.\n" );

    /* Must have exclusive access to the tree encapsulation */
    if ( pthread_mutex_lock( &mctreeP->mct_lock ) != 0 )
	warning( "mclient_prune: can't get lock" );

    /* Need to walk the tree to find nodes to delete.
       Can't modify the tree while walking it, so the nodes will be
       added to a linked list, up to some maximum number.  Later we'll
       walk the list and remove the indicated nodes.
    */
    Mclient_prune_listC = 0;
    time( &Mclient_prune_time );
    Mclient_prune_before = Mclient_prune_time - MCLIENT_PRUNE_AGE;
    avlwalk( mctreeP->mct_treeP, mclient_prune_node );

    /* Now remove each node that's on the prune list.  We can use the data
       in the node itself as a key to avldelete, since the key is only
       used for comparison prior to deletion.
    */
    if ( Debug > 1 )
	printf( "  mclient_prune found %d node%s.\n",
	         Mclient_prune_listC,
		 Mclient_prune_listC == 1 ? "" : "s" );

    while ( ( lnodeP = Mclient_prune_list.llh_firstP ) != NULL ) {
	mcnodeP = (MCNODE *)lnodeP->lln_valP;

	/* Remove this agent block from the tree (and thus from existence).
	   Note that we use the tree pointer from the node and not the
	   assumed one; why not.
	*/
	if ( avldelete( mcnodeP->mcn_mctreeP->mct_treeP,
		       mcnodeP->mcn_mclient.mc_mcname.mn_nameP ) < 0 )
	    warning( "mclient_prune: avldelete failure." );

	/* And remove the node... since we already have exclusive access,
	   no need to lock the list.
	*/
	ll_rem_nolock( lnodeP );
    }

    mctreeP->mct_purgeC += Mclient_prune_listC;
    ++mctreeP->mct_pruneC;

    /* Set the next prune time */
    time( &timenow );
    if ( ( Mclient_prune_listC >= MCLIENT_PRUNE_MAX ) &&
         ( MCLIENT_PRUNE_INTERVAL_SHORT < Mclient_prune_interval ) )
	/* Need another pruning sooner than normal */
	mctreeP->mct_nptime = timenow + MCLIENT_PRUNE_INTERVAL_SHORT;
    else
	/* Do standard interval */
    mctreeP->mct_nptime = timenow + Mclient_prune_interval;

    /* Done with this tree's pruning */
    if ( Debug > 1 )
	printf( "mclient_prune done, next in %ld.\n",
	         mctreeP->mct_nptime - timenow );
    if ( pthread_mutex_unlock( &mctreeP->mct_lock ) != 0 )
	warning( "mclient_prune: can't release lock" );

    /* Find the tree that wants the soonest pruning, and schedule it */
    for ( n = 0; Mctrees[n] != NULL; ++n ) {
	/* Initialize next prune time if not already set */
	if ( Mctrees[n]->mct_nptime == 0 )
	    Mctrees[n]->mct_nptime = timenow + (n * 30 ) +
				MCLIENT_PRUNE_INTERVAL_SHORT;

	/* Select lowest time */
	if ( Mctrees[n]->mct_nptime < mctreeP->mct_nptime )
	    mctreeP = Mctrees[n];
    }

    /* Reschedule the prune function for the selected tree */
    if ( !timer_add( mctreeP->mct_nptime, mclient_prune, mctreeP ) )
	warning( "mclient_prune: timer_add failed for mclient_prune." );

    /* done */
}
/*

*//* mclient_prune_node( treeP, nodeP, depth )

	Check a mail client node for pruning

Accepts :

	treeP		Ptr to AVL tree
	nodeP		Ptr to the node to write
	depth		depth of the node

Returns :

	<value>		0 to continue
			non-zero to stop

Notes :

	Per avlwalk() interface; see that.

*/

static int
mclient_prune_node ARGLIST( (treeP, nodeP, depth) )
   NFARG( AVLTREE	*treeP )	/* Ptr to tree */
   NFARG( MCNODE	*nodeP )	/* Ptr to node */
    FARG( int		depth )		/* Depth of node */
{
	MCLIENT		*mcP;

    /* NULL node signals end of walk */
    if ( nodeP == NULL )
	return ( 0 );			/* no special processing */

    /* Skip this node if it's got any accesses against it */
    if ( nodeP->mcn_accessF )
	return ( 0 );

    mcP = &nodeP->mcn_mclient;

    /* Skip this node if it is some kind of permanent type */
    if ( ( mcP->mc_flags & (MCF_AGG|MCF_PROTO) ) != 0 )
	return ( 0 );

    /* Do further check according to the simple state */
    switch ( mcP->mc_sstate ) {
	case MCS_GREY:			/* Greylisted */
	    /* Skip this node only if it is still in the grace period. */
	    if ( ( Mclient_prune_time - mcP->mc_sstime ) < MCLIENT_PRUNE_GREY )
		return ( 0 );
	    break;

	case MCS_BLOCK_C_ADM:		/* Admin blocks never go away */
	case MCS_BLOCK_G_ADM:
	case MCS_BLOCK_S_ADM:
	    return ( 0 );
	    
	case MCS_NONE:			/* Generic handling */
	case MCS_BLOCK_C_AUTO:
	case MCS_BLOCK_G_AUTO:
	case MCS_BLOCK_S_AUTO:
	    /* Skip this node if it has been heard from recently */
	    if ( mcP->mc_lctime >= Mclient_prune_before )
		return ( 0 );
	    break;

	default:
	    warning( "mclient_prune_node: unknown mclient state %d, removing.",
		        mcP->mc_sstate );
	    break;
    }

    /* Anything with an override state doesn't get pruned. */
    if ( mcP->mc_ostate != MCO_NONE )
	return ( 0 );

    /* We'll want to delete this node.  Add it to the prune list. */
    ll_addvt( &Mclient_prune_list, nodeP );

    /* Bump count... and be done if we've hit the max */
    if ( ++Mclient_prune_listC >= MCLIENT_PRUNE_MAX )
	return ( 1 );

    return ( 0 );
}
/*

*//* mctree_write( mctreeP )

	Write a mail client tree to disk

Accepts :

	mctreeP		Ptr to the tree to write

Returns :

	<nothing>

Notes :

	Is most likely to be called via timer function, scheduled via
	mctree_write_sched().  Can, however, be called directly if
	caller pays attention to certain parameters such as the
	scheduled write count.

*/

static void
mctree_write ARGLIST( (mctreeP) )
    FARG( MCTREE	*mctreeP )	/* Tree handle */
{
	BOOL		okF;
	int		i;
	AVLTREE		*treeP;		/* The underlying AVL tree */
	MCWRITE		winfo;		/* Tree-writing info */
	char		fname[ PATH_MAX +1];  /* Basename */
	char		tmpname[ PATH_MAX +1 ]; /* temp name */

    treeP = mctreeP->mct_treeP;

    /* Must have exclusive access to the tree encapsulation */
    if ( pthread_mutex_lock( &mctreeP->mct_lock ) != 0 )
	warning( "mctree_write: can't get lock" );

    /* Calculate the target name */
    if ( snprintf( &fname[0], sizeof(fname), "%s/mctree_%s",
		  MVMCAS_RDBDIR,
		  mctreeP->mct_tnameP ) >= sizeof(fname) ) {
	warning( "mctree_write: file name too large for type %s",
		  mctreeP->mct_tnameP );
    }

    /* And the tmp name */
    else if ( snprintf( &tmpname[0], sizeof(tmpname), "%s.tmp",
		        &fname[0] ) >= sizeof(tmpname) ) {
	warning( "mctree_write: temp file name too large." );
    }
    else if ( ( winfo.mcw_fP = fopen( &tmpname[0], "w" ) ) == NULL ) {
	warning( "mctree_write: can't open \"%s\"", &fname[0] );
    }
    else {
	/* Do the write-walk */
	winfo.mcw_okF = TRUE;
	treeP->t_walkP = (void *)&winfo;
	avlwalk( treeP, mctree_write_node );

	fclose( winfo.mcw_fP );

	if ( !winfo.mcw_okF ) {
	    warning( "mctree_write failed, tmp file not deleted." );
	}
	else {
	    /* Generate n backup copies */
	    if ( !file_rotate( &fname[0], &tmpname[0], 3 ) )
		;
	}
    }

    /* Decrement scheduled writes */
    if ( mctreeP->mct_writeC > 0 )
	--mctreeP->mct_writeC;
    else
	warning( "mctree_write: writeC is somehow zero.  BUG!" );

    if ( pthread_mutex_unlock( &mctreeP->mct_lock ) != 0 )
	warning( "mctree_write: can't release lock" );

    /* done */
}
/*

*//* mctree_write_node( treeP, nodeP, depth )

	Write a mail client node on behalf of AVL walk 

Accepts :

	treeP		Ptr to AVL tree
	nodeP		Ptr to the node to write
	depth		depth of the node

Returns :

	<value>		0 to continue
			non-zero to stop

Notes :

	Per avlwalk() interface; see that.

*/

static int
mctree_write_node ARGLIST( (treeP, nodeP, depth) )
   NFARG( AVLTREE	*treeP )	/* Ptr to tree */
   NFARG( MCNODE	*nodeP )	/* Ptr to node */
    FARG( int		depth )		/* Depth of node */
{
	BOOL		okF;
	int		i;
	FILE		*ofP;		/* Output file ptr */
	MCWRITE		*mcwP;		/* Write info */
	MCLIENT		*mcP;		/* Mail client info */
	UBYTE		mcnamestr[ MCNAME_BSMAX + 20 ];

    /* NULL node signals end of walk */
    if ( nodeP == NULL )
	return ( 0 );			/* no special processing */

    mcwP = (MCWRITE *)treeP->t_walkP;
    mcP = &nodeP->mcn_mclient;

    /* Check flags and don't write if:
       - this is a prototype (it doesn't get dynamically written);
    */
    if ( ( mcP->mc_flags & (MCF_PROTO) ) != 0 )
	return ( 0 );

    i = mcname_str( &mcP->mc_mcname, &mcnamestr[0], sizeof(mcnamestr) );
    if ( ( i < 0 ) || ( i >= sizeof(mcnamestr) ) ) {
	warning( "mctree_write_node can't make mail client name string." );
	okF = FALSE;
    }
    else
	okF = TRUE;

    ofP = mcwP->mcw_fP;

    okF = okF && ( fprintf( ofP, "MCLIENT %s", &mcnamestr[0] ) > 0 );

    /* Begin values */
    okF = okF && ( fprintf( ofP, " ;" ) > 0 );

    /* Add any flags on same line */
    if ( ( mcP->mc_flags & MCF_AGG ) != 0 )
	okF = okF && ( fprintf( ofP, " AGG ;" ) > 0 );
    if ( ( mcP->mc_flags & MCF_AUTHREQ ) != 0 )
	okF = okF && ( fprintf( ofP, " AUTHREQ ;" ) > 0 );
    if ( ( mcP->mc_flags & MCF_DNSBL ) != 0 )
	okF = okF && ( fprintf( ofP, " DNSBL ;" ) > 0 );
    if ( ( mcP->mc_flags & MCF_RELAYOK ) != 0 )
	okF = okF && ( fprintf( ofP, " RELAYOK ;" ) > 0 );
    if ( ( mcP->mc_flags & MCF_TRANSIENT ) != 0 )
	okF = okF && ( fprintf( ofP, " TRANSIENT ;" ) > 0 );
    if ( ( mcP->mc_flags & MCF_NO_RGOV ) != 0 )
	okF = okF && ( fprintf( ofP, " NO-RGOV ;" ) > 0 );
    okF = okF && ( fputc( '\n', ofP ) != EOF );

    /* State info */
    switch ( mcP->mc_sstate ) {
	case MCS_NONE:			/* No specific state */
	    break;

	case MCS_GREY:			/* In greylisting period */
	    okF = okF && ( fprintf( ofP, " T.GL %lu ;\n", mcP->mc_sstime ) > 0 );
	    break;

	case MCS_BLOCK_C_ADM:		/* Block connection, administrative */
	    okF = okF && ( fprintf( ofP, " BLOCK conn admin %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	case MCS_BLOCK_C_AUTO:		/* Block connection, automatic */
	    okF = okF && ( fprintf( ofP, " BLOCK conn auto %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	case MCS_BLOCK_G_ADM:		/* Block at greeting, administrative */
	    okF = okF && ( fprintf( ofP, " BLOCK greet admin %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	case MCS_BLOCK_G_AUTO:		/* Block at greeting, automatic */
	    okF = okF && ( fprintf( ofP, " BLOCK greet auto %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	case MCS_BLOCK_S_ADM:		/* Block with dialog, administrative */
	    okF = okF && ( fprintf( ofP, " BLOCK smtp admin %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	case MCS_BLOCK_S_AUTO:		/* Block with dialog, automatic */
	    okF = okF && ( fprintf( ofP, " BLOCK smtp auto %lu %s ;",
				    mcP->mc_sstime,
				    "1234" ) > 0 );
	    break;

	default:
	    warning( "mctree_write_node -- unknown mc_sstate %d", 
		      mcP->mc_sstate );
	    okF = FALSE;
	    break;
    }


    /* Override state info */
    switch ( mcP->mc_ostate ) {
	case MCO_NONE:			/* None set */
	    break;

	case MCO_PASS:			/* Client gets a pass */
	    okF = okF && ( fprintf( ofP, " PASS %lu ;\n", mcP->mc_ostime ) > 0 );
	    break;

	default:
	    warning( "mctree_write_node -- unknown mc_ostate %d", 
		      mcP->mc_ostate );
	    okF = FALSE;
	    break;
    }

    if ( mcP->mc_pgd_sec != 0 ) {
	okF = okF &&
	      ( fprintf( ofP, " PGD %d %d %d ;",
			(int)mcP->mc_pgd_sec,
			(int)mcP->mc_pgd_wcon,
			(int)mcP->mc_pgd_wday ) > 0 );

	okF = okF &&
	      ( fprintf( ofP, " PGDWIN %d %ld ;",
			(int)mcP->mc_pgd_wconC,
			(long)mcP->mc_pgd_wtime ) > 0 );
    }

    /* Connection times and counts */
    okF = okF && ( fprintf( ofP, " CONN %ld ;", mcP->mc_connC ) > 0 );
    if ( mcP->mc_fctime != 0 )
	okF = okF &&
		( fprintf( ofP, " FCTIME %ld ;", mcP->mc_fctime ) > 0 );
    if ( mcP->mc_lctime != 0 )
	okF = okF &&
		( fprintf( ofP, " LCTIME %ld ;", mcP->mc_lctime ) > 0 );
    okF = okF && ( fputc( '\n', ofP ) != EOF );


    /* RCPT-TO counters and governor */
    okF = okF &&
	    ( fprintf( ofP, " RCPT %ld %ld ;",
		      mcP->mc_grcptC, mcP->mc_brcptC ) > 0 );
    if ( mcP->mc_rcptgov != 0 )
	okF = okF && ( fprintf( ofP, " RGOV %d %ld ;",
			 mcP->mc_rcptgov, mcP->mc_rgtime ) > 0 );
    okF = okF && ( fputc( '\n', ofP ) != EOF );

    /* End entry */
    okF = okF && ( fprintf( ofP, " END\n" ) > 0 );

    if ( okF )
	return ( 0 );

    /* Some error; mark and be done */
    mcwP->mcw_okF = FALSE;
    return ( 1 );
}
/*

*//* mctree_write_sched( mctreeP, mdelay )

	Schedules a write for a mail client tree

Accepts :

	mctreeP		Ptr to the tree to write
	mdelay		Minutes to delay until writing

Returns :

	<nothing>

Notes :

	Called to schedule a write.  mdelay has several meanings:

	  -1 -- do a write asap, regardless of any scheduled writes.

	   0 -- schedule a write with normal delay.

	  +n -- if no writes are scheduled, delay n minutes.

*/

static void
mctree_write_sched ARGLIST( (mctreeP, mdelay) )
   NFARG( MCTREE	*mctreeP )	/* Tree handle */
    FARG( int		mdelay )	/* Minutes delay */
{
	time_t		when;
	int		sdelay;		/* Seconds to delay */

    /* Lock the mctree while we act on the request */
    if ( pthread_mutex_lock( &mctreeP->mct_lock ) != 0 )
	warning( "mctree_write_sched: can't get lock" );

    /* Interpret mdelay */
    if ( mdelay == 0 )
	/* Use default */
	mdelay = 47;

    if ( mdelay < 0 ) {
	/* Wants a write asap. */
	sdelay = 5;
    }
    else {
	/* Wants a delayed write but only if none is already scheduled. */
	if ( mctreeP->mct_writeC > 0 ) {
	    if ( pthread_mutex_unlock( &mctreeP->mct_lock ) != 0 )
		warning( "mctree_write_sched: can't release lock" );
	    return;
	}
	sdelay = 60 * mdelay;
    }

    /* Schedule the event */
    time( &when );
    when += sdelay;

    if ( timer_add( when, mctree_write, (void *)mctreeP ) ) {
	++mctreeP->mct_writeC;
	if ( Debug > 0 )
	    warning( "mctree_write for %s scheduled.",
			 mctreeP->mct_tnameP );
    }
    else {
	warning( "mctree_write_sched: timer_add failed for %s",
		      mctreeP->mct_tnameP );
    }

    if ( pthread_mutex_unlock( &mctreeP->mct_lock ) != 0 )
	warning( "mctree_write_sched: can't release lock" );

    /* done */
}
