/* mcoccur.c -- mail client occurances

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This module provides functions to handle mail client occurrances:

 - keeping track of connections from a mail client
 - keeping track of mvmcas agents that are interested in a mail client

Also in this module are some functions relating to agents.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <ctype.h>
#include <pthread.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-avl.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "mvmcas-misc.h"
#include "mcdb.h"
#include "mcoccur.h"
#include "timer.h"

/* Local definitions */

    /* Structure of a non-persistent agent (np-agent) node. */
typedef struct {
    AVLNODE	npa_avl;	/* AVL node header */
    pthread_mutex_t npa_lock;	/* A lock for this node */
    AGENT	npa_agent;	/* Agent info */
    int		npa_accessC;	/* Access count */
    time_t	npa_heard;	/* Time last heard from */
}  NPANODE;


    /* Structure of a mail client occurance node */
typedef struct {
    AVLNODE	mon_avl;	/* AVL node header */
    pthread_mutex_t mon_lock;	/* Lock for this node */
    MCOCCUR	mon_occ;	/* Occurance info for this mail client */
}  MCONODE;


    /* Each mail client occurance has a list of agents that are interested
       in it.  This is the structure of the agent interest block.
    */
typedef struct {
    UWORD	ia_flags;	/* Flags */
    AGENT	*ia_agentP;	/* Ptr to the agent occurance */
}  IAGENT;



    /* Each agent occurance has a list of mail clients that it is connected
       to.  This is kept so that we can remove the agents from mail clients
       when the agent goes away
    */
typedef struct {
    MCOCCUR	*im_mcoccP;	/* Mail client occurance */
}  IMCLIENT;



/* External data referenced */

extern	int	Debug;		/* Debugging level */
extern	in_addr_t IP4mask[];
extern	long	Rgov_decay_int;		/* How often (seconds) to decay rgov */
extern	long	Rgov_decay_max;		/* Max interval to apply decay over */
extern	int	Rgov_decay_mult;	/* Decay rgov by multiplying by this */
extern	int	Rgov_decay_div;		/*   and then dividing by this */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* AVL support functions for AVL trees */

    /* These are for mail client occurance with keys that are IPv4 addresses */
static	int	avl_mco_ip4_cmpkey();
static	AVLNODE	*avl_mco_ip4_mknode();
static	int	avl_mco_ip4_rmnode();

    /* These are mail client occurance with keys that are simple strings */
static	int	avl_mco_str_cmpkey();
static	AVLNODE	*avl_mco_str_mknode();
static	int	avl_mco_str_rmnode();

    /* These are for the np-agent tree */
static	int	avl_npagent_cmpkey();
static	AVLNODE	*avl_npagent_mknode();
static	int	avl_npagent_rmnode();

    /* Initialize generic part of mail client occurance node */
static	void	avl_mconode_init PROTO( (MCONODE *nodeP) );

    /* Prune the npagent tree */
static	void	npagent_prune PROTO( (void *argP) );

    /* Called for each node by avlwalk on behalf of npagent_prune */
static	int	npagent_prune_node();


/* Private data */

    /* How often to prune np-agents, in seconds */
static	int	Npa_prune_interval = NPA_PRUNE_INTERVAL;

    /* How long to keep old agents, in seconds */
static	int	Npa_timeout = NPA_TIMEOUT;

    /* List to hold npagent nodes that are being pruned */
static	LL_HEAD	Npagent_prune_list;

    /* Non-persistent agent AVL tree */
static	AVLTREE	*Npagent_treeP;

    /* Mail client occurance AVL trees, one per mail client type */
static	AVLTREE	*Mcotree_authP;		/* For auth id */
static	AVLTREE	*Mcotree_ipP;		/* For IPv4 */


/*

*//* mcoccur_access( mcnameP )

	Access a mail client occurance

Accepts :

	mcnameP		Ptr to MCNAME for the mail client

Returns :

	<value>		Ptr to MCOCCUR block if found
			FALSE if some error.

Notes :

	The returned MCOCCUR block is locked until released via
	mcoccur_release().  

*/

MCOCCUR *
mcoccur_access ARGLIST( (mcnameP) )
    FARG( MCNAME	*mcnameP )	/* Mail client name */
{
	MCOCCUR		*mcoccP;

    switch ( mcnameP->mn_type ) {
	case MCT_AUTH:
	    mcoccP = mcoccur_access_auth( (UBYTE *)mcnameP->mn_nameP );
	    break;

	case MCT_IP:
	    mcoccP = mcoccur_access_ip( (MCNAME_IP4 *)mcnameP->mn_nameP );
	    break;

	case MCT_IPv6:
	    warning( "mcoccur_access: IPv6 type not supported." );
	    mcoccP = NULL;
	    break;

	default:
	    warning( "mcoccur_access: unknown mctype %d",
		     mcnameP->mn_type );
	    mcoccP = NULL;
	    break;
    }

    return ( mcoccP );
}
/*

*//* mcoccur_access_auth( authnameP )

	Get mail client occurance info for AUTH mail client type

Accepts :

	authnameP	Ptr to auth name string

Returns :

	<value>		Ptr to MCOCCUR block if successful
			NULL if not (some error)

Notes :


*/

MCOCCUR *
mcoccur_access_auth ARGLIST( (authnameP) )
    FARG( UBYTE		*authnameP )	/* Auth name */
{
	MCONODE		*mconodeP;	/* AVL node ptr */

    /* Try to find the entry, keeping the tree locked until we are all done. */
    mconodeP = (MCONODE *)avlfind_nounlock( Mcotree_authP, (void *)authnameP );
    if ( mconodeP == NULL ) {
	/* Can't find it, need to create it. */
	avlinsert_nolock( Mcotree_authP, (void *)authnameP, (void *)NULL );
	mconodeP = (MCONODE *)avlfind_nolock( Mcotree_authP, (void *)authnameP );
	if ( mconodeP == NULL ) {
	    /* Error.. */
	    if ( pthread_mutex_unlock( &Mcotree_authP->t_lock ) != 0 )
		warning( "mcoccur_access_auth: can't unlock tree in failure case" );
	    return ( NULL );
	}
	mconodeP->mon_occ.mco_mcname.mn_type = MCT_AUTH;
    }

    /* Lock & return */
    if ( pthread_mutex_lock( &mconodeP->mon_lock ) != 0 )
	warning( "mcoccur_access_auth: can't lock node" );

    if ( pthread_mutex_unlock( &Mcotree_authP->t_lock ) != 0 )
	warning( "mcoccur_access_auth: can't unlock tree." );

    return ( &mconodeP->mon_occ );
}
/*

*//* mcoccur_access_ip( mcnameP )

	Get client info for IP mail client type

Accepts :

	ipnameP		Name of client, MCNAME_IP4 format.

Returns :

	<value>		Ptr to MCOCCUR block if successful
			NULL if not (some error)

Notes :


*/

MCOCCUR *
mcoccur_access_ip ARGLIST( (ipnameP) )
    FARG( MCNAME_IP4	*ipnameP )	/* Mail client name (IP address) */
{
	MCONODE		*mconodeP;	/* AVL node ptr */

    /* Try to find the entry, keeping the tree locked */
    mconodeP = (MCONODE *)avlfind_nounlock( Mcotree_ipP, (void *)ipnameP );
    if ( mconodeP == NULL ) {
	/* Can't find it, need to create it. */
	avlinsert_nolock( Mcotree_ipP, (void *)ipnameP, (void *)NULL );
	mconodeP = (MCONODE *)avlfind_nolock( Mcotree_ipP, (void *)ipnameP );
	if ( mconodeP == NULL ) {
	    /* Error.. */
	    if ( pthread_mutex_unlock( &Mcotree_ipP->t_lock ) != 0 )
		warning( "mcoccur_access_ip: can't unlock tree in failure case" );
	    return ( NULL );
	}
    }

    /* Lock & return */
    if ( pthread_mutex_lock( &mconodeP->mon_lock ) != 0 )
	warning( "mcoccur_access_ip: can't lock node" );

    if ( pthread_mutex_unlock( &Mcotree_ipP->t_lock ) != 0 )
	warning( "mcoccur_access_ip: can't unlock tree." );

    return ( &mconodeP->mon_occ );
}
/*

*//* mcoccur_agint_add( mcoccP, agentP, flags )

	Establish an agent interest in a mail client occurance

Accepts :

	mcoccP		Ptr to mail client occurance data
	agentP		Ptr to agent occurance data
	flags		The agent interest flags

Returns :

	<nothing>

Notes :


*/

void
mcoccur_agint_add ARGLIST( (mcoccP, agentP, flags) )
   NFARG( MCOCCUR	*mcoccP )	/* mail client occurance */
   NFARG( AGENT		*agentP )	/* Agent occurance info */
    FARG( UWORD		flags )		/* Flags to apply */
{
	int		i;
	LL_NODE		*lnodeP;	/* Linked list node */
	IAGENT		*iaP;		/* Interested agent node */
	AGENT		*iagentP;	/* Agent in iaP */
	IMCLIENT	*imcP;		/* Interested mail client node */
	MCOCCUR		*imcoccP;	/* mc occurance node in imcP */

    /* NB: we can look through these linked lists since we have access to
       the structure; locking the list isn't necessary since anything that
       looks at the list will have to have access to the structure.
    */

    /* Link the agent interest into the mail occurance.  First find
       the spot (and make sure it isn't there already).
    */
    for ( lnodeP = mcoccP->mco_agents.llh_firstP; lnodeP != NULL;
		lnodeP = lnodeP->lln_nextP ) {
	iaP = (IAGENT *)lnodeP->lln_valP;
	iagentP = iaP->ia_agentP;
	if ( agentP->ag_infoP == iagentP->ag_infoP )
	    break;
    }
    if ( ( lnodeP == NULL ) || ( i < 0 ) ) {
	/* Insert new node here.  Make it first. */
	iaP = (IAGENT *)erualloc( "mcoccur_agint_add", "IAGENT",
					sizeof(IAGENT) );
	if ( iaP != NULL ) {
	    iaP->ia_flags = flags;
	    iaP->ia_agentP = agentP;

	    /* Add new node before lnodeP (or at end if NULL) */
	    ll_addvb( &mcoccP->mco_agents, lnodeP, iaP );
	}
    }


    /* Link the interested mail client into the agent occurance */
    i = 1;
    for ( lnodeP = agentP->ag_mclients.llh_firstP; lnodeP != NULL;
		lnodeP = lnodeP->lln_nextP ) {
	imcP = (IMCLIENT *)lnodeP->lln_valP;
	imcoccP = imcP->im_mcoccP;

	/* Silly comparison of node values */
	if ( imcoccP < mcoccP ) {
	    i = -1;
	    break;
	}
	if ( imcoccP == mcoccP ) {
	    i = 0;
	    break;
	}
    }
    if ( ( lnodeP == NULL ) || ( i < 0 ) ) {
	/* Insert new node here.  Make it, first. */
	imcP = (IMCLIENT *)erualloc( "mcoccur_agint_add", "IMCLIENT",
					sizeof(IMCLIENT) );
	if ( imcP != NULL ) {
	    imcP->im_mcoccP = mcoccP;

	    /* Add new node before lnodeP (or at end if NULL) */
	    ll_addvb( &agentP->ag_mclients, lnodeP, imcP );
	}
    }

    /* done */
}
/*

*//* mcoccur_agint_remove( mcoccP, agentP )

	Remove an agent interest in a mail client occurance

Accepts :

	mcoccP		Ptr to mail client occurance data
	agentP		Ptr to agent occurance data

Returns :

	<nothing>

Notes :

	This does *not* release or end or other way update the mcoccP
	  mail occurance info.  If any of that is needed, caller should
	  do it.

*/

void
mcoccur_agint_remove ARGLIST( (mcoccP, agentP) )
   NFARG( MCOCCUR	*mcoccP )	/* mail client occurance */
    FARG( AGENT		*agentP )	/* Agent occurance info */
{
	LL_NODE		*lnodeP;	/* Linked list node */
	IAGENT		*iaP;		/* Interested agent node */
	AGENT		*iagentP;	/* Agent in iaP */
	IMCLIENT	*imcP;		/* Interested mail client node */
	MCOCCUR		*imcoccP;	/* mc occurance node in imcP */

    /* NB: we can look through these linked lists since we have access to
       the structure; locking the list isn't necessary since anything that
       looks at the list will have to have access to the structure.
    */

    /* Find this agent in the mail occurance's agent interest list. */
    for ( lnodeP = mcoccP->mco_agents.llh_firstP; lnodeP != NULL;
		lnodeP = lnodeP->lln_nextP ) {
	iaP = (IAGENT *)lnodeP->lln_valP;
	iagentP = iaP->ia_agentP;
	if ( agentP->ag_infoP == iagentP->ag_infoP )
	    break;
    }
    if ( lnodeP != NULL ) {
	ll_rem( lnodeP );
	derualloc( iaP, sizeof(IAGENT) );
    }
    else
	warning( "agent not found in mail occurance" );

    /* Find this mail client in the agent's mail client list */
    for ( lnodeP = agentP->ag_mclients.llh_firstP; lnodeP != NULL;
		lnodeP = lnodeP->lln_nextP ) {
	imcP = (IMCLIENT *)lnodeP->lln_valP;
	imcoccP = imcP->im_mcoccP;

	/* Just compare the node values */
	if ( imcoccP == mcoccP )
	    break;
    }
    if ( lnodeP != NULL ) {
	ll_rem( lnodeP );
	derualloc( imcP, sizeof(IMCLIENT) );
    }
    else
	warning( "mail occurance not found in agent" );

    /* done */
}
/*

*//* mcoccur_agint_remove_all( agentP )

	Remove all mail occurance interests from an agent

Accepts :

	agentP		Ptr to agent occurance data

Returns :

	<nothing>

Notes :

	i.e., clean up for an agent going away.

	Unlike mcoccur_agint_remove (which removes a single link),
	  this function ends each mail occurance that it finds.  It's
	  assumed that this function is cleaning up the mail occurance
	  links for any agent that isn't doing it themselves, so it
	  will release the mail client occurances.

*/

void
mcoccur_agint_remove_all ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Agent occurance info */
{
	LL_NODE		*lnodeP;	/* Linked list node */
	IMCLIENT	*imcP;		/* Interested mail client node */
	MCOCCUR		*imcoccP;	/* mc occurance node in imcP */

    /* NB: we can look through these linked lists since we have access to
       the structure; locking the list isn't necessary since anything that
       looks at the list will have to have access to the structure.
    */

    /* Go through the agent's list of associated mail clients, and
       remove each one.
    */
    while ( ( lnodeP = agentP->ag_mclients.llh_firstP ) != NULL ) {
	imcP = (IMCLIENT *)lnodeP->lln_valP;
	imcoccP = imcP->im_mcoccP;

	/* Remove the single association between the agent and
	   this first mail client occurance.
	*/
	mcoccur_agint_remove( imcoccP, agentP );

	/* Access, end, and release the occurance */
	imcoccP = mcoccur_access( &imcoccP->mco_mcname );
	if ( imcoccP == NULL ) {
	    warning( "mcoccur_agint_release_all: can't access mail occurance" );
	}
	else {
	    mcoccur_end( imcoccP, NULL );
	    mcoccur_release( imcoccP );
	}
    }

    /* done */
}
/*

*//* mcoccur_begin( mcoccP, heldmcP )

	Process the beginning of a mail client occurance

Accepts :

	mcoccP		Ptr to occurance info
	heldmcP		Ptr to mail client db entry, if held

Returns :

	<nothing>

Notes :

	This function will need access to the corresponding mail client
	db entry.  If the caller has that access, it needs to be passed
	here, otherwise the attempt to gain the access in this function
	could deadlock.

*/

void
mcoccur_begin ARGLIST( (mcoccP, heldmcP) )
   NFARG( MCOCCUR	*mcoccP )	/* mail client occurance */
    FARG( MCLIENT	*heldmcP )	/* Mail client if held */
{
	BOOL		changedF;	/* TRUE if mclient needs update */
	int		n;
	int		d;
	long		lv;
	time_t		timenow;
	time_t		tdiff;
	MCLIENT		*mcP;

    time( &timenow );

    /* Do any needed fuddling with the mail client */
    if ( heldmcP != NULL )
	mcP = heldmcP;
    else {
	mcP = mcdb_mclient_access( &mcoccP->mco_mcname );
	if ( mcP == NULL ) {
	    warning( "mcoccur_access: can't access client." );
	    return;
	}
    }

    changedF = FALSE;			/* Presume no change */

    /* Do any simple state adjustment needed */
    switch ( mcP->mc_sstate ) {
	case MCS_NONE:			/* No specific state */
	    break;

	case MCS_GREY:			/* In greylisting period */
	    /* See if we've passed the graylist period */
	    if ( timenow >= mcP->mc_sstime ) {
		mcP->mc_sstate = MCS_NONE;
		changedF = TRUE;
	    }
	    break;

	case MCS_BLOCK_C_ADM:		/* Various block states */
	case MCS_BLOCK_C_AUTO:
	case MCS_BLOCK_G_ADM:
	case MCS_BLOCK_G_AUTO:
	case MCS_BLOCK_S_ADM:
	case MCS_BLOCK_S_AUTO:
	    /* See if we've passed the end-of-block time */
	    if ( mcP->mc_sstime != 0 ) {
		if ( timenow >= mcP->mc_sstime ) {
		    mcP->mc_sstate = MCS_NONE;
		    changedF = TRUE;
		}
	    }
	    break;

	default:
	    warning( "mcoccur_begin -- unknown mc_sstate %d", 
				mcP->mc_sstate );
	    break;
    }


    /* Do any override state adjustment needed */
    switch ( mcP->mc_ostate ) {
	case MCO_NONE:			/* No specific state */
	    break;

	case MCO_PASS:			/* Has a pass */
	    /* See if we've passed the end of the period */
	    if ( timenow >= mcP->mc_ostime ) {
		mcP->mc_ostate = MCO_NONE;
		changedF = TRUE;
	    }
	    break;

	default:
	    warning( "mcoccur_begin -- unknown mc_ostate %d", 
				mcP->mc_ostate );
	    break;
    }


    /* Check to see if we ran off the pre-greeting delay window. */
    if ( ( mcP->mc_pgd_sec > 0 ) && ( mcP->mc_pgd_wconC > 0 ) ) {
	/* Test against explicit or default connection and days limits. */
	n = ( mcP->mc_pgd_wcon > 0 ) ? mcP->mc_pgd_wcon : MCLIENT_PGD_WCON;
	d = ( mcP->mc_pgd_wday > 0 ) ? mcP->mc_pgd_wday : MCLIENT_PGD_WDAY;

	if ( ( mcP->mc_pgd_wconC > n ) ||
	     ( timenow > ( mcP->mc_pgd_wtime + ( d * 86400L ) ) ) ) {
	    mcP->mc_pgd_wconC = 0;
	    mcP->mc_pgd_wtime = 0;
	}
    }

    /* If there's any rcpt-to governor, may need to decay it */
    if ( ( mcP->mc_rcptgov != 0 ) &&
         ( mcP->mc_rgtime != 0 ) &&
	 ( Rgov_decay_int > 0 ) ) {

	/* Calculate decay period, and decay it. */
	tdiff = timenow - mcP->mc_rgtime;
	if ( tdiff > Rgov_decay_max )
	    tdiff = Rgov_decay_max;

	if ( tdiff > Rgov_decay_int ) {

#if 0	/* here's how we would iterate over it */
	    for ( ; tdiff > Rgov_decay_int; tdiff -= Rgov_decay_int ) {
		mcP->mc_rcptgov = (mcP->mc_rcpt_gov * Rgov_decay_mult)/
					Rgov_decay_div;
	    }
#else	/* And here's how we would simply calculate it.  Beware of overflow.
	   Note that the two methods will produce different results because
	   the integer truncation occurs at different points.
	*/
	    n = tdiff/Rgov_decay_int;
	    lv = mcP->mc_rcptgov;
	    lv *= (n * Rgov_decay_mult);
	    lv /= (n * Rgov_decay_div);
	    mcP->mc_rcptgov = lv;
#endif
	    /* Remember the last-decayed time */
	    mcP->mc_rgtime = timenow;
	}
    }

    /* Update client's seen counters and times */
    mcP->mc_lctime = timenow;
    if ( mcP->mc_fctime == 0 )
	mcP->mc_fctime = mcP->mc_lctime;
    ++mcP->mc_connC;
    changedF = TRUE;

    if ( changedF )
	/* Mark the client as changed so that the update flag will
	   be forced when it is released.
	*/
	mcdb_mclient_update( mcP );

    if ( heldmcP == NULL )
	mcdb_mclient_release( mcP, FALSE );


    /* done */
}
/*

*//* mcoccur_end( mcoccP, heldmcP )

	Process the end of a mail client occurance

Accepts :

	mcoccP		Ptr to occurance info
	heldmcP		Ptr to mail client db entry, if held

Returns :

	<nothing>

Notes :

	This function could conceivably need access to the corresponding
	mail client db entry.  If the caller has that access, it needs to be
	passed here, otherwise the attempt to gain the access in this
	function could deadlock.

*/

void
mcoccur_end ARGLIST( (mcoccP, heldmcP) )
   NFARG( MCOCCUR	*mcoccP )	/* mail client occurance */
    FARG( MCLIENT	*heldmcP )	/* Mail client if held */
{
    /* We don't really do anything here, but it is a complement
       to mcoccur_begin() where we might actually want to do something,
       so we maintain the calling protocol.
    */

    /* done */
}
/*

*//* mcoccur_init()

	Initialize the mcoccur module.

Accepts :


Returns :

	<value>		TRUE if successful;
			FALSE if not.

Notes :


*/

BOOL
mcoccur_init NOARGLIST
{
	time_t		timenow;

    /* Init the tree for auth IDs */
    Mcotree_authP = avlnew( NULL,
			   avl_mco_str_cmpkey,
			   avl_mco_str_mknode,
			   avl_mco_str_rmnode );
    if ( Mcotree_authP == NULL )
	return ( FALSE );

    /* Init the tree for IP addresses */
    Mcotree_ipP = avlnew( NULL,
			   avl_mco_ip4_cmpkey,
			   avl_mco_ip4_mknode,
			   avl_mco_ip4_rmnode );
    if ( Mcotree_ipP == NULL )
	return ( FALSE );

    /* Init the agent tree */
    Npagent_treeP = avlnew( NULL,
			   avl_npagent_cmpkey,
			   avl_npagent_mknode,
			   avl_npagent_rmnode );
    if ( Npagent_treeP == NULL )
	return ( FALSE );

    /* Init the np-agent prune list used */
    ll_init( &Npagent_prune_list );

    /* Shorten the np-agent timeout and  prune interval if debugging */
    if ( Debug > 1 ) {
	Npa_prune_interval = 30;
	Npa_timeout = 45;
    }

    /* Schedule the first pruning */
    time( &timenow );
    if ( !timer_add( timenow + Npa_prune_interval, npagent_prune, NULL ) ) {
	warning( "mcoccur_init: timer_add failed for npagent_prune." );
	return ( FALSE );
    }
    
    return ( TRUE );
}
/*

*//* mcoccur_notify( flags, fmtrtc, nbsP, mcoccP, agentP, mcP )

	Send a notice to agents interested in a mail client occurance

Accepts :

	flags		Flags to match interest flags
	fmtrtc		The routine to call to format the message
	nbsP		BSTR to store the message into
	mcoccP		Mail client occurance handle
	agentP		Mail agent occurance handle, or NULL
	mcP		Mail client db handle, or NULL

Returns :

	<nothing>

Notes :

	If the fmtrtc is NULL, the already-formatted string is assumed
	  to be in nbsP.  Otherwise fmtrtc will be called for each agent
	  being notified (as there may be agent-specific info to convey).
	  fmtrtc is called with:

	    (nbsP, mcoccP, iagentP, mcP)

	  and is expected to return TRUE on success, FALSE otherwise.
	  iagentP is the pointer to the agent occurance for the
	  specific agent being notified

	The caller is assumed to have access to a mail client occurance.

	Caller may pass an agent occurance ptr to force a notification to
	  that agent, whether or not the agent is on the mail client
	  occurance's interested agents list.

	If the caller has access to an agent occurance, it may be passed
	  here (as agentP) so that this function won't try to access it
	  and result in a deadlock.  Relatedly, if agentP is non-NULL,
	  it's assumed that it wants to be notified.

	If the caller has access to the relevant mail client database
	  entry, and that entry is needed to format the message, it
	  should be passed here so that the formatting function won't
	  need to try to access it and deadlock as a result.

*/

void
mcoccur_notify ARGLIST( (flags, fmtrtc, nbsP, mcoccP, agentP, mcP) )
   NFARG( UWORD		flags )		/* Interest flags */
   NFARG( BOOL		(*fmtrtc)() )	/* Formatting routine to call */
   NFARG( BSTR		*nbsP )		/* BSTR to use for the message */
   NFARG( MCOCCUR	*mcoccP )	/* Mail client occurance */
   NFARG( AGENT		*agentP )	/* Agent if any */
    FARG( MCLIENT	*mcP )		/* Mail client from mcdb, if any */
{
	BOOL		sendF;		/* whether we send to an agent */
	LL_NODE		*lnodeP;	/* Linked list node */
	IAGENT		*iaP;		/* Interested agent node */
	AGENT		*iagentP;	/* Agent in iaP */

    /* Look through the agents attached to the mail client occurance.
       Note that we don't have to lock the linked list since we have the
       occurance itself locked.

       Note also that we can point to and look at agent occurances that
       are connected to this node without gaining access to them; the
       agent occurances can not go away without being removed from this
       node, and that can not be done without gaining access to this
       node.
    */
    for ( lnodeP = mcoccP->mco_agents.llh_firstP; lnodeP != NULL;
		lnodeP = lnodeP->lln_nextP ) {
	iaP = (IAGENT *)lnodeP->lln_valP;
	iagentP = iaP->ia_agentP;

	if ( ( iaP->ia_flags & IA_NEXT ) != 0 ) {
	    /* Agent wants next message; clear the flag */
	    iaP->ia_flags &= ~(IA_NEXT);
	    sendF = TRUE;
	}
	else if ( ( iaP->ia_flags & flags ) != 0 ) {
	    sendF = TRUE;		/* some interest flags match. */
	}
	else if ( ( iaP->ia_flags & IA_ALL ) != 0 ) {
	    sendF = TRUE;		/* Agent wants all messages */
	}
	else
	    sendF = FALSE;

	/* Suppress sending message if the agent will get a forced
	   notification outside of this loop.
	*/
	if ( iagentP == agentP )
	    sendF = FALSE;

	/* Format if necessary */
	if ( sendF && ( fmtrtc != NULL ) )
		sendF = (*fmtrtc)(nbsP, mcoccP, iagentP, mcP);

	if ( sendF ) {
	    /* Send the message */
	    (*iagentP->ag_ctlP->ac_send)( iagentP,
					  nbsP->bs_bP, nbsP->bs_bC );
	}
    }

    /* Send a notification to specific agent if requested */
    if ( agentP != NULL ) {
	if ( ( fmtrtc == NULL ) ||
	     (*fmtrtc)(nbsP, mcoccP, iagentP, mcP) )

	    (*agentP->ag_ctlP->ac_send)( agentP,
					  nbsP->bs_bP, nbsP->bs_bC );
    }

    /* done */
}
/*

*//* mcoccur_release( mcoccP )

	Release mail client occurance block

Accepts :

	mcoccP		Ptr to the occurance block to release

Returns :

	<nothing>

Notes :

	Deletes the occurance data if no longer in use.

*/

void
mcoccur_release ARGLIST( (mcoccP) )
    FARG( MCOCCUR	*mcoccP )	/* Occurance block to release */
{
	MCONODE		*mconodeP;	/* The AVL node */
	AVLTREE		*treeP;		/* Tree it belongs in */
	UBYTE		mcname[50];

    mconodeP = (MCONODE *)mcoccP->mco_odbP;

    /* Remove the occurance data if it's ended */
    if ( mcoccP->mco_agents.llh_nodeC > 0 ) {
	/* Unlock */
	if ( pthread_mutex_unlock( &mconodeP->mon_lock ) != 0 )
	    warning( "mcoccur_release: can't release lock" );
    }
    else {
	/* Do a safety check to see that no agents are linked here.  This
	   would be some kind of bug if so.
	*/
	if ( mcoccP->mco_agents.llh_firstP != NULL ) {
	    warning( "mcoccur_release: zero count but has agents.  BUG!" );
	    if ( pthread_mutex_unlock( &mconodeP->mon_lock ) != 0 )
		warning( "mcoccur_release: can't release lock" );
	}
	else {
	    switch ( mcoccP->mco_mcname.mn_type ) {
		case MCT_AUTH:
		    treeP = Mcotree_authP;
		    break;

		case MCT_IP:
		    treeP = Mcotree_ipP;
		    break;

		default:
		    warning( "mcoccur_release: bad mn_type %d.  BUG!",
			    mcoccP->mco_mcname.mn_type );
		    if ( pthread_mutex_unlock( &mconodeP->mon_lock ) != 0 )
			warning( "mcoccur_release: can't release lock" );
		    return;
	    }

	    if ( Debug > 1 ) {
		if ( mcname_str( &mcoccP->mco_mcname, &mcname[0],
				 sizeof(mcname) ) >= sizeof(mcname) )
		    strcpy( (char *)&mcname[0], "<unknown>" );
		printf( "mcoccur_release: deleting %s\n", &mcname[0] );
	    }

	    /* This will destroy the lock (releasing it) */
	    if ( avldelete( treeP, mcoccP->mco_mcname.mn_nameP ) < 0 )
		warning( "mcoccur_release: avldelete failed." );
	}
    }

    /* done */
}
/*

*//* mcoccur_remove_all( mcoccP )

	Remove all interest connections from a mail client

Accepts :

	mcoccP		Ptr to the occurance block

Returns :

	<nothing>

Notes :

	Removes all interest agent links;

	sets occurance count to zero.

*/

void
mcoccur_remove_all ARGLIST( (mcoccP) )
    FARG( MCOCCUR	*mcoccP )	/* Occurance block to release */
{
	LL_NODE		*lnodeP;	/* Linked list node */
	IAGENT		*iaP;		/* Interested agent node */
	AGENT		*iagentP;	/* Agent in iaP */

    /* Remove all of the agents */
    while ( ( lnodeP = mcoccP->mco_agents.llh_firstP ) != NULL ) {
	iaP = (IAGENT *)lnodeP->lln_valP;
	iagentP = iaP->ia_agentP;

	mcoccur_agint_remove( mcoccP, iagentP );

	/* Make sure the agent still isn't first on the list, just a
	   dumb check but who knows.
	*/
	if ( lnodeP == mcoccP->mco_agents.llh_firstP ) {
	    warning( "mcoccur_remove_all: agent is still there.  BUG!" );
	    break;
	}
    }

    /* done */
}
/*

*//* mcoccur_stats( bsP )

	Print some brief stats into a buffer

Accepts :

	bsP		BSTR to print into

Returns :

	<nothing>

Notes :

	bsP is already prepped by caller (this function will likely
	  be appending to it)

	Status should be reasonably short and sweet, and formatted
	  as part of a STATS messsage.

*/

void
mcoccur_stats ARGLIST( (bsP) )
    FARG( BSTR		*bsP )		/* BSTR to print into */
{
    bstr_printf( bsP, "  Occ: npa %ld, auth %ld, ip %ld ;\n",
		Npagent_treeP->t_nodeC,
		Mcotree_authP->t_nodeC,
		Mcotree_ipP->t_nodeC );
}
/*

*//* npagent_access( actlP, infoP, infoL )

	Access a non-persistent agent occurance

Accepts :

	actlP		Ptr to agent control structure (for agent type)
	infoP		Ptr to agent info
	infoL		Lenth of agent info

Returns :

	<value>		ptr to AGENT structure if successful
			NULL if not

Notes :

	A non-persistent agent is one that does not have a long-lived
	  context in this program.  We need to cache information about
	  the agent so that we can continue to communicate with it.  An
	  np-agent may be heard from in different contexts, and the
	  owner of that context can not completely manage it.  This
	  would, e.g., correspond to a non-connected communications
	  channel (such as UDP datagrams).

	The returned AGENT structure is part of the AVL tree node.

	The agent info (infoP,infoL) represents both a unique identity for
	  the agent and enough information for the agent to be contacted.
	  It must not be dependant on its location: the info is copied
	  into the np-agent's AVL node and used from there.  When looking
	  up an np-agent, all bytes of this info are compared; a non-equal
	  comparison will mean a non-match.

	A np-agent is aged out if it hasn't been heard from in some
	  amount of time.  The act of accessing the agent will update
	  the heard-from time.  np-agent accesses via this routine will
	  also be counted and correspondingly un-counted when the
	  npagent_release function is called; any agent with an
	  outstanding count will not be aged out (deleted).  Caller
	  must be sure to release the np-agent.

	non-persistent agents need to be locked while they are being
	  manipulated, but should later be unlocked.  This function
	  returns the agent in a locked state.  A np-agent may have
	  multiple accesses against it, but only one lock at a time.

*/

AGENT *
npagent_access ARGLIST( (actlP, infoP, infoL) )
   NFARG( AGENT_CTL	*actlP )	/* Agent control info */
   NFARG( void		*infoP )	/* Agent identifying data */
    FARG( int		infoL )		/* Length of agent ID */
{
	NPANODE		*agnodeP;	/* Agent node */
	AGENT		akey;		/* Static agent key */

    /* Note that the tree lock will be retained until we are done,
       excluding other agent tree operations.
    */

    /* Build a key for the AVL access */
    akey.ag_ctlP = actlP;
    akey.ag_infoP = infoP;
    akey.ag_infoL = infoL;

    /* Try to find the entry */
    agnodeP = (NPANODE *)avlfind_nounlock( Npagent_treeP, (void *)&akey );
    if ( agnodeP == NULL ) {
	/* Can't find it, need to create it. */
	avlinsert_nolock( Npagent_treeP, (void *)&akey, (void *)NULL );
	agnodeP = (NPANODE *)avlfind_nolock( Npagent_treeP, (void *)&akey );
	if ( agnodeP == NULL ) {
	    /* Error.. */
	    if ( pthread_mutex_unlock( &Npagent_treeP->t_lock ) != 0 )
		warning( "npagent_access: can't unlock tree in failure case" );
	    return ( NULL );
	}
    }

    /* Update & return */
    npagent_lock( &agnodeP->npa_agent );

    ++agnodeP->npa_accessC;
    time( &agnodeP->npa_heard );
    
    if ( pthread_mutex_unlock( &Npagent_treeP->t_lock ) != 0 )
	warning( "npagent_access: can't unlock tree." );

    return ( &agnodeP->npa_agent );
}
/*

*//* npagent_lock( agentP )

	Acquire a lock on a non-persistent agent.

Accepts :

	agentP		Ptr to agent occurance block to lock

Returns :

	<value>		TRUE if OK
			FALSE if problem.

Notes :

	Locks the underlying AVL node.

	A non-persistent agent may need locking since it can be accessed
	  by multiple threads, as opposed to a persistent agent which
	  has only one thread owner.

*/

BOOL
npagent_lock ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Ptr to block to lock */
{
	int		r;
	NPANODE		*agnodeP;	/* The AVL node */

    agnodeP = (NPANODE *)agentP->ag_dbP;
    if ( ( r = pthread_mutex_lock( &agnodeP->npa_lock ) ) != 0 ) {
	warning( "npagent_lock: can't acquire lock, error %d.", r );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* npagent_release( agentP )

	Release access to non-persistent agent

Accepts :

	agentP		Ptr to agent occurance block to release

Returns :

	<nothing>

Notes :

	See npagent_access()

	It's assumed that the np-agent has been locked by the caller.
	  This function releases that lock.

*/

void
npagent_release ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Ptr to block to release */
{
	NPANODE		*agnodeP;	/* The AVL node */

    agnodeP = (NPANODE *)agentP->ag_dbP;

    if ( --agnodeP->npa_accessC < 0 ) {
	warning( "npagent_release: access count lt 0, programming error." );
	agnodeP->npa_accessC = 0;
    }

    npagent_unlock( agentP );

    /* done */
}
/*

*//* npagent_unlock( agentP )

	Release a lock on a non-persistent agent.

Accepts :

	agentP		Ptr to agent occurance block to unlock

Returns :

	<value>		TRUE if OK
			FALSE if problem.

Notes :

	Unlocks the underlying AVL node.

	A non-persistent agent may need locking since it can be accessed
	  by multiple threads, as opposed to a persistent agent which
	  has only one thread owner.

*/

BOOL
npagent_unlock ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Ptr to block to lock */
{
	int		r;
	NPANODE		*agnodeP;	/* The AVL node */

    agnodeP = (NPANODE *)agentP->ag_dbP;
    if ( ( r = pthread_mutex_unlock( &agnodeP->npa_lock ) ) != 0 ) {
	warning( "npagent_unlock: can't release lock, error %d.", r );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* pagent_create( actlP, infoP )

	Create a new persistent agent block

Accepts :

	actlP		Ptr to agent control structure (for agent type)
	infoP		The info associated with this specific agent.

Returns :

	<value>		ptr to AGENT structure if successful
			NULL if not

Notes :

	A persistent agent is one that stays around, i.e. one that we don't
	have to cache information for.  It only exists in a single context,
	and it can be completely managed by its owner, including removing it
	when it goes away.  This would, e.g., correspond to a connected
	communication channel (such as TCP).

	The agent info is something that is meaningful to the agent
	handler.

*/

AGENT *
pagent_create ARGLIST( (actlP, infoP) )
   NFARG( AGENT_CTL	*actlP )	/* Agent control info */
    FARG( void		*infoP )	/* Info for the specific agent */
{
	AGENT		*agentP;

    /* Allocate */
    agentP = (AGENT *)emalloc( "agent_create", "agent block", sizeof(AGENT) );
    if ( agentP == NULL )
	return ( NULL );

    /* Fill it in */
    agentP->ag_ctlP = actlP;		/* Ptr to agent control */
    agentP->ag_infoP = infoP;		/* Specific info */
    ll_init( &agentP->ag_mclients );	/* List of interested mail clients */

    return ( agentP );
}
/*

*//* pagent_delete( agentP )

	Deletes a persistent agent.

Accepts :

	agentP		Ptr to agent block to release

Returns :

	<nothing>

Notes :

	Calls the agent control's ac_delinfo callback which may elect
	  to remove any associated info (ag_infoP).

	Other than the above, does not do any related agent cleanup.
	Caller must ensure that other agent cleanups have been done.

*/

void
pagent_delete ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Ptr to block to release */
{
    /* Use the control callback function to clean up any info */
    (*agentP->ag_ctlP->ac_delinfo)( agentP );

    /* Finalize the interested mail clients list */
    ll_finit( &agentP->ag_mclients );

    /* Delete the agent */
    dealloc( agentP );
}
/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* avl_mco_ip4_cmpkey( keyP, nodeP )

	Compare mail client keys for mail avl tree routines,
	  for a mail client occurance tree where
	  the client identifier is an IP address

Accepts :

	keyP		Ptr to the subject key (an MCNAME_IP4 value)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:


*/


static int
avl_mco_ip4_cmpkey( keyP, nodeP )
	MCNAME_IP4	*keyP;		/* Key to look for */
	MCONODE		*nodeP;		/* Node to compare against */
{
	MCNAME_IP4	*nkeyP;		/* The key in the node */

    /* Compare the IP4 address/mask data */
    nkeyP = (MCNAME_IP4 *)nodeP->mon_occ.mco_mcname.mn_nameP;
    return ( mcname_ip4_cmp( keyP, nkeyP ) );
}
/*

*//* avl_mco_ip4_mknode( treeP, keyP, dataP, enodeP )

	Make a tree node on behalf of AVL tree routines,
	  for a mail client occurance tree where the client identifier is an
	  IP address

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (an MCNAME_IP4 value)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
avl_mco_ip4_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	MCNAME_IP4	*keyP;		/* Key for the new node (IP) */
	void		*dataP;		/* Addr of new data (n/a) */
	MCONODE		*enodeP;	/* Existing node... */
{
	int		nodesize;	/* Size of node required */
	MCONODE		*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the name/key) */
    nodesize = sizeof(MCONODE) + sizeof( MCNAME_IP4 );
    nodeP = (MCONODE *)emalloc( "avl_mco_ip4_mknode", "avl node",
					 nodesize );

    /* Setup the node struct */
    nodeP->mon_occ.mco_mcname.mn_nameP = (void *)&nodeP[1];
    *(MCNAME_IP4 *)nodeP->mon_occ.mco_mcname.mn_nameP = *keyP;
    nodeP->mon_occ.mco_mcname.mn_type = MCT_IP;

    avl_mconode_init( nodeP );		/* Finish initializing the node */

    return( (AVLNODE *)nodeP );
}
/*

*//* avl_mco_ip4_rmnode( treeP, nodeP )

	Delete a mail client avl node on behalf of AVL tree routines,
	  for a mail client occurance tree where
	  the client identifier is an IP address

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :

*/

static int
avl_mco_ip4_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	MCONODE		*nodeP;
{
    if ( nodeP->mon_occ.mco_mcname.mn_type != MCT_IP )
	warning( "avl_mco_ip4_rmnode: node type should be %d, but is %d",
		  MCT_IP, nodeP->mon_occ.mco_mcname.mn_type );

    if ( nodeP->mon_occ.mco_agents.llh_nodeC != 0 )
	warning( "avl_mco_ip4_rmnode: occurance count should be 0, but is %d",
		nodeP->mon_occ.mco_agents.llh_nodeC );

    /* Stupid pthreads, makes you unlock before you can destroy.
       this sequence ought to be ok, not to mention atomic.
    */
    if ( pthread_mutex_unlock( &nodeP->mon_lock ) != 0 )
	warning( "avl_mco_ip4_rmnode can't release lock." );

    if ( pthread_mutex_destroy( &nodeP->mon_lock ) != 0 )
	warning( "avl_mco_ip4_rmnode can not destroy lock." );

    /* Finalize interested agents list */
    ll_finit( &nodeP->mon_occ.mco_agents );

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* avl_mco_str_cmpkey( keyP, nodeP )

	Compare mail client keys for avl tree routines,
	  for a mail client occurance tree where
	  the client identifier is a simple string.

Accepts :

	keyP		Ptr to the subject key (a string)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:


*/


static int
avl_mco_str_cmpkey( keyP, nodeP )
	char		*keyP;		/* Key to look for */
	MCONODE		*nodeP;		/* Node to compare against */
{
	int		cmpval;

    /* Do case insensitive string compare */
    cmpval = stricmp( keyP, (char *)nodeP->mon_occ.mco_mcname.mn_nameP );

    return ( cmpval );
}
/*

*//* avl_mco_str_mknode( treeP, keyP, dataP, enodeP )

	Make a mail client avl node on behalf of AVL tree routines,
	  for a mail client occurance tree where
	  the client identifier is a simple string.

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (a string)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
avl_mco_str_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	char		*keyP;		/* Key for the new node (string) */
	void		*dataP;		/* Addr of new data (n/a) */
	MCONODE		*enodeP;	/* Existing node... */
{
	int		nodesize;	/* Size of node required */
	MCONODE		*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the name/key) */
    nodesize = sizeof(MCONODE) + strlen( keyP ) + 1;
    nodeP = (MCONODE *)emalloc( "avl_mco_str_mknode", "avl node",
					 nodesize );

    /* Setup the node struct */
    nodeP->mon_occ.mco_mcname.mn_nameP = (void *)&nodeP[1];
    strcpy( (char *)nodeP->mon_occ.mco_mcname.mn_nameP, keyP );

    avl_mconode_init( nodeP );		/* Finish initializing the node */

    return( (AVLNODE *)nodeP );
}
/*

*//* avl_mco_str_rmnode( treeP, nodeP )

	Delete a mail client avl node on behalf of AVL tree routines,
	  for a mail client occurance tree where
	  the client identifier is a simple string

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :


*/

static int
avl_mco_str_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	MCONODE		*nodeP;
{
    if ( nodeP->mon_occ.mco_mcname.mn_type != MCT_AUTH )
	warning( "avl_mco_str_rmnode: node type should be %d, but is %d",
		  MCT_IP, nodeP->mon_occ.mco_mcname.mn_type );

    if ( nodeP->mon_occ.mco_agents.llh_nodeC != 0 )
	warning( "avl_mco_str_rmnode: occurance count should be 0, but is %d",
		nodeP->mon_occ.mco_agents.llh_nodeC );

    /* Stupid pthreads. */
    if ( pthread_mutex_unlock( &nodeP->mon_lock ) != 0 )
	warning( "avl_mco_str_rmnode can't release lock." );
    if ( pthread_mutex_destroy( &nodeP->mon_lock ) != 0 )
	warning( "avl_mco_str_rmnode can not destroy lock." );

    /* Finalize interested agents list */
    ll_finit( &nodeP->mon_occ.mco_agents );

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* avl_mconode_init( nodeP )

	Initialize the non-specific part of an avl node in a
	  mail client occurance tree.

Accepts :

	nodeP		Ptr to MCONODE to finish initializing

Returns :

	<nothing>

Notes :


*/

static void
avl_mconode_init ARGLIST( (nodeP) )
    FARG( MCONODE	*nodeP )	/* Node to init */
{
    if ( pthread_mutex_init( &nodeP->mon_lock, NULL ) != 0 )
	warning( "avl_mconode_init: can't init lock" );

    /* Provide a hook back to the AVL node */
    nodeP->mon_occ.mco_odbP = (void *)nodeP;

    ll_init( &nodeP->mon_occ.mco_agents );	/* Interested agents list */
}
/*

*//* avl_npagent_cmpkey( keyP, nodeP )

	Compare mail client keys for mail avl tree routines,
	  for np-agent tree.

Accepts :

	keyP		Ptr to the subject key (an AGENT value)
	nodeP		Ptr to existing node to test key against

Returns :

	< value >	-1, 0, or 1, as dictated by the rules of
			interface between the avl tree routines and
			the key comparison routine.

Notes:

	The only significant part(s) of the agent for comparison purposes
	are:
	   - the agent control pointer
	   - the agent info pointer.
	   - the agent info length

*/


static int
avl_npagent_cmpkey( keyP, nodeP )
	AGENT		*keyP;		/* Key to look for */
	NPANODE		*nodeP;		/* Node to compare against */
{
	int		cmpval;
	AGENT		*nkeyP;		/* The key in the node */

    nkeyP = &nodeP->npa_agent;

    /* Agents must have the same control */
    if ( keyP->ag_ctlP != nkeyP->ag_ctlP )
	return ( -1 );

    /* Next sort is on info length */
    if ( keyP->ag_infoL < nkeyP->ag_infoL )
	return ( -1 );
    if ( keyP->ag_infoL > nkeyP->ag_infoL )
	return ( 1 );

    /* Next sort is on info data */
    cmpval = memcmp( keyP->ag_infoP, nkeyP->ag_infoP, keyP->ag_infoL );
    if ( cmpval == 0 )
	return ( 0 );
    if ( cmpval < 0 )
	return ( -1 );

    return ( 1 );
}
/*

*//* avl_npagent_mknode( treeP, keyP, dataP, enodeP )

	Make a tree node on behalf of AVL tree routines,
	  for np-agent tree.

Accepts :

	treeP		Address of tree header block
	keyP		The key for the new node (an AGENT value)
	dataP		Address of data (not used)
	enodeP		Address of any existing node

Returns :

	<value>		Address of the new node

Notes :


*/

static AVLNODE *
avl_npagent_mknode( treeP, keyP, dataP, enodeP )
	AVLTREE		*treeP;		/* Addr of tree header block */
	AGENT		*keyP;		/* Key for the new node (IP) */
	void		*dataP;		/* Addr of new data (n/a) */
	NPANODE		*enodeP;	/* Existing node... */
{
	int		nodesize;	/* Size of node required */
	NPANODE		*nodeP;		/* New node that we make */

    /* If enodeP is non-NULL, this is a call to overwrite the data.
       That should never happen, so return NULL in that case. */
    if ( enodeP != NULL )
	return( (AVLNODE *)NULL );

    /* Calculate new node size (we co-allocate space for the id) */
    nodesize = sizeof(NPANODE) + keyP->ag_infoL;
    nodeP = (NPANODE *)emalloc( "avl_npagent_mknode", "avl node",
					 nodesize );

    /* Setup the node struct */
    if ( pthread_mutex_init( &nodeP->npa_lock, NULL ) != 0 )
	warning( "avl_npagent_mknode: can't init lock" );
    nodeP->npa_accessC = 0;
    nodeP->npa_heard = 0L;

    /* Initialize the agent within the node */
    nodeP->npa_agent.ag_ctlP = keyP->ag_ctlP;
    nodeP->npa_agent.ag_infoL = keyP->ag_infoL;
    nodeP->npa_agent.ag_infoP = (void *)&nodeP[1];
    memcpy( nodeP->npa_agent.ag_infoP, keyP->ag_infoP, keyP->ag_infoL );
    ll_init( &nodeP->npa_agent.ag_mclients );

    /* Set up reference pointer back to the AVL node */
    nodeP->npa_agent.ag_dbP = nodeP;

    return( (AVLNODE *)nodeP );
}
/*

*//* avl_npagent_rmnode( treeP, nodeP )

	Delete a mail client avl node on behalf of AVL tree routines,
	  for np-agent tree.

Accepts :

	treeP		Address of tree header block
	nodeP		Address of the node being removed

Returns :

	<value>			0 for success, -1 for failure

Notes :

	The thing deleting this node should have locked it first.

*/

static int
avl_npagent_rmnode( treeP, nodeP )
	AVLTREE		*treeP;
	NPANODE		*nodeP;
{
    if ( nodeP->npa_agent.ag_mclients.llh_firstP != NULL )
	warning( "avl_npagent_rmnode: mclient list not empty" );

    if ( nodeP->npa_accessC != 0 )
	warning( "avl_npagent_rmnode: access count should be 0, but is %d",
		nodeP->npa_accessC );

    /* Finalize the interested agents list */
    ll_finit( &nodeP->npa_agent.ag_mclients );

    /* Remove the node lock.  Dumb pthreads (have I said this before?)
       doesn't let you destroy & unlock in one operation.  (or if it
       does, it's me that's dumb for not knowing how.)
    */
    if ( pthread_mutex_unlock( &nodeP->npa_lock ) != 0 )
	warning( "avl_npagent_rmnode can't release lock." );
    if ( pthread_mutex_destroy( &nodeP->npa_lock ) != 0 )
	warning( "avl_npagent_rmnode can not destroy lock." );

    /* Deallocate the node */
    dealloc( nodeP );

    /* Return OK */
    return( 0 );
}
/*

*//* npagent_prune( argP )

	Prune the np-agent tree

Accepts :

	argP		Dummy argument

Returns :

	<nothing>

Notes :

	This is called via a timer function.  The timer function passes
	an opaque argument, so this function is defined to accept one.

	Reschedules itself.

*/

static void
npagent_prune ARGLIST( (argP) )
    FARG( void		*argP )		/* Dummy argument */
{
	time_t		timenow;
	LL_NODE		*lnodeP;	/* Linked list node */
	AGENT		*agentP;
	char		orig[100];

    /* Note that the tree lock will be kept during this operation;
       this will prevent other npagent tree manipulations until we're
       done.
    */

    if ( Debug > 1 )
	printf( "npagent_prune here.\n" );

    /* Need to walk the np-agent tree to find np-agents to delete.
       Can't modify the tree while walking it, so the agents will be
       added to the prune linked list, from which we will later delete
       them one at a time.  We'll pass the current time to the walk
       function, so it doesn't have to do a system call for every node.
    */
    time( &timenow );
    Npagent_treeP->t_walkP = (void *)&timenow;
    avlwalk_nounlock( Npagent_treeP, npagent_prune_node );

    /* Now remove each agent that's on the prune list.  We can use
       the agent as a key to avldelete, since the key is only used
       for comparison prior to deletion.
    */
    while ( ( lnodeP = Npagent_prune_list.llh_firstP ) != NULL ) {
	agentP = (AGENT *)lnodeP->lln_valP;

	if ( Debug > 0 ) {
	    if ( !(*agentP->ag_ctlP->ac_origstr)(agentP,
						 &orig[0], sizeof(orig) ) )
		strcpy( &orig[0], "<unknown>" );
	    printf( "  pruning agent: %s\n", &orig[0] );
	}

	/* Unlink this agent from any mail client occurances */
	mcoccur_agint_remove_all( agentP );

	/* Call the agent delinfo function to clean up any attached
	   info that the agent handler might know about.
	*/
	(*agentP->ag_ctlP->ac_delinfo)( agentP );

	/* Remove this agent block from the tree (and thus from existence).
	   Remember that the tree is still locked here.
	*/
	if ( avldelete_nolock( Npagent_treeP, agentP ) < 0 )
	    warning( "npagent_prune: avldelete failure." );

	/* And remove the node... since we already have the np-agent
	   stuff locked, no need to lock the list
	*/
	ll_rem_nolock( lnodeP );
    }

    /* Reschedule the prune function.. */
    time( &timenow );
    if ( !timer_add( timenow + Npa_prune_interval, npagent_prune, NULL ) ) {
	warning( "npagent_prune: timer_add failed for npagent_prune." );
    }

    /* OK to unlock the tree. */
    if ( pthread_mutex_unlock( &Npagent_treeP->t_lock ) != 0 )
	warning( "npagent_prune: can't release Npagent_treeP lock." );

    if ( Debug > 1 )
	printf( "npagent_prune done.\n" );

    /* done */
}
/*

*//* npagent_prune_node( treeP, nodeP, depth )

	Write a mail client node on behalf of AVL walk 

Accepts :

	treeP		Ptr to AVL tree
	nodeP		Ptr to the node to write
	depth		depth of the node

Returns :

	<value>		0 to continue
			non-zero to stop

Notes :

	Per avlwalk() interface; see that.

	Each np-agent is locked as it is added to the to-delete list.
	  The deletion function will unlock it.

*/

static int
npagent_prune_node ARGLIST( (treeP, nodeP, depth) )
   NFARG( AVLTREE	*treeP )	/* Ptr to tree */
   NFARG( NPANODE	*nodeP )	/* Ptr to node */
    FARG( int		depth )		/* Depth of node */
{
	int		r;
	time_t		timenow;

    /* NULL node signals end of walk */
    if ( nodeP == NULL )
	return ( 0 );			/* no special processing */

    /* We don't have to lock the node to look at it: the only possible
       contention is via a release, since an access can't happen while
       the tree is locked (as it is here).  If a release happens while
       we are looking at the node, it means we might miss pruning it
       until later, but otherwise no harm.
    */

    /* Skip this node if it has been heard from recently
       or if it has any accesses against it.
    */
    timenow = *(time_t *)Npagent_treeP->t_walkP;
    if ( ( nodeP->npa_heard >= (timenow - Npa_timeout) ) ||
         ( nodeP->npa_accessC > 0 ) )
	return ( 0 );

    /* We'll want to delete this node.  Add it to the prune list,
       locking it first.
    */
    if ( ( r = pthread_mutex_lock( &nodeP->npa_lock ) ) != 0 ) {
	warning( "npagent_prune_node: can't lock node, error %d", r );

	/* Return without pruning, which means this will probably keep
	   happening.  Which is probably good, since if it does happen
	   it's a bug that needs to be fixed.
	*/
	return ( 0 );
    }

    ll_addvt( &Npagent_prune_list, &nodeP->npa_agent );

    return ( 0 );
}
