/* mvmcas-command.c -- command processing for mvmcas

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

This module provides the command processing for mvmcas.

Command processing is done on behalf of the individual server modules
(tcp, udp, what-have-you).

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdarg.h>
#include <ctype.h>
#include <pthread.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmfmisc.h"

#include "mvmcas.h"
#include "mvmcas-version.h"
#include "mvmcas-misc.h"
#include "mcdb.h"
#include "mcoccur.h"
#include "mvmcas-command.h"


/* Local definitions */

    /* Command table entry */
typedef struct {
    char	*cm_nameP;		/* Verb */
    STSCODE	(*cm_rtc)();		/* Routine to call */
}  CMTE;


/* External data referenced */

extern	int	Debug;		/* Debugging level */
extern	int	Rgov_bad_incr;		/* Bad rgov increment */
extern	int	Rgov_bad_limit;		/* Bad rgov limit */
extern	int	Rgov_good_incr;		/* Good rgov increment */
extern	int	Rgov_good_limit;	/* Good rgov limit */
extern	struct timeval Starttime;	/* Wall clock when we start */
extern	long	Tcp_connC;		/* TCP connections */
extern	long	Udp_connC;		/* UDP connections */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Command processing routines */
static	STSCODE	cmd_done();		/* "done" or "quit" */
static	STSCODE	cmd_hi();		/* "hi" */
static	STSCODE	cmd_mclient_change();	/* "mclient-change" */
static	STSCODE	cmd_mclient_delete();	/* "mclient-delete" */
static	STSCODE	cmd_mclient_info();	/* "mclient-info" */
static	STSCODE	cmd_mclient_replace();	/* "mclient-replace" */

#ifdef	MML_MALLOC_DEBUGGING
static	STSCODE	cmd_memdump();		/* "memdump" */
#endif	/* MML_MALLOC_DEBUGGING */

static	STSCODE	cmd_start();		/* "start" */
static	STSCODE	cmd_stats();		/* "stats" */
static	STSCODE	cmd_stop();		/* "stop" */
static	STSCODE	cmd_sync();		/* "sync" */
static	STSCODE	cmd_update();		/* "update" */

    /* Agent notification formatting function(s) */
static	BOOL	agnotfmt_info PROTO( (BSTR *bsP, MCOCCUR *mcoccP,
				      AGENT *agentP, MCLIENT *heldmcP) );
static	BOOL	agnotfmt_info_x PROTO( (BSTR *bsP, MCOCCUR *mcoccP,
				      AGENT *agentP, MCLIENT *heldmcP) );
static	BOOL	agnotfmt_stats PROTO( (BSTR *bsP, MCOCCUR *mcoccP,
				       AGENT *agentP, MCLIENT *heldmcP) );

    /* Get a command block */
static	MCCMD *mccmd_get PROTO( (void) );

    /* Process command elements to set mclient info */
static	BOOL	mccmd_mclient_set PROTO( (MCCMD *mccP, MCLIENT *mcP,
					  BOOL initF) );

    /* Parse mail client name, and one for each kind of value */
static	BOOL	mccmd_parse_mcname PROTO( (MCCMD *mccP, MCNAME_S *mcnmsP) );
static	BOOL	mccmd_parse_mcname_bs PROTO( (MCCMD *mccP, MCNAME_S *mcnmsP) );
static	BOOL	mccmd_parse_mcname_ip PROTO( (MCCMD *mccP, MCNAME_S *mcnmsP) );


    /* Start processing a command */
static	CMTE	*mccmd_preamble PROTO( (MCCMD *mccP, UBYTE *cmdbufP) );

    /* Release a client block */
static	void	mccmd_release PROTO( (MCCMD *mccP) );

    /* Return a status via command block */
static	void	mccmd_return PROTO( (MCCMD *mccP, char *fmtP, ...) );

    /* Return an extended INFO message for a mail client */
static	STSCODE	mccmd_return_info_x PROTO( (MCCMD *mccP, MCLIENT *mcP) );

    /* Get a token from command buffer, with error processing */
static	BOOL	mccmd_token PROTO ( (MCCMD *mccP, UBYTE *tokP, int tokL) );

    /* Get a token from command buffer, interpret as int */
static	BOOL	mccmd_token_int PROTO( (MCCMD *mccP, long *longP) );

    /* Get a token from command buffer, interpret as unsigned int */
static	BOOL	mccmd_token_uint PROTO( (MCCMD *mccP, ULONG *longP) );

    /* Process an update sequence */
static	STSCODE	mccmd_update PROTO( (MCCMD *mccP, BOOL stopF) );

/* Private data */

static	MCCMD	*Mccmd_listP;	/* List of free command blocks */
static	pthread_mutex_t	Mccmd_lock = PTHREAD_MUTEX_INITIALIZER;

    /* Command table */
static	CMTE	Cmdtbl[] = {
/*	name			rtc			*/
{	"done",			cmd_done		},
{	"hi",			cmd_hi			},
{	"mclient-change",	cmd_mclient_change	},
{	"mclient-delete",	cmd_mclient_delete	},
{	"mclient-info",		cmd_mclient_info	},
{	"mclient-replace",	cmd_mclient_replace	},

#ifdef	MML_MALLOC_DEBUGGING
{	"memdump",		cmd_memdump		},
#endif	/* MML_MALLOC_DEBUGGING */

{	"quit",			cmd_done		},
{	"start",		cmd_start		},
{	"stats",		cmd_stats		},
{	"stop",			cmd_stop		},
{	"sync",			cmd_sync		},
{	"update",		cmd_update		},

{	NULL,			NULL			}
};


/*

*//* mvmcas_command( mccP, cmdbufP )

	Process a command

Accepts :

	mccP		Command context block
	cmdbufP		Ptr to buffer with command string to process

Returns :

	<value>		SCxxx status code


Notes :

	If there is any immediate response to be returned to the
	initiator (agent), it is stored in the response string in
	the command block.  Some commands do not have immediate
	responses but simply trigger one or more later asynchronous
	messages.

*/

STSCODE
mvmcas_command ARGLIST( (mccP, cmdbufP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( UBYTE		*cmdbufP )	/* Command buffer */
{
	STSCODE		status;
	CMTE		*cmteP;		/* Command table entry */

    if ( Debug > 0 )
	printf( "mvmcas_command: %s\n", cmdbufP );

    cmteP = mccmd_preamble( mccP, cmdbufP );
    if ( cmteP == NULL ) {
	status = SCOK;
    }
    else {
	/* Hand off to the dispatch routine */
	status = (*cmteP->cm_rtc)( mccP );
    }

    /* If there's a direct response, return it to the agent */
    if ( mccP->mc_rstrP->bs_bC > 0 ) {
	(*mccP->mc_agentP->ag_ctlP->ac_send)( mccP->mc_agentP,
					    mccP->mc_rstrP->bs_bP,
					    mccP->mc_rstrP->bs_bC );
    }

    /* Translate benign status(es) to OK */
    if ( status == SCNOTDONE )
	status = SCOK;

    return ( status );
}
/*

*//* mvmcas_command_context_begin( agentP )

	Begin a command processing context

Accepts :

	agentP		Ptr to the agent data

Returns :

	<value>		Ptr to MCCMD command context structure
			NULL if failure

Notes :


*/

MCCMD *
mvmcas_command_context_begin ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* Ptr to the agent block */
{
	MCCMD		*mccP;		/* Command block */

    /* Get command block for this command sequence */
    mccP = mccmd_get();

    /* Fill in */
    mccP->mc_agentP = agentP;

    return ( mccP );
}
/*

*//* mvmcas_command_context_end( mccP )

	End a command processing context

Accepts :

	mccP		Command context block

Returns :

	<nothing>

Notes :


*/

void
mvmcas_command_context_end ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
    /* Release the command block for another use. */
    mccmd_release( mccP );
}



/***********************************************************************
 *                                                                     *
 *                     Command processing routines                     *
 *                                                                     *
 ***********************************************************************/

/* Following are individual command processing functions.  They are
   defined as:

	STSCODE cmd_xxx( mccP )

    Each accepts:

	mccP		The pointer to the command context block

    and returns

	<value>		STSCODE result code.


    Much of the returned info is via the mccP.

*/


/*************************************************************
 * cmd_done -- the "done" or "quit" command.

	done (or quit)

    Says that the agent is finished.

*/
static STSCODE
cmd_done ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
    mccmd_return( mccP, "TEXT - see you later." );

    /* Remove all connections to mail clients. */
    mcoccur_agint_remove_all( mccP->mc_agentP );

    /* Note: if this is not a stateful agent, the agent will be removed
       at some later time by the pruning timed event.
    */

    return ( SCDONE );
}
/*************************************************************/



/*************************************************************
 * cmd_hi -- the "hi" command.

	hi

    Says "hi" back; performs no other function.

*/
static STSCODE
cmd_hi ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	char		*rP;

    rP = solbat( (char *)mccP->mc_cmdP, TRUE );
    if ( *rP == NUL )
	mccmd_return( mccP, "HI - yourself." );
    else
	mccmd_return( mccP, "HI - to your \"%s\"", rP );

    return ( SCOK );
}
/*************************************************************/



/*************************************************************
 * cmd_mclient_change -- the "mclient-change" command.

	mclient-change mcname ; args ; END

    Changes specific info for a mail client.

    Returns an INFO message with the new client data.

*/
static STSCODE
cmd_mclient_change ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;
	BOOL		okF;
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */
	MCOCCUR		*mcoccP;	/* Occurance entry */

    /* Get mail client name */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    status = SCOK;

    /* Update the mclient from the command arguments */
    okF = mccmd_mclient_set( mccP, mcP, FALSE );

    /* If OK, return an INFO message */
    if ( okF )
	status = mccmd_return_info_x( mccP, mcP );

    /* Release client, update depends on success */
    mcdb_mclient_release( mcP, okF );

    return ( status );
}
/*************************************************************/



/*************************************************************
 * cmd_mclient_delete -- the "mclient-delete" command.

	mclient-delete mcname

    Delete the specified mail client from memory

*/
static STSCODE
cmd_mclient_delete ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;
	BOOL		cmpF;
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */
	MCOCCUR		*mcoccP;	/* Occurance entry */

    /* Get mail client name */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    /* Only delete if the name is the same.  e.g. if the mclient is
       an aggregate when we asked for a specific, we shouldn't delete.
    */
    switch ( mcns.mns_mcname.mn_type ) {
	case MCT_AUTH:
	    cmpF = ( stricmp( (char *)mcns.mns_mcname.mn_nameP,
			      (char *)mcP->mc_mcname.mn_nameP ) == 0 );
	    break;

	case MCT_IP:
	    cmpF = ( mcname_ip4_cmp( (MCNAME_IP4 *)mcns.mns_mcname.mn_nameP,
				     (MCNAME_IP4 *)mcP->mc_mcname.mn_nameP )
					   == 0 );
	    break;

	case MCT_IPv6:
	default:
	    warning( "cmd_mclient_delete: bad case.  BUG!" );
	    cmpF = FALSE;
	    break;
    }
    if ( !cmpF ) {
	mccmd_return( mccP, "RQERR - must delete exact target." );
	status = SCNOTDONE;
    }
    else {
	mcoccP = mcoccur_access( &mcP->mc_mcname );
	if ( mcoccP == NULL ) {
	    mccmd_return( mccP, "ERROR - can't get mail client occurance." );
	    status = SCNOTDONE;
	}
	else {
	    /* Remove all agent occurances */
	    mcoccur_remove_all( mcoccP );

	    mcoccur_release( mcoccP );

	    status = SCOK;
	}
    }

    if ( status == SCOK ) {
	if ( mcdb_mclient_delete( mcP ) )
	    mccmd_return( mccP, "OK" );
	else
	    mccmd_return( mccP, "ERROR -- deletion error." );
    }
    else
	mcdb_mclient_release( mcP, FALSE );

    return ( status );
}
/*************************************************************/



/*************************************************************
 * cmd_mclient_info -- the "mclient-info" command.

	mclient-info mcname

    Asks for info about a mail client.  Triggers an INFO message
      response.

    Similar to "START" except:
      - No restrictions on mail client (e.g., can be an aggregate);
      - no side-effects (no update of occurances, etc);
      - INFO message is only given to the sending agent, not to all
        interested agents.

*/
static STSCODE
cmd_mclient_info ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */

    /* Get mail client name */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    /* Format and return the info message */
    status = mccmd_return_info_x( mccP, mcP );

    mcdb_mclient_release( mcP, FALSE );

    return ( status );
}
/*************************************************************/



/*************************************************************
 * cmd_mclient_replace -- the "mclient-replace" command.

	mclient-replace mcname ; args ; END

    Replaces mclient, and updates with new info.

    Returns an INFO message with the new client data.

*/
static STSCODE
cmd_mclient_replace ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;
	BOOL		okF;
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */

    /* Get mail client name */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    status = SCOK;

    /* Update the mclient from the command arguments */
    okF = mccmd_mclient_set( mccP, mcP, TRUE );

    /* If OK, return an INFO message */
    if ( okF )
	status = mccmd_return_info_x( mccP, mcP );

    /* Release client, update depends on success */
    mcdb_mclient_release( mcP, okF );

    return ( status );
}
/*************************************************************/



#ifdef	MML_MALLOC_DEBUGGING
/*************************************************************
 * cmd_memdump -- the "memdump" command.

	memdump

    Does a memory debugging dump

*/
static STSCODE
cmd_memdump ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	FILE		*ofP;

    ofP = fopen( "/var/tmp/mvmcas.memdump", "w" );
    if ( ofP == NULL ) {
	mccmd_return( mccP, "ERROR -- error opening file." );
    }
    else {
	if ( !mml_malloc_dump( ofP ) )
	    mccmd_return( mccP, "ERROR -- dump returned error." );
	else
	    mccmd_return( mccP, "OK" );
	fclose( ofP );
    }

    return ( SCOK );
}
/*************************************************************/
#endif	/* MML_MALLOC_DEBUGGING */



/*************************************************************
 * cmd_start -- the "start" command.

	start <mcname> <etc>

    Begins an occurance of the mail client.

*/
static STSCODE
cmd_start ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;
	UWORD		flags;
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */
	MCOCCUR		*mcoccP;	/* Occurance entry */

    /* Get mail client name that is starting */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    /* Do some checks */
    if ( mcns.mns_mcname.mn_type == MCT_IP ) {
	if ( mcns.mns_data.nm_ip4.nip_masklen != 32 ) {
	    mccmd_return( mccP, "RQERR - IP masklen must be 32 for \"start\"" );
	    return ( SCNOTDONE );
	}
    }

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    status = SCOK;

    /* Get occurance data for the mclient's name.  Note that this
       may be different than the input name, e.g. for an aggregate.
    */
    mcoccP = mcoccur_access( &mcP->mc_mcname );
    if ( mcoccP == NULL ) {
	mccmd_return( mccP, "ERROR - can't get mail client occurance." );
	status = SCNOTDONE;
    }
    else {
	/* Handle new occurance */
	mcoccur_begin( mcoccP, mcP );

	/* Connect with the agent */
	flags = IA_SIGCHG;
	if ( mccP->mc_statefulF )
	    flags |= IA_ALL;
	mcoccur_agint_add( mcoccP, mccP->mc_agentP, flags );

	/* Fire off info messages to interested agents */
	mcoccur_notify( 0, agnotfmt_info,
		        mccP->mc_bsP,
			mcoccP, mccP->mc_agentP, mcP );

	mcoccur_release( mcoccP );
    }

    mcdb_mclient_release( mcP, FALSE );
    return ( status );
}
/*************************************************************/


/*************************************************************
 * cmd_stats -- the "stats" command.

	stats

    Asks for statistics.  May be enhanced at some point to take
      a 'subtype' argument

    Formats a stats message and sends to the agent.

*/
static STSCODE
cmd_stats ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
    /* Format and return stats message to the agent */
    if ( agnotfmt_stats( mccP->mc_bsP, NULL, mccP->mc_agentP, NULL) )
	(*mccP->mc_agentP->ag_ctlP->ac_send)( mccP->mc_agentP,
					      mccP->mc_bsP->bs_bP,
					      mccP->mc_bsP->bs_bC );
	else
	    mccmd_return( mccP, "ERROR - error formating response." );

    return ( SCOK );
}
/*************************************************************/



/*************************************************************
 * cmd_stop -- the "stop" command.

	stop <mcname> <etc>

    Stop this mail client occurance and update its status

*/
static STSCODE
cmd_stop ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;

    /* Hand off to the update-processing function */
    status = mccmd_update( mccP, TRUE );

    return ( status );
}
/*************************************************************/




/*************************************************************
 * cmd_sync -- the "sync" command.

	sync

    Syncs any memory data to disk, if relevant

*/
static STSCODE
cmd_sync ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;

    /* Tell mcdb to sync */
    mcdb_sync();

    mccmd_return( mccP, "OK" );

    return ( status );
}
/*************************************************************/




/*************************************************************
 * cmd_update -- the "update" command.

	update <mcname> <etc>

    Update the status of this mail client occurance

*/
static STSCODE
cmd_update ARGLIST( (mccP) )
    FARG( MCCMD		*mccP )		/* Command context block */
{
	STSCODE		status;

    /* Hand off to the update-processing function */
    status = mccmd_update( mccP, FALSE );

    return ( status );
}
/*************************************************************/


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* agnotfmt_info( bsP, mcoccP, agentP, heldmcP )

	Agent notification formatting: INFO message

Accepts :

	bsP		BSTR to format the message into
	mcoccP		Ptr to mail client occurance
	agentP		Ptr to occurance for agent to whom notice is being sent
	heldmcP		Ptr to mail client db entry, if held

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Calling sequence per mcoccur_notify

*/

static BOOL
agnotfmt_info ARGLIST( (bsP, mcoccP, agentP, heldmcP) )
   NFARG( BSTR		*bsP )		/* BSTR to format into */
   NFARG( MCOCCUR	*mcoccP )	/* Mail client occurance */
   NFARG( AGENT		*agentP )	/* Agent occurance */
    FARG( MCLIENT	*heldmcP )	/* Mail client if held */
{
	BOOL		okF;
	int		i;
	MCLIENT		*mcP;		/* Mail client info */
	UBYTE		mcnamestr[ MCNAME_BSMAX + 20 ];

    i = mcname_str( &mcoccP->mco_mcname, &mcnamestr[0], sizeof(mcnamestr) );
    if ( i < 0 ) {
	warning( "agnotfmt_info -- mcname_str returned error." );
	return ( FALSE );
    }
    if ( i >= sizeof( mcnamestr ) ) {
	warning( "agnotfmt_info -- mcname too large for buffer." );
	return ( FALSE );
    }

    bsP->bs_bC = 0;

    /* Need to get mail client occurance if not passed */
    if ( heldmcP != NULL )
	mcP = heldmcP;
    else {
	mcP = mcdb_mclient_access( &mcoccP->mco_mcname );
	if ( mcP == NULL )
	    return ( FALSE );
    }

    okF = TRUE;				/* Presume success */

    bstr_printf( bsP, "INFO %s ; OCC %d ;",
		&mcnamestr[0],
		mcoccP->mco_agents.llh_nodeC );

    /* Add any relevant flags */
    if ( ( mcP->mc_flags & MCF_DNSBL ) != 0 )
	bstr_printf( bsP, " DNSBL ;" );
    if ( ( mcP->mc_flags & MCF_RELAYOK ) != 0 )
	bstr_printf( bsP, " RELAYOK ;" );

    /* Give anything based on simple state */
    switch ( mcP->mc_sstate ) {
	case MCS_NONE:			/* No specific state */
	    break;

	case MCS_GREY:			/* In greylisting period */
	    bstr_printf( bsP, " T.GL %lu ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_C_ADM:		/* Block connection, administrative */
	    bstr_printf( bsP, " BLOCK conn admin %lu 0 ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_C_AUTO:		/* Block connection, automatic */
	    bstr_printf( bsP, " BLOCK conn auto %lu 0 ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_G_ADM:		/* Block at greeting, administrative */
	    bstr_printf( bsP, " BLOCK greet admin %lu 0 ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_G_AUTO:		/* Block at greeting, automatic */
	    bstr_printf( bsP, " BLOCK greet auto %lu 0 ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_S_ADM:		/* Block with dialog, administrative */
	    bstr_printf( bsP, " BLOCK smtp admin %lu 0 ;", mcP->mc_sstime );
	    break;

	case MCS_BLOCK_S_AUTO:		/* Block with dialog, automatic */
	    bstr_printf( bsP, " BLOCK smtp auto %lu 0 ;", mcP->mc_sstime );
	    break;

	default:
	    warning( "agnotfmt_info -- unknown mc_sstate %d", 
		      mcP->mc_sstate );
	    okF = FALSE;
	    break;
    }

    /* Give anything based on override state */
    switch ( mcP->mc_ostate ) {
	case MCO_NONE:			/* No override state */
	    break;

	case MCO_PASS:			/* A "pass" on some checks */
	    bstr_printf( bsP, " PASS %lu ;", mcP->mc_ostime );
	    break;

	default:
	    warning( "agnotfmt_info -- unknown mc_ostate %d", 
		      mcP->mc_ostate );
	    okF = FALSE;
	    break;
    }

    /* Any pre-greeting delay. */
    if ( okF && ( mcP->mc_pgd_sec > 0 ) ) {
	/* Give low number if in the pass window. */
	if ( mcP->mc_pgd_wconC > 0 )
	    bstr_printf( bsP, " T.PGD 0 ;" );
	else
	    bstr_printf( bsP, " T.PGD %d ;", mcP->mc_pgd_sec );
    }

    /* End */
    bstr_printf( bsP, " END" );

    /* Release locally obtained mclient if not passed in */
    if ( heldmcP == NULL )
	mcdb_mclient_release( mcP, FALSE );

    return ( okF );
}
/*

*//* agnotfmt_info_x( bsP, mcoccP, agentP, heldmcP )

	Agent notification formatting: INFO message, extended

Accepts :

	bsP		BSTR to format the message into
	mcoccP		Ptr to mail client occurance
	agentP		Ptr to occurance for agent to whom notice is being sent
	heldmcP		Ptr to mail client db entry, if held

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Calling sequence per mcoccur_notify

*/

static BOOL
agnotfmt_info_x ARGLIST( (bsP, mcoccP, agentP, heldmcP) )
   NFARG( BSTR		*bsP )		/* BSTR to format into */
   NFARG( MCOCCUR	*mcoccP )	/* Mail client occurance */
   NFARG( AGENT		*agentP )	/* Agent occurance */
    FARG( MCLIENT	*heldmcP )	/* Mail client if held */
{
	BOOL		okF;
	BOOL		wantnlF;	/* Want newline */
	MCLIENT		*mcP;		/* Mail client info */

    /* Need to get mail client occurance if not passed */
    if ( heldmcP != NULL )
	mcP = heldmcP;
    else {
	mcP = mcdb_mclient_access( &mcoccP->mco_mcname );
	if ( mcP == NULL )
	    return ( FALSE );
    }

    /* Begin with the basic info, with hackery to extended it. */
    okF = agnotfmt_info( bsP, mcoccP, agentP, mcP );
    if ( okF && ( bsP->bs_bC >= 4 ) &&
	 ( strnicmp( (char *)&bsP->bs_bP[ bsP->bs_bC -4 ],
	     " END", 4 ) == 0 ) )
	bsP->bs_bC -= 4;
    bstr_putb( bsP, '\n' );

    /* Any pre-greeting delay parms */
    if ( okF && ( mcP->mc_pgd_sec > 0 ) ) {
	bstr_printf( bsP, " PGD %d %d %d ;",
		       mcP->mc_pgd_sec,
		       mcP->mc_pgd_wcon,
		       mcP->mc_pgd_wday );
	bstr_printf( bsP, " PGDWIN %d %ld ;\n",
		       mcP->mc_pgd_wconC,
		       mcP->mc_pgd_wtime );
    }

    /* Add bookkeepping stuff */
    wantnlF = FALSE;
    if ( okF && ( ( mcP->mc_grcptC > 0 ) || ( mcP->mc_brcptC > 0 ) ) ) {
	bstr_printf( bsP, " RCPT %ld %ld ;",
		      mcP->mc_grcptC, mcP->mc_brcptC );
	wantnlF = TRUE;
    }
    if ( okF && ( mcP->mc_rcptgov != 0 ) ) {
	bstr_printf( bsP, " RGOV %d %ld;",
		     mcP->mc_rcptgov, mcP->mc_rgtime );
	wantnlF = TRUE;
    }
    if ( wantnlF )
	bstr_putb( bsP, '\n' );

    wantnlF = FALSE;
    if ( okF && ( mcP->mc_fctime != 0 ) ) {
	bstr_printf( bsP, " FCTIME %ld ;", mcP->mc_fctime );
	wantnlF = TRUE;
    }
    if ( okF && ( mcP->mc_lctime != 0 ) ) {
	bstr_printf( bsP, " LCTIME %ld ;", mcP->mc_lctime );
	wantnlF = TRUE;
    }
    if ( okF && ( mcP->mc_connC != 0 ) ) {
	bstr_printf( bsP, " CONN %ld ;", mcP->mc_connC );
	wantnlF = TRUE;
    }
    if ( wantnlF )
	bstr_putb( bsP, '\n' );

    /* Flags that don't get reported in standard info */
    wantnlF = FALSE;
    if ( okF && ( ( mcP->mc_flags & MCF_AGG ) != 0 ) ) {
	bstr_printf( bsP, " AGG ;" );
	wantnlF = TRUE;
    }
    if ( okF && ( ( mcP->mc_flags & MCF_PROTO ) != 0 ) ) {
	bstr_printf( bsP, " PROTO ;" );
	wantnlF = TRUE;
    }
    if ( okF && ( ( mcP->mc_flags & MCF_TRANSIENT ) != 0 ) ) {
	bstr_printf( bsP, " TRANSIENT ;" );
	wantnlF = TRUE;
    }
    if ( okF && ( ( mcP->mc_flags & MCF_NO_RGOV ) != 0 ) ) {
	bstr_printf( bsP, " NO-RGOV ;" );
	wantnlF = TRUE;
    }
    if ( wantnlF )
	bstr_putb( bsP, '\n' );


    /* End */
    bstr_printf( bsP, " END" );

    /* Release locally obtained mclient if not passed in */
    if ( heldmcP == NULL )
	mcdb_mclient_release( mcP, FALSE );

    return ( okF );
}
/*

*//* agnotfmt_stats( bsP, mcoccP, agentP, heldmcP )

	Agent notification formatting: statistics message

Accepts :

	bsP		BSTR to format the message into
	mcoccP		Ptr to mail client occurance
	agentP		Ptr to occurance for agent to whom notice is being sent
	heldmcP		Ptr to mail client db entry, if held

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Calling sequence per mcoccur_notify;

	mcoccP is unused and should be passed as NULL

	heldmcP is unused and should be passed as NULL.

*/

static BOOL
agnotfmt_stats ARGLIST( (bsP, mcoccP, agentP, heldmcP) )
   NFARG( BSTR		*bsP )		/* BSTR to format into */
   NFARG( MCOCCUR	*mcoccP )	/* Mail client occurance */
   NFARG( AGENT		*agentP )	/* Agent occurance */
    FARG( MCLIENT	*heldmcP )	/* Mail client if held */
{
	int		i;
	struct rusage	ru;

    bsP->bs_bC = 0;

    bstr_printf( bsP, "TEXT stats ; mvmcas %s ;\n", MVMCAS_VERSION );

    /* Generic stats */
    bstr_printf( bsP, "  started %ld ;\n", Starttime.tv_sec );

    bstr_printf( bsP, "  TCP req: %ld ;\n", Tcp_connC );
    bstr_printf( bsP, "  UDP req: %ld ;\n", Udp_connC );

    /* mclient stats */
    mcdb_mclient_stats( bsP );

    /* Occurance stats */
    mcoccur_stats( bsP );

    /* Process stats */
    if ( getrusage( RUSAGE_SELF, &ru ) == 0 ) {
	bstr_printf( bsP, "  user-ms %ld ;",
		       ( ru.ru_utime.tv_sec * 1000 ) +
		       ( ru.ru_utime.tv_usec / 1000 ) );

	bstr_printf( bsP, " sys-ms %ld ;",
		       ( ru.ru_stime.tv_sec * 1000 ) +
		       ( ru.ru_stime.tv_usec / 1000 ) );

	bstr_printf( bsP, " rss-kb %ld ;\n", ru.ru_maxrss );
    }

    /* End */
    bstr_printf( bsP, "END" );

    return ( TRUE );
}
/*

*//* mccmd_get()

	Get a command block

Accepts :



Returns :

	<value>		PTR to MCCMD block
			NULL if error

Notes :

	Keeps free command blocks on a free list, which is tried first.

*/

static MCCMD *
mccmd_get NOARGLIST
{
	MCCMD	*mccP;			/* Client block pointer */

    /* Try the free list first */
    if ( pthread_mutex_lock( &Mccmd_lock ) != 0 )
	warning( "mccmd_get: can't get Mccmd_lock" );
    if ( ( mccP = Mccmd_listP ) != NULL ) {
	Mccmd_listP = mccP->mc_nextP;
    }
    else {
	/* No free, allocate new one */
	mccP = (MCCMD *)emalloc( "mccmd_get", "command block", 
				      sizeof(MCCMD) );
	if ( mccP == NULL ) {
	    warning( "mccmd_get can't allocate command block." );
	}
	else {
	    mccP->mc_rstrP = bstr_new( "command result string", 500, 2000 );
	    if ( mccP->mc_rstrP == NULL ) {
		warning( "mccmd_get can't allocate result string buffer" );
		dealloc( mccP );
		mccP = NULL;
	    }
	    mccP->mc_bsP = bstr_new( "command scratch string", 200, 2000 );
	    if ( mccP->mc_bsP == NULL ) {
		warning( "mccmd_get can't allocate scratch string buffer" );
		bstr_free( mccP->mc_rstrP );
		dealloc( mccP );
		mccP = NULL;
	    }
	}
    }

    if ( pthread_mutex_unlock( &Mccmd_lock ) != 0 )
	warning( "mccmd_get: can't release Mccmd_lock" );

    if ( mccP != NULL ) {
	/* Finish initializing */
	mccP->mc_statefulF = FALSE;
	mccP->mc_cmdP = NULL;
    }

    return ( mccP );
}
/*

*//* mccmd_mclient_set( mccP, mcP, initF )

	Process commands to set new info in a mail client

Accepts :

	mccP		Ptr to command context block
	mcP		Ptr to mail client
	initF		TRUE to completely initialize the client

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	on FALSE return, an error message will be returned.

	The parsing code may duplicate similar code elsewhere.
	  Should really coalesce these somehow.

*/

static BOOL
mccmd_mclient_set ARGLIST( (mccP, mcP, initF) )
   NFARG( MCCMD		*mccP )		/* Command block */
   NFARG( MCLIENT	*mcP )		/* Mail client info */
    FARG( BOOL		initF )		/* For fresh initialization */
{
	BOOL		skipF;		/* Skipping to semicolon */
	long		lv1;
	ULONG		ulv1, ulv2, ulv3;
	MCSSTATE	state1, state2;
	char		*errP;		/* Any error string */
	UBYTE		tok[ MCNAME_BSMAX + 20 ];
	UBYTE		tok1[ MCNAME_BSMAX + 20 ];
	UBYTE		tok2[ MCNAME_BSMAX + 20 ];
	UBYTE		tok3[ MCNAME_BSMAX + 20 ];
	MCLIENT		mc;		/* Local mail client data */

    /* Either start with the passed data, or from scratch */
    if ( initF )
	mclient_init( &mc );
    else
	memcpy( &mc, mcP, sizeof(MCLIENT) );


    /* Now get the data for this mail client.  This is probably pretty
       ugly.
    */
    skipF = TRUE;			/* Skip to first semi, first */
    errP = NULL;			/* No error yet */
    for ( ; errP == NULL ; ) {
	if ( ! mccmd_token( mccP, &tok[0], sizeof(tok) ) ) {
	    errP = "token too large";
	    break;
	}
	if ( tok[0] == NUL )
	    break;			/* No more tokens */
	if ( skipF ) {
	    /* Skipping to semi... */
	    if ( ( tok[0] != ';' ) || ( tok[1] != NUL ) )
		errP = "expected semicolon";
	    else
		skipF = FALSE;		/* Found it */
	    continue;
	}

	/* Toggle the skip flag back on; anything else handled here (valid
	   or not) will need to look for semicolon.
	*/
	skipF = TRUE;

	/* Process token.  Check end first.. */
	if ( stricmp( (char *)&tok[0], "END" ) == 0 )
	    break;

	else if ( stricmp( (char *)&tok[0], "-AGG" ) == 0 )
	    /* Remove aggregate flag */
	    mc.mc_flags &= ~MCF_AGG;

	else if ( stricmp( (char *)&tok[0], "AGG" ) == 0 )
	    /* Flag as an aggregate */
	    mc.mc_flags |= MCF_AGG;

	else if ( stricmp( (char *)&tok[0], "-AUTHREQ" ) == 0 )
	    /* Remove authreq flag */
	    mc.mc_flags &= ~MCF_AUTHREQ;

	else if ( stricmp( (char *)&tok[0], "AUTHREQ" ) == 0 )
	    /* AUTH is required for this source. */
	    mc.mc_flags |= MCF_AUTHREQ;

	else if ( stricmp( (char *)&tok[0], "BLOCK" ) == 0 ) {
	    /* Blocked:  type how end code */
	    if ( !mccmd_token( mccP, &tok1[0], sizeof(tok1) ) ) {
		errP = "token too large for BLOCK";
	    }
	    else if ( ( tok1[0] == ';' ) && ( tok1[1] == NUL ) )
		errP = "too few arguments to BLOCK";
	    else if ( !mccmd_token( mccP, &tok2[0], sizeof(tok2) ) ) {
		errP = "token too large for BLOCK";
	    }
	    else if ( ( tok2[0] == ';' ) && ( tok2[1] == NUL ) )
		errP = "too few arguments to BLOCK";
	    else if ( !mccmd_token_uint( mccP, &ulv1 ) )
		errP = "bad time argument for BLOCK";
	    else if ( !mccmd_token( mccP, &tok3[0], sizeof(tok3) ) )
		errP = "token too large for BLOCK";
	    else if ( ( tok3[0] == ';' ) && ( tok3[1] == NUL ) ) {
		errP = "too few arguments to BLOCK";
	    }
	    else {
		if ( stricmp( (char *)&tok1[0], "conn" ) == 0 ) {
		    state1 = MCS_BLOCK_C_ADM;   state2 = MCS_BLOCK_C_AUTO;
		}
		else if ( stricmp( (char *)&tok1[0], "greet" ) == 0 ) {
		    state1 = MCS_BLOCK_G_ADM;   state2 = MCS_BLOCK_G_AUTO;
		}
		else if ( stricmp( (char *)&tok1[0], "smtp" ) == 0 ) {
		    state1 = MCS_BLOCK_S_ADM;   state2 = MCS_BLOCK_S_AUTO;
		}
		else {
		    errP = "invalid BLOCK type";
		    state1 = MCS_NONE;
		}

		if ( state1 == MCS_NONE )
		    ;
		else if ( stricmp( (char *)&tok2[0], "admin" ) == 0 )
		    mc.mc_sstate = state1;
		else if ( stricmp( (char *)&tok2[0], "auto" ) == 0 )
		    mc.mc_sstate = state2;
		else {
		    errP = "invalid BLOCK source";
		    mc.mc_sstate = MCS_NONE;
		}

		if ( mc.mc_sstate != MCS_NONE ) {
		    mc.mc_sstime = ulv1;
TODO		    /* Fill in the code from tok3 */
		}
	    }
	}

	else if ( stricmp( (char *)&tok[0], "CONN" ) == 0 ) {
	    /* Connection count */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad argument to CONN";
		skipF = ( ulv1 != 1 );
	    }
	    else
		mcP->mc_connC = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "-DNSBL" ) == 0 )
	    /* Remove DNSBL flag */
	    mc.mc_flags &= ~MCF_DNSBL;

	else if ( stricmp( (char *)&tok[0], "DNSBL" ) == 0 )
	    /* DNSBL check should be done for this source */
	    mc.mc_flags |= MCF_DNSBL;

	else if ( stricmp( (char *)&tok[0], "FCTIME" ) == 0 ) {
	    /* First connect time */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad argument to FCTIME";
		skipF = ( ulv1 != 1 );
	    }
	    else
		mc.mc_fctime = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "LCTIME" ) == 0 ) {
	    /* Last connect time */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad argument to LCTIME";
		skipF = ( ulv1 != 1 );
	    }
	    else
		mc.mc_lctime = ulv1;
	}

	else if ( stricmp( (char *)&tok[0], "-NO-RGOV" ) == 0 )
	    /* Remove NO-RGOV flag */
	    mc.mc_flags &= ~MCF_NO_RGOV;

	else if ( stricmp( (char *)&tok[0], "NO-RGOV" ) == 0 )
	    /* Suppress rcpt-governing for this client */
	    mc.mc_flags |= MCF_NO_RGOV;

	else if ( stricmp( (char *)&tok[0], "PASS" ) == 0 ) {
	    /* Give a "pass" on some checks.  The value, as with
	       greylisting, is either an interval or an end date.
	       It may also be 0 to remove the pass.
	    */
	    if ( mccmd_token_uint( mccP, &ulv1 ) ) {
		if ( ulv1 == 0 )
		    mc.mc_ostate = MCO_NONE;
		else {
		    mc.mc_ostate = MCO_PASS;
		    mc.mc_ostime = ulv1;
		}
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		errP = "bad argument to PASS";
		skipF = ( ulv1 != 1 );
	    }
	}

	else if ( stricmp( (char *)&tok[0], "PGD" ) == 0 ) {
	    /* Pre-greeting delay parameters:
	       PGD seconds wcon wdays
	    */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad first argument to PGD";
		skipF = ( ulv1 != 1 );
	    }
	    else if ( !mccmd_token_uint( mccP, &ulv2 ) ) {
		errP = "bad second argument to PGD";
		skipF = ( ulv2 != 1 );
	    }
	    else if ( !mccmd_token_uint( mccP, &ulv3 ) ) {
		errP = "bad third argument to PGD";
		skipF = ( ulv3 != 1 );
	    }
	    else {
		mc.mc_pgd_sec = ulv1;
		mc.mc_pgd_wcon = ulv2;
		mc.mc_pgd_wday = ulv3;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "PGDWIN" ) == 0 ) {
	    /* Pre-greeting delay window status:
	       PGDWIN conn wtime
	    */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad first argument to PGDWIN";
		skipF = ( ulv1 != 1 );
	    }
	    else if ( !mccmd_token_uint( mccP, &ulv2 ) ) {
		errP = "bad second argument to PGDWIN";
		skipF = ( ulv2 != 1 );
	    }
	    else {
		mc.mc_pgd_wconC = ulv1;
		mc.mc_pgd_wtime = ulv2;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "-PROTO" ) == 0 )
	    /* Remove prototype flag */
	    mc.mc_flags &= ~MCF_PROTO;

	else if ( stricmp( (char *)&tok[0], "PROTO" ) == 0 )
	    /* Flag as a prototype */
	    mc.mc_flags |= MCF_PROTO;

	else if ( stricmp( (char *)&tok[0], "RCPT" ) == 0 ) {
	    /* RCPT-TO counters */
	    if ( !mccmd_token_uint( mccP, &ulv1 ) ) {
		errP = "bad first argument to RCPT";
		skipF = ( ulv1 != 1 );
	    }
	    else if ( !mccmd_token_uint( mccP, &ulv2 ) ) {
		errP = "bad second argument to RCPT";
		skipF = ( ulv2 != 1 );
	    }
	    else {
		mc.mc_grcptC = ulv1;
		mc.mc_brcptC = ulv2;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "-RELAYOK" ) == 0 )
	    /* Remove relayok flag */
	    mc.mc_flags &= ~MCF_RELAYOK;

	else if ( stricmp( (char *)&tok[0], "RELAYOK" ) == 0 )
	    /* Relay is OK for this */
	    mc.mc_flags |= MCF_RELAYOK;

	else if ( stricmp( (char *)&tok[0], "RGOV" ) == 0 ) {
	    /* RCPT-TO governor value and last decay time*/
	    if ( !mccmd_token_int( mccP, &lv1 ) ) {
		errP = "bad first argument to RGOV";
		skipF = ( lv1 != 1 );
	    }
	    else if ( !mccmd_token_uint( mccP, &ulv2 ) ) {
		errP = "bad second argument to RGOV";
		skipF = ( ulv2 != 1 );
	    }
	    else {
		mc.mc_rcptgov = lv1;
		mc.mc_rgtime = ulv2;
	    }
	}

	else if ( stricmp( (char *)&tok[0], "-TRANSIENT" ) == 0 )
	    /* Remove transient flag */
	    mc.mc_flags &= ~MCF_TRANSIENT;

	else if ( stricmp( (char *)&tok[0], "TRANSIENT" ) == 0 )
	    /* This (or any child created from it) is transient */
	    mc.mc_flags |= MCF_TRANSIENT;

	else if ( stricmp( (char *)&tok[0], "T.GL" ) == 0 ) {
	    /* Specifies greylisting; value is either end time or 
		interval: at any rate, it's a number.
	    */
	    if ( mccmd_token_uint( mccP, &ulv1 ) ) {
		mc.mc_sstate = MCS_GREY;
		mc.mc_sstime = ulv1;
	    }
	    else {
		/* Problem with input value; set skip flag accordingly */
		errP = "bad argument to T.GL";
		skipF = ( ulv1 != 1 );
	    }
	}

	else {
	    /* Unrecognized; ignore by skipping to semi. */
	    errP = "unrecognized token";
	}
    }

    if ( errP != NULL ) {
	mccmd_return( mccP, "RQERR -  %s", errP );
	return ( FALSE );
    }

    /* Copy any changeable stuff back from local mclient to real one */
    mcP->mc_flags = mc.mc_flags;
    mcP->mc_pgd_sec = mc.mc_pgd_sec;
    mcP->mc_pgd_wday = mc.mc_pgd_wday;
    mcP->mc_pgd_wcon = mc.mc_pgd_wcon;
    mcP->mc_pgd_wconC = mc.mc_pgd_wconC;
    mcP->mc_pgd_wtime = mc.mc_pgd_wtime;
    mcP->mc_sstate = mc.mc_sstate;
    mcP->mc_sstime = mc.mc_sstime;
    mcP->mc_ostate = mc.mc_ostate;
    mcP->mc_ostime = mc.mc_ostime;
    mcP->mc_grcptC = mc.mc_grcptC;
    mcP->mc_brcptC = mc.mc_brcptC;
    mcP->mc_rcptgov = mc.mc_rcptgov;

    return ( TRUE );
}
/*

*//* mccmd_parse_mcname( mccP, mcnmsP )

	Parse mail client name

Accepts :

	mccP		Ptr to command context block
	mcnmsP		Ptr to static name block to fill in.

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

static BOOL
mccmd_parse_mcname ARGLIST( (mccP, mcnmsP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( MCNAME_S	*mcnmsP )	/* The name block to fill in */
{
	int		i;
	UBYTE		ntype[20];	/* Client name type */

    /* Get the name type */
    i = strtoken_ws( &mccP->mc_cmdP, &ntype[0], sizeof(ntype), FALSE );
    if ( i >= sizeof(ntype) ) {
	mccmd_return( mccP, "RQERR - mail client name type too long." );
	return ( FALSE );
    }

    /* Initialize the static pointer to point to the contained data */
    mcnmsP->mns_mcname.mn_nameP = (void *)&mcnmsP->mns_data;

    /* Decode name type and process accordingly */
    if ( stricmp( (char *)&ntype[0], "auth" ) == 0 ) {
	/* Auth name */
	mcnmsP->mns_mcname.mn_type = MCT_AUTH;
	return ( mccmd_parse_mcname_bs( mccP, mcnmsP ) );
    }
    else if ( stricmp( (char *)&ntype[0], "ip" ) == 0 ) {
	/* IPv4 address */
	mcnmsP->mns_mcname.mn_type = MCT_IP;
	return ( mccmd_parse_mcname_ip( mccP, mcnmsP ) );
    }

    mccmd_return( mccP, "RQERR - unknown mail client name type." );
    return ( FALSE );
}
/*

*//* mccmd_parse_mcname_bs( mccP, mcnmsP )

	Parse mail client name that is a byte string

Accepts :

	mccP		Ptr to command context block
	mcnmsP		Ptr to name block

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

static BOOL
mccmd_parse_mcname_bs ARGLIST( (mccP, mcnmsP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( MCNAME_S	*mcnmsP )	/* The name block to fill in */
{
	int		i;

    /* Get the string */
    i = strtoken_ws( &mccP->mc_cmdP,
		     &mcnmsP->mns_data.nm_bs[0], MCNAME_BSMAX,
		     FALSE );
    if ( i >= MCNAME_BSMAX )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* mccmd_parse_mcname_ip( mccP, mcnmsP )

	Parse mail client name that is an IP address

Accepts :

	mccP		Ptr to command context block
	mcnmsP		Ptr to name block

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :


*/

static BOOL
mccmd_parse_mcname_ip ARGLIST( (mccP, mcnmsP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( MCNAME_S	*mcnmsP )	/* The name block to fill in */
{
	int		i;
	UBYTE		*mP;		/* Ptr to any mask string */
	UBYTE		ipbuf[50];	/* Place to store the IP string */

    /* Get the IP string to work with */
    i = strtoken_ws( &mccP->mc_cmdP, &ipbuf[0], sizeof(ipbuf), FALSE );
    if ( i >= sizeof(ipbuf) ) {
	mccmd_return( mccP, "RQERR - IP address string too long." );
	return ( FALSE );
    }

    /* Point to masklen suffix or default to /32 */
    mP = (UBYTE *)strchr( (char *)&ipbuf[0], '/' );
    if ( mP != NULL )
	*mP++ = NUL;
    else
	mP = (UBYTE *)"32";

    if ( !ip_decode_str( &ipbuf[0], &mcnmsP->mns_data.nm_ip4.nip_addr ) ) {
	mccmd_return( mccP, "RQERR - Bad IP address string." );
	return( FALSE );
    }
    else if ( ( ( i = atod( mP ) ) < 0 ) || ( i > 32 ) ) {
	warning( "mccmd_parse_mcname_ip: bad IP masklen \"%d\"", i );
	return ( FALSE );
    }

    mcnmsP->mns_data.nm_ip4.nip_masklen = i;

    return ( TRUE );
}
/*

*//* mccmd_preamble( mccP, cmdbufP )

	Does the initial command processing work

Accepts :

	mccP		Command context block
	cmdbufP		Ptr to the command buffer

Returns :

	<value>		Ptr to the command table entry if found
			NULL if not

Notes :

	Sets up for processing a command by setting up the parsing
	context, extracting the command verb, and looking it up.

*/

static CMTE *
mccmd_preamble ARGLIST( (mccP, cmdbufP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( UBYTE		*cmdbufP )	/* Command buffer */
{
	int		i;
	CMTE		*cmteP;
	char		verb[20];	/* Command verb */

    /* Set up the context */
    mccP->mc_cmdP = cmdbufP;		/* Command buffer point */
    mccP->mc_rstrP->bs_bC = 0;		/* Clear the return string */

    /* Get the command verb */
    i = strtoken_ws( &mccP->mc_cmdP, &verb[0], sizeof(verb), FALSE );
    if ( i == 0 ) {
	/* Nothing in buffer, could be a blank line */
	mccmd_return( mccP, "TEXT - Eh?" );
	return ( NULL );
    }

    if ( i >= sizeof(verb) ) {
	mccmd_return( mccP, "RQERR - Command verb is too big." );
	return ( NULL );
    }

    /* Look up this verb in the command table */
    for( cmteP = &Cmdtbl[0]; cmteP->cm_nameP != NULL; ++cmteP ) {
	if ( stricmp( (char *)&verb[0], (char *)cmteP->cm_nameP ) == 0 )
	    break;
    }

    if ( cmteP->cm_nameP == NULL ) {
	mccmd_return( mccP, "RQERR - Command unrecognized." );
	return ( NULL );
    }

    return ( cmteP );
}
/*

*//* mccmd_release( mccP )

	Release a command block

Accepts :

	mccP		Ptr to MCCMD block to release

Returns :

	<void>

Notes :

	Free command blocks are kept on a list for quick access.

*/

static void
mccmd_release ARGLIST( (mccP) )
    FARG( MCCMD	*mccP )			/* Block to release */
{
    if ( pthread_mutex_lock( &Mccmd_lock ) != 0 )
	warning( "mccmd_release: can't get Mccmd_lock" );

    mccP->mc_nextP = Mccmd_listP;
    Mccmd_listP = mccP;

    if ( pthread_mutex_unlock( &Mccmd_lock ) != 0 )
	warning( "mccmd_release: can't release Mccmd_lock" );
}
/*

*//* mccmd_return( mccP, fmtP, args... )

	Return a status via command context block

Accepts :

	mccP		Command context block
	fmtP		Format string for text return
	args...		Any arguments to the format string

Returns :

	<nothing>

Notes :


*/

static void
mccmd_return ARGLIST( (mccP, fmtP, va_decl) )
   NFARG( MCCMD		*mccP )		/* Command context block */
   NFARG( char		*fmtP )		/* Format string */
    FARG( ... )				/* stdc variable arguments */
{
	BSTR		*bsP;
	va_list		ap;		/* Varargs arg ptr */

    /* Start varargs */
    va_start( ap, fmtP );

    bsP = mccP->mc_rstrP;
    bsP->bs_bC = 0;

    bstr_vprintf( bsP, fmtP, ap );

    if ( Debug > 0 ) {
	printf( "Return: %s\n", bsP->bs_bP );
    }

    bstr_putb( bsP, '\n' );
    bstr_nulterm( bsP );

    va_end( ap );
}
/*

*//* mccmd_return_info_x( mccP, mcP )

	Return an extended info message for a mail client

Accepts :

	mccP		Command context block
	mcP		Ptr to mail client block

Returns :

	<value>		SCxxx status code

Notes :


*/

static STSCODE
mccmd_return_info_x ARGLIST( (mccP, mcP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( MCLIENT	*mcP )		/* Mail client info */
{
	STSCODE		status;
	MCOCCUR		*mcoccP;	/* Mail client occurance */

    /* Get occurance data for the mclient's name.  Note that this
       may be different than the input name, e.g. for an aggregate.
    */
    mcoccP = mcoccur_access( &mcP->mc_mcname );
    if ( mcoccP == NULL ) {
	mccmd_return( mccP, "ERROR - can't get mail client occurance." );
	status = SCNOTDONE;
    }
    else {
	/* Format and return info message to the agent */
	if ( agnotfmt_info_x( mccP->mc_bsP, mcoccP, mccP->mc_agentP, mcP) )
	    (*mccP->mc_agentP->ag_ctlP->ac_send)( mccP->mc_agentP,
						  mccP->mc_bsP->bs_bP,
						  mccP->mc_bsP->bs_bC );
	else
	    mccmd_return( mccP, "ERROR - error formating response." );

	status = SCOK;

	mcoccur_release( mcoccP );
    }

    return ( status );
}
/*

*//* mccmd_token( mccP, tokP, tokL )

	Get next token from a command message

Accepts :

	mccP		Command context block
	tokP		Where to put the token
	tokL		Length of token buffer

Returns :

	<value>		TRUE if successful
			FALSE otherwise.

Notes :

	More or less a wrapper around strtoken_ws() from mml;
	as with strtoken_ws,
	  - the pointer at mccP->mc_cmdP is updated to point after the
	    extracted token.
	  - the returned token is NUL-terminated.

	The FALSE return will happen if the buffer is too small for
	  the token.  A warning message may be given, but no return
	  is done to the agent.

*/

static BOOL
mccmd_token ARGLIST( (mccP, tokP, tokL) )
   NFARG( MCCMD		*mccP )		/* Command context block */
   NFARG( UBYTE		*tokP )		/* Where to put the token */
    FARG( int		tokL )		/* Length of token buffer */
{
	int		i;
	UBYTE		*bufP;

    /* Get a copy of the message pointer, in case we need it for warning */
    bufP = mccP->mc_cmdP;

    /* Do token-getting */
    i = strtoken_ws( &mccP->mc_cmdP, tokP, tokL, TRUE );
    if ( i >= tokL ) {
	warning( "mccmd_token: token too large at \"%s\".", bufP );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* mccmd_token_int( mccP, longP )

	Get the next token from a command message, interpret as int

Accepts :

	mccP		Command context block
	longP		Where to store the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a sign or a digit.

	For a FALSE return, the return value *longP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

static BOOL
mccmd_token_int ARGLIST( (mccP, longP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( long		*longP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mccmd_token( mccP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*longP = 0;
	return ( FALSE );
    }

    if ( ( tok[0] == '-' ) || ( tok[0] == '+' ) ||
	 ( isdigit( tok[0] ) ) ) {
	/* Input looks OK; get value and return it. */
	*longP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NUL ) )
	*longP = 1;
    else
	*longP = 0;

    return ( FALSE );
}
/*

*//* mccmd_token_uint( mccP, ulongP )

	Get the next token from a command message, interpret as unsigned int

Accepts :

	mccP		Command context block
	ulongP		Where to store the value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	At present the only validity check that is done is to see if the
	string begins with a digit or a plus

	For a FALSE return, the return value *ulongP also conveys meaning:
	  0 - nothing special
	  1 - input was a semicolon token.

*/

static BOOL
mccmd_token_uint ARGLIST( (mccP, ulongP) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( ULONG		*ulongP )	/* Where to store the result */
{
	UBYTE		tok[50];	/* Token for integer value */

    if ( !mccmd_token( mccP, &tok[0], sizeof(tok) ) ) {
	/* Bad input. */
	*ulongP = 0;
	return ( FALSE );
    }

    if ( isdigit( tok[0] ) || ( tok[0] == '+' ) ) {
	/* Input looks OK; get value and return it. */
	*ulongP = atod( (char *)&tok[0] );
	return ( TRUE );
    }

    /* Not valid input; check for semicolon token */
    if ( ( tok[0] == ';' ) && ( tok[1] == NUL ) )
	*ulongP = 1;
    else
	*ulongP = 0;

    return ( FALSE );
}
/*

*//* mccmd_update( mccP, stopF )

	Processes an update-like message

Accepts :

	mccP		Command context block
	stopF		Whether the command implies "stop"

Returns :

	<value>		SCxxx status code, including:

Notes :

	Does the work of the "stop" and "update" commands; the
	  "stopF" flag is passed to imply that the mail client is
	  stopping (i.e., it's the "stop" command).  Even if stopF
	  if FALSE, the command text itself may also indicate a
	  mail client stoppage.

*/

static STSCODE
mccmd_update ARGLIST( (mccP, stopF) )
   NFARG( MCCMD		*mccP )		/* Command context block */
    FARG( BOOL		stopF )		/* Stop by default */
{
	STSCODE		status;
	BOOL		skipF;		/* Skipping to semicolon */
	BOOL		dataF;		/* If DATA attempted */
	BOOL		doneF;		/* If agent is done */
	BOOL		rgovF;		/* Whether to apply rcpt-gov */
	BOOL		conndF;		/* If we got connection data */
	BOOL		cd_greetF;	/* Connd: was issued a greeting */
	BOOL		cd_validF;	/* Connd: probably valid connection */
	int		n;
	int		rcptgov;	/* Copy of rcpt-to governor */
	MCNAME_S	mcns;		/* Client name */
	MCLIENT		*mcP;		/* Mail client from db */
	MCOCCUR		*mcoccP;	/* Occurance entry */
	UBYTE		tok[ 100 ];
	UBYTE		tok1[ 100 ];
	UBYTE		tok2[ 100 ];

    /* Get mail client name involved */
    if ( ! mccmd_parse_mcname( mccP, &mcns ) )
	return ( SCNOTDONE );

    mcP = mcdb_mclient_access( &mcns.mns_mcname );
    if ( mcP == NULL ) {
	mccmd_return( mccP, "ERROR - can't access client." );
	return ( SCNOTDONE );
    }

    rcptgov = mcP->mc_rcptgov;

    status = SCOK;
    dataF = FALSE;
    doneF = FALSE;
    conndF = FALSE;

    /* Parse the updating message; stop when we see END */
    skipF = TRUE;			/* Skip to first semi first */
    for ( ; ; ) {
	if ( ! mccmd_token( mccP, &tok[0], sizeof(tok) ) ) {
	    /* Token too large; skip to the next semi. */
	    skipF = TRUE;
	    continue;			/* too large, warning given */
	}
	if ( tok[0] == NUL )
	    break;			/* No more tokens */
	if ( skipF ) {
	    /* Skipping to semi... */
	    if ( ( tok[0] != ';' ) || ( tok[1] != NUL ) ) {
		warning( "mccmd_update: expected semi, got \"%s\"",
			&tok[0] );
	    }
	    else
		skipF = FALSE;		/* Found it */
	    continue;
	}

	/* Toggle the skip flag back on; anything else handled here (valid
	   or not) will need to look for semicolon.
	*/
	skipF = TRUE;

	/* Process token.  Check end first.. */
	if ( stricmp( (char *)&tok[0], "END" ) == 0 )
	    break;

	else if ( stricmp( (char *)&tok[0], "CONND" ) == 0 ) {
	    /* An interior series of keywords addressing how the
	       mail client behaved during its connection.
	       Loop on the keywords thru ';' or end of data.
	    */
	    skipF = FALSE;		/* Entirely handled here. */

	    conndF = TRUE;
	    cd_greetF = FALSE;
	    cd_validF = FALSE;

	    for ( ; ; ) {
		if ( ! mccmd_token( mccP, &tok1[0], sizeof(tok1) ) )
		    break;
		if ( ( tok1[0] == ';' ) && ( tok1[1] == NUL ) )
		    break;

		if ( stricmp( (char *)&tok1[0], "GREET" ) == 0 ) {
		    /* Client was issued a greeting */
		    cd_greetF = TRUE;
		}
		else if ( stricmp( (char *)&tok1[0], "PGDISC" ) == 0 ) {
		    /* Client disconnected before greeting */
		    ;
		}
		else if ( stricmp( (char *)&tok1[0], "PGTALK" ) == 0 ) {
		    /* Client talked before greeting */
		    ;
		}
		else if ( stricmp( (char *)&tok1[0], "REFUSED" ) == 0 ) {
		    /* Agent didn't want to talk to this client */
		    ;
		}
		else if ( stricmp( (char *)&tok1[0], "VALID" ) == 0 ) {
		    /* Agent thinks the connection was valid */
		    cd_validF = TRUE;
		}
		else {
		    /* Unknown connection note */
		    warning( "mccmd_update: unknown CONND item \"%s\"",
			    &tok1[0] );
		}
	    }
	}

	else if ( stricmp( (char *)&tok[0], "DATA" ) == 0 ) 
	    /* Mail client gave DATA */
	    dataF = TRUE;

	else if ( stricmp( (char *)&tok[0], "DONE" ) == 0 )
	    /* Agent is done. */
	    doneF = TRUE;

	else if ( stricmp( (char *)&tok[0], "RCPT" ) == 0 ) {
	    /* Sequence (or aggregation) of RCPT validity */
	    for ( ; ; ) {
		if ( ! mccmd_token( mccP, &tok1[0], sizeof(tok1) ) )
		    continue;
		if ( ( tok1[0] == NUL ) ||
		     ( ( tok1[0] == ';' ) && ( tok1[1] == NUL ) ) ) {
		    skipF = FALSE;
		    break;
		}
		if ( tok1[0] == '-' ) {
		    /* Count of invalid RCPT TO */
		    if ( tok1[1] == NUL )
			n = 1;
		    else if ( ( n = atod( &tok1[1] ) ) < 0 )
			n = 0;

		    /* Accumulate lifetime counter */
		    mcP->mc_brcptC += n;

		    /* Apply the bad increment */
		    n *= Rgov_bad_incr;
		    rcptgov += n;
		    if ( rcptgov > Rgov_bad_limit )
			rcptgov = Rgov_bad_limit;
		}
		else if ( tok1[0] == '+' ) {
		    /* Count of valid RCPT TO */
		    if ( tok1[1] == NUL )
			n = 1;
		    else if ( ( n = atod( &tok1[1] ) ) < 0 )
			n = 0;

		    /* Accumulate lifetime counter */
		    mcP->mc_grcptC += n;

		    /* Apply the goodfactor and decay the governor */
		    n *= Rgov_good_incr;
		    rcptgov += n;
		    if ( rcptgov < Rgov_good_limit )
			rcptgov = Rgov_good_limit;
		}
		else if ( tok1[0] == '=' ) {
		    /* Count of accepted foreign RCPT TO */
TODO		    /* Currently ignored */
		    ;
		}
		else {
		    warning( "mccmd_update: invalid \"%s\" in RCPT.",
			      &tok1[0] );
		    break;
		}
	    }
	}

	else if ( stricmp( (char *)&tok[0], "STOP" ) == 0 )
	    /* Mail client is stopping */
	    stopF = TRUE;

	else {
	    /* Unrecognized.  Ignore by skipping to semi. */
	    warning( "unrecognized \"%s\" in update.", &tok[0] );
	}
    }

    /* If we're here, we found the "END" or ran out of input. */

    /* Apply any connection analysis */
    if ( conndF ) {
	if ( cd_validF ) {
	    /* Connection was deemed valid (e.g. not just a probe). */

	    /* Update and/or set the PGD window */
	    if ( mcP->mc_pgd_wconC++ == 0 )
		time( &mcP->mc_pgd_wtime );
	}
    }

    /* Apply any rcpt governor triggers */
    rgovF = dataF;			/* Whether to act */
    if ( ( mcP->mc_flags & MCF_NO_RGOV ) != 0 )
	rgovF = FALSE;			/* Suppress acting on rcpt-gov */
    if ( rcptgov != mcP->mc_rcptgov )
	mclient_rcptgov_set( mcP, rcptgov, rgovF );

    /* If the agent is done, remove it and all its mail client connections */
    if ( doneF ) {
	mcdb_mclient_release( mcP, TRUE );
	mcoccur_agint_remove_all( mccP->mc_agentP );
    }

    else {
	/* If the mail client is stopping, remove it */
	if ( stopF ) {
	    /* Get occurance data for the mclient's name.  Note that this
	       may be different than the input name, e.g. for an aggregate.
	    */
	    mcoccP = mcoccur_access( &mcP->mc_mcname );
	    if ( mcoccP == NULL ) {
		mccmd_return( mccP,
			     "ERROR - can't get mail client occurance." );
		status = SCNOTDONE;
	    }
	    else {
		/* Disconnect the agent from the mail client occurance */
		mcoccur_agint_remove( mcoccP, mccP->mc_agentP );

		/* Do end-of-occurance processing */
		mcoccur_end( mcoccP, mcP );

		/* Fire off info messages to interested agents */
		mcoccur_notify( 0, agnotfmt_info,
			       mccP->mc_bsP,
			       mcoccP, mccP->mc_agentP, mcP );

		mcoccur_release( mcoccP );
	    }
	}

	mcdb_mclient_release( mcP, TRUE );
    }

    return ( status );
}
