/* mvmcas -- mvmf mail client assessment server

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

A mail client assessment server

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#include <signal.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "cdb.h"

#include "mvmf.h"
#include "thread.h"
#include "timer.h"
#include "mvmcas.h"
#include "mvmcas-version.h"
#include "controls.h"

#include "data.h"
#include "errwarn.h"


/* Local definitions */

    /* Specifies a service definition - a service to start up */
typedef struct {
    SERVE_TYPE	sd_type;		/* Type */
    char	*sd_nameP;		/* name to listen on (e.g. hostname) */
    long	sd_int1;		/* Int arg, if relevant (e.g. port) */
}  SERVEDEF;


/* External data referenced */

extern	CNTRLINT Ctrl_admin;		/* Control: admin */
extern	CNTRLINT Ctrl_pipe_allow;	/* Control: pipe_allow */
extern	int	Debug;			/* Debugging level */
extern	char	*MyhostnameP;		/* Our hostname */
extern	struct passwd	*MypwdP;	/* Our pwd info */
extern	OPMODE	Opmode;			/* Operating mode */
extern	int	Rgov_bad_incr;		/* Bad rgov increment */
extern	int	Rgov_bad_limit;		/* Bad rgov limit */
extern	int	Rgov_good_incr;		/* Good rgov increment */
extern	int	Rgov_good_limit;	/* Good rgov limit */
extern	long	Rgov_decay_int;		/* How often (seconds) to decay rgov */
extern	long	Rgov_decay_max;		/* Max interval to apply decay over */
extern	int	Rgov_decay_mult;	/* Decay rgov by multiplying by this */
extern	int	Rgov_decay_div;		/*   and then dividing by this */
extern	struct timeval Starttime;	/* Wall clock when we start */
extern	long	Tcp_connC;		/* TCP connections */
extern	long	Udp_connC;		/* UDP connections */
extern	char	*UsehostnameP;		/* Hostname to use (e.g. for greeting) */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Add a service to the list of servers to start up */
static	BOOL	add_service PROTO( (SERVE_TYPE type, char *nameP) );

    /* Read any startup file */
static	BOOL	read_start_file PROTO( (void) );

    /* Initialize stuff */
static	BOOL	mvmcas_init PROTO( (void) );

    /* Normal mode control */
static	STSCODE	mvmcas_normal PROTO( (void) );

static	void	usage();



/* Private data */

static	LL_HEAD	Serve_list;		/* Linked list of services to start */
static	BOOL	StartfileF;		/* If startfile was explicitly given */
static	char	*StartfileP;		/* Name of startup file */

int
main( argc, argv )
	int		argc;
	char		**argv;
{
	char		usageOK;
	int		optch;
	STSCODE		status;
	struct timezone	tz;		/* Dummy timezone */

    /* First thing: get resource info at program start. */
    gettimeofday( &Starttime, &tz );

    ll_init( &Serve_list );

    StartfileF = FALSE;
    StartfileP = MVMCAS_RCFILE;
    usageOK = TRUE;

    while ( usageOK &&
	   ( optch = getopt( argc, argv, "C:h:T:U:vx" ) ) != EOF ) {

	switch ( optch ) {
	    case	'C':		/* Configuration file */
	        StartfileP = optarg;
		StartfileF = TRUE;
		break;

	    case	'h':		/* My hostname to use */
		if ( UsehostnameP == NULL )
		    UsehostnameP = optarg;
		else {
		    fprintf( stderr, "Duplicate switch \"-h\".\n" );
		    usageOK = FALSE;
		}
		break;

	    case	'T':		/* Add a TCP service */
		usageOK = add_service( SERVE_TCP, optarg );
		break;

	    case	'U':		/* Add a UDP service */
		usageOK = add_service( SERVE_UDP, optarg );
		break;

	    case	'v':		/* Print version info */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-v\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_VERSION;
		break;

	    case	'x':		/* Enable debugging */
		++Debug;
		break;

	    default:
		usageOK = FALSE;
		break;
	}
    }

    /* If no operating mode selected, set to normal */
    if ( usageOK && ( Opmode == OM_NONE ) )
	Opmode = OM_NORMAL;

    /* No more command line arguments allowed */
    if ( usageOK && ( optind < argc ) )
	usageOK = FALSE;

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    if ( UsehostnameP == NULL )
	UsehostnameP = MyhostnameP;

    /* Proceed now depending on the mode */
    switch ( Opmode ) {
	case	OM_NORMAL:
	    if ( !mvmcas_init() )
		status = SCERR;
	    else
		status = mvmcas_normal();
	    break;

	case	OM_VERSION:
	    printf( "mvmcas version %s\n", MVMCAS_VERSION );
	    status = SCOK;
	    break;

	default:
	    warning( "Error: unknown Opmode %d.  This is a bug.", Opmode );
	    status = SCERR;
	    break;
    }

    exit( 0 );
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/


static void
usage()
{
    printf( "usage: mvmcas [flags]\n" );
    printf( "    Flags are:\n" );
    printf( "       -C file          Specify a startup file to use.  This is\n" );
    printf( "                          the full path to the startup file.\n" );
    printf( "       -h hostname      Specify my hostname to use.\n" );
    printf( "       -T host[:port]   Specify a TCP service to start.\n" );
    printf( "       -U host[:port]   Specify a UDP service to start.\n" );
    printf( "       -v               Give version information.\n" );
    printf( "       -x               Increment debugging level.\n" );
    printf( "\n" );
}
/*

*//* add_service( type, nameP )

	Add a service to the list of services we start up

Accepts :

	type		Service type
	nameP		Name string describing the service.

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
add_service ARGLIST( (type, nameP) )
   NFARG( SERVE_TYPE	type )		/* Service type */
    FARG( char		*nameP )	/* Service name */
{
	SERVEDEF	*defP;		/* Service definition block */
	char		*str1P;

    defP = (SERVEDEF *)emalloc( "add_service", "service block",
			         sizeof(SERVEDEF) );
    if ( defP == NULL )
	return ( FALSE );
    defP->sd_type = type;

    switch( type ) {
	case SERVE_TCP:			/* TCP service */
	case SERVE_UDP:			/* UDP service */
	    /* The string at nameP is
		    hostname[:port]
	       if port is not given, a default is used.
	    */

	    /* Copy the name so that we can write into it.  Will probably
	       waste some bytes here if :port is given.
	    */
	    defP->sd_nameP = newstr( "add_service", "sd_nameP", nameP );
	    if ( defP->sd_nameP == NULL ) {
		dealloc( defP );
		defP = NULL;
	    }
	    else {
		/* name is "name:port" -- "name" or "name:" will
		   default the port.
		*/
		str1P = strrchr( defP->sd_nameP, ':' );
		if ( str1P != NULL ) {
		    *str1P++ = NUL;
		    if ( *str1P == NUL )
			str1P = NULL;
		}
		if ( str1P != NULL )
		    defP->sd_int1 = atod( str1P );
		else
		    defP->sd_int1 = MVMCAS_PORT;
	    }
	    break;


	default:
	    warning( "add_service: bad service type %d, programming error.",
			 type );
	    dealloc( defP );
	    defP = NULL;
	    break;
    }

    if ( defP == NULL )
	return ( FALSE );

    ll_addvt( &Serve_list, defP );

    return ( TRUE );
}
/*

*//* read_start_file()

	Read the startup file

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
read_start_file NOARGLIST
{
	BOOL		okF;
	long		lv;
	SERVE_TYPE	servetype;
	FILE		*fP;		/* For reading */
	UBYTE		*cP;		/* Context in the input */
	UBYTE		tok[50];

    fP = fopen( StartfileP, "r" );
    if ( fP == NULL ) {
	/* No file is OK, unless specifically specified */
	if ( StartfileF )
	    warning( "read_start_file: can't open \"%s\"", StartfileP );
	return ( !StartfileF );
    }

    /* Use a generic BSTR for the input */
    for ( okF = TRUE; okF; ) {
	/* Get next line */
	if ( fgetbline( fP, Gbs1P ) < 0 )
	    break;

	cP = Gbs1P->bs_bP;

	if ( ! (okF = ws_token( &cP, &tok[0], sizeof(tok) ) ) )
	    break;

	/* Skip comments and blank lines */
	if ( ( tok[0] == NUL ) ||
	     ( stricmp( (char *)&tok[0], "#" ) == 0 ) ||
	     ( stricmp( (char *)&tok[0], "//" ) == 0 ) )
	    continue;

	else if ( stricmp( (char *)&tok[0], "rgov_bad_incr" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_bad_incr" );
	    else
		Rgov_bad_incr = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_bad_limit" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_bad_limit" );
	    else
		Rgov_bad_limit = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_decay_div" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_decay_div" );
	    else
		Rgov_decay_div = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_decay_int" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_decay_int" );
	    else
		Rgov_decay_int = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_decay_max" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_decay_max" );
	    else
		Rgov_decay_max = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_decay_mult" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_decay_mult" );
	    else
		Rgov_decay_mult = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_good_incr" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_good_incr" );
	    else
		Rgov_good_incr = lv;
	}

	else if ( stricmp( (char *)&tok[0], "rgov_good_limit" ) == 0 ) {
	    if ( !ws_token_int( &cP, &lv ) )
		warning( "Invalid input for rgov_good_limit" );
	    else
		Rgov_good_limit = lv;
	}

	else if ( stricmp( (char *)&tok[0], "serve" ) == 0 ) {
	    /* serve tcp|udp namestring */
	    if ( ! (okF = ws_token( &cP, &tok[0], sizeof(tok) ) ) )
		break;
	    if ( stricmp( (char *)&tok[0], "tcp" ) == 0 )
		servetype = SERVE_TCP;
	    else if ( stricmp( (char *)&tok[0], "udp" ) == 0 )
		servetype = SERVE_UDP;
	    else {
		warning( "read_start_file: unknown serve type \"%s\"",
			  &tok[0] );
		okF = FALSE;
		break;
	    }

	    if ( ! (okF = ws_token( &cP, &tok[0], sizeof(tok) ) ) )
		break;
	    okF = add_service( servetype, (char *)&tok[0] );

	    if ( *solbat( (char *)cP, FALSE ) != NUL )
		warning( "junk at end of line: %s", Gbs1P->bs_bP );
	}
	else {
	    warning( "read_start_file: unknown line: %s", Gbs1P->bs_bP );
	}
    }

    fclose( fP );

    return ( okF );
}
/*

*//* mvmcas_init()

	Do general initialization

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
mvmcas_init NOARGLIST
{
	BOOL		okF;
	char		path[ PATH_MAX +1 ];

    /* Initialize mvmcas data */
    okF = mvmcas_data_init();

    /* Initialize lib stuff */
    okF = okF && data_init();

    /* Define controls needed */
    if ( okF &&
         ( cntrl_define_str( estr_tmp_str( "log_private_name" ),
			     NULL, FALSE ) == NULL ) )
	okF = FALSE;

    if ( okF &&
         ( cntrl_define_str( estr_tmp_str( "log_program_name" ),
			     estr_tmp_str( "mvmcas" ), FALSE ) == NULL ) )
	okF = FALSE;

    /* Read any startup file */
    okF = okF && read_start_file();

    /* Make sure there are services defined */
    if ( okF && ( Serve_list.llh_firstP == NULL ) ) {
	warning( "no services defined." );
	okF = FALSE;
    }

    /* Initialize DNSBL */
    okF = okF && dnsbl_init();

    /* Initialize threads handling */
    okF = okF && thread_init();

    /* Initialize the various server modules */
    okF = okF && serve_tcp_init();
    okF = okF && serve_udp_init();

    /* Init timers module */
    okF = okF && timer_init();

    /* Initialize database interface */
    okF = okF && mcdb_init();

    /* Init client occurance module */
    okF = okF && mcoccur_init();

    return ( okF );
}
/*

*//* mvmcas_normal()

	Run in normal mode

Accepts :



Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmcas_normal NOARGLIST
{
	LL_NODE		*lnodeP;
	SERVEDEF	*defP;

    for ( lnodeP = Serve_list.llh_firstP;
	    lnodeP != NULL;
	       lnodeP = lnodeP->lln_nextP ) {
	defP = (SERVEDEF *)lnodeP->lln_valP;
	switch( defP->sd_type ) {
	    case SERVE_TCP:
		serve_tcp( defP->sd_nameP, defP->sd_int1 );
		break;

	    case SERVE_UDP:
		serve_udp( defP->sd_nameP, defP->sd_int1 );
		break;

	    default:
		warning( "unknown service type %d -- programming error.",
			  defP->sd_type );
		return ( SCNOTDONE );
	}
    }

    /* Idle ... service threads will take care of things. */
    for ( ; ; ) {
	sleep( 60 );
    }

    return ( SCNOTDONE );
}
