/* serve-tcp.c -- Serve tcp requests

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Responsible for taking TCP/IP requests and handling them.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <ctype.h>
#include <pthread.h>
#include <time.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-bio.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "mvmcas-command.h"
#include "mcoccur.h"
#include "thread.h"
#include "serve-tcp.h"


/* Local definitions */

    /* The control structure for the main tcp_service thread */
typedef struct {
    int		ts_fd;			/* fd to accept connections on */
}  TCP_SERVE;


    /* This is the control structure for servicing of a particular client. */
typedef struct tcp_client {
    struct tcp_client *tc_nextP;	/* For free list management */
    int		tc_fd;			/* The fd for the client */
    struct sockaddr tc_sock;		/* Socket data for client */
    int		tc_sockL;		/* Length of socket data */
    MML_IN	*tc_inP;		/* Input stack */
    BIO		tc_in;			/* Input stream */
    BIO		tc_out;			/* Output stream */
    BSTR	*tc_ibufP;		/* Input buffer */
}  TCP_CLIENT;



/* External data referenced */

extern	int	Debug;			/* Debugging level */
extern	long	Tcp_connC;		/* TCP connections */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Agent control functions */
static	void	agent_tcp_delinfo();
static	BOOL	agent_tcp_origstr();
static	void	agent_tcp_send();

    /* Serve a TCP client */
static	void	*serve_tcp_client PROTO( (TCP_CLIENT *tcP) );

    /* The service loop: get new client connections */
static	void	*serve_tcp_loop PROTO( (TCP_SERVE *tsP) );

    /* Get a client block */
static	TCP_CLIENT *tcb_get PROTO( (void) );

    /* Release a client block */
static	void	tcb_release PROTO( (TCP_CLIENT *ucP) );


/* Private data */

static	TCP_CLIENT	*Tcb_listP;	/* List of free client blocks */
static	pthread_mutex_t	Tcb_lock = PTHREAD_MUTEX_INITIALIZER;
static	AGENT_CTL	Tcp_agent;  /* Agent control block */

    /* Generic lock for quick operations herein */
static	pthread_mutex_t	Tcp_genlock = PTHREAD_MUTEX_INITIALIZER;

/*

*//* serve_tcp( hostP, portN )

	Initialize and fire up the TCP request service.

Accepts :

	hostP		The host name string to accept TCP requests on
	portN		The port number to accept TCP requests on

Returns :

	<status>	TRUE if service started successfully
			FALSE if not

Notes :

	Does the initial setup and then starts a thread which
	does the service loop.  That thread, in turn, will start
	a new thread for each connection.

	hostP may be specified as NULL to bind to all.

*/

BOOL
serve_tcp ARGLIST( (hostP, portN) )
   NFARG( char		*hostP )	/* Host or IP string */
    FARG( int		portN )		/* Port number */
{
	int		sfd;
	TCP_SERVE	*tsP;		/* Handle for the service */
	struct hostent	*hP;		/* For the host */
	struct sockaddr_in sin;		/* For the socket */

    if ( Debug > 0 )
	fprintf( stderr, "serve_tcp: start host \"%s\" port %d\n",
			  hostP, portN );

    /* Might as well check this first */
    if ( hostP != NULL ) {
	hP = gethostbyname( hostP );
	if ( hP == NULL ) {
	    warning( "serve_tcp unknown host \"%s\"", hostP );
	    return ( FALSE );
	}
    }

    /* Grab a socket */
    sfd = socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
    if ( sfd < 0 ) {
	warning( "serve_tcp can't get socket, errno=%d", errno );
	return ( FALSE );
    }

    /* Fill in host info */
    memset( &sin, 0, sizeof(sin) );
    if ( hostP != NULL ) {
	sin.sin_family = hP->h_addrtype;
	memcpy( (char *)&sin.sin_addr, hP->h_addr, hP->h_length );
    }
    else {
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl( INADDR_ANY );
    }

    /* Fill in port */
    sin.sin_port = htons( portN );

    /* bind the socket */
    if ( bind( sfd, (struct sockaddr *)&sin, sizeof(sin) ) < 0 ) {
	warning( "serve_tcp could not bind to port %d, errno=%d",
		    portN, errno );
	close( sfd );
	return ( FALSE );
    }
    listen( sfd, 0 );

    /* Grab a data structure for our use */
    tsP = (TCP_SERVE *)emalloc( "serve_tcp", "tsP", sizeof(TCP_SERVE) );
    if ( tsP == NULL ) {
	close( sfd );
	return ( FALSE );
    }

    /* Fill it in for use */
    tsP->ts_fd = sfd;

    /* Fire up the service thread */
    thread_start( serve_tcp_loop, tsP );

    return ( TRUE );
}
/*

*//* serve_tcp_init()

	Initialize the TCP service module

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
serve_tcp_init NOARGLIST
{
    /* Set up the agent control for this service */
    Tcp_agent.ac_delinfo = agent_tcp_delinfo;
    Tcp_agent.ac_send = agent_tcp_send;
    Tcp_agent.ac_origstr = agent_tcp_origstr;

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* agent_tcp_delinfo( agentP )

	Delete the info from an agent

Accepts :

	agentP		Agent to delete the info for

Returns :

	<nothing>

Notes :

	Handler for the UDP Agent control block, with interface as
	specified by the agent handling stuff.

*/

static void
agent_tcp_delinfo ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* The agent */
{
    /* This function is a no-op */

    /* done */
}
/*

*//* agent_tcp_origstr( agentP, bufP, bufL )

	Provide a string representing the agent's origin

Accepts :

	agentP		Ptr to agent block
	bufP		Where to store the string
	bufL		Length of bufP

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Handler for the TCP Agent control block, with interface as
	specified by the agent handling stuff.

*/

static BOOL
agent_tcp_origstr ARGLIST( (agentP, bufP, bufL) )
   NFARG( AGENT		*agentP )	/* The agent */
   NFARG( UBYTE		*bufP )		/* Buffer for the string */
    FARG( int		bufL )		/* Size of the buffer */
{
	TCP_CLIENT	*tcP;

    tcP = (TCP_CLIENT *)agentP->ag_infoP;

    if ( snprintf( (char *)bufP, bufL, "%s:%d", 
             inet_ntoa( ((struct sockaddr_in *)&tcP->tc_sock)->sin_addr ),
	     ntohs( ((struct sockaddr_in *)&tcP->tc_sock)->sin_port )
	          ) >= bufL )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* agent_tcp_send( agentP, bufP, bufL )

	Send a data buffer to an agent

Accepts :

	agentP		Ptr to agent block
	bufP		Ptr to data to send to the agent
	bufL		Length of the data to send

Returns :

	<nothing>

Notes :

	Handler for the TCP Agent control block, with interface as
	specified by the agent handling stuff.

	Messages to agents are sent blind, without immediate regard to
	error.

*/

static void
agent_tcp_send ARGLIST( (agentP, bufP, bufL) )
   NFARG( AGENT		*agentP )	/* The agent */
   NFARG( UBYTE		*bufP )		/* Buffer to send */
    FARG( int		bufL )		/* Length to send */
{
	TCP_CLIENT	*tcP;
	BIO		*bioP;

    tcP = (TCP_CLIENT *)agentP->ag_infoP;
    bioP = &tcP->tc_out;

    bio_putn( bioP, bufP, bufL );
    if ( bufP[ bufL -1] != '\n' )
	bio_putb( bioP, '\n' );
    bio_drain( bioP );
}
/*

*//* serve_tcp_client( tcP )

	TCP service for a client

Accepts :

	tcP		Control structure for the TCP/IP client

Returns :

	<value>		Some pointer, ignore it

Notes :

	This is the entry point for a thread that handles the request
	from a particular client.

	When finished, returns to the thread loop.

*/

static void *
serve_tcp_client ARGLIST( (tcP) )
    FARG( TCP_CLIENT	*tcP )		/* Client control */
{
	STSCODE		status;
	int		len;
	AGENT		*agentP;	/* Agent block */
	MCCMD		*mccP;		/* Command context */

    if ( Debug > 0 ) {
	warning( "Got TCP request from %s port %d.",
		inet_ntoa( ((struct sockaddr_in *)&tcP->tc_sock)->sin_addr),
		ntohs( ((struct sockaddr_in *)&tcP->tc_sock)->sin_port) );
    }

    status = SCOK;
    mccP = NULL;
    agentP = NULL;

    /* Set up the rest of the stuff we need in the client block */
    if ( ( tcP->tc_inP = in_new() ) == NULL )
	status = SCNOTDONE;
    else {
	bio_init( &tcP->tc_in, tcP->tc_fd );
	bio_init( &tcP->tc_out, tcP->tc_fd );
	if ( !in_stream_push_bio( tcP->tc_inP,
				  &tcP->tc_in, "tcp client input" ) )
	    status = SCNOTDONE;
    }

    if ( status == SCOK ) {
	if ( ! ( bio_putstr( &tcP->tc_out, (UBYTE *)"220 Ready\n" ) &&
	         bio_drain( &tcP->tc_out ) ) )
	    status = SCNOTDONE;
    }

    /* Make agent */
    agentP = pagent_create( &Tcp_agent, tcP );
    if ( agentP == NULL )
	status = SCERR;

    /* Get a command block */
    if ( status == SCOK ) {
	mccP = mvmcas_command_context_begin( agentP );
	if ( mccP == NULL )
	    status = SCERR;
    }


    if ( status == SCOK ) {
	/* Initial timeout */
	bio_timeout_set( &tcP->tc_in, 10, 0L );
	mccP->mc_statefulF = TRUE;
    }

    while ( status == SCOK ) {
	len = igetbtrec( tcP->tc_inP, tcP->tc_ibufP,
			 (TXR_END_CRLF)|(TXR_END_LF), NULL );
	if ( len < 0 ) {
	    if ( tcP->tc_in.bio_result == IOR_TIMEOUT ) {
		bio_putstr( &tcP->tc_out, "TEXT - Timed out.\n" );
	    }
	    else {
		bio_putstr( &tcP->tc_out, "TEXT - Input error.\n" );
	    }
	    bio_drain( &tcP->tc_out );
	    status = SCNOTDONE;
	    break;
	}

#if 0
	if ( len == 0 ) {
	    bio_putstr( &tcP->tc_out, "TEXT - End of input.\n" );
	    bio_drain( &tcP->tc_out );
	    break;
	}
#endif

	status = mvmcas_command( mccP, tcP->tc_ibufP->bs_bP );

#if 0
	/* If there's a direct response, return it */
	if ( mccP->mc_rstrP->bs_bC > 0 ) {
	    bio_putstr( &tcP->tc_out, mccP->mc_rstrP->bs_bP );
	    bio_drain( &tcP->tc_out );
	}
#endif

	/* Raise the timeout for non-first command */
	bio_timeout_set( &tcP->tc_in, 300, 0L );
    }

    if ( Debug > 0 ) {
	warning( "Ending TCP session with %s port %d.",
		inet_ntoa( ((struct sockaddr_in *)&tcP->tc_sock)->sin_addr),
		ntohs( ((struct sockaddr_in *)&tcP->tc_sock)->sin_port) );
    }

    /* Clean up */
    if ( mccP != NULL )
	mvmcas_command_context_end( mccP );

    if ( agentP != NULL ) {
	/* Remove any connections to mail clients.  This may be
	   redundant if any command did it already.
	*/
	mcoccur_agint_remove_all( agentP );

	pagent_delete( agentP );
    }

    if ( tcP->tc_inP != NULL )
	in_destroy( tcP->tc_inP );

    close( tcP->tc_fd );

    tcb_release( tcP );			/* Release client block */

    return ( NULL );			/* Return to the thread loop */
}
/*

*//* serve_tcp_loop( tsP )

	The tcp service loop

Accepts :

	tsP		Control structure for the TCP/IP service loop

Returns :

	<value>		Some opaque pointer, but ignore it

Notes :

	This thread listens for new client connections for a particular
	listening configuration.

	When done, it returns back to the thread loop .

*/

static void *
serve_tcp_loop ARGLIST( (tsP) )
    FARG( TCP_SERVE	*tsP )		/* Service control */
{
	TCP_CLIENT	*tcP;		/* client info */

    /* Loop taking requests */
    for ( ; ; ) {
	tcP = tcb_get();		/* Get client block */
	if ( tcP == NULL )
	    break;

	tcP->tc_sockL = sizeof( tcP->tc_sock );
	tcP->tc_fd = accept( tsP->ts_fd, &tcP->tc_sock, &tcP->tc_sockL );

	if ( tcP->tc_fd < 0 ) {
	    warning( "serve_tcp_loop accept failed, errno=%d", errno );
	    tcb_release( tcP );
	    break;
	}

#if 0	/* It's probably OK to miss a count now and then. */
	if ( pthread_mutex_lock( &Tcp_genlock ) != 0 )
	    warning( "serve_tcp_loop: can't get Tcp_genlock" );
#endif
	++Tcp_connC;
#if 0
	if ( pthread_mutex_unlock( &Tcp_genlock ) != 0 )
	    warning( "serve_tcp_loop: can't release Tcp_genlock" );
#endif

	if ( thread_start( serve_tcp_client, tcP ) < 0 ) {	
	    warning( "serve_tcp_loop can't start serve_tcp_client thread." );
	    tcb_release( tcP );
	    break;
	}
    }

    return ( NULL );			/* Back to thread loop */
}
/*

*//* tcb_get()

	Get a TCP client block

Accepts :



Returns :

	<value>		PTR to TCP_CLIENT block
			NULL if error

Notes :

	Keeps free client blocks on a free list, which is tried first.

*/

static TCP_CLIENT *
tcb_get NOARGLIST
{
	TCP_CLIENT	*tcP;		/* Client block pointer */

    /* Try the free list first */
    if ( pthread_mutex_lock( &Tcb_lock ) != 0 )
	warning( "tcb_get: can't get Tcb_lock" );
    if ( ( tcP = Tcb_listP ) != NULL ) {
	Tcb_listP = tcP->tc_nextP;
    }
    else {
	/* No free, allocate new one */
	tcP = (TCP_CLIENT *)emalloc( "tcb_get", "client block", 
				      sizeof(TCP_CLIENT) );
	if ( tcP == NULL ) {
	    warning( "tcb_get can't allocate client block." );
	}
	else {
	    tcP->tc_ibufP = bstr_new( "tcp client input", 1000, 1000 );
	    if ( tcP->tc_ibufP == NULL ) {
		warning( "tcb_get can't allocate input buffer" );
		dealloc( tcP );
		tcP = NULL;
	    }
	}
    }

    if ( pthread_mutex_unlock( &Tcb_lock ) != 0 )
	warning( "tcb_get: can't release Tcb_lock" );

    if ( tcP != NULL ) {
	/* Initialize some stuff */
	tcP->tc_inP = NULL;
    }

    return ( tcP );
}
/*

*//* tcb_release( tcP )

	Release a TCP client block

Accepts :

	tcP		Ptr to TCP_CLIENT block to release

Returns :

	<void>

Notes :

	Free client blocks are kept on a list for quick access.

*/

static void
tcb_release ARGLIST( (tcP) )
    FARG( TCP_CLIENT	*tcP )		/* Block to release */
{
    if ( pthread_mutex_lock( &Tcb_lock ) != 0 )
	warning( "tcb_release: can't get Tcb_lock" );

    tcP->tc_nextP = Tcb_listP;
    Tcb_listP = tcP;

    if ( pthread_mutex_unlock( &Tcb_lock ) != 0 )
	warning( "tcb_release: can't get Tcb_lock" );
}
