/* serve-udp.c -- Serve udp request

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Responsible for taking UDP requests and handling them.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <ctype.h>
#include <time.h>
#include <pthread.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "mvmcas-command.h"
#include "mcoccur.h"
#include "thread.h"
#include "serve-udp.h"


/* Local definitions */

    /* The control structure for the main udp_service thread */
typedef struct {
    int		us_fd;			/* fd to accept connections on */
}  UDP_SERVE;


    /* This is the control structure for servicing of a particular client. */
typedef struct udp_client {
    struct udp_client *uc_nextP;	/* For free list management */
    struct sockaddr uc_sock;		/* Socket data for client */
    int		uc_sockL;		/* Length of socket data */
    BSTR	*uc_ibufP;		/* Input buffer */
}  UDP_CLIENT;



/* External data referenced */

extern	int	Debug;			/* Debugging level */
extern	long	Udp_connC;		/* UDP connections */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Agent control functions */
static	void	agent_udp_delinfo();
static	BOOL	agent_udp_origstr();
static	void	agent_udp_send();

    /* Serve a UDP client */
static	void	*serve_udp_client PROTO( (UDP_CLIENT *ucP) );

    /* The service loop: get new client connections */
static	void	*serve_udp_loop PROTO( (UDP_SERVE *tsP) );

    /* Get a client block */
static	UDP_CLIENT *ucb_get PROTO( (void) );

    /* Release a client block */
static	void	ucb_release PROTO( (UDP_CLIENT *ucP) );


/* Private data */

static	UDP_CLIENT	*Ucb_listP;	/* List of free client blocks */
static	pthread_mutex_t	Ucb_lock = PTHREAD_MUTEX_INITIALIZER;
static	AGENT_CTL	Udp_agent;  /* Agent control block */

static	int	Sock_ofd = -1;		/* Socket for output */
static	pthread_mutex_t Sock_ofd_lock = PTHREAD_MUTEX_INITIALIZER;

    /* Generic lock for quick operations herein */
static	pthread_mutex_t	Udp_genlock = PTHREAD_MUTEX_INITIALIZER;

/*

*//* serve_udp( hostP, portN )

	Initialize and fire up the UDP request service.

Accepts :

	hostP		The host name string to accept UDP requests on
	portN		The port number to accept UDP requests on

Returns :

	<status>	TRUE if service started successfully
			FALSE if not

Notes :

	Does the initial setup and then starts a thread which
	does the service loop.  That thread, in turn, will start
	a new thread for each connection.

	hostP may be specified as NULL to bind to all.

*/

BOOL
serve_udp ARGLIST( (hostP, portN) )
   NFARG( char		*hostP )	/* Host or IP string */
    FARG( int		portN )		/* Port number */
{
	int		sfd;
	UDP_SERVE	*usP;		/* Handle for the service */
	struct hostent	*hP;		/* For the host */
	struct sockaddr_in sin;		/* For the socket */

    if ( Debug > 0 )
	fprintf( stderr, "serve_udp: start host \"%s\" port %d\n",
			  hostP, portN );

    /* Might as well check this first */
    if ( hostP != NULL ) {
	hP = gethostbyname( hostP );
	if ( hP == NULL ) {
	    warning( "serve_udp unknown host \"%s\"", hostP );
	    return ( FALSE );
	}
    }

    /* Grab a socket */
    sfd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
    if ( sfd < 0 ) {
	warning( "serve_udp can't get socket, errno=%d", errno );
	return ( FALSE );
    }

    /* Fill in host info */
    memset( &sin, 0, sizeof(sin) );
    if ( hostP != NULL ) {
	sin.sin_family = hP->h_addrtype;
	memcpy( (char *)&sin.sin_addr, hP->h_addr, hP->h_length );
    }
    else {
	sin.sin_family = AF_INET;
	sin.sin_addr.s_addr = htonl( INADDR_ANY );
    }

    /* Fill in port */
    sin.sin_port = htons( portN );

    /* bind the socket */
    if ( bind( sfd, (struct sockaddr *)&sin, sizeof(sin) ) < 0 ) {
	warning( "serve_udp could not bind to port %d, errno=%d",
		    sfd, errno );
	close( sfd );
	return ( FALSE );
    }

    /* This will also be our output socket */
    Sock_ofd = sfd;

    /* Grab a data structure for our use */
    usP = (UDP_SERVE *)emalloc( "serve_udp", "usP", sizeof(UDP_SERVE) );
    if ( usP == NULL ) {
	close( sfd );
	return ( FALSE );
    }

    /* Fill it in for use */
    usP->us_fd = sfd;

    /* Fire up the service thread */
    thread_start( serve_udp_loop, usP );

    return ( TRUE );
}
/*

*//* serve_udp_init()

	Initialize the UDP service module

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
serve_udp_init NOARGLIST
{
    /* Set up the agent control for this service */
    Udp_agent.ac_send = agent_udp_send;
    Udp_agent.ac_origstr = agent_udp_origstr;
    Udp_agent.ac_delinfo = agent_udp_delinfo;

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* agent_udp_delinfo( agentP )

	Delete the info from an agent

Accepts :

	agentP		Agent to delete the info for

Returns :

	<nothing>

Notes :

	Handler for the UDP Agent control block, with interface as
	specified by the agent handling stuff.

*/

static void
agent_udp_delinfo ARGLIST( (agentP) )
    FARG( AGENT		*agentP )	/* The agent */
{
    /* This function is a no-op */

    /* done */
}
/*

*//* agent_udp_origstr( agentP, bufP, bufL )

	Provide a string representing the agent's origin

Accepts :

	agentP		Ptr to agent block
	bufP		Where to store the string
	bufL		Length of bufP

Returns :

	<value>		TRUE if OK
			FALSE if not

Notes :

	Handler for the UDP Agent control block, with interface as
	specified by the agent handling stuff.

*/

static BOOL
agent_udp_origstr ARGLIST( (agentP, bufP, bufL) )
   NFARG( AGENT		*agentP )	/* The agent */
   NFARG( UBYTE		*bufP )		/* Buffer for the string */
    FARG( int		bufL )		/* Size of the buffer */
{
	struct sockaddr_in *siP;

    siP = (struct sockaddr_in *)agentP->ag_infoP;

    if ( snprintf( (char *)bufP, bufL, "%s:%d",
		    inet_ntoa( siP->sin_addr ),
		    ntohs( siP->sin_port ) ) >= bufL )
	return ( FALSE );

    return ( TRUE );
}
/*

*//* agent_udp_send( agentP, bufP, bufL )

	Send a data buffer to an agent

Accepts :

	agentP		Ptr to agent block
	bufP		Ptr to data to send to the agent
	bufL		Length of the data to send

Returns :

	<nothing>

Notes :

	Handler for the UDP Agent control block, with interface as
	specified by the agent handling stuff.

	Messages to agents are sent blind, without immediate regard to
	error.

*/

static void
agent_udp_send ARGLIST( (agentP, bufP, bufL) )
   NFARG( AGENT		*agentP )	/* The agent */
   NFARG( UBYTE		*bufP )		/* Buffer to send */
    FARG( int		bufL )		/* Length to send */
{
	int		len;

#if 0
	int		fd;		/* handle to send with */

    /* Grab a socket to send a response with */
    fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
    if ( fd >= 0 ) {
	len = sendto( fd,
		      (char *)bufP, bufL,
		      0,
		      (struct sockaddr *)agentP->ag_infoP,
		         sizeof( struct sockaddr ) );
	close( fd );
    }
#else
    if ( pthread_mutex_lock( &Sock_ofd_lock ) != 0 )
	warning( "agent_udp_send: can't get Sock_ofd_lock" );
    if ( Sock_ofd < 0 ) {
	/* Grab a socket to send a response with */
	Sock_ofd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
	if ( Sock_ofd < 0 ) {
	    if ( pthread_mutex_unlock( &Sock_ofd_lock ) != 0 )
		warning( "agent_udp_send: can't release Sock_ofd_lock" );
	    return;
	}
    }

    len = sendto( Sock_ofd,
		  (char *)bufP, bufL,
		  0,
		  (struct sockaddr *)agentP->ag_infoP,
		  sizeof( struct sockaddr ) );

    if ( Debug > 0 ) {
	printf( "udp sent (%d): ", bufL );
	fwrite( bufP, 1, bufL, stdout );
	if ( ( bufL == 0 ) || ( bufP[bufL-1] != '\n' ) )
	    fputc( '\n', stdout );
    }

    if ( pthread_mutex_unlock( &Sock_ofd_lock ) != 0 )
	warning( "agent_udp_send: can't release Sock_ofd_lock" );
#endif

    if ( len != bufL )
	warning( "agent_udp_send: send error (len=%d bufL=%d)",
		  len, bufL );

    /* done */
}
/*

*//* serve_udp_client( ucP )

	UDP service for a client

Accepts :

	ucP		Control structure for the UDP client

Returns :

	<value>		Some pointer, ignore it

Notes :

	This is the entry point for a thread that handles the request
	from a particular client.

	When done, returns back to the thread loop.

*/

static void *
serve_udp_client ARGLIST( (ucP) )
    FARG( UDP_CLIENT	*ucP )		/* Client control */
{
	STSCODE		status;
	int		len;
	int		rlen;		/* Length to send */
	int		rfd;		/* Response fd */
	UBYTE		*rbufP;		/* Buffer to send */
	AGENT		*agentP;	/* Agent block */
	MCCMD		*mccP;		/* Command context */
	UBYTE		buf[100];	/* In case ... */


    if ( Debug > 0 ) {
	warning( "Got UDP request from %s port %d.",
		inet_ntoa( ((struct sockaddr_in *)&ucP->uc_sock)->sin_addr),
		ntohs( ((struct sockaddr_in *)&ucP->uc_sock)->sin_port) );
    }

    status = SCOK;
    mccP = NULL;

    /* Access this non-persistent agent */
    agentP = npagent_access( &Udp_agent,
			      (void *)&ucP->uc_sock,
			      sizeof( struct sockaddr_in ) );

    if ( agentP == NULL )
	status = SCERR;
    else {
	/* Get command block */
	mccP = mvmcas_command_context_begin( agentP );
	if ( mccP == NULL )
	    status = SCERR;
	else {
	    /* Process the command */
	    status = mvmcas_command( mccP, ucP->uc_ibufP->bs_bP );

	    /* Clean up the command */
	    mvmcas_command_context_end( mccP );
	}

	/* Release the agent */
	npagent_release( agentP );
    }

    ucb_release( ucP );			/* Free client block */

    return ( NULL );			/* Back to the thread loop. */
}
/*

*//* serve_udp_loop( usP )

	The udp service loop

Accepts :

	usP		Control structure for the UDP/IP service loop

Returns :

	<value>		Some opaque pointer, but ignore it

Notes :

	This thread listens for new client connections for a particular
	listening configuration.

	When finished, returns to the thread loop.

*/

static void *
serve_udp_loop ARGLIST( (usP) )
    FARG( UDP_SERVE	*usP )		/* Service control */
{
	int		len;
	UDP_CLIENT	*ucP;		/* client info */

    /* Loop taking requests */
    for ( ; ; ) {
	ucP = ucb_get();		/* Get client block */
	if ( ucP == NULL )
	    break;
	
	ucP->uc_sockL = sizeof( ucP->uc_sock );

	len = recvfrom( usP->us_fd,
		        ucP->uc_ibufP->bs_bP, ucP->uc_ibufP->bs_bL,
			MSG_WAITALL,
			&ucP->uc_sock, &ucP->uc_sockL );

	if ( len < 0 ) {
	    break;
	}

#if 0	/* It's probably OK to miss a count now and then. */
	if ( pthread_mutex_lock( &Udp_genlock ) != 0 )
	    warning( "serve_udp_loop: can't get Udp_genlock" );
#endif
	++Udp_connC;
#if 0
	if ( pthread_mutex_unlock( &Udp_genlock ) != 0 )
	    warning( "serve_udp_loop: can't release Udp_genlock" );
#endif

	ucP->uc_ibufP->bs_bC = len;
	bstr_nulterm( ucP->uc_ibufP );

	if ( thread_start( serve_udp_client, ucP ) < 0 ) {	
	    warning( "serve_udp_loop can't start serve_udp_client thread." );
	    ucb_release( ucP );
	    break;
	}
    }

    return ( NULL );			/* Back to thread loop. */
}
/*

*//* ucb_get()

	Get a UDP client block

Accepts :



Returns :

	<value>		PTR to UDP_CLIENT block
			NULL if error

Notes :

	Keeps free client blocks on a free list, which is tried first.

*/

static UDP_CLIENT *
ucb_get NOARGLIST
{
	UDP_CLIENT	*ucP;		/* Client block pointer */

    /* Try the free list first */
    if ( pthread_mutex_lock( &Ucb_lock ) != 0 )
	warning( "ucb_get: can't get Ucb_lock" );
    if ( ( ucP = Ucb_listP ) != NULL ) {
	Ucb_listP = ucP->uc_nextP;
    }
    else {
	/* No free, allocate new one */
	ucP = (UDP_CLIENT *)emalloc( "ucb_get", "client block", 
				      sizeof(UDP_CLIENT) );
	if ( ucP == NULL ) {
	    warning( "ucb_get can't allocate client block." );
	}
	else {
	    ucP->uc_ibufP = bstr_new( "udp client input", 1000, 1000 );
	    if ( ucP->uc_ibufP == NULL ) {
		warning( "ucb_get can't allocate input buffer" );
		dealloc( ucP );
		ucP = NULL;
	    }
	}
    }

    /* Clear out the sockaddr, just to wipe out any stray bytes. */
    if ( ucP != NULL )
	memset( &ucP->uc_sock, 0, sizeof(ucP->uc_sock) );

    if ( pthread_mutex_unlock( &Ucb_lock ) != 0 )
	warning( "ucb_get: can't release Ucb_lock" );

    return ( ucP );
}
/*

*//* ucb_release( ucP )

	Release a UDP client block

Accepts :

	ucP		Ptr to UDP_CLIENT block to release

Returns :

	<void>

Notes :

	Free client blocks are kept on a list for quick access.

*/

static void
ucb_release ARGLIST( (ucP) )
    FARG( UDP_CLIENT	*ucP )		/* Block to release */
{
    if ( pthread_mutex_lock( &Ucb_lock ) != 0 )
	warning( "ucb_release: can't get Ucb_lock" );

    ucP->uc_nextP = Ucb_listP;
    Ucb_listP = ucP;

    if ( pthread_mutex_unlock( &Ucb_lock ) != 0 )
	warning( "ucb_release: can't release Ucb_lock" );
}
