/* thread.c -- Threads support for mvmcas

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Supplies some glue routines for managing threads

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <errno.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/stat.h>
#include <time.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-dns.h>
#include <mml/mml-ll.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "thread.h"


/* Local definitions */


#ifndef	MML_PTHREADS
# error	MML_PTHREADS is required!
#endif	/* MML_PTHREADS */

    /* An internal thread object */
typedef struct {
    pthread_t	to_thread;		/* Thread control */
    pthread_attr_t to_attr;		/* Just in case */
    pthread_cond_t to_go;		/* Tells the thread to go */
    pthread_mutex_t to_go_lock;		/* A lock for the go condition */
    void	*(*to_rtc)();		/* Routine to call */
    void	*to_argP;		/* Argument for the rtc */
}  THREAD;


/* External data referenced */

extern	int	Debug;			/* Debugging level */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* The thread loop for each thread */
static	void	*thread_loop PROTO( (void *toP) );


/* Private data */

static	LL_HEAD	Ready_list;		/* Threads that are ready to go */

/*

*//* thread_init()

	Initialize the threads module

Accepts :



Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
thread_init NOARGLIST
{
    /* Initialize the threads list */
    ll_init( &Ready_list );

    return ( TRUE );
}
/*

*//* thread_start( rtc, argP )

	Starts a new thread

Accepts :

	rtc		The function to call for the new thread
	argP		An argument to pass to the new thread

Returns :

	<value>		>= 0:  some info
			-1 error

Notes :


*/


int
thread_start ARGLIST( (rtc, argP) )
   NFARG( void		*(*rtc)() )	/* The routine to call */
    FARG( void		*argP )		/* Argument to give to thread */
{
	int		i;
	LL_NODE		*lnodeP;	/* Linked list node */
	THREAD		*toP;		/* Thread object */


    /* Let's see if there's a thread on the ready list */
    if ( pthread_mutex_lock( &Ready_list.llh_lock ) != 0 )
	warning( "thread_start: can't get ready list lock" );
    if ( ( lnodeP = Ready_list.llh_firstP ) != NULL ) {
	/* There's one on the ready list, grab it and remove it */
	toP = (THREAD *)lnodeP->lln_valP;
	ll_rem_nolock( lnodeP );
    }
    else
	toP = NULL;

    if ( pthread_mutex_unlock( &Ready_list.llh_lock ) != 0 )
	warning( "thread_start: can't release ready list lock" );

    /* If we found a ready thread, activate it.  */
    if ( toP != NULL ) {
	if ( pthread_mutex_lock( &toP->to_go_lock ) != 0 )
	    warning( "thread_start: can't get thread lock" );
	toP->to_rtc = rtc;		/* Tell the thread where to go */
	toP->to_argP = argP;		/*  and what to take with it.. */

	/* Tell it to go. */
	if ( Debug > 1 )
	    warning( "waking thread %lx\n", toP->to_thread );
	if ( pthread_cond_signal( &toP->to_go ) != 0 ) {
	    warning( "thread_start: pthread_cond_signal error %d", errno );
	}
	if ( pthread_mutex_unlock( &toP->to_go_lock ) != 0 )
	    warning( "thread_start: can't release thread lock" );
    }
    else {
	/* There is no ready thread.  Create a new one. */
	toP = (THREAD *)emalloc( "thread_start", "thread object",
				sizeof( THREAD ) );
	if ( toP == NULL ) 
	    return ( -1 );

	if ( pthread_attr_init( &toP->to_attr ) != 0 )
	    warning( "thread_start: attr_init failed for thread attr" );
	if ( pthread_attr_setdetachstate( &toP->to_attr,
					 PTHREAD_CREATE_DETACHED ) != 0 )
	    warning( "thread_start: failed to set detached state" );

	if ( pthread_cond_init( &toP->to_go, NULL ) != 0 )
	    warning( "thread_start: condition init failed." );
	if ( pthread_mutex_init( &toP->to_go_lock, NULL ) != 0 )
	    warning( "thread_start: can't init new thread lock" );

	/* Start the thread by giving it to the thread loop. */
	toP->to_rtc = rtc;
	toP->to_argP = argP;
	i = pthread_create( &toP->to_thread, &toP->to_attr,
			    thread_loop, toP );
	if ( i != 0 ) {
	    /* Thread failed, not much we can do */
	    pthread_attr_destroy( &toP->to_attr );
	    pthread_cond_destroy( &toP->to_go );
	    pthread_mutex_destroy( &toP->to_go_lock );

	    dealloc( toP );
	    return ( -1 );
	}
    }

    return ( 0 );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* thread_loop( toP )

	A loop that a thread sits in

Accepts :

	toP		The thread object.

Returns :

	<value>		A pointer to NULL

Notes :

	All threads go into the thread loop, where they repeatedly
	  run and wait for another task.  The task is given to them
	  in their thread object, which is the argument passed.

	Is invoked via pthread_create, which dictates the call
	  and return conventions.

*/

static void *
thread_loop ARGLIST( (argP) )
    FARG( void		*argP )		/* The thread pointer */
{
	int		i;
	THREAD		*toP;

    /* Contortions to satisfy function prototypes */
    toP = (THREAD *)argP;

    /* Setup anything that needs setting up */
    toP->to_thread = pthread_self();

    if ( Debug > 1 )
	warning( "thread_loop starting thread %lx\n", toP->to_thread );

    /* Loop to run the task and wait for another. */
    for ( ; ; ) {
	/* Do the task (invoke the rtc) */
	(*toP->to_rtc)( toP->to_argP );

	/* We're done-- add this to the end of the ready list */
	if ( pthread_mutex_lock( &toP->to_go_lock ) != 0 )
	    warning( "thread_loop: lock failed." );
	if ( ll_addvt( &Ready_list, toP ) == NULL ) {
TODO	    /* Probably need some error message here */
	    break;
	}

	/* Wait for another task.  Loop 'til valid wakeup. */
	toP->to_rtc = NULL;
	for ( ; ; ) {
	    i = pthread_cond_wait( &toP->to_go, &toP->to_go_lock );
	    if ( i != 0 ) {
		warning( "thread_loop: pthread_cond_wait returned %d thread %lx",
			  i, toP->to_thread );
	    }
	    else if ( toP->to_rtc != NULL )
		break;
	}
	if ( pthread_mutex_unlock( &toP->to_go_lock ) != 0 )
	    warning( "thread_loop: pthread_mutex_unlock failed." );

	/* Ready to go, so loop again.  Note that the thread handler
	   that woke this thread up will have removed it from the ready
	   list.
	*/
    }

    /* Here to die. */
    pthread_exit( NULL );
    return ( NULL );
}
