/* timer.c -- timer functions for mvmcas

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Stuff that is done on a timed basis.

*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <pthread.h>
#include <ctype.h>
#include <sys/stat.h>
#include <time.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-dns.h>

#include "mvmf.h"
#include "mvmcas.h"
#include "thread.h"
#include "timer.h"

/* Local definitions */

    /* A timer event node, i.e. in the timer event queue */
typedef struct {
    time_t	te_when;	/* When it happens */
    void	(*te_what)();	/* What happens */
    void	*te_argP;	/* 'What' gets this */
}  TIMER_EVENT;

/* External data referenced */

extern	int	Debug;			/* Debugging level */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* The timer service thread; process the timer queue */
static	void	*timer_loop PROTO( (void *argP) );


/* Private data */

static	LL_HEAD	Tqueue;			/* Timer queue */


/*

*//* timer_add( when, what, argP )

	Add a timer event

Accepts :

	when		time_t time when the event should happen
	what		routine-to-call
	argP		Some opaque argument to give to 'what' 

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	No de-duplication of 'what' is attempted.

*/

BOOL
timer_add ARGLIST( (when, what, argP) )
   NFARG( time_t	when )		/* When to fire */
   NFARG( void		(*what)() )	/* What to fire */
    FARG( void		*argP )		/* Something to give to what */
{
	LL_NODE		*lnodeP;	/* Linked list node */
	TIMER_EVENT	*teP;		/* New timer event */

    /* Create the new event info */
    teP = (TIMER_EVENT *)erualloc( "timer_add", "timer event",
				   sizeof(TIMER_EVENT) );
    if ( teP == NULL )
	return ( FALSE );
    teP->te_when = when;
    teP->te_what = what;
    teP->te_argP = argP;

    /* Need access to the queue */
    if ( pthread_mutex_lock( &Tqueue.llh_lock ) != 0 )
	warning( "timer_add: can't get thread queue lock" );

    /* Link the new event into the queue, in order of time value. */
    for ( lnodeP = Tqueue.llh_firstP; lnodeP != NULL;
			lnodeP = lnodeP->lln_nextP ) {
	if ( when < ((TIMER_EVENT *)lnodeP->lln_valP)->te_when )
	    break;
    }

    /* Add new data before the node found (or at end, if none) */
    ll_addvb( &Tqueue, lnodeP, teP );

    if ( pthread_mutex_unlock( &Tqueue.llh_lock ) != 0 )
	warning( "timer_add: can't release thread queue lock" );


TODO	/* Should probably wake up the timer thread */

    return ( TRUE );
}
/*

*//* timer_init()

	Initialize the timer module

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
timer_init NOARGLIST
{
    ll_init( &Tqueue );

    thread_start( timer_loop, (void *)NULL );

    return ( TRUE );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* timer_loop( argP )

	The timer service thread

Accepts :

	argP		Currently nothing; thread interface requires it.

Returns :

	<value>		Some opaque pointer, but ignore it

Notes :

	This thread processes events in the timer queue when they are
	ready.

	When finished, it returns back to the thread loop.

*/

static void *
timer_loop ARGLIST( (argP) )
    FARG( void		*argP )		/* thread argument */
{
	int		t;		/* Time to sleep */
	time_t		timenow;
	LL_NODE		*lnodeP;	/* Linked list node */
	TIMER_EVENT	*teP;		/* New timer event */

    if ( Debug > 0 )
	warning( "timer_loop started." );

    /* Process timer events */
    for ( ; ; ) {
	/* Lock the queue so we can check it. */
	if ( pthread_mutex_lock( &Tqueue.llh_lock ) != 0 )
	    warning( "timer_loop: can't get thread queue lock" );
	lnodeP = Tqueue.llh_firstP;
	if ( lnodeP == NULL ) {
	    /* Nothing on the queue; sleep some default time. */
	    t = 10;
	}
	else {
	    time( &timenow );
	    teP = (TIMER_EVENT *)lnodeP->lln_valP;
	    if ( teP->te_when <= timenow ) {

		/* First thing on queue is ready.  Remove it (we
		   already have the list locked.)
		*/
		ll_rem_nolock( lnodeP );

		/* Can unlock the queue now */
		if ( pthread_mutex_unlock( &Tqueue.llh_lock ) != 0 )
		    warning( "timer_loop: can't release thread queue lock" );

		/* Process the timer */
		(*teP->te_what)(teP->te_argP);

		/* Done with this; release for quick re-use */
		derualloc( teP, sizeof(TIMER_EVENT) );

		/* Loop to try next item on the queue */
		continue;
	    }
	    else {
		/* Next thing is not ready yet.  Calculate sleep time */
		t = ( teP->te_when - timenow );
		if ( ( t <= 0 ) || ( t > 60 ) )
		    t = 60;
	    }
	}

	/* Unlock the queue and wait some more. */
	if ( pthread_mutex_unlock( &Tqueue.llh_lock ) != 0 )
	    warning( "timer_loop: can't release thread queue lock" );

	sleep( t );
    }

    return ( NULL );			/* Back to the thread loop */
}
