/* mvmda-command.c -- command processing stuff for mvmda

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Contains functions to process individual kinds of command lines,
as well as the driver code for command processing.

*/

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>

#include <mml/mml.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-ttio.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "mvmfmisc.h"
#include "mvmda.h"
#include "msgpart.h"
#include "omm.h"
#include "msgstore.h"
#include "mvmda-command.h"
#include "dnsbl.h"

#include "mfl.h"
#include "mfl-token.h"
#include "mfl-parse.h"
#include "mfl-sieve.h"
#include "mfl-exec.h"

#include "errwarn.h"

/* Local definitions */

    /* Flags for commands in command table */
#define	CF_INTER	0x01		/* For interactive use only */



/* External data referenced */

extern	BOOL	DebugF;			/* Debugging flag */
extern	BOOL	Mfl_checkF;		/* If check mode (parse only) */
extern	OMM	*OmmP;			/* Open mail message handle */
extern	OPMODE	Opmode;			/* Operating mode */
extern	char	Promptcont[];		/* Continuation prompt */
extern	char	Promptmain[];		/* Main prompt */

/* External routines used */

extern	char	*getenv();

/* Local data publicly available */



/* Local routines and forward references */

    /* The command driver */
static	STSCODE	cmd PROTO( (MML_IN *inP) );

    /* Command processing routines */
static	STSCODE	cmd_debug PROTO( (MML_IN *inP) );
static	STSCODE	cmd_echo PROTO( (MML_IN *inP) );
static	STSCODE	cmd_exit PROTO( (MML_IN *inP) );
static	STSCODE	cmd_help PROTO( (MML_IN *inP) );
static	STSCODE	cmd_parse PROTO( (MML_IN *inP) );
static	STSCODE	cmd_prompt PROTO( (MML_IN *inP) );
static	STSCODE	cmd_wrmail PROTO( (MML_IN *inP) );

    /* Check for end of line next. */
static	STSCODE	cmds_eol PROTO( (MML_IN *inP) );

    /* Get a command line token */
static	char	*cmds_token PROTO( (MML_IN *inP) );

/* Private data */

static	BOOL	CmtkeolF = FALSE;	/* Set if last token was EOL */

    /* Table for commands */
static struct {
    char	*cm_strP;		/* Command string */
    STSCODE	(*cm_rtc)();		/* Command handler routine to call */
    WORD	cm_flags;		/* Any flags */
    char	*cm_explP;		/* Short explanation string */
    }	Cmdtbl[] = {

    {	"debug",	cmd_debug,	0,
    		"Turn debugging on or off"		},

    {	"echo",		cmd_echo,	CF_INTER,
    		"Translate input to output"		},

    {	"exit",		cmd_exit,	0,
    		"Exit the program"			},

    {	"help",		cmd_help,	CF_INTER,
    		"Give a list of commands"		},

    {   "parse",	cmd_parse,	CF_INTER,
    		"Parse input and show resulting parse tree" },

    {	"prompt",	cmd_prompt,	CF_INTER,
    		"Set main and/or continuation prompts"	},

    {	"wrmail",	cmd_wrmail,	CF_INTER,
    		"Write a mail message"			},

    {   NULL,	NULL,	0		}
    		   };


/*

*//* mf_command(inP, sieveF)

	Input and process the next command

Accepts :

	inP		Ptr to input stream
	sieveF		Whether to start in Sieve mode

Returns :

	<value>		SCxx status code

Notes :

	If operating mode (Opmode) is "check" (syntax check),
	or if the global Mfl_checkF flag is set,
	the execution of parse trees and commands is inhibited.

*/

STSCODE
mf_command ARGLIST( (inP, sieveF) )
   NFARG( MML_IN	*inP )		/* Input stack */
    FARG( BOOL		sieveF )	/* Whether to start in Sieve mode */
{
	int		status;
	TOKEN		*tokP;		/* Ptr to token */
	PTNODE		*treeP;
	MML_INS		*instrP;	/* Ptr to input stream block */
	TTCMDSTR	*ttstrP;	/* Ptr to tty stream */

    /* Check for interactive prompting at the top level */
    if ( ( inP->in_level == 1 ) && ( Opmode == OM_INTERACTIVE ) ) {
	instrP = inP->in_streamP;
	ttstrP = (TTCMDSTR *)instrP->is_handleP;
	ttstrP->ts_cmlP->tc_contF = FALSE; /* Not continuation line */

	/* Read until something significant is in the input stream, and
	   push it back.  This is a cheezy way to set up the continuation
	   prompt.
	*/
	in_charback( inP, in_nonblank( inP, FALSE ) );
	ttstrP->ts_cmlP->tc_contF = TRUE; /* Now continuation line */
    }

    /* Precheck first token */
    tokP = mfl_token( inP );
    if ( tokP == NULL )
	status = SCERR;

    else if ( ( tokP->tk_type == TK_CHAR ) &&
	      ( tokP->tk_bstrP->bs_bP[0] == ':' ) )
        /* Built-in command */
	status = cmd( inP );

    else if ( tokP->tk_type == TK_EOL )
    	/* Ignore blank lines */
	status = SCOK;

    else {
	/* Assume to be some language syntax */

	/* Push back the token for reuse */
	mfl_tokback( inP, tokP );

	/* Parse the next thing */
	if ( sieveF )
	    status = parse_sieve( inP, &treeP,
				 SBIS_ACTION | SBIS_CONTROL | SBIS_TEST );
	else
	    status = mfl_parse( inP, &treeP );

	/* If successful parse, deal with it */
	if ( status == SCOK ) {
	    if ( Mfl_checkF || ( Opmode == OM_CHECK ) ) {
		/* Syntax check only: just delete the tree */
		parsetree_free( treeP );
	    }
	    else {
		/* Execute the parse tree (and delete it) */
		status = exec_tree( treeP, (Opmode == OM_INTERACTIVE) );
	    }
	}
    }

    return( status );
}





/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* cmd(inP)

	Input and process a command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	This is called whenever a command line is expected, e.g.
	when a command introducer is seen.

*/

static STSCODE
cmd ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;
	int		len;
	int		cmdX;		/* Command table index */
	int		matX;		/* Index of any match */
	char		*tstrP;		/* Token string */
	char		*nameP;		/* Command name ptr */

    /* Input command token */
    tstrP = cmds_token( inP );
    if ( tstrP == NULL )
	return ( SCOK );

    /* Regardless of token type, try to look it up */
    matX = -1;				/* No match */
    for( cmdX = 0; ( nameP = Cmdtbl[cmdX].cm_strP ) != NULL; ++cmdX ) {
	/* Skip interactive-only commands if not interactive mode */
	if ( ( Opmode != OM_INTERACTIVE ) &&
	     ( ( Cmdtbl[cmdX].cm_flags & CF_INTER ) != 0 ) )
	    continue;

	/* Exact match always wins */
        if ( stricmp( nameP, tstrP ) == 0 )
	    break;

	/* Otherwise try partial match */
	else if ( ( (len = strlen( tstrP ) ) < strlen( nameP ) ) &&
		  ( strnicmp( nameP, tstrP, len ) == 0 ) ) {
	    if ( matX == -1 )
	    	matX = cmdX;
	    else
	    	matX = -2;
	}
    }

    /* See if command seen */
    if ( nameP == NULL )
        cmdX = matX;

    if ( cmdX >= 0 )
        status = (*Cmdtbl[cmdX].cm_rtc)( inP );
    else {
    	warning( "%s command: %s", 
			   cmdX== -2 ? "ambiguous" : "invalid",
			   		tstrP );
	status = SCNOTDONE;
    }

    if ( status == SCNOTDONE ) {
    	/* Command not completed -- flush the command line. */
	for ( ; !CmtkeolF ; )
	    (void)cmds_token( inP );

	status = SCOK;			/* Nothing seriously wrong... */
    }

    return( status );
}
/*

*//* cmd_debug(inP)

	Input and process the "debug" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	debug on
		turns on debugging
	debug off
		turns off debugging


*/

static STSCODE
cmd_debug ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;
	BOOL		flag;		/* New flag value */
	char		*tstrP;		/* String */

    status = SCOK;

    /* Get the "on" or "off" keyword. */
    tstrP = cmds_token( inP );
    if ( ( tstrP != NULL ) &&
	 ( stricmp( tstrP, "on" ) == 0 ) )
    	flag = TRUE;
    else if ( ( tstrP != NULL ) &&
	      ( stricmp( tstrP, "off" ) == 0 ) )
    	flag = FALSE;
    else {
    	warning( "\"debug\" requires the word ON or OFF." );
	status = SCNOTDONE;
    }

    if ( status == SCOK )
	status = cmds_eol( inP );

    if ( ( status == SCOK ) && ( Opmode != OM_CHECK ) ) {
    	/* Set the flag */
	DebugF = flag;

	warning( "Debug is now %s", flag?"ON":"OFF" );
    }

    return( status );
}
/*

*//* cmd_echo(inP)

	Input and process the "echo" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	echo stuff to end of line token by token

	If debug flag is set, dumps info about the token.

 	There exists the danger that macro substitution will place a
	newline into the input stream, causing the echo to terminate
	prematurely.  This is not prevented in this implementation.

*/

static STSCODE
cmd_echo ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	BOOL		firstF;		/* If first one */
	TOKEN		*tokP;		/* Ptr to token */

    /* Loop getting tokens until end-of-line */
    for( firstF = TRUE; ; ) {
	tokP = mfl_token( inP );
	if ( ( tokP == NULL ) || ( tokP->tk_type == TK_EOL ) )
	    break;

	if ( Opmode == OM_CHECK )
	    continue;			/* Syntax check only */

	if ( DebugF ) {
	    /* Full-on debugging */
	    printf( "addr=0x%lx type=%d(%s) tok=\"%s\"\n", 
		           tokP,
		           tokP->tk_type, mfl_toktype(tokP->tk_type),
		           tokP->tk_bstrP->bs_bP );

	    /* Print raw equivalent if provided */
	    if ( tokP->tk_rawF )
		printf( "  raw=\"%s\"\n", tokP->tk_rawP->bs_bP );
	}
	else {
	    /* Just print the token */
	    if ( !firstF )
		putchar( ' ' );
	    else
		firstF = FALSE;

	    /* Print raw if given, otherwise print translated token */
	    if ( tokP->tk_rawF )
		printf( "%s", tokP->tk_rawP->bs_bP );
	    else
		printf( "%s", tokP->tk_bstrP->bs_bP );
	}
    }
    if ( Opmode != OM_CHECK )
	putchar( '\n' );

    return( SCOK );
}
/*

*//* cmd_exit(inP)

	Input and process the "exit" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	We simply return SCERR to stop command processing.

*/

static STSCODE
cmd_exit ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
    cmds_eol( inP );			/* Make sure end of line */
    if ( Opmode == OM_CHECK )
	return ( SCOK );		/* Syntax check only */

    return ( SCERR );			/* This will cause an exit */
}
/*

*//* cmd_help(inP)

	Input and process the "help" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

*/

static STSCODE
cmd_help ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	int		cmdX;		/* Command table index */
	char		*nameP;		/* Command name ptr */

    if ( Opmode != OM_CHECK ) {
	/* Dump the command table */
	for( cmdX = 0; ( nameP = Cmdtbl[cmdX].cm_strP ) != NULL; ++cmdX )
            printf( "%-10s %s\n", nameP, Cmdtbl[cmdX].cm_explP );
    }

    return( SCOK );
}
/*

*//* cmd_parse(inP)

	Input and process the "parse" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	parse <some stuff>
	    parses language construct, outputs parse tree

*/

static STSCODE
cmd_parse ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;		/* Result */
	BOOL		sieveF;
	PTNODE		*treeP;		/* Ptr to generated tree */

    /* Parse whatever follows */
    sieveF = cntrl_int_get( estr_tmp_str( "mfl_mode_sieve" ), 0 );
    if ( sieveF )
	status = parse_sieve( inP, &treeP,
				 SBIS_ACTION | SBIS_CONTROL | SBIS_TEST );
    else
	status = mfl_parse( inP, &treeP );

    if ( status == SCOK ) {
    	/* Display the parse tree */
	if ( Opmode != OM_CHECK )
	    parsetree_print( stdout, 0, treeP );

	/* Delete the parse tree */
	parsetree_free( treeP );
    }

    return( status );
}
/*

*//* cmd_prompt(inP)

	Input and process the "prompt" command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	prompt
	    displays current prompts

	prompt string [, cstring ]
	    sets new prompt string and optional continuation prompt

*/

static STSCODE
cmd_prompt ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;
	BOOL		mainF;		/* If main given */
	BOOL		contF;		/* If continuation given */
	char		*tstrP;		/* Token string */
	char		pmain[MFL_TOKSIZE];	/* Main prompt */
	char		pcont[MFL_TOKSIZE];	/* Continuation prompt */

    status = SCOK;
    mainF = contF = FALSE;

    tstrP = cmds_token( inP );
    if ( tstrP == NULL ) {
	if ( Opmode != OM_CHECK ) {
	    printf( "Main prompt is \"%s\"\n", &Promptmain[0] );
	    printf( "Continuation prompt is \"%s\"\n", &Promptcont[0] );
	}
    }
    else {
	strncpy( &pmain[0], tstrP, MFL_TOKSIZE );
	mainF = TRUE;
	tstrP = cmds_token( inP );

	/* Check for comma and secondary prompt */
	if ( ( tstrP != NULL ) &&
	     ( strcmp( tstrP, "," ) == 0 ) ) {

	    tstrP = cmds_token( inP );
	    if ( tstrP != NULL ) {
	        contF = TRUE;
		strcpy( &pcont[0], tstrP );
		tstrP = cmds_token( inP );
	    }
	}

	/* At this point, we should have input the end of line. */
	if ( tstrP == NULL ) {
	    if ( Opmode != OM_CHECK ) {
		if ( mainF )
	    	    strcpy( &Promptmain[0], &pmain[0] );
		if ( contF )
	    	    strcpy( &Promptcont[0], &pcont[0] );
	    }
	}
	else
	    status = cmds_eol( inP );
    }

    if ( status != SCOK )
    	warning( "syntax error in Prompt command." );

    return( status );
}
/*

*//* cmd_wrmail(inP)

	Write a mail message to a folder

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCxx status code

Notes :

	Internal function for command processing module.

	wrmail folder
	    Writes current mail message to folder

*/

static STSCODE
cmd_wrmail ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	STSCODE		status;
	char		*tstrP;		/* Token string */
	char		fname[ PATH_MAX +1 ];  /* Filename */

    status = SCNOTDONE;

    /* Get the filename */
    tstrP = cmds_token( inP );
    if ( tstrP == NULL )
	printf( "Filename required.\n" );
    else {
	strncpy( &fname[0], tstrP, PATH_MAX );
	cmds_eol( inP );		/* Skip to input */

	if ( Opmode != OM_CHECK ) {
	    if ( OmmP == NULL ) 
		warning( "There is no mail file open." );
	    else if ( msg_write( OmmP, &fname[0] ) )
		status = SCOK;
	}
    }

    return ( status );
}
/*

*//* cmds_eol(inP)

	Check for end-of-line next in stream

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		SCOK if end-of-line next
			SCNOTDONE if not.

Notes :

	Internal support function for command processing module.

	Makes sure the next thing in the input is an end-of-line.


*/

static STSCODE
cmds_eol ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	char		*tstrP;		/* Ptr to token string */

    /* Get end-of-line */
    tstrP = cmds_token( inP );
    if ( tstrP != NULL ) {
	warning( "end of line expected." );
	return( SCNOTDONE );
    }
 
    return ( SCOK );
}
/*

*//* cmds_token(inP)

	Get a token for command

Accepts :

	inP		Ptr to input stack

Returns :

	<value>		Ptr to token string if not EOL
			NULL if EOL

Notes :

	Internal support function for command processing module.


*/

static char *
cmds_token ARGLIST( (inP) )
    FARG( MML_IN	*inP )		/* Input stack */
{
	TOKEN		*tokP;		/* Ptr to token */

    /* Get next thing */
    tokP = mfl_token( inP );

    /* If no token or end of line, return null */
    if ( ( tokP == NULL ) || ( tokP->tk_type == TK_EOL ) ) {
	CmtkeolF = TRUE;		/* Saw EOL token */
	return ( NULL );
    }

    /* Some non-eol token- just return the pointer to it. */
    CmtkeolF = FALSE;
    return ( (char *)tokP->tk_bstrP->bs_bP );
}
