/* mvmda-mfl-bif.c -- MFL built-in-functions for mvmda

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Defines & implements MFL built-in function(s) for mvmda.

The stuff to run an mvmda/mfl script is essentially a clone of the
mfl_running stuff.  These are to support very old mvmda script files,
which have a mixture of pure MFL and mvmda commands (commands prefixed
by ':').  Only people who used mvmda before MFL was reasonably well
developed will want these things, and we won't really talk about them..

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <mml/mml.h>

#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>
#include <mml/mml-symtab.h>

#include "mvmf.h"
#include "controls.h"
#include "mfl.h"
#include "mfl-pp.h"
#include "mfl-token.h"
#include "mfl-type.h"
#include "mfl-parse.h"
#include "mfl-eval.h"
#include "mfl-func.h"
#include "mfl-misc.h"

#include "data.h"
#include "errwarn.h"

/* Local definitions */


/* External data referenced */

extern	BOOL	Mfl_checkF;		/* For MFL check mode */
extern	BOOL	Mfl_traceF;		/* Whether tracing */

/* External routines used */


/* Local routines and forward references */


    /* Built-in functions contained herein */
static	STSCODE	bif_mvmda_run_fname();

    /* Run mfl/mvmda code from a named file */
static	STSCODE	mvmda_mfl_run_fname PROTO( (VALUE *rvalP, ESTR *fnesP,
					    ESTR *msesP) );

    /* Run mfl/mvmda code from a file pointer */
static	STSCODE	mvmda_mfl_run_fp PROTO( (VALUE *rvalP, FILE *fP,
					 ESTR *msesP, char *fnameP) );

    /* Run mfl/mvmda code from an input stream */
static	STSCODE	mvmda_mfl_run_in PROTO( (VALUE *rvalP, MML_IN *inP,
					 ESTR *msesP) );


/* Local data publicly available */



/* Private data */


    /* Table of built-in functions to define */
static	BIFINF	Bifinf_mvmda[] = {
/*  function name               handler function		*/
{  "$mvmda_run_fname",	bif_mvmda_run_fname	},

{  NULL,		NULL			}
		};


/*

*//* bif_mvmda_init(args)

	Initialize the mvmda built-in-functions module

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

BOOL
bif_mvmda_init NOARGLIST
{
    /* Register our functions */
    if ( !mfl_func_bifreg( &Bifinf_mvmda[0] ) )
	return ( FALSE );

    return ( TRUE );
}


/***********************************************************************
 *                                                                     *
 *                     Internal built-in functions                     *
 *                                                                     *
 ***********************************************************************/

/* Built-in functions are called with standard arguments, like this:

     built-in-function( fnameP, rvalP, treeP )

	fnameP is the function name that was referenced
	rvalP is a VALUE: where the return value is placed.
	treeP is the address of the argument list.

  The function returns a STSCODE status code.

*/


/*

*//* bif_mvmda_run_fname -- Run mfl/mvmda code from a named file

    BOOL $mvmda_run_fname( string fname, string modes )

    The named file is opened and mfl inside it is run

    modes is a modes string per the mvmda_mfl_run_xxx functions.

*/

static STSCODE
bif_mvmda_run_fname ARGLIST( (fnameP, rvalP, treeP) )
   NFARG( char		*fnameP )	/* Function name */
   NFARG( VALUE		*rvalP )	/* Return value ptr */
    FARG( PTNODE	*treeP )	/* Argument tree */
{
	STSCODE		status;
	BOOL		execedF;	/* If we did the exec */
	ESTR		*fnesP;		/* The file name */
	ESTR		*msesP;		/* Modes string */
	VALUE		fnval, msval;	/* fname and modes string values */

    execedF = FALSE;

    /* Get the arguments */
    if ( ( fnesP = func_treearg_estr( fnameP, treeP, 0, &fnval ) ) == NULL ) {
	warning( "%s: first argument must be valid filename string.", fnameP );
	status = SCNOTDONE;
    }
    else {
	msesP = func_treearg_estr( fnameP, treeP, 1, &msval );
	if ( msesP == NULL ) {
	    warning( "%s: second argument must be valid modes string.",
			fnameP );
	    status = SCNOTDONE;
	}
	else {
	    /* Run the file */
	    execedF = TRUE;
	    status = mvmda_mfl_run_fname( rvalP, fnesP, msesP );

	    val_release( &msval );
	}
	val_release( &fnval );
    }

    if ( !execedF )
	val_set_int( rvalP, 4, 0 );

    return ( status );
}


/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/


/*

*//* mvmda_mfl_run_fname(rvalP, fnesP, msesP)

	Run mfl/mvmda code that's in a named file

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	fnesP		Name of the file to run
	msesP		Modes string

Returns :

	<value>		status code

Notes :

	Basically a clone of mfl_run_fname(), but using mvmda_mfl_run_fp
	  rather than mfl_run_fp.  

	msesP is a string with characters indicating some operational
	  modes.  An 'O' character indicates that any following lowercase
	  characters, up to the next uppercase character, affect how to deal
	  with file opening issues, as follows:

		'e'	Return done (SCDONE) if file can't be opened
		'w'	Give warning if file can't be opened

	msesP is also passed along to mvmda_mfl_run_fp().

	See notes at mvmda_mfl_run_fp() as well.

*/

STSCODE
mvmda_mfl_run_fname ARGLIST( (rvalP, fnesP, msesP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( ESTR		*fnesP )	/* Name of file to open */
    FARG( ESTR		*msesP )	/* Modes characters string */
{
	STSCODE		status;
	UBYTE		ch;
	int		len;
	BOOL		warnedF;
	BOOL		mrelF;		/* If mode char is relevant */
	UBYTE		*strP;
	FILE		*fP;		/* File pointer */
	char		fname[PATH_MAX +1 ]; /* Actual filename */

    /* Get a nul-terminated copy of fname, which is what fopen() needs. */
    len = estr_len( fnesP );
    if ( ( len + 1 ) >= sizeof(fname) ) {
	warning( "mvmda_mfl_run_fname: filename too long." );
	status = SCNOTDONE;
    }
    else {
	memcpy( &fname[0], fnesP->es_strP, len );
	fname[len] = NUL;

	fP = fopen( &fname[0], "r" );
	if ( fP != NULL ) {
	    /* Successful open; hand off to mvmda_mfl_run_fp */
	    status = mvmda_mfl_run_fp( rvalP, fP, msesP, &fname[0] );
	    fclose( fP );
	}
	else {
	    /* Can't open.  Work with the modes string in continuing. */
	    warnedF = FALSE;
	    mrelF = FALSE;
	    status = SCNOTDONE;
	    strP = msesP->es_strP;
	    len = estr_len( msesP );
	    while ( len-- > 0 ) {
		ch = *strP++;
		if ( !mrelF ) {
		    /* Not in relevant area; just wait for an 'O' */
		    mrelF = ( ch == 'O' );
		    continue;
		}

		/* In relevant area; be done if out of it. */
		if ( !islower(ch) )
		    break;

		switch ( ch ) {
		    case 'e':		/* Return error (done) */
			status = SCDONE;
			break;

		    case 'w':		/* Issue warning */
			if ( !warnedF ) {
			    warning( "mvmda_mfl_run_fname: Can't open file \"%s\"",
						&fname[0] );
			    warnedF = TRUE;
			}
			break;

		    default:
			break;
		}
	    }
	}
    }

    if ( rvalP != NULL )
	val_set_int( rvalP, 4, ( status == SCOK ) ? 1 : 0 );

    if ( status == SCNOTDONE )
	status = SCOK;
	
    return ( status );
}
/*

*//* mvmda_mfl_run_fp(rvalP, fP, msesP, fnameP)

	Run mfl/mvmda code that's in an opened file

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	fP		File pointer for the open file
	msesP		Modes string
	fnameP		The name of the file or input (for logging purposes)

Returns :

	<value>		status code

Notes :

	Basically a clone of mfl_run_fname(), but using mvmda_mfl_run_in
	  rather than mfl_run_in.  

	The file is NOT closed by this function.

	fnameP is a NUL-terminated string for exposition.  It may be passed
	  as NULL.

	The msesP string is used to affect operation and is passed
	  along to mvmda_mfl_run_in().

	See notes at calls mvmda_mfl_run_in().

*/

STSCODE
mvmda_mfl_run_fp ARGLIST( (rvalP, fP, msesP, fnameP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( FILE		*fP )		/* Input file */
   NFARG( ESTR		*msesP )	/* Modes string */
    FARG( char		*fnameP )	/* Name of the input */
{
	STSCODE		status;
	MML_IN		*inP;		/* Input stack */

    /* Create input stack */
    if ( ( inP = in_new() ) == NULL ) {
	warning( "mvmda_mfl_run_fp: in_new() failed." );
	return ( SCERR );
    }

    /* Push the string */
    if ( !in_stream_push_file_nc( inP, fP, fnameP ) ) {
	warning( "mvmda_mfl_run_fp: in_stream_push_fp() failed." );
	in_destroy( inP );
	return ( SCERR );
    }

    /* mfl-interp the input stack */
    status = mvmda_mfl_run_in( rvalP, inP, msesP );

    /* Remove the input stack */
    in_destroy( inP );

    /* Be done */
    return ( status );
}
/*

*//* mvmda_mfl_run_in(rvalP, inP, msesP)

	Run mfl/mvmda code that's in an input stack

Accepts :

	rvalP		If non-NULL, a VALUE to use for the result.
	inP		Ptr to input stack to get MFL code from
	msesP		Modes string

Returns :

	<value>		status code

Notes :

	Basically a clone of mfl_run_in(), but using mvmda's mf_command()
	  loop rather than a straight mfl parse/exec loop.

	If rvalP is non-NULL, this provides a VALUE to use for
	the result of tree execution.  Only the final VALUE is returned
	in a multi-statement string.

	Note Well: if caller supplies this, the final value is
	not released.  Caller may need to call val_finish() or
	val_release() .

	If Mfl_checkF is true (parse-only mode), only parses, doesn't
	exec.

	modesP is a string of modes characters.  An uppercase character
	  indicates what area the immediately following string of lowercase
	  characters applies to.  Here:

	  'M' specifies mfl modes:
	      'c' starts in c-like mode
	      'p' signifies parse-only (forces Mfl_checkF for the duration)
	      's' starts in sieve mode

*/

STSCODE
mvmda_mfl_run_in ARGLIST( (rvalP, inP, msesP) )
   NFARG( VALUE		*rvalP )	/* Where to return value */
   NFARG( MML_IN	*inP )		/* Ptr to input stack */
    FARG( ESTR		*msesP )	/* Modes string */
{
	STSCODE		status;
	BOOL		valF;		/* TRUE if any value present */
	BOOL		checkF;		/* Saved value of Mfl_checkF */
	BOOL		mrelF;		/* If in the relevant modes area */
	char		mflmode;	/* 0 for C-like, 1 for sieve, -1 unk */
	UBYTE		ch;
	int		len;
	UBYTE		*strP;		/* Generic string pointer */
	PTNODE		*treeP;		/* Parse tree */
	VALUE		*valP;		/* VALUE pointer to use */
	VALUE		val;		/* Scratch value if needed */
	TOKEN		*pbtokP;	/* Saved token context head */

    /* Save current pushedback token state */
    pbtokP = mfl_tokback_save();

    /* Save Mfl_checkF and interpret any modes characters */
    checkF = Mfl_checkF;
    mflmode = -1;			/* Dunno the mfl mode yet */
    strP = msesP->es_strP;
    len = estr_len( msesP );
    mrelF = FALSE;

    while ( len-- > 0) {
	ch = *strP++;
	if ( !mrelF ) {
	    /* Just looking for the 'M' area */
	    mrelF = ( ch == 'M' );
	    continue;
	}

	/* Be done if the area is ended (no more lowercase letters) */
	if ( !islower( ch ) )
	    break;
	switch ( ch ) {
	    case 'c':			/* C-like mode */
		mflmode = 0;
		break;

	    case 'p':			/* Force parse-only mode */
		Mfl_checkF = TRUE;
		break;

	    case 's':			/* Start in sieve mode */
		mflmode = 1;
		break;

	    default:			/* Ignore anything unrecognized */
		break;
	}
    }

    /* If we didn't get an explicit mfl mode, consult a control */
    if ( ( mflmode < 0 ) )
	mflmode = cntrl_int_get( estr_tmp_str( "mfl_mode_sieve" ), 0 );

    /* mvmda/mfl command/exec loop until the input is used up.
       or until some finished status is returned.
    */
    for ( ; ; ) {
	status = mf_command( inP, mflmode );
	if ( ( status != SCOK ) &&
	     ( status != SCNOTDONE ) )
	    break;
    }

    /* SCERR might mean EOF */
    if ( ( status == SCERR ) && ( inP->in_result == IOR_EOF ) )
	status = SCOK;

    /* Restore Mfl_checkF value */
    Mfl_checkF = checkF;

    /* Restore pushed-back token state */
    mfl_tokback_restore( pbtokP );

    /* Finish value if not giving it back to caller */
    if ( valF && ( rvalP == NULL ) )
	val_finish( valP );

    return ( status );
}

