/* mvmda.c -- mvmf Mail Delivery Agent

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

The mail delivery agent utility

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-symtab.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-ttio.h>
#include <mml/mml-dns.h>

#include "mvmf.h"

#include "mfl.h"
#include "mfl-misc.h"
#include "mfl-token.h"
#include "mfl-pp.h"
#include "mfl-func.h"

#include "mvmda.h"
#include "controls.h"
#include "mvmfmisc.h"
#include "mvmda-version.h"
#include "mvmda-command.h"
#include "mvmda-mfl-bif.h"

#include "msgpart.h"
#include "omm.h"
#include "dnsbl.h"
#include "msgstore.h"
#include "data.h"
#include "errwarn.h"


/* Local definitions */



/* External data referenced */

extern	BOOL	Mfl_checkF;		/* If check mode (parse only) */
extern	BOOL	Mfl_testF;		/* If test mode (non-delivery) */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */

/* External routines used */



/* Local data publicly available */

	BOOL	DebugF = FALSE;		/* Debugging flag */
	char	*MailboxP = NULL;	/* Default mailbox */
	OMM	*OmmP;			/* The open mail message handle */
	BOOL	PassthroughF = FALSE;	/* -p "passthrough" mode */

#ifdef	QMAIL
	BOOL	QmailF = FALSE;		/* If running directly from qmail */
#endif	/* QMAIL */

	    /* Operating mode */
	OPMODE	Opmode = OM_NONE;

	    /* Continuation prompt */
	char	Promptcont[MFL_TOKSIZE] = "mvmda_? ";

	    /* Main prompt */
	char	Promptmain[MFL_TOKSIZE] = "mvmda> ";


/* Local routines and forward references */

    /* Parse-only (check) mode control */
static	STSCODE	mvmda_check PROTO( (void) );

    /* Controls-printing (controls) mode control */
static	STSCODE	mvmda_controls PROTO( (void) );

    /* A function to print a control, via cntrl_walk */
static	void	mvmda_control_print();

    /* Produce ending statistics */
static	void	mvmda_endstats PROTO( (void) );

    /* Exit */
static	void	mvmda_exit PROTO( (int exitcode) );

    /* Initialize mvmda stuff */
static	BOOL	mvmda_init PROTO( (void) );
static	BOOL	mvmda_init_rc PROTO( (void) );
static	BOOL	mvmda_init_startup PROTO( (void) );
static	BOOL	mvmda_init_user PROTO( (void) );

    /* Interactive mode control */
static	STSCODE	mvmda_interactive PROTO( (void) );

    /* Publish the operating mode to MFL */
static	STSCODE	mvmda_mode_set PROTO( (char *modenameP) );

    /* Normal mode control */
static	STSCODE	mvmda_normal PROTO( (void) );

    /* signal/exit catcher for interactive mode */
static	void	sig_exit PROTO( (int signum) );

static	void	usage();


/* Private data */

static	int	ArgC;			/* Copy of argc from main */
static	char	**ArgV;			/* Copy of argv from main */
static	LL_HEAD	Opt_postcmds;		/* Post-sysrc mfl cmds to execute */
static	LL_HEAD	Opt_precmds;		/* Pre-sysrc mfl cmds to execute */
static	struct passwd	*MypwdP;	/* Passwd entry for user */
static	BOOL	SieveF = FALSE;		/* -S "sieve" mode */
static	struct timeval Starttime;	/* Wall clock when we start */
static	char	*UsernameP = NULL;	/* Specified username. */
static	int	VerC;			/* -v or -V counter */

    /* Table of integer controls to define */
static	CNTRLDEF_INT Ctdef_int[] = {
/*
 ------- name -------	val	admin	admin	user	user
				 low	 high	 low	high
*/
{ "admin",		  0,	    0,	    1,	   0,	   0	},
{ "pipe_allow",		  0,	    0,	    1,	   0,	   0	},

{ NULL,			  0,	    0,	    0,	   0,	   0	}
		};


    /* Table of string controls to define */
static	CNTRLDEF_STR Ctdef_str[] = {
/*
 ------- name -------		--------- val --------		adminF
*/
{ "data_directory",		NULL,				 FALSE	},
{ "email_address_primary",	NULL,				 FALSE	},
{ "log_private_name",		NULL,				 FALSE	},
{ "log_program_name",		"mvmda",			 FALSE	},

{ NULL,			  NULL,				     0	}
		};


#ifdef	MVMDA_BUILTIN_SYSRC		/* Default code if system-wide
					   rc file can't be opened.  Only if
					   configured.
					*/
static	UBYTE	Sys_mfl[] =
"\n\
// Start by setting up some values depending on what mode we are in\n\
//  The '= 0' initializers aren't necessary, but they look good.\n\
char $_rc_do_settings = 0;\n\
char $_rc_do_scripts = 0;\n\
char $_rc_do_message = 0;\n\
string $_rc_script_mode;\n\
\n\
// Most likely first\n\
if ( $OPMODE == \"NORMAL\" ) {\n\
    $_rc_do_settings = $_rc_do_scripts = $_rc_do_message = 1;\n\
    $_rc_script_mode = \"Ow\";\n\
}\n\
else if ( $OPMODE == \"CHECK\" ) {\n\
    $_rc_do_settings = $_rc_do_scripts = 1;\n\
    $_rc_script_mode = \"MpOw\";\n\
}\n\
else if ( $OPMODE == \"CONTROLS\" ) {\n\
    $_rc_do_settings = 1;\n\
}\n\
else if ( $OPMODE == \"INTERACTIVE\" ) {\n\
    $_rc_do_settings = 1;\n\
}\n\
else {\n\
    pv$ \"sysrc: bad value for $OPMODE: \";\n\
    pv$ $OPMODE;\n\
    return;\n\
}\n\
\n\
// Other scratch variables for the rc file\n\
string $_rc_s, *$_rc_sP;\n\
int $_rc_i;\n\
\n\
// If we are supposed to do controls settings, do them now.\n\
if ( $_rc_do_settings ) {\n\
    $_rc_s = $env_variable( \"MVMF_LIBDIR\" );\n\
    $mfl_incdir_add( \"system\", $_rc_s + \"/include\" );\n\
    $mfl_hookdir_add( $_rc_s + \"/mvmda/hooks\" );\n\
\n\
    // Next might include running a default user-rc script.\n\
    $_rc_sP = $env_homedir();\n\
    if ( $_rc_sP != 0 )\n\
	$mfl_protect_admin(\n\
    	    $mfl_run_fname( *$_rc_sP + \"/.mvmda-rc.mfl\", \"Mc\" )\n\
	);\n\
}\n\
\n\
// If we need to open the incoming message, do it.\n\
if ( $_rc_do_message ) {\n\
    if ( !$msg_open_fname( \"\", 0 ) )\n\
	return;\n\
}\n\
\n\
// Run any scripts now in the previously set mode.\n\
// One might look for default scriptfile name(s) here, or\n\
// use a system-wide user script if none are given, etc.\n\
if ( $_rc_do_scripts ) {\n\
    for ( $_rc_i = $ArgN; $_rc_i < $ArgC; ++$_rc_i )\n\
	$mfl_protect_admin(\n\
	    $mfl_run_fname( $ArgV[ $_rc_i ], $_rc_script_mode )\n\
	);\n\
}\n\
" ;

static	BOOL	Sys_mfl_forceF = FALSE;

#endif	/* MVMDA_BUILTIN_SYSRC */



int
main( argc, argv )
	int		argc;
	char		**argv;
{
	char		usageOK;
	int		exitcode;
	int		optch;
	STSCODE		status;

    if ( !mvmda_init_startup() )	/* first-thing init */
	mvmda_exit( MEX_SOFT );

    usageOK = TRUE;

#ifdef	QMAIL
    while ( usageOK &&
	   ( optch = getopt( argc, argv, "bcd:e:E:il:m:npqr:s:SvV" ) ) != EOF ) {
#else
    while ( usageOK &&
	   ( optch = getopt( argc, argv, "bcd:e:E:il:m:npr:s:SvV" ) ) != EOF ) {
#endif	/* QMAIL */

	switch ( optch ) {
	    case	'b':
#ifdef	MVMDA_BUILTIN_SYSRC_FORCEABLE
		Sys_mfl_forceF = TRUE;
#else
		fprintf( stderr, "illegal option: -b" );
		usageOK = FALSE;
#endif	/* MVMDA_BUILTIN_SYSRC_FORCEABLE */
		break;


	    case	'c':		/* Parse-only mode */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-c\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_CHECK;
		break;


	    case	'd':		/* Destination username */
		if ( UsernameP != NULL ) {
		    fprintf( stderr, "Duplicate -d given.\n" );
		    usageOK = FALSE;
		}
		else
		    UsernameP = optarg;
		break;


	    case	'e':		/* pre-sysrc mfl command */
		/* Must remember the command for later execution */
		if ( ll_addvt( &Opt_precmds, optarg ) == NULL ) {
		    fprintf( stderr, "Can't remember 'e' argument: \"%s\"\n",
					optarg );
		    usageOK = FALSE;
		}
		break;


	    case	'E':		/* post-sysrc mfl command */
		/* Remember the command */
		if ( ll_addvt( &Opt_postcmds, optarg ) == NULL ) {
		    fprintf( stderr, "Can't remember 'E' argument: \"%s\"\n",
					optarg );
		    usageOK = FALSE;
		}
		break;


	    case	'i':		/* Interactive mode */
		if ( Opmode != OM_NONE ) {
		    fprintf( stderr, "Conflicting switch \"-i\".\n" );
		    usageOK = FALSE;
		}
		else
		    Opmode = OM_INTERACTIVE;

		break;

	    case	'l':		/* Folder layout type */
		/* Obsolete option.  Will be phased out. */
		break;

	    case	'm':		/* Default mailbox */
		MailboxP = optarg;
		break;

	    case	'n':		/* Test mode */
		Mfl_testF = TRUE;
		break;

	    case	'p':		/* Pass-through mode */
		PassthroughF = TRUE;
		break;

#ifdef	QMAIL
	    case	'q':		/* Running under .qmail */
		QmailF = TRUE;
		break;
#endif	/* QMAIL */

	    case	'r':		/* Folder/FS root */
		/* Obsolete options.  Will be phased out. */
		break;

	    case	's':		/* Default storage type */
		/* Obsolete option.  Will be phased out. */
		break;

	    case	'S':		/* Start in SIEVE mode */
		SieveF = TRUE;
		break;

	    case	'v':		/* Print version info */
		if ( ( Opmode != OM_NONE ) &&
		     ( Opmode != OM_VERSION ) ) {
		    fprintf( stderr, "Conflicting switch \"-v\".\n" );
		    usageOK = FALSE;
		}
		else {
		    Opmode = OM_VERSION;
		    ++VerC;
		}
		break;

	    case	'V':		/* Print version and controls */
		if ( ( Opmode != OM_NONE ) &&
		     ( Opmode != OM_CONTROLS ) ) {
		    fprintf( stderr, "Conflicting switch \"-V\".\n" );
		    usageOK = FALSE;
		}
		else {
		    Opmode = OM_CONTROLS;
		    ++VerC;
		}
		break;

	    default:
		usageOK = FALSE;
		break;
	}
    }

    /* If no operating mode selected, set to normal */
    if ( usageOK && ( Opmode == OM_NONE ) )
	Opmode = OM_NORMAL;

    if ( !usageOK ) {
	usage();
	mvmda_exit( MEX_SOFT );
    }

    if ( !mvmda_init() )
	mvmda_exit( MEX_SOFT );

    /* Make arguments available to mvmda and to mfl */
    ArgC = argc;
    ArgV = argv;
    mfl_program_args( argc, argv, optind );

    /* Proceed now depending on the mode */
    switch ( Opmode ) {
	case	OM_CHECK:
	    status = mvmda_check();
	    break;

	case	OM_INTERACTIVE:
	    status = mvmda_interactive();
	    break;

	case	OM_NORMAL:
	    status = mvmda_normal();
	    break;

	case	OM_VERSION:
	    /* Don't even run the sysrc ... just print the version */
	    printf( "mvmda version %s\n", MVMDA_VERSION );
	    status = SCOK;
	    break;

	case	OM_CONTROLS:
	    status = mvmda_controls();
	    break;

	default:
	    warning( "Error: unknown Opmode %d.  This is a bug.", Opmode );
	    status = SCERR;
	    break;
    }


    /* Exit code must reflect what happened to the message.  Because
       a sieve script can attempt multiple deliveries and only fail
       on one, it's difficult to know how to dispose of the message
       in the presence of failures.  But we approach it this way:

     - At the time of any delivery error, the MFL hook
       "$hook_delivery_error" is called.  This hook can put in a
       request for a message refusal by calling $msg_reject().

     - Requests for refusal are noted in the OMM flags om_wantbounceF
       and om_wantrejectF.  The latter is for SMTP-time rejection.

     - If we're running at SMTP time (e.g. as an LMTP agent), then
       if either flag is set, we indicate an SMTP rejection.
       NOTE WELL: SMTP-time and LMTP are still not here.. this is
       for later.

     - If we are not running at SMTP time, we only generate a bounce
       if the om_wantbounceF flag is set.  We also take additional
       bounce-suppression steps, up to and including not honoring
       them at all (see the code).

     The whole issue of bounces at non-SMTP-time is problematic yet
     can't be ignored entirely, no matter what some folks proclaim.
     Rejection at SMTP-time is thorny enough.

     Regardless of bounce behaviour:

     - If there were any successful deliveries, dequeue the mail,

     - If there were any hard errors, dequeue the mail.  Don't want to
       keep trying the same delivery.

     - If there were no deliveries or errors at all, that's OK in
       pass-through mode (go ahead and pass through).  Otherwise
       it's weird, try later.

     - if none of the above, leave in the queue for later retry.

       The actual program logic may be arranged somewhat differently
       than that expressed above.

    */

    if ( OmmP == NULL ) {
	/* No open mail message, exit with the final command status */
	exitcode = ( status == SCOK ? MEX_OK : MEX_SOFT );
    }
    else if ( Opmode == OM_NORMAL ) {

	/* Only do default delivery stuff if "normal" mode.. presumably
	   in interactive mode the user has already done what they
	   want to do.
	*/

	/* If we delivered at all, suppress any non-SMTP bounces. */
	if ( OmmP->om_deliveryC > 0 )
	    OmmP->om_wantbounceF = FALSE;

	/* If we're running at SMTP time, handle any bounces now.
	   Note: orphaned code, never invoked since we don't have a
	   working LMTP engine yet.  It's coming!  (tho I imagine
	   this will all be changed when that happens.)
	*/
TODO
	if ( FALSE ) {
	    if ( OmmP->om_wantbounceF || OmmP->om_wantrejectF ) {
		;			/* Handle it */
	    }
	}
	else {
	    /* Not SMTP time.  Generate a bounce if it's been requested.
	       Note that msg_bounce() does further anti-bounce checks.
	    */
	    if ( OmmP->om_wantbounceF ) {
		if ( Mfl_testF )
		    printf( "test: bounce.\n" );
		else
		    msg_bounce( OmmP, "Error(s) occured during delivery." );
	    }
	}


	/* Decide how to dispose of the message. */
	if ( Mfl_testF )
	    exitcode = MEX_OK;
	else if ( OmmP->om_bouncedF )
	    exitcode = MEX_HARD;	/* Bounce attempted, hard error. */
	else if ( OmmP->om_deliveryC == 0 ) {
	    if ( OmmP->om_hardfailC != 0 )
		/* No delivery, hard failure, fail hard (no more tries) */
		exitcode = MEX_HARD;
	    else if ( OmmP->om_softfailC != 0 )
		/* No delivery, soft error, fail softly (for retry) */
		exitcode = MEX_SOFT;
	    else {
		/* No deliveries attempted at all.  If pass-through mode,
		   that is ok.
		*/
		if ( PassthroughF )
		    exitcode = MEX_OKCONT;
		else {
		    /* Strange situation.  I guess we'll fail softly. */
		    exitcode = MEX_SOFT;
		}
	    }
	}
	else {
	    /* There were deliveries, so we'll call it OK and dequeue
	       the message.
	    */
	    exitcode = MEX_OK;
	}

	/* Give ending stats */
	mvmda_endstats();
    }

    mvmda_exit( exitcode );

    exit(0);			/* Shouldn't get here */
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/

static void
usage()
{
    printf( "usage: mvmda [flags] [args]\n" );
    printf( "    Flags are:\n" );
#ifdef	MVMDA_BUILTIN_SYSRC_FORCEABLE
    printf( "       -b               Force the use of built-in rc script.\n" );
#endif	/* MVMDA_BUILTIN_SYSRC_FORCEABLE */
    printf( "       -c               \"Check\" mode: syntax check on script.\n" );
    printf( "       -e string        Give 'string' to MFL via $Opts_e.\n" );
    printf( "       -E string        Give 'string' to MFL via $Opts_E.\n" );
    printf( "       -i               Interactive mode\n" );
    printf( "       -m  path         Default mailbox name.  [INBOX]\n" );
    printf( "       -n               Run in test mode (report but don't take actions)\n" );
    printf( "       -p               Pass-through mode\n" );
#ifdef	QMAIL
    printf( "       -q               If being called by qmail-local\n" );
#endif	/* QMAIL */
    printf( "       -S               Set 'mfl_mode_sieve' control to 1.\n" );
    printf( "       -v               Give version information.\n" );
    printf( "       -V               Show running configuration.\n" );
    printf( "\n" );
    printf( "args are anything that might be used by the system-wide mfl script,\n" );
    printf( "  and might, for example, be script name(s).\n" );
    printf( "  Your system administrator will have decided this.\n" );
    printf( "\n" );
    printf( "The following options are obsolete and do nothing.  They will\n" );
    printf( "  soon be phased out entirely:\n" );
    printf( "       -l  type         Folder layout type (FS, Maildir++)\n" );
    printf( "       -r  path         Path to default folder root\n" );
    printf( "       -s  type         Default storage type (mbox, Maildir)\n" );
}
/*

*//* mvmda_check()

	Run in "check" mode to check script syntax

Accepts :


Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmda_check NOARGLIST
{
	STSCODE		status;
	int		ch;
	int		err;		/* errno copy */
	int		argN;
	FILE		*sfP;		/* Script file */

    /* Publish the operating mode */
    if ( ( status = mvmda_mode_set( "CHECK" ) ) != SCOK )
	return ( status );

    /* The rc file processing will do the rest */
    if ( !mvmda_init_rc() )
	return ( SCERR );

    return ( SCOK );
}
/*

*//* mvmda_controls()

	Run in "controls" mode to emit the controls settings

Accepts :


Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmda_controls NOARGLIST
{
	STSCODE		status;
	int		ch;
	int		err;		/* errno copy */
	int		argN;
	FILE		*sfP;		/* Script file */

    /* Publish the operating mode */
    if ( ( status = mvmda_mode_set( "CONTROLS" ) ) != SCOK )
	return ( status );

    /* Do sysrc */
    if ( !mvmda_init_rc() )
	return ( SCERR );

    /* Print the version and controls now */
    printf( "// mvmda version %s\n\n", MVMDA_VERSION );
    printf( "// system-wide script file: %s\n\n", MVMDA_SYSRC );

#ifdef	MVMDA_BUILTIN_SYSRC
    printf( "// The built-in rc script has been enabled.\n\n" );
#else
    printf( "// There is NO built-in rc script.\n\n" );
#endif	/* MVMDA_BUILTIN_SYSRC */

    cntrl_walk( mvmda_control_print );

#ifdef	MVMDA_BUILTIN_SYSRC
    if ( VerC > 1 ) {
	printf( "\n// Built-in rc script:\n\n" );
	printf( "%s\n", &Sys_mfl[0] );
    }
#endif	/* MVMDA_BUILTIN_SYSRC */


    return ( SCOK );
}
/*

*//* mvmda_control_print( ctlP )

	Print a control for control-printing mode

Accepts :

	ctlP		Ptr to control struct

Returns :

	<nothing>

Notes :

	Called via cntrl_walk()

*/

static void
mvmda_control_print ARGLIST( (ctlP) )
    FARG( CNTRL		*ctlP )		/* The control */
{
	CNTRLINT	*ciP;		/* For integer control */
	CNTRLSTR	*csP;		/* For string control */

    switch ( ctlP->ctl_type ) {
	case CTL_INT:
	    ciP = (CNTRLINT *)ctlP->ctl_infoP;
	    printf( "$control_int_set( \"%s\", %d );\n\n",
		   ctlP->ctl_nameP,
		   ciP->cv_int );
	    break;

	case CTL_STR:
	    csP = (CNTRLSTR *)ctlP->ctl_infoP;
	    if ( csP->cs_strP == NULL )
		printf( "$control_string_set( \"%s\", 0 );\n\n",
			  ctlP->ctl_nameP );
	    else
TODO		/* Really need to quote special chars here */
		printf( "$control_string_set( \"%s\", \"%s\" );\n\n",
		          ctlP->ctl_nameP,
		          csP->cs_strP );
	    break;

	default:
	    fprintf( stderr, "mvmda_control_print: unknown control type %d",
			ctlP->ctl_type );
	    break;
    }

    /* done */
}
/*

*//* mvmda_endstats()

	Produce ending statistics for a normal run

Accepts :

	<none>

Returns :

	<nothing>

Notes :

	Logs statistics when done and relevant.

*/

void
mvmda_endstats NOARGLIST
{
	int	lbX;			/* logbuf index */
	struct timeval	endtime;	/* Wall clock when we end */
	struct timezone tz;		/* Dummy */
	struct rusage	ru;		/* Resource usage stuff */
	char	logbuf[ 200 ];		/* Big buffer for log info */

    lbX = 0;

    /* Log elapsed time in milliseconds */
    gettimeofday( &endtime, &tz );
    lbX += sprintf( &logbuf[lbX], "elapse-ms %ld",
		   ((endtime.tv_sec - Starttime.tv_sec) *1000) +
		   ((endtime.tv_usec - Starttime.tv_usec)/1000) );
    

    if ( getrusage( RUSAGE_SELF, &ru ) == 0 ) {
	lbX += sprintf( &logbuf[lbX], ", user-ms %ld",
		       ( ru.ru_utime.tv_sec * 1000 ) +
		       ( ru.ru_utime.tv_usec / 1000 ) );

	lbX += sprintf( &logbuf[lbX], ", sys-ms %ld",
		       ( ru.ru_stime.tv_sec * 1000 ) +
		       ( ru.ru_stime.tv_usec / 1000 ) );

	lbX += sprintf( &logbuf[lbX], ", rss-kb %ld", ru.ru_maxrss );
    }

    logbuf[ lbX ] = NUL;		/* Yeah, not necessary */

    if ( Mfl_testF )
	printf( "Stats: %s\n", logbuf );
    else
	log_both( "Stats: %s", logbuf );
}
/*

*//* mvmda_exit( exitcode )

	Exits with appropriate code.

Accepts :

	exitcode	MEX_xxx exit code.

Returns :

	<nothing>


Notes :

	Uses the MFL hook $hook_exitcode(), if it is defined,
	to translate the exit code from an internal value to whatever
	the hook wants it to be.

	If there is no such hook, the passed exitcode is used
	as-is.

*/

static void
mvmda_exit ARGLIST( (exitcode) )
    FARG( int		exitcode )	/* Exit code */
{
	BOOL		hookedF;
	char		mflcmd[200];	/* MFL code string */
	VALUE		hookval;

    hookedF = FALSE;			/* No hook processed */

    /* Use exitcode hook if we can find one */
    if ( mfl_func_hook_find( "$hook_exitcode" ) ) {
	if ( snprintf( &mflcmd[0], sizeof(mflcmd),
		      "%s( %d );",
		      "$hook_exitcode",
		       exitcode ) >= sizeof(mflcmd) ) {
	    warning( "mvmda_exit: mflcmd buffer too small!" );
	}
	else {
	    mfl_run_str( &hookval, estr_tmp_str( (UBYTE *)&mflcmd[0] ),
			 estr_tmp_str( (UBYTE *)"Mc" ) );
	    exitcode = val_get_int( &hookval );
	    val_finish( &hookval );
	    hookedF = TRUE;
	}
    }

    exit( exitcode );
}
/*

*//* mvmda_init()

	Do general initialization

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	This handles the initialization of things that must be done
	once the program is ready to go: after options have been
	collected and sanitized.

*/

static BOOL
mvmda_init NOARGLIST
{
	int		i;
	char		path[ PATH_MAX +1 ];

    /* Initialize stuff about the user (perhaps becoming someone) */
    if ( !mvmda_init_user() )
	return ( FALSE );

    /* Define controls used by this application */
    if ( !cntrl_table_def_int( &Ctdef_int[0] ) ) {
	fprintf( stderr, "Problem defining integer controls." );
	return ( FALSE );
    }
    if ( !cntrl_table_def_str( &Ctdef_str[0] ) ) {
	fprintf( stderr, "Problem defining string controls." );
	return ( FALSE );
    }

    /* Make a guess here: if user shell looks like an interactive shell,
       we'll enable some system access stuff.  Looking in /etc/shells
       isn't good enough for various reasons.  At any rate this can
       be overridden in system-wide rc.
    */
    if ( ( strcmp( MypwdP->pw_shell, "/bin/bash" ) == 0 ) ||
         ( strcmp( MypwdP->pw_shell, "/bin/csh" ) == 0 ) ||
         ( strcmp( MypwdP->pw_shell, "/bin/ksh" ) == 0 ) ||
         ( strcmp( MypwdP->pw_shell, "/bin/sh" ) == 0 ) ||
         ( strcmp( MypwdP->pw_shell, "/bin/tcsh" ) == 0 ) ||
         ( strcmp( MypwdP->pw_shell, "/bin/zsh" ) == 0 ) ) {

	/* May pipe */
	cntrl_int_setlim( estr_tmp_str( "pipe_allow" ), CA_ADMIN, 0, 1 );
	cntrl_int_set( estr_tmp_str( "pipe_allow" ), CA_ADMIN, 1 );
    }

    /* Initialize lib stuff */
    if ( !data_init() )
	return ( FALSE );

    /* Initialize msgstore stuff */
    if ( !msgstore_init() )
	return ( FALSE );

    /* Initialize DNSBL */
    dnsbl_init();

    /* Initialize language stuff */
    if ( mfl_init() != SCOK )
	return ( FALSE );
    if ( !bif_mvmda_init() )
	return ( FALSE );

    /* Set the mfl mode control according to -S (SieveF) */
    cntrl_int_set( estr_tmp_str( "mfl_mode_sieve" ),
		   CA_ADMIN,
		   SieveF ? 1 : 0 );

    /* Make configuration information available via environment variables. */
    if ( setenv( "MVMF_LIBDIR", MVMF_LIBDIR, 1 ) != 0 ) {
	warning( "mvmda_init: can't set environment variable MVMF_LIBDIR" );
	return ( FALSE );
    }

    if ( setenv( "MVMF_SYSDIR", MVMF_SYSDIR, 1 ) != 0 ) {
	warning( "mvmda_init: can't set environment variable MVMF_SYSDIR" );
	return ( FALSE );
    }

    /* Give the '-e' and '-E' lists to MFL */
    if ( !mfl_array_from_list_strings( "$Opts_e", 0, &Opt_precmds ) ) {
	fprintf( stderr, "Can't initialize $Opts_e variables (for -e)\n" );
	return ( FALSE );
    }
    if ( !mfl_array_from_list_strings( "$Opts_E", 0, &Opt_postcmds ) ) {
	fprintf( stderr, "Can't initialize $Opts_E variables (for -E)\n" );
	return ( FALSE );
    }

    /* Set up default mailbox if needed */
    if ( MailboxP == NULL ) {
	MailboxP = "INBOX";
    }

    /* Setup the "data_directory" control if it hasn't been done */
    if ( cntrl_str_get( estr_tmp_str( "data_directory" ) ) == NULL ) {
	i = snprintf( &path[0], PATH_MAX, "%s/.%s",
				MypwdP->pw_dir,
				cntrl_str_get(
				   estr_tmp_str( "log_program_name" ) ) );
	if ( i < PATH_MAX )
	    cntrl_str_set( estr_tmp_str( "data_directory" ),
			    CA_ADMIN,
			    estr_tmp_str( &path[0] ) );
    }

    /* Done without returning, must have succeeded */
    return ( TRUE );
}
/*

*//* mvmda_init_rc()

	Initialize from rc file(s)

Accepts :

	<nothing>

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Subservient to mvmda_init, processes the rc files
	(system-wide and user-level startup files).

	Also responsible for executing any mfl code snippets
	supplied on the command line.

*/

static BOOL
mvmda_init_rc NOARGLIST
{
	STSCODE		status;
	FILE		*fP;		/* File handle */

    /* The system-level rc file executes in admin mode. */
    cntrl_int_set( estr_tmp_str( "admin" ), CA_ADMIN, 1 );

    /* If we haven't forced the internal one, check for system-wide RC file */
#ifdef	MVMDA_BUILTIN_SYSRC
    fP = NULL;
    if ( ! Sys_mfl_forceF )
#endif	/* MVMDA_BUILTIN_SYSRC */
	fP = fopen( MVMDA_SYSRC, "r" );

    if ( fP != NULL ) {
	status = mfl_run_fp( NULL, fP, estr_tmp_str( (UBYTE *)"Mc" ),
			      MVMDA_SYSRC );
	fclose( fP );
    }
    else {

#ifdef	MVMDA_BUILTIN_SYSRC
# ifdef	MVMDA_BUILTIN_SYSRC_FORCEABLE
	/* The built-in sysrc is available, but must be forced with -b */
	if ( !Sys_mfl_forceF ) {
	    warning( "mvmda_init_rc: system RC file \"%s\" is not available.\n",
		   MVMDA_SYSRC );
	    warning( "you must use -b to enable the internal script." );
	    status = SCERR;
	}
	else
# endif	/* MVMDA_BUILTIN_SYSRC_FORCEABLE */
	status = mfl_run_str( NULL,
			      estr_tmp_str( &Sys_mfl[0] ),
			      estr_tmp_str( (UBYTE *)"Mc" ) );

#else	/* MVMDA_BUILTIN_SYSRC */
	warning( "mvmda_init_rc: system RC file \"%s\" is not available.\n",
		   MVMDA_SYSRC );
	status = SCERR;

#endif	/* MVMDA_BUILTIN_SYSRC */

    }

    /* Back out of admin mode now. */
    cntrl_int_set( estr_tmp_str( "admin" ), CA_ADMIN, 0 );

    return ( ( status == SCOK ) );
}
/*

*//* mvmda_init_startup()

	Do initialization on startup

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	This is intended for initialization of data immediately upon
	startup; it prepares for getting options and whatever else might
	be needed right off the bat.

*/

static BOOL
mvmda_init_startup NOARGLIST
{
	struct timezone	tz;		/* Dummy timezone */

    /* First thing: get resource info at program start. */
    gettimeofday( &Starttime, &tz );

    /* Other signals that need to be different than default */
    signal( SIGPIPE, SIG_IGN );		/* Default terminates process; we
					   can check write failure just
					   fine, thank you.
					*/

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();


    /* Init the lists for mfl commands given on the command line */
    ll_init( &Opt_postcmds );
    ll_init( &Opt_precmds );

    return ( TRUE );
}
/*

*//* mvmda_init_user()

	Initialize stuff about the user environment

Accepts :


Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	There's a bunch of fussy user-related initialization that
	we do.  It's forked into its own function for whatever reason.

*/

static BOOL
mvmda_init_user NOARGLIST
{
	int		i;
	int		grpC;
	uid_t		myuid;
	gid_t		egroup;
	char		*userP;
	struct passwd	*pwdP;
	UBYTE		grpname[50];
	gid_t		groups[ NGROUPS +1 ];

    /* If we are root, and there was a delivery username specified,
       we will attempt to become that user.  It's assumed we're root for
       a reason...
    */
    if ( ( geteuid() == 0 ) && ( UsernameP != NULL ) ) {
	pwdP = getpwnam( UsernameP );
	if ( pwdP == NULL ) {
	    fprintf( stderr,
		    "mvmda_init_user: invalid username specified: \"%s\"\n",
			UsernameP );
	    return ( FALSE );
	}

	/* We'll want to set up this process with the group list
	   corresponding to all of the groups that the user is
	   a member of, plus any additional that we are supposed to
	   set.
	*/
	grpC = NGROUPS;
	if ( getgrouplist( pwdP->pw_name, pwdP->pw_gid,
			   &groups[0], &grpC ) < 0 ) {
	    perror( "getgrouplist" );
	    return ( FALSE );
	}

	/* Attempt to put any setgid effective gid (as indicated by a
	   non-zero egid, so don't setgid 0 I guess) into the groups list.
        */
	egroup = getegid();
	if ( egroup != 0 ) {
	    for ( i = 0; i < grpC; ++i )
		if ( groups[i] == egroup )
		    break;
	    if ( ( i == grpC ) && ( grpC < NGROUPS ) )
		groups[ grpC++ ] = egroup;
	}
	if ( setgroups( grpC, &groups[0] ) < 0 ) {
	    perror( "setgroups" );
	    return( FALSE );
	}
	/* Set the group to the user's base group */
	if ( setgid( pwdP->pw_gid ) < 0 ) {
	    perror( "setgid" );
	    return ( FALSE );
	}

	/* Now become the desired user. */
	if ( setuid( pwdP->pw_uid ) < 0 ) {
	    perror( "setuid" );
	    return ( FALSE );
	}

	/* We'll want to cd to that user's homedir.  If it fails, we'll
	   just go to /tmp.
	*/
	if ( chdir( pwdP->pw_dir ) < 0 ) {
	    warning( "Can't cd to \"%s\" when becoming \"%s\" -- %d.\n",
		      pwdP->pw_dir,
		      UsernameP,
		      errno );
	    chdir( "/tmp" );
	}
    }

    /* Get info about the user.  Note, some of this is repeated in
       general data initialization later (data_init()), but using
       global data.
    */
    myuid = getuid();			/* My uid */

    /* If there's a USER environment variable and it matches my uid,
       we'll use that info.
    */
    userP = getenv( "USER" );
    if ( userP != NULL ) {
	MypwdP = getpwnam( userP );
	if ( ( MypwdP == NULL ) || ( MypwdP->pw_uid != myuid ) ) {
	    /* Not a match, use first user matching my uid */
	    MypwdP = getpwuid( myuid );
	}
    }
    else
	MypwdP = getpwuid( myuid );
    if ( MypwdP == NULL ) {
	fprintf( stderr, "mvmda_init_user: can't find user info!\n" );
	return ( FALSE );
    }

    /* Force the USER environment variable to match who we think we
       are at this point.
    */
    if ( ( userP == NULL ) ||
         ( strcmp( userP, MypwdP->pw_name ) != 0 ) ) {
	if ( setenv( "USER", MypwdP->pw_name, 1 ) != 0 ) {
	    fprintf( stderr,
		   "mvmda_init_user: can't set USER environment variable." );
	    return ( FALSE );
	}
	userP = getenv( "USER" );
	if ( ( userP == NULL ) ||	/* Being anal here */
	     ( strcmp( userP, MypwdP->pw_name ) != 0 ) ) {
	    fprintf( stderr,
		"mvmda_init_user: problem with USER environment variable." );
	    return ( FALSE );
	}
    }

    /* If a delivery username was specified on the command line, it
       must match who we are running as.
    */
    if ( ( UsernameP != NULL ) &&
         ( strcmp( UsernameP, userP ) != 0 ) ) {
	fprintf( stderr, "mvmda_init: wanted username \"%s\" is not us.\n",
		 UsernameP );
	return ( FALSE );
    }

    /* We made it. */
    return ( TRUE );
}
/*

*//* mvmda_interactive()

	Run in interactive mode

Accepts :



Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmda_interactive NOARGLIST
{
	STSCODE		status;
	TTCMDLN		*cmlP;		/* Command line object */
	MML_IN		*inP;		/* Input stack object */

    /* Publish the mode name */
    if ( ( status = mvmda_mode_set( "INTERACTIVE" ) ) != SCOK )
	return ( status );

    /* It's OK to start logging now */
    log_begin();

    /* Process any startup (rc) files */
    if ( !mvmda_init_rc() )
	return ( FALSE );

    /* Create an input stack */
    inP = in_new();
    if ( inP == NULL ) {
	fprintf( stderr, "in_new() failed.\n" );
	return ( SCERR );
    }

    /* Set up command line stream */
    cmlP = tt_cmdline_new( &Promptmain[0], &Promptcont[0] );
    if ( cmlP == NULL ) {
	fprintf( stderr, "mvmda_interactive: can't create cmdline stream\n" );
	return ( SCERR );
    }

    if ( !in_stream_push_ttcmd( inP, cmlP, "command line" ) ) {
	fprintf( stderr, "mvmda_interactive: can't push cmdline stream\n" );
	return ( SCERR );
    }

    tt_init();			/* Init terminal */

    /* Catch signal(s) so we can clean up before exit */
    signal( SIGINT, sig_exit );

    printf( "\n" );
    printf( ":xxx to execute command xxx (:help for help);\n" );
    printf( "%sxxxx for preprocessor directive;\n", MFL_PP_INTRODUCER );
    printf( "anything else is a language element.\n" );
    printf( "\n" );

    /* Loop doing command processing */
    for ( ; ; ) {
	status = mf_command( inP,
		     cntrl_int_get( estr_tmp_str( "mfl_mode_sieve" ), 0 ) );
	if ( ( status != SCOK ) &&
	     ( status != SCNOTDONE ) &&
	     ( status != SCDONE ) )
	    break;
    }

    tt_fini();			/* Finit term */

    /* Get rid of the input stack we were using */
    in_destroy( inP );

    return ( SCOK );
}
/*

*//* mvmda_mode_set( modenameP )

	Publish the operating mode to MFL

Accepts :

	modenameP	Name of the operational mode.

Returns :

	<value>		SCxx status code

Notes :

	Mode name should be the mode name suffix, e.g. "NORMAL" and
	  must not contain any weird characters that need quoting.

	The mode name is published to MFL as the value of a string
	  variable named $OPMODE, so that the system-wide RC can act
	  on it.

	Uses Gbs1P BSTR as scratch.

*/

static STSCODE
mvmda_mode_set ARGLIST( (modenameP) )
    FARG( char		*modenameP )	/* Mode name mnemonic to publish */
{
    /* Publish the mode name in MFL as $OPMODE */
    Gbs1P->bs_bC = 0;
    if ( ! bstr_printf( Gbs1P, "string $OPMODE = \"%s\";\n", modenameP ) ) {
	warning( "mvmda_via_sysrc: can't publish mode name (%s)!", modenameP );
	return ( SCERR );
    }
    bstr_nulterm( Gbs1P );

    return ( mfl_run_str( NULL, estr_tmp_bstr( Gbs1P ),
			   estr_tmp_str( (UBYTE *)"Mc" ) ) );
}
/*

*//* mvmda_normal()

	Run in normal mode

Accepts :


Returns :

	<value>		SCxx status code

Notes :


*/

static STSCODE
mvmda_normal NOARGLIST
{
	STSCODE		status;
	int		err;
	int		argN;
	FILE		*sfP;		/* Script file */

    /* Publish the operating mode */
    if ( ( status = mvmda_mode_set( "NORMAL" ) ) != SCOK )
	return ( status );

    /* It's OK to start logging now */
    log_begin();

    /* Process any startup (rc) files */
    if ( !mvmda_init_rc() )
	return ( FALSE );

    /* If there's an implicit keep, try our default mailbox */
    if ( ( OmmP != NULL ) && ( OmmP->om_ikeepF ) ) {
	if ( Mfl_testF )
	    printf( "test: implicit keep\n" );
	else if ( PassthroughF ) {
	    /* In pass-through mode implicit keep is just continuing. */
	    /* but log it.. */
	    log_both( "Action: implicit-keep <pass-through> succeeded." );
	}
	else if ( MailboxP != NULL ) {
	    if ( msg_write( OmmP, MailboxP ) ) {
		OmmP->om_ikeepF = FALSE;
		log_both( "Action: implicit-keep \"%s\" succeeded.",
			   (MailboxP != NULL) ? MailboxP : "<inbox>" );
	    }
	    else {
		log_both( "Action: implicit-keep \"%s\" failed.",
			   (MailboxP != NULL) ? MailboxP : "<inbox>" );
	    }
	}
    }

    if ( ( OmmP == NULL ) ||
         ( ( OmmP->om_deliveryC == 0 ) && !Mfl_testF ) )
	return ( SCNOTDONE );
    else
	return ( SCOK );
}
/*

*//* sig_exit( signum )

	Catch signal to exit.. interactive mode

Accepts :

	signum		Signal number caught

Returns :

	<exits>

Notes :


*/

static void
sig_exit ARGLIST( (signum) )
    FARG( int		signum )	/* Signal number */
{
    tt_fini();			/* Finit term */

    /* Attempt to print a notice about the signal, I suppose */
    fprintf( stderr, "mvmda caught signal %d\n", signum );
    exit( 1 );
}
