/* qmail-smtpd-compat.c -- support for mvmtr's qmail-smtpd compatibility

	Copyright 2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Functions and data supporting qmail-smtpd compatibility in mvmtr.

Not all of qmail-smtpd compatibility can be isolated here, of course.  This
is a place to put some of the separable code and data.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/fcntl.h>

#include <mml/mml.h>
#include <mml/mml-alloc.h>
#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-bio.h>
#include <mml/mml-avl.h>

#include "cdb.h"

#include "mvmf.h"

    /* This entire module should only be compiled at all if
       qmail-smtpd compatibility is enabled.  But we'll be sure..
    */
#ifdef	MVMTR_QMAIL_COMPATIBILITY


#include "mvmtr.h"
#include "controls.h"
#include "msgpart.h"
#include "omm.h"
#include "xmtp.h"
#include "qmail-smtpd-compat.h"

#include "errwarn.h"


/* Local definitions */



/* External data referenced */

	BSTR	*Gbs1P;			/* Some general use BSTRs */
	BSTR	*Gbs2P;

	char	*UsehostnameP;		/* Hostname to use (e.g. for greeting) */


/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Read an integer value from a file */
static	STSCODE	qc_file_int PROTO( (char *fnameP, long *valP,
				    char *defaultP) );

    /* Read a file into a string */
static	STSCODE	qc_file_string PROTO( (char *fnameP, UBYTE **strPP,
				       UBYTE *defaultP,
				       BOOL multiF, BOOL endnlF) );

    /* Read a string from an open file */
static	STSCODE	qc_fp_string PROTO( (FILE *fP, char *fnameP,
				      UBYTE **strPP,
				      BOOL multiF, BOOL endnlF) );

    /* Read a file that is a bunch of keys, one per line */
static	STSCODE	qc_file_keys PROTO ( (char *fnameP, AVLTREE **treePP,
				      BOOL reqF) );

/* Private data */

static	struct cdb Mrcpthosts_cdb;	/* cdb access for morercpthosts.cdb */
static	BOOL	MrcpthostsF = FALSE;	/* Whether morercpthosts.cdb is open */

static	AVLTREE	*Badfrom_treeP;		/* Tree for badmailfrom */
static	AVLTREE	*Rcpthosts_treeP;	/* Tree for rcpthosts */

/*

*//* qmc_smtpd_check_mail_from( mbxP, acceptance )

	Check on "mail from" argument 

Accepts :

	mbxP		Ptr to mailbox info
	acceptance	Any already-determined acceptance value.

Returns :

	<value>		MDV_XXX decision value

Notes :

	Checks against any badmailfrom file previously read.  The
	only decision made is to reject.

	Should only be called if the current acceptance value can
	  be overridden.

*/


int
qmc_smtpd_check_mail_from ARGLIST( (mbxP, acceptance) )
   NFARG( XMTPMBX	*mbxP )		/* Mailbox to remember */
    FARG( int		acceptance )	/* Current acceptance value */
{
	UBYTE		ch;
	UBYTE		*strP;
	UBYTE		*sP, *dP;	/* Source/destination for copy */

    /* Return status quo if no badmailfrom file */
    if ( Badfrom_treeP == NULL )
	return ( acceptance );

    /* Get a copy of the passed name so we can lowercase it */
    strP = tmpstr( mbxP->mb_len + 1 );
    for ( sP = mbxP->mb_pathP, dP = strP; ( ch = *sP++ ) != NUL; )
	*dP++ = isupper( ch ) ? tolower( ch ) : ch;
    *dP = NUL;

    /* Check for exact match and for wildcard match */
    if ( ( avlfind( Badfrom_treeP, strP ) != NULL ) ||
         ( avlfind( Badfrom_treeP, strP + mbxP->mb_atX ) != NULL ) )
	return ( MDV_REJECT );

    return ( acceptance );
}
/*

*//* qmc_smtpd_check_rcpt_to( mbxP, acceptance )

	Check on "rcpt to" argument 

Accepts :

	mbxP		Ptr to mailbox info
	acceptance	Current acceptance state

Returns :

	<value>		MDV_XXX decision value

Notes :

	Checks against any rcpthosts or morercpthosts.cdb files.

	If neither file is present, this function makes no judgement.

	If either file is present, this function makes will do a
	  rcpthosts-style match against it or them.  If the domain
	  is found, it will simply return the current acceptance
	  state (making no further judgement).  If the domain is not
	  found, it will return a reject state.  Because of this,
	  caller should only invoke this function if the current
	  acceptance state is overrideable with a reject.

*/


int
qmc_smtpd_check_rcpt_to ARGLIST( (mbxP, acceptance) )
   NFARG( XMTPMBX	*mbxP )		/* Mailbox to remember */
    FARG( int		acceptance )	/* Current MDV_XXX state */
{
	UBYTE		ch;
	UBYTE		*strP;
	UBYTE		*sP, *dP;	/* Source/destination for copy,
					   + other scratch use.
					*/

    /* Return no judgement if no files */
    if ( ( Rcpthosts_treeP == NULL ) && ( !MrcpthostsF ) )
	return ( acceptance );

    /* Get a copy of the passed domain name in lowercase */
    strP = tmpstr( mbxP->mb_len - mbxP->mb_atX );
    for ( sP = &mbxP->mb_pathP[ mbxP->mb_atX +1 ], dP = strP;
	 ( ch = *sP++ ) != NUL; )
	*dP++ = isupper( ch ) ? tolower( ch ) : ch;
    *dP = NUL;

    /* Check the Rcpthosts file first.  exact match and wildcard match. */
    if ( Rcpthosts_treeP != NULL ) {
	for ( sP = strP; *sP != NUL; ) {
	    if ( avlfind( Rcpthosts_treeP, sP ) != NULL )
		return ( acceptance );
	    /* Skip to next '.' and try wild subdomain */
	    while ( ( ch = *++sP ) != NUL )
		if ( ch == '.' )
		    break;
	}
    }

    /* Now try the morrcpthosts.cdb file -- full and wildcard matches */
    if ( MrcpthostsF ) {
	for ( sP = strP; *sP != NUL; ) {
	    if ( cdb_find( &Mrcpthosts_cdb, sP, strlen( (char *)sP ) ) > 0 )
		return ( acceptance );
	    /* Skip to next '.' and try wild subdomain */
	    while ( ( ch = *++sP ) != NUL )
		if ( ch == '.' )
		    break;
	}
    }

    /* Not matched in either file- must reject. */
    return ( MDV_REJECT );
}
/*

*//* qmc_smtpd_init()

	Initialize qmail-smtpd compatibility

Accepts :



Returns :

	<value>		SCxxx status code:
			   SCOK to keep going
			   something else to stop.

Notes :

	Reads all relevant qmail-smtpd control files and assimilates them
	into mvmtr (+qmail compatibility) meaningfulness.

*/

STSCODE
qmc_smtpd_init NOARGLIST
{
	STSCODE		status;
	int		fd;
	long		longval;
	UBYTE		*strP;

    /* File: "me" -- optional.
       Contains a string that is my default hostname.
    */
    status = qc_file_string( QMAILDIR "/control/me", &strP, (UBYTE *)"",
			     FALSE, FALSE );
    if ( status != SCOK ) {
	warning( "qmc_smtpd_init: problem reading control/me" );
	return ( SCNOTDONE );
    }
    if ( strP != NULL ) {
	/* New value for the hostname we use. */
	if ( UsehostnameP != NULL )
	    dealloc( UsehostnameP );
	UsehostnameP = newstr( "qmc_smtpd_init", "UsehostnameP",
			       (char *)strP );
	if ( UsehostnameP == NULL )
	    return ( SCNOTDONE );
    }


    /* File: "smtpgreeting" -- The user-supplied part of the SMTP
       greeting banner.  This sets the "smtp_server_greeting"
       control.

       Differences from qmail-smtpd:
         - Should not include the hostname.  mvmtr always supplies the
	   hostname in the greeting.
	 - May be multi-line.
	 - qmail-smtpd requires some value for this (either from this
	   file, or defaulted from "me").  Since mvmtr always supplies
	   the hostname in the greeting, we don't make user-supplied
	   greeting info mandatory.
    */
    status = qc_file_string( QMAILDIR "/control/smtpgreeting",
			     &strP, (UBYTE *)"",
			     TRUE, FALSE );
    if ( status != SCOK ) {
	warning( "qmc_smtpd_init: problem reading control/smtpgreeting" );
	return ( SCNOTDONE );
    }
    if ( ( strP != NULL ) &&
         ( !cntrl_str_set( estr_tmp_str( "smtp_server_greeting" ),
			   CA_USER,
			   estr_tmp_str( (char *)strP ) ) ) ) {
	warning( "qmc_smtpd_init: problem setting smtp_server_greeting control" );
	return ( SCNOTDONE );
    }


    /* File: "localiphost" -- optional.
       Contains the hostname to use instead of a local IP address
         in a 'RCPT TO' string, when such address is seen in
	 square brackets.  e.g. mem@[10.0.0.1] will be replaced
	 with mem@localiphost.
       If missing, uses the value of "me".
    */
    status = qc_file_string( QMAILDIR "/control/localiphost",
			     &strP,
			     UsehostnameP == NULL ? (UBYTE *)"" :
						(UBYTE *)UsehostnameP,
			     FALSE, FALSE );
    if ( status != SCOK ) {
	warning( "qmc_smtpd_init: problem reading control/localiphost" );
	return ( SCNOTDONE );
    }
    if ( ( strP != NULL ) &&
         ( !cntrl_str_set( estr_tmp_str( "smtp_server_localip_host" ),
			   CA_USER,
			   estr_tmp_str( (char *)strP ) ) ) ) {
	warning( "qmc_smtpd_init: problem setting smtp_server_localip_host control" );
	return ( SCNOTDONE );
    }



    /* File: "timeoutsmtpd" -- optional.
       Contains the timeout for waiting for data from client.
    */
    status = qc_file_int( QMAILDIR "/control/timeoutsmtpd",
			  &longval, NULL );
    if ( status == SCOK ) {
	if ( ( longval < 0 ) || ( longval > 1200 ) )
	    warning( "qmc_smtpd_init: bad control/timeoutsmtpd value %d",
		      longval );
	else
	    cntrl_int_set( estr_tmp_str( "smtp_server_timeout" ),
			   CA_USER, longval );
    }
    else if ( status == SCDONE )
	status = SCOK;
    else {
	warning( "qmc_smtpd_init: problem getting control/timeoutsmtpd value" );
	return ( SCNOTDONE );
    }



    /* File: "rcpthosts" -- optional.
       Contains a list of domain names (one per line) that we will
         receive mail for.  Domain name beginning with "." is a
	 wildcard for subdomains within that domain.
       Any 'RCPT TO' host not listed in rcpthosts will not be
         accepted (subject to further tests).
    */
    status = qc_file_keys( QMAILDIR "/control/rcpthosts",
			   &Rcpthosts_treeP, FALSE );
    if ( status != SCOK ) {
	warning( "qmc_smtpd_init: problem with control/rcpthosts." );
	return ( SCNOTDONE );
    }


    /* File: "badmailfrom" -- optional.
       Contains a list of unacceptable sender addresses, one per line.
         The 'MAIL FROM' address is checked against this list.
	 Line of the form "@domain" is a wildcard for a domain.
    */
    status = qc_file_keys( QMAILDIR "/control/badmailfrom",
			   &Badfrom_treeP, FALSE );
    if ( status != SCOK ) {
	warning( "qmc_smtpd_init: problem with control/badmailfrom." );
	return ( SCNOTDONE );
    }


    /* File: "databytes" -- optional.
       Maximum size of data accepted by smtpd.  Default is 0.
       May be overriden by environment variable DATABYTES.
    */
    strP = (UBYTE *)getenv( "DATABYTES" );
    if ( ( strP != NULL ) && ( *strP != NUL ) ) {
	longval = atod( strP );
	status = SCOK;
    }
    else
	status = qc_file_int( QMAILDIR "/control/databytes",
			      &longval, NULL );

    if ( status == SCOK ) {
	if ( ( longval < 500000L ) && ( longval != 0 ) )
	    warning( "qmc_smtpd_init: bad control/databytes value %d",
		      longval );
	else
	    cntrl_int_set( estr_tmp_str( "smtp_server_msgsize_max" ),
			    CA_USER, longval );
    }
    else if ( status == SCDONE )
	status = SCOK;
    else {
	warning( "qmc_smtpd_init: problem getting control/databytes value" );
	return ( SCNOTDONE );
    }


    /* File: "morercpthosts.cdb" -- optional.
       Generated from "morercpthosts" -- contains a CDB database
         of more domain names.  This is effectively a supplement to
         "rcpthosts" and behaves the same way.
    */
    fd = open( QMAILDIR "/control/morercpthosts.cdb", O_RDONLY );
    if ( fd < 0 ) {
	/* Fail if some problem other than nonexist */
	if ( errno != ENOENT ) {
	    warning( "qmc_smtpd_init: problem opening morercpthosts.cdb" );
	    return ( SCNOTDONE );
	}
    }
    else {
	cdb_init( &Mrcpthosts_cdb, fd );
	MrcpthostsF = TRUE;
    }


    return ( status );
}

/***********************************************************************
 *                                                                     *
 *                     Internal support routines                       *
 *                                                                     *
 ***********************************************************************/

/*

*//* qc_file_int( fnameP, valP, defaultP )

	Reads an integer value from a file

Accepts :

	fnameP		The filename
	valP		Where to store the value
	defaultP	If non-NULL, a default string with the value.

Returns :

	<value>		SCxxx status code, one of:
			   SCOK -- value read or defaulted successfully
			   SCDONE -- No file
			   SCNOTDONE -- some problem.

Notes :

	The value is taken from the first non-blank non-comment line in
	  the file (via qc_file_string()).

	On returns other than SCOK, the value at *valP is untouched.

	defaultP is NULL if there is no default.  A string pointer is
	  used for a default since it can convey both a default and
	  the presence of a default.  Not to mention that it hooks
	  into qc_file_string() easily.

	See qc_file_string() for any other relevant notes

*/

static STSCODE
qc_file_int ARGLIST( (fnameP, valP, defaultP) )
   NFARG( char		*fnameP )	/* Name of control file */
   NFARG( long		*valP )		/* Where to store the value */
    FARG( char		*defaultP )	/* Default string */
{
	STSCODE		status;
	char		*strP;		/* Value string from the file */

    /* Use qc_file_string to read the first significant line of the
       file.  This will cause a string allocation on success; trade
       the overhead for common code.
    */
    status = qc_file_string( fnameP, (UBYTE **)&strP, (UBYTE *)defaultP,
			     FALSE, FALSE );
    if ( status == SCOK ) {
	/* Got something; interpret as int.  Could probably check
	   syntax here if we wanted to be anal.
	*/
	*valP = atod( strP );
	dealloc( strP );
    }

    return ( status );
}
/*

*//* qc_file_string( fnameP, strPP, defaultP, multiF, endnlF )

	Read all or part of a file into a string

Accepts :

	fnameP		The filename
	strPP		Where to put the string pointer
	defaultP	Ptr to default string
	multiF		If multi-line string is acceptable
	endnlF		Whether to terminate string with newline

Returns :

	<value>		SCxxx status code, one of:
			   SCOK -- string read or defaulted successfully
			   SCDONE -- No file, no default
			   SCNOTDONE -- some problem.

Notes :

	*strPP will contain freshly allocated string copy, or NULL.

	A defaultP of NULL will mean there is no default, and a missing
	  file will return SCDONE.

	A defaultP pointing to an empty string ("") will be taken to
	  mean that the default value is the NULL string, and a
	  missing file will return SCOK but with a NULL string pointer.

	Other parameters are as for qc_fp_string, which this function
	  uses.  See that.

*/

static STSCODE
qc_file_string ARGLIST( (fnameP, strPP, defaultP, multiF, endnlF) )
   NFARG( char		*fnameP )	/* Name of control file */
   NFARG( UBYTE		**strPP )	/* Ptr to string pointer */
   NFARG( UBYTE		*defaultP )	/* Default string */
   NFARG( BOOL		multiF )	/* If multi-line OK */
    FARG( BOOL		endnlF )	/* Whether to terminate with newline */
{
	STSCODE		status;
	BOOL		copyF;		/* If we need to make a copy */
	FILE		*fP;
	UBYTE		*retP;		/* Ptr to string to copy for return */

    fP = fopen( fnameP, "r" );
    if ( fP == NULL ) {
	/* File is missing or unreadable. */
	if ( defaultP == NULL )
	    status = SCDONE;		/* There is no default */
	else {
	    status = SCOK;
	    if ( *defaultP == NUL )
		retP = NULL;		/* The default is the NULL string */
	    else {
		retP = defaultP;	/* The default is the default */
		copyF = TRUE;		/* Will need to make a copy */
	    }
	}
    }
    else {
	/* Use qc_fp_string() to read the first string */
	status = qc_fp_string( fP, fnameP, &retP, multiF, endnlF );

	fclose( fP );

	/* on EOF, qc_fp_string() will return NULL pointer.  Default to
	   a blank string in this case.
	*/
	if ( ( status == SCOK ) && ( retP == NULL ) ) {
	    retP = "";
	    copyF = TRUE;		/* Need to make a copy */
	}
	else
	    copyF = FALSE;		/* Already have a copy */
    }

    /* Make a copy of the string if needed */
    if ( ( status == SCOK ) && ( retP != NULL ) && copyF ) {
	retP = (UBYTE *)newstr( "qc_file_string", "return string",
			         (char *)retP );
	if ( retP == NULL )
	    status = SCERR;
    }

    if ( status != SCOK )
	retP = NULL;

    *strPP = retP;

    return ( status );
}
/*

*//* qc_fp_string( fP, fnameP, strPP, multiF, endnlF )

	Read from an open file into a string

Accepts :

	fP		Ptr to open file handle
	fnameP		Name of file (for messages)
	strPP		Where to put the string pointer
	multiF		If multi-line string is acceptable
	endnlF		Whether to terminate string with newline

Returns :

	<value>		SCxxx status code, one of:
			   SCOK -- string read or EOF
			   SCNOTDONE -- some problem.

Notes :

	*strPP will contain freshly allocated string copy, or NULL
	   if end of file.

	If multiF is FALSE, only the next non-comment non-empty line
	  of the file will be used.  Otherwise all non-comment non-empty
	  lines will be concatenated (up to a reasonable limit).  Note
	  that an empty line is a line with nothing on it, as opposed to
	  a blank line which would be a non-empty line with just whitespace
	  on it.

	Any line in the file may be preceeded with '\' to take the rest of
	  the line literally.  This provides a way of escaping comment
	  and blank lines.

	Any line in the file may be ended with '\' to have the following
	  line (non-empty line) treated as a virtual continuation (no
	  newline between them).

	If endnlF is TRUE, the string if read from the file is always
	  terminated with a newline.  If endnlF, the string if read
	  from the file always has all terminating newlines stripped.
	  Default string, if used, will not be altered.

	Uses scratch BSTRs Gbs1P and Gbs2P

*/

static STSCODE
qc_fp_string ARGLIST( (fP, fnameP, strPP, multiF, endnlF) )
   NFARG( FILE		*fP )		/* Open file pointer */
   NFARG( char		*fnameP )	/* Name for messages */
   NFARG( UBYTE		**strPP )	/* Ptr to string pointer */
   NFARG( BOOL		multiF )	/* If multi-line OK */
    FARG( BOOL		endnlF )	/* Whether to terminate with newline */
{
	STSCODE		status;
	BOOL		eofF;		/* TRUE if end of file */
	UBYTE		*bP;		/* Convenient ptr */
	UBYTE		*retP;		/* Ptr to string to copy for return */

    Gbs2P->bs_bC = 0;			/* We'll build the string here */
    eofF = TRUE;			/* No data yet */

    for ( ; ; ) {
	/* Get next line */
	if ( fgetbline( fP, Gbs1P ) < 0 )
	    break;

	/* Get pointer to significant start of line, and skip
	   empty or comment line.
	*/
	bP = Gbs1P->bs_bP;
	if ( *bP == '\\' )
	    ++bP;
	else if ( ( *bP == '#' ) || ( *bP == NUL ) )
	    continue;

	/* Valid line, store it into the string we're generating */
	eofF = FALSE;			/* Not end of file */
	bstr_cat( Gbs2P, (char *)bP );

	/* If the line ends with backslash, it's continued */
	if ( ( Gbs2P->bs_bC > 0 ) &&
	     ( Gbs2P->bs_bP[ Gbs2P->bs_bC -1 ] == '\\' ) ) {
	    -- Gbs2P->bs_bC;
	}
	else {
	    /* Not continued-- done if not multline */
	    if ( !multiF )
		break;

	    /* Multi-line.  Add newline to this. */
	    bstr_putb( Gbs2P, '\n' );
	}
    }

    /* If end of file, return null pointer */
    if ( eofF ) {
	*strPP = NULL;
	return ( SCOK );
    }

    /* Honor the end-newline flag */
    if ( endnlF ) {
	if ( ( Gbs2P->bs_bC == 0 ) ||
	     ( Gbs2P->bs_bP[ Gbs2P->bs_bC -1 ] != '\n' ) )
	    bstr_putb( Gbs2P, '\n' );
    }
    else {
	while ( ( Gbs2P->bs_bC > 0 ) &&
	        ( Gbs2P->bs_bP[ Gbs2P->bs_bC -1 ] == '\n' ) )
	    --Gbs2P->bs_bC;
    }

    /* In any case, NUL-terminate it */
    bstr_nulterm( Gbs2P );

    /* Check for overflow */
    if ( Gbs2P->bs_bC > Gbs2P->bs_bM ) {
	warning( "string%s too large in %s",
	          multiF ? "s" : "",
		  fnameP );
	status = SCNOTDONE;
    }
    else {
	/* OK to return what we got. */
	retP = Gbs2P->bs_bP;
	status = SCOK;
    }

    /* Fill in the return string pointer */
    if ( status == SCOK ) {
	if ( retP == NULL )
	    *strPP = NULL;
	else {
	    /* Make a copy of the string we're returning */
	    bP = (UBYTE *)newstr( "qc_fp_string", "return string",
				 (char *)retP );
	    if ( bP == NULL )
		status = SCNOTDONE;
	    else
		*strPP = bP;
	}
    }

    if ( status != SCOK )
	*strPP = NULL;

    return ( status );
}
/*

*//* qc_file_keys( fnameP, treePP, reqF )

	Read a file that is a bunch of keys, one per line

Accepts :

	fnameP		The filename
	treePP		Ptr to the AVL tree pointer
	reqF		TRUE if the file is mandatory

Returns :

	<value>		SCxxx status code, one of:
			   SCOK -- File read or, if not required, missing
			   SCDONE -- No file when required
			   SCNOTDONE -- some problem.

Notes :

	The keys are the non-blank non-comment lines in the file (via
	  qc_file_string()).

	All keys will be forced to lowercase.

	See qc_file_string() for any other relevant notes

*/

static STSCODE
qc_file_keys ARGLIST( (fnameP, treePP, reqF) )
   NFARG( char		*fnameP )	/* Name of control file */
   NFARG( AVLTREE	**treePP )	/* Ptr to tree ptr */
    FARG( BOOL		reqF )		/* TRUE if file is required */
{
	STSCODE		status;
	UBYTE		ch;
	UBYTE		*strP;		/* String ptr */
	UBYTE		*cP;		/* Copy */
	FILE		*fP;

    fP = fopen( fnameP, "r" );
    if ( fP == NULL )
	return ( reqF ? SCDONE : SCOK );

    /* Loop reading keys and adding them to the tree.  Tree will be
       initialized if needed.
    */
    for ( status = SCOK; status == SCOK; ) {
	status = qc_fp_string( fP, fnameP, &strP, FALSE, FALSE );
	if ( ( status != SCOK ) || ( strP == NULL ) )
	    break;			/* Problem or EOF */

	/* Got a key.  Initialize the tree if necessary */
	if ( *treePP == NULL ) {
	    *treePP = avlnew( NULL,
			      mml_avl_strp_cmpkey,
			      mml_avl_strp_mknode,
			      mml_avl_strp_rmnode );
	    if ( *treePP == NULL ) {
		status = SCERR;
		break;
	    }
	}

	/* Make the key lowercase */
	for ( cP = strP; ( ( ch = *cP ) != NUL ); ++cP )
	    if ( isupper( ch ) )
		*cP = tolower( ch );

	/* Insert */
	switch ( avlinsert( *treePP, strP, NULL ) ) {
	    case -1:			/* Failure */
		status = SCERR;
		break;

	    case 0:			/* Success */
		break;

	    case 1:			/* Duplicate */
		dealloc( strP );
		break;

	    default:
		warning( "qc_file_keys: unexpected return from avlinsert." );
		status = SCERR;
		break;
	}
    }

    fclose( fP );

    return ( status );
}


/*

*//* end of qmail compatibility conditional */

#endif		/* MVMTR_QMAIL_COMPATIBILITY */
