/* cdbgen.c -- make a .cdb file from a generic input file

	Copyright 2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Similar to the 'cdbmake' bundled with djb's cdb, but with a different input
format (I mean, who wants to maintain a file with counted records?)

An alternative implementation could simply emit a format that cdbmake
wants, and pipe to cdbmake, but that's needlessly separatist.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <fcntl.h>
#include <ctype.h>

#include <mml/mml.h>

#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>

#include "cdb_make.h"
#include "cdb.h"

/* Local definitions */



/* External data referenced */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */

/* External routines used */



/* Local data publicly available */



/* Local routines and forward references */

    /* Add a record to CDB file */
static	BOOL	cdb_addrec PROTO( (struct cdb_make *cmP,
				   UBYTE *keyP, int keyL,
				   UBYTE *valP, int valL) );

    /* Do the cdbgen operation */
static	BOOL	cdbgen PROTO( (int ofd, FILE *ifP) );

static	void	usage();


/* Private data */

static	char	*KvsepP;		/* Key/value separator string */

int
main( argc, argv )
	int		argc;
	char		**argv;
{
	BOOL		usageOK;
	BOOL		okF;
	int		optch;
	int		ofd;		/* Output file */
	char		*ofnameP;	/* Output filename */
	char		*kvfnameP;	/* Key/value filename */
	FILE		*kvfP;		/* Key/value file ptr */

    usageOK = TRUE;
    kvfnameP = NULL;
    ofnameP = NULL;

    while ( usageOK &&
	   ( optch = getopt( argc, argv, "o:s:" ) ) != EOF ) {
	switch ( optch ) {
	    case	'o':		/* Output filename */
		ofnameP = optarg;
		break;

	    case	's':		/* Key/Value separator string */
		KvsepP = optarg;
		break;

	    default:
		usageOK = FALSE;
		break;
	}
    }

    if ( usageOK && ( optind < argc ) )
	kvfnameP = argv[ optind++ ];

    /* No more command line arguments allowed */
    if ( usageOK && ( optind < argc ) )
	usageOK = FALSE;

    /* Output filename is required */
    if ( usageOK && ( ofnameP == NULL ) ) {
	fprintf( stderr, "Output filename (via -o) is required.\n" );
	usageOK = FALSE;
    }

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();

    /* Default key/value separator */
    if ( KvsepP == NULL )
	KvsepP = ":";

    /* Open up input and proceed */
    if ( kvfnameP != NULL ) {
	if ( ( kvfP = fopen( kvfnameP, "r" ) ) == NULL ) {
	    fprintf( stderr, "Can't open key/value file \"%s\".\n",
				kvfnameP );
	    perror( "fopen" );
	    exit( 1 );
	}
    }
    else
	kvfP = stdin;

    ofd = open( ofnameP, O_WRONLY|O_TRUNC|O_CREAT, 0444 );
    if ( ofd < 0 ) {
	fprintf( stderr, "Can't open output file \"%s\"\n", ofnameP );
	perror( "open" );
	exit( 1 );
    }

    okF = cdbgen( ofd, kvfP );

    if ( kvfnameP != NULL )
	fclose( kvfP );

    if ( !okF )
	unlink( ofnameP );

    if ( close( ofd ) < 0 ) {
	fprintf( stderr, "close of \"%s\" failed.\n", ofnameP );
	perror( "close" );
	okF = FALSE;
    }

    exit( okF ? 0 : 1 );
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/

static void
usage()
{
    printf( "usage: cdbgen [flags] [inputfile]\n" );
    printf( "    Flags are:\n" );
    printf( "       -o output-file   The name of the cdb file to create.\n" );
    printf( "       -s sep           Specify key/value separator string.\n" );
    printf( "\n" );
}
/*

*//* cdb_addrec( cmP, keyP, keyL, valP, valL )

	Add a record to the cdb file

Accepts :

	cmP		Ptr to cdb_make struct
	keyP		Ptr to key to add
	keyL		Length of key
	valP		Ptr to value for key
	valL		Length of value

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
cdb_addrec ARGLIST( (cmP, keyP, keyL, valP, valL) )
   NFARG( struct cdb_make *cmP )	/* cdb_make thing */
   NFARG( UBYTE		*keyP )		/* Ptr to key */
   NFARG( int		keyL )		/* Key length */
   NFARG( UBYTE		*valP )		/* Ptr to value */
    FARG( int		valL )		/* Length of value */
{
	int		i;
	ULONG		hashval;

    /* Start the cdb record add */
    if ( cdb_make_addbegin( cmP, keyL, valL ) < 0 ) {
	fprintf( stderr, "cdb_make_addbegin failed (write error?)\n" );
	perror( "cdb_make_addbegin" );
	return ( FALSE );
    }

    /* Calculate hash value from key while writing the key */
    hashval = CDB_HASHSTART;
    for ( i = 0; i < keyL; ++i ) {
	if ( buffer_PUTC( &cmP->b, ((char *)keyP)[i] ) < 0 ) {
	    fprintf( stderr, "Error writing key.\n" );
	    return ( FALSE );
	}
	hashval = cdb_hashadd( hashval, keyP[i] );
    }

    /* Write the data buffer via the cdb_make handle */
    for ( i = 0; i < valL; ++i ) {
	if ( buffer_PUTC( &cmP->b, ((char *)valP)[i] ) < 0 ) {
	    fprintf( stderr, "Error writing value.\n" );
	    return ( FALSE );
	}
    }

    /* End the cdb record add */
    if ( cdb_make_addend( cmP, keyL, valL, hashval ) < 0 ) {
	fprintf( stderr, "cdb_make_addend failed (write error?)\n" );
	perror( "cdb_make_addend" );
	return ( FALSE );
    }

    return ( TRUE );
}
/*

*//* cdbgen( ofd, ifP )

	Do cdb generation from input file to output file

Accepts :

	ofd		file descriptor for output file
	ifP		FILE ptr for intput file

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :


*/

static BOOL
cdbgen ARGLIST( (ofd, ifP) )
   NFARG( int		ofd )		/* output file descriptor */
    FARG( FILE		*ifP )		/* Input FILE ptr */
{
	BOOL		okF;
	int		seplen;		/* Length of separator string */
	int		keylen;		/* Length of key string */
	int		vlen;		/* Length of value string */
	int		len;
	long		lineC;
	UBYTE		*bP;		/* Ptr to byte string */
	UBYTE		*vP;		/* Ptr to value part of string */
	BSTR		*bsP;		/* BSTR buffer for input lines */
	struct cdb_make	cm;		/* handle for cdb_make ops */

    /* Init the cdb_make handle */
    if ( cdb_make_start( &cm, ofd ) < 0 ) {
	fprintf( stderr, "cdb_make_start failed.\n" );
	perror( "cdb_make_start" );
	return ( FALSE );
    }

    /* Create BSTR to use */
    bsP = bstr_new( "cdbgen", 1000, 1000 );
    if ( bsP == NULL ) {
	fprintf( stderr, "Can't allocate bsP\n" );
	return ( FALSE );
    }

    seplen = strlen( KvsepP );

    /* Loop getting and processing records */
    for ( lineC = 1, okF = TRUE; okF; ++lineC) {
	len = fgetbtrecc( ifP, bsP,
		 ((TXR_END_CRLF)|(TXR_END_CR)|(TXR_END_LF)),
		 NULL );
	if ( len < 0 )
	    break;

	/* Error if overflow. */
	if ( bsP->bs_bC >= bsP->bs_bM ) {
	    fprintf( stderr, "Line %ld too large- aborted.\n", lineC );
	    okF = FALSE;
	    break;
	}

	/* Ignore if comment or blank line.  A comment is signalled
	   by a '#' leading character.  A leading '\' char always
	   escapes the next.
	*/
	bP = bsP->bs_bP;
	if ( ( *bP == '#' ) || ( *solbat( (char *)bP, TRUE ) == NUL ) )
	    continue;
	if ( *bP == '\\' ) {
	    if ( *solbat( (char *)++bP, TRUE ) == NUL )
		continue;
	}

	/* Locate key/data separator */
	vP = (UBYTE *)strstr( (char *)bP, KvsepP );
	if ( vP == NULL ) {
	    fprintf( stderr, "Line %ld format error: no separator string.\n",
				lineC );
	    okF = FALSE;
	    break;
	}
	*vP = NUL;
	keylen = vP - bP;
	vP += seplen;
	vlen = &bsP->bs_bP[bsP->bs_bC] - vP;

       okF = cdb_addrec( &cm, bP, keylen, vP, vlen );
    }

    if ( okF ) {
	if ( cdb_make_finish( &cm ) < 0 ) {
	    fprintf( stderr, "cdb_make_finish error.\n" );
	    perror( "cdb_make_finish" );
	    okF = FALSE;
	}
    }

    return ( okF );
}
