/* dnsblc.c -- Check an IP address against DNSBL(s)

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Checks an IP address against a specific DNSBL or against all known.

	dnsblc ipaddr
		check all known DNSBLs

	dnsblc ipaddr dnsbl rcode
		check against dnsbl/rcode

	(see dnsbl.conf)

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <pwd.h>
#include <signal.h>

#include "mvmf.h"

#include <mml/mml.h>

#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-dns.h>

#include "msgpart.h"
#include "omm.h"
#include "dnsbl.h"
#include "data.h"


/* Local definitions */

typedef enum {				/* Private operating modes */
    POM_NONE,				/* Not defined */
    POM_NORMAL,				/* Normal */
    POM_VERBOSE,			/* Show all results */
    POM_QUIET,				/* Don't show results at all */
}  POPMODE;


/* External data referenced */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */

/* External routines used */



/* Local data publicly available */

    /* To satisfy linker */
OMM	*OmmP;

/* Local routines and forward references */

static	BOOL	blcheck PROTO( (char *BlnameP, char *rnameP) );
static	void	blcheckpf PROTO( (char *BlnameP, char *rnameP ) );

static	void	usage();


/* Private data */

static	char	*BlnameP;		/* DNSBL name to check */
static	int	MatchC = 0;		/* Match counter */
static	POPMODE	Opmode = POM_NONE;	/* Operating mode */
static	char		*RnameP;	/* DNSBL result name to check */
static	char	*TestipP;		/* The IP addr to check */

int
main( argC, argv )
	int		argC;
	char		**argv;
{
	BOOL		usageOK;
	int		optch;
	char		*ipaddrP;	/* IP addr to check */

    usageOK = TRUE;

    while ( usageOK &&
	   ( optch = getopt( argC, argv, "qv" ) ) != EOF ) {
	switch ( optch ) {
	    case	'q':		/* quiet mode */
		if ( Opmode != POM_NONE ) {
		    fprintf( stderr, "Conflicting flag -%c.\n", optch );
		    usageOK = FALSE;
		}
		else {
		    Opmode = POM_QUIET;
		}
		break;


	    case	'v':		/* verbose mode */
		if ( Opmode != POM_NONE ) {
		    fprintf( stderr, "Conflicting flag -%c.\n", optch );
		    usageOK = FALSE;
		}
		else {
		    Opmode = POM_VERBOSE;
		}
		break;


	    default:
		usageOK = FALSE;
		break;
	}
    }

    if ( usageOK ) {
	/* Get the IP address to check */
	if ( optind < argC ) {
	    ipaddrP = argv[ optind++ ];

	    /* Get optional blname and result code */
	    if ( optind < argC ) {
		BlnameP = argv[ optind++ ];
		if ( optind < argC )
		    RnameP = argv[ optind++ ];
		else
		    RnameP = NULL;
	    }
	    else
		BlnameP = NULL;
	}
	else
	    /* No IP address given-- bad usage */
	    usageOK = FALSE;
    }


    /* Bad if any unconsumed args */
    if ( usageOK && ( optind < argC ) )
	usageOK = FALSE;

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    /* Set the operating mode if not given */
    if ( Opmode == POM_NONE ) {
	if ( RnameP != NULL )
	    /* Specific result and dnsbl -- be verbose */
	    Opmode = POM_VERBOSE;
	else
	    /* Wildcard result name -- do normal mode */
	    Opmode = POM_NORMAL;
    }

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();

    /* Initialize all the crap we carry around because of using
       generic mvmf stuff
    */
    data_init();
    dnsbl_init();

    /* Set up IP address */
    TestipP = ipaddrP;

    /* Walk the DNSBL database */
    dnsbl_walk( blcheck );

    /* If any matches, exit 0.  No matches, exit 1. */
    exit ( ( MatchC > 0 ) ? 0 : 1 );
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/

static void
usage()
{
    printf( "\n" );
    printf( "usage: dnsblc [flags] ip-addr [blname [rcode] ]\n" );
    printf( "    checks against registered DNSBLs\n" );
    printf( "    Flags are:\n" );
    printf( "       -q               Quiet: check but do not print results.\n" );
    printf( "       -v               Verbose: print result whether listed or not.\n" );
    printf( "\n" );
    printf( "DNSBLs are registered in the system-wide file\n" );
    printf( "    \"%s\"\n", LIBDNSBL );
}
/*

*//* blcheck( BlnameP, rnameP )

	Check against a specific blacklist, report result if positive

Accepts :

	BlnameP		DNSBL name
	rnameP		Resultcode name

Returns :

	<value>		TRUE always

Notes :

	Called as the routine-to-call for dnsbl_walk()

	Simply calls blcheckpf

*/

static BOOL
blcheck ARGLIST( (blnameP, rnameP) )
   NFARG( char		*blnameP )	/* The DNSBL name */
    FARG( char		*rnameP )	/* The resultcode name */
{
    /* Note that a final termination call will be made with
        (NULL, NULL) -- so we ignore a call with a null blname
    */
    if ( blnameP != NULL ) {
	/* Hand off to blcheckpf after verifying dnsbl and
	   result code names if specified
	*/
	if ( ( ( BlnameP == NULL ) ||
               ( strcmp( BlnameP, blnameP ) == 0 ) ) &&
	     ( ( RnameP == NULL ) ||
	       ( strcmp( RnameP, rnameP ) == 0 ) ) )
	    blcheckpf( blnameP, rnameP );
    }

    return ( TRUE );
}
/*

*//* blcheckpf( blnameP, rnameP )

	Check against a specific blacklist, maybe report result

Accepts :

	blnameP		DNSBL name
	rnameP		Resultcode name

Returns :

	<nothing>

Notes :


*/

static void
blcheckpf ARGLIST( (blnameP, rnameP) )
   NFARG( char		*blnameP )	/* The DNSBL name */
    FARG( char		*rnameP )	/* The resultcode name */
{
	BOOL		status;

    if ( ( blnameP == NULL ) || ( rnameP == NULL ) )
	return;

    /* Get status */
    status = dnsbl_check_ip( TestipP, blnameP, rnameP );

    /* If positive, bump the match count */
    if ( status )
	++MatchC;

    /* Decide whether to print */
    if ( ( Opmode != POM_QUIET ) &&
	 ( ( Opmode == POM_VERBOSE ) ||
	   ( ( Opmode == POM_NORMAL ) && status ) ) ) {

	printf( "%s against %s/%s:  %s\n",
	        TestipP,
	        blnameP, rnameP,
	        status ? "positive" : "negative" );
    }
}
