/* mvmdalogsum.c -- Summarize a mvmda logfile

	Copyright 2003,2004,2005,2006,2007
		by Mark E. Mallett, MV Communications, Inc.

	See the "LICENSE" file for terms.

Summarizes an mvmda log file.

*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <ctype.h>

#include "mvmf.h"

#include <mml/mml.h>

#include <mml/mml-ll.h>
#include <mml/mml-str.h>
#include <mml/mml-in.h>
#include <mml/mml-iopf.h>
#include <mml/mml-avldb.h>
#include <mml/mml-alloc.h>

#include "msgpart.h"
#include "omm.h"

/* Local definitions */

    /* The name to associate with the totals/summary */
#define	TOTALSNAME	"**all**"

    /* avldb data element types */
enum {
    DT_LOGSUM = 100,
};

    /* A log summary entry */
typedef struct {
	/* Counters */
    int		ls_actionC;		/* Total actions */
    int		ls_discardC;		/* Discard */
    int		ls_fileintoC;		/* Fileinto */
    int		ls_filespamC;		/* Fileinto spam folder */
    int		ls_ikeepC;		/* Implicit keep */
    int		ls_keepC;		/* keep */
    int		ls_redirectC;		/* Redirect */
    int		ls_rejectC;		/* Reject */
    int		ls_vacC;		/* Vacation actions */
    int		ls_vacinhC;		/* Vacations inhibited */
}  LOGSUM;

/* External data referenced */

extern	char	*optarg;		/* getopt() argument */
extern	int	optind;			/* getopt() index */

/* External routines used */



/* Local data publicly available */

	OMM	*OmmP;			/* The open mail message handle */

/* Local routines and forward references */

	void	lfcollect PROTO( (FILE *lfP, char *lfnameP) );
	BOOL	lfcollectline PROTO( (char *lfnameP, int lineN,
				      char *unameP,
				      TKNS *tknsP, int basetX, int count) );
	void	lfreport PROTO( (void) );
	void	lfsum PROTO( (FILE *lfP, char *lfnameP) );
	LOGSUM	*logsum_get PROTO( (char *nameP) );
	int	report_logsum PROTO( (int nodetype, char *nameP, LOGSUM **lsPP) );


static	void	usage();


/* Private data */

static	int	RepC;			/* # entries reported */

int
main( argc, argv )
	int		argc;
	char		**argv;
{
	char		usageOK;
	int		optch;
	char		*lfnameP;	/* Logfile name */
	FILE		*lfP;		/* Logfile FILE ptr */

    usageOK = TRUE;
    lfnameP = NULL;

    while ( usageOK &&
	   ( optch = getopt( argc, argv, "x" ) ) != EOF ) {
	switch ( optch ) {
	    case	'x':
		exit(1);

	    default:
		usageOK = FALSE;
		break;
	}
    }

    if ( usageOK && ( optind < argc ) )
	lfnameP = argv[ optind++ ];

    /* No more command line arguments allowed */
    if ( usageOK && ( optind < argc ) )
	usageOK = FALSE;

    if ( !usageOK ) {
	usage();
	exit( 1 );
    }

    /* This brings in pthreads stubs for mml use since we don't
       use pthreads in this application.
    */
    mml_fthread();

    /* Init the avldb */
    avldb_init();

    /* Open logfile if specified */
    if ( lfnameP != NULL ) {
	lfP = fopen( lfnameP, "r" );
	if ( lfP == NULL ) {
	    fprintf( stderr, "Can't open file: \"%s\" -- %s\n",
		    lfnameP, strerror( errno ) );
	    exit( 1 );
	}

	lfsum( lfP, lfnameP );
	fclose( lfP );
    }
    else {
	lfsum( stdin, "<stdin>" );
    }

    exit( 0 );
}
/*

*//* usage()

	Print usage message

Accepts :

Returns :

*/

static void
usage()
{
    printf( "usage: mvmdalogsum [flags] [logfile]\n" );
    printf( "    Flags are:\n" );
    printf( "       -x               dummy flag-- exits.\n" );
    printf( "\n" );
}
/*

*//* lfcollect( lfP, lfnameP )

	Do logfile collection

Accepts :

	lfP		FILE ptr for open logfile
	lfnameP		A descriptive name for the file

Returns :

	<nothing>


Notes :


*/

void
lfcollect ARGLIST( (lfP, lfnameP) )
   NFARG( FILE		*lfP )		/* Open file ptr */
    FARG( char		*lfnameP )	/* Descriptive name */
{
	int		cC;		/* Character count */
	int		lineN;		/* Line number */
	int		bufset;		/* Current buffer set */
	int		count;		/* Repeat count */
	int		basetX;		/* Base token index */
	char		*unameP;	/* Username */
	UBYTE		*tbP;		/* raw token buffer ptr */
	BSTR		*bstrP;		/* Ptr from current bstrpool[] */
	TKNS		*tknsP;		/* Ptr from current tknspool[] */
	BSTR		*bstrpool[2];	/* BSTRs for input */
	BSTR		*tkbstrpool[2];	/* BSTRs for tokenizing */
	TKNS		*tknspool[2];	/* Tokens pool */

    /* When reading/tokenizing input lines we alternate between two
       string/token buffers.  This is because of syslog's propensity
       to log strings like "last message repeated n times".  In order
       to properly process one of those things, we have to have a copy
       of the last message to reprocess.  So we alternate between sets
       of buffers.
    */

    /* Initialize the buffers */
    for ( bufset = 0; bufset < 2; ++bufset ) {
	/* Input buffer */
	bstrpool[bufset] = bstr_new( "lfcollect input buffer",
				      1000, 500 );
	if ( bstrpool[bufset] == NULL )
	    return;

	/* Token buffer */
	tkbstrpool[bufset] = bstr_new( "lfcollect tokens buffer",
				      1000, 2000 );
	if ( tkbstrpool[bufset] == NULL )
	    return;

	/* Token ptr */
	tknspool[bufset] = tkns_new( tkbstrpool[bufset], 50,
					(UBYTE *)"", (UBYTE *)" \t,;:",
					FALSE );
	if ( tknspool[bufset] == NULL )
	    return;
    }

    bufset = 0;				/* Initial buffer set */

    for ( lineN = 1; ; ++lineN) {
	/* Select next buffer set */
	bstrP = bstrpool[bufset];
	tknsP = tknspool[bufset];

	/* Count (number of times to apply this line) should be 1 */
	count = 1;

	cC = fgetbline( lfP, bstrP );
	if ( cC < 0 )
	    break;

	/* Tokenize */
	tkline( bstrP, tknsP );
	tbP = tknsP->tp_bsP->bs_bP;

	/* Distinguish between a syslog format and a private log format */
	if ( isalpha( tbP[ tknsP->tp_tbP[0].tb_startX] ) ) {
	    /* Syslog format */

	    /* Raw lines will look something like this:
Aug 14 03:08:50 iridium mvmda[29144]: User: 101/mem Action: implicit-keep "Mailbox" succeeded.
Aug 14 03:09:05 iridium last message repeated 2 times
	       and tokenized at colons and such.
	    */

	    /* Check for "repeated n times" syslog thing. */
	    if ( ( tknsP->tp_tC == 11 ) &&
	         ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[6].tb_startX]), "last" ) == 0 ) &&
	         ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[7].tb_startX]), "message" ) == 0 ) &&
	         ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[8].tb_startX]), "repeated" ) == 0 ) &&
	         ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[10].tb_startX]), "times" ) == 0 ) ) {

		/* If this is the first line, there is no line to repeat */
		if ( lineN == 1 )
		    continue;

		/* Do the previous buffer another n times */
		count = atoi( (char *)&(tbP[ tknsP->tp_tbP[10].tb_startX]) );
		if ( count < 1 ) {
		    continue;
		}

		/* Select other buffer set */
		bufset = 1 - bufset;
		bstrP = bstrpool[bufset];
		tknsP = tknspool[bufset];
		tbP = tknsP->tp_bsP->bs_bP;
	    }
	    /* Check for a valid log line with data */
	    if ( ( tknsP->tp_tC > 10 ) &&
	         ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[7].tb_startX]), "User" ) == 0 ) ) {
		basetX = 9;		/* Base token index of info */
		unameP = (char *)&(tbP[ tknsP->tp_tbP[8].tb_startX]);
	    }
	    else {
		/* No valid data, skip */
		continue;
	    }
	}
	else {
	    /* Private-log format */

	    /* Log lines will look like this:
20050205.015221 User: 101/mem Action: fileinto "M/zs" succeeded.
20050205.015221 User: 101/mem Stats: elapse-ms 1175, user-ms 8, sys-ms 16, rss-kb 1136
20050205.015258 User: 101/mem Action: fileinto "M/geez" succeeded.
20050205.015258 User: 101/mem Stats: elapse-ms 1321, user-ms 21, sys-ms 0, rss-kb 1128

	    The "User nn/mmm" part may not be present; if it's missing,
	    we just supply something fake.

	    */
	    if ( tknsP->tp_tC > 3 ) {
		if ( ( tknsP->tp_tC > 5 ) &&
	             ( strcmp( (char *)&(tbP[ tknsP->tp_tbP[1].tb_startX]), "User" ) == 0 ) ) {
		    basetX = 3;
		    unameP = (char *)&(tbP[ tknsP->tp_tbP[2].tb_startX]);
		}
		else {
		    unameP = "me";
		    basetX = 1;
		}
	    }
	    else {
		fprintf( stderr, "Too few tokens in plog line.\n" );
		break;
	    }
	}

	/* Common code (syslog and private log) */
	if ( !lfcollectline( lfnameP, lineN, unameP, tknsP, basetX, count ) )
	    break;
    }
}
/*

*//* lfcollectline( lfnameP, lineN, unameP, tknsP, basetX, count )

	Collect information from a log line

Accepts :

	lfnameP		Descriptive name for logfile input
	lineN		Line number in log file (for reporting)
	unameP		Username string
	tksnP		Ptr to tokens pool handle
	basetX		Base token index of where info starts
	count		Repeat count to apply to this line

Returns :

	<value>		TRUE if OK to proceed
			FALSE if not

Notes :


*/

BOOL
lfcollectline ARGLIST( (lfnameP, lineN, unameP, tknsP, basetX, count) )
   NFARG( char		*lfnameP )	/* Input name */
   NFARG( int		lineN )		/* Line number */
   NFARG( char		*unameP )	/* Username string */
   NFARG( TKNS		*tknsP )	/* Tokens pool */
   NFARG( int		basetX )	/* Base token number */
    FARG( int		count )		/* Repeat count */
{
	char		*kwdP;		/* Ptr to first keyword */
	char		*arg1P;		/* Ptr to first arg */
	char		*arg2P;		/* Ptr to second arg */
	char		*ukeyP;		/* Key for entry */
	UBYTE		*tbP;		/* raw token buffer ptr */
	LOGSUM		*lsuP;		/* Logsum ptr for user */
	LOGSUM		*lsgP;		/* Logsum ptr for global info */

    tbP = tknsP->tp_bsP->bs_bP;		/* Ptr to raw buffer */

    /* Get initial keyword */
    kwdP = (char *)&(tbP[ tknsP->tp_tbP[basetX].tb_startX]);

    /* Get any first token argument */
    if ( tknsP->tp_tC >= basetX )
	arg1P = (char *)&(tbP[ tknsP->tp_tbP[basetX +1].tb_startX]);
    else
	arg1P = NULL;

    /* Get any second token argument */
    if ( tknsP->tp_tC >= basetX )
	arg2P = (char *)&(tbP[ tknsP->tp_tbP[basetX +2].tb_startX]);
    else
	arg2P = NULL;

    /* The username comes in as uid/name .. for the key, just use
       the name.
    */
    ukeyP = strchr( unameP, '/' );
    if ( ukeyP != NULL )
	++ukeyP;
    else
	ukeyP = unameP;

    /* Get the logsum entries for this user and for the global accounting */
    if ( ( lsuP = logsum_get( ukeyP ) ) == NULL )
	return ( FALSE );

    if ( ( lsgP = logsum_get( TOTALSNAME ) ) == NULL )
	return ( FALSE );

    /* Process according to keyword */
#if 0
    printf( "Line %d  unameP=\"%s\"  kwdP=\"%s\"    arg1P=\"%s\"\n",
	   lineN, unameP,
	   kwdP,
	   (arg1P != NULL) ? arg1P : "<none>" );
#endif


    if ( stricmp( kwdP, "Action" ) == 0 ) {
	if ( arg1P == NULL ) {
	    fprintf( stderr, "%s line %d: no action specified.\n",
		        lfnameP, lineN );
	    return ( TRUE );
	}

	else if ( ( stricmp( arg1P, "discard" ) == 0 ) ||
	          ( stricmp( arg1P, "dnsbl_drop" ) == 0 ) ) {
	    lsgP->ls_discardC += count;
	    lsuP->ls_discardC += count;
	}
	else if ( stricmp( arg1P, "implicit-keep" ) == 0 ) {
	    lsgP->ls_ikeepC += count;
	    lsuP->ls_ikeepC += count;
	}
	else if ( ( stricmp( arg1P, "fileinto" ) == 0 ) ||
	          ( stricmp( arg1P, "dnsbl_file" ) == 0 ) ) {
	    lsgP->ls_fileintoC += count;
	    lsuP->ls_fileintoC += count;

	    /* See if it looks like a spam folder */
	    if ( ( arg2P != NULL ) &&
	         ( strmatch_exact( NULL,
				   (UBYTE *)"spam", (UBYTE *)arg2P,
				   MF_SUBSTR|MF_IGNCASE ) ) ) {
		lsgP->ls_filespamC += count;
		lsuP->ls_filespamC += count;
	    }
	}
	else if ( stricmp( arg1P, "keep" ) == 0 ) {
	    lsgP->ls_keepC += count;
	    lsuP->ls_keepC += count;
	}
	else if ( stricmp( arg1P, "redirect" ) == 0 ) {
	    lsgP->ls_redirectC += count;
	    lsuP->ls_redirectC += count;
	}
	else if ( stricmp( arg1P, "reject" ) == 0 ) {
	    lsgP->ls_rejectC += count;
	    lsuP->ls_rejectC += count;
	}
	else if ( stricmp( arg1P, "vacation" ) == 0 ) {
	    lsgP->ls_vacC += count;
	    lsuP->ls_vacC += count;
	    if ( ( arg2P != NULL ) &&
	         ( stricmp( arg2P, "inhibited" ) == 0 ) ) {
		lsgP->ls_vacinhC += count;
		lsuP->ls_vacinhC += count;
	    }
	}
	else {
	    fprintf( stderr, "%s line %d: unknown action \"%s\"\n",
		    lfnameP, lineN, arg1P );
	    return ( TRUE );
	}

	/* Bump globals */
	lsgP->ls_actionC += count;
	lsuP->ls_actionC += count;
    }
    else if ( stricmp( kwdP, "Stats" ) == 0 ) {
	;
    }

    return ( TRUE );
}
/*

*//* lfreport()

	Do the report

Accepts :



Returns :

	<nothing>


Notes :


*/

void
lfreport NOARGLIST
{
    /* Clear the reported counter */
    RepC = 0;

    /* Walk through items */
    avldb_walk( report_logsum );
}
/*

*//* lfsum( lfP, lfnameP )

	Do logfile summary

Accepts :

	lfP		FILE ptr for open logfile
	lfnameP		A descriptive name for the file

Returns :

	<nothing>


Notes :


*/

void
lfsum ARGLIST( (lfP, lfnameP) )
   NFARG( FILE		*lfP )		/* Open file ptr */
    FARG( char		*lfnameP )	/* Descriptive name */
{
    /* Collect the info */
    lfcollect( lfP, lfnameP );

    /* Report */
    lfreport();
}
/*

*//* logsum_get( nameP )

	Find/create a named logsum block

Accepts :

	nameP		Ptr to name for block

Returns :

	<value>		Ptr to block
			NULL if problem

Notes :

	Finds a named block if it's already registered,
	inserts the named block if not.

*/

LOGSUM *
logsum_get ARGLIST( (nameP) )
    FARG( char		*nameP )	/* Name associated with block */
{
	LOGSUM		*lsP;		/* Ptr to block */
	LOGSUM		**lsPP;		/* Ptr to block ptr */

    /* Try to find existing name/block */
    lsPP = avldb_find( DT_LOGSUM, nameP );
    if ( lsPP == NULL ) {
	/* Doesn't exist- make new one. */
	lsP = (LOGSUM *)emalloc( "logsum_get", "logsum block",
					sizeof(LOGSUM) );
	if ( lsP != NULL ) {
	    lsP->ls_actionC = 0;
	    lsP->ls_discardC = 0;
	    lsP->ls_fileintoC = 0;
	    lsP->ls_filespamC = 0;
	    lsP->ls_ikeepC = 0;
	    lsP->ls_keepC = 0;
	    lsP->ls_redirectC = 0;
	    lsP->ls_rejectC = 0;
	    lsP->ls_vacC = 0;
	    lsP->ls_vacinhC =0;

	    /* Insert name in database */
	    lsPP = avldb_insert( DT_LOGSUM, nameP );
	    if ( lsPP == NULL ) {
		fprintf( stderr, "logsum_get: avldb_insert failed for \"%s\"\n",
					nameP );
		lsP = NULL;
	    }
	    else {
		*lsPP = lsP;
	    }
	}
    }
    else {
	/* Found it */
	lsP = *lsPP;
    }

    return ( lsP );
}
/*

*//* report_logsum( nodetype, nameP, lsPP )

	Report one LOGSUM entry

Accepts :

	nodetype	Type of node in avldb
	nameP		Name associated with node
	lsPP		Ptr to LOGSUM ptr 

Returns :

	<value>		TRUE if OK
			FALSE if not.

Notes :

	Follows interface to avldb_walk

*/

int
report_logsum ARGLIST( (nodetype, nameP, lsPP) )
   NFARG( int		nodetype )	/* Node type */
   NFARG( char		*nameP )	/* Name associated with node */
    FARG( LOGSUM	**lsPP )	/* Ptr to LOGSUM pointer */
{
	LOGSUM		*lsP;		/* Ptr to LOGSUM block */

    if ( nodetype == DT_LOGSUM ) {
	lsP = *lsPP;

	/* Give title if needed */
	if ( ++RepC == 1 ) {
	    printf( "%-18s  %5s  %5s  %5s  %5s  %5s  %5s  %5s  %5s  %5s  %5s\n\n",
		       " - name -",
		       "total",
		       "ikeep",
		       "keep",
		       "file",
		       "fspam",
		       "disc",
		       "rej",
		       "redir",
		       "vac",
		       "vacih"
		       );
	}

        printf( "%-18s  %5d  %5d  %5d  %5d  %5d  %5d  %5d  %5d  %5d  %5d\n",
		      nameP,
		      lsP->ls_actionC,
		      lsP->ls_ikeepC,
		      lsP->ls_keepC,
		      lsP->ls_fileintoC,
		      lsP->ls_filespamC,
		      lsP->ls_discardC,
		      lsP->ls_rejectC,
		      lsP->ls_redirectC,
		      lsP->ls_vacC,
		      lsP->ls_vacinhC
	       );
    }

    return ( TRUE );
}
